//ContainmentRegion.cpp - Body of associated class.

#include "ContainmentRegion.h"
#include <math.h>   
#include <iostream.h>
#include <stdio.h>
/* =======================================================================================
	The ContainmentRegion is used to store information about a target containment region which
	exists at a particular momement in time.  The region may then be displayed or expanded.


	This software was developed as part of a grant by ONR 311, Office of Naval Research, 
	800 North Quincy Street, Arlington, Virginia 22217, for development of an "Application 
	of Volume Modeling Techniques to Dynamic Containment Regions for Naval Applications".
	Grant was awarded to Arizona State University, PRISM Laboratory, Department of Computer 
	Science and Engineering, Dr. Greg Nielson and Gary Graf.


	Author: Gary Graf, PRISM Laboratory, Arizona State University.
	Version:  1.0, 10 Sep 2002

  Copyright 2002, all rights reserved.
  
	Notes:
=========================================================================================*/


		//A set of constants to enumerate the options for how the region should be expanded.
//TODO:  Add more than spherical options.
	int const static SPHERICAL = 0;


		//A set of constants to enumerate the options for building simple containments.  In each
		//  case the volume is centered in the middle of the containment region.
		//  FILL_NONE - The containment region is created but not filled.
		//  FILL_VOLUME - A sphere is added whose diameter is equal to the maximum dimension of the
		//			containment region's latitude or longitude coordinates. The sphere is then used 
		//			to fill the volume. The sphere is clipped	where	it extends beyond the edges of the 
		//			volume.
		//  FILL_AREA - A sphere is added whose diameter is equal to the minimum dimension of the
		//			containment region's latitude or longitude coordinates.  The sphere is then used 
		//			to fill the volume.  The sphere is clipped where it extends beyond the edges of 
		//			the depth limits of the volume.
		//  LIMIT_VOLUME - A sphere whose diameter is equal to the minimum dimension of the
		//			containment region, including the depth component..  The sphere is then used 
		//			to fill the volume.  No clipping occurs.
		//	Note:  Future flags may include mods for filling with ellipsoid shapes.  Irregular
		//		shapes can be made by using FILL_NONE, then writing the data directly to the 
		//		volumeData vector.
	int const static FILL_NONE    = 0;
	int const static FILL_VOLUME  = 1;
	int const static FILL_AREA    = 2;
	int const static LIMIT_VOLUME = 3;

		//A constant to multiply times degrees to get radians.
	float const static DEG_TO_RAD = 0.017453292f;
		//A constant to convert yards to nautical miles.  Note that one nautical mile (nm) is
		//	 also equal to one degree of latitude.
	float const static YARDS_PER_NM = 2000.0f;
		//A constant to convert minutes of latitude/longitude to degrees.
	float const static MIN_PER_DEGREE = 60.0f;

		//The default max depth in yards
	float const static MAX_DEPTH_YDS = 1000.0f;

/**************************************************************************************
---------------------------------------------------------------------------------------
		Constructors / Destructors
---------------------------------------------------------------------------------------
**************************************************************************************/

	/*
		The default constructor establishes an enumerated volume of approximately, 200 kyds x
		200 kyds, 500 yd spacing, origin at 25 deg N, 25 deg E, with no ID number.
	*/
CContainmentRegion::CContainmentRegion() : CEnumeratedVolume() {

	assocTargetID = 0;
	fillType = FILL_NONE;
	regionSources = NULL;
	isModified = false;


	regionCenter[0] = volumeDimension[0] / 2;
	regionCenter[1] = volumeDimension[1] / 2;

	regionCenter[2] = volumeDimension[2] / 2;
}	//end constructor CContainmentRegion()


	/*
		This constructor will take a min latitude and longitude along with the required grid
		dimensions to compute the final extent of the region.  It is primarily used in the
		construction of resultant containment regions following the expansion of another region.  

		@param newID - An integer which is a unique identifier for the containment region.
		@param source - A pointer to a string of charecters.  This is inteded to be used as
				a verbal identifier of the region.  It can be concatinated to identify intersection
				regions as the sum of it parts.  Its exact use is undefined.
		@param refTime - A long integer which represents the time when the given dimensions of the
				volume data for the region are valid.
		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param latDim - An integer which specifies how many grid cell the region should occupy
				in latitude.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param lonDim - An integer which specifies how many grid cell the region should occupy
				in longitude.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
		@param colors  - A pointer to a three element float array specifying the color to be used 
				in rendering this particular containment region.  The nature of the color space 
				used and the spread of values are not specified.  The information contained in 
				in the colors[] array are only stored to be later passed to the rendering engine.
		@param trans  - A float specifying the transparancy (or opacity) of the containment region
				when rendered.  The nature of the transparancy and the spread of values are not 
				specified.  The information contained is only stored to be later passed to the 
				rendering engine.
		@param tgtID - An integer which is an identifier for the target related to this containment
				region.  The ID relates internally to a CTargetMotionData object and is used to
				retrieve speed information needed to calculate an expansion volume.
	*/
CContainmentRegion::CContainmentRegion(int newID, char* source, long int refTime, 
													float mnLat, int latDim,
												   float mnLon, int lonDim, 
													float rngErr, float *colors,
													float trans, int tgtID) : 
						  CEnumeratedVolume(newID, refTime, mnLat, latDim,	mnLon, lonDim, rngErr, 
												 colors, trans){
	assocTargetID = tgtID;
	fillType = FILL_NONE;
	regionSources = source;
	isModified = false;

	regionCenter[0] = (volumeDimension[0] - 1) / 2.0;
	regionCenter[1] = (volumeDimension[1] - 1) / 2.0;
	regionCenter[2] = (volumeDimension[2] - 1) / 2.0;
}	//end constructor CContainmentRegion(int, char*, long int, float, int, float, int, 
	//													float, float*, float, int)


	/*
		This is the fully specified constructor.

		@param newID - An integer which is a unique identifier for the containment region.
		@param source - A pointer to a string of charecters.  This is inteded to be used as
				a verbal identifier of the region.  It can be concatinated to identify intersection
				regions as the sum of it parts.  Its exact use is undefined.
		@param refTime - A long integer which represents the time when the given dimensions of the
				volume data for the region are valid.
		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param mxLat  - A float which is the maximum latitude boundary of the region.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param mxLon  - A float which is the maximum longitude boundary of the region.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
		@param colors  - A pointer to a three element float array specifying the color to be used 
				in rendering this particular containment region.  The nature of the color space 
				used and the spread of values are not specified.  The information contained in 
				in the colors[] array are only stored to be later passed to the rendering engine.
		@param trans  - A float specifying the transparancy (or opacity) of the containment region
				when rendered.  The nature of the transparancy and the spread of values are not 
				specified.  The information contained is only stored to be later passed to the 
				rendering engine.
		@param tgtID - An integer which is an identifier for the target related to this containment
				region.  The ID relates internally to a CTargetMotionData object and is used to
				retrieve speed information needed to calculate an expansion volume.
		@param fill - An integer which is an enumerated constant (see listings of constants in
				the remarks at the top of the class) describing what to do with the interior volume
				after instantiation.  A value of FILL_NONE ( 0 ) is most commonly used.
	*/
CContainmentRegion::CContainmentRegion(int newID, char* source, long int refTime, 
													float mnLat, float mxLat,
													float mnLon, float mxLon, 
													float rngErr, float *colors,
													float trans, int tgtID, int fill) : 
			CEnumeratedVolume(newID, refTime, mnLat, mxLat,	mnLon, mxLon, rngErr, colors,
												 trans) {

	regionSources = source;
	isModified = false;

		//The - 1 below accounts for removal of the zero offset in the array
	regionCenter[0] = (float)(volumeDimension[0] - 1) / 2.0;
	regionCenter[1] = (float)(volumeDimension[1] - 1) / 2.0;
	regionCenter[2] = (float)(volumeDimension[2] - 1) / 2.0;


	assocTargetID = tgtID;
	if((fill >= FILL_NONE) && (fill <= LIMIT_VOLUME)){
		fillType = fill;
		fillRegion(fill);
	}	//end if
	else
		fillType = FILL_NONE;
}	//end constructor CContainmentRegion(int, time, float, float, float, float, float, int, int)


	//Destructor
CContainmentRegion::~CContainmentRegion(){
	if(!surfacePointList.empty())
		surfacePointList.clear();
	if(!surfaceArcList.empty())
		surfaceArcList.clear();
	if(!surfaceQuadList.empty())
		surfaceQuadList.clear();
	if (regionSources)
	  delete[] regionSources;
}	//end destructor ~CContainmentRegion()


/**************************************************************************************
---------------------------------------------------------------------------------------
		Public methods - General
---------------------------------------------------------------------------------------
**************************************************************************************/

	/*
		This method will make a containment volume and inside of it construct a modified tubular 
			section representing a pseudo hyperbolic conic section.  THIS METHOD IS USED FOR
			DEVELOPMENT PURPOSES ONLY.
	*/
void CContainmentRegion::makeTubularVolume(){
	float thirdOfDepth = (float)(volumeDimension[2] - 1) / 3.0;
	float halfOfDepth  = (float)(volumeDimension[2] - 1) / 2.0;
	float maxLon = (float)(volumeDimension[0] - 1);
	float maxLat = (float)(volumeDimension[1] - 1);
	float radiusOffset = (maxLat - regionCenter[1] - halfOfDepth) / maxLon;
	float centerStep = thirdOfDepth / maxLon;

	float centerlineDepth;
	float minRadiusSqr, maxRadiusSqr;
	double yMin, yMax, z, tubeValue; 
	double yMinSqr, yMaxSqr, ySqr, zSqr; 

		//Clear out any old data and construct new geometry
	volumeData.clear();
	long int midOffset = volumeDimension[1] * volumeDimension[0];

	for(int k=0; k < volumeDimension[2]; ++k){
		for(int j=0; j < volumeDimension[1]; ++j){
			ySqr = (j - regionCenter[1]) * (j - regionCenter[1]);
			for(int i=0; i < volumeDimension[0]; ++i){

				centerlineDepth = thirdOfDepth - i * centerStep;
					//Eliminate points above tube
				if(k >= (centerlineDepth - 0.5)){
					z = k - centerlineDepth;
					zSqr = z * z;
					tubeValue = ySqr + zSqr;

					yMax = halfOfDepth  + i * radiusOffset;
					yMaxSqr = yMax * yMax;
					yMin = yMax - thirdOfDepth;
					yMinSqr = yMin * yMin;

					maxRadiusSqr = yMaxSqr + 0.5;
					minRadiusSqr = yMinSqr + 0.5;

					if((tubeValue - minRadiusSqr >= 0) && (tubeValue - maxRadiusSqr <= 0))
						volumeData.push_back(1);	
					else
						volumeData.push_back(0);
				}	//end if
				else
					volumeData.push_back(0);
			}	//end for i
		}	//end for j
	}	//end for k
		//Standard expansion volumes determine the surface points when the fillRegion()
		//  method is called.  For this specialized volume, you must explicitly call it.
	findSurfacePoints();
}	//end makeTubularVolume()


	/*
	  This method will search the volume for points which lie on the surface, including
	    all points which lie on the boundary of the volume.  A surface point is defined as
		 a location where the data is greater than zero and at least one point adjacent, not
		 counting diagonals, is equal to zero.  When it finds a surface point
	    it stores the corresponding vector and array indices in the appropriate surface list.  
		 The method automatically flags boundary points as surface points and then tests the
		 interior points.
	*/	
void CContainmentRegion::findSurfacePoints(){
		//Clear any potentially old data
	if(!surfacePointList.empty())
		surfacePointList.clear();
	if(!surfaceArcList.empty())
		surfaceArcList.clear();
	if(!surfaceQuadList.empty())
		surfaceQuadList.clear();

	int surfaceHashValue = 0;

	long int lonOffset, latOffset, depthOffset;
	long int midOffset = volumeDimension[1] * volumeDimension[0];

		//Walk through every point in volume looking for solid objects

	for(int k=0; k < volumeDimension[2]; ++k){
		depthOffset = k * midOffset;
		for(int j=0; j < volumeDimension[1]; ++j){
			latOffset = j * volumeDimension[0] + depthOffset;
			for(int i=0; i < volumeDimension[0]; ++i){
				lonOffset = i + latOffset;
					//If a point in the volume, test to see if it is on the surface
				if(volumeData[lonOffset] > 0){
					surfaceHashValue = findPointFacing(i, j, k, midOffset);
						//Add surface points to the list, ignore the rest.
					if(surfaceHashValue > 0) {
						addSurfacePointToList(lonOffset, i, j, k, surfaceHashValue);

					}				}	//end if
			}	//end for i
		}	//end for j
	}	//end for k
}	//end findSurfacePoints()



	/*
		Newly expanded regions have a shell created by plotting the appropriate points of the 
			template (faces, arcs, and quadrants) but the interior is characterized by a hollow 
			space matching the original volume as well as broken and trapped hollows between the 
			original volume and the new outer shell. To define the correct surface for rendering 
			the interior regions must all be filled to a value of one or the surface determination 
			algorithm will incorrectly mark interior hollows as exterior surfaces. A simple flood 
			fill of the interior will not work because the space may be cut into isolated regions 
			by the plotting of intersecting arcs and quads. 
		This method solves the problem of creating a solid filled shape from irregular surface and
			point data by reversing the traditional flood fill concept.  In this method the outside
			is tested and if it is empty it is marked with a value of two.  After all the outside
			spaces are marked, the volume is scanned and the outside spaces are set to their proper
			value of zero while all other values are presumed to be interior points and set to one.
		NOTE:  If the shape itself originally had an interior void, this algorithm will eliminate
			the void entirely.
	*/
void CContainmentRegion::floodFillRegion(){
	long int midOffset = volumeDimension[1] * volumeDimension[0];
	int i = 0;
	int j = 0;

		//To ensure you find all the exterior points, start the flood by testing everyting at the
		//  outer edge of the bounding box.

		//Test top and bottom planes
	int k = 0;
	for(j = 0; j < volumeDimension[1]; ++j){
		for(i = 0; i < volumeDimension[0]; ++i)
			markExteriorPoint(i, j, k, midOffset);
	}	//end for

	k = volumeDimension[2] - 1;
	for(j = 0; j < volumeDimension[1]; ++j){
		for(i = 0; i < volumeDimension[0]; ++i){
			markExteriorPoint(i, j, k, midOffset);
		}	//end for
	}	//end for

		//Test east and west planes
	i = 0;
	for(k = 1; k < volumeDimension[2] - 1; ++k){	   //Note: k = 0 and k = vd[] -1 already tested
		for(j = 0; j < volumeDimension[1]; ++j){
			markExteriorPoint(i, j, k, midOffset);
		}	//end for
	}	//end for

	i = volumeDimension[0] - 1;
	for(k = 1; k < volumeDimension[2] - 1; ++k){	   //Note: k = 0 and k = vd[] -1 already tested
		for(j = 0; j < volumeDimension[1]; ++j){
			markExteriorPoint(i, j, k, midOffset);
		}	//end for
	}	//end for

		//Test north and south planes
	j = 0;
	for(k = 1; k < volumeDimension[2] - 1; ++k){	   //Note: k = 0 and k = vd[] -1 already tested
		for(i = 1; i < volumeDimension[0] - 1; ++i){	//Note: i = 0 and i = vd[] -1 already tested
			markExteriorPoint(i, j, k, midOffset);
		}	//end for
	}	//end for

	j = volumeDimension[1] - 1;
	for(k = 1; k < volumeDimension[2] - 1; ++k){	   //Note: k = 0 and k = vd[] -1 already tested
		for(i = 1; i < volumeDimension[0] - 1; ++i){	//Note: i = 0 and i = vd[] -1 already tested
			markExteriorPoint(i, j, k, midOffset);
		}	//end for
	}	//end for


		//Now that all exterior points have been marked, invert the values to convert the exterior
		//  points to zeros and the remaining (interior) points to ones.
	long int offset = 0;
	for(k = 0; k < volumeDimension[2]; ++k){
		for(j = 0; j < volumeDimension[1]; ++j){
			for(i = 0; i < volumeDimension[0]; ++i){
				offset = getOffset(i, j, k, midOffset);
				if(volumeData[offset] == 2)
					volumeData[offset] = 0;
				else
					volumeData[offset] = 1;
			}	//end for
		}	//end for
	}	//end for
}	//end floodFillRegion()




	/*  
		This method will expand a containment region as a function of time.  The amount of
			expansion is first determined by getting the difference between the regions current 
			reference time and the the passed time.  This time difference is the multiplied
			by the target speed, as referenced in the regions associated TargetMotionData object,
			to derive a distance figure.
		The expansion is performed by first creating a cubical enumerated volume (possible clipped
			in depth) to act as a template.  Within this volume, data values are assigned resulting
			in a centered, spherical region of ones, zero outside the sphere.  The surface points
			of the template are catagorized as lying on a particular quadrant or along a cardinal
			meridian.
		The template is then 'walked' over the surface of the region being expanded.  Depending on
			the local facing of the current surface point, the appropriate points, arcs or quads
			from the template are mapped into a new blank region using appropriate offsets from both
			the original region as well as the template.  To enhance speed performance, only the 
			mininum number of surface points needed to develop the expanded shell are used, both on
			the region and template.
		When the template has been applied to the entire surface, the result region will consist of
			an outer shell representing the maximum extent of the expansion. Inside the shell though,
			the data will be a fractured combination of ones and zeros resulting from overlapping 
			template data.  The method floodFillRegion() is them called to fill the gaps and restore
			the data coherence of the expaned containment region.
			
		@param targetInfo  - A CTargetMotionData object which holds information as to how the region
									is to be expanded as a function of elapsed time.
		@param endTime - A long integer representing the termination or current time of expansion.
		@returns - A CContainmentRegion object which is the original region after expansion
							for the elapsed time. If the expansion distance (time differential times the
							target speed) is less than the allowed range error in the display, the
							object returned is the original region. Otherwise, a new object is returned 
							which is an expanded region.
	*/
CContainmentRegion* CContainmentRegion::expandRegion(CTargetMotionData *targetInfo, long int endTime){

		//The value of resultRgn is returned, so set it to input now in case no expansion is done.
	CContainmentRegion *resultRgn = this;

		//If target info is really available, expand it.  Otherwise return the original region.
	if(targetInfo != NULL){
			//Get parameters needed to index into the three containment regions
			//Original region

			//Time is passed as seconds, for speed measure convert to hours.
		float elapsedTimeHr = ((float)endTime - (float)timeReference) / 3600.0;
			//Get expansion distances in nautical miles
		float maxDistanceNM = targetInfo->maxSpeed * elapsedTimeHr;

			//If time difference is too small to allow expansion, exit with original
		if((maxDistanceNM * YARDS_PER_NM) <= allowedRangeError)
			return this;

		//  The factor  of * 2.0 (below) increases the dimensions to account for target motion both
			//	 towards or away from the center.
		int newGridSize = (int)(maxDistanceNM * YARDS_PER_NM * 2 / allowedRangeError + 0.5) + 1;

//TODO:  Include more sophisticated expansion techniques other than the simple spherical
//			expansion delineated below.

			//Create a containment region the size of the expansion, with nothing inside, that is large
			//  enough to account for the construction of an expansion volume base
			//  on the ellapsed time and target speed.
			//  The factor  of * 2.0 increases the dimensions to account for target motion both
			//	 towards or away from the center.
		CExpansionTemplate *expansionTemplate = new CExpansionTemplate(0,	newGridSize, 0, newGridSize, 
															allowedRangeError, targetInfo, 
															elapsedTimeHr, SPHERICAL);

			//Get info to build region to hold results and set up variables for the later looping
			//  calculations  Note: The latitude or longitude of the
			//  expansion region might not match the earlier constructor since expansion regions have 
			//  their size adjusted if necessary to ensure that an odd number of grid points exist.
		int *expDim = expansionTemplate->getGridDimensions();
		long int expSizeOfGridArea = expDim[1] * expDim[0];

			//Calculate the correct change in longitude. The design of the expansion
			//  volume is centered at zero latitude to assist in making uniform shapes and so
			//  the resultant longitude values must be adjusted as they relate to the initial
			//  containment region.  The divide by two is because the expansion volume assumes target
			//  motion in either direction and this longitude correction will only be used in one.
		float modSpanLon = expansionTemplate->getSpanLat() / cos(centerLat * DEG_TO_RAD) / 2.0;

			//Create a containment to hold the final answer
		char *newsrc=new char[strlen(regionSources)+1];
		strcpy(newsrc,regionSources);
		resultRgn = new CContainmentRegion(volumeID, newsrc, endTime,  
			maxExtent.minLat - expansionTemplate->getSpanLat() / 2.0, 
			expDim[1] + volumeDimension[1] - 1,
			maxExtent.minLon - modSpanLon, 
			expDim[0] + volumeDimension[0] - 1,
			allowedRangeError, regnColor,
			transparancy, assocTargetID);

			//Find the girdpoint at the center of the expansion region.  Although regionCenter[]
			//  is a float, for expansion regions the value is always odd so using an int for
			//  expLonCenter results in no rounding error.
		int expLonCenter = (int)expansionTemplate->regionCenter[0];
		int expLatCenter = (int)expansionTemplate->regionCenter[1];

			//Set up result region related variables for the later looping calculations.
		long int resultDepthOffset, resultLatOffset, resultLonOffset;
		int resultDepth;
		int* resultDim = resultRgn->getGridDimensions();
		long int resultSizeOfGridArea = resultDim[1] * resultDim[0];
		//cout<<"Result region size: "<<resultDim[0]<<" x "<<resultDim[1]<<" x "<<resultDim[2]<<endl;
		//cout<<endl;


			//Varibles to track the corresponding grid index position in the result region.
			//  The result depth will be the source region depth plus/minus the expansion
			//  region depth.  The result lat (lon) will be the source lat (lon) plus the
			//  offset between the source and result (all of which == resultLat(Lon)Index)
			//  plus/minus the expansion region lat (lon) offset.
		long int resultLatIndex = 0;
		long int resultLonIndex = 0;

			//Find grid offset difference between source region and result region
		int lonRegionGridToResultGrid = resultRgn->getGridFromLon(maxExtent.minLon);
		int latRegionGridToResultGrid = resultRgn->getGridFromLat(maxExtent.minLat);

		int depthMidpoint = expDim[2]/2;

			//Get variables to track across the surface of the original region.
		CContainmentRegion::LIST_INDEX_VECTOR::iterator regionSurfIterator;
		CContainmentRegion::ListIndices nextRegionSurfPt;


			//Get variables to track across the surface of the expansion region
		CExpansionTemplate::LIST_INDEX_VECTOR::iterator expSurfIterator;
		CExpansionTemplate::ListIndices nextExpSurfPt;


			//Start by walking through the planar surface points on the original containment region
		/*		for (regionSurfIterator = surfacePointList.begin(); regionSurfIterator != surfacePointList.end(); ++regionSurfIterator){
			nextRegionSurfPt = (*regionSurfIterator);
				//Find corresponding offset to the resultant containment region.
			resultLatIndex = nextRegionSurfPt.latIndex + latRegionGridToResultGrid;
			resultLonIndex = nextRegionSurfPt.lonIndex + lonRegionGridToResultGrid;









			resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
			//Do west point
			resultLonIndex    = resultLonIndex - expansionTemplate->expDistWest;
			addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
					     resultDim[0], resultRgn);
			//Add east point
			resultLonIndex    = resultLonIndex + expansionTemplate->expDistEast;
			addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
					     resultDim[0], resultRgn);




			resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
			//Add south point
			resultLatIndex    = resultLatIndex - expansionTemplate->expDistSouth;
			addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
					     resultDim[0], resultRgn);
			//Add north point
			resultLatIndex    = resultLatIndex + expansionTemplate->expDistNorth;
			addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
					     resultDim[0], resultRgn);

			
			//Add down point
			resultDepth = nextRegionSurfPt.depthIndex + expansionTemplate->expDistDown;
			//Ignore depth calculations which will lie outside resultant containment
			if(resultDepth >= resultDim[2])
			  resultDepth = resultDim[2] - 1;
			resultDepthOffset = resultDepth * resultSizeOfGridArea;
			addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
					     resultDim[0], resultRgn);
			//Add up point
			resultDepth = nextRegionSurfPt.depthIndex - expansionTemplate->expDistUp;
			//Ignore depth calculations which will lie outside resultant containment
			if(resultDepth < 0)
			  resultDepth = 0;
			resultDepthOffset = resultDepth * resultSizeOfGridArea;
			addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
					     resultDim[0], resultRgn);
			
			
			
			


		*/




			/*				//Compute the resultant point for each surface point dependent on facing.
			switch(nextRegionSurfPt.facing){
				case 1:		//Opens to the east
					resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
					resultLonIndex    = resultLonIndex + expansionTemplate->expDistEast;
					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
					break;

				case 2:		//Opens to the west
					resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
					resultLonIndex    = resultLonIndex - expansionTemplate->expDistWest;
					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
					break;


				case 3:		//Opens to the east and west
					resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						//Do west point
					resultLonIndex    = resultLonIndex - expansionTemplate->expDistWest;
					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
						//Add east point
					resultLonIndex    = resultLonIndex + expansionTemplate->expDistEast;
					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
					break;


				case 4:		//Opens to the north
					resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
					resultLatIndex    = resultLatIndex + expansionTemplate->expDistNorth;
					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
					break;

				case 8:		//Opens to the south
					resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
					resultLatIndex    = resultLatIndex - expansionTemplate->expDistSouth;
					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
					break;

				case 12:		//Opens to the north and south
					resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						//Add south point
					resultLatIndex    = resultLatIndex - expansionTemplate->expDistSouth;
					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
						//Add north point
					resultLatIndex    = resultLatIndex + expansionTemplate->expDistNorth;
					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
					break;

				case 16:		//Opens to the up
					resultDepth = nextRegionSurfPt.depthIndex - expansionTemplate->expDistUp;
						//Ignore depth calculations which will lie outside resultant containment
					if(resultDepth < 0)
						resultDepth = 0;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
					break;

				case 32:		//Opens to the down
					resultDepth = nextRegionSurfPt.depthIndex + expansionTemplate->expDistDown;
						//Ignore depth calculations which will lie outside resultant containment
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] - 1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;

					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
					break;

				case 48:		//Opens to the up and down
						//Add down point
					resultDepth = nextRegionSurfPt.depthIndex + expansionTemplate->expDistDown;
						//Ignore depth calculations which will lie outside resultant containment
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] - 1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
						//Add up point
					resultDepth = nextRegionSurfPt.depthIndex - expansionTemplate->expDistUp;
						//Ignore depth calculations which will lie outside resultant containment
					if(resultDepth < 0)
						resultDepth = 0;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					addDataPointToRegion(resultLonIndex, resultLatIndex, resultDepthOffset, 
													 resultDim[0], resultRgn);
					break;


				default:
					cout<<"Error in surfacePointList, not a point index: "<<nextRegionSurfPt.facing<<endl;
					break;
					}	//end switch*/
		//}	//end for surface iterator (points)


			//Now walk through every surface point of the original containment region on an edge
		/*	for (regionSurfIterator = surfaceArcList.begin(); regionSurfIterator != surfaceArcList.end(); ++regionSurfIterator){
			nextRegionSurfPt = (*regionSurfIterator);
				//Find corresponding offset to the resultant containment region.
			resultLatIndex = nextRegionSurfPt.latIndex + latRegionGridToResultGrid;
			resultLonIndex = nextRegionSurfPt.lonIndex + lonRegionGridToResultGrid;







			//BEGIN CRUDE HACK






			//Now walk through every arc point of the NE expansion volume
			for (expSurfIterator = expansionTemplate->arcsNE.begin(); expSurfIterator != expansionTemplate->arcsNE.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  //For arcs in the NE plane, all depth indices will align
			  resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
			  resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
			  resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
			  resultRgn->volumeData[resultLonOffset] = 1;
			}	//end for expansion iterator
			//Now walk through every arc point of the NW expansion volume
			for (expSurfIterator = expansionTemplate->arcsNW.begin(); expSurfIterator != expansionTemplate->arcsNW.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  //For arcs in the NW plane, all depth indices will align
			  resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
			  resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
			  resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
			  resultRgn->volumeData[resultLonOffset] = 1;
			}	//end for expansion iterator
			//Now walk through every arc point of the SW expansion volume
			for (expSurfIterator = expansionTemplate->arcsSW.begin(); expSurfIterator != expansionTemplate->arcsSW.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  //For arcs in the SW plane, all depth indices will align
			  resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
			  resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
			  resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
			  resultRgn->volumeData[resultLonOffset] = 1;
			}	//end for expansion iterator
						//Now walk through every arc point of the SE expansion volume
			for (expSurfIterator = expansionTemplate->arcsSE.begin(); expSurfIterator != expansionTemplate->arcsSE.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  //For arcs in the SE plane, all depth indices will align
			  resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
			  resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
			  resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
			  resultRgn->volumeData[resultLonOffset] = 1;
			}	//end for expansion iterator
			

			//Now walk through every arc point of the DW expansion volume
			for (expSurfIterator = expansionTemplate->arcsDW.begin(); expSurfIterator != expansionTemplate->arcsDW.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  //Find out out depth indices align
			  resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
			  //Flatten the arc along the surface and floor.
			  //Latitude index unchanged in this plane
			  if(resultDepth < 0)
			    resultDepth = 0;
			  if(resultDepth >= resultDim[2])
			    resultDepth = resultDim[2] -1;
			  resultDepthOffset = resultDepth * resultSizeOfGridArea;
			  resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
			  resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
			  resultRgn->volumeData[resultLonOffset] = 1;
			}	//end for expansion iterator
			//Now walk through every arc point of the UW expansion volume
			for (expSurfIterator = expansionTemplate->arcsUW.begin(); expSurfIterator != expansionTemplate->arcsUW.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  //Find out out depth indices align
			  resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
			  //Flatten the arc along the surface and floor.
			  //Latitude index unchanged in this plane
			  if(resultDepth < 0)
			    resultDepth = 0;
			  if(resultDepth >= resultDim[2])
			    resultDepth = resultDim[2] -1;
			  resultDepthOffset = resultDepth * resultSizeOfGridArea;
			  resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
			  resultLonIndex = nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex;
			  resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
			  resultRgn->volumeData[resultLonOffset] = 1;
			}	//end for expansion iterator
			//Now walk through every arc point of the UE expansion volume
			for (expSurfIterator = expansionTemplate->arcsUE.begin(); expSurfIterator != expansionTemplate->arcsUE.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  //Find out out depth indices align
			  resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
			  //Flatten the arc along the surface and floor.
			  //Latitude index unchanged in this plane
			  if(resultDepth < 0)
			    resultDepth = 0;
			  if(resultDepth >= resultDim[2])
			    resultDepth = resultDim[2] -1;
			  resultDepthOffset = resultDepth * resultSizeOfGridArea;
			  resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
			  resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
			  resultRgn->volumeData[resultLonOffset] = 1;
			}	//end for expansion iterator
			//Now walk through every arc point of the DE expansion volume
			for (expSurfIterator = expansionTemplate->arcsDE.begin(); expSurfIterator != expansionTemplate->arcsDE.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  //Find out out depth indices align
			  resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
			  //Flatten the arc along the surface and floor.
			  //Latitude index unchanged in this plane
			  if(resultDepth < 0)
			    resultDepth = 0;
			  if(resultDepth >= resultDim[2])
			    resultDepth = resultDim[2] -1;
			  resultDepthOffset = resultDepth * resultSizeOfGridArea;
			  resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
			  resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
			  resultRgn->volumeData[resultLonOffset] = 1;
			}	//end for expansion iterator
			

			//Now walk through every arc point of the DS expansion volume
			for (expSurfIterator = expansionTemplate->arcsDS.begin(); expSurfIterator != expansionTemplate->arcsDS.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  
			  //Find out out depth indices align
			  resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
			  //Flatten the arc along the surface and floor.
			  //Longitude index unchanged in this plane
			  if(resultDepth < 0)
			    resultDepth = 0;
			  if(resultDepth >= resultDim[2])
			    resultDepth = resultDim[2] -1;
			  resultDepthOffset = resultDepth * resultSizeOfGridArea;
			  resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
			  resultLonOffset = resultLonIndex + resultLatOffset;
			}	//end for expansion iterator
			resultRgn->volumeData[resultLonOffset] = 1;
			//Now walk through every arc point of the US expansion volume
			for (expSurfIterator = expansionTemplate->arcsUS.begin(); expSurfIterator != expansionTemplate->arcsUS.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  
			  //Find out out depth indices align
			  resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
			  //Flatten the arc along the surface and floor.
			  //Longitude index unchanged in this plane
			  if(resultDepth < 0)
			    resultDepth = 0;
			  if(resultDepth >= resultDim[2])
			    resultDepth = resultDim[2] -1;
			  resultDepthOffset = resultDepth * resultSizeOfGridArea;
			  resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
			  resultLonOffset = resultLonIndex + resultLatOffset;
			  resultRgn->volumeData[resultLonOffset] = 1;
			}	//end for expansion iterator
			//Now walk through every arc point of the UN expansion volume
			for (expSurfIterator = expansionTemplate->arcsUN.begin(); expSurfIterator != expansionTemplate->arcsUN.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  
			  //Find out out depth indices align
			  resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
			  //Flatten the arc along the surface and floor.
			  //Longitude index unchanged in this plane
			  if(resultDepth < 0)
			    resultDepth = 0;
			  if(resultDepth >= resultDim[2])
			    resultDepth = resultDim[2] -1;
			  resultDepthOffset = resultDepth * resultSizeOfGridArea;
			  resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
			  resultLonOffset = resultLonIndex + resultLatOffset;
			  resultRgn->volumeData[resultLonOffset] = 1;
			}	//end for expansion iterator
			//Now walk through every arc point of the DN expansion volume
			for (expSurfIterator = expansionTemplate->arcsDN.begin(); expSurfIterator != expansionTemplate->arcsDN.end(); ++expSurfIterator){
			  nextExpSurfPt = (*expSurfIterator);
			  
			  //Find out out depth indices align
			  resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
			  //Flatten the arc along the surface and floor.
			  //Longitude index unchanged in this plane
			  if(resultDepth < 0)
			    resultDepth = 0;
			  if(resultDepth >= resultDim[2])
			    resultDepth = resultDim[2] -1;
			  resultDepthOffset = resultDepth * resultSizeOfGridArea;
			  resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
			  resultLonOffset = resultLonIndex + resultLatOffset;
			  resultRgn->volumeData[resultLonOffset] = 1;
			}	//end for expansion iterator
			






*/











				//Compute the resultant arcs for each surface point dependent on facing.
			/*switch(nextRegionSurfPt.facing){
				case 1:		//Edge to the NE
						//Now walk through every arc point of the expansion volume
					for (expSurfIterator = expansionTemplate->arcsNE.begin(); expSurfIterator != expansionTemplate->arcsNE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the NE plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 2:		//Edge to the NW
						//Now walk through every arc point of the expansion volume
					for (expSurfIterator = expansionTemplate->arcsNW.begin(); expSurfIterator != expansionTemplate->arcsNW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the NW plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 3:		//Edge to the NE and NW (T-shape in plane)
						//Now walk through every arc point of the NE expansion volume
					for (expSurfIterator = expansionTemplate->arcsNE.begin(); expSurfIterator != expansionTemplate->arcsNE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the NE plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the NW expansion volume
					for (expSurfIterator = expansionTemplate->arcsNW.begin(); expSurfIterator != expansionTemplate->arcsNW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the NW plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 4:		//Edge to the SE
						//Now walk through every arc point of the expansion volume
					for (expSurfIterator = expansionTemplate->arcsSE.begin(); expSurfIterator != expansionTemplate->arcsSE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the SE plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 5:		//Edge to the SE and NE (T-shape in plane)
						//Now walk through every arc point of the SE expansion volume
					for (expSurfIterator = expansionTemplate->arcsSE.begin(); expSurfIterator != expansionTemplate->arcsSE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the SE plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the NE expansion volume
					for (expSurfIterator = expansionTemplate->arcsNE.begin(); expSurfIterator != expansionTemplate->arcsNE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the NE plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 8:		//Edge to the SW
						//Now walk through every arc point of the expansion volume
					for (expSurfIterator = expansionTemplate->arcsSW.begin(); expSurfIterator != expansionTemplate->arcsSW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the SW plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 10:		//Edge to the SW and NW (T-shape in plane)
						//Now walk through every arc point of the SW expansion volume
					for (expSurfIterator = expansionTemplate->arcsSW.begin(); expSurfIterator != expansionTemplate->arcsSW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the SW plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the NW expansion volume
					for (expSurfIterator = expansionTemplate->arcsNW.begin(); expSurfIterator != expansionTemplate->arcsNW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the NW plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 12:		//Edge to the SW and SE (T-shape in plane)
						//Now walk through every arc point of the SW expansion volume
					for (expSurfIterator = expansionTemplate->arcsSW.begin(); expSurfIterator != expansionTemplate->arcsSW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the SW plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the SE expansion volume
					for (expSurfIterator = expansionTemplate->arcsSE.begin(); expSurfIterator != expansionTemplate->arcsSE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the SE plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 15:		//Edge to the SW, NW, NE and SE (Line)
						//Now walk through every arc point of the NE expansion volume
					for (expSurfIterator = expansionTemplate->arcsNE.begin(); expSurfIterator != expansionTemplate->arcsNE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the NE plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the NW expansion volume
					for (expSurfIterator = expansionTemplate->arcsNW.begin(); expSurfIterator != expansionTemplate->arcsNW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the NW plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the SW expansion volume
					for (expSurfIterator = expansionTemplate->arcsSW.begin(); expSurfIterator != expansionTemplate->arcsSW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the SW plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the SE expansion volume
					for (expSurfIterator = expansionTemplate->arcsSE.begin(); expSurfIterator != expansionTemplate->arcsSE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
							//For arcs in the SE plane, all depth indices will align
						resultDepthOffset = nextRegionSurfPt.depthIndex * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 16:		//Edge to the UE
						//Now walk through every arc point of the UE expansion volume
					for (expSurfIterator = expansionTemplate->arcsUE.begin(); expSurfIterator != expansionTemplate->arcsUE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 32:			//Edge to the UW
						//Now walk through every arc point of the UW expansion volume
					for (expSurfIterator = expansionTemplate->arcsUW.begin(); expSurfIterator != expansionTemplate->arcsUW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonIndex = nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 48:			//Edge to the UW and UE (T-shape in plane)
						//Now walk through every arc point of the UW expansion volume
					for (expSurfIterator = expansionTemplate->arcsUW.begin(); expSurfIterator != expansionTemplate->arcsUW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonIndex = nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the UE expansion volume
					for (expSurfIterator = expansionTemplate->arcsUE.begin(); expSurfIterator != expansionTemplate->arcsUE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 64:		//Edge to the DE
						//Now walk through every arc point of the DE expansion volume
					for (expSurfIterator = expansionTemplate->arcsDE.begin(); expSurfIterator != expansionTemplate->arcsDE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 80:		//Edge to the DE and UE (T-shape in plane)
						//Now walk through every arc point of the DE expansion volume
					for (expSurfIterator = expansionTemplate->arcsDE.begin(); expSurfIterator != expansionTemplate->arcsDE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the UE expansion volume
					for (expSurfIterator = expansionTemplate->arcsUE.begin(); expSurfIterator != expansionTemplate->arcsUE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 128:		//Edge to the DW
						//Now walk through every arc point of the DW expansion volume
					for (expSurfIterator = expansionTemplate->arcsDW.begin(); expSurfIterator != expansionTemplate->arcsDW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 160:		//Edge to the DW and UW (T-shape in plane)
						//Now walk through every arc point of the DW expansion volume
					for (expSurfIterator = expansionTemplate->arcsDW.begin(); expSurfIterator != expansionTemplate->arcsDW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the UW expansion volume
					for (expSurfIterator = expansionTemplate->arcsUW.begin(); expSurfIterator != expansionTemplate->arcsUW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonIndex = nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 192:		//Edge to the DW and DE (T-shape in plane)
						//Now walk through every arc point of the DW expansion volume
					for (expSurfIterator = expansionTemplate->arcsDW.begin(); expSurfIterator != expansionTemplate->arcsDW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the DE expansion volume
					for (expSurfIterator = expansionTemplate->arcsDE.begin(); expSurfIterator != expansionTemplate->arcsDE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 240:		//Edge to the DW, UW, UE and DE (Line)
						//Now walk through every arc point of the DW expansion volume
					for (expSurfIterator = expansionTemplate->arcsDW.begin(); expSurfIterator != expansionTemplate->arcsDW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the UW expansion volume
					for (expSurfIterator = expansionTemplate->arcsUW.begin(); expSurfIterator != expansionTemplate->arcsUW.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonIndex = nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the UE expansion volume
					for (expSurfIterator = expansionTemplate->arcsUE.begin(); expSurfIterator != expansionTemplate->arcsUE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the DE expansion volume
					for (expSurfIterator = expansionTemplate->arcsDE.begin(); expSurfIterator != expansionTemplate->arcsDE.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);
						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Latitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = resultLatIndex * resultDim[0] + resultDepthOffset;
						resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex) + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 256:		//Edge to the UN
						//Now walk through every arc point of the UN expansion volume
					for (expSurfIterator = expansionTemplate->arcsUN.begin(); expSurfIterator != expansionTemplate->arcsUN.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 512:		//Edge to the US
						//Now walk through every arc point of the US expansion volume
					for (expSurfIterator = expansionTemplate->arcsUS.begin(); expSurfIterator != expansionTemplate->arcsUS.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 768:		//Edge to the US and UN (T-shape in plane)
						//Now walk through every arc point of the US expansion volume
					for (expSurfIterator = expansionTemplate->arcsUS.begin(); expSurfIterator != expansionTemplate->arcsUS.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the UN expansion volume
					for (expSurfIterator = expansionTemplate->arcsUN.begin(); expSurfIterator != expansionTemplate->arcsUN.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 1024:		//Edge to the DN
						//Now walk through every arc point of the expansion volume
					for (expSurfIterator = expansionTemplate->arcsDN.begin(); expSurfIterator != expansionTemplate->arcsDN.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 1280:		//Edge to the DN and UN (T-shape in plane)
						//Now walk through every arc point of the DN expansion volume
					for (expSurfIterator = expansionTemplate->arcsDN.begin(); expSurfIterator != expansionTemplate->arcsDN.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the UN expansion volume
					for (expSurfIterator = expansionTemplate->arcsUN.begin(); expSurfIterator != expansionTemplate->arcsUN.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 2048:		//Edge to the DS
						//Now walk through every arc point of the expansion volume
					for (expSurfIterator = expansionTemplate->arcsDS.begin(); expSurfIterator != expansionTemplate->arcsDS.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 2560:		//Edge to the DS and US (T-shape in plane)
						//Now walk through every arc point of the DS expansion volume
					for (expSurfIterator = expansionTemplate->arcsDS.begin(); expSurfIterator != expansionTemplate->arcsDS.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the US expansion volume
					for (expSurfIterator = expansionTemplate->arcsUS.begin(); expSurfIterator != expansionTemplate->arcsUS.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 3072:		//Edge to the DS and DN (T-shape in plane)
						//Now walk through every arc point of the DS expansion volume
					for (expSurfIterator = expansionTemplate->arcsDS.begin(); expSurfIterator != expansionTemplate->arcsDS.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the DN expansion volume
					for (expSurfIterator = expansionTemplate->arcsDN.begin(); expSurfIterator != expansionTemplate->arcsDN.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				case 3840:		//Edge to the DS, US, UN and DN (Line)
						//Now walk through every arc point of the DS expansion volume
					for (expSurfIterator = expansionTemplate->arcsDS.begin(); expSurfIterator != expansionTemplate->arcsDS.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
					}	//end for expansion iterator
						resultRgn->volumeData[resultLonOffset] = 1;
						//Now walk through every arc point of the US expansion volume
					for (expSurfIterator = expansionTemplate->arcsUS.begin(); expSurfIterator != expansionTemplate->arcsUS.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the UN expansion volume
					for (expSurfIterator = expansionTemplate->arcsUN.begin(); expSurfIterator != expansionTemplate->arcsUN.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
						//Now walk through every arc point of the DN expansion volume
					for (expSurfIterator = expansionTemplate->arcsDN.begin(); expSurfIterator != expansionTemplate->arcsDN.end(); ++expSurfIterator){
						nextExpSurfPt = (*expSurfIterator);

						 //Find out out depth indices align
						resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
							//Flatten the arc along the surface and floor.
							//Longitude index unchanged in this plane
						if(resultDepth < 0)
							resultDepth = 0;
						if(resultDepth >= resultDim[2])
							resultDepth = resultDim[2] -1;
						resultDepthOffset = resultDepth * resultSizeOfGridArea;
						resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex) * resultDim[0] + resultDepthOffset;
						resultLonOffset = resultLonIndex + resultLatOffset;
						resultRgn->volumeData[resultLonOffset] = 1;
					}	//end for expansion iterator
					break;

				default:
					cout<<"Error in surfaceArcList, not an arc index: "<<nextRegionSurfPt.facing<<endl;
					break;
			}*/	//end switch
			//}	//end for surface iterator (arcs)

			//Now walk through every surface point of the original containment region on the corner of a cube.


		for (regionSurfIterator = surfacePointList.begin(); regionSurfIterator != surfacePointList.end(); ++regionSurfIterator){
			nextRegionSurfPt = (*regionSurfIterator);
				//Find corresponding offset to the resultant containment region.
			resultLatIndex = nextRegionSurfPt.latIndex + latRegionGridToResultGrid;
			resultLonIndex = nextRegionSurfPt.lonIndex + lonRegionGridToResultGrid;

				//Compute the resultant arcs for each surface point dependent on facing.
			//if((nextRegionSurfPt.facing & 1) == 1){	//Edge to the UNE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUNE.begin(); expSurfIterator != expansionTemplate->quadsUNE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;

					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;
				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 2) == 2){	//Edge to the UNW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUNW.begin(); expSurfIterator != expansionTemplate->quadsUNW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;
				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 4) == 4){	//Edge to the USE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUSE.begin(); expSurfIterator != expansionTemplate->quadsUSE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 8) == 8){	//Edge to the USW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUSW.begin(); expSurfIterator != expansionTemplate->quadsUSW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 16) == 16){	//Edge to the DNE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDNE.begin(); expSurfIterator != expansionTemplate->quadsDNE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 32) == 32){	//Edge to the DNW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDNW.begin(); expSurfIterator != expansionTemplate->quadsDNW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 64) == 64){	//Edge to the DSE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDSE.begin(); expSurfIterator != expansionTemplate->quadsDSE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 128) == 128){	//Edge to the DSW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDSW.begin(); expSurfIterator != expansionTemplate->quadsDSW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if
		}	//end for surface iterator (quads)



			//Now walk through every surface point of the original containment region on the corner of a cube.
		for (regionSurfIterator = surfaceArcList.begin(); regionSurfIterator != surfaceArcList.end(); ++regionSurfIterator){
			nextRegionSurfPt = (*regionSurfIterator);
				//Find corresponding offset to the resultant containment region.
			resultLatIndex = nextRegionSurfPt.latIndex + latRegionGridToResultGrid;
			resultLonIndex = nextRegionSurfPt.lonIndex + lonRegionGridToResultGrid;

				//Compute the resultant arcs for each surface point dependent on facing.
			//if((nextRegionSurfPt.facing & 1) == 1){	//Edge to the UNE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUNE.begin(); expSurfIterator != expansionTemplate->quadsUNE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 2) == 2){	//Edge to the UNW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUNW.begin(); expSurfIterator != expansionTemplate->quadsUNW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 4) == 4){	//Edge to the USE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUSE.begin(); expSurfIterator != expansionTemplate->quadsUSE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 8) == 8){	//Edge to the USW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUSW.begin(); expSurfIterator != expansionTemplate->quadsUSW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 16) == 16){	//Edge to the DNE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDNE.begin(); expSurfIterator != expansionTemplate->quadsDNE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 32) == 32){	//Edge to the DNW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDNW.begin(); expSurfIterator != expansionTemplate->quadsDNW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 64) == 64){	//Edge to the DSE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDSE.begin(); expSurfIterator != expansionTemplate->quadsDSE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 128) == 128){	//Edge to the DSW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDSW.begin(); expSurfIterator != expansionTemplate->quadsDSW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if
		}	//end for surface iterator (quads)



















			//Now walk through every surface point of the original containment region on the corner of a cube.
		for (regionSurfIterator = surfaceQuadList.begin(); regionSurfIterator != surfaceQuadList.end(); ++regionSurfIterator){
			nextRegionSurfPt = (*regionSurfIterator);
				//Find corresponding offset to the resultant containment region.
			resultLatIndex = nextRegionSurfPt.latIndex + latRegionGridToResultGrid;
			resultLonIndex = nextRegionSurfPt.lonIndex + lonRegionGridToResultGrid;

				//Compute the resultant arcs for each surface point dependent on facing.
			//if((nextRegionSurfPt.facing & 1) == 1){	//Edge to the UNE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUNE.begin(); expSurfIterator != expansionTemplate->quadsUNE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 2) == 2){	//Edge to the UNW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUNW.begin(); expSurfIterator != expansionTemplate->quadsUNW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 4) == 4){	//Edge to the USE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUSE.begin(); expSurfIterator != expansionTemplate->quadsUSE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 8) == 8){	//Edge to the USW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsUSW.begin(); expSurfIterator != expansionTemplate->quadsUSW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 16) == 16){	//Edge to the DNE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDNE.begin(); expSurfIterator != expansionTemplate->quadsDNE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 32) == 32){	//Edge to the DNW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDNW.begin(); expSurfIterator != expansionTemplate->quadsDNW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;
					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;

				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 64) == 64){	//Edge to the DSE
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDSE.begin(); expSurfIterator != expansionTemplate->quadsDSE.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;

					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;
				}	//end for expansion iterator
				//}	//end if

				//if((nextRegionSurfPt.facing & 128) == 128){	//Edge to the DSW
					//Now walk through every arc point of the expansion volume
				for (expSurfIterator = expansionTemplate->quadsDSW.begin(); expSurfIterator != expansionTemplate->quadsDSW.end(); ++expSurfIterator){
					nextExpSurfPt = (*expSurfIterator);
						//Find out out depth indices align
					resultDepth = nextRegionSurfPt.depthIndex + nextExpSurfPt.depthIndex - depthMidpoint;
						//Flatten the quad along the surface and floor.
					if(resultDepth < 0)
						resultDepth = 0;
					if(resultDepth >= resultDim[2])
						resultDepth = resultDim[2] -1;
					resultDepthOffset = resultDepth * resultSizeOfGridArea;

					resultLatOffset = (nextExpSurfPt.latIndex - expLatCenter + resultLatIndex);
					if (resultLatOffset<0) resultLatOffset = 0;
					if (resultLatOffset>=resultDim[1]) resultLatOffset = resultDim[1]-1;
					resultLatOffset = resultLatOffset * resultDim[0] + resultDepthOffset;
					
					resultLonOffset = (nextExpSurfPt.lonIndex - expLonCenter + resultLonIndex); 
					if (resultLonOffset<0) resultLonOffset = 0;
					if (resultLonOffset>=resultDim[0]) resultLonOffset = resultDim[0]-1;
					resultLonOffset+= resultLatOffset;
					resultRgn->volumeData[resultLonOffset] = 1;
				}	//end for expansion iterator
				//}	//end if
		}	//end for surface iterator (quads)


		delete expansionTemplate;
		resultRgn->floodFillRegion();
		resultRgn->findSurfacePoints();

	}	//end if

	return resultRgn;
}	//end expandRegion()




	/*
		This method will take two containment regions and return a containment region which is
		  the intersection of the two input regions.  The source name is set to a concatination
		  of the first region's name plus the second region's name.  The target ID for both of the
		  original regions is presumed to be the same and internally the new target ID is set to
		  the ID of the first region.  No attempt is made to validate if it is logical that the 
		  intersection of these two regions makes sense (e.g., different targets).  If the two 
		  regions do not intersect, the method returns NULL.

		@param newID - An integer which is a unique identifier for the resultant containment region.
		@param refTime - A long integer which represents the time when the given dimensions of the
				volume data for the region are valid.
		@param colors  - A pointer to a three element float array specifying the color to be used 
				in rendering this particular containment region.  The nature of the color space 
				used and the spread of values are not specified.  The information contained in 
				in the colors[] array are only stored to be later passed to the rendering engine.
		@param trans  - A float specifying the transparancy (or opacity) of the containment region
				when rendered.  The nature of the transparancy and the spread of values are not 
				specified.  The information contained is only stored to be later passed to the 
				rendering engine.
		@param firstRegion - The first CContainmentRegion object which is to be considered for
				determination of the intersection volume.
		@param secondRegion - The second CContainmentRegion object which is to be considered for
				determination of the intersection volume.
	*/
	CContainmentRegion* CContainmentRegion::intersectRegions(int newID, long int refTime, 
																	float *colors, float trans, 
																	CContainmentRegion *firstRegion, 
																	CContainmentRegion *secondRegion){

		CContainmentRegion* resultRegion = NULL;

			//Find smallest grid spacing
		float minErr = firstRegion->getAllowedRangeError();
		if(minErr > secondRegion->getAllowedRangeError())
			minErr = secondRegion->getAllowedRangeError();

			//Find overlap between the two grids
		FloatRect resultSize = firstRegion->getExtent();
		FloatRect secondExtent = secondRegion->getExtent();

			//Set resultSize to the most limiting of the two regions (i.e., overlap)
		if(resultSize.minLon < secondExtent.minLon)
			resultSize.minLon = secondExtent.minLon;
		if(resultSize.maxLon > secondExtent.maxLon)
			resultSize.maxLon = secondExtent.maxLon;

		if(resultSize.minLat < secondExtent.minLat)
			resultSize.minLat = secondExtent.minLat;
		if(resultSize.maxLat > secondExtent.maxLat)
			resultSize.maxLat = secondExtent.maxLat;

			//Test to see if overlap exists
		if((resultSize.minLon > resultSize.maxLon) || (resultSize.minLat > resultSize.maxLat))
			return resultRegion;

			//Make new name by concatinating the two source names
		//char *newSource = strcat(firstRegion->getSourceString(), secondRegion->getSourceString());
		char *newSource;
		if (strcmp(firstRegion->getSourceString(), secondRegion->getSourceString())) {
 		  newSource = new char[strlen(firstRegion->getSourceString())+strlen(secondRegion->getSourceString())+1];
		  strcpy(newSource,firstRegion->getSourceString());
		  strcat(newSource,secondRegion->getSourceString());
	 	} else {
		  newSource = new char[strlen(firstRegion->getSourceString())+1];
		  strcpy(newSource,firstRegion->getSourceString());
		}
		//Create new region for intersection
		resultRegion = new CContainmentRegion(newID, newSource, refTime, 
													resultSize.minLat, resultSize.maxLat,
													resultSize.minLon, resultSize.maxLon, 
													minErr, colors, trans, 
													firstRegion->getTargetID(), FILL_NONE);

		int* firstDims  = firstRegion->getGridDimensions();
		int* secondDims = secondRegion->getGridDimensions();
		int* resultDims = resultRegion->getGridDimensions();
		long int firstMidOffset  = firstDims[1] * firstDims[0];
		long int secondMidOffset = secondDims[1] * secondDims[0];
		long int midOffset = resultDims[1] * resultDims[0];

		long int lonOffset, latOffset, depthOffset;
		float targetLon, targetLat, targetDepth;
		int targetLonGrid, targetLatGrid, targetDepthGrid;
		long int targetOffset;


			//Scan the result grid and using lat and lon (avoids gaps if grid spacing is dissimilar),
			//  querry each of the source regions to determine if the volume exists.
		for(int k = 0; k < resultDims[2]; ++k){
			depthOffset = k * midOffset;
			targetDepth = resultRegion->getDepthFromGrid(k);
			for(int j = 0; j < resultDims[1]; ++j){
				latOffset = j * resultRegion->volumeDimension[0] + depthOffset;
				targetLat = resultRegion->getLatFromGrid(j);
				for(int i = 0; i < resultDims[0]; ++i){
					lonOffset = i + latOffset;
					targetLon = resultRegion->getLonFromGrid(i);

						//Get corresponding data from first region
					targetLonGrid   = firstRegion->getGridFromLon(targetLon);
					targetLatGrid   = firstRegion->getGridFromLat(targetLat);
					targetDepthGrid = firstRegion->getGridFromDepth(targetDepth);
					targetOffset    = firstRegion->getOffset(targetLonGrid, targetLatGrid, 
																			targetDepthGrid, firstMidOffset);
						//The -1 ensures that in the result, all values except overlap areas are < 1
					resultRegion->volumeData[lonOffset] = firstRegion->volumeData[targetOffset] - 1;

						//Repeat for the second region
					targetLonGrid   = secondRegion->getGridFromLon(targetLon);
					targetLatGrid   = secondRegion->getGridFromLat(targetLat);
					targetDepthGrid = firstRegion->getGridFromDepth(targetDepth);
					targetOffset    = secondRegion->getOffset(targetLonGrid, targetLatGrid, 
																			targetDepthGrid, secondMidOffset);
					resultRegion->volumeData[lonOffset] = resultRegion->volumeData[lonOffset] + secondRegion->volumeData[targetOffset];
				}	//end for
			}	//end for
		}	//end for

		resultRegion->floodFillRegion();
		resultRegion->findSurfacePoints();

		return resultRegion;
	}	//end intersectRegions()



/**************************************************************************************
---------------------------------------------------------------------------------------
		Public methods - Accessors & Manipulators
---------------------------------------------------------------------------------------
**************************************************************************************/


	/*
		This method will take a list of vertices and a list of triangles which make up a shape
		  defined by a mesh bounded surface, and convert that shape into an enumerated volume.
		  The method converts the position of a surface point into a parametric equation where
		  P0*t + (P1 * r + P2 * (1 - r)) * ( 1 - t) describes the point on the surface.  Values
		  of t and r are selected so that each step through the equation results in an evaluation
		  within one half the allowed range error.  As the outer loop walks the value of r between
		  P1 and P2, the effective length of the t leg changes (terminus point of t is a point
		  on the r leg) so the value must be adjusted as r is incremented.  This new terminal
		  point is identified by the v subsrcipt.
		The method will guarantee that all the grid points on the surface of the resultant are
			filled and continuous as long as the underlying mesh had no gaps or spaces greater than 
			half the allowable range error.

		@param verticeList - A pointer to a float array of n X 3 objects, representing n vertices
			with three (lon, lat, depth) values for each.
		@param numVerts - An integer specifying the number of vertices in the source mesh.
		@param cornerList - A pointer to an integer array of m X 3 objects, representing m triangles
			with three corner indices for each.  The values of the indices relate to the array
			index, n, of the verticeList.
		@param numTriangles - An integer specifying the number of triangles in the source mesh.
  */
void CContainmentRegion::makeRegionFromMesh(float (*verticeList)[3], int numVerts, 
														  int (*cornerList)[3], int numTriangles){
	int (*gridList)[3] = new int[numVerts][3];
	long int sizeOfGridArea = volumeDimension[1] * volumeDimension[0];

	float P0X, P0Y, P0Z;
	float P1X, P1Y, P1Z;
	float P2X, P2Y, P2Z;
	float PvX, PvY, PvZ;

	float pointV[3] = {0.0, 0.0, 0.0};
	float yardsP0v, yardsP12;
	float t = 0.0;
	float r = 0.0;
	float tStep, rStep;
	float tLat, tLon, tDepth;
	int lonGrid, latGrid, depthGrid;


		//Translate the vertice list into a grid coordinate list and to save time, add each
		//  vertex to the suface of the enumerated voulume.
	for(int i = 0; i < numVerts; ++i){
		gridList[i][0] = getGridFromLon(verticeList[i][0]);
		gridList[i][1] = getGridFromLat(verticeList[i][1]);
		gridList[i][2] = getGridFromDepth(verticeList[i][2]);
		addSurfacePointToRegion(gridList[i][0], gridList[i][1], gridList[i][2], sizeOfGridArea);
	}	//end for


		//Convert each triangle
	for(int j = 0; j < numTriangles; ++j){
			//Find relative lengths to later compute parametric parameters
		P0X = verticeList[ cornerList[j][0] ][0];
		P0Y = verticeList[ cornerList[j][0] ][1];
		P0Z = verticeList[ cornerList[j][0] ][2];

		P1X = verticeList[ cornerList[j][1] ][0];
		P1Y = verticeList[ cornerList[j][1] ][1];
		P1Z = verticeList[ cornerList[j][1] ][2];

		P2X = verticeList[ cornerList[j][2] ][0];
		P2Y = verticeList[ cornerList[j][2] ][1];
		P2Z = verticeList[ cornerList[j][2] ][2];

		PvX = P1X;
		PvY = P1Y;
		PvZ = P1Z;

			//Find the distance, in yards, of two legs of the triangle.
		yardsP0v = distanceBetweenPoints(verticeList[ cornerList[j][0] ], verticeList[ cornerList[j][1] ]);
		yardsP12 = distanceBetweenPoints(verticeList[ cornerList[j][1] ], verticeList[ cornerList[j][2] ]);

			//Find what percentage of the triangle leg is half the allowed error.
		rStep = allowedRangeError * 0.5 / yardsP12;
		tStep = allowedRangeError * 0.5 / yardsP0v;

			//Since vertex already done, skip t = r = 0 and go into loop to walk down first leg
			//  of trangle evaluating at different t points.
		r = 0.0;
		t = tStep;
		while(r < 1.0){
				//Scan down t leg.
			while(t < 1.0){
				tLon   = P0X * ( 1 - t) + PvX * t;
				tLat   = P0Y * ( 1 - t) + PvY * t;
				tDepth = P0Z * ( 1 - t) + PvZ * t;
				lonGrid   = getGridFromLon(tLon);
				latGrid   = getGridFromLat(tLat);
				depthGrid = getGridFromDepth(tDepth);
				addSurfacePointToRegion(lonGrid, latGrid, depthGrid, sizeOfGridArea);
				t += tStep;		//Increment the t iterator
			}	//end while t

				//Move a step along the r leg and compute new Pv for the t leg.
			r += rStep;

			PvX = P1X * ( 1 - r) + P2X * r;
			PvY = P1Y * ( 1 - r) + P2Y * r;
			PvZ = P1Z * ( 1 - r) + P2Z * r;

				//Find the distance, in yards to the new endpoint of the t leg
			pointV[0] = PvX;
			pointV[1] = PvY;
			pointV[2] = PvZ;
			yardsP0v = distanceBetweenPoints(verticeList[ cornerList[j][0] ], pointV);

			tStep = allowedRangeError * 0.5 / yardsP0v;
			t = tStep;			//Reset the t iterator
		}	//end while r
	}	//end for
	floodFillRegion();
	findSurfacePoints();
}	//end makeRegionFromMesh()


	/*
		This method takes two positions within the containment region and computes the distance, 
		  in yards, between them.

		@param point1 - A three element float array which is longitude, latitude, and depth of 
			the first point to be evaluated.
		@param point2 - A three element float array which is longitude, latitude, and depth of 
			the second point to be evaluated.
		@returns - A float which is the distance, in yards, between points one and two.
	*/
float CContainmentRegion::distanceBetweenPoints(float point1[3], float point2[3]){
	float latInYards = (point1[1] - point2[1]) * MIN_PER_DEGREE * YARDS_PER_NM;
	float lonInYards = (point1[0] - point2[0]) * cos(centerLat * DEG_TO_RAD) * MIN_PER_DEGREE * YARDS_PER_NM;
	float depthInYards = (point1[2] - point2[2]) / 3.0;

	return sqrt(latInYards * latInYards + lonInYards * lonInYards + depthInYards * depthInYards);
}	//end distanceBetweenPoints(float[3], float[3])


	/*
		This method takes the position of a surface point which is already defined in grid coordinates
		  and adds it to the volume data.

		@param lon - An integer which is the X grid index (longitude) of the volume data for the region.
		@param lat - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param depth - An integer which is the Z grid index (depth) of the volume data for the region.
		@param offset - A long integer which is the product of the X and Y grid indices for the
								volume data for the region.  Precomputed and passed to avoid repetition.
	*/
void CContainmentRegion::addSurfacePointToRegion(int lon, int lat, int depth, long int offset){
	volumeData[lon + lat * volumeDimension[0] + depth * offset] = 1;
}	//end addSurfacePointToRegion(int, int, int, long int)


	/*
		Simple accessor, returns string of the name of the region.

		@returns - A pointer to characters which is the name of the region.
	*/
char* CContainmentRegion::getSourceString(){
	char* result = regionSources;

	return result;
}	//end getSourceString()


	/*
		Simple manipulator, sets target ID number associated with the region.

		@param ID - An integer which is the new targetID to use with the region.
	*/
void CContainmentRegion::setTargetID(int ID){
	assocTargetID = ID;
}	//end setTargetID()


	/*
		Simple accessor, returns target ID number associated with the region.

		@returns - An integer which is the targetID number associated with the region.
	*/
int CContainmentRegion::getTargetID(){
	return assocTargetID;
}	//end getTargetID()


/**************************************************************************************
---------------------------------------------------------------------------------------
		Private methods
---------------------------------------------------------------------------------------
**************************************************************************************/

	/*
		This method will take the containment region and fill it with a spherical expansion
	     volume as specified by the fillType variable.  The fillType is an integer which is an 
		  enumerated constant (see listings of constants in the remarks at the top of the class) 
		  describing what to do with the interior volume after instantiation.  A value of 
		  FILL_NONE ( 0 ) is most commonly used and if specified, this method will clear
		  any existing volume data.

		@param fill - An enumerated integer, the options for which is identified in the
							list of class constants.  Specifies how the sphere should fill the
							cubic region.
	*/
void CContainmentRegion::fillRegion(int fill){
		//Find the regions lat/lon size
	int areaDim = volumeDimension[0];

		//If fill none, clear the surface list
	if(fill == FILL_NONE){
		if(!volumeData.empty())
			volumeData.clear();
		for (long int x = 0; x < vectorSize; x++)
			volumeData.push_back(0);	
	}	//end if

		//Find appropriate lat/lon limits
	else if(fill == FILL_VOLUME)
		if(volumeDimension[1] > areaDim)
			areaDim = volumeDimension[1];

	else if(fill == FILL_AREA)
		if(volumeDimension[1] < areaDim)
			areaDim = volumeDimension[1];

		//If depth include, find limits
	else if(fill == LIMIT_VOLUME){
		if(volumeDimension[1] < areaDim)
			areaDim = volumeDimension[1];
		if(volumeDimension[2] < areaDim)
			areaDim = volumeDimension[2];
	}	//end if

		//Gerneate a sphere within the containment region and update vector
	long int lonOffset, latOffset, depthOffset;
	float radius = (float)(areaDim - 1) / 2.0;	//The -1 adjusts for dimension including edge pt.
	double x, y, z, sphereValue;
	double ySqr, zSqr; 
	double radiusSqr = radius * radius;

	if((fill > FILL_NONE) && (fill <= LIMIT_VOLUME)){	//Create volume		
			//Clear the existing data from the vector to add new data.
		volumeData.clear();
		long int midOffset = volumeDimension[1] * volumeDimension[0];

		for(int k=0; k < volumeDimension[2]; ++k){
			z = k - regionCenter[2];
			zSqr = z * z;
			depthOffset = k * midOffset;
			for(int j=0; j < volumeDimension[1]; ++j){
				y = j - regionCenter[1];
				ySqr = y * y;
				latOffset = j * volumeDimension[0] + depthOffset;
				for(int i=0; i < volumeDimension[0]; ++i){
					x = i - regionCenter[0];
					sphereValue = x*x + ySqr + zSqr - radiusSqr;
					lonOffset = i + latOffset;
					if(sphereValue <= 0)
						volumeData.push_back(1);	
					else
						volumeData.push_back(0);
				}	//end for i
			}	//end for j
		}	//end for k
		findSurfacePoints();
	}	//end if
}	//end fillRegion()



	/*	
	  This method will test a point of the volume to ascertain to ascertain the local
	    topology.  If the point is not on the surface, it will return a hash value of zero.
		 A point is considered to be on the surface if any adjacent grid point (NOT counting 
		 diagonals) are zero.  Surface points return a hash value identifying which adjacent
		 grid points are zero.  By use of a bitwise summation, the local topography of the surface 
		 can be ascertained.  The bit values are:

			1  = Surface opens to the EAST
			2  = Surface opens to the WEST
			4  = Surface opens to the NORTH
			8  = Surface opens to the SOUTH
			16 = Surface opens to the UP
			32 = Surface opens to the DOWN

	A surface point with a final hash value of one of the above powers of two means that
	  the surface is a flat face and subsequent expansion will be by face points. A hash value 
	  which is a combination of two opposing directions (1 & 2, 4 & 8, or 16 & 32) imply the 
	  surface is a thin flat sheet at the point in question.  Subsequent expansion will be by 
	  completing an entire sphere about the point. A hash value which is a combination of two 
	  non-opposing directions (1 & anything but 2, 4 & anything but 8, or 16 & anything but 32) 
	  define a ledge like structure.  Subsequent expansion will be by arcs.  Finally, a hash
	  value which represents the combination of any three, one from each set of similar pairs
	  ( 1 (or 2) & 4 (or 8) & 16 (or 32) ) define a corner cube and subsequent expansion will
	  be by quads.

		@param i - An integer which is the X grid index (longitude) of the volume data for the region.
		@param j - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param k - An integer which is the Z grid index (depth) of the volume data for the region.
		@param midOff - A long integer which is the product of the X and Y grid indices for the
								volume data for the region.  Precomputed and passed to avoid repetition.
		@returns - An integer which is a binary encoded hash value representing which adajacent
								points are open.
	*/	
int CContainmentRegion::findPointFacing(int i, int j, int k, long int midOff){
		//Start by presuming it is an interior point
	int hash = 0;

		//Find and record any exterior surfaces
		//East (lon) side is tested to see if it is a suface
	if(i == volumeDimension[0] - 1){	//Avoid exceeding array bounds
		hash = (hash | 1);
		//printf("east bound\n");
	} else if(volumeData[getOffset(i + 1, j, k, midOff)] == 0) {
		hash = (hash | 1);
		//printf("east empty\n");
	}
		//West (lon) side is tested to see if it is a suface
	if(i == 0){	                  //Avoid exceeding array bounds
		hash = (hash | 2);
		//printf("west bound\n");
	} else if(volumeData[getOffset(i - 1, j, k, midOff)] == 0) {
		hash = (hash | 2);
		//printf("west empty\n");
	}
		//North (lat) side is tested to see if it is a suface
	if(j == volumeDimension[1] - 1){	//Avoid exceeding array bounds
		hash = (hash | 4);
		//printf("north bound\n");
	} else if(volumeData[getOffset(i, j + 1, k, midOff)] == 0) {
		hash = (hash | 4);
		//printf("north empty\n");
	}
		//South (lat) side is tested to see if it is a suface
	if(j == 0){	                  //Avoid exceeding array bounds
		hash = (hash | 8);
		//printf("south bound\n");
	} else if(volumeData[getOffset(i, j - 1, k, midOff)] == 0) {
		hash = (hash | 8);
		//printf("south empty\n");
	}
		//Up (depth) side is tested to see if it is a suface
	if(k == 0){	                  //Avoid exceeding array bounds
	  //printf("up bound\n");
		hash = (hash | 16);
	} else if(volumeData[getOffset(i, j, k - 1, midOff)] == 0) {
		hash = (hash | 16);
		//printf("up empty\n");
	}
		//Down (depth) side is tested to see if it is at the floor
	if(k == volumeDimension[2] - 1){	//Avoid exceeding array bounds
	  //printf("down bound\n");
		hash = (hash | 32);
	} else if(volumeData[getOffset(i, j, k + 1, midOff)] == 0) {
		hash = (hash | 32);
		//printf("down empty\n");
	}
	return hash;
}	//end findPointFacing()



	/*
	  Takes a surface point and its associated hash value and quantifies the local topology at that
		point.  The topology information is later determines which spherical segments will need to 
		be plotted if the volume is expanded and how the surface mesh should be recreated from the
		enumerated volume.  The original hash value is used as an entry argument into a lookup 
		table which will output one of the following new binary codes identifying the local
		topology.  The values are catagoried into three generic types:  (1) The surface opens from 
		the referenced point only in the direction normal to the surface (point), (2) The surface
		opens in two orthagonal directions, effectively defining a 90 degree ledge (arc), or
		(3)  The surface opens in three orthagonal directions, effectively defining a point on
		the tip of a corner (quad).

	  pointHash encoding (plotting new or expanded surface requies only a single point):
	    1  = Plot only point directly east.
	    2  = Plot only point directly west.
	    4  = Plot only point directly north.
	    8  = Plot only point directly south.
	    16 = Plot only point directly up.
	    32 = Plot only point directly down.

	  arcHash encoding (plotting new or expanded surface requies an arc segment):
	    1    = Plot arc of points in NE corner of plane.
	    2    = Plot arc of points in NW corner of plane.
	    4    = Plot arc of points in SE corner of plane.
	    8    = Plot arc of points in SW corner of plane.
	    16   = Plot arc of points in UE corner of plane.
	    32   = Plot arc of points in UW corner of plane.
	    64   = Plot arc of points in DE corner of plane.
	    128  = Plot arc of points in DW corner of plane.
	    256  = Plot arc of points in UN corner of plane.
	    512  = Plot arc of points in US corner of plane.
	    1024 = Plot arc of points in DN corner of plane.
	    2048 = Plot arc of points in DS corner of plane.

	  quadHash encoding (plotting new or expanded surface requies a spherical quadrant):
	    1    = Plot all the points in UNE quadrant of volume.
	    2    = Plot all the points in UNW quadrant of volume.
	    4    = Plot all the points in USE quadrant of volume.
	    8    = Plot all the points in USW quadrant of volume.
	    16   = Plot all the points in DNE quadrant of volume.
	    32   = Plot all the points in DNW quadrant of volume.
	    64   = Plot all the points in DSE quadrant of volume.
	    128  = Plot all the points in DSW quadrant of volume.

		The original hash value is translated into one of the possibilities above and then the
			data associated with the point is recorded into a list of like points along with the
			new translated hash value.
		See also the documetation for the method findPointFacing() which develops the original
			hash code used as the entry argument.

		@param vecIndex - A long integer which is the vector offset equivalent to grid[i][j][k].
		@param lnIndex - An integer which is the X grid index (longitude) of the volume data for the region.
		@param ltIndex - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param depIndex - An integer which is the Z grid index (depth) of the volume data for the region.
		@param hash - An integer which is the original hash value descibing the local topology.
			1  = Surface opens to the EAST
			2  = Surface opens to the WEST
			4  = Surface opens to the NORTH
			8  = Surface opens to the SOUTH
			16 = Surface opens to the UP
			32 = Surface opens to the DOWN
	*/
void CContainmentRegion::addSurfacePointToList(long int vecIndex, int lnIndex, 
															  int ltIndex, int depIndex, int hash){
	ListIndices newIndex;
		newIndex.vectorIndex = vecIndex;
		newIndex.lonIndex = lnIndex;
		newIndex.latIndex = ltIndex;
		newIndex.depthIndex = depIndex;
		newIndex.facing = hash;

		int pointHash[3] = {0, 0, 0};

		//Creates a lookup table to translate hash values into arcs and quads.  The
		//  first argument is the original hash value.  The second value is the index
		//  for (0 = points, 1 = arcs, 2 = quads).

		/*int pointArcQuadIndex[64][3] = { { 0,    0, 0}, {1,  0,   0}, {2,   0,   0}, {3,   0,   0}, 
						 { 4,    0, 0}, {0,4095,   0}, {0,4095,   0}, {0,4095,   0}, 
						 { 8,    0, 0}, {0,4095,   0}, {0,4095,   0}, {0,4095,   0}, 
						 {12,    0, 0}, {0,4095,   0}, {0,4095,   0}, {0,4095,   0},
						 
						 {16,    0, 0}, {0,4095,   0}, {0,4095,   0}, {0,4095,   0}, 
						 { 0, 4095, 0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255}, 
						 { 0, 4095, 0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255}, 
						 { 0, 4095, 0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255},
						 
						 {32,    0, 0}, {0,4095,   0}, {0,4095,   0}, {0,4095,   0}, 
						 { 0, 4095, 0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255}, 
						 { 0, 4095, 0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255}, 
						 { 0, 4095, 0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255},
						 
						 {48,    0, 0}, {0,4095,   0}, {0,4095,   0}, {0,4095,   0}, 
						 { 0, 4095, 0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255}, 
						 { 0, 4095, 0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255}, 
						 { 0, 4095, 0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255}};
		*/

		int pointArcQuadIndex[64][3] = { { 0,    0, 0}, {1,  0,   0}, {2,   0,   0}, {3,   0,   0}, 
						 { 4,    0, 0}, {0,  1,   0}, {0,   2,   0}, {0,   3,   0}, 
						 { 8,    0, 0}, {0,  4,   0}, {0,   8,   0}, {0,  12,   0}, 
						 {12,    0, 0}, {0,  5,   0}, {0,  10,   0}, {0,  15,   0},
						 
						 {16,    0, 0}, {0, 16,   0}, {0,  32,   0}, {0,  48,   0}, 
						 { 0,  256, 0}, {0,  0,   1}, {0,   0,   2}, {0,   0,   3}, 
						 { 0,  512, 0}, {0,  0,   4}, {0,   0,   8}, {0,   0,  12}, 
						 { 0,  768, 0}, {0,  0,   5}, {0,   0,  10}, {0,   0,  15},
						 
						 {32,    0, 0}, {0, 64,   0}, {0, 128,   0}, {0, 192,   0}, 
						 { 0, 1024, 0}, {0,  0,  16}, {0,   0,  32}, {0,   0,  48}, 
						 { 0, 2048, 0}, {0,  0,  64}, {0,   0, 128}, {0,   0, 192}, 
						 { 0, 3072, 0}, {0,  0,  80}, {0,   0, 160}, {0,   0, 240},
						 
						 {48,    0, 0}, {0, 80,   0}, {0, 160,   0}, {0, 240,   0}, 
						 { 0, 1280, 0}, {0,  0,  17}, {0,   0,  34}, {0,   0,  51}, 
						 { 0, 2560, 0}, {0,  0,  68}, {0,   0, 136}, {0,   0, 204}, 
						 { 0, 3840, 0}, {0,  0,  85}, {0,   0, 170}, {0,   0, 255}};
		
		
		/*int pointArcQuadIndex[64][3] = { { 0,    0, 0}, {1,  0,   0}, {2,   0,   0}, {3,  0,   0}, 
						 { 4,    0, 0}, {0,  0, 255}, {0,   0, 255}, {0,  0, 255}, 
						 { 8,    0, 0}, {0,  0, 255}, {0,   0, 255}, {0,  0, 255}, 
						 {12,    0, 0}, {0,  0, 255}, {0,   0, 255}, {0,  0, 255},
						 
						 {16,    0,  0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255}, 
						 { 0,    0,255}, {0,  0,   1}, {0,   0,   2}, {0,   0,   3}, 
						 { 0,    0,255}, {0,  0,   4}, {0,   0,   8}, {0,   0,  12}, 
						 { 0,    0,255}, {0,  0,   5}, {0,   0,  10}, {0,   0,  15},
						 
						 {32,    0, 0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255}, 
						 { 0, 0, 255}, {0,  0,  16}, {0,   0,  32}, {0,   0,  48}, 
						 { 0, 0, 255}, {0,  0,  64}, {0,   0, 128}, {0,   0, 192}, 
						 { 0, 0,  255}, {0,  0,  80}, {0,   0, 160}, {0,   0, 240},
						 
						 {48, 0,   0}, {0,  0, 255}, {0,   0, 255}, {0,   0, 255}, 
						 { 0, 0, 255}, {0,  0,  17}, {0,   0,  34}, {0,   0,  51}, 
						 { 0, 0, 255}, {0,  0,  68}, {0,   0, 136}, {0,   0, 204}, 
						 { 0, 0, 255}, {0,  0,  85}, {0,   0, 170}, {0,   0, 255}};
		

		*/
		//Add point to appropriate specialty list
	pointHash[0] = pointArcQuadIndex[hash][0];
	pointHash[1] = pointArcQuadIndex[hash][1];
	pointHash[2] = pointArcQuadIndex[hash][2];

		//Surface points
	if(pointHash[0] > 0){
		newIndex.facing = pointHash[0];
		surfacePointList.push_back(newIndex);
	}	//end if

		//Edge areas
	else if(pointHash[1] > 0){
		newIndex.facing = pointHash[1];
		surfaceArcList.push_back(newIndex);
	}	//end else if

		//Surface corners
	else{
		newIndex.facing = pointHash[2];
		surfaceQuadList.push_back(newIndex);
	}	//end else
}	//end addSurfacePointToList()


	/*
		The recursive portion of the flood fill algorithm.  If point tested is zero (an empty
			exterior point), it is set to two and then the six adjacent points are tested.  When
			complete, all exterior contiguous points will have a value of two.

		@param lon - An integer which is the X grid index (longitude) of the volume data for the region.
		@param lat - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param depth - An integer which is the Z grid index (depth) of the volume data for the region.
		@param depthOffset - A long integer which is the product of the X and Y grid indices for the
								volume data for the region.  Precomputed and passed to avoid repetition.
	*/
void CContainmentRegion::markExteriorPoint(int lon, int lat, int depth, int depthOffset){
	int x = 0;

	long int lonOffset = getOffset(lon, lat, depth, depthOffset);
	if(volumeData[lonOffset] < 1){
		volumeData[lonOffset] = 2;
		x = lon + 1;
		if(x <  volumeDimension[0] - 1)
			markExteriorPoint(x, lat, depth, depthOffset);

		x = lon - 1;
		if(x > 0)
			markExteriorPoint(x, lat, depth, depthOffset);

		x = lat + 1;
		if(x <  volumeDimension[1] - 1)
			markExteriorPoint(lon, x, depth, depthOffset);

		x = lat - 1;
		if(x > 0)
			markExteriorPoint(lon, x, depth, depthOffset);

		x = depth + 1;
		if(x < volumeDimension[2] -1)
			markExteriorPoint(lon, lat, x, depthOffset);

		x = depth - 1;
		if(x > 0)
			markExteriorPoint(lon, lat, x, depthOffset);
	}	//end if
}	//end markExteriorPoint()


	/*
		Data for the enumerated volumes is stored in a vector rather than a three dimensional
		  array.  This method compresses the repeated need to set data points in a volume.  It
		  calculates the vector offsets corresponding to the three dimensional input (longitude,
		  latitude, and depth) and sets the data value for that point to one.

  		@param lonIdx   - A long integer of the enumerated volume's equivalent 3D array grid index
								  in the X (longitude) direction.
  		@param latIdx   - A long integer of the enumerated volume's equivalent 3D array grid index
								  in the Y (latitude) direction.
  		@param depthIdx - A long integer of the enumerated volume's equivalent 3D array grid index
								  in the Z (depth) direction.
  		@param lonDim   - An integer of the enumerated volume's equivalent 3D array grid size
								  in the X (longitude) direction.
		@param region  - The CContainmentRegion object which is to have a point in its data set
									updated.
	*/
void CContainmentRegion::addDataPointToRegion(long int lonIdx, long int latIdx, long int depthIdx, 
													 int lonDim, CContainmentRegion *region){
	long int finalIndex = lonIdx + (latIdx * lonDim) + depthIdx;
	region->volumeData[finalIndex] = 1;
}	//end addDataPointToRegion()


