//ContainmentRegion.h - Specification for the associated class.


#ifndef ContainmentRegion_h
#define ContainmentRegion_h

#include "EnumeratedVolume.h"
#include "TargetMotionData.h"
#include "ExpansionTemplate.h"
#include <vector>

/* =======================================================================================
	The ContainmentRegion is used to store information about a target containment region which
	exists at a particular momement in time.  The region may then be displayed or expanded.

	This software was developed as part of a grant by ONR 311, Office of Naval Research, 
	800 North Quincy Street, Arlington, Virginia 22217, for development of an "Application 
	of Volume Modeling Techniques to Dynamic Containment Regions for Naval Applications".
	Grant was awarded to Arizona State University, Department of Computer Science and Engineering,
	Dr. Greg Nielson and Gary Graf.


	Author: Gary Graf, Arizona State University.
	Version:  1.0b, 14 Apr 2002

  All rights reserved.
  
	Notes:
=========================================================================================*/

class CContainmentRegion : public CEnumeratedVolume
{
/***************************  Public class varibles **************************************/	
public:
		//Grid coordinates of the center of the region.
	float regionCenter[3];

		//This structure is to hold the various array indices, including a computed
		//  vector index, for the containment region.  As each surface, or other special
		//  point in the volume is located, the indices are loaded into this structure
		//  so they can be store, en masse, in a vector (e.g., list of edge points).
	struct ListIndices {
		long int vectorIndex;
		int latIndex;
		int lonIndex;
		int depthIndex;
		int facing;
	};
		//Define a vector to hold the list of points which define the outside surface
		//  of the containment region.
	typedef std::vector<ListIndices> LIST_INDEX_VECTOR;

		//Vectors which will hold the list of all the points which lie on the outside surface
		//  of the containment region and whose points on a flat face, shelf, or corner, respectively.
	LIST_INDEX_VECTOR surfacePointList;
	LIST_INDEX_VECTOR surfaceArcList;
	LIST_INDEX_VECTOR surfaceQuadList;

		//This variable identifies if the containment region has been modified (expanded and/or
		//  clipped) for display.  Needed to decide which regions must be deleted from memory
		//  when the display list is cleared.
	bool isModified;

	
/***************************  Private class varibles ***********************************/ 
private:
		//A string concatinating all the names of the data sources which define the region.
	char* regionSources;

		//Is the ID of the target
	int assocTargetID;

		//Flag to define the type of region fill used.  See the list of the const static variables
		//  in the .cpp file to identify the available options.
	int fillType;

/***************************  Public  methods ********************************************/
public:
		//Constructors and destructors

	/*
		The default constructor establishes an enumerated volume of approximately, 200 kyds x
		200 kyds, 500 yd spacing, origin at 25 deg N, 25 deg E, with no ID number.
	*/
	CContainmentRegion();


	/*
		This constructor will take a min latitude and longitude along with the required grid
		dimensions to compute the final extent of the region.  It is primarily used in the
		construction of resultant containment regions following the expansion of another region.  

		@param newID - An integer which is a unique identifier for the containment region.
		@param source - A pointer to a string of charecters.  This is inteded to be used as
				a verbal identifier of the region.  It can be concatinated to identify intersection
				regions as the sum of it parts.  Its exact use is undefined.
		@param refTime - A long integer which represents the time when the given dimensions of the
				volume data for the region are valid.
		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param latDim - An integer which specifies how many grid cell the region should occupy
				in latitude.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param lonDim - An integer which specifies how many grid cell the region should occupy
				in longitude.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
		@param colors  - A pointer to a three element float array specifying the color to be used 
				in rendering this particular containment region.  The nature of the color space 
				used and the spread of values are not specified.  The information contained in 
				in the colors[] array are only stored to be later passed to the rendering engine.
		@param trans  - A float specifying the transparancy (or opacity) of the containment region
				when rendered.  The nature of the transparancy and the spread of values are not 
				specified.  The information contained is only stored to be later passed to the 
				rendering engine.
		@param tgtID - An integer which is an identifier for the target related to this containment
				region.  The ID relates internally to a CTargetMotionData object and is used to
				retrieve speed information needed to calculate an expansion volume.
	*/
	CContainmentRegion(int, char*, long int, float, float, float, float, float, float*,
												 float, int, int);

	/*
		This is the fully specified constructor.

		@param newID - An integer which is a unique identifier for the containment region.
		@param source - A pointer to a string of charecters.  This is inteded to be used as
				a verbal identifier of the region.  It can be concatinated to identify intersection
				regions as the sum of it parts.  Its exact use is undefined.
		@param refTime - A long integer which represents the time when the given dimensions of the
				volume data for the region are valid.
		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param mxLat  - A float which is the maximum latitude boundary of the region.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param mxLon  - A float which is the maximum longitude boundary of the region.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
		@param colors  - A pointer to a three element float array specifying the color to be used 
				in rendering this particular containment region.  The nature of the color space 
				used and the spread of values are not specified.  The information contained in 
				in the colors[] array are only stored to be later passed to the rendering engine.
		@param trans  - A float specifying the transparancy (or opacity) of the containment region
				when rendered.  The nature of the transparancy and the spread of values are not 
				specified.  The information contained is only stored to be later passed to the 
				rendering engine.
		@param tgtID - An integer which is an identifier for the target related to this containment
				region.  The ID relates internally to a CTargetMotionData object and is used to
				retrieve speed information needed to calculate an expansion volume.
		@param fill - An integer which is an enumerated constant (see listings of constants in
				the remarks at the top of the class) describing what to do with the interior volume
				after instantiation.  A value of FILL_NONE ( 0 ) is most commonly used.
	*/
	CContainmentRegion(int, char*, long int, float, int,   float, int,   float, float*,
												 float, int);
	~CContainmentRegion();



		//General Methods

	/*
		This method will make a containment volume and inside of it construct a modified tubular 
			section representing a pseudo hyperbolic conic section.  THIS METHOD IS USED FOR
			DEVELOPMENT PURPOSES ONLY.
	*/
	void makeTubularVolume();



	/*
	  This method will search the volume for points which lie on the surface, including
	    all points which lie on the boundary of the volume.  A surface point is defined as
		 a location where the data is greater than zero and at least one point adjacent, not
		 counting diagonals, is equal to zero.  When it finds a surface point
	    it stores the corresponding vector and array indices in the appropriate surface list.  
		 The method automatically flags boundary points as surface points and then tests the
		 interior points.
	*/	
	void findSurfacePoints();


	/*
		Newly expanded regions have a shell created by plotting the appropriate points of the 
			template (faces, arcs, and quadrants) but the interior is characterized by a hollow 
			space matching the original volume as well as broken and trapped hollows between the 
			original volume and the new outer shell. To define the correct surface for rendering 
			the interior regions must all be filled to a value of one or the surface determination 
			algorithm will incorrectly mark interior hollows as exterior surfaces. A simple flood 
			fill of the interior will not work because the space may be cut into isolated regions 
			by the plotting of intersecting arcs and quads. 
		This method solves the problem of creating a solid filled shape from irregular surface and
			point data by reversing the traditional flood fill concept.  In this method the outside
			is tested and if it is empty it is marked with a value of two.  After all the outside
			spaces are marked, the volume is scanned and the outside spaces are set to their proper
			value of zero while all other values are presumed to be interior points and set to one.
		NOTE:  If the shape itself originally had an interior void, this algorithm will eliminate
			the void entirely.
	*/
	void floodFillRegion();



	/*  
		This method will expand a containment region as a function of time.  The amount of
			expansion is first determined by getting the difference between the regions current 
			reference time and the the passed time.  This time difference is the multiplied
			by the target speed, as referenced in the regions associated TargetMotionData object,
			to derive a distance figure.
		The expansion is performed by first creating a cubical enumerated volume (possible clipped
			in depth) to act as a template.  Within this volume, data values are assigned resulting
			in a centered, spherical region of ones, zero outside the sphere.  The surface points
			of the template are catagorized as lying on a particular quadrant or along a cardinal
			meridian.
		The template is then 'walked' over the surface of the region being expanded.  Depending on
			the local facing of the current surface point, the appropriate points, arcs or quads
			from the template are mapped into a new blank region using appropriate offsets from both
			the original region as well as the template.  To enhance speed performance, only the 
			mininum number of surface points needed to develop the expanded shell are used, both on
			the region and template.
		When the template has been applied to the entire surface, the result region will consist of
			an outer shell representing the maximum extent of the expansion. Inside the shell though,
			the data will be a fractured combination of ones and zeros resulting from overlapping 
			template data.  The method floodFillRegion() is them called to fill the gaps and restore
			the data coherence of the expaned containment region.
			
		@param targetInfo  - A CTargetMotionData object which holds information as to how the region
									is to be expanded as a function of elapsed time.
		@param endTime - A long integer representing the termination or current time of expansion.
		@returns - A CContainmentRegion object which is the original passed region after expansion
							for the elapsed time. If the expansion distance (time differential times the
							target speed) is less than the allowed range error in the display, the
							object returned is the original region. Otherwise, a new object is returned 
							which is an expanded region.
	*/
	CContainmentRegion* expandRegion(CTargetMotionData*, long int);


	/*
		This method will take two containment regions and return a containment region which is
		  the intersection of the two input regions.  The source name is set to a concatination
		  of the first region's name plus the second region's name.  The target ID for both of the
		  original regions is presumed to be the same and internally the new target ID is set to
		  the ID of the first region.  No attempt is made to validate if it is logical that the 
		  intersection of these two regions makes sense (e.g., different targets).  If the two 
		  regions do not intersect, the method returns NULL.

		@param newID - An integer which is a unique identifier for the resultant containment region.
		@param refTime - A long integer which represents the time when the given dimensions of the
				volume data for the region are valid.
		@param colors  - A pointer to a three element float array specifying the color to be used 
				in rendering this particular containment region.  The nature of the color space 
				used and the spread of values are not specified.  The information contained in 
				in the colors[] array are only stored to be later passed to the rendering engine.
		@param trans  - A float specifying the transparancy (or opacity) of the containment region
				when rendered.  The nature of the transparancy and the spread of values are not 
				specified.  The information contained is only stored to be later passed to the 
				rendering engine.
		@param firstRegion - The first CContainmentRegion object which is to be considered for
				determination of the intersection volume.
		@param secondRegion - The second CContainmentRegion object which is to be considered for
				determination of the intersection volume.
	*/
	static CContainmentRegion* intersectRegions(int, long int, float*, float,
													CContainmentRegion*, CContainmentRegion*);


	/*
		This method will take a list of vertices and a list of triangles which make up a shape
		  defined by a mesh bounded surface, and convert that shape into an enumerated volume.
		  The method converts the position of a surface point into a parametric equation where
		  P0*t + (P1 * r + P2 * (1 - r)) * ( 1 - t) describes the point on the surface.  Values
		  of t and r are selected so that each step through the equation results in an evaluation
		  within one half the allowed range error.  As the outer loop walks the value of r between
		  P1 and P2, the effective length of the t leg changes (terminus point of t is a point
		  on the r leg) so the value must be adjusted as r is incremented.  This new terminal
		  point is identified by the v subsrcipt.
		The method will guarantee that all the grid points on the surface of the resultant are
			filled and continuous as long as the underlying mesh had no gaps or spaces greater than 
			half the allowable range error.

		@param verticeList - A pointer to a float array of n X 3 objects, representing n vertices
			with three (lon, lat, depth) values for each.
		@param numVerts - An integer specifying the number of vertices in the source mesh.
		@param cornerList - A pointer to an integer array of m X 3 objects, representing m triangles
			with three corner indices for each.  The values of the indices relate to the array
			index, n, of the verticeList.
		@param numTriangles - An integer specifying the number of triangles in the source mesh.
  */
	void makeRegionFromMesh(float (*verticeList)[3], int, int (*cornerList)[3], int);

	/*
		This method takes two positions within the containment region and computes the distance, 
		  in yards, between them.

		@param point1 - A three element float array which is longitude, latitude, and depth of 
			the first point to be evaluated.
		@param point2 - A three element float array which is longitude, latitude, and depth of 
			the second point to be evaluated.
		@returns - A float which is the distance, in yards, between points one and two.
	*/
	float distanceBetweenPoints(float[3], float[3]);

	/*
		This method takes the position of a surface point which is already defined in grid coordinates
		  and adds it to the volume data.

		@param lon - An integer which is the X grid index (longitude) of the volume data for the region.
		@param lat - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param depth - An integer which is the Z grid index (depth) of the volume data for the region.
		@param offset - A long integer which is the product of the X and Y grid indices for the
								volume data for the region.  Precomputed and passed to avoid repetition.
	*/
	void addSurfacePointToRegion(int, int, int, long int);


	/*
		Simple accessor, returns string of the name of the region.

		@returns - A pointer to characters which is the name of the region.
	*/
	char* getSourceString();


	/*
		Simple manipulator, sets target ID number associated with the region.

		@param ID - An integer which is the new targetID to use with the region.
	*/
	void setTargetID(int ID);


	/*
		Simple accessor, returns target ID number associated with the region.

		@returns - An integer which is the targetID number associated with the region.
	*/
	int getTargetID();


/***************************  Private methods ******************************************/

private:
	/*
		This method will take the containment region and fill it with a spherical expansion
	     volume as specified by the fillType variable.  The fillType is an integer which is an 
		  enumerated constant (see listings of constants in the remarks at the top of the class) 
		  describing what to do with the interior volume after instantiation.  A value of 
		  FILL_NONE ( 0 ) is most commonly used and if specified, this method will clear
		  any existing volume data.

		@param fill - An enumerated integer, the options for which is identified in the
							list of class constants.  Specifies how the sphere should fill the
							cubic region.
	*/
	void fillRegion(int fill);


	/*	
	  This method will test a point of the volume to ascertain to ascertain the local
	    topology.  If the point is not on the surface, it will return a hash value of zero.
		 A point is considered to be on the surface if any adjacent grid point (NOT counting 
		 diagonals) are zero.  Surface points return a hash value identifying which adjacent
		 grid points are zero.  By use of a bitwise summation, the local topography of the surface 
		 can be ascertained.  The bit values are:

			1  = Surface opens to the EAST
			2  = Surface opens to the WEST
			4  = Surface opens to the NORTH
			8  = Surface opens to the SOUTH
			16 = Surface opens to the UP
			32 = Surface opens to the DOWN

	A surface point with a final hash value of one of the above powers of two means that
	  the surface is a flat face and subsequent expansion will be by face points. A hash value 
	  which is a combination of two opposing directions (1 & 2, 4 & 8, or 16 & 32) imply the 
	  surface is a thin flat sheet at the point in question.  Subsequent expansion will be by 
	  completing an entire sphere about the point. A hash value which is a combination of two 
	  non-opposing directions (1 & anything but 2, 4 & anything but 8, or 16 & anything but 32) 
	  define a ledge like structure.  Subsequent expansion will be by arcs.  Finally, a hash
	  value which represents the combination of any three, one from each set of similar pairs
	  ( 1 (or 2) & 4 (or 8) & 16 (or 32) ) define a corner cube and subsequent expansion will
	  be by quads.

		@param i - An integer which is the X grid index (longitude) of the volume data for the region.
		@param j - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param k - An integer which is the Z grid index (depth) of the volume data for the region.
		@param midOff - A long integer which is the product of the X and Y grid indices for the
								volume data for the region.  Precomputed and passed to avoid repetition.
		@returns - An integer which is a binary encoded hash value representing which adajacent
								points are open.
	*/	
	int findPointFacing(int, int, int, long int);



	/*
	  Takes a surface point and its associated hash value and quantifies the local topology at that
		point.  The topology information is later determines which spherical segments will need to 
		be plotted if the volume is expanded and how the surface mesh should be recreated from the
		enumerated volume.  The original hash value is used as an entry argument into a lookup 
		table which will output one of the following new binary codes identifying the local
		topology.  The values are catagoried into three generic types:  (1) The surface opens from 
		the referenced point only in the direction normal to the surface (point), (2) The surface
		opens in two orthagonal directions, effectively defining a 90 degree ledge (arc), or
		(3)  The surface opens in three orthagonal directions, effectively defining a point on
		the tip of a corner (quad).

	  pointHash encoding (plotting new or expanded surface requies only a single point):
	    1  = Plot only point directly east.
	    2  = Plot only point directly west.
	    4  = Plot only point directly north.
	    8  = Plot only point directly south.
	    16 = Plot only point directly up.
	    32 = Plot only point directly down.

	  arcHash encoding (plotting new or expanded surface requies an arc segment):
	    1    = Plot arc of points in NE corner of plane.
	    2    = Plot arc of points in NW corner of plane.
	    4    = Plot arc of points in SE corner of plane.
	    8    = Plot arc of points in SW corner of plane.
	    16   = Plot arc of points in UE corner of plane.
	    32   = Plot arc of points in UW corner of plane.
	    64   = Plot arc of points in DE corner of plane.
	    128  = Plot arc of points in DW corner of plane.
	    256  = Plot arc of points in UN corner of plane.
	    512  = Plot arc of points in US corner of plane.
	    1024 = Plot arc of points in DN corner of plane.
	    2048 = Plot arc of points in DS corner of plane.

	  quadHash encoding (plotting new or expanded surface requies a spherical quadrant):
	    1    = Plot all the points in UNE quadrant of volume.
	    2    = Plot all the points in UNW quadrant of volume.
	    4    = Plot all the points in USE quadrant of volume.
	    8    = Plot all the points in USW quadrant of volume.
	    16   = Plot all the points in DNE quadrant of volume.
	    32   = Plot all the points in DNW quadrant of volume.
	    64   = Plot all the points in DSE quadrant of volume.
	    128  = Plot all the points in DSW quadrant of volume.

		The original hash value is translated into one of the possibilities above and then the
			data associated with the point is recorded into a list of like points along with the
			new translated hash value.
		See also the documetation for the method findPointFacing() which develops the original
			hash code used as the entry argument.

		@param vecIndex - A long integer which is the vector offset equivalent to grid[i][j][k].
		@param lnIndex - An integer which is the X grid index (longitude) of the volume data for the region.
		@param ltIndex - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param depIndex - An integer which is the Z grid index (depth) of the volume data for the region.
		@param hash - An integer which is the original hash value descibing the local topology.
	*/
	void addSurfacePointToList(long int, int, int, int, int);


	/*
		The recursive portion of the flood fill algorithm.  If point tested is zero (an empty
			exterior point), it is set to two and then the six adjacent points are tested.  When
			complete, all exterior contiguous points will have a value of two.

		@param lon - An integer which is the X grid index (longitude) of the volume data for the region.
		@param lat - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param depth - An integer which is the Z grid index (depth) of the volume data for the region.
		@param depthOffset - A long integer which is the product of the X and Y grid indices for the
								volume data for the region.  Precomputed and passed to avoid repetition.
	*/
	void markExteriorPoint(int, int, int, int);


	/*
		Data for the enumerated volumes is stored in a vector rather than a three dimensional
		  array.  This method compresses the repeated need to set data points in a volume.  It
		  calculates the vector offsets corresponding to the three dimensional input (longitude,
		  latitude, and depth) and sets the data value for that point to one.

  		@param lonIdx   - A long integer of the enumerated volume's equivalent 3D array grid index
								  in the X (longitude) direction.
  		@param latIdx   - A long integer of the enumerated volume's equivalent 3D array grid index
								  in the Y (latitude) direction.
  		@param depthIdx - A long integer of the enumerated volume's equivalent 3D array grid index
								  in the Z (depth) direction.
  		@param lonDim   - An integer of the enumerated volume's equivalent 3D array grid size
								  in the X (longitude) direction.
		@param region  - The CContainmentRegion object which is to have a point in its data set
									updated.
	*/
void addDataPointToRegion(long int, long int, long int, int, CContainmentRegion*);

};	//end class CContainmentRegion specifications
#endif

