// sharedMemObj.hpp		A Shared Memory Object base class definition

#ifndef _SHAR_MEM_H_
#define _SHAR_MEM_H_

// 2002, Fernando Das-Neves, Virginia Tech UVAG Group //

#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>


/*! By inheriting fromg this class, an otherwise ordinary object is located in shared memory instead of the heap.
 The purpose of this class is to allow communication between two processes without having to handle raw arrays or structs.
 One process, the "master", can treat the object in a normal way
 all the other processes (the "clients") must declare the object as a pointer,
 and instantiate the object with the special version of new that requires the
 shared memory block ID as a parameter. */
class ObjectInSharedMemory {
	protected:
	  /*! Contains the number of references to this object.
	     IMPORTANT!! DO NOT INITIALIZE ref_count in default constructor! */
	  int ref_count;
	  ObjectInSharedMemory() {}
	public:
	  /// ID of shared memory block where the object is located
	  int shmid;
          typedef enum { CREATE_WITH_KEY } CREATE_SEGMENT;
	  /*! This version of new is used by the master (whoever creates the object for the first time), when
              we want to specify a key id for the shared mem block, and then we can use the same ID to create the object in the client.
	      This operator will allocate a shared memory block. */
	  void *operator new(size_t blocksize, const CREATE_SEGMENT, key_t segment_key = IPC_PRIVATE) throw(); // used by the master
	  /*! This version of new is used by the master (whoever creates the object for the first time), when
              we don't care about the key ID to use for the shared mem block. Once the memory block is allocated, the segment id
              is stored in shmid. The client in a separate process will have to get the
	      key from somewhere else.
	      This operator will allocate a shared memory block. */
	  void *operator new(size_t blocksize) throw();
	  /*! This is the version of new used by the clients. It really doesn't allocate more memory, only gets access to
	     the block of memory allocated by the master, and the object contained within. To do this it needs the
	     key ID of the existing shared memory block */
	  void *operator new(size_t,key_t existing_segment_shmkey) throw();
	  /// delete() will decrease ref_count, and if ref_count reaches 0, it will free the shared mem block
	  void operator delete(void *);
};


#endif
