#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/SbLinear.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoOutput.h>
#include <Inventor/SoDB.h>
#include <stdlib.h>

#define DEG_TO_RAD(d) (d*6.2831853f/360.0f)

const char* ERROR_ROTATION = "Invalid rotation angle";
const char* ERROR_SCALING = "Invalid scaling value";
const char* ERROR_TRANSLATION = "Invalid translation value";

void printUsage() {
  printf("USAGE:\nivtransform [OPTIONS] input_file -o output_name\n");
  printf("OPTIONS:	\n");
  printf("\t-c\tCenters object\n");
  printf("\t-r rotX rotY rotZ\tRotates object by (rotX,rotY,rotZ) degrees\n");
  printf("\t-s scaleX scaleY scaleZ\tScales object by (scaleX,scaleY,scaleZ) units\n");
  printf("\t-t transX transY transZ\tMoves object by (transX,transY,transZ) units\n");
  printf("\t-m\tPrints transformation matrix that will be applied to the object\n");
  printf("\t-b\tSaves object in Inventor's binary file format, instead of ASCII.\n");
}

void exitError(const char* errmsg) {
  printf("ERROR: %s.\n",errmsg);
  printUsage();
  exit(-1);
}


SoCallbackAction::Response transformFn(void* matrix, SoCallbackAction*, const SoNode* coords) {
  SbVec3f dest, *points = ((SoCoordinate3*)coords)->point.startEditing();
  int numpts = ((SoCoordinate3*)coords)->point.getNum();
  for(int i=0; i<numpts; i++) {
    ((SbMatrix*)matrix)->multVecMatrix(*points,dest);
    *points=dest;
    points++;
  }
  ((SoCoordinate3*)coords)->point.finishEditing();
  return SoCallbackAction::CONTINUE;
}

int main(int argc, char* argv[]) {
        SoDB::init();
        SbBool saveAsBinary=FALSE;
        bool centerObject = false,
             printTransformationMatrix = false;
        int currparam = 1;
        char* lastdigit;
	SbMatrix result;
        SbMatrix m;
        SbRotation rotation;

	SoCallbackAction applyTransform;

        float valx,valy,valz;

        char *iv_infilename = NULL, *iv_outfilename = NULL;

        SoSeparator *objectRoot;

        result.makeIdentity();

        while(currparam < argc) {
        	if(strcasecmp(argv[currparam],"--help")==0 || strcasecmp(argv[currparam],"-h")==0) {
			printf("ivtransform reads an inventor file,applies transformations and saves the result.\n");
                        printUsage();
                        exit(-1);
	        } else if(strcasecmp(argv[currparam],"-b")==0) {
                  saveAsBinary=TRUE;
                } else if(strcasecmp(argv[currparam],"-m")==0) {
                  printTransformationMatrix = true;
                } else if(strcasecmp(argv[currparam],"-r")==0) {
                	if(currparam+3 <argc) {
				++currparam;
                		valx = strtod(argv[currparam],&lastdigit);
                                if(valx==0 && argv[currparam]==lastdigit) {
                                  exitError(ERROR_ROTATION);
                                }
                                if(valx!=0) {
                                	rotation.setValue(SbVec3f(1,0,0),DEG_TO_RAD(valx));
                                	m.setRotate(rotation);
                                	result.multLeft(m);
                                }
                                ++currparam;
                		valy = strtod(argv[currparam],&lastdigit);
                                if(valy==0 && argv[currparam]==lastdigit) {
                                  exitError(ERROR_ROTATION);
                                }
                                if(valy!=0) {
                                	rotation.setValue(SbVec3f(0,1,0),DEG_TO_RAD(valy));
                                	m.setRotate(rotation);
                                	result.multLeft(m);
                                }
                                ++currparam;
                                valz = strtod(argv[currparam],&lastdigit);
                                if(valz==0 && argv[currparam]==lastdigit) {
                                  exitError(ERROR_ROTATION);
                                }
                                if(valz!=0) {
                                	rotation.setValue(SbVec3f(0,0,1),DEG_TO_RAD(valz));
                                	m.setRotate(rotation);
                                	result.multLeft(m);
                                }
			} else {
                           exitError("Not enough values for rotation (there must be 3)");
                        }
        	} else if(strcasecmp(argv[currparam],"-s")==0) {
                	if(currparam+3 < argc) {
                 		++currparam;
				valx = strtod(argv[currparam],&lastdigit);
                                if(valx==0 && argv[currparam]==lastdigit) {
                                  exitError(ERROR_SCALING);
                                }
                                ++currparam;
                                valy = strtod(argv[currparam],&lastdigit);
                                if(valy==0 && argv[currparam]==lastdigit) {
                                  exitError(ERROR_SCALING);
                                }
                                ++currparam;
                                valz = strtod(argv[currparam],&lastdigit);
                                if(valz==0 && argv[currparam]==lastdigit) {
                                  exitError(ERROR_SCALING);
                                }
				m.setScale(SbVec3f(valx,valy,valz));
                                result.multLeft(m);
                        } else {
                           exitError("Not enough values for scaling (there must be 3)");
                        }
        	} else if(strcasecmp(argv[currparam],"-t")==0) {
                	if(currparam+3 < argc) {
                 		++currparam;
				valx = strtod(argv[currparam],&lastdigit);
                                if(valx==0 && argv[currparam]==lastdigit) {
                                  exitError(ERROR_TRANSLATION);
                                }
                                ++currparam;
				valy = strtod(argv[currparam],&lastdigit);
                                if(valx==0 && argv[currparam]==lastdigit) {
                                  exitError(ERROR_TRANSLATION);
                                }
                                ++currparam;
                                valz = strtod(argv[currparam],&lastdigit);
                                if(valx==0 && argv[currparam]==lastdigit) {
                                  exitError(ERROR_TRANSLATION);
                                }
                                m.setTranslate(SbVec3f(valx,valy,valz));
                                result.multLeft(m);
                        } else {
                           exitError("Not enough values for translation (there must be 3)");
                        }
    		} else if(strcasecmp(argv[currparam],"-c")==0) {
      		  centerObject = true;
                } else if(strcasecmp(argv[currparam],"-o")==0) {
                   iv_outfilename = argv[++currparam];
                }else {
                   iv_infilename = argv[currparam];
                }
                ++currparam;
    	}

        if(iv_infilename == NULL) {
          exitError("Filename of file to transform missing");
        }

        if(iv_outfilename == NULL) {
          exitError("Filename to save result missing");
        }

        /* read inventor object to transform */
        SoInput infile;
        if(infile.openFile(iv_infilename) == FALSE) {
          exitError("Can't open input inventor file");
        }

        printf("Reading %s...\n",iv_infilename);
        objectRoot = SoDB::readAll(&infile);
        infile.closeFile();
        if(objectRoot == NULL) {
          exitError("Can't read input inventor file. Probably it is not an valid Inventor file.");
        }

        objectRoot->ref();
        printf("Applying transformations...\n");
        
        if(printTransformationMatrix) {
        	printf("Transformation Matrix:\n");
        	result.print(stdout);
        	printf("\n");
        }
	applyTransform.addPostCallback(SoCoordinate3::getClassTypeId(),transformFn,&result);
	applyTransform.addPostCallback(SoNormal::getClassTypeId(),transformFn,&result);
        applyTransform.apply(objectRoot);

        if(centerObject) {
                printf("Centering Object...\n");
                SoGetBoundingBoxAction bboxAction(SbViewportRegion(1,1));
		bboxAction.apply(objectRoot);
        	SbBox3f bbox = bboxAction.getBoundingBox();
                result.setTranslate(SbVec3f(-(bbox.getMin()[0]+bbox.getMax()[0])/2.0f,
                                -(bbox.getMin()[1]+bbox.getMax()[1])/2.0f,
                                -(bbox.getMin()[2]+bbox.getMax()[2])/2.0f));
		applyTransform.apply(objectRoot);
        }
        printf("Saving result as %s...\n",iv_outfilename);
        SoWriteAction writer;
        writer.getOutput()->openFile(iv_outfilename);
        writer.getOutput()->setBinary(saveAsBinary);
        writer.apply(objectRoot);
     	writer.getOutput()->closeFile();
        printf("Done.\n");
}
