#ifndef TRACKVIEW_H
#define TRACKVIEW_H

#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoMultipleCopy.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoTransformSeparator.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoScale.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoLineSet.h>
#include <Inventor/actions/SoWriteAction.h>
#include <list>
#include "track.h"
#include "NameToTrackMapping.h"
#include "IvSwitchEngine.h"
#include "IvColorizeGrayscalesAction.h"

#include "ASU/ContainmentRegion.h"


// 2002, Virginia Tech UVAG Group
// 2002, NUWC Group

	
   /*! This is an Inventor object that can take a 3D gray object and color it
       , or change the color of an already-colored object and change 
       its color depending on a threat class. */ 
   class Iv3dIconColorizer: public IvColorizeGrayscalesAction {
   public:
     // These are the colors used for bounding boxes, traces, and any object associated to a track *
     /// color of friendly contacts in HSV. Use FRIENDLY_RGBCOLOR for the same color in RGB
     static const float FRIENDLY_HSVCOLOR[3];
     /// color of enemy contacts in HSV. Use HOSTILE_RGBCOLOR for the same color in RGB
     static const float HOSTILE_HSVCOLOR[3];
     /// color of unclassified contacts in HSV. Use UNKNOWN_THREAT_RGBCOLOR for the same color in RGB
     static const float UNKNOWN_THREAT_HSVCOLOR[3];
     /// color for the ownship      
     static const float OWNSHIP_RGBCOLOR[3];
     ///color of friendly contacts in RGB. Use FRIENDLY_HSVCOLOR for the same color in HSV
     static const float FRIENDLY_RGBCOLOR[3];
     /// color of enemy contacts in RGB. Use HOSTILE_HSVCOLOR for the same color in HSV
     static const float HOSTILE_RGBCOLOR[3];
     /// color of unclassified contacts in HSV. Use UNKNOWN_THREAT_HSVCOLOR for the same color in HSV
     static const float UNKNOWN_THREAT_RGBCOLOR[3];
     /// color for fired weapons (torpedoes, missiles, etc)
     static const float WEAPON_RGBCOLOR[3];
      void colorize3dIconAs(SoNode* node, ThreatClass threat);
      void change3dIconColorTo(SoNode* node, ThreatClass threat);
      const float* getRGBColorForThreatClass(ThreatClass threat) const;
   };


/*! An IvTrackView is a view (visual representation) of a TrackTraceDecorator.
  An IvTrackView has an Inventor object that is used as marker, and it will appear at the position indicated every
  time IvTrackView::update() is called.
  Since it is a subclass of Observer, it can listen for event notifications from a TrackTraceDecorator.
*/
class IvTrackView: public SoSwitch, public Observer  {
  /// holds the model track which  (IvTrackView is the track's view)
  TrackTraceDecorator *model;
protected:
  /// position of marker at the 'head' of the track trail
  SbVec3f posOfLastMarker;
  /// holds the minimum separation between 2 markers
  double dist_step;
  /// This variable holds the scaling factor to apply to the marker positions, but not to the markers (the Inventor objects)
  float scaleFactor[3];
  /// Inventor object used as a marker. It must lay flat on the XY plane, pointing at +Y
  SoNode *marker;
  /// pointer to the object that holds the point coordinates to draw the trace
  SoMFVec3f *tracePoints;
  /// pointer to the object that holds the color for the trace line
  SoMFUInt32 *traceColor;
  /// Inventor object that holds the transform to position the marker in multiple places
  SoMultipleCopy *markerPositions;
  /// Inventor object that draws a trace joining the past positions of the contact
  SoLineSet * trace;
  /// This is the inventor action used to set the color of the markers according to the threal level (Friendly, Enemy, Unknown)
  Iv3dIconColorizer *colorizer;
  virtual ~IvTrackView();
  /*! Add a marker at the position indicated by pos. This happens automatically
  when update() is called */
  virtual void addMark(const Coord& pos);
  void removeLastMark();
  void addLineSegment(Coord* pos);
  void removeLastLineSegment();
public:
  IvTrackView();
  /// Set which object to use as a marker
  void setMarker(SoNode* aMarker);
  /*! aColorizer will be used to change the color of the track marker when the threat class changes.
      This method must be called BEFORE setMarker() for the color to be set an updated correctly. */
  void setTraceColorUsing(Iv3dIconColorizer *aColorizer);
  void setTraceColor(float r, float g, float b);
  /// Tell the IvTrackView which TrackTraceDecorator to watch
  virtual void setModel(Observable* aTrack);
  /// returns a TrackTraceDecorator
  ObservableTrack* getModel() const { return model; }
  /// Set the minimum distance that must separate the center of two consecutive markers in a track
  void setMinMarkerDistanceSeparation(double distance) { dist_step = distance; }
  double getMinMarkerDistanceSeparation() const { return dist_step; }
  /*! update() is called by the TrackTraceDecorator every time something happens in the Track.
    The info parameter in this case can be casted to Coord* */
  virtual void update(void *info);
  void scaleBy(float newScaleFactorX, float newScaleFactorY, float newScaleFactorZ);
  void hide() {whichChild=SO_SWITCH_NONE;}
  void show() {whichChild=SO_SWITCH_ALL;}
};

/*! This is the base class for all classes that handle Inventor objects to show a contact's position,
    or its Area of Uncertainty (AOU).
    Important: IvContactView is designed to be a Observer of an instance of DiscreteTrack .
*/
class IvContactView: public SoSwitch, public Observer {
  SO_NODE_ABSTRACT_HEADER(IvContactView);
protected:
  DiscreteTrack* model;
  IvSwitchEngine *which;
  SoSFInt32 invisible_count;
  SoSFBool outOfRange;
  SoSFBool userVisible;
  Iv3dIconColorizer *colorizer;
  bool grow;
  bool selfIntersect;
  /// returns a integer that is the bitwise OR of all event types the view is interested. See TrackEvent and DiscreteTrackEvent for a list of possible event types.
  virtual int getEventsOfInterest() const;
public:
  //field
  //if a program does not know how to render a node, it looks at the alternate rep field
  SoSFNode alternateRep;
  /// Initializes this class.
  static void initClass();
  ///boolean, set show or no-show
  static void setShowIntersecting(bool show);
  ///toggle intersections on/off
  static void toggleShowIntersecting() {setShowIntersecting(!((SoSFBool*)SoDB::getGlobalField("showIntersecting"))->getValue());}
  IvContactView();
  ///set track to toggle and observe
  virtual void setModel(Observable* aTrack) { setModel(aTrack,0); }
  virtual void setModel(Observable* aTrack, unsigned int userBitMask);
  ObservableTrack* getModel() const { return model; }
  /// Must return an Inventor field with the position of the contact's visual representation
  virtual SoSFVec3f& getPosition() const =0;
  /*! handles the hiding/showing of the contact icon by turning the inherited SoSwitch on/off, and nothing else.
    If a subclass redefines this method and doesn't call  IvContactView::update(),
    then it must handle visibility in the redefined method through the whichChild field of SoSwitch. */
  virtual void update(void *evt);
  /*gets the 2D convex hull of the current representation of the contact as well as its depth bounds*/
  virtual CContainmentRegion *getContainmentRegion() {return NULL;};
  virtual CContainmentRegion *makeContainmentRegion() {return NULL;};
  ///increases/decreases the number of objects telling this object to be invisible
  virtual void setVisible(bool vis);
  /// returns true if the inventor object is visible (drawn), or false if not.
  virtual bool isVisible() const { return (whichChild.getValue()!=SO_SWITCH_NONE); };
  ///makes the object invisible
  virtual void hide() {userVisible.setValue(false);}
  ///makes the object visible
  virtual void show() {userVisible.setValue(true);}
  ///makes the object grow or not grow
  virtual void setGrow(bool do_grow) {grow=do_grow;}
  virtual bool getGrow() {return grow;}
  virtual void setSelfIntersect(bool do_self) {selfIntersect=do_self;}
  virtual bool getSelfIntersect() {return selfIntersect;}
  //true if contact is within its time range ie events in past and future
  virtual bool isAvailable() const { return !outOfRange.getValue(); };
  /*! aColorizer will be used to change the color of the contact marker when the threat class changes.
      This method must be called BEFORE setMarker() for the marker color to be set an updated correctly. */
  virtual void setMarkerColorUsing(Iv3dIconColorizer* aColorizer) { colorizer = aColorizer; }

SoEXTENDER public:
    //overwritten functions
    //overwrite the write function to update alternate rep before the node is written out to file
    virtual void write(SoWriteAction *action);
    virtual void scaleBy(float newScaleFactorX, float newScaleFactorY, float newScaleFactorZ) =0;
};


/*! This is the object used to handle the visual representation of any object with a concrete, point-like position.
   It can use any Inventor object (marker) to show the current position of a contact.
   Since it is a subclass of Observer, it can listen for event notifications from a DiscreteTrack.
   The model can be any DiscreteTrack or its subclasses, as long as the steps are instances of class PreciseContact. */
class IvPreciseContactView: public IvContactView {
protected:
  SoNode *marker;
  ///soRotationXYZ holds the rotation variable to apply to the marker, but only in the X and Y planes.  Z not supported.
  SoRotationXYZ *rotationX, *rotationY;
  ///SoTranslation holds the translation position sum to apply to the marker
  SoTranslation* pos;
  /// scaleFactor holds the scaling factor to apply to the marker positions, but not to the markers (the Inventor objects)
  float scaleFactor[3];
public:
  IvPreciseContactView();
  ///get2DBounds returns the bounding box for the marker in 2D, extruded.
  virtual CContainmentRegion *getContainmentRegion();
  ///setMarker creates a new marker.
  virtual void setMarker(SoNode* aMarker);
  /*!update(void *info) handles the hiding/showing of the contact icon by turning the inherited SoSwitch on/off, and nothing else.
    If a subclass redefines this method and doesn't call  IvContactView::update(),
    then it must handle visibility in the redefined method through the whichChild field of SoSwitch. */
  virtual void update(void *info);
  ///getPosition() Returns the marker's position
  virtual SoSFVec3f& getPosition() const { return pos->translation; }
  ///scaleBy applies scale factors to the precise AOU region
  virtual void scaleBy(float newScaleFactorX, float newScaleFactorY, float newScaleFactorZ);
};


/*! An Inventor node to show a Torpedo */
class IvWPNView: public IvContactView {
protected:
  SoNode *marker;
  ///soRotationXYZ holds the rotation variable to apply to the marker, but only in the X and Y planes.  Z not supported.
  SoRotationXYZ *rotationX, *rotationY;
  ///SoTranslation holds the translation position sum to apply to the marker
  SoTranslation* pos;
  /// scaleFactor holds the scaling factor to apply to the marker positions, but not to the markers (the Inventor objects)
  float scaleFactor[3];
  SoCoordinate3 *covPts;
  SoMaterial *covMaterial;
  SoTransformSeparator *sep;
  SoSeparator *fence;
  int fenceLength, fenceHeight, fenceWidth;
public:
  IvWPNView();
  ///get2DBounds returns the bounding box for the marker in 2D, extruded.
  virtual CContainmentRegion *getContainmentRegion();

  void setColor(float r, float g, float b);
 // void setFence(IvContactView &*f){fence = f;};
 // void showFence(){fence->show();};

  ///setMarker creates a new marker.
  virtual void setMarker(SoNode* aMarker);
  /*!update(void *info) handles the hiding/showing of the contact icon by turning the inherited SoSwitch on/off, and nothing else.
    If a subclass redefines this method and doesn't call  IvContactView::update(),
    then it must handle visibility in the redefined method through the whichChild field of SoSwitch. */
  virtual void update(void *info);
  ///getPosition() Returns the marker's position
  virtual SoSFVec3f& getPosition() const { return pos->translation; }
  ///scaleBy applies scale factors to the precise AOU region
  virtual void scaleBy(float newScaleFactorX, float newScaleFactorY, float newScaleFactorZ);

  void rangebear2xy(float range, float bearing, float& x, float& y);
  void createFence();
};


/*! Objects of this class are used to represent contact for which we do not know a precise position,
    but we know an volume, called Area of Uncertainty (AOU), where the contact is located within.
    IvAOUView assumes the model is a DiscreteTrack whose steps are instances of ContactFromSensor. */
class IvAOUView: public IvContactView {
protected:
  ///markerPOS is the inventor object showing the area of uncertainty
  static SoSeparator *markerPOS;
  static SoSeparator *markerLOB;
  SoSwitch* poslobSwitch;
  SoTranslation* pos;
  CContainmentRegion *region;
  /// This variable holds the scaling factor to apply to the marker positions, but not to the markers (the Inventor objects)
  float scaleFactor[3];


  //POS stuff
  SoMaterial * posMaterial;
  SoRotationXYZ *rotationXY;
  SoScale* scale;
  ContactFromSensor *curPos;

  //LOB stuff
  SoMaterial * lobMaterial;
  SoCoordinate3 *points;
  ContactFromSensor *curLob;

  //region stuff
  SoMaterial * regionMaterial;
  SoCoordinate3 *regionPoints;
  SoIndexedFaceSet *regionFaces;

  float depth_min;
  float depth_max;
public:
  IvAOUView(float depth_max, float depth_min=0);
  virtual void setMarkerColorUsing(Iv3dIconColorizer* aColorizer);
  ///return 2D bounds of AOU contact
  virtual CContainmentRegion *getContainmentRegion();
  virtual CContainmentRegion *makeContainmentRegion();
  void setColor(float r, float g, float b);
  void addMeshFromRegion();
  ///update AOU contact info
  virtual void update(void *info);
  ///return position of AOU
  virtual SoSFVec3f& getPosition() const { return pos->translation; }
  ///scaleBy applies scale factors to the AOU region
  virtual void scaleBy(float newScaleFactorX, float newScaleFactorY, float newScaleFactorZ);
};


/*! Objects of this class are used to represent merged contacts
    IvAOUView assumes the model is a DiscreteTrack whose steps are instances of HypotheticalContact. */
class IvHypotheticalView: public IvContactView {
protected:

  static SoMaterial *material;
  SoTranslation *pos;
  CContainmentRegion *region;
 /// This variable holds the scaling factor to apply to the marker positions, but not to the markers (the Inventor objects)
  float scaleFactor[3];
  /// This variable holds the coordinates of the contact
  SoCoordinate3 *points;
  SoIndexedFaceSet *faces;
  HypotheticalContact *cur;
  typedef list<void *> AssocList;
  AssocList assoc;
  float depth_min;
  float depth_max;
  NameToTrackMapping *map;
  void removeAssoc();
  void makeAssoc(char *name);
  void makeIntersection();

public:
  IvHypotheticalView(NameToTrackMapping *mapping);
  ///get2DBounds returns the bounding box for the marker in 2D, extruded.
  virtual CContainmentRegion *getContainmentRegion();
  virtual CContainmentRegion *makeContainmentRegion();
  /*!update(void *info) handles the hiding/showing of the contact icon by turning the inherited SoSwitch on/off, and nothing else.
    If a subclass redefines this method and doesn't call  IvContactView::update(),
    then it must handle visibility in the redefined method through the whichChild field of SoSwitch. */
  virtual void update(void *info);
  ///getPosition() Returns the marker's position
  virtual SoSFVec3f& getPosition() const { return pos->translation; }
  ///scaleBy applies scale factors to the AOU region
  virtual void scaleBy(float newScaleFactorX, float newScaleFactorY, float newScaleFactorZ);
  void addMeshFromRegion();
};


#endif

