//EnumeratedVolume.cpp - Body of associated class.

#include "EnumeratedVolume.h"
#include <math.h>   

/* =======================================================================================
	The EnumeratedVolume is a data construction used to represent a real world piece of ocean
		as a three dimensional grid of data points.  This is a base class to provide the basic
		information needed to define where the volume resides in the real world and how that
		same volume is translated into a numeric grid, the data space for computational
		manipulation.


	This software was developed as part of a grant by ONR 311, Office of Naval Research, 
	800 North Quincy Street, Arlington, Virginia 22217, for development of an "Application 
	of Volume Modeling Techniques to Dynamic Containment Regions for Naval Applications".
	Grant was awarded to Arizona State University, PRISM Laboratory, Department of Computer 
	Science and Engineering, Dr. Greg Nielson and Gary Graf.


	Author: Gary Graf, PRISM Laboratory, Arizona State University.
	Version:  1.0, 10 Sep 2002

  Copyright 2002, all rights reserved.
  
	Notes:
	1) See warning in method computeVolumeDimensions().
=========================================================================================*/

		//A constant to multiply times degrees to get radians.
	float const static DEG_TO_RAD = 0.017453292f;
		//A constant to convert yards to nautical miles.  Note that one nautical mile (nm) is
		//	 also equal to one degree of latitude.
	float const static YARDS_PER_NM = 2000.0f;
		//A constant to convert minutes of latitude/longitude to degrees.
	float const static MIN_PER_DEGREE = 60.0f;

		//The default max depth in yards
	float const static MAX_DEPTH_YDS = 1000.0f;


/**************************************************************************************
---------------------------------------------------------------------------------------
		Constructors / Destructors
---------------------------------------------------------------------------------------
**************************************************************************************/

	/*
		The default constructor establishes an enumerated volume of approximately, 200 kyds x
		200 kyds, 500 yd spacing, origin at 25 deg N, 25 deg E, with no ID number.
	*/
CEnumeratedVolume::CEnumeratedVolume(){
		//Set class parameters
	volumeID = 0;
	timeReference = 0;

		//Corresponds to an extent of roughly 200 kyd x 200 kyd.
	maxExtent.minLat = 25.0;
	maxExtent.minLon = 25.0;
	maxExtent.maxLat = 25.83f;
	maxExtent.maxLon = 25.919f;
	
	spanLat = 0.83f;
	spanLon = 0.4595f;
	centerLat = 25.415f;

	allowedRangeError = 500.0;

	displayVol = false;

	computeVolumeDimensions();

	vectorSize = volumeDimension[0] * volumeDimension[1] * volumeDimension[2];
   volumeData.reserve(vectorSize);
	zeroVolumeData();
	
	regnColor[0] = 0.0f;
	regnColor[1] = 0.5f;
	regnColor[2] = 0.5f;
	transparancy = 1.0f;
}	//end constructor CEnumeratedVolume()


	/*
		This constructor will take a min latitude and longitude along with the required grid
		dimensions to compute the final extent of the region.  It is primarily used in the
		construction of resultant containment regions following the expansion of another region.  

		@param newID - An integer which is a unique identifier for the containment region.
		@param refTime - A long integer which represents the time when the given dimensions of the
				volume data for the region are valid.
		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param latDim - An integer which specifies how many grid cell the region should occupy
				in latitude.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param lonDim - An integer which specifies how many grid cell the region should occupy
				in longitude.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
		@param colors  - A pointer to a three element float array specifying the color to be used 
				in rendering this particular containment region.  The nature of the color space 
				used and the spread of values are not specified.  The information contained in 
				in the colors[] array are only stored to be later passed to the rendering engine.
		@param trans  - A float specifying the transparancy (or opacity) of the containment region
				when rendered.  The nature of the transparancy and the spread of values are not 
				specified.  The information contained is only stored to be later passed to the 
				rendering engine.
	*/
CEnumeratedVolume::CEnumeratedVolume(int newID, long int refTime, float mnLat, int latDim,
												 float mnLon, int lonDim, float rngErr,
												 float *colors, float trans){
		//From passed grid size, compute the correct latitude and longitude
	float latDiff = (latDim - 1) * rngErr / MIN_PER_DEGREE / YARDS_PER_NM;
	float mxLat = mnLat + latDiff;
	float lonDiff = ((lonDim - 1) / cos((mnLat + latDiff / 2.0) * DEG_TO_RAD)) 
		               * rngErr / MIN_PER_DEGREE / YARDS_PER_NM;
	float mxLon = mnLon + lonDiff;

		//Set class parameters
	volumeID = newID;
	timeReference = refTime;

	maxExtent.minLat = mnLat;
	maxExtent.maxLat = mxLat;
	maxExtent.minLon = mnLon;
	maxExtent.maxLon = mxLon;

	allowedRangeError = rngErr;
	spanLat = maxExtent.maxLat - maxExtent.minLat;
	spanLon = maxExtent.maxLon - maxExtent.minLon;
	centerLat = (spanLat / 2.0) + maxExtent.minLat;

	displayVol = false;

	volumeDimension[0] = lonDim;
	volumeDimension[1] = latDim;

	lonToGridScale = (volumeDimension[0] - 1) / spanLon; 
	latToGridScale = (volumeDimension[1] - 1) / spanLat;

		//Depth is always considered constant for all volumes.
	float depth = MAX_DEPTH_YDS/allowedRangeError + 0.5;
	volumeDimension[2] = (int)depth + 1;  //The +1 add a point for the boundary

	vectorSize = volumeDimension[0] * volumeDimension[1] * volumeDimension[2];
   volumeData.reserve(vectorSize);
	zeroVolumeData();

	regnColor[0] = colors[0];
	regnColor[1] = colors[1];
	regnColor[2] = colors[2];
	transparancy = trans;
}	//end constructor CEnumeratedVolume(int, long int, float, int, float, int,   
	//										      float, float*, float)



	/*
		This constructor will take a min latitude and longitude along with the required grid
		dimensions to compute the final extent of the region.  It is used in the
		construction of expansion regions. 

		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param latDim - An integer which specifies how many grid cell the region should occupy
				in latitude.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param lonDim - An integer which specifies how many grid cell the region should occupy
				in longitude.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
	*/
CEnumeratedVolume::CEnumeratedVolume(float mnLat, int latDim, 
												 float mnLon, int lonDim, float rngErr){
		//Set default class parameters (not used with expansion regions)
	volumeID = 0;
	timeReference = 0;

	regnColor[0] = 0.0f;
	regnColor[1] = 0.0f;
	regnColor[2] = 0.0f;
	transparancy = 1.0f;

		//From passed grid size, compute the correct latitude and longitude
	float latDiff = (latDim - 1) * rngErr / MIN_PER_DEGREE / YARDS_PER_NM;
	float mxLat = mnLat + latDiff;
	float lonDiff = ((lonDim - 1) / cos((mnLat + latDiff / 2.0) * DEG_TO_RAD)) 
		               * rngErr / MIN_PER_DEGREE / YARDS_PER_NM;
	float mxLon = mnLon + lonDiff;

		//Set corrected class parameters
	maxExtent.minLat = mnLat;
	maxExtent.maxLat = mxLat;
	maxExtent.minLon = mnLon;
	maxExtent.maxLon = mxLon;

	allowedRangeError = rngErr;
	spanLat = maxExtent.maxLat - maxExtent.minLat;
	spanLon = maxExtent.maxLon - maxExtent.minLon;
	centerLat = (spanLat / 2.0) + maxExtent.minLat;

	displayVol = false;

	volumeDimension[0] = lonDim;
	volumeDimension[1] = latDim;

	lonToGridScale = (volumeDimension[0] - 1) / spanLon; 
	latToGridScale = (volumeDimension[1] - 1) / spanLat;

		//Depth is always considered constant for all volumes.
	float depth = MAX_DEPTH_YDS/allowedRangeError + 0.5;
	volumeDimension[2] = (int)depth + 1;  //The +1 add a point for the boundary

	vectorSize = volumeDimension[0] * volumeDimension[1] * volumeDimension[2];
   volumeData.reserve(vectorSize);
	zeroVolumeData();
}	//end constructor CEnumeratedVolume(float, int, float, int, float)


	/*
		This is the fully specified constructor.

		@param newID - An integer which is a unique identifier for the containment region.
		@param refTime - A long integer which represents the time when the given dimensions of the
				volume data for the region are valid.
		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param mxLat  - A float which is the maximum latitude boundary of the region.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param mxLon  - A float which is the maximum longitude boundary of the region.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
		@param colors  - A pointer to a three element float array specifying the color to be used 
				in rendering this particular containment region.  The nature of the color space 
				used and the spread of values are not specified.  The information contained in 
				in the colors[] array are only stored to be later passed to the rendering engine.
		@param trans  - A float specifying the transparancy (or opacity) of the containment region
				when rendered.  The nature of the transparancy and the spread of values are not 
				specified.  The information contained is only stored to be later passed to the 
				rendering engine.
	*/
CEnumeratedVolume::CEnumeratedVolume(int newID, long int refTime, float mnLat, float mxLat,
												 float mnLon, float mxLon, float rngErr,
												 float *colors, float trans){
		//Set class parameters
	volumeID = newID;
	timeReference = refTime;


//TODO: Include exception handling if depth input not consistent vice the simplistic
//	approach used below.
		
	if(mnLat <= mxLat){
		maxExtent.minLat = mnLat;
		maxExtent.maxLat = mxLat;
	}
	else{
		maxExtent.minLat = mxLat;
		maxExtent.maxLat = mnLat;
	}

	if(mnLon <= mxLon){
		maxExtent.minLon = mnLon;
		maxExtent.maxLon = mxLon;
	}
	else{
		maxExtent.minLon = mxLon;
		maxExtent.maxLon = mnLon;
	}

	allowedRangeError = rngErr;
	spanLat = maxExtent.maxLat - maxExtent.minLat;
	spanLon = maxExtent.maxLon - maxExtent.minLon;
	centerLat = (spanLat / 2.0) + maxExtent.minLat;

	displayVol = false;

	computeVolumeDimensions();

	vectorSize = volumeDimension[0] * volumeDimension[1] * volumeDimension[2];
   volumeData.reserve(vectorSize);
	zeroVolumeData();

	regnColor[0] = colors[0];
	regnColor[1] = colors[1];
	regnColor[2] = colors[2];
	transparancy = trans;
}	//end constructor CEnumeratedVolume(int, long int, float, float, float, float, 
	//												float, float*, float)


	//Destructor
CEnumeratedVolume::~CEnumeratedVolume(){
	if(!volumeData.empty())
		volumeData.clear();
}	//end destructor ~CEnumeratedVolume()


/**************************************************************************************
---------------------------------------------------------------------------------------
		Public methods - General
---------------------------------------------------------------------------------------
**************************************************************************************/

	/*
		Method will zero all of the data for the enumerated volume but leave the data
			structure of the underlying vector intact.
	*/
void  CEnumeratedVolume::zeroVolumeData(){
	if(!volumeData.empty())
		volumeData.clear();
	for (long int x = 0; x < vectorSize; x++)
		volumeData.push_back(0);	
}	// zeroVolumeData()


	/*	
	  This method streamlines the repetitive calculation of the vector offset for a set of lon, 
		lat & depth grid coordinates.  Effectively translates the 3D array (grid[lon][lat][depth]) 
		into the equivalent linear vector position.

		@param ln - An integer which is the X grid index (longitude) of the volume data for the region.
		@param lt - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param dp - An integer which is the Z grid index (depth) of the volume data for the region.
		@param midOff - A long integer which is the product of the X and Y grid indices for the
								volume data for the region.  Precomputed and passed to avoid repetition.
		@returns - A long integer which is the vector offset equivalent to grid[i][j][k].
	*/	
long int CEnumeratedVolume::getOffset(int ln, int lt, int dp, long int midOff){
	return ln + lt * volumeDimension[0] + dp * midOff;
}	//end getOffset()


/**************************************************************************************
---------------------------------------------------------------------------------------
		Public methods - Accessors & Manipulators
---------------------------------------------------------------------------------------
**************************************************************************************/

	/*
		Simple accessor will allow access to the volume ID number.

		@retuns - An integer which is a unique identifier for the enumerated volume.
	*/
int  CEnumeratedVolume::getVolumeID(){
	int result = volumeID;
	return result;
}	// end getVolumeID()


	/*
		Simple manipulator will allow changing the volume ID number.

		@param ID - An integer which is a unique identifier for the enumerated volume.
	*/
void CEnumeratedVolume::setVolumeID(int ID){
		volumeID = ID;
}	//end setVolumeID()


	/*
		Simple accessor will allow access to the data in the FloatRect structure which holds data 
			about the size and real world coordinates of the enumerated volume. Setting values 
			is only allowed as the object is instantiated.

		@retuns - A FloatRect object which holds data about the size and real world coordinates
			of the enumerated volume.
	*/
CEnumeratedVolume::FloatRect CEnumeratedVolume::getExtent(){
	return maxExtent;
}	//end getExtent()


	/*
		Simple accessor will return the latitude of the midpoint of the region. Setting values is 
			only allowed when the object as instantiated.

		@retuns - A float which is the latitude of the midpoint of the region.
	*/
float CEnumeratedVolume::getCenterLat(){
	return centerLat;
}	//end getCenterLat()


	/*
		Simple accessor will return the difference between the maximum and mininum latitudes of 
			the region. Setting values is only allowed when the object as instantiated.

		@retuns - A float which is the difference between the maximum and mininum latitude.
	*/
float CEnumeratedVolume::getSpanLat(){
	return spanLat;
}	//end getSpanLat()


	/*
		Simple accessor will return the difference between the maximum and mininum longitudes of 
			the region. Setting values is only allowed when the object as instantiated.

		@retuns - A float which is the difference between the maximum and mininum longitude.
	*/
float CEnumeratedVolume::getSpanLon(){
	return spanLon;
}	//end getSpanLon()


	/*
		Simple accessor will allow access to the size of the underlying grid structure of
			the enumerated volume. Setting values is only allowed when the object as instantiated.

		@retuns - A pointer to a three element integer array, the compones of which hold the
				size of the grid in dimensions of [longitude][latitude][depth].
	*/
int* CEnumeratedVolume::getGridDimensions(){
	int* result = new int[3];

	result[0] = volumeDimension[0];
	result[1] = volumeDimension[1];
	result[2] = volumeDimension[2];

	return result;
}	//end getGridDimensions()


	/*
		Simple accessor will return the maximum allowable range error which this volume is using
			to detrimine grid spacing.  Setting values is only allowed when the object as instantiated.

		@retuns - A float which is the maximum allowable range error of the volume.
	*/
float CEnumeratedVolume::getAllowedRangeError(){
		return allowedRangeError;
}	//end getAllowedRangeError()


	/*
		Method is to allow quick conversion from a real world coordinate to the corresponding
			grid point along the latitude axis of the enumerated volume.  To improve speed 
			performance, error checking is not done.  If the passed latitude is outside the area 
			defined by the enumerated volume, the returned grid coordinates could be negative or
			a value greater than the actual dimensions of the grid structure.

		@param lat - A float representing a latitude.
		@retuns - An integer which is the latitude element of the grid index for the point
			which corresponds closest to the passed latitude.
	*/
int   CEnumeratedVolume::getGridFromLat(float lat){
	float temp = lat - maxExtent.minLat;
	int result = (int)(temp * latToGridScale + 0.5);

	return result;
}	//end getGridFromLat()


	/*
		Method is to allow quick conversion from grid point along the latitude axis of the 
			enumerated volume to the corresponding real world coordinate. 

		@param grid - An integer representing a grid index in the latitude dirction.
		@retuns - A float which is the corresponding latitude of the passed grid index.
	*/
float CEnumeratedVolume::getLatFromGrid(int grid){
	float result = (float)grid / latToGridScale + maxExtent.minLat;

	return result;
}	//end getLatFromGrid()


	/*
		Method is to allow quick conversion from a real world coordinate to the corresponding
			grid point along the longitude axis of the enumerated volume.  To improve speed 
			performance, error checking is not done.  If the passed longitude is outside the area 
			defined by the enumerated volume, the returned grid coordinates could be negative or
			a value greater than the actual dimensions of the grid structure.

		@param lon - A float representing a longitude.
		@retuns - An integer which is the longitude element of the grid index for the point
			which corresponds closest to the passed longitude.
	*/
int   CEnumeratedVolume::getGridFromLon(float lon){
	float temp = lon - maxExtent.minLon;
	int result = (int)(temp * lonToGridScale + 0.5);

	return result;
}	//end getGridFromLon()


	/*
		Method is to allow quick conversion from grid point along the longitude axis of the 
			enumerated volume to the corresponding real world coordinate. 

		@param grid - An integer representing a grid index in the longitude dirction.
		@retuns - A float which is the corresponding longitude of the passed grid index.
	*/
float CEnumeratedVolume::getLonFromGrid(int grid){
	float result = (float)grid / lonToGridScale + maxExtent.minLon;

	return result;
}	//end getLonFromGrid()


	/*
		Method is to allow quick conversion from a real world coordinate to the corresponding
			grid point along the depth axis of the enumerated volume.  To improve speed 
			performance, error checking is not done.  If the passed depth is outside the area 
			defined by the enumerated volume, the returned grid coordinates could be negative or
			a value greater than the actual dimensions of the grid structure.
		NOTE:  Depth values are in feet (not yards).

		@param depth - A float representing a depth.
		@retuns - An integer which is the depth element of the grid index for the point
			which corresponds closest to the passed depth.
	*/
int   CEnumeratedVolume::getGridFromDepth(float depth){
	float temp = depth / 3.0 / allowedRangeError;
	int result = (int)(temp + 0.5);

	return result;
}	//end getGridFromDepth()


	/*
		Method is to allow quick conversion from grid point along the depth axis of the 
			enumerated volume to the corresponding real world coordinate. 
		NOTE:  Depth values are in feet (not yards).

		@param grid - An integer representing a grid index in the depth dirction.
		@retuns - A float which is the corresponding depth of the passed grid index.
	*/
float CEnumeratedVolume::getDepthFromGrid(int grid){
	float result = (float)(grid) * allowedRangeError * 3.0;

	return result;
}	//end getDepthFromGrid()


	/*
		Simple manipulator to set the time reference related to the volume.

		@param time - A long integer which is the time reference related to the volume.
	*/
void CEnumeratedVolume::setReferenceTime(long int time){
	timeReference = time;
}	//end setReferenceTime(


	/*
		Simple accessor will return the time reference related to the volume.

		@retuns - A long integer which is the time reference related to the volume.
	*/
long int CEnumeratedVolume::getReferenceTime(){
	return timeReference;
}	//end getReferenceTime()


	/*
		Simple manipulator to set the color that the volume is to be rendered with.  The
			passed parameter is assumed to be a three element array, but no specific limits
			are placed on the color space to be used or on the range of values passed.

		@param newColor - A pointer to float array which is the color that the volume is to be 
				rendered with.
	*/
void CEnumeratedVolume::setColor(float* newColor){
}	//end setColor()


	/*
		Simple accessor will return the color that the volume is to be rendered with.  The
			value is assumed to be a three element array, but no specific limits
			are placed on the color space to be used or on the range of values returned.

		@retuns - A pointer to float array which is the color that the volume is to be 
				rendered with.
	*/
float* CEnumeratedVolume::getColor(){
		float *color = regnColor;
		return color;
}	//end getColor()


	/*
		Simple manipulator to set the transparancy (or opacity) that the volume is to be rendered 
			with.  There are no specific limits placed on the range of values passed.

		@param trans - A float which indicates the degree of transparancy (or opacity) that the 
				volume is to be rendered with.
	*/
void CEnumeratedVolume::setTransparancy(float trans){
	transparancy = trans;
}	//end	setTransparancy()


	/*
		Simple accessor will return the transparancy (or opacity) that the volume is to be rendered 
			with.  There are no specific limits placed on the range of values returned.

		@retuns - A float which indicates the degree of transparancy (or opacity) that the 
				volume is to be rendered with.
	*/
float CEnumeratedVolume::getTransparancy(){
	return transparancy;
}	//end getTransparancy()



/**************************************************************************************
---------------------------------------------------------------------------------------
		Protected methods
---------------------------------------------------------------------------------------
**************************************************************************************/


	/*
	The method will take the dimensions of the volume, as determined by the maximum extent of
		the latitude and longitude, and convert the real world coordinate space to a enumerated 
		grid with the minimum latitude and longitude values at zero in the grid.
	WARNING:  The algorithm will fail if the latitude inputs exceed 90 degres.  For
	  use in polar regions, use a transverse coordinate system or adjust algorithm accordingly.
	 The method will also compute the conversion factors to directly relate a point of latitude
	  or longitude to their respective grid coordinates.
	*/
void CEnumeratedVolume::computeVolumeDimensions(){
		//Compute enumerated volume size in the latitude (N-S) (Y) direction. First convert to the
		//latitude to yards (1 deg lat = 60 min, 1 min lat = 2000 yds).
	float latDiffYds = spanLat * MIN_PER_DEGREE * YARDS_PER_NM;
		//Convert the net spread in yards into a uniform grid.
		//  The +1 below ensures a grid point at either extreme of the extent
	volumeDimension[1] = (int)(latDiffYds/allowedRangeError + 0.5) + 1;

		//Compute enumerated volume size in the longitude (E-W) (X) direction. The length of one minute of
		//  longitude is equal to 1 nm * cos(latitude).
	float lonDiffYds = spanLon * cos(centerLat * DEG_TO_RAD) * MIN_PER_DEGREE * YARDS_PER_NM;	//Convert to yards (1 deg 
																						   //lon = 60 min, 1 min
																						   //lon = 2000 yds * cos(lat)).
	volumeDimension[0] = (int)(lonDiffYds/allowedRangeError + 0.5) + 1;

	lonToGridScale = (volumeDimension[0] - 1) / spanLon; 
	latToGridScale = (volumeDimension[1] - 1) / spanLat;

		//Depth is always considered constant for all volumes.
	float depth = MAX_DEPTH_YDS/allowedRangeError + 0.5;
	volumeDimension[2] = (int)depth + 1;  //The +1 add a point for the boundary
}	//end computeVolumeDimensions()


//end class CEnumeratedVolume

