//ExpansionTemplate.h - Specification for the associated class.


#ifndef ExpansionTemplate_h
#define ExpansionTemplate_h

#include "EnumeratedVolume.h"
#include "TargetMotionData.h"
#include <vector>

/* =======================================================================================
	The ExpansionTemplate is used to store information about a target containment region which
	exists at a particular momement in time.  The region may then be displayed or expanded.

	This software was developed as part of a grant by ONR 311, Office of Naval Research,
	800 North Quincy Street, Arlington, Virginia 22217, for development of an "Application 
	of Volume Modeling Techniques to Dynamic Containment Regions for Naval Applications".
	Grant was awarded to Arizona State University, Department of Computer Science and Engineering,
	Dr. Greg Nielson and Gary Graf.


	Author: Gary Graf, Arizona State University.
	Version:  1.0b, 14 Apr 2002

  All rights reserved.
  
	Notes:
=========================================================================================*/

/*!The ExpansionTemplate is used to store information about a target containment region which
exists at a particular momement in time.  The region may then be displayed or expanded.
*/
class CExpansionTemplate : public CEnumeratedVolume
{
/***************************  Public class varibles **************************************/	
public:
		/*!This structure is to hold the various array indices, including a computed
		  vector index, for the containment region.  As each surface, or other special
		  point in the volume is located, the indices are loaded into this structure
		  so they can be store, en masse, in a vector (e.g., list of edge points).
		  */
	struct ListIndices {
		long int vectorIndex;
		int latIndex;
		int lonIndex;
		int depthIndex;
	};

		//The number of grid points from the center to the outer edge of the expansion volume.
		//  Six different variables are used in case the expansion template is not spherical.
	int expDistNorth;
	int expDistSouth;
	int expDistEast;
	int expDistWest;
	int expDistUp;
	int expDistDown;

		//Define a vector to hold the list of points which define the outside surface
		//  of the containment region.
	typedef std::vector<ListIndices> LIST_INDEX_VECTOR;

		//The vector which will hold the list of all the points which define the outside surface
		//  of the expansion template.
	LIST_INDEX_VECTOR surfaceList;
//	LIST_INDEX_VECTOR surfaceFaceList;

		//The vector which will hold the list of all the points which define arcs on the outside 
		//  surface of the expansion template.
	LIST_INDEX_VECTOR arcsNE;
	LIST_INDEX_VECTOR arcsNW;
	LIST_INDEX_VECTOR arcsSE;
	LIST_INDEX_VECTOR arcsSW;
	LIST_INDEX_VECTOR arcsUE;
	LIST_INDEX_VECTOR arcsUW;
	LIST_INDEX_VECTOR arcsDE;
	LIST_INDEX_VECTOR arcsDW;
	LIST_INDEX_VECTOR arcsUN;
	LIST_INDEX_VECTOR arcsUS;
	LIST_INDEX_VECTOR arcsDN;
	LIST_INDEX_VECTOR arcsDS;

		//The vector which will hold the list of all the points which define quadrants on the 
		//  outside surface of the expansion template.
	LIST_INDEX_VECTOR quadsUNE;
	LIST_INDEX_VECTOR quadsUNW;
	LIST_INDEX_VECTOR quadsUSE;
	LIST_INDEX_VECTOR quadsUSW;
	LIST_INDEX_VECTOR quadsDNE;
	LIST_INDEX_VECTOR quadsDNW;
	LIST_INDEX_VECTOR quadsDSE;
	LIST_INDEX_VECTOR quadsDSW;

		//Grid coordinates of the center of the region.
	float regionCenter[3];
	
/***************************  Private class varibles ***********************************/ 
private:
		//Characteristics of target used for expansion.
	CTargetMotionData *assocTarget;

		//Time interval (in hours) over which expansion is calculated.
	float elapsedTimeHr;


/***************************  Public  methods ********************************************/
public:
		//Constructors and destructors


	/*!
		This constructor will take a min latitude and longitude along with the required grid
		dimensions to compute the final extent of the region.  It is used in the
		construction of simple spherical expansion templates.  

		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param latDim - An integer which specifies how many grid cell the region should occupy
				in latitude.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param lonDim - An integer which specifies how many grid cell the region should occupy
				in longitude.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
		@param target - An pointer to a CTargetMotionData object which provides the necessary
				information on target speed and motion needed to develop an expansion volume.
		@param timeHr - A float which is the ellapsed time to calculate the expansion.
		@param type - An enumerated integer which identifies how the exansion is developed.  The
				enumerated options are noted in documentation with the class constants.
	*/
	CExpansionTemplate(float, int, float, int, float, CTargetMotionData*, float, int);


		//!Destructor
	~CExpansionTemplate();


/***************************  Private methods ******************************************/
private:

	/*!
	  Method takes a volume and ensures that it has the odd number of grid points in each
		dimension required to correctly allign the center of the expansion template with
		the containment volume which is being expanded.  Geographic coordinates and grid
		coordinates are both adjusted as necessary.
	*/
	void adjustExpansionVolume();


	/*!
	  This method will take the expansion template and fill its data values with a spherical
			cloud of points.

		@param maxDist - A float representing the radius of the spherical data within the template,
					expressed as a measure of nautical miles.
	*/
	void fillSphericalRegion(float);


	/*
	  This method will search the volume for points which lie on the surface, including
	    all points which lie on the boundary of the volume.  A surface point is defined as
		 a location where the data is greater than zero and at least one point adjacent, not
		 counting diagonals, is equal to zero.  When it finds a surface point
	    it stores the corresponding vector and array indices in the appropriate surface list.  
		 The method automatically flags boundary points as surface points and then tests the
		 interior points.
	*/	
	void findSurfacePoints();


	/*
	  This method takes all the surface points and sorts them into arcs and quads.  All of the
	    surface points of the templante may be catagorized based on their relationship to the
		 ordinal planes which cut through the template.  That catagorization is later useful in
		 determining how the points of the template relate to the points in the grid of the 
		 containment region.

	  A face point is a point which lies on the surface of the expansion template and at the 
	    intersection of two ordinal planes.  Since the expansion template is defined by a bounding
		 rectangle, those face points are the surface points closest to the center of each face
		 of the bounding rectangle (but may or may not be on the surface of the rectangle).  With
		 six faces to the bounding rectange, there are six face points.  Face points are determined
		 in the method "addSurfacePointToList(long int, int, int, int)" for efficiency and not
		 handled in this method.  With only six points to be concerned with, they are not sorted
		 with respect to a sepcific orientation.
		  
	  An arc is defined as the sequece of points making up a line on the surface of the template 
	    which also lies in an ordinal plane cutting through the center of the template.  The 
		 points are further subdivided into which quadrant of the intersecting plane they lie in.  
		 The position of the point is then stored separately in a list appropriate to that arcs 
		 orientation.  With three planes cutting the volume and four quadrants per plane, there are 
		 12 possible arcs defined.  The sorting is relative to the position in the real world as
		 a two element combination of north, south, east, west, up and down.

	  A quad is defined as a set of points which line on the surface of the template.  The points
	    are subdivided into quadrants of the bounding rectangle relative to the position with
		 respect to the three possible ordinal planes cutting through the region.  With three
		 planes there are eight possible quadrants.  The sorting is relative to the position in the 
		 real world as a three element combination of north, south, east, west, up and down.

	  Note that face points are also the terminal ends of arcs and so those points show in both
	    lists.  Similarly, arcs are the terminal edges of quads and so show in both lists.  The
		 duplication is done to ensure that when mapping out a new expansion region that no gaps
		 are inadvertently generated by the inability to plot the appropriate boundary point.
	*/	
	void findArcsAndQuads();


	/*	
	  This method will test a point which is internal to the volume to ascertain if the point 
	    is on the surface.  A non-zero point is considered to be on the surface if any adjacent 
		 grid point (not counting diagonals) are zero.

		@param i - An integer which is the X grid index (longitude) of the volume data for the region.
		@param j - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param k - An integer which is the Z grid index (depth) of the volume data for the region.
		@param midOff - A long integer which is the product of the X and Y grid indices for the
								volume data for the region.  Precomputed and passed to avoid repetition.
		@returns - A boolean which is true if the point is on the surface, false otherwise.
	*/	
	bool pointIsOnSurface(int, int, int, long int);


	/*
		Takes a surface point and adds it to the point list.  Points are not catagorized with
			respect to facing.

		@param vecIndex - A long integer which is the vector offset equivalent to grid[i][j][k].
		@param lnIndex - An integer which is the X grid index (longitude) of the volume data for the region.
		@param ltIndex - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param depIndex - An integer which is the Z grid index (depth) of the volume data for the region.
	*/
	void addSurfacePointToList(long int, int, int, int);

};	//end class CExpansionTemplate specifications
#endif

