//MeshFromEnumeratedVolume.cpp - Body of associated class.

#include "MeshFromEnumeratedVolume.h"
#include <iostream.h>

/* =======================================================================================
	The MeshFromEnumeratedVolume is used to convert a enumerated volume as encapsulated as a
	CContainmentRegion and translate the surface of that volume into a triangular mesh.  The
	list of vertices and surface triangles is then available for use by standard graphics
	programs.

	This software was developed as part of a grant by ONR 311, Office of Naval Research, 
	800 North Quincy Street, Arlington, Virginia 22217, for development of an "Application 
	of Volume Modeling Techniques to Dynamic Containment Regions for Naval Applications".
	Grant was awarded to Arizona State University, PRISM Laboratory, Department of Computer 
	Science and Engineering, Dr. Greg Nielson and Gary Graf.


	Author: Gary Graf, PRISM Laboratory, Arizona State University.
	Version:  1.0, 10 Sep 2002

  Copyright 2002, all rights reserved.
  
	Notes:
=========================================================================================*/

		//The default max depth in yards
	float const static MAX_DEPTH_FT = 3000.0f;


/**************************************************************************************
---------------------------------------------------------------------------------------
		Constructors / Destructors
---------------------------------------------------------------------------------------
**************************************************************************************/

		/*
			The default constructor takes the containment region to convert and calls the
				appropriate methods to initiate conversion.

			@param target - A pointer to the CContainmentRegion which is to be translated
				back into a triangular mesh.
		*/
CMeshFromEnumeratedVolume::CMeshFromEnumeratedVolume(CContainmentRegion *target){
	clearLocalLists();

		//If valid containment, begin processing
	if(target != NULL){
		region = target;
		maxDepth = MAX_DEPTH_FT;
		regionDim = region->getGridDimensions();
		sizeOfGridArea = regionDim[0] * regionDim[1];

			//Find lat/lon offsets for a half grid step
		float halfLon  = region->getSpanLon() / (regionDim[0] - 1) / 2.0;
		float halfLat  = region->getSpanLat() / (regionDim[1] - 1) / 2.0;
		float halfDpth = maxDepth / (regionDim[2] - 1) / 2.0;

			//Create non-decimated triangles for flat surfaces
		basicSurfaceMesh(halfLon, halfLat, halfDpth);
			//Create non-decimated triangles for edges.
		basicEdgeMesh(halfLon, halfLat, halfDpth);
			//Create non-decimated triangles for corners.
		basicCornerMesh(halfLon, halfLat, halfDpth);

			//Blend meshes together
		blendMeshes();
	}	//end if
}	//end constructor CMeshFromEnumeratedVolume()

	
	//Destructor
CMeshFromEnumeratedVolume::~CMeshFromEnumeratedVolume(){
	clearLocalLists();

	//delete region;
	delete [] finalVerticeList;
	delete [] finalCornerList;
	delete regionDim;
}	//end destructor ~CMeshFromEnumeratedVolume()


/**************************************************************************************
---------------------------------------------------------------------------------------
		Private methods
---------------------------------------------------------------------------------------
**************************************************************************************/


	/*
		Several temporary data structures are created to effect the data conversion.  This 
			method clears all the structures and disposes of unused memory.  It is called in
			both initialization and destruction.
	*/
void CMeshFromEnumeratedVolume::clearLocalLists(){
		//Get variables to track across the local geometry sets.
	LIST_LOCAL_MESH::iterator localMeshIterator;
	LocalMeshData nextVerticeSet;

		//Clear any potentially old data
	if(!localPointMeshList.empty()){
			//Walk through the local lists to clear memory
		for (localMeshIterator = localPointMeshList.begin(); localMeshIterator != localPointMeshList.end(); ++localMeshIterator){
			nextVerticeSet = (*localMeshIterator);

			delete [] nextVerticeSet.localVerticeList;
			delete [] nextVerticeSet.localTriangleList;
		}	//end for
		localPointMeshList.clear();
	}	//end if

	if(!localArcMeshList.empty()){
			//Walk through the local lists to clear memory
		for (localMeshIterator = localArcMeshList.begin(); localMeshIterator != localArcMeshList.end(); ++localMeshIterator){
			nextVerticeSet = (*localMeshIterator);

			delete [] nextVerticeSet.localVerticeList;
			delete [] nextVerticeSet.localTriangleList;
		}	//end for
		localArcMeshList.clear();
	}	//end if

	if(!localQuadMeshList.empty()){
			//Walk through the local lists to clear memory
		for (localMeshIterator = localQuadMeshList.begin(); localMeshIterator != localQuadMeshList.end(); ++localMeshIterator){
			nextVerticeSet = (*localMeshIterator);

			delete [] nextVerticeSet.localVerticeList;
			delete [] nextVerticeSet.localTriangleList;
		}	//end for
		localQuadMeshList.clear();
	}	//end if
}	//end clearLocalLists()


		/*
			This method scans through the list of surface points of the containment region and creates
			  a small two triangle mesh piece.  The piece floats a half grid unit above the surface
			  point with the four corner vertices a half grid unit in either direction.  If the patch
			  is on the surface or at max depth it is forced to conform with those depth limits.
		*/
	void CMeshFromEnumeratedVolume::basicSurfaceMesh(float halfLon, float halfLat, float halfDpth){
		int i, j;
		int bend;

			//The net relative alterations for each vertex in form [vertex][lon,lat,depth]
		int deltaGrid[4][3];
			//The net relative alterations for each vertex in form [vertex][lon,lat,depth] for
			// those grids where the corner will be bent down
		int altDeltaGrid[4][3];
			//Relative positions of vertices with respect to surface point.  The values represent
			//  the direction (lon,lat,depth) of the point in the number of steps of a grid.  The basic
			//  values are predicted on an upward facing surface point.  The array is four sets of 
			//  coner points with the three value (lon,lat,depth) of each. 
		int verticeOffset[4][3]  = {{-1, 1, -1}, {1, 1, -1}, {-1, -1, -1}, {1, -1, -1}};
			//Relative positions of vertices with respect to surface point.  The values represent
			//  the direction (lon,lat,depth) of the point in the number of steps of a grid.  The basic
			//  values are predicted on an upward facing surface point.  The array is four sets of 
			//  coner points with the three value (lon,lat,depth) of each. This array is used for
			//  the additional four vertices used when a corner is bent down.
		int altVerticeOffset[4][3]  = {{0, 1, -1}, {1, 0, -1}, {-1, 0, -1}, {0, -1, -1}};
			//The rotation offset is the change of position of a vertex (lon,lat,depth) dependent on the
			//  facing of the surface point.  The first array index is the facing (0 = east, 1 = west,
			//  2 = north, 3 = south, 4 = up, 5 = down).
		int rotationOffset[6][3] = {{-1, 1, 1}, {1, 1, -1}, {1, -1, 1}, {1, 1, -1}, {1, 1, 1}, {-1, 1, -1}};
			//The rotation is the change of the relative offset of a vertex (lon,lat,depth) dependent on the
			//  facing of the surface point.  The offset is considered a movement of the existing
			//  lat/lon/depth location relative to the basline upward facing.  The first array index is the 
			//  facing (0 = east, 1 = west, 2 = north, 3 = south, 4 = up, 5 = down).  The second
			//  index is reference to which baseline index is substituted for this face (0 = lon from
			//  the baseline, 1 = lat, 2 = depth).
		int rotationFace[6][3] = {{2, 1, 0}, {2, 1, 0}, {0, 2, 1}, {0, 2, 1}, {0, 1, 2}, {0, 1, 2}};

			//Start by walking through the planar surface points on the original containment region
for (regionSurfIterator = region->surfacePointList.begin(); regionSurfIterator != region->surfacePointList.end(); ++regionSurfIterator){
			nextRegionSurfPt = (*regionSurfIterator);

				//Compute the resultant point for each surface point dependent on facing.
			switch(nextRegionSurfPt.facing){
				case 1:		//Opens to the east
					bend = testBentCorner();
					if(bend > 0){
							//Reset the indices for the bent corners
						if(bend & 2)
							verticeOffset[0][2] = 0;
						if(bend & 4)
							verticeOffset[1][2] = 0;
						if(bend & 8)
							verticeOffset[2][2] = 0;
						if(bend & 16)
							verticeOffset[3][2] = 0;

							//Process each of the four base vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j){
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
								altDeltaGrid[i][j] = altVerticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
							}	//end for

						recordBentSurfaceMeshData(deltaGrid, altDeltaGrid, halfLon, halfLat, halfDpth);

							//Restore the indices for the bent corners
						for(i = 0; i < 4; ++i)
							verticeOffset[i][2] = -1;
					}	//end if
					else{
							//Process each of the four vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j)
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];

						recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					}	//end else
					break;

				case 2:		//Opens to the west
					bend = testBentCorner();
					if(bend > 0){
							//Reset the indices for the bent corners
						if(bend & 2)
							verticeOffset[0][2] = 0;
						if(bend & 4)
							verticeOffset[1][2] = 0;
						if(bend & 8)
							verticeOffset[2][2] = 0;
						if(bend & 16)
							verticeOffset[3][2] = 0;

							//Process each of the four base vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j){
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
								altDeltaGrid[i][j] = altVerticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
							}	//end for

						recordBentSurfaceMeshData(deltaGrid, altDeltaGrid, halfLon, halfLat, halfDpth);

							//Restore the indices for the bent corners
						for(i = 0; i < 4; ++i)
							verticeOffset[i][2] = -1;
					}	//end if
					else{
							//Process each of the four vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j)
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];

						recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					}	//end else
					break;

				case 3:		//Opens to the east and west
						//Process each of the four vertices to the east, then each vertex coordinate
					for(i = 0; i < 4; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
					recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the four vertices to the west, then each vertex coordinate
					for(i = 0; i < 4; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
					recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 4:		//Opens to the north
					bend = testBentCorner();
					if(bend > 0){
							//Reset the indices for the bent corners
						if(bend & 2)
							verticeOffset[0][2] = 0;
						if(bend & 4)
							verticeOffset[1][2] = 0;
						if(bend & 8)
							verticeOffset[2][2] = 0;
						if(bend & 16)
							verticeOffset[3][2] = 0;

							//Process each of the four base vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j){
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
								altDeltaGrid[i][j] = altVerticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
							}	//end for

						recordBentSurfaceMeshData(deltaGrid, altDeltaGrid, halfLon, halfLat, halfDpth);

							//Restore the indices for the bent corners
						for(i = 0; i < 4; ++i)
							verticeOffset[i][2] = -1;
					}	//end if
					else{
							//Process each of the four vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j)
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];

						recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					}	//end else
					break;

				case 8:		//Opens to the south
					bend = testBentCorner();
					if(bend > 0){
							//Reset the indices for the bent corners
						if(bend & 2)
							verticeOffset[0][2] = 0;
						if(bend & 4)
							verticeOffset[1][2] = 0;
						if(bend & 8)
							verticeOffset[2][2] = 0;
						if(bend & 16)
							verticeOffset[3][2] = 0;

							//Process each of the four base vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j){
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
								altDeltaGrid[i][j] = altVerticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
							}	//end for

						recordBentSurfaceMeshData(deltaGrid, altDeltaGrid, halfLon, halfLat, halfDpth);

							//Restore the indices for the bent corners
						for(i = 0; i < 4; ++i)
							verticeOffset[i][2] = -1;
					}	//end if
					else{
							//Process each of the four vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j)
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];

						recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					}	//end else
					break;

				case 12:		//Opens to the north and south
						//Process each of the four vertices to the north, then each vertex coordinate
					for(i = 0; i < 4; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
					recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the four vertices to the south, then each vertex coordinate
					for(i = 0; i < 4; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
					recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 16:		//Opens to the up
					bend = testBentCorner();
					if(bend > 0){
							//Reset the indices for the bent corners
						if(bend & 2)
							verticeOffset[0][2] = 0;
						if(bend & 4)
							verticeOffset[1][2] = 0;
						if(bend & 8)
							verticeOffset[2][2] = 0;
						if(bend & 16)
							verticeOffset[3][2] = 0;

							//Process each of the four base vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j){
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
								altDeltaGrid[i][j] = altVerticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
							}	//end for

						recordBentSurfaceMeshData(deltaGrid, altDeltaGrid, halfLon, halfLat, halfDpth);

							//Restore the indices for the bent corners
						for(i = 0; i < 4; ++i)
							verticeOffset[i][2] = -1;
					}	//end if
					else{
							//Process each of the four vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j)
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];

						recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					}	//end else
					break;

				case 32:		//Opens to the down
					bend = testBentCorner();
					if(bend > 0){
							//Reset the indices for the bent corners
						if(bend & 2)
							verticeOffset[0][2] = 0;
						if(bend & 4)
							verticeOffset[1][2] = 0;
						if(bend & 8)
							verticeOffset[2][2] = 0;
						if(bend & 16)
							verticeOffset[3][2] = 0;

							//Process each of the four base vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j){
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
								altDeltaGrid[i][j] = altVerticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
							}	//end for

						recordBentSurfaceMeshData(deltaGrid, altDeltaGrid, halfLon, halfLat, halfDpth);

							//Restore the indices for the bent corners
						for(i = 0; i < 4; ++i)
							verticeOffset[i][2] = -1;
					}	//end if
					else{
							//Process each of the four vertices, then each vertex coordinate
						for(i = 0; i < 4; ++i)
							for(j = 0; j < 3; ++j)
								deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];

						recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					}	//end else
					break;


				case 48:		//Opens to the up and down
						//Process each of the four vertices to the up, then each vertex coordinate
					for(i = 0; i < 4; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
					recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the four vertices to the down, then each vertex coordinate
					for(i = 0; i < 4; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
					recordSurfaceMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				default:
					cout<<"basicSurfaceMesh():  Error in surfacePointList for region."<<endl;
					break;
			}	//end switch
		}	//end for
	}	//end basicSurfaceMesh()



		/*
			This method takes the relative vertice positions and records in the vertex and corner
			arrays the actual values.  It then stores the results in the vector.
		*/
	void CMeshFromEnumeratedVolume::recordSurfaceMeshData(int grid[4][3], float halfLon, 
																			float halfLat, float halfDpth){
			//Set up storage variables
		LocalMeshData tempMesh;								//Mesh consists of:
		tempMesh.localTriangleList  = new int[2][3];	//Two triangles of three vertices each, total of
		tempMesh.localNumTriangles = 2;
		tempMesh.localVerticeList = new float[4][3];	//Four vertices with three coordinates each.
		tempMesh.localNumVerts = 4;

			//Convert each of the offsets for the four vertices into real coordinates
		for(int i = 0; i < 4; ++i){
			tempMesh.localVerticeList[i][0] = region->getLonFromGrid(nextRegionSurfPt.lonIndex) +	//Start Lon
																(grid[i][0] * halfLon);					   //Delta Lon
			tempMesh.localVerticeList[i][1] = region->getLatFromGrid(nextRegionSurfPt.latIndex) +	//Start Lat
																(grid[i][1] * halfLat);					   //Delta Lat
			tempMesh.localVerticeList[i][2] = region->getDepthFromGrid(nextRegionSurfPt.depthIndex) + 
																(grid[i][2] * halfDpth);


				//Constrain to suface and floor
			if(tempMesh.localVerticeList[i][2] < 0)
				tempMesh.localVerticeList[i][2] = 0;
			if(tempMesh.localVerticeList[i][2] > MAX_DEPTH_FT)
				tempMesh.localVerticeList[i][2] = MAX_DEPTH_FT;
			}	//end for

			//Define the triangles that make up the face (wound counter-clockwise).
		tempMesh.localTriangleList[0][0]  = 0;
		tempMesh.localTriangleList[0][1]  = 2;
		tempMesh.localTriangleList[0][2]  = 1;

		tempMesh.localTriangleList[1][0]  = 3;
		tempMesh.localTriangleList[1][1]  = 1;
		tempMesh.localTriangleList[1][2]  = 2;

		localPointMeshList.push_back(tempMesh);
	}	//end recordSurfaceMeshData()


		/*
			This method takes the relative vertice positions and records in the vertex and corner
			arrays the actual values.  The vertices are adjusted depending on which corners must
			be bent down.  Additional vertices and corners are added to account for the bending.
			It then stores the results in the vector.
		*/
	void CMeshFromEnumeratedVolume::recordBentSurfaceMeshData(int grid[4][3], int altGrid[4][3],
																				 float halfLon, float halfLat, float halfDpth){
		int i;

			//Set up storage variables
		LocalMeshData tempMesh;								//Mesh consists of:
		tempMesh.localTriangleList  = new int[6][3];	//Six triangles of three vertices each, total of
		tempMesh.localNumTriangles = 6;
		tempMesh.localVerticeList = new float[8][3];	//Eight vertices with three coordinates each.
		tempMesh.localNumVerts = 8;

			//Convert each of the offsets for the four original vertices into real coordinates
		for(i = 0; i < 4; ++i){
			tempMesh.localVerticeList[i][0] = region->getLonFromGrid(nextRegionSurfPt.lonIndex) +	//Start Lon
																(grid[i][0] * halfLon);					   //Delta Lon
			tempMesh.localVerticeList[i][1] = region->getLatFromGrid(nextRegionSurfPt.latIndex) +	//Start Lat
																(grid[i][1] * halfLat);					   //Delta Lat
			tempMesh.localVerticeList[i][2] = region-> getDepthFromGrid(nextRegionSurfPt.depthIndex) + 
																(grid[i][2] * halfDpth);

			//Add each of the four additional vertices as real coordinates.
			tempMesh.localVerticeList[i + 4][0] = region->getLonFromGrid(nextRegionSurfPt.lonIndex) +	//Start Lon
																(altGrid[i][0] * halfLon);					   //Delta Lon
			tempMesh.localVerticeList[i + 4][1] = region->getLatFromGrid(nextRegionSurfPt.latIndex) +	//Start Lat
																(altGrid[i][1] * halfLat);					   //Delta Lat
			tempMesh.localVerticeList[i + 4][2] = region-> getDepthFromGrid(nextRegionSurfPt.depthIndex) + 
																(altGrid[i][2] * halfDpth);
		}	//end for

			//Constrain to suface and floor
		for(i = 0; i < 8; ++i){
			if(tempMesh.localVerticeList[i][2] < 0)
				tempMesh.localVerticeList[i][2] = 0;
			if(tempMesh.localVerticeList[i][2] > MAX_DEPTH_FT)
				tempMesh.localVerticeList[i][2] = MAX_DEPTH_FT;
		}	//end for

			//Define the triangles that make up the face (wound counter-clockwise).
		tempMesh.localTriangleList[0][0]  = 0;
		tempMesh.localTriangleList[0][1]  = 6;
		tempMesh.localTriangleList[0][2]  = 4;

		tempMesh.localTriangleList[1][0]  = 2;
		tempMesh.localTriangleList[1][1]  = 7;
		tempMesh.localTriangleList[1][2]  = 6;

		tempMesh.localTriangleList[2][0]  = 3;
		tempMesh.localTriangleList[2][1]  = 5;
		tempMesh.localTriangleList[2][2]  = 7;

		tempMesh.localTriangleList[3][0]  = 1;
		tempMesh.localTriangleList[3][1]  = 4;
		tempMesh.localTriangleList[3][2]  = 5;

		tempMesh.localTriangleList[4][0]  = 4;
		tempMesh.localTriangleList[4][1]  = 6;
		tempMesh.localTriangleList[4][2]  = 5;

		tempMesh.localTriangleList[5][0]  = 7;
		tempMesh.localTriangleList[5][1]  = 5;
		tempMesh.localTriangleList[5][2]  = 6;

		localPointMeshList.push_back(tempMesh);
	}	//end recordBentSurfaceMeshData()



		/*
			This method takes a surface point and checks to see if the mesh will overhang a
			void in the volume.  If so, it sets a return value which will later bend the corner 
			down to blend in with the corners or edges below.  The method looks at each corner of 
			the intended mesh and returns a integer which is a binary representation of which 
			corners should be bent, as follows:
				0  = no bending
				2  = corner 0 bent
				4  = corner 1 bent
				8  = corner 2 bent
				16 = corner 3 bent
		*/
	int CMeshFromEnumeratedVolume::testBentCorner(){
		int result = 0;
		int testLon = nextRegionSurfPt.lonIndex;
		int testLat = nextRegionSurfPt.latIndex;
		int testDph = nextRegionSurfPt.depthIndex;
		int *testDim = region->getGridDimensions();
		long int testMidOffset = testDim[1] * testDim[0];
		long int testPt = 0;

			//Compute the resultant test point for each surface point dependent on facing.
		switch(nextRegionSurfPt.facing){
			case 1:		//Opens to the east
					//Check if US or UN corners should be bent down
				if(testDph > 0){
					if(testLat > 0){	//US Corner
						testPt = region->getOffset(testLon, testLat - 1, testDph - 1, testMidOffset);
						if(region->volumeData[testPt] == 0)
							result += 8;
					}	//end if
					if(testLat < testDim[1] - 1){	//UN Corner
						testPt = region->getOffset(testLon, testLat + 1, testDph - 1, testMidOffset);
						if(region->volumeData[testPt] == 0)
							result += 2;
					}	//end if
				}	//end if
					//Check if DS or DN corners should be bent down
				if(testDph < testDim[2] - 1){
					if(testLat > 0){	//DS Corner
						testPt = region->getOffset(testLon, testLat - 1, testDph + 1, testMidOffset);
						if(region->volumeData[testPt] == 0)
							result += 16;
					}	//end if
					if(testLat < testDim[1] - 1){	//DN Corner
						testPt = region->getOffset(testLon, testLat + 1, testDph + 1, testMidOffset);
						if(region->volumeData[testPt] == 0)
							result += 4;
					}	//end if
				}	//end if
				break;

			case 2:		//Opens to the west
						//Check if US or UN corners should be bent down
					if(testDph > 0){
						if(testLat > 0){	//US Corner
							testPt = region->getOffset(testLon, testLat - 1, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 16;
						}	//end if
						if(testLat < testDim[1] - 1){	//UN Corner
							testPt = region->getOffset(testLon, testLat + 1, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 4;
						}	//end if
					}	//end if
						//Check if DS or DN corners should be bent down
					if(testDph < testDim[2] - 1){
						if(testLat > 0){	//DS Corner
							testPt = region->getOffset(testLon, testLat - 1, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 8;
						}	//end if
						if(testLat < testDim[1] - 1){	//DN Corner
							testPt = region->getOffset(testLon, testLat + 1, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 2;
						}	//end if
					}	//end if
				break;

			case 4:		//Opens to the north
						//Check if UW or DW corners should be bent down
					if(testLon > 0){
						if(testDph > 0){	//UW Corner
							testPt = region->getOffset(testLon - 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 8;
						}	//end if
						if(testDph < testDim[2] - 1){	//DW Corner
							testPt = region->getOffset(testLon - 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 2;
						}	//end if
					}	//end if
						//Check if UE or DE corners should be bent down
					if(testLon < testDim[0] - 1){
						if(testDph > 0){	//UE Corner
							testPt = region->getOffset(testLon + 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 16;
						}	//end if
						if(testDph < testDim[2] - 1){	//DE Corner
							testPt = region->getOffset(testLon + 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 4;
						}	//end if
					}	//end if
				break;

			case 8:		//Opens to the south
						//Check if UW or DW corners should be bent down
					if(testLon > 0){
						if(testDph > 0){	//UW Corner
							testPt = region->getOffset(testLon - 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 2;
						}	//end if
						if(testDph < testDim[2] - 1){	//DW Corner
							testPt = region->getOffset(testLon - 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 8;
						}	//end if
					}	//end if
						//Check if UE or DE corners should be bent down
					if(testLon < testDim[0] - 1){
						if(testDph > 0){	//UE Corner
							testPt = region->getOffset(testLon + 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 4;
						}	//end if
						if(testDph < testDim[2] - 1){	//DE Corner
							testPt = region->getOffset(testLon + 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 16;
						}	//end if
					}	//end if
				break;

			case 16:		//Opens to the up
					//Check if SW or NW corners should be bent down
				if(testLon > 0){
					if(testLat > 0){	//SW Corner
						testPt = region->getOffset(testLon - 1, testLat - 1, testDph, testMidOffset);
						if(region->volumeData[testPt] == 0)
							result += 8;
					}	//end if
					if(testLat < testDim[1] - 1){	//NW Corner
						testPt = region->getOffset(testLon - 1, testLat + 1, testDph, testMidOffset);
						if(region->volumeData[testPt] == 0)
							result += 2;
					}	//end if
				}	//end if
					//Check if SE or NE corners should be bent down
				if(testLon < testDim[0] - 1){
					if(testLat > 0){	//SE Corner
						testPt = region->getOffset(testLon + 1, testLat - 1, testDph, testMidOffset);
						if(region->volumeData[testPt] == 0)
							result += 16;
					}	//end if
					if(testLat < testDim[1] - 1){	//NE Corner
						testPt = region->getOffset(testLon + 1, testLat + 1, testDph, testMidOffset);
						if(region->volumeData[testPt] == 0)
							result += 4;
					}	//end if
				}	//end if
				break;

			case 32:		//Opens to the down
						//Check if SW or NW corners should be bent down
					if(testLon > 0){
						if(testLat > 0){	//SW Corner
							testPt = region->getOffset(testLon - 1, testLat - 1, testDph, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 16;
						}	//end if
						if(testLat < testDim[1] - 1){	//NW Corner
							testPt = region->getOffset(testLon - 1, testLat + 1, testDph, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 4;
						}	//end if
					}	//end if
						//Check if SE or NE corners should be bent down
					if(testLon < testDim[0] - 1){
						if(testLat > 0){	//SE Corner
							testPt = region->getOffset(testLon + 1, testLat - 1, testDph, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 8;
						}	//end if
						if(testLat < testDim[1] - 1){	//NE Corner
							testPt = region->getOffset(testLon + 1, testLat + 1, testDph, testMidOffset);
							if(region->volumeData[testPt] == 0)
								result += 2;
						}	//end if
					}	//end if
				break;

			default:
				cout<<"testBentCorner():  Error in facing for region."<<endl;
				break;
		}	//end switch

		return result;
	}	//end testBentCorner()



		/*
			This method scans through the list of edge points of the containment region and creates
			  a small six triangle mesh piece.  The piece floats a half grid unit above the surface
			  in the shape of a beveled angle.  The corner vertices are a half grid unit in any 
			  direction.  If the patch is on the surface or at max depth it is forced to conform 
			  with those depth limits.
		*/
	void CMeshFromEnumeratedVolume::basicEdgeMesh(float halfLon, float halfLat, float halfDpth){
		int i, j;
			//The net relative alterations for each vertex in form [vertex][lon,lat,depth]
		int deltaGrid[8][3];
			//Relative positions of vertices with respect to surface point.  The values represent
			//  the direction (lon,lat,depth) of the point in the number of steps of a grid.  The basic
			//  values are predicted on an UE facing edge point.  The array is eight sets of 
			//  coner points with the three value (lon,lat,depth) of each. 
		int verticeOffset[8][3] = {{-1, -1, -1}, {-1, 1, -1}, {0, -1, -1}, {0, 1, -1}, 
											{1, -1, 0},   {1, 1, 0},   {1, -1, 1},  {1, 1, 1}};
			//The rotation offset is the change of position of a vertex (lon,lat,depth) dependent on the
			//  facing of the surface point.  The array indices are in the order of facing (0 = NE, 
			//  1 = NW, 2 = SE, 3 = SW, 4 = UE, 5 = UW, 6 = DE, 7 = DW, 8 = UN, 9 = US, 10 = DN, 11 = DS).
		int rotationOffset[12][3] =  {{1, -1,  1}, {1,  1,  1}, {-1, -1,  1}, {-1, 1,  1}, 
												{1,  1,  1}, {1,  1, -1}, {-1,  1,  1}, {-1, 1, -1},
												{1, -1, -1}, {1, -1,  1}, { 1,  1, -1}, { 1, 1,  1}};
			//The rotation is the change of the relative offset of a vertex (lon,lat,depth) dependent on the
			//  facing of the surface point.  The offset is considered a movement of the existing
			//  lat/lon/depth location relative to the basline UE facing.  The array indices are in the 
			//  order of facing (0 = NE, 1 = NW, 2 = SE, 3 = SW, 4 = UE, 5 = UW, 6 = DE, 7 = DW, 
			//  8 = UN, 9 = US, 10 = DN, 11 = DS).
		int rotationFace[12][3] = {{0, 2, 1}, {2, 0, 1}, {2, 0, 1}, {0, 2, 1}, 
											{0, 1, 2}, {2, 1, 0}, {2, 1, 0}, {0, 1, 2},
											{1, 2, 0}, {1, 0, 2}, {1, 0, 2}, {1, 2, 0}};

			//Start by walking through the planar surface points on the original containment region
for (regionSurfIterator = region->surfaceArcList.begin(); regionSurfIterator != region->surfaceArcList.end(); ++regionSurfIterator){
			nextRegionSurfPt = (*regionSurfIterator);
				//Compute the resultant point for each surface point dependent on facing.
			switch(nextRegionSurfPt.facing){
				case 1:		//Edge to the NE
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(0, 0, halfLon, halfLat, halfDpth);
					break;

				case 2:		//Edge to the NW
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(1, 0, halfLon, halfLat, halfDpth);
					break;

				case 3:		//Edge to the NE and NW (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(0, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(1, 0, halfLon, halfLat, halfDpth);
					break;

				case 4:		//Edge to the SE
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(2, 0, halfLon, halfLat, halfDpth);
					break;

				case 5:		//Edge to the SE and NE (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(0, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(2, 0, halfLon, halfLat, halfDpth);
					break;

				case 8:		//Edge to the SW
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(3, 0, halfLon, halfLat, halfDpth);
					break;

				case 10:		//Edge to the SW and NW (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(1, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(3, 0, halfLon, halfLat, halfDpth);
					break;

				case 12:		//Edge to the SW and SE (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(2, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(3, 0, halfLon, halfLat, halfDpth);
					break;

				case 15:		//Edge to the SW, NW, NE and SE (Line)
						//Process each of the eight vertices to the NE, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the eight vertices to the NW, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the eight vertices to the SE, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the eight vertices to the SW, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(0, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(1, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(2, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(3, 0, halfLon, halfLat, halfDpth);
					break;

				case 16:		//Edge to the UE
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(4, 0, halfLon, halfLat, halfDpth);
					break;

				case 32:			//Edge to the UW
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(5, 0, halfLon, halfLat, halfDpth);
					break;

				case 48:			//Edge to the UW and UE (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(4, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(5, 0, halfLon, halfLat, halfDpth);
					break;

				case 64:		//Edge to the DE
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[6][j] ] * rotationOffset[6][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(6, 0, halfLon, halfLat, halfDpth);
					break;

				case 80:		//Edge to the DE and UE (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(4, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(6, 0, halfLon, halfLat, halfDpth);
					break;

				case 128:		//Edge to the DW
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[7][j] ] * rotationOffset[7][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(7, 0, halfLon, halfLat, halfDpth);
					break;

				case 160:		//Edge to the DW and UW (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(5, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(7, 0, halfLon, halfLat, halfDpth);
					break;

				case 192:		//Edge to the DW and DE (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(6, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(7, 0, halfLon, halfLat, halfDpth);
					break;

				case 240:		//Edge to the DW, UW, UE and DE (Line)
						//Process each of the eight vertices to the UE, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the eight vertices to the UW, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the eight vertices to the DE, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[6][j] ] * rotationOffset[6][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the eight vertices to the DW, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[7][j] ] * rotationOffset[7][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(4, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(5, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(6, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(7, 0, halfLon, halfLat, halfDpth);
					break;

				case 256:		//Edge to the UN
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[8][j] ] * rotationOffset[8][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(8, 0, halfLon, halfLat, halfDpth);
					break;

				case 512:		//Edge to the US
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[9][j] ] * rotationOffset[9][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(9, 0, halfLon, halfLat, halfDpth);
					break;

				case 768:		//Edge to the US and UN (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(8, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(9, 0, halfLon, halfLat, halfDpth);
					break;

				case 1024:		//Edge to the DN
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[10][j] ] * rotationOffset[10][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(10, 0, halfLon, halfLat, halfDpth);
					break;

				case 1280:		//Edge to the DN and UN (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(10, 0, halfLon, halfLat, halfDpth);
					break;

				case 2048:		//Edge to the DS
						//Process each of the eight vertices, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[11][j] ] * rotationOffset[11][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(11, 0, halfLon, halfLat, halfDpth);
					break;

				case 2560:		//Edge to the DS and US (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(9, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(11, 0, halfLon, halfLat, halfDpth);
					break;

				case 3072:		//Edge to the DS and DN (T-shape in plane)
					sharpEdgeMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(10, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(11, 0, halfLon, halfLat, halfDpth);
					break;

				case 3840:		//Edge to the DS, US, UN and DN (Line)
						//Process each of the eight vertices to the UN, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[8][j] ] * rotationOffset[8][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the eight vertices to the US, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[9][j] ] * rotationOffset[9][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the eight vertices to the DN, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[10][j] ] * rotationOffset[10][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the eight vertices to the DS, then each vertex coordinate
					for(i = 0; i < 8; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[11][j] ] * rotationOffset[11][j];
					recordEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(8, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(9, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(10, 0, halfLon, halfLat, halfDpth);
					testForHolePatch(11, 0, halfLon, halfLat, halfDpth);
					break;

				default:
					cout<<"basicEdgeMesh():  Error in surfaceArcList for region."<<endl;
					break;
			}	//end switch
		}	//end for
	}	//end basicEdgeMesh()


		/*
			This method takes the relative vertice positions and records in the vertex and corner
			arrays the actual values.  It then stores the results in the vector.
		*/
	void CMeshFromEnumeratedVolume::recordEdgeMeshData(int grid[8][3], float halfLon, 
																		float halfLat, float halfDpth){
			//Set up storage variables
		LocalMeshData tempMesh;
		tempMesh.localTriangleList  = new int[6][3];
		tempMesh.localNumTriangles = 6;
		tempMesh.localVerticeList = new float[8][3];
		tempMesh.localNumVerts = 8;

			//Convert each of the offsets into real coordinates
		for(int i = 0; i < 8; ++i){
			tempMesh.localVerticeList[i][0] = region->getLonFromGrid(nextRegionSurfPt.lonIndex) +
																grid[i][0] * halfLon;
			tempMesh.localVerticeList[i][1] = region->getLatFromGrid(nextRegionSurfPt.latIndex) +
																grid[i][1] * halfLat;
			tempMesh.localVerticeList[i][2] = region-> getDepthFromGrid(nextRegionSurfPt.depthIndex) + 
																grid[i][2] * halfDpth;
				//Constrain to suface and floor
			if(tempMesh.localVerticeList[i][2] < 0)
				tempMesh.localVerticeList[i][2] = 0;
			if(tempMesh.localVerticeList[i][2] > MAX_DEPTH_FT)
				tempMesh.localVerticeList[i][2] = MAX_DEPTH_FT;
		}	//end for

			//Define the triangles that make up the face (wound counter-clockwise).
		tempMesh.localTriangleList[0][0]  = 1;
		tempMesh.localTriangleList[0][1]  = 0;
		tempMesh.localTriangleList[0][2]  = 3;

		tempMesh.localTriangleList[1][0]  = 2;
		tempMesh.localTriangleList[1][1]  = 3;
		tempMesh.localTriangleList[1][2]  = 0;

		tempMesh.localTriangleList[2][0]  = 3;
		tempMesh.localTriangleList[2][1]  = 2;
		tempMesh.localTriangleList[2][2]  = 5;

		tempMesh.localTriangleList[3][0]  = 4;
		tempMesh.localTriangleList[3][1]  = 5;
		tempMesh.localTriangleList[3][2]  = 2;

		tempMesh.localTriangleList[4][0]  = 5;
		tempMesh.localTriangleList[4][1]  = 4;
		tempMesh.localTriangleList[4][2]  = 7;

		tempMesh.localTriangleList[5][0]  = 6;
		tempMesh.localTriangleList[5][1]  = 7;
		tempMesh.localTriangleList[5][2]  = 4;

		localArcMeshList.push_back(tempMesh);
	}	//end recordEdgeMeshData()


		/*
			This method takes the sharp edges where the end of an object is represented by a
			  planar piece and creates a small eight triangle mesh piece to wrap around the edge
			  of the plane.  The piece floats a half grid unit above the surface
			  in the shape of a hollow house.  The corner vertices are a half grid unit in any 
			  direction.  If the patch is on the surface or at max depth it is forced to conform 
			  with those depth limits.
		*/
	void CMeshFromEnumeratedVolume::sharpEdgeMesh(float halfLon, float halfLat, float halfDpth){
		int i, j;
			//The net relative alterations for each vertex in form [vertex][lon,lat,depth]
		int deltaGrid[10][3];
			//Relative positions of vertices with respect to surface point.  The values represent
			//  the direction (lon,lat,depth) of the point in the number of steps of a grid.  The basic
			//  values are predicted on an UE/DE facing edge.  The array is ten sets of 
			//  coner points with the three value (lon,lat,depth) of each. 
		int verticeOffset[10][3] =   {{-1, -1, -1}, {-1, 1, -1}, {0, -1, -1}, {0, 1, -1}, 
												{ 1, -1,  0}, { 1, 1,  0}, {0, -1,  1}, {0, 1,  1},
												{-1, -1,  1}, {-1, 1,  1}};
			//The rotation offset is the change of position of a vertex (lon,lat,depth) dependent on the
			//  facing of the surface point.  The array indices are in the order of facing: 
			//  (0 = NE/NW, 1 = SE/NE,  2 = SW/NW,  3 = SW/SE, 
			//   4 = UW/UE, 5 = DE/UE,  6 = DW/UW,  7 = DW/DE,
			//   8 = US/UN, 9 = DN/UN, 10 = DS/US, 11 = DS/DN).
		int rotationOffset[12][3] =  {{ 1,  1,  1}, { 1, -1,  1}, {-1,  1,  1}, {-1, -1,  1}, 
												{ 1,  1, -1}, { 1,  1,  1}, {-1,  1, -1}, {-1,  1,  1},
												{-1,  1, -1}, {-1,  1,  1}, {-1, -1, -1}, {-1, -1,  1}};
			//The rotation is the change of the relative offset of a vertex (lon,lat,depth) dependent on the
			//  facing of the surface point.  The offset is considered a movement of the existing
			//  lat/lon/depth location relative to the basline UE facing.  The array indices are in the 
			//  order of facing: 
			//  (0 = NE/NW, 1 = SE/NE,  2 = SW/NW,  3 = SW/SE, 
			//   4 = UW/UE, 5 = DE/UE,  6 = DW/UW,  7 = DW/DE,
			//   8 = US/UN, 9 = DN/UN, 10 = DS/US, 11 = DS/DN).
		int rotationFace[12][3] = {{2, 0, 1}, {0, 2, 1}, {0, 2, 1}, {2, 0, 1}, 
											{2, 1, 0}, {0, 1, 2}, {0, 1, 2}, {2, 1, 0},
											{1, 2, 0}, {1, 0, 2}, {1, 0, 2}, {1, 2, 0}};

				//Compute the resultant point for each surface point dependent on facing.
			switch(nextRegionSurfPt.facing){
				case 3:		//Edge to the NE and NW (T-shape in plane)
						//Process each of the eight vertices to the NE, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 5:		//Edge to the SE and NE (T-shape in plane)
						//Process each of the eight vertices to the NE, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 10:		//Edge to the SW and NW (T-shape in plane)
						//Process each of the eight vertices to the NW, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 12:		//Edge to the SW and SE (T-shape in plane)
						//Process each of the eight vertices to the SE, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 48:			//Edge to the UW and UE (T-shape in plane)
						//Process each of the eight vertices to the UE, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 80:		//Edge to the DE and UE (T-shape in plane)
						//Process each of the eight vertices to the UE, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 160:		//Edge to the DW and UW (T-shape in plane)
						//Process each of the eight vertices to the UW, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[6][j] ] * rotationOffset[6][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 192:		//Edge to the DW and DE (T-shape in plane)
						//Process each of the eight vertices to the DE, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[7][j] ] * rotationOffset[7][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 768:		//Edge to the US and UN (T-shape in plane)
						//Process each of the eight vertices to the UN, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[8][j] ] * rotationOffset[8][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 1280:		//Edge to the DN and UN (T-shape in plane)
						//Process each of the eight vertices to the UN, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[9][j] ] * rotationOffset[9][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 2560:		//Edge to the DS and US (T-shape in plane)
						//Process each of the eight vertices to the US, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[10][j] ] * rotationOffset[10][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 3072:		//Edge to the DS and DN (T-shape in plane)
						//Process each of the eight vertices to the DN, then each vertex coordinate
					for(i = 0; i < 10; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[11][j] ] * rotationOffset[11][j];
					recordSharpEdgeMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				default:
					cout<<"sharpEdgeMesh():  Error in surfaceArcList for region."<<endl;
					break;
			}	//end switch
	}	//end sharpEdgeMesh()


		/*
			This method takes the relative vertice positions and records in the vertex and corner
			arrays the actual values.  It then stores the results in the vector.
		*/
	void CMeshFromEnumeratedVolume::recordSharpEdgeMeshData(int grid[10][3], float halfLon, 
																			  float halfLat, float halfDpth){
			//Set up storage variables
		LocalMeshData tempMesh;
		tempMesh.localTriangleList  = new int[8][3];
		tempMesh.localNumTriangles = 8;
		tempMesh.localVerticeList = new float[10][3];
		tempMesh.localNumVerts = 10;

			//Convert each of the offsets into real coordinates
		for(int i = 0; i < 10; ++i){
			tempMesh.localVerticeList[i][0] = region->getLonFromGrid(nextRegionSurfPt.lonIndex) +
																grid[i][0] * halfLon;
			tempMesh.localVerticeList[i][1] = region->getLatFromGrid(nextRegionSurfPt.latIndex) +
																grid[i][1] * halfLat;
			tempMesh.localVerticeList[i][2] = region-> getDepthFromGrid(nextRegionSurfPt.depthIndex) + 
																grid[i][2] * halfDpth;
				//Constrain to suface and floor
			if(tempMesh.localVerticeList[i][2] < 0)
				tempMesh.localVerticeList[i][2] = 0;
			if(tempMesh.localVerticeList[i][2] > MAX_DEPTH_FT)
				tempMesh.localVerticeList[i][2] = MAX_DEPTH_FT;
		}	//end for

			//Define the triangles that make up the face (wound counter-clockwise).
		tempMesh.localTriangleList[0][0]  = 1;
		tempMesh.localTriangleList[0][1]  = 0;
		tempMesh.localTriangleList[0][2]  = 3;

		tempMesh.localTriangleList[1][0]  = 2;
		tempMesh.localTriangleList[1][1]  = 3;
		tempMesh.localTriangleList[1][2]  = 0;

		tempMesh.localTriangleList[2][0]  = 3;
		tempMesh.localTriangleList[2][1]  = 2;
		tempMesh.localTriangleList[2][2]  = 5;

		tempMesh.localTriangleList[3][0]  = 4;
		tempMesh.localTriangleList[3][1]  = 5;
		tempMesh.localTriangleList[3][2]  = 2;

		tempMesh.localTriangleList[4][0]  = 5;
		tempMesh.localTriangleList[4][1]  = 4;
		tempMesh.localTriangleList[4][2]  = 7;

		tempMesh.localTriangleList[5][0]  = 6;
		tempMesh.localTriangleList[5][1]  = 7;
		tempMesh.localTriangleList[5][2]  = 4;

		tempMesh.localTriangleList[6][0]  = 7;
		tempMesh.localTriangleList[6][1]  = 6;
		tempMesh.localTriangleList[6][2]  = 9;

		tempMesh.localTriangleList[7][0]  = 8;
		tempMesh.localTriangleList[7][1]  = 9;
		tempMesh.localTriangleList[7][2]  = 6;

		localArcMeshList.push_back(tempMesh);
	}	//end recordSharpEdgeMeshData()



		/*
			This method scans through the list of edge points of the containment region and creates
			  a small six triangle mesh piece.  The piece floats a half grid unit above the surface
			  in the shape of a beveled angle.  The corner vertices are a half grid unit in any 
			  direction.  If the patch is on the surface or at max depth it is forced to conform 
			  with those depth limits.
		*/
	void CMeshFromEnumeratedVolume::basicCornerMesh(float halfLon, float halfLat, float halfDpth){
		int i, j;
			//The net relative alterations for each vertex in form [vertex][lon,lat,depth]
		int deltaGrid[12][3];
			//Relative positions of vertices with respect to surface point.  The values represent
			//  the direction (lon,lat,depth) of the point in the number of steps of a grid.  The basic
			//  values are predicted on an USE facing edge point.  The array is eight sets of 
			//  coner points with the three value (lon,lat,depth) of each. 
		int verticeOffset[12][3] =   {{-1, -1, 1}, {-1, -1, 0}, {-1, 0, -1}, {-1, 1, -1}, 
												{ 0, -1, 1}, { 0, -1, 0}, { 0, 0, -1}, { 0, 1, -1},
												{ 1,  0, 1}, { 1,  0, 0}, { 1, 1,  0}, { 1, 1,  1}};
			//The rotation offset is the change of position of a vertex (lon,lat,depth) dependent on the
			//  facing of the surface point.  The array indices are in the order of facing:
			//  (0 = UNE, 1 = UNW, 2 = USE, 3 = USW, 
			//   4 = DNE, 5 = DNW, 6 = DSE, 7 = DSW).
		int rotationOffset[8][3] =   {{-1, 1,  1}, {-1, -1,  1}, { 1, 1,  1}, {1, -1,  1}, 
												{-1, 1, -1}, {-1, -1, -1}, { 1, 1, -1}, {1, -1, -1}};
			//The rotation is the change of the relative offset of a vertex (lon,lat,depth) dependent on the
			//  facing of the surface point.  The offset is considered a movement of the existing
			//  lat/lon/depth location relative to the basline UE facing.  The array indices are in the 
			//  order of facing:
			//  (0 = UNE, 1 = UNW, 2 = USE, 3 = USW, 
			//   4 = DNE, 5 = DNW, 6 = DSE, 7 = DSW).
		int rotationFace[8][3] =  {{1, 0, 2}, {0, 1, 2}, {0, 1, 2}, {1, 0, 2}, 
											{1, 0, 2}, {0, 1, 2}, {0, 1, 2}, {1, 0, 2}};

			//Start by walking through the planar surface points on the original containment region
for (regionSurfIterator = region->surfaceQuadList.begin(); regionSurfIterator != region->surfaceQuadList.end(); ++regionSurfIterator){
			nextRegionSurfPt = (*regionSurfIterator);

				//Compute the resultant point for each corner point dependent on facing.
			switch(nextRegionSurfPt.facing){
				case 1:		//Corner to the UNE
						//Process each of the twelve vertices to the UNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(0,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(4,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(8,  1, halfLon, halfLat, halfDpth);
					break;

				case 2:		//Corner to the UNW
						//Process each of the twelve vertices to the UNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(1,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(5,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(8,  2, halfLon, halfLat, halfDpth);
					break;

				case 3:		//Corner to the UNE and UNW (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(0,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(1,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(4,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(5,  1, halfLon, halfLat, halfDpth);
					break;

				case 4:		//Corner to the USE
						//Process each of the twelve vertices to the USE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(2,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(4,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(9,  2, halfLon, halfLat, halfDpth);
					break;

				case 5:		//Corner to the USE and UNE (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(0,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(2,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(8,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(9,  2, halfLon, halfLat, halfDpth);
					break;

				case 8:		//Corner to the USW
						//Process each of the twelve vertices to the USW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(3,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(5,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(9,  1, halfLon, halfLat, halfDpth);
					break;

				case 10:		//Corner to the USW and UNW (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(1,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(3,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(8,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(9,  1, halfLon, halfLat, halfDpth);
					break;

				case 12:		//Corner to the USW and USE (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(2,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(3,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(4,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(5,  2, halfLon, halfLat, halfDpth);
					break;

				case 15:		//Corner to the USW, UNW, UNE and USE (Point at end of line)
						//Process each of the twelve vertices to the USW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the UNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the UNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the USE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(0,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(1,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(2,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(3,  1, halfLon, halfLat, halfDpth);
					break;

				case 16:		//Corner to the DNE
						//Process each of the twelve vertices to the DNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(0,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(6,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(10, 2, halfLon, halfLat, halfDpth);
					break;

				case 17:		//Corner to the UNE and DNE (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(4 , 2, halfLon, halfLat, halfDpth);
					testForHolePatch(6,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(8,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(10, 2, halfLon, halfLat, halfDpth);
					break;

				case 32:		//Corner to the DNW
						//Process each of the twelve vertices to the DNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(1,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(7,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(10, 1, halfLon, halfLat, halfDpth);
					break;

				case 34:		//Corner to the UNW and DNW (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(5,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(7,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(8,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(10, 1, halfLon, halfLat, halfDpth);
					break;

				case 48:		//Corner to the DNW and DNE (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(0,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(1,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(6,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(7,  2, halfLon, halfLat, halfDpth);
					break;

				case 51:		//Corner to the DNW, UNW, UNE and DNE (Point at end of line)
						//Process each of the twelve vertices to the DNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the UNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the UNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(4,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(5,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(6,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(7,  2, halfLon, halfLat, halfDpth);
					break;

				case 64:		//Corner to the DSE
						//Process each of the twelve vertices to the DSE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[6][j] ] * rotationOffset[6][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(2,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(6,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(11, 1, halfLon, halfLat, halfDpth);
					break;

				case 68:		//Corner to the USE and DSE (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(4, 1, halfLon, halfLat, halfDpth);
					testForHolePatch(6, 2, halfLon, halfLat, halfDpth);
					testForHolePatch(9, 2, halfLon, halfLat, halfDpth);
					testForHolePatch(11,1, halfLon, halfLat, halfDpth);
					break;

				case 80:		//Corner to the DSE and DNE (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(0, 2, halfLon, halfLat, halfDpth);
					testForHolePatch(2, 1, halfLon, halfLat, halfDpth);
					testForHolePatch(10,2, halfLon, halfLat, halfDpth);
					testForHolePatch(11,1, halfLon, halfLat, halfDpth);
					break;

				case 85:		//Corner to the DSE, USE, DNE and UNE  (Point at end of line)
						//Process each of the twelve vertices to the DSE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[6][j] ] * rotationOffset[6][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the USE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the UNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(8,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(9,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(10, 2, halfLon, halfLat, halfDpth);
					testForHolePatch(11, 1, halfLon, halfLat, halfDpth);
					break;

				case 128:	//Corner to the DSW
						//Process each of the twelve vertices to the DSW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[7][j] ] * rotationOffset[7][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(3,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(7,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(11, 2, halfLon, halfLat, halfDpth);
					break;

				case 136:	//Corner to the DSW and USW (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(5, 2, halfLon, halfLat, halfDpth);
					testForHolePatch(7, 1, halfLon, halfLat, halfDpth);
					testForHolePatch(9, 1, halfLon, halfLat, halfDpth);
					testForHolePatch(11,2, halfLon, halfLat, halfDpth);
					break;

				case 160:	//Corner to the DNW and DSW (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(1,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(3,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(10, 1, halfLon, halfLat, halfDpth);
					testForHolePatch(11, 2, halfLon, halfLat, halfDpth);
					break;

				case 170:	//Corner to the UNW, DNW, USW and DSW  (Point at end of line)
						//Process each of the twelve vertices to the UNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the USW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DSW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[7][j] ] * rotationOffset[7][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(8,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(9,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(10, 1, halfLon, halfLat, halfDpth);
					testForHolePatch(11, 2, halfLon, halfLat, halfDpth);
					break;

				case 192:	//Corner to the DSW and DSE (L-shape in plane)
					sharpCornerMesh(halfLon, halfLat, halfDpth);
					testForHolePatch(2,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(3,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(6,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(7,  1, halfLon, halfLat, halfDpth);
					break;

				case 204:	//Corner to the DSW, USW, USE and DSE  (Point at end of line)
						//Process each of the twelve vertices to the DSW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[7][j] ] * rotationOffset[7][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the USW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the USE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DSE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[6][j] ] * rotationOffset[6][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(4,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(5,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(6,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(7,  1, halfLon, halfLat, halfDpth);
					break;

				case 240:	//Corner to the DNW, DSW, DSE and DNE  (Point at end of line)
						//Process each of the twelve vertices to the DNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DSW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[7][j] ] * rotationOffset[7][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DSE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[6][j] ] * rotationOffset[6][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					testForHolePatch(0,  2, halfLon, halfLat, halfDpth);
					testForHolePatch(1,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(2,  1, halfLon, halfLat, halfDpth);
					testForHolePatch(3,  2, halfLon, halfLat, halfDpth);
					break;

				case 255:	//Corner to all quads  (Point sitting by itself)
						//Process each of the twelve vertices to the UNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the UNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the USE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the USW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DSE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[6][j] ] * rotationOffset[6][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
						//Process each of the twelve vertices to the DSW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[7][j] ] * rotationOffset[7][j];
					recordCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				default:
					cout<<"basicCornerMesh():  Error in surfaceQuadList ("<<nextRegionSurfPt.facing<<")for region."<<endl;
					break;
			}	//end switch
		}	//end for
	}	//end basicCornerMesh()



		/*
			This method takes the relative vertice positions and records in the vertex and corner
			arrays the actual values.  It then stores the results in the vector.
		*/
	void CMeshFromEnumeratedVolume::recordCornerMeshData(int grid[12][3], float halfLon, 
																		  float halfLat, float halfDpth){
			//Set up storage variables
		LocalMeshData tempMesh;
		tempMesh.localTriangleList  = new int[13][3];
		tempMesh.localNumTriangles = 13;
		tempMesh.localVerticeList = new float[12][3];
		tempMesh.localNumVerts = 12;

			//Convert each of the offsets into real coordinates
		for(int i = 0; i < 12; ++i){
			tempMesh.localVerticeList[i][0] = region->getLonFromGrid(nextRegionSurfPt.lonIndex) +
																grid[i][0] * halfLon;
			tempMesh.localVerticeList[i][1] = region->getLatFromGrid(nextRegionSurfPt.latIndex) +
																grid[i][1] * halfLat;
			tempMesh.localVerticeList[i][2] = region-> getDepthFromGrid(nextRegionSurfPt.depthIndex) + 
																grid[i][2] * halfDpth;

				//Constrain to suface and floor
			if(tempMesh.localVerticeList[i][2] < 0)
				tempMesh.localVerticeList[i][2] = 0;
			if(tempMesh.localVerticeList[i][2] > MAX_DEPTH_FT)
				tempMesh.localVerticeList[i][2] = MAX_DEPTH_FT;
		}	//end for

			//Define the triangles that make up the face (wound counter-clockwise).
		tempMesh.localTriangleList[0][0]  = 3;
		tempMesh.localTriangleList[0][1]  = 2;
		tempMesh.localTriangleList[0][2]  = 7;

		tempMesh.localTriangleList[1][0]  = 6;
		tempMesh.localTriangleList[1][1]  = 7;
		tempMesh.localTriangleList[1][2]  = 2;

		tempMesh.localTriangleList[2][0]  = 2;
		tempMesh.localTriangleList[2][1]  = 1;
		tempMesh.localTriangleList[2][2]  = 6;

		tempMesh.localTriangleList[3][0]  = 6;
		tempMesh.localTriangleList[3][1]  = 1;
		tempMesh.localTriangleList[3][2]  = 5;

		tempMesh.localTriangleList[4][0]  = 1;
		tempMesh.localTriangleList[4][1]  = 0;
		tempMesh.localTriangleList[4][2]  = 5;

		tempMesh.localTriangleList[5][0]  = 5;
		tempMesh.localTriangleList[5][1]  = 0;
		tempMesh.localTriangleList[5][2]  = 4;

		tempMesh.localTriangleList[6][0]  = 5;
		tempMesh.localTriangleList[6][1]  = 4;
		tempMesh.localTriangleList[6][2]  = 9;

		tempMesh.localTriangleList[7][0]  = 9;
		tempMesh.localTriangleList[7][1]  = 4;
		tempMesh.localTriangleList[7][2]  = 8;

		tempMesh.localTriangleList[8][0]  = 9;
		tempMesh.localTriangleList[8][1]  = 8;
		tempMesh.localTriangleList[8][2]  = 10;

		tempMesh.localTriangleList[9][0]  = 10;
		tempMesh.localTriangleList[9][1]  = 8;
		tempMesh.localTriangleList[9][2]  = 11;

		tempMesh.localTriangleList[10][0]  = 7;
		tempMesh.localTriangleList[10][1]  = 6;
		tempMesh.localTriangleList[10][2]  = 10;

		tempMesh.localTriangleList[11][0]  = 10;
		tempMesh.localTriangleList[11][1]  = 6;
		tempMesh.localTriangleList[11][2]  = 9;

		tempMesh.localTriangleList[12][0]  = 6;
		tempMesh.localTriangleList[12][1]  = 5;
		tempMesh.localTriangleList[12][2]  = 9;

		localQuadMeshList.push_back(tempMesh);
	}	//end recordCornerMeshData()


		/*
			This method takes the sharp edges where the end of an object is represented by a
			  planar piece and creates a small fourteen triangle mesh piece to wrap around the
			  corner of the plane.  The piece floats a half grid unit above the surface
			  in the shape of a hollow house.  The corner vertices are a half grid unit in any 
			  direction.  If the patch is on the surface or at max depth it is forced to conform 
			  with those depth limits.
		*/
	void CMeshFromEnumeratedVolume::sharpCornerMesh(float halfLon, float halfLat, float halfDpth){
		int i, j;
			//The net relative alterations for each vertex in form [vertex][lon,lat,depth]
		int deltaGrid[12][3];
			//Relative positions of vertices with respect to surface point.  The values represent
			//  the direction (lon,lat,depth) of the point in the number of steps of a grid.  The basic
			//  values are predicted on an UNE/USE facing edge point.  The array is eight sets of 
			//  coner points with the three value (lon,lat,depth) of each. 
		int verticeOffset[12][3] =   {{-1, -1, 1}, {-1, -1, 0}, {-1,  0, -1}, {-1, 1,  0}, 
												{-1,  1, 1}, { 0, -1, 1}, { 0, -1,  0}, { 0, 0, -1},
												{ 0,  1, 0}, { 0,  1, 1}, { 1,  0,  1}, { 1, 0,  0}};
			//The rotation offset is the change of position of a vertex (lon,lat,depth) dependent on the
			//  facing of the surface point.  The array indices are in the order of facing: 
			//  (0 = UNE/UNW, 1 = USE/UNE,  2 = USW/UNW,  3 = USW/USE, 
			//   4 = UNE/DNE, 5 = UNW/DNW,  6 = DNW/DNE,  7 = USE/DSE,
			//   8 = DSE/DNE, 9 = DSW/USW, 10 = DNW/DSW, 11 = DSW/DSE).
		int rotationOffset[12][3] =  {{-1,  1,  1}, { 1,  1,  1}, { 1,  1, -1}, {-1,  1, -1}, 
												{ 1, -1,  1}, { 1,  1,  1}, {-1, -1,  1}, {-1, -1,  1},
												{-1,  1,  1}, {-1,  1,  1}, {-1,  1, -1}, {-1, -1, -1}};
			//The rotation is the change of the relative offset of a vertex (lon,lat,depth) dependent on the
			//  facing of the surface point.  The offset is considered a movement of the existing
			//  lat/lon/depth location relative to the basline UE facing.  The array indices are in the 
			//  order of facing: 
			//  (0 = UNE/UNW, 1 = USE/UNE,  2 = USW/UNW,  3 = USW/USE, 
			//   4 = UNE/DNE, 5 = UNW/DNW,  6 = DNW/DNE,  7 = USE/DSE,
			//   8 = DSE/DNE, 9 = DSW/USW, 10 = DNW/DSW, 11 = DSW/DSE).
		int rotationFace[12][3] = {{1, 0, 2}, {0, 1, 2}, {2, 1, 0}, {1, 2, 0}, 
											{0, 2, 1}, {2, 0, 1}, {1, 2, 0}, {2, 0, 1},
											{2, 1, 0}, {0, 2, 1}, {0, 1, 2}, {1, 0, 2}};

		//Compute the resultant point for each corner point dependent on facing.
			switch(nextRegionSurfPt.facing){
				case 3:		//Corner to the UNE and UNW (L-shape in plane)
						//Process each of the twelve vertices to the UNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[0][j] ] * rotationOffset[0][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 5:		//Corner to the USE and UNE (L-shape in plane)
						//Process each of the twelve vertices to the USE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[1][j] ] * rotationOffset[1][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 10:		//Corner to the USW and UNW (L-shape in plane)
						//Process each of the twelve vertices to the USW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[2][j] ] * rotationOffset[2][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 12:		//Corner to the USW and USE (L-shape in plane)
						//Process each of the twelve vertices to the USW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[3][j] ] * rotationOffset[3][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 17:		//Corner to the UNE and DNE (L-shape in plane)
						//Process each of the twelve vertices to the UNE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[4][j] ] * rotationOffset[4][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 34:		//Corner to the UNW and DNW (L-shape in plane)
						//Process each of the twelve vertices to the UNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[5][j] ] * rotationOffset[5][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 48:		//Corner to the DNW and DNE (L-shape in plane)
						//Process each of the twelve vertices to the DNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[6][j] ] * rotationOffset[6][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 68:		//Corner to the USE and DSE (L-shape in plane)
						//Process each of the twelve vertices to the USE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[7][j] ] * rotationOffset[7][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 80:		//Corner to the DSE and DNE (L-shape in plane)
						//Process each of the twelve vertices to the DSE, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[8][j] ] * rotationOffset[8][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 136:	//Corner to the DSW and USW (L-shape in plane)
						//Process each of the twelve vertices to the DSW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[9][j] ] * rotationOffset[9][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 160:	//Corner to the DNW and DSW (L-shape in plane)
						//Process each of the twelve vertices to the DNW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[10][j] ] * rotationOffset[10][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				case 192:	//Corner to the DSW and DSE (L-shape in plane)
						//Process each of the twelve vertices to the DSW, then each vertex coordinate
					for(i = 0; i < 12; ++i)
						for(j = 0; j < 3; ++j)
							deltaGrid[i][j] = verticeOffset[i][ rotationFace[11][j] ] * rotationOffset[11][j];
					recordSharpCornerMeshData(deltaGrid,	halfLon, halfLat, halfDpth);
					break;

				default:
					cout<<"sharpCornerMesh():  Error in surfaceArcList for region."<<endl;
					break;
			}	//end switch
	}	//end sharpCornerMesh()


		/*
			This method takes the relative vertice positions and records in the vertex and corner
			arrays the actual values.  It then stores the results in the vector.
		*/
	void CMeshFromEnumeratedVolume::recordSharpCornerMeshData(int grid[14][3], float halfLon, 
																				 float halfLat, float halfDpth){
			//Set up storage variables
		LocalMeshData tempMesh;
		tempMesh.localTriangleList  = new int[14][3];
		tempMesh.localNumTriangles = 14;
		tempMesh.localVerticeList = new float[12][3];
		tempMesh.localNumVerts = 12;

			//Convert each of the offsets into real coordinates
		for(int i = 0; i < 12; ++i){
			tempMesh.localVerticeList[i][0] = region->getLonFromGrid(nextRegionSurfPt.lonIndex) +
																grid[i][0] * halfLon;
			tempMesh.localVerticeList[i][1] = region->getLatFromGrid(nextRegionSurfPt.latIndex) +
																grid[i][1] * halfLat;
			tempMesh.localVerticeList[i][2] = region-> getDepthFromGrid(nextRegionSurfPt.depthIndex) + 
																grid[i][2] * halfDpth;

				//Constrain to suface and floor
			if(tempMesh.localVerticeList[i][2] < 0)
				tempMesh.localVerticeList[i][2] = 0;
			if(tempMesh.localVerticeList[i][2] > MAX_DEPTH_FT)
				tempMesh.localVerticeList[i][2] = MAX_DEPTH_FT;
		}	//end for

			//Define the triangles that make up the face (wound counter-clockwise).
		tempMesh.localTriangleList[0][0]  = 0;
		tempMesh.localTriangleList[0][1]  = 5;
		tempMesh.localTriangleList[0][2]  = 6;

		tempMesh.localTriangleList[1][0]  = 6;
		tempMesh.localTriangleList[1][1]  = 1;
		tempMesh.localTriangleList[1][2]  = 0;

		tempMesh.localTriangleList[2][0]  = 1;
		tempMesh.localTriangleList[2][1]  = 6;
		tempMesh.localTriangleList[2][2]  = 7;

		tempMesh.localTriangleList[3][0]  = 7;
		tempMesh.localTriangleList[3][1]  = 2;
		tempMesh.localTriangleList[3][2]  = 1;

		tempMesh.localTriangleList[4][0]  = 2;
		tempMesh.localTriangleList[4][1]  = 7;
		tempMesh.localTriangleList[4][2]  = 8;

		tempMesh.localTriangleList[5][0]  = 8;
		tempMesh.localTriangleList[5][1]  = 3;
		tempMesh.localTriangleList[5][2]  = 2;

		tempMesh.localTriangleList[6][0]  = 3;
		tempMesh.localTriangleList[6][1]  = 8;
		tempMesh.localTriangleList[6][2]  = 9;

		tempMesh.localTriangleList[7][0]  = 9;
		tempMesh.localTriangleList[7][1]  = 4;
		tempMesh.localTriangleList[7][2]  = 3;

		tempMesh.localTriangleList[8][0]  = 5;
		tempMesh.localTriangleList[8][1]  = 10;
		tempMesh.localTriangleList[8][2]  = 11;

		tempMesh.localTriangleList[9][0]  = 11;
		tempMesh.localTriangleList[9][1]  = 6;
		tempMesh.localTriangleList[9][2]  = 5;

		tempMesh.localTriangleList[10][0]  = 11;
		tempMesh.localTriangleList[10][1]  = 10;
		tempMesh.localTriangleList[10][2]  = 9;

		tempMesh.localTriangleList[11][0]  = 9;
		tempMesh.localTriangleList[11][1]  = 8;
		tempMesh.localTriangleList[11][2]  = 11;

		tempMesh.localTriangleList[12][0]  = 7;
		tempMesh.localTriangleList[12][1]  = 6;
		tempMesh.localTriangleList[12][2]  = 11;

		tempMesh.localTriangleList[13][0]  = 11;
		tempMesh.localTriangleList[13][1]  = 8;
		tempMesh.localTriangleList[13][2]  = 7;

		localQuadMeshList.push_back(tempMesh);
	}	//end recordSharpCornerMeshData()



		/*
			The possibility exists that a small triangular hole will exist if a beveled edge or
			  corner buts against a flat face.  This method tests for that condition and will plug
			  in an extra triangle to fill the hole if necessary.  The search is keyed to the facing
			  of the edge:
				  0 = NE, 1 = NW, 2 = SE, 3 = SW,  4 = UE,  5 = UW, 
				  6 = DE, 7 = DW, 8 = UN, 9 = US, 10 = DN, 11 = DS
			To test connectivity, the side parameter is assessed by imagining your right hand laying
			  over the edge.  Your palm rests on the face specified by the first directional letter 
			  (e.g., north for a NE edge), with your fingers curling over to the side flagged by the
			  second directional letter (e.g., east for a NE edge).  In this position, the first side
			  is in the direction of the thumb, and the second side in the opposite direction.
				  0 = both sides, 1 = 1st side, 2 = second side.

			NOTE:  You do not have to check for bounds violation along the ends of the edge since
			  if it existed on a boundary it would be a closed corner and not an exposed edge.
		*/
	void CMeshFromEnumeratedVolume::testForHolePatch(int direction, int side, 
																	 float halfLon, float halfLat, float halfDpth){
		int result = 0;
		int testLon = nextRegionSurfPt.lonIndex;
		int testLat = nextRegionSurfPt.latIndex;
		int testDph = nextRegionSurfPt.depthIndex;
		int *testDim = region->getGridDimensions();
		long int testMidOffset = testDim[1] * testDim[0];
		long int testPt = 0;
		int grid[3][3] = {{1, 1, 1}, {1, 1, 1}, {1, 1, 1}};

		switch (direction){
			case 0:	//Beveled edge to NE
						//First make sure it is not an outside edge
					if((testLon < testDim[0] - 1) && (testLat < testDim[1] - 1)){
							//Test down end first
						if(side != 2){
							testPt = region->getOffset(testLon + 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon, testLat + 1, testDph + 1, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] = 0;
									grid[0][1] = 1;
									grid[0][2] = 1;

									grid[1][0] = 1;
									grid[1][1] = 1;
									grid[1][2] = 1;

									grid[2][0] = 1;
									grid[2][1] = 0;
									grid[2][2] = 1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the up end
						if(side != 1){
							testPt = region->getOffset(testLon + 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon, testLat + 1, testDph - 1, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  1;
									grid[0][1] =  0;
									grid[0][2] = -1;

									grid[1][0] =  1;
									grid[1][1] =  1;
									grid[1][2] = -1;

									grid[2][0] =  0;
									grid[2][1] =  1;
									grid[2][2] = -1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			case 1:	//Beveled edge to NW
						//First make sure it is not an outside edge
					if((testLon > 0) && (testLat < testDim[1] - 1)){
							//Test up end first
						if(side != 2){
							testPt = region->getOffset(testLon - 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon, testLat + 1, testDph - 1, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  0;
									grid[0][1] =  1;
									grid[0][2] = -1;

									grid[1][0] = -1;
									grid[1][1] =  1;
									grid[1][2] = -1;

									grid[2][0] = -1;
									grid[2][1] =  0;
									grid[2][2] = -1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the down end
						if(side != 1){
							testPt = region->getOffset(testLon - 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon, testLat + 1, testDph + 1, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] = -1;
									grid[0][1] =  0;
									grid[0][2] =  1;

									grid[1][0] = -1;
									grid[1][1] =  1;
									grid[1][2] =  1;

									grid[2][0] =  0;
									grid[2][1] =  1;
									grid[2][2] =  1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			case 2:	//Beveled edge to SE
						//First make sure it is not an outside edge
					if((testLon < testDim[0] - 1) && (testLat > 0)){
							//Test up end first
						if(side != 2){
							testPt = region->getOffset(testLon + 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon, testLat - 1, testDph - 1, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  0;
									grid[0][1] = -1;
									grid[0][2] = -1;

									grid[1][0] =  1;
									grid[1][1] = -1;
									grid[1][2] = -1;

									grid[2][0] =  1;
									grid[2][1] =  0;
									grid[2][2] = -1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the down end
						if(side != 1){
							testPt = region->getOffset(testLon + 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon, testLat - 1, testDph + 1, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  1;
									grid[0][1] =  0;
									grid[0][2] =  1;

									grid[1][0] =  1;
									grid[1][1] = -1;
									grid[1][2] =  1;

									grid[2][0] =  0;
									grid[2][1] = -1;
									grid[2][2] =  1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			case 3:	//Beveled edge to SW
						//First make sure it is not an outside edge
					if((testLon > 0) && (testLat > 0)){
							//Test down end first
						if(side != 2){
							testPt = region->getOffset(testLon - 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon, testLat - 1, testDph + 1, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  0;
									grid[0][1] = -1;
									grid[0][2] =  1;

									grid[1][0] = -1;
									grid[1][1] = -1;
									grid[1][2] =  1;

									grid[2][0] = -1;
									grid[2][1] =  0;
									grid[2][2] =  1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the up end
						if(side != 1){
							testPt = region->getOffset(testLon - 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon, testLat - 1, testDph - 1, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] = -1;
									grid[0][1] =  0;
									grid[0][2] = -1;

									grid[1][0] = -1;
									grid[1][1] = -1;
									grid[1][2] = -1;

									grid[2][0] =  0;
									grid[2][1] = -1;
									grid[2][2] = -1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			case 4:	//Beveled edge to UE
						//First make sure it is not an outside edge
					if((testLon < testDim[0] - 1) && (testDph > 0)){
							//Test north end first
						if(side != 2){
							testPt = region->getOffset(testLon, testLat + 1, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon + 1, testLat + 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  0;
									grid[0][1] =  1;
									grid[0][2] = -1;

									grid[1][0] =  1;
									grid[1][1] =  1;
									grid[1][2] = -1;

									grid[2][0] =  1;
									grid[2][1] =  1;
									grid[2][2] =  0;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the south end
						if(side != 1){
							testPt = region->getOffset(testLon, testLat - 1, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon + 1, testLat - 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  1;
									grid[0][1] = -1;
									grid[0][2] =  0;

									grid[1][0] =  1;
									grid[1][1] = -1;
									grid[1][2] = -1;

									grid[2][0] =  0;
									grid[2][1] = -1;
									grid[2][2] = -1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			case 5:	//Beveled edge to UW
						//First make sure it is not an outside edge
					if((testLon > 0) && (testDph > 0)){
							//Test south end first
						if(side != 2){
							testPt = region->getOffset(testLon, testLat - 1, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon - 1, testLat - 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  0;
									grid[0][1] = -1;
									grid[0][2] = -1;

									grid[1][0] = -1;
									grid[1][1] = -1;
									grid[1][2] = -1;

									grid[2][0] = -1;
									grid[2][1] = -1;
									grid[2][2] =  0;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the north end
						if(side != 1){
							testPt = region->getOffset(testLon, testLat + 1, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon - 1, testLat + 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] = -1;
									grid[0][1] =  1;
									grid[0][2] =  0;

									grid[1][0] = -1;
									grid[1][1] =  1;
									grid[1][2] = -1;

									grid[2][0] =  0;
									grid[2][1] =  1;
									grid[2][2] = -1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			case 6:	//Beveled edge to DE
						//First make sure it is not an outside edge
					if((testLon < testDim[0] - 1) && (testDph < testDim[2] - 1)){
							//Test south end first
						if(side != 2){
							testPt = region->getOffset(testLon, testLat - 1, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon + 1, testLat - 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  0;
									grid[0][1] = -1;
									grid[0][2] =  1;

									grid[1][0] =  1;
									grid[1][1] = -1;
									grid[1][2] =  1;

									grid[2][0] =  1;
									grid[2][1] = -1;
									grid[2][2] =  0;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the north end
						if(side != 1){
							testPt = region->getOffset(testLon, testLat + 1, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon + 1, testLat + 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  1;
									grid[0][1] =  1;
									grid[0][2] =  0;

									grid[1][0] =  1;
									grid[1][1] =  1;
									grid[1][2] =  1;

									grid[2][0] =  0;
									grid[2][1] =  1;
									grid[2][2] =  1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			case 7:	//Beveled edge to DW
						//First make sure it is not an outside edge
					if((testLon > 0) && (testDph < testDim[2] - 1)){
							//Test north end first
						if(side != 2){
							testPt = region->getOffset(testLon, testLat + 1, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon - 1, testLat + 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  0;
									grid[0][1] =  1;
									grid[0][2] =  1;

									grid[1][0] = -1;
									grid[1][1] =  1;
									grid[1][2] =  1;

									grid[2][0] = -1;
									grid[2][1] =  1;
									grid[2][2] =  0;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the south end
						if(side != 1){
							testPt = region->getOffset(testLon, testLat - 1, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon - 1, testLat - 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] = -1;
									grid[0][1] = -1;
									grid[0][2] =  0;

									grid[1][0] = -1;
									grid[1][1] = -1;
									grid[1][2] =  1;

									grid[2][0] =  0;
									grid[2][1] = -1;
									grid[2][2] =  1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			case 8:	//Beveled edge to UN
						//First make sure it is not an outside edge
					if((testLat < testDim[1] - 1) && (testDph > 0)){
							//Test west end first
						if(side != 2){
							testPt = region->getOffset(testLon - 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon - 1, testLat + 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] = -1;
									grid[0][1] =  0;
									grid[0][2] = -1;

									grid[1][0] = -1;
									grid[1][1] =  1;
									grid[1][2] = -1;

									grid[2][0] = -1;
									grid[2][1] =  1;
									grid[2][2] =  0;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the east end
						if(side != 1){
							testPt = region->getOffset(testLon + 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon + 1, testLat + 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  1;
									grid[0][1] =  1;
									grid[0][2] =  0;

									grid[1][0] =  1;
									grid[1][1] =  1;
									grid[1][2] = -1;

									grid[2][0] =  1;
									grid[2][1] =  0;
									grid[2][2] = -1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			case 9:	//Beveled edge to US
						//First make sure it is not an outside edge
					if((testLat > 0) && (testDph > 0)){
							//Test east end first
						if(side != 2){
							testPt = region->getOffset(testLon + 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon + 1, testLat - 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  1;
									grid[0][1] =  0;
									grid[0][2] = -1;

									grid[1][0] =  1;
									grid[1][1] = -1;
									grid[1][2] = -1;

									grid[2][0] =  1;
									grid[2][1] = -1;
									grid[2][2] =  0;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the west end
						if(side != 1){
							testPt = region->getOffset(testLon - 1, testLat, testDph - 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon - 1, testLat - 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] = -1;
									grid[0][1] = -1;
									grid[0][2] =  0;

									grid[1][0] = -1;
									grid[1][1] = -1;
									grid[1][2] = -1;

									grid[2][0] = -1;
									grid[2][1] =  0;
									grid[2][2] = -1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			case 10:	//Beveled edge to DN
						//First make sure it is not an outside edge
					if((testLat < testDim[1] - 1) && (testDph < testDim[2] - 1)){
							//Test east end first
						if(side != 2){
							testPt = region->getOffset(testLon + 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon + 1, testLat + 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] = 1;
									grid[0][1] = 0;
									grid[0][2] = 1;

									grid[1][0] = 1;
									grid[1][1] = 1;
									grid[1][2] = 1;

									grid[2][0] = 1;
									grid[2][1] = 1;
									grid[2][2] = 0;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the west end
						if(side != 1){
							testPt = region->getOffset(testLon - 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon - 1, testLat + 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] = -1;
									grid[0][1] =  1;
									grid[0][2] =  0;

									grid[1][0] = -1;
									grid[1][1] =  1;
									grid[1][2] =  1;

									grid[2][0] = -1;
									grid[2][1] =  0;
									grid[2][2] =  1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			case 11:	//Beveled edge to DS
						//First make sure it is not an outside edge
					if((testLat > 0) && (testDph < testDim[2] - 1)){
							//Test west end first
						if(side != 2){
							testPt = region->getOffset(testLon - 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon - 1, testLat - 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] = -1;
									grid[0][1] =  0;
									grid[0][2] =  1;

									grid[1][0] = -1;
									grid[1][1] = -1;
									grid[1][2] =  1;

									grid[2][0] = -1;
									grid[2][1] = -1;
									grid[2][2] =  0;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
	
							//Test the east end
						if(side != 1){
							testPt = region->getOffset(testLon + 1, testLat, testDph + 1, testMidOffset);
							if(region->volumeData[testPt] > 0){ //Now test if north point exists.
								testPt = region->getOffset(testLon + 1, testLat - 1, testDph, testMidOffset);
								if(region->volumeData[testPt] > 0){ //Hole exists, patch.
									grid[0][0] =  1;
									grid[0][1] = -1;
									grid[0][2] =  0;

									grid[1][0] =  1;
									grid[1][1] = -1;
									grid[1][2] =  1;

									grid[2][0] =  1;
									grid[2][1] =  0;
									grid[2][2] =  1;

									recordHolePatchMeshData(grid, halfLon, halfLat,halfDpth);
								}	//end if
							}	//end if 
						}	//end if
					}	//end if
					break;

			default:
				cout<<"testForHolePatch():  Error in direction."<<endl;
				break;
		}	//end switch
	}	//end testForHolePatch()



	void CMeshFromEnumeratedVolume::recordHolePatchMeshData(int grid[3][3], float halfLon, 
																			  float halfLat, float halfDpth){
			//Set up storage variables
		LocalMeshData tempMesh;
		tempMesh.localTriangleList  = new int[1][3];
		tempMesh.localNumTriangles = 1;
		tempMesh.localVerticeList = new float[3][3];
		tempMesh.localNumVerts = 3;

			//Convert each of the offsets into real coordinates
		for(int i = 0; i < 3; ++i){
			tempMesh.localVerticeList[i][0] = region->getLonFromGrid(nextRegionSurfPt.lonIndex) +
																grid[i][0] * halfLon;
			tempMesh.localVerticeList[i][1] = region->getLatFromGrid(nextRegionSurfPt.latIndex) +
																grid[i][1] * halfLat;
			tempMesh.localVerticeList[i][2] = region-> getDepthFromGrid(nextRegionSurfPt.depthIndex) + 
																grid[i][2] * halfDpth;
			//hooray for tpd
			if (tempMesh.localVerticeList[i][2]<0) tempMesh.localVerticeList[i][2]=0;
			if (tempMesh.localVerticeList[i][2]>MAX_DEPTH_FT) tempMesh.localVerticeList[i][2]=MAX_DEPTH_FT;
		}	//end for


			//Define the triangles that make up the face (wound counter-clockwise).
		tempMesh.localTriangleList[0][0]  = 2;
		tempMesh.localTriangleList[0][1]  = 1;
		tempMesh.localTriangleList[0][2]  = 0;

		localQuadMeshList.push_back(tempMesh);
	}	//end recordHolePatchMeshData()
	
int insertPoint(float ptlist[][3],int *curnum,float pt[3]) {
  int i;
  for (i=0;i<*curnum;i++) {
    if (ptlist[i][0]==pt[0])
      if (ptlist[i][1]==pt[1])
	if (ptlist[i][2]==pt[2])
	  return i;
  }
  ptlist[*curnum][0]=pt[0];
  ptlist[*curnum][1]=pt[1];
  ptlist[*curnum][2]=pt[2];
  int ret = *curnum;
  (*curnum)++;
  return ret;
}
		/*
			This method takes the individual point, arc, and corner mesh lists and creates
			a single mesh array for the containment region.
		*/
	void CMeshFromEnumeratedVolume::blendMeshes(){
		int i, j;
		int vertIndexOffset = 0;
		int triIndexOffset = 0;

		int currentVert = 0;
		int currentCorner = 0;
		LIST_LOCAL_MESH::iterator localMeshIterator;
		LocalMeshData nextLocalMesh;

		finalVertCount = 0;
		finalTriangleCount = 0;

			//Scan each list of local meshes to tally up the total number of vertices and triangles used.
			//  A direct computation cannot be used since special cases may add or remove triangles.
		for (localMeshIterator = localPointMeshList.begin(); localMeshIterator != localPointMeshList.end(); ++localMeshIterator){
			nextLocalMesh = (*localMeshIterator);
				//Find the number of vertices and triangles used
			finalVertCount += nextLocalMesh.localNumVerts;
			finalTriangleCount += nextLocalMesh.localNumTriangles;
		}	//end for (iterator)
		for (localMeshIterator = localArcMeshList.begin(); localMeshIterator != localArcMeshList.end(); ++localMeshIterator){
			nextLocalMesh = (*localMeshIterator);
				//Find the number of vertices and triangles used
			finalVertCount += nextLocalMesh.localNumVerts;
			finalTriangleCount += nextLocalMesh.localNumTriangles;
		}	//end for (iterator)
		for (localMeshIterator = localQuadMeshList.begin(); localMeshIterator != localQuadMeshList.end(); ++localMeshIterator){
			nextLocalMesh = (*localMeshIterator);
				//Find the number of vertices and triangles used
			finalVertCount += nextLocalMesh.localNumVerts;
			finalTriangleCount += nextLocalMesh.localNumTriangles;
		}	//end for (iterator)

			//Allocate arrays for data to be passed outside enumerated voulume classes
		finalVerticeList = new float[finalVertCount][3];
		finalCornerList = new int[finalTriangleCount][3];
		int *localvertmap;

			//Start by walking through the planar surface points
		for (localMeshIterator = localPointMeshList.begin(); localMeshIterator != localPointMeshList.end(); ++localMeshIterator){
			nextLocalMesh = (*localMeshIterator);
				//Add vert coordinates to list
			localvertmap=new int[nextLocalMesh.localNumVerts];
			for(i = 0; i < nextLocalMesh.localNumVerts; ++i){
			  localvertmap[i]=insertPoint(finalVerticeList,&vertIndexOffset,nextLocalMesh.localVerticeList[i]);
				//for(j = 0; j < 3; ++j){
				//	finalVerticeList[vertIndexOffset + i][j] = nextLocalMesh.localVerticeList[i][j];
				//}	//end for
			}	//end for

				//Add triangle indices to list
			for(i = 0; i < nextLocalMesh.localNumTriangles; ++i){
				for(j = 0; j < 3; ++j){
				  finalCornerList[triIndexOffset + i][j] = localvertmap[nextLocalMesh.localTriangleList[i][j]]; //+ vertIndexOffset;
				}	//end for
			}	//end for

			delete[] localvertmap;

			//vertIndexOffset += nextLocalMesh.localNumVerts;
			triIndexOffset  += nextLocalMesh.localNumTriangles;
		}	//end for (interator)

			//Add mesh data from the edge points
		for (localMeshIterator = localArcMeshList.begin(); localMeshIterator != localArcMeshList.end(); ++localMeshIterator){
			nextLocalMesh = (*localMeshIterator);
				//Add vert coordinates to list
			localvertmap=new int[nextLocalMesh.localNumVerts];
			for(i = 0; i < nextLocalMesh.localNumVerts; ++i){
			  localvertmap[i]=insertPoint(finalVerticeList,&vertIndexOffset,nextLocalMesh.localVerticeList[i]);
				//for(j = 0; j < 3; ++j){
				//	finalVerticeList[vertIndexOffset + i][j] = nextLocalMesh.localVerticeList[i][j];
				//}	//end for
			}	//end for


				//Add triangle indices to list
			for(i = 0; i < nextLocalMesh.localNumTriangles; ++i){
				for(j = 0; j < 3; ++j){
				  finalCornerList[triIndexOffset + i][j] = localvertmap[nextLocalMesh.localTriangleList[i][j]]; //+ vertIndexOffset;
				  //					finalCornerList[triIndexOffset + i][j] = nextLocalMesh.localTriangleList[i][j] + 
				  //  vertIndexOffset;
				}	//end for
			}	//end for
			delete[] localvertmap;
			//vertIndexOffset += nextLocalMesh.localNumVerts;
			triIndexOffset  += nextLocalMesh.localNumTriangles;
		}	//end for (interator)



			//Add mesh data from the corner points
		for (localMeshIterator = localQuadMeshList.begin(); localMeshIterator != localQuadMeshList.end(); ++localMeshIterator){
			nextLocalMesh = (*localMeshIterator);
				//Add vert coordinates to list
			localvertmap=new int[nextLocalMesh.localNumVerts];
			for(i = 0; i < nextLocalMesh.localNumVerts; ++i){
			  localvertmap[i]=insertPoint(finalVerticeList,&vertIndexOffset,nextLocalMesh.localVerticeList[i]);
				//for(j = 0; j < 3; ++j){
				//	finalVerticeList[vertIndexOffset + i][j] = nextLocalMesh.localVerticeList[i][j];
				//}	//end for
			}	//end for

				//Add triangle indices to list
			for(i = 0; i < nextLocalMesh.localNumTriangles; ++i){
				for(j = 0; j < 3; ++j){
				  finalCornerList[triIndexOffset + i][j] = localvertmap[nextLocalMesh.localTriangleList[i][j]]; //+ vertIndexOffset;
				  //					finalCornerList[triIndexOffset + i][j] = nextLocalMesh.localTriangleList[i][j] + 
				  //  vertIndexOffset;
				}	//end for
			}	//end for
			delete[] localvertmap;
			//vertIndexOffset += nextLocalMesh.localNumVerts;
			triIndexOffset  += nextLocalMesh.localNumTriangles;
		}	//end for (interator)
		finalVertCount=vertIndexOffset;

	}	//end blendMeshes()


/**************************************************************************************
---------------------------------------------------------------------------------------
		Public methods - General
---------------------------------------------------------------------------------------
**************************************************************************************/

		/*
			This method provides a vehicle to retrieve the arrays of vertices and
				triangles created from the class.  Fixed structures are not usable size the
				size of the dataset is unknown beforehand.  The user passes pointers to
				the method which then gets them to point to the correct data structures.

			@param vList - A pointer to an array of n x 3 vertices.  The value n refers to the
					total number of vertices in the containment mesh, and the three refers to
					the value of the longitude/latitude/depth.
			@param triList - A pointer to an array of m x 3 triangles.  The value m refers to the
					total number of triangles in the containment mesh, and the three refers to
					the value of the index for each corner vertice.  The index relates directly
					to the number (n) of the vertice passed in the vList[n][3] pointer.
		*/
	void CMeshFromEnumeratedVolume::getMesh(float (*vList)[3], int (*triList)[3]){
		int i, j;

		for(i = 0; i < finalVertCount; ++i){
			for(j = 0; j < 3; ++j){
				vList[i][j] = finalVerticeList[i][j];
			}	//end for
		}	//end for

		for(i = 0; i < finalTriangleCount; ++i){
			for(j = 0; j < 3; ++j){
				triList[i][j] = finalCornerList[i][j];
			}	//end for
		}	//end for
	}	//end
	

		/*
			A simple accessor.  I returns the number (n) of vertices in the containment mesh.  It
				is needed to correctly size the pointer to the array of vertices.

			@returns - An integer reflecting the size of the vertice array.
		*/
	int CMeshFromEnumeratedVolume::getNumVertsInMesh(){
		return finalVertCount;}


		/*
			A simple accessor.  I returns the number (m) of triangles in the containment mesh.  It
				is needed to correctly size the pointer to the array of triangles.

			@returns - An integer reflecting the size of the triangle array.
		*/
	int CMeshFromEnumeratedVolume::getNumTrianglesInMesh(){
		return finalTriangleCount;}

	//end class CMeshFromEnumeratedVolume
