#ifndef BQUEUE_H
#define BQUEUE_H

#include <stdlib.h>
#include <string.h>

  /// A queue of fixed maximum length, using a vector for internal storage.
   class BoundedQueue {
   public:		
      /// The queue internal storage to hold the items pushed to the queue.
      char *items;
      /// Points to first item in the queue
      char *head;
      /// Points to the last item in the queue
      char *tail;
      /// End of the queue, always equal to items+itemSize*capacity 		
      char *limit;
      /// Size in bytes of an item in the queue		
      unsigned int itemSize;
      /// Maximum number of items the queue can hold
      unsigned int capacity;
      /// Current number of items the queue holds, 0<=count<=capacity
      unsigned int count;
   
   protected:
      /* Given a valid pointer to an item inside the queue, it moves the pointer one item forward in the queue, wrapping around if needed. */
      void next(char*& pos) const {
         pos+=itemSize;
         if(pos == limit) pos = items;
      
      }
      /* Given a valid pointer to an item inside the queue, it moves the pointer one item back in the queue, wrapping around if needed. */
      void prev(char*& pos) const {
         if(pos ==items) pos = limit-itemSize;
         else pos-=itemSize;
      }
   
   public:
      bool isEmpty() const { 
         return count==0; };
   
      bool isFull() const { 
         return count==capacity;
      }

      /*! Copies and appends the element to the tail of the queue.
          \param item	Pointer to an area where to copy the item from.
          \return true if the queue is not empty and an element was copied to item, or false if the queue is empty.
      */    
      bool push(void* const item) { 
         if(!isFull()) { 
            memcpy(tail,item,itemSize);
            next(tail);
            count++;
            return true; 
         } 
         else 
            return false; 
      }

      /*! Copies the element at the head of the queue, and moves the head to the next element in the queue.
          \param item	Pointer to an area where to copy the item at the head of the queue.
          \return true if the queue is not empty and an element was copied to item, or false if the queue is empty.
      */  
      bool pop(void* item) {
         if(!isEmpty()) {
            memcpy(item,head,itemSize);
            next(head);
            count--;
            return true;
         }
         else
            return false;
      }

      /*! Copies the element at the head of the queue, without discarding it.
          \param item	Pointer to an area where to copy the item at the head of the queue.
          \return true if the queue is not empty and an element was copied to item, or false if the queue is empty.
      */  
      bool peek(void* item) {
         if(!isEmpty()) {
            memcpy(item,head,itemSize);
            return true;
         }
         else
            return false;
      }
   
      /// Discards the element in the head of the queue
      bool skip() {	
         if(!isEmpty()) {
            next(head);
            count--;
            return true;
         } 
         else 
            return false;
      }
   

      /// Empties the queue. 
      void empty() {
         tail = head;
         count = 0;
      }
   
      /// Returns a pointer to the last item in the queue. Doesn't check if the queue is empty.
      char* last() {
         return tail-itemSize;
      }

      /// Returns the current number of elements in the queue.
      unsigned int size() const { 
         return count; }
   
      /*! Given a pointer to a position inside the queue' internal storage in BoundedQueue::items, 
         will return the offset of that position from the beginning of the queue. */
      unsigned int queueOfsFromQueuePosition(void* const queuePos) const {
         return ((char*)queuePos-items)/itemSize;
      }
   
      /*! Given an position number, it will return a pointer to the item in that position inside the queue's internal storage. */
      char* queuePositionFromQueueOfs(unsigned int pos) const {
         return items+pos*itemSize;
      }
   
      /*! Will remove any item from the queue, ignoring the queue's semantic.
       \param abs_pos	An item index from the beginning of the queue's internal storage (BoundedQueue::items) 
       \param removedItem A pointer to an area where to copy the removed item, or NULL if the copy is not needed.
      */     
      bool removeStraightFromBuffer(unsigned int abs_pos, void* removedItem);

      /*! Creates a bounded queue that will hold 'howManyItems' items, each of size 'queueItemSizeInBytes'.
         \param queueItemSizeInBytes The number of bytes (sizeof) of the type of items the queue will hold. All items will be of the same size.
         \param howManyItems  The maximum number of items the queue will be able to hold. Call resizeTo() to change the queue limit.
      */
      BoundedQueue(unsigned int queueItemSizeInBytes,unsigned int howManyItems):capacity(howManyItems),itemSize(queueItemSizeInBytes) {
      
         unsigned int no_bytes=howManyItems*queueItemSizeInBytes;
         items = (char*)malloc(no_bytes);
         limit = items+no_bytes;
         head = items;
         tail = items;
         count = 0;
      }
      /*! Expands or contracts the queue, to a new maximum of 'howManyItems' items. */
      void resizeTo(unsigned int howManyItems);
   
      ~BoundedQueue() {
         if(items) free(items);
      }
   
   };

#endif
