#include "FeederConnection.h"
#include <string.h>
#include <sys/types.h>
#include <sys/un.h> // socketaddr_un for AF_UNIX, use <netinet/in.h> for AF_INET

const char *FeederConnection::SOCK_ADDRESS  = "/tmp/feederSocket";  /* addr to connect */



int FeederConnection::startFeeder() { 
      struct sockaddr_un unixSocket;

      if(sockfd == -1)
        if((sockfd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0) {
         perror("socket");
         return -1;
      }
	
      unixSocket.sun_family = AF_UNIX;
      strcpy(unixSocket.sun_path,SOCK_ADDRESS);

      if (connect(sockfd, (const sockaddr*)&unixSocket, sizeof(unixSocket)) < 0) {
         return -1;
      }

      unlink(SOCK_ADDRESS); // in case it terminates or crashes

      if(allocateMarshallingBuffer() == 0) {
        closeConnection();
        fprintf(stderr,"FeederConnection: allocateMarshallingBuffer() couldn't allocate memory buffer used for serialization.\n");
        return -1;
      }

      // initialize socket polling struct; we will use it when callin poll()
      pollSocket.fd = sockfd;
      pollSocket.events = POLLIN;

      return sockfd;
}

unsigned int FeederConnection::allocateMarshallingBuffer() {
  /* we really don't the size of an event because the contact name can be any lenght,
     so here we get a 'big enough' buffer */
  enum { MAX_PACKET_SIZE = 10240 };
  // destroy all buffer, if any
  destroyMarshallingBuffer();
  marshallingBuffer  = new char[sizeof(int)+MAX_PACKET_SIZE];
  if(marshallingBuffer != NULL)
    bufferSize = sizeof(int)+MAX_PACKET_SIZE;
  else bufferSize = 0;
  return bufferSize;
}

void FeederConnection::destroyMarshallingBuffer() {
  if(marshallingBuffer != NULL) {
	delete[] marshallingBuffer;
	marshallingBuffer = NULL;
        bufferSize = 0;
  }
}

int FeederConnection::connectToFeeder() {
  struct sockaddr_un my_addr, their_addr;

  if(sockfd != -1) {
    fprintf(stderr,"ERROR: FeederConnection already connected, second call is ignored.\n");
    return -1;
  }

  if((sockfd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
    perror("connectToFeeder(): socket");
    return -1;
  }

  memset(&my_addr,0,sizeof(my_addr));
  my_addr.sun_family = AF_UNIX;
  strcpy(my_addr.sun_path, SOCK_ADDRESS);

  unlink(SOCK_ADDRESS); /* unlink in case it already exists */

  int len = sizeof(my_addr.sun_family) + strlen(my_addr.sun_path);
  if (bind(sockfd, (const sockaddr*)&my_addr, len) < 0) {
    printf("ERROR: Couldn't open connection to listen for data.\nProbable cause: there was a file %s already.\nCheck if the file %s exits; if so, delete it and start again.\n\n",SOCK_ADDRESS,SOCK_ADDRESS);
    return -1;
  }

  if (listen(sockfd, 5) < 0) {
     perror("connectToFeeder(): listen");
     return -1;
  }

  if(allocateMarshallingBuffer() == 0) {
     fprintf(stderr,"FeederConnection: allocateMarshallingBuffer() couldn't allocate memory\n");
     return -1;
  }

  int acceptedFileDescriptor;
#ifdef __sgi  
  int fromlen = sizeof(struct sockaddr_un);
#else
  socklen_t fromlen = sizeof(struct sockaddr_un);	
#endif    
  if ((acceptedFileDescriptor = accept(sockfd, (sockaddr*)&their_addr, &fromlen)) < 0) {
    perror("connectToFeeder(): accept");
    return -1;
  } else { // if we are here, then everything went OK
    close(sockfd);
    sockfd = acceptedFileDescriptor;
    // initialize socket polling struct; we will use it when callin poll()
    pollSocket.fd = sockfd;
    pollSocket.events = POLLIN;
    return sockfd;
  }
}

int FeederConnection::sendEvent(const ContactEvent& evt) const {
   // first 2 bytes are payload size (excluding the 2 bytes)
   unsigned int packet_size = evt.packTo(marshallingBuffer+sizeof(packet_size));
   memcpy(marshallingBuffer,&packet_size,sizeof(packet_size));
   return write(sockfd, marshallingBuffer, sizeof(packet_size)+ packet_size);
}

bool FeederConnection::isDataAvailable() {
     return poll(&pollSocket,1,5) > 0;
}

int FeederConnection::readEvent(ContactEvent& evt) {
      int packet_size, bytes_read;
      // read packet size (variable) in first 2 bytes. Packet size does not count these 2 bytes
      bytes_read = recv(sockfd, &packet_size, sizeof(packet_size), MSG_DONTWAIT);
      if(bytes_read == -1) return -1;
      if(bytes_read != 0) {
        // truncate packet length if needed, to avoid buffer overrun
        if(packet_size > bufferSize) packet_size = bufferSize;
        // read packet load
        bytes_read = recv(sockfd, marshallingBuffer, packet_size, MSG_DONTWAIT) ;
        evt.unpackFrom(marshallingBuffer);
      }
      return bytes_read;
}

int FeederConnection::closeConnection() {
      if(sockfd != -1) {
        int result = close(sockfd);
        unlink(SOCK_ADDRESS); //we need to do this because we use a PF_UNIX socket type
        sockfd = -1;
        return result;
      }
      else return 0;
}

bool FeederConnection::sendByte(unsigned char aByte) const {
  return write(sockfd,&aByte,sizeof(aByte)) == sizeof(aByte);
}

bool FeederConnection::readByte(unsigned char& aByte) {
  return recv(sockfd,&aByte,sizeof(unsigned char),MSG_DONTWAIT) == sizeof(unsigned char);
}


FeederConnection::~FeederConnection() {
  closeConnection();
  destroyMarshallingBuffer();
};
