#include "IvColorizeGrayscalesAction.h"
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoMaterial.h>



   void  IvColorizeGrayscalesAction::colorizeSoMFColorField(SoMFColor& mfgrayscale, const float* hsvNewColor) {
      int i;
      float *color;
      SbColor *colors = mfgrayscale.startEditing();
      color = (float*)colors->getValue();
   /* Here we are using a cheap-and-fast trick to detect 
     when the current node is a node that we already processed, and 
     that happens to be referenced more than once in the scene graph.
     Since this algorithm generates only 3 possible types of HSV triplets:
     (hue,1,0), (hue,1,1), and (hue,1,0), and hue is always in the [0,1) range,
     any node with whose color has r=g=b is a node that we didn't process yet.
     If r!=g or r!=b or g!=b, then we already transformed the gray levels 
     in this node, and so we skip the node. */
      if (color[0]== color[1] && color[1]== color[2]) {
         if (*color < 0.5) {
            colors->setHSVValue(hsvNewColor[0],1.0f,2*color[0]);
         } 
         else {
            colors->setHSVValue(hsvNewColor[0],color[0],1.0f);
         }
         for(i=1; i<mfgrayscale.getNum(); i++) {
            colors++;
            color = (float*)colors->getValue();
            if (*color < 0.5) {
               colors->setHSVValue(hsvNewColor[0],1.0f,2*color[0]);
            }
            else {
               colors->setHSVValue(hsvNewColor[0],color[0],1.0f);
            }
         }
      }
      mfgrayscale.finishEditing();
   }

   void IvColorizeGrayscalesAction::changeHueInSoMFColorField(SoMFColor& mfcolor, const float* newHue) {
      int i;
      float h,s,v;
      SbColor *colors = mfcolor.startEditing();
   /* Here we change ONLY the hue of the color, assuming that
     the Saturation and Value components are correct. */
      for(i=0; i<mfcolor.getNum(); i++) {  
         colors->getHSVValue(h,s,v);
         if(h != *newHue) {
            colors->setHSVValue(*newHue,s,v);
         }
         colors++;
      }
      mfcolor.finishEditing();
   }

   SoCallbackAction::Response IvColorizeGrayscalesAction::changeMaterialFromGrayscaleToColor(void* hsvColor, SoCallbackAction*, const SoNode* mat) {
      colorizeSoMFColorField(((SoMaterial*)mat)->ambientColor,(float*)hsvColor);
      colorizeSoMFColorField(((SoMaterial*)mat)->diffuseColor,(float*)hsvColor);
      colorizeSoMFColorField(((SoMaterial*)mat)->specularColor,(float*)hsvColor);
      colorizeSoMFColorField(((SoMaterial*)mat)->emissiveColor,(float*)hsvColor);
      return SoCallbackAction::CONTINUE;
   }

   SoCallbackAction::Response IvColorizeGrayscalesAction::changeBaseFromGrayscaleToColor(void* hsvColor,SoCallbackAction*, const SoNode* baseColor) {
      colorizeSoMFColorField(((SoBaseColor*)baseColor)->rgb,(float*)hsvColor);
      return SoCallbackAction::CONTINUE;
   }

   SoCallbackAction::Response IvColorizeGrayscalesAction::changeMaterialToNewHue(void* newHue, SoCallbackAction*, const SoNode* mat) {
      changeHueInSoMFColorField(((SoMaterial*)mat)->ambientColor,(float*)newHue);
      changeHueInSoMFColorField(((SoMaterial*)mat)->diffuseColor,(float*)newHue);
      changeHueInSoMFColorField(((SoMaterial*)mat)->specularColor,(float*)newHue);
      changeHueInSoMFColorField(((SoMaterial*)mat)->emissiveColor,(float*)newHue);
      return SoCallbackAction::CONTINUE;
   }


   SoCallbackAction::Response IvColorizeGrayscalesAction::changeBaseColorToNewHue(void* newHue,SoCallbackAction*, const SoNode* baseColor) {
      changeHueInSoMFColorField(((SoBaseColor*)baseColor)->rgb,(float*)newHue);
      return SoCallbackAction::CONTINUE;
   }

   void IvColorizeGrayscalesAction::setCallbacksToWorkWithGrays() {
      colorMode = GRAY_TO_COLOR;
      addPostCallback(SoBaseColor::getClassTypeId(),changeBaseFromGrayscaleToColor,this->hsvColor);
      addPostCallback(SoMaterial::getClassTypeId(),changeMaterialFromGrayscaleToColor,this->hsvColor);
   }

   void IvColorizeGrayscalesAction::setCallbacksToWorkWithColors() {
      colorMode = COLOR_TO_COLOR;
      addPostCallback(SoBaseColor::getClassTypeId(),changeBaseColorToNewHue,this->hsvColor);
      addPostCallback(SoMaterial::getClassTypeId(),changeMaterialToNewHue,this->hsvColor);
   }


   IvColorizeGrayscalesAction::IvColorizeGrayscalesAction():SoCallbackAction() {
      setCallbacksToWorkWithGrays();
   }

   void IvColorizeGrayscalesAction::applyRGBColor(SoNode *node, const SbColor& rgbNewColor) {
      rgbNewColor.getHSVValue(hsvColor);
      if(hsvColor[0]>=1.0f)
         hsvColor[0] -= 1.0f; // guarantees hue is in the [0,1) range
      if(colorMode != GRAY_TO_COLOR)
         setCallbacksToWorkWithGrays();
      SoCallbackAction::apply(node);
   }

   void IvColorizeGrayscalesAction::applyHSVColor(SoNode *node, float h, float s, float v) {
      hsvColor[0] = h;
      if(hsvColor[0]>=1.0f)
         hsvColor[0] -= 1.0f; // guarantees hue is in the [0,1) range
      hsvColor[1] = s;
      hsvColor[2] = v;
      if(colorMode != GRAY_TO_COLOR)
         setCallbacksToWorkWithGrays();
      SoCallbackAction::apply(node);
   }

   void IvColorizeGrayscalesAction::changeColorHue(SoNode *node, float newHueValue) {
      hsvColor[0]=newHueValue;
      if(colorMode != COLOR_TO_COLOR)
         setCallbacksToWorkWithColors();
      SoCallbackAction::apply(node);
   }

