#ifndef COLORIZE_GRAYS_H
#define COLORIZE_GRAYS_H

#include <Inventor/actions/SoCallbackAction.h>

/*! This class is an Inventor action that will change
    any SoBaseColor or Material node color from grayscale
    to a gradient of another color.
    This object gives a gray object a color in a reduced color space:
    all colors are either HSV(hue,saturation,1), HSV(hue,1,value). 
    If a node has already been 'colored' using this action, then
    you can use changeColorHue() to change the color without
    affecting the Saturation and Value components. In this
    way changing color many times will not affect the 
    'brightness'  of the object.
    applyHSVColor() and applyRGBColor() will only work with graylevels,
    they will skip any color node that is not in grays. */
   class IvColorizeGrayscalesAction: public SoCallbackAction {
   private:
      virtual void apply(SoNode*) { 
      };
      virtual void apply(SoPath*) { 
      };
      virtual void apply(const SoPathList &, SbBool) { 
      };
   protected:
      typedef enum { GRAY_TO_COLOR, COLOR_TO_COLOR } ColorizingMode;
      ColorizingMode colorMode;
      float hsvColor[3];
      static void colorizeSoMFColorField(SoMFColor& mfcolor, const float* newcolor);
      static SoCallbackAction::Response changeMaterialFromGrayscaleToColor(void* newColor,SoCallbackAction*, const SoNode* mat);
      static SoCallbackAction::Response changeBaseFromGrayscaleToColor(void* newColor,SoCallbackAction*, const SoNode* baseColor);
      static void changeHueInSoMFColorField(SoMFColor& mfcolor, const float* newHue);
      static SoCallbackAction::Response changeMaterialToNewHue(void* newHue,SoCallbackAction*, const SoNode* mat);
      static SoCallbackAction::Response changeBaseColorToNewHue(void* newHue,SoCallbackAction*, const SoNode* baseColor);
   	void setCallbacksToWorkWithGrays();
   	void setCallbacksToWorkWithColors();
   public:
      IvColorizeGrayscalesAction();
      /*! Goes through all the color nodes that are children of 'node', and changes
          the color of the gray ones using the (h,s,v) parameters. Currently 
          only h is used, s and v are ignored. */
      virtual void applyHSVColor(SoNode *node, float h, float s=0, float v=0);
    /*! Goes through all the color nodes that are children of 'node', and changes
          the color of the gray ones to the hue value equivalent to the (r,g,b) color
          passed as parameters. */  
      virtual void applyRGBColor(SoNode *node, const SbColor& rgbNewColor);
    /*! This function is meant to be called once an object has been colored
       using applyHSVColor() or applyRGBColor() to change the Hue without
       changing the S and V in the (H,S,V) color representation. */
      virtual void changeColorHue(SoNode *node, float newHueValue);
   };

#endif
