#include "IvTiledBathymetry.h"
#include <string.h>
#include <strings.h>
#include <pthread.h>
#include <Inventor/SoDB.h>
#include <Inventor/SoInput.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoGroup.h>
#include "BoundedQueue.h"

#include <Inventor/nodes/SoCallback.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/elements/SoCacheElement.h>

#include <sys/time.h>

#include <GL/gl.h>


/* HACK-HACK-HACK, this is a temporary filler till the MapTile's bbox is truly 3D */
#define BBOX_Z_PLACEHOLDER 0

/* if DRAW_MISSING_TILE_MARKER is defined, then a missing tile will be marked with a stippled pattern */
// #define DRAW_MISSING_TILE_MARKER 

/*****************  Box3dFloat ****************/

   bool Box3dFloat::overlaps(const Rectangle2dFloat& rectangle2) {
      return !(rectangle2.leftBottom.x > getMax()[0] ||
         rectangle2.rightTop.x < getMin()[0] ||
         rectangle2.leftBottom.y > getMax()[1] ||
         rectangle2.rightTop.y < getMin()[1]) ;
   }


/*****************  BathymetryTileLoaderThread *****************/


/*!  BathymetryTileLoaderThread is a class that encapsulates the creation and handling
     of a POSIX thread to asynchronously load map tiles (Open Inventor objects).
    This class is internally used by the class IvTiledBathymetry.
    Requests to load tiles are received by calling requestTiles(). Every time
    a tile is loaded, the callback set by the application with setTileLoadedNotifyCallback()
    will be called.
    It is possible to cancel the loading or some or all the tiles that have not
    been loaded yet by calling BathymetryTileLoaderThread::cancelAllRequestedTiles().
 */ 
   struct BathymetryTileLoaderThread {
   protected:
      /// mutex used to protect the pending requests queue
      pthread_mutex_t in_sync;
      /// mutex used to protect the processed requests queue
      pthread_mutex_t out_sync;
      /// mutex used to protect to control the way we finish the thread   	   	
      pthread_mutex_t quit_mutex;
   	/// condition variable to signal when the inbound queue has a something to process
      pthread_cond_t inqueue_not_empty;
    	/// condition variable to signal when the inbound queue has space to store more loaded tiles
      pthread_cond_t outqueue_not_full;
   	/// This is the prefix (if any) of all tile filenames
      char *TILE_PREFIX;
      bool quit;
      bool thread_is_dead;
   	/// This struct is only used inside BathymetryTileLoaderThread, and is what we store in inqueue, the pending requests queue
      struct InQueueItem {
         TileKeyRange range;
         void *tileUserInfo;
      };
   
      /// This is a queue we use to store the client's requests to load tiles till we can process them
      BoundedQueue inqueue;
   
     /// This struct is only used inside BathymetryTileLoaderThread, and is what we store in outqueue, the processed requests queue 
      struct OutQueueItem {
         SoNode* tile;
         void *tileUserInfo;
      };
   	/// This is the queue where we put all the tiles we have succesfully loaded.
      BoundedQueue outqueue;
      /// thread function 
      static void *loader(void* self);
   public:
      /// Type of callback the application can set to be called every time a tile is loaded. 
      typedef void(*TileLoadedCallback)(void* callbackArg, const TileKeyRange& range, unsigned int tileRangeIndex, bool success);
      /// Type of callback the application can set to be called when a tile range is cancelled
      typedef void(*RangeCancelledCallback)(void* callbackArg, const TileKeyRange& range, void* rangeUserDara);
      /// thread id
      pthread_t id;
   protected:
      TileLoadedCallback loadNotifyCallback;
      void *loadNotifierArg;
      RangeCancelledCallback cancelNotifyCallback;
      void *cancelNotifierArg;
      bool cancelLoading;
   public:
      /*! The constructor creates the thread.
          \param queueSize		Maximum number of tiles that can be simultaneously requested.
          \param tileFilenamePrefix	An optional prefix that all tile filenames have at the beginning.
                                    The tiles are assummed to be named tileFilenamePrefix+tile_number+".iv" .
      */
      BathymetryTileLoaderThread(unsigned int queueSize,const char* tileFilenamePrefix);
   	/*! Changes the maximum number of pending tile requests that can be stored, and cancels all current requests. */
      void changeInputQueueSize(unsigned int size);
   	/*! Changes the maximum number of loaded tiles that can be stored. 
   	   The size of the output queue must always be at least the size of the input queue. */
      void changeOutputQueueSize(unsigned int size);
   	/*! Sets a callback to be called by the thread every times it tries to load a tile.
   	   The callback must be of type  BathymetryTileLoaderThread::TileLoadedCallback.
   	   \param  cb	Pointer to the function to be called
   		\param  arg The argument passed to requestTiles() when the tile range was requested.
   		Once the tile is loaded (or not, if not found), then range passed 
   		in requestTiles() and the value (inside range) of the tile are passed as parameters
   		to the callback. The last parameter of the callback will be true if the tile was succesfully loaded.
   	*/	
      void setTileLoadedNotifyCallback(TileLoadedCallback cb, void* arg) {
         loadNotifyCallback = cb; loadNotifierArg = arg; }
   	/*! Sets a callback that will be called every time a tile range (or part of it) that
   	   was requested with requestTiles() is cancelled, because the client called one of the
   	   cancel... methods, or because some other method inside the thread has cancelled that range.
   	   \param cb	The callback to be called when a tile range is cancelled
   		\param arg	The argument passed to requestTiles() when the tile range was requested.
   	*/	
      void setRangeCancelledNotifyCallback(RangeCancelledCallback cb, void* arg) {
         cancelNotifyCallback = cb; cancelNotifierArg = arg; }
   	/*! Requests a tile range to be asynchronously loaded. This call will return immediately.
   	    \param tiles	One or more tiles to load
   		 \param someUserTileInfo	Optional information to pass to the callbacks when the tile is loaded, or the tile range is cancelled.
   		<br>See also:
           \ref BathymetryTileLoaderThread::setTileLoadedNotifyCallback()
    		  \ref BathymetryTileLoaderThread::setRangeCancelledNotifyCallback()
         \return true if the tile range was scheduled to be loaded, or
                 false if the pending requests queue if full.
      */
      bool requestTiles(const TileKeyRange& tiles,void* const someUserTileInfo);
   	/*! Cancels the loading all pending tile ranges. Will discard all tiles partially
   	    loaded ranges that were not picked yet by the application by calling
   	    getNextLoadedTile().
   	*/
      void cancelAllRequestedTiles();
   	/*! Retrieves the next loaded tile, in the order in which they were requested.
   	    \param tile	A pointer to a SoNode* where to put the tile (inventor object) loaded.
   	    \param userInfo	A pointer to the void* passed as parameter in requestTiles() for the range the tile belongs.
   	    <br>
   	    Both  tile and userInfo will be NULL if no tile is available.
   	*/
      void getNextLoadedTile(SoNode** tile, void** userInfo);
   	/*! Kills the thread, and discards all loaded tiles that were not picked by calling getNextLoadedTile().
            \param	timeout_in_secs Maximum time to wait for the thread to die "nicely", before we just kill it.
        */
      void die(unsigned int timeout_in_secs=UINT_MAX);
      ~BathymetryTileLoaderThread() { 
         if(!quit) die(4); }
   };



/**************  Class BathymetryTileLoaderThread **************/

   BathymetryTileLoaderThread::BathymetryTileLoaderThread(unsigned int queueSize, const char* tileFilenamePrefix):quit(false),thread_is_dead(false),inqueue(sizeof(InQueueItem),queueSize),outqueue(sizeof(OutQueueItem),queueSize),
   loadNotifyCallback(NULL),cancelNotifyCallback(NULL) {
      if(tileFilenamePrefix!=NULL) {	
        // make a copy of the prefix - it will be discarded by loader, which keeps a local copy.
         TILE_PREFIX=new char[strlen(tileFilenamePrefix)+1];	
         strcpy(TILE_PREFIX,tileFilenamePrefix);
      } 
      else {
         TILE_PREFIX=NULL;
      }
   	// Create the thread that will do the loading
      pthread_mutex_init(&in_sync,NULL);
      pthread_mutex_init(&out_sync,NULL);
      pthread_cond_init(&inqueue_not_empty,NULL);
      pthread_cond_init(&outqueue_not_full,NULL);
      pthread_mutex_init(&quit_mutex,NULL);
      pthread_attr_t attr;
      pthread_attr_init(&attr);
      pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);
      pthread_create(&id,&attr,BathymetryTileLoaderThread::loader,(void*)this);  
   }

// get the ascii representation of an unsigned int. Returns a pointer to the end of the string
   char* uitoa(unsigned int i,char *str) {
   	// get digits in reverse order
      char *chrptr=str;	
      do {
         *chrptr=i%10+'0';
         i/=10;
         ++chrptr;
      } while(i>0);
      *chrptr = '\0'; // NULL-terminate string
   	// reverse string to get digits in right order
      char *p2 = chrptr-1, c;
      while(p2>str) {
         c = *p2;
         *p2 = *str;
         *str = c;
         --p2;
         ++str;
      }
      return chrptr;
   }

/* This is the function that does the tile loading. It retrieves pending tile ranges
   from inqueue, and tries to load all the tiles in the range. Every time it loads
   a tile it puts the new tile in outqueue. After it loaded a tile or failed to do so,
   it will call the callback stored in loadNotifyCallback if it is not NULL.
*/
   void*  BathymetryTileLoaderThread::loader(void* myself) {
      char path[PATH_MAX];
      unsigned int LEN_TILE_PREFIX;
      SoInput file;
      SoNode *tile;
      InQueueItem currentRange;
      OutQueueItem loadedTile;
      BathymetryTileLoaderThread *self = (BathymetryTileLoaderThread*)myself;
      TILE_KEY_NUMBER nextTile = MAX_TILE_KEY_NUMBER;
      currentRange.range.end = MIN_TILE_KEY_NUMBER;
      self->cancelLoading = false;
      if(self->TILE_PREFIX!=NULL) {
         strncpy(path,self->TILE_PREFIX,PATH_MAX-10);
         LEN_TILE_PREFIX=strlen(path);
         delete[] self->TILE_PREFIX;
       	// Discard the filename prefix, the only copy we need we keep it in path      	
         self->TILE_PREFIX=NULL;
      }
      else {
         LEN_TILE_PREFIX=0;
      }
   
      /* main loop */
      while(!self->quit) {
         if(nextTile > currentRange.range.end) {
            pthread_mutex_lock(&self->in_sync);
            if(self->inqueue.isEmpty()) {
               pthread_cond_wait(&self->inqueue_not_empty,&self->in_sync);
            }
             /* when the thread is awaken, the 'in' queue can still be empty, because
               it may have been awaken by die(), which calls cancelAllRequestedTiles()
               before calling pthread_cond_signal(). That's why we check again here.	*/
            if(self->inqueue.isEmpty()) {
               continue;
            }
            self->inqueue.peek(&currentRange);
            pthread_mutex_unlock(&self->in_sync);
            nextTile = currentRange.range.begin;
         }
         strcpy(uitoa(nextTile,path+LEN_TILE_PREFIX),".iv");
      
      
         if(file.openFile(path) != FALSE) {
            if(SoDB::read(&file,tile) == FALSE)
               tile=NULL;
            file.closeFile();
         } 
         else {
            tile=NULL;
         }
         if(!self->cancelLoading) {
            if(tile != NULL) {
               pthread_mutex_lock(&self->out_sync);
               if(self->outqueue.isFull()) {
                  pthread_cond_wait(&self->outqueue_not_full,&self->out_sync);
               }
               loadedTile.tile = tile;
               loadedTile.tileUserInfo = currentRange.tileUserInfo;
               self->outqueue.push(&loadedTile);
               pthread_mutex_unlock(&self->out_sync);
            } 
            if(self->loadNotifyCallback != NULL) {
               (*self->loadNotifyCallback)(self->loadNotifierArg,currentRange.range,nextTile,tile!=NULL);
            } 
            nextTile++;
            if(nextTile > currentRange.range.end) {
               pthread_mutex_lock(&self->in_sync);
               self->inqueue.skip();
               pthread_mutex_unlock(&self->in_sync);
            }
         } 
         else {
            if(tile != NULL) tile->unref();
            nextTile = MAX_TILE_KEY_NUMBER;
            self->cancelLoading = false;
         }
      }
      self->thread_is_dead = true;
      return NULL;
   }


   bool BathymetryTileLoaderThread::requestTiles( const TileKeyRange& tiles, void* const someUserTileInfo) {
      pthread_mutex_lock(&in_sync);
      if(!inqueue.isFull()) {
         InQueueItem newRange;
         newRange.range = tiles;
         newRange.tileUserInfo = someUserTileInfo;
         inqueue.push(&newRange);
         pthread_mutex_unlock(&in_sync);
         pthread_cond_signal(&inqueue_not_empty);
         return true;
      }
      else {
         pthread_mutex_unlock(&in_sync);
         return false;
      }
   }

   void BathymetryTileLoaderThread::cancelAllRequestedTiles() {
      InQueueItem item;
      pthread_mutex_lock(&in_sync);
      cancelLoading = true;
   	// remove all requests from inqueue
      while(!inqueue.isEmpty()) {
         inqueue.pop(&item);
         if(cancelNotifyCallback != NULL)
            (*cancelNotifyCallback)(cancelNotifierArg,item.range,item.tileUserInfo);
      }
      pthread_mutex_unlock(&in_sync);
      // clean up outqueue
      pthread_mutex_lock(&out_sync);
      while(!outqueue.isEmpty()) {
         if(((OutQueueItem*)outqueue.head)->tile != NULL) ((OutQueueItem*)outqueue.head)->tile->unref();
         outqueue.skip();
      }
      pthread_mutex_unlock(&out_sync);
      pthread_cond_signal(&outqueue_not_full);
   }

	/// Get the next tile item from the loaded tiles queue
   void BathymetryTileLoaderThread::getNextLoadedTile(SoNode** tile, void** userInfo) {
      OutQueueItem item;
      pthread_mutex_lock(&out_sync);
      if(!outqueue.isEmpty()) {
         outqueue.pop(&item);
         *tile = item.tile;
         *userInfo = item.tileUserInfo;
      }
      else {
         *tile = NULL;
         *userInfo = NULL;
      }
      pthread_mutex_unlock(&out_sync);
      pthread_cond_signal(&outqueue_not_full);
   }

   void BathymetryTileLoaderThread::die(unsigned int timeout_in_secs) {
      struct timeval start_time, now;
      if(!quit) {
      /* aquire a "quit" lock, to be sure that
         not two threads try to kill the worker thread at the same time */
         pthread_mutex_lock(&quit_mutex);
         // check again, just in case there is/was more than one thread executing die()
         if(!quit) {
            quit = true;
            cancelAllRequestedTiles();
            // wake up the loading thread, just in case it's waiting
            pthread_cond_signal(&inqueue_not_empty);
             // spin-lock, waiting for the thread to die
            /* here we wait up to N secs for the other thread to die "nicely".
               If it's still stuck doing something after N secs, we just kill it. */
            gettimeofday(&start_time,NULL);
            do {
               gettimeofday(&now,NULL);
            } while(!thread_is_dead && now.tv_sec-start_time.tv_sec < timeout_in_secs);
            if(!thread_is_dead) {
               pthread_cancel(id);
            }
            // clean up
            pthread_cond_destroy(&inqueue_not_empty);
            pthread_cond_destroy(&outqueue_not_full);
            pthread_mutex_destroy(&in_sync);
            pthread_mutex_destroy(&out_sync);
         
         }
         pthread_mutex_unlock(&quit_mutex);
      }
   }

   void BathymetryTileLoaderThread::changeInputQueueSize(unsigned int size) {
      inqueue.resizeTo(size);
   }

   void BathymetryTileLoaderThread::changeOutputQueueSize(unsigned int size) {
      outqueue.resizeTo(size);
   }

/**************  Class IvTiledBathymetry **************/

   SO_NODE_SOURCE(IvTiledBathymetry);

   void IvTiledBathymetry::initClass() {
   // Initialize type id variables
      SO_NODE_INIT_CLASS(IvTiledBathymetry, SoGroup, "Group");
   }

   void IvTiledBathymetry::AnotherTileIsReady(void* myself, const TileKeyRange& range, unsigned int currentTileKey, bool succesfullyLoaded) {
      MapTile::pointer theCell;
      IvTiledBathymetry* self = (IvTiledBathymetry*)myself;
      if(succesfullyLoaded) {
         SoNode* theTileNode;
         MapTile::pointer* theMapTileArray;
         self->workerThread->getNextLoadedTile(&theTileNode,(void**)&theMapTileArray);
         // get the tile plus the "extra info" array (bounding boxes in the MapTile struct)
         theCell = theMapTileArray[currentTileKey-range.begin];
         theCell->tile = theTileNode;
         theTileNode->ref();
      
      //prevent this node from being deleted until setTileColorscaleDepth returns by incrementing ref count
         theTileNode->ref();
         self->setTileColorscaleDepth(theTileNode);//bring about to current colorscale
         theTileNode->unref();
      
        /* The tile's bounding box was already calculated by IvTiledBathymetry::requestTiles(),
           so here we don't bother calculating it again.
        */
         if(self->loadNotifyCallback != NULL) {
            (*self->loadNotifyCallback)(theCell,self->loadNotifierArg);
         }
      }
      else {
         /* when the tile bathymetry couldn't be loaded, we fill the place with a tile
       that indicates we don't have a bathymetry for this tile */
         Point2dUint pos = self->worldMapTileToWorldColRow(currentTileKey);
         pos.x-=self->leftBottomVisibleTileCoord.x-1;
         pos.y-=self->leftBottomVisibleTileCoord.y-1;
         theCell = self->tiles[pos.y*self->cachedAreaSize.x+pos.x];
          #ifdef DRAW_MISSING_TILE_MARKER 
         theCell->tile = new SoCallback;
         ((SoCallback*)theCell->tile)->setCallback(IvTiledBathymetry::drawMissingTile,(void*)&(theCell->bbox));
         theCell->tile->ref();
         #else
         theCell->tile = NULL;
	 #endif
      }

     #ifdef BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP
      /* if we already got all the tiles in a range, then throw away
         the array of MapTiles for that range, that was allocated in requestTiles() */
      if(currentTileKey == range.end) {
        delete[] (MapTile::pointer*)theMapTileArray;
      }
     #endif
      //if its not already there, add this tile to the scenegraph if its in the visible area
      if (theCell) {
         if (theCell->bbox.overlaps(self->worldMapBBox)) {
            if (theCell->tile) {
               self->addChild(theCell->tile);
               //force a redraw
               self->touch();
            }
         }
      }
   }

#ifdef BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP

   void IvTiledBathymetry::DiscardMapTileArray(void*, const TileKeyRange&, void* theMapTileArray) {
      delete[] (MapTile::pointer*)theMapTileArray;
   }

#endif

// Initializes all the values needes to load and handle the tiles, and starts the loading thread.
   void IvTiledBathymetry::initInstance(const Rectangle2dFloat& mapArea, const Point2dFloat& theTileSize, const Point2dUint& visibleAreaInColRows,
   float leftSlack, float bottomSlack,
   unsigned int plusOutskirtRows, unsigned int plusOutskirtCols,
   const char* tileFilenamePrefix) {
      drawEdge=true;
      edgeColor[0]=edgeColor[1]=edgeColor[2]=1.0f;
      worldMapBBox=mapArea;
      tileSize=theTileSize;
      returnFastBBox=false;
      visibleAreaSize=visibleAreaInColRows;
      outskirtSize.y=plusOutskirtCols;
      outskirtSize.x=plusOutskirtRows;
      cachedAreaSize.y=visibleAreaSize.y+plusOutskirtRows+plusOutskirtRows;
      cachedAreaSize.x=visibleAreaSize.x+plusOutskirtCols+plusOutskirtCols;
      visibleArea.setValue(plusOutskirtCols,plusOutskirtRows,plusOutskirtCols+visibleAreaSize.x-1,plusOutskirtRows+visibleAreaSize.y-1);
    // calculate # of complete tiles that fit in the map after removing slack
      if(leftSlack > 0) {
         worldMapBBox.leftBottom.x -= tileSize.y-leftSlack;
      }
      if(bottomSlack > 0) {
         worldMapBBox.leftBottom.y -= tileSize.y-bottomSlack;
      }
      worldMapSize.x = (unsigned int)floor((worldMapBBox.rightTop.x - worldMapBBox.leftBottom.x)/tileSize.x);
      worldMapSize.y = (unsigned int)floor((worldMapBBox.rightTop.y - worldMapBBox.leftBottom.y)/tileSize.y);
   
      float slack = worldMapBBox.rightTop.x-worldMapBBox.leftBottom.x-worldMapSize.x*tileSize.x;
      if(slack > 0.000001) {
         worldMapBBox.rightTop.x += tileSize.x-slack;
         worldMapSize.x++;
      }
      slack = worldMapBBox.rightTop.y-worldMapBBox.leftBottom.y-worldMapSize.y*tileSize.y;
      if(slack > 0.000001) {
         worldMapBBox.rightTop.y += tileSize.y-slack;
         worldMapSize.y++;
      }
      unsigned int size = cachedAreaSize.y*cachedAreaSize.x;
      tiles = (MapTile::pointer*)malloc(sizeof(MapTile::pointer)*size);
      for(unsigned int i=0; i<size; i++)
         tiles[i]=new MapTile;
      backTileBuffer = (MapTile::pointer*)malloc(sizeof(MapTile::pointer)*size);
      bzero(backTileBuffer,sizeof(MapTile::pointer)*size);
      loadNotifyCallback = NULL;
      offsetLeftBottomVisibleArea=visibleArea.leftBottom.y*cachedAreaSize.x+visibleArea.leftBottom.x;
      // start async tile loading thread
      workerThread =  new BathymetryTileLoaderThread(size,tileFilenamePrefix);
      workerThread->setTileLoadedNotifyCallback(IvTiledBathymetry::AnotherTileIsReady,this);
   #ifdef BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP
      workerThread->setRangeCancelledNotifyCallback(IvTiledBathymetry::DiscardMapTileArray,this);
   #endif
   }


   IvTiledBathymetry::IvTiledBathymetry():tiles(NULL),backTileBuffer(NULL),
   visibleArea(0,0,0,0),visibleAreaSize(0,0),returnFastBBox(false),leftBottomVisibleTileCoord(0,0),
   visibleAreaBBox(0,0,0,0),outskirtSize(0,0),cachedAreaSize(0,0),
   worldMapSize(0,0),worldMapBBox(0,0,0,0), drawEdge(false),
   loadNotifyCallback(NULL),mindepth(0),maxdepth(0)  {
      SO_NODE_CONSTRUCTOR(IvTiledBathymetry);
      /* adds the alternate representation field.
         If a program does not know how to render a node, it looks at the alternate rep field. */
      SO_NODE_ADD_FIELD(alternateRep,(NULL));
   }

   IvTiledBathymetry::IvTiledBathymetry(const Rectangle2dFloat& worldMapArea,unsigned int rows, unsigned int cols, const char* tileFilenamePrefix):tiles(NULL),
   worldMapBBox(worldMapArea),drawEdge(true),worldMapSize(0,0),cachedAreaSize(cols,rows),visibleAreaSize(cols,rows), returnFastBBox(false),
   leftBottomVisibleTileCoord(0,0), loadNotifyCallback(NULL) {
      SO_NODE_CONSTRUCTOR(IvTiledBathymetry);
      /* adds the alternate representation field.
         If a program does not know how to render a node, it looks at the alternate rep field. */
      SO_NODE_ADD_FIELD(alternateRep,(NULL));
      edgeColor[0]=edgeColor[1]=edgeColor[2]=1.0f;
      unsigned int size = rows*cols;
      outskirtSize.y = outskirtSize.x = 0;
      tiles = (MapTile::pointer*)malloc(sizeof(MapTile::pointer)*size);
      bzero(tiles,sizeof(MapTile::pointer)*size);
      backTileBuffer = (MapTile::pointer*)malloc(sizeof(MapTile::pointer)*size);
      bzero(backTileBuffer,sizeof(MapTile::pointer)*size);
      loadNotifyCallback = NULL;
   	// start async tile loading thread
      workerThread =  new BathymetryTileLoaderThread(size,tileFilenamePrefix);
   	// notify me after a tile has been succesfully loaded
      workerThread->setTileLoadedNotifyCallback(IvTiledBathymetry::AnotherTileIsReady,this);
   #ifdef BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP
   	// notify me after a range has been cancelled - we need to dispose some memory there
      workerThread->setRangeCancelledNotifyCallback(IvTiledBathymetry::DiscardMapTileArray,this);
   #endif
   }

   IvTiledBathymetry::IvTiledBathymetry(const Rectangle2dFloat& mapArea, const Point2dFloat& theTileSize,
   unsigned int visibleRows, unsigned int visibleCols,
   unsigned int plusOutskirtRows, unsigned int plusOutskirtCols,
   float leftSlack, float bottomSlack, const char* tileFilenamePrefix) {
      SO_NODE_CONSTRUCTOR(IvTiledBathymetry);
      /* adds the alternate representation field.
         If a program does not know how to render a node, it looks at the alternate rep field. */
      SO_NODE_ADD_FIELD(alternateRep,(NULL));
      initInstance(mapArea,theTileSize,Point2dUint(visibleRows,visibleCols),leftSlack, bottomSlack,plusOutskirtRows,plusOutskirtCols, tileFilenamePrefix);
   }



   IvTiledBathymetry::~IvTiledBathymetry() {
   // wait for worker thread to end
      if(workerThread) {
         delete workerThread;
      }
      if(tiles) {
         free(tiles);
         free(backTileBuffer);
      }
   }

 /*  Resizes the map to hold at most (howManyRows+2*plusOutskirtRows)*(howManyCols+2*plusOutskirtCols) tiles.
     the visibleArea will be set to a size of [howManyVisibleRows,howManyVisibleCols],
     and the map will also cache at most 'plusOutskirtRows' at the top and bottom of the visible area,
     and 'plusOutskirtCols' to the left and right of the visible area.
     This method updates the instance variables outskirtSize.y, outskirtSize.x, cachedAreaSize.y and cachedAreaSize.x.
  */
   void IvTiledBathymetry::resizeTo(unsigned int howManyVisibleRows,unsigned int howManyVisibleCols,
   unsigned int plusOutskirtRows, unsigned int plusOutskirtCols ) {
      outskirtSize.y = plusOutskirtRows;
      outskirtSize.x = plusOutskirtCols;
      resizeTo(howManyVisibleRows,howManyVisibleCols);
   }


 /*! Resizes the map to hold at most (howManyRows+2*this->outskirtSize.y)*(howManyCols+2*this->outskirtSize.x) tiles.
     the visibleArea will be set to a size of [howManyVisibleRows,howManyVisibleCols],
     and the map will also cache at most 'plusOutskirtRows' at the top and bottom of the visible area,
     and 'plusOutskirtCols' to the left and right of the visible area.
     This method updates the instance variables cachedAreaSize.y and cachedAreaSize.x.
  */
   void IvTiledBathymetry::resizeTo(unsigned int howManyVisibleRows,unsigned int howManyVisibleCols) {
      workerThread->cancelAllRequestedTiles();
      MapTile::pointer* otherTiles;
      unsigned int oldAreaSize = cachedAreaSize.y*cachedAreaSize.x;
      cachedAreaSize.y = howManyVisibleRows+outskirtSize.y+outskirtSize.y;
      cachedAreaSize.x = howManyVisibleCols+outskirtSize.x+outskirtSize.x;
      unsigned int newAreaSize = cachedAreaSize.y*cachedAreaSize.x;
      visibleArea.leftBottom.setValue(outskirtSize.x,outskirtSize.y);
      visibleArea.rightTop.setValue(howManyVisibleCols+outskirtSize.x-1,howManyVisibleRows+outskirtSize.y-1);
      offsetLeftBottomVisibleArea=visibleArea.leftBottom.y*cachedAreaSize.x+visibleArea.leftBottom.x;
      if(oldAreaSize > newAreaSize) {
         // new area is smaller; throw away tiles we won't need
         otherTiles = tiles+newAreaSize;
         for(unsigned int i = newAreaSize; i<oldAreaSize; i++) {
            if(*otherTiles) {
               if((*otherTiles)->tile != NULL)
                  (*otherTiles)->tile->unref();
               delete *otherTiles;
               *otherTiles=NULL;
            }
            otherTiles++;
         }
         tiles = (MapTile::pointer*)realloc(tiles,sizeof(MapTile::pointer)*newAreaSize);
         free(backTileBuffer);
         backTileBuffer=(MapTile::pointer*)malloc(sizeof(MapTile::pointer)*newAreaSize);
      }
      else if(oldAreaSize < newAreaSize) {
         // new area is bigger, we need more space
         register unsigned int i;
         // throw away all inventor map tiles in the old map
         for(i = 0; i<oldAreaSize; i++) {
            if(tiles[i]->tile != NULL) {
               tiles[i]->tile->unref();
               tiles[i]->tile = NULL;
            }
         }
         // since the new map is bigger, create the MapTile objects needed for the new tiles
         tiles = (MapTile::pointer*)realloc(tiles,sizeof(MapTile::pointer)*newAreaSize);
         free(backTileBuffer);
         backTileBuffer=(MapTile::pointer*)malloc(sizeof(MapTile::pointer)*newAreaSize);
         otherTiles = tiles+oldAreaSize;
         for(i = oldAreaSize; i<newAreaSize; i++) {
            *otherTiles = new MapTile;
            (*otherTiles)->tile = NULL;
            otherTiles++;
         }
      }
   
      // invalidate known bbox of cached area
      visibleAreaBBox.leftBottom.setValue(0,0);
      visibleAreaBBox.rightTop.setValue(0,0);
   
      workerThread->changeInputQueueSize(newAreaSize);
      workerThread->changeOutputQueueSize(newAreaSize);
   }

   void IvTiledBathymetry::clearTileCache() {
      MapTile::pointer *mapTile = tiles,
         *lastTile = tiles+cachedAreaSize.x*cachedAreaSize.y;
      while(mapTile!=lastTile) {
         if((*mapTile) == NULL)
            *mapTile = new MapTile;
         else if((*mapTile)->tile!=NULL) {
            (*mapTile)->tile->unref();
            (*mapTile)->tile=NULL;
         }
         mapTile++;
      }
   }

   void IvTiledBathymetry::fillVisibleAreaWithTiles(const Rectangle2dFloat& newVisibleArea, unsigned int plusOutskirtRows, unsigned int plusOutskirtCols) {
      MapTile::pointer *mapTiles;
      Point2dUint leftBottomColRow,rightTopColRow;
      Point2dFloat corner;
      TileKeyRange range;
      Rectangle2dFloat constrainedVisibleArea = newVisibleArea;
      // constrainedVisibleArea is the part of the visible area that fits inside the map
      newVisibleArea.isect(worldMapBBox,constrainedVisibleArea);
      /* shrink the top right corner of the visible area just a little bit,
         so we get the calculation of the number of tiles right */
      corner=constrainedVisibleArea.rightTop;
      corner.x-=0.0001;
      corner.y-=0.0001;
      leftBottomColRow=worldPointToWorldColRow(constrainedVisibleArea.leftBottom);
      rightTopColRow=worldPointToWorldColRow(corner);
      // calculate the coordinate of the lower left corner of the constrained visible area
      corner.setValue(worldMapBBox.leftBottom.x+leftBottomColRow.x*tileSize.x,worldMapBBox.leftBottom.y+leftBottomColRow.y*tileSize.y);
      // see if the number of visible rows is still fits in the number of tiles we can hold
      //visibleAreaSize.setValue(rightTopColRow.y-leftBottomColRow.y+1,rightTopColRow.x-leftBottomColRow.x+1);
      visibleAreaSize.setValue(rightTopColRow.x-leftBottomColRow.x+1,rightTopColRow.y-leftBottomColRow.y+1);
   	// if the new total number of tiles is different, then resize the tiles array and update nCachedRows,nCachedCols
      if(cachedAreaSize.y*cachedAreaSize.x != (visibleAreaSize.y+plusOutskirtRows+plusOutskirtRows)*(visibleAreaSize.x+plusOutskirtCols+plusOutskirtCols)) {
         resizeTo(visibleAreaSize.y,visibleAreaSize.x,plusOutskirtRows,plusOutskirtCols);
      } 
      else {
         clearTileCache();
      }
      /* pointer to the leftmost tile in a range, from the array of MapTiles in tiles.
   	   We pass this pointer to requestTiles(), so when the tile is loaded we know where to put it */
      MapTile::pointer *pLeftTile = tiles+visibleArea.leftBottom.y*cachedAreaSize.x+visibleArea.leftBottom.x;
      // calculate bounding box of new visible area
      visibleAreaBBox.setValue(corner.x,corner.y,
         corner.x+visibleAreaSize.x*tileSize.x,corner.y+visibleAreaSize.y*tileSize.y);
      visibleArea.setValue((cachedAreaSize.y-visibleAreaSize.y)/2,(cachedAreaSize.x-visibleAreaSize.x)/2,
         (cachedAreaSize.y+visibleAreaSize.y)/2,(cachedAreaSize.x+visibleAreaSize.x)/2);
      // keep the (col,row) of the bottom left tile in the visible area
      leftBottomVisibleTileCoord=leftBottomColRow;
      // now, fill the visible area with tiles
      float leftedge=corner.x;
      range.begin=worldColRowToWorldMapTile(leftBottomColRow);
      for(register unsigned int i=0; i<visibleAreaSize.y; i++) {
         range.end=range.begin+visibleAreaSize.x-1;
      #ifdef BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP
      	/* Get an array to hold all the MapTiles in range in a known place; 
      	   they can shift around inside the mapTiles array when the map shifts,
      	   so we use this array to hold them where we know they won't move */
         mapTiles = new MapTile::pointer[visibleAreaSize.x];
         memcpy(mapTiles,pLeftTile,visibleAreaSize.x*sizeof(MapTile::pointer));
      #else
         mapTiles=pLeftTile;
      #endif
      	/* Here we recalculate each tile's bounding box and we put them in an array
      	   that we send a a parameter of requestTiles(), so when the tile is loaded
      	   we don't have to find the bbox again.
      	   We do this because is faster to calculate the bboxes here, rathen than
      	   doing it at the receiving end, that will have to scan each Inventor object
      	   to calculate its bounding box.
      	   This ONLY works iff all the tiles are of exactly the same size.
       	   The bbox array will be deleted by  AnotherBathymetryTileIsReady()
            when all tiles in the array have been loaded, or by the BathymetryTileRangeCancelled()
            callback when a whole tile range is cancelled. */
         for(register unsigned j=0; j< visibleAreaSize.x; j++) {
            mapTiles[j]->bbox.getLeftBottom().setValue(corner.x,corner.y,BBOX_Z_PLACEHOLDER);
            mapTiles[j]->bbox.getRightTop().setValue(corner.x+tileSize.x,corner.y+tileSize.y,BBOX_Z_PLACEHOLDER);
            corner.x+=tileSize.x;
         }
      //clear the old tiles
         removeAllChildren();
      //add the new tiles that we already have
         MapTile::pointer *nodeToRender = tiles+offsetLeftBottomVisibleArea;
         unsigned int nextRow = cachedAreaSize.x-visibleAreaSize.x;
         for(register unsigned int i=0;i<visibleAreaSize.y;i++) {
            for(register unsigned int j=0;j<visibleAreaSize.x;j++) {
               if(*nodeToRender != NULL && (*nodeToRender)->tile != NULL) {
                  if (findChild((*nodeToRender)->tile)==-1) {
                     addChild((*nodeToRender)->tile);
                  }
               }
               nodeToRender++;
            } 
            nodeToRender+=nextRow;
         }
      //the rest of the tiles will be added as they come in
         workerThread->requestTiles(range,(void*)mapTiles);
         corner.x=leftedge;
         corner.y+=tileSize.y;
         pLeftTile+=cachedAreaSize.x;
         range.begin+=worldMapSize.x;
      }
   }

   void IvTiledBathymetry::shiftMapBy(int xshift, int yshift) {
   	/* check if we're shifting by an area bigger than the cache. If so, then
   	  simply invalidate the tile cache and fill the visibleArea. */
      if(abs(xshift) >= cachedAreaSize.x || abs(yshift) >= cachedAreaSize.y) {
         workerThread->cancelAllRequestedTiles();
         Rectangle2dFloat r;
         r.moveBy(xshift*tileSize.x,yshift*tileSize.y);
         fillVisibleAreaWithTiles(r);
      }
      else {
         register int i,j,k,l,m, cellPos, xstart, xend, ystart, yend;
         MapTile::pointer *mapTiles;
         TileKeyRange range;
         Rectangle2dInt shiftedCachedArea(-xshift,-yshift,cachedAreaSize.x-xshift-1,cachedAreaSize.y-yshift-1), cachedArea;
         if(shiftedCachedArea.leftBottom.x < 0) shiftedCachedArea.leftBottom.x = 0;
         if(shiftedCachedArea.leftBottom.y < 0) shiftedCachedArea.leftBottom.y = 0;
         if(shiftedCachedArea.rightTop.x >= cachedAreaSize.x) shiftedCachedArea.rightTop.x=cachedAreaSize.x-1;
         if(shiftedCachedArea.rightTop.y >= cachedAreaSize.y) shiftedCachedArea.rightTop.y=cachedAreaSize.y-1;
         cachedArea=shiftedCachedArea;
         cachedArea.moveBy(xshift,yshift);
         Rectangle2dInt tilesToFill[4];
         shiftedCachedArea.gridDiff(cachedArea,tilesToFill);
      
      #ifndef ATYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP
         /* stop loading pending tiles */
         workerThread->cancelAllRequestedTiles();
      #endif
      
         /* Step 1: backup all tiles unique to shiftedCachedArea that will be overwritten after shifting */
      
         l=0;
         for(i=0; i<4; i++) {
            if(tilesToFill[i].isValidRectangle()) {
               for(j=tilesToFill[i].leftBottom.y; j<=tilesToFill[i].rightTop.y; j++) {
                  cellPos = j*cachedAreaSize.x+tilesToFill[i].leftBottom.x;
                  for(k=tilesToFill[i].leftBottom.x; k<=tilesToFill[i].rightTop.x; k++) {
                     backTileBuffer[l++]=tiles[cellPos++];
                  }
               }
            }
         }
      
         /* Step 2: move all tiles that are still inside the cached area after shifting to their new position */
      
         if(xshift >= 0) {
            k=1;
            xstart=shiftedCachedArea.leftBottom.x;
            xend=shiftedCachedArea.rightTop.x+1;
         }
         else {
            k=-1;
            xstart=shiftedCachedArea.rightTop.x;
            xend=shiftedCachedArea.leftBottom.x-1;
         }
         if(yshift >= 0) {
            l=1;
            ystart=shiftedCachedArea.leftBottom.y;
            yend=shiftedCachedArea.rightTop.y+1;
         }
         else {
            l=-1;
            ystart=shiftedCachedArea.rightTop.y;
            yend=shiftedCachedArea.leftBottom.y-1;
         }
      
         for(i=ystart; i!=yend; i+=l) {
            for(j=xstart; j!=xend; j+=k) {
               tiles[i*cachedAreaSize.x+j]=tiles[(i+yshift)*cachedAreaSize.x+j+xshift];
            }
         }
      
      #ifdef BATYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP
         /* stop loading pending tiles */
         workerThread->cancelAllRequestedTiles();
      #endif
      
         /* Step 3: Fill all "empty" tiles with the tiles we kept in step 1 */
      
         Rectangle2dInt freeTiles[4];
          /* Moving the tiles in visible area for shifting is going to leave some empty tiles.
            Find where to get the tiles to fill the empty Area (tilesToFill)
            that will be left after we shift the visible area */
         cachedArea.gridDiff(shiftedCachedArea,freeTiles);
         /* by now, freeTiles[0..3] have the rectangles or
            areas inside cachedArea that hold MapTiles that are free after shifting */
      
         l=0;
         for(i=0; i<4; i++) {
            if(freeTiles[i].isValidRectangle()) {
               for(j=freeTiles[i].leftBottom.y; j<=freeTiles[i].rightTop.y; j++) {
                  cellPos = j*cachedAreaSize.x+freeTiles[i].leftBottom.x;
                  for(k=freeTiles[i].leftBottom.x; k<=freeTiles[i].rightTop.x; k++) {
                     tiles[cellPos++]=backTileBuffer[l++];
                  }
               }
            }
         }
      
         /* step 4: discard inventor objects in all the positions in cachedArea that are 'empty' after shifting */
         cachedArea.setValue(0,0,cachedAreaSize.x-1,cachedAreaSize.y-1);
         cachedArea.gridDiff(shiftedCachedArea,freeTiles);
         for(i=0; i<4; i++) {
            if(freeTiles[i].isValidRectangle()) {
               for(j=freeTiles[i].leftBottom.y; j<=freeTiles[i].rightTop.y; j++) {
                  cellPos = j*cachedAreaSize.x+freeTiles[i].leftBottom.x;
                  for(k=freeTiles[i].leftBottom.x; k<=freeTiles[i].rightTop.x; k++) {
                     if(tiles[cellPos]->tile != NULL) {
                        tiles[cellPos]->tile->unref();
                        tiles[cellPos]->tile=NULL;
                     }
                     cellPos++;
                  }
               }
            }
         }
      
         /* Step 5: Find ranges of "holes" without tiles (the inventor objects) in VisibleArea,
            calculate their bounding boxes, and request the tiles */
         // adjust (row,col) of left corner of visible area inside the tiles array to account for shifting
         leftBottomVisibleTileCoord.moveBy(xshift,yshift);
         // adjust the bbox of visible area to account for shifting
         visibleAreaBBox.moveBy(xshift*tileSize.x,yshift*tileSize.y);
      
      	/* find the subsection of visible area that we need to fill (maybe all of it,
      	   maybe a subsection, if the visible area is partially out of the map) */
         Rectangle2dInt subsetOfVisibleArea,
            worldMap(0,0,worldMapSize.x-1,worldMapSize.y-1);
         // get the intersection between the world map and the visible area, in case the visible area is partially out of the map
         worldMap.isect(Rectangle2dInt(leftBottomVisibleTileCoord.x,leftBottomVisibleTileCoord.y,
            leftBottomVisibleTileCoord.x+visibleAreaSize.x-1,
            leftBottomVisibleTileCoord.y+visibleAreaSize.y-1), subsetOfVisibleArea);
      
         for(i= subsetOfVisibleArea.leftBottom.y; i<=subsetOfVisibleArea.rightTop.y; i++) {
            // here 'j' is the column number of the leftmost column of the visible area in the world map
            j=subsetOfVisibleArea.leftBottom.x;
            // 'k' is the position of (i,offsetLeftBottomVisibleArea) in the tiles array
            k=(i-leftBottomVisibleTileCoord.y+visibleArea.leftBottom.y)*cachedAreaSize.x-leftBottomVisibleTileCoord.x+visibleArea.leftBottom.x;
            while(j<=subsetOfVisibleArea.rightTop.x) {
               // find beginning of a tile key range with empty tiles
               while(j<=subsetOfVisibleArea.rightTop.x && tiles[k+j]->tile != NULL) {
                  j++;
               }
               if(j<=subsetOfVisibleArea.rightTop.x) {
                  m=j;
                  cellPos=k+j; // take note of the beginning column of range as index inside the cached area
                  range.begin=worldColRowToWorldMapTile(j,i);
                  // find end of tile key range with empty tiles
                  do {
                     j++;
                  } while(j<=subsetOfVisibleArea.rightTop.x && tiles[k+j]->tile == NULL);
                  range.end=worldColRowToWorldMapTile(j-1,i);
                  /* Here we recalculate each tile's bounding box and we put them in an array
                  that we send a a parameter of requestTiles(), so when the tile is loaded
                  we don't have to find the bbox again.
                  We do this because is faster to calculate the bboxes here, rathen than
                  doing it at the receiving end, that will have to scan each Inventor object
                  to calculate its bounding box.
                  This ONLY works iff all the tiles are of exactly the same size.
                  The bbox array will be deleted by  AnotherBathymetryTileIsReady()
                  when all tiles in the array have been loaded, or by the BathymetryTileRangeCancelled()
                   callback when a whole tile range is cancelled. */
               #ifdef BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP
                  mapTiles = new MapTile::pointer[range.end-range.begin+1];
                  mapTiles[0]=tiles[cellPos];
               #else
                  mapTiles = tiles+cellPos;
               #endif
                  mapTiles[0]->bbox.getLeftBottom().setValue(visibleAreaBBox.leftBottom.x+(m-subsetOfVisibleArea.leftBottom.x)*tileSize.x,
                     visibleAreaBBox.leftBottom.y+(i-subsetOfVisibleArea.leftBottom.y)*tileSize.y,BBOX_Z_PLACEHOLDER);
                  mapTiles[0]->bbox.getRightTop().setValue(mapTiles[0]->bbox.getLeftBottom()[0]+tileSize.x,
                     mapTiles[0]->bbox.getLeftBottom()[1]+tileSize.y,BBOX_Z_PLACEHOLDER);
                  xend=range.end-range.begin+1;
                  for(xstart=1; xstart<xend; xstart++) {
                     cellPos++;
                  #ifdef BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP
                     mapTiles[xstart]=tiles[cellPos];
                  #endif
                     mapTiles[xstart]->bbox=mapTiles[xstart-1]->bbox;
                     mapTiles[xstart]->bbox.getLeftBottom()[0]+=tileSize.x;
                     mapTiles[xstart]->bbox.getRightTop()[0]+=tileSize.x;
                  }
                  workerThread->requestTiles(range,mapTiles);
               }
            }
         }
      //clear the old tiles
         removeAllChildren();
      //add the new tiles that we already have
         MapTile::pointer *nodeToRender = tiles+offsetLeftBottomVisibleArea;
         unsigned int nextRow = cachedAreaSize.x-visibleAreaSize.x;
         for(register unsigned int i=0;i<visibleAreaSize.y;i++) {
            for(register unsigned int j=0;j<visibleAreaSize.x;j++) {
               if(*nodeToRender != NULL && (*nodeToRender)->tile != NULL) {
                  if (findChild((*nodeToRender)->tile)==-1) {
                     addChild((*nodeToRender)->tile);
                  }
               }
               nodeToRender++;
            }
            nodeToRender+=nextRow;
         }
      //any other tiles will be displayed as they come in	 
      }
   }

   Point2dUint IvTiledBathymetry::worldPointToWorldColRow(const Point2dFloat& pt) const {
      return Point2dUint((int)floor((pt.x-worldMapBBox.leftBottom.x)/tileSize.x),(int)floor((pt.y-worldMapBBox.leftBottom.y)/tileSize.y));
   }

   TILE_KEY_NUMBER IvTiledBathymetry::worldColRowToWorldMapTile(const Point2dUint& pos) const {
      return pos.y*worldMapSize.x+pos.x;
   }

   TILE_KEY_NUMBER IvTiledBathymetry::worldColRowToWorldMapTile(unsigned int col, unsigned int row) const {
      return row*worldMapSize.x+col;
   }

   Point2dUint IvTiledBathymetry::worldMapTileToWorldColRow(TILE_KEY_NUMBER tile_number) const {
      return Point2dUint(tile_number%worldMapSize.x,tile_number/worldMapSize.x);
   }


   bool IvTiledBathymetry::worldColRowToCachedColRow(const Point2dUint& worldColRow, Point2dUint& relativeColRow) {
      if(worldColRow.x < leftBottomVisibleTileCoord.x || worldColRow.y < leftBottomVisibleTileCoord.y ||
      worldColRow.x > leftBottomVisibleTileCoord.x+visibleAreaSize.x || worldColRow.y > leftBottomVisibleTileCoord.x+visibleAreaSize.y) {
         return false;
      } 
      else {
         relativeColRow.setValue(worldColRow.x-leftBottomVisibleTileCoord.x,worldColRow.y-leftBottomVisibleTileCoord.y);
         return true;
      }
   }


//No longer necessary now that its a group node
/*   void IvTiledBathymetry::GLRender(SoGLRenderAction *action) {
      int numIndices;
      const int *indices;

      MapTile::pointer *nodeToRender = tiles+offsetLeftBottomVisibleArea;
      unsigned int nextRow = cachedAreaSize.x-visibleAreaSize.x;

      SoAction::PathCode pc = action->getPathCode(numIndices, indices);
      // draw the edge of the visible area
      if(drawEdge) {
        GLboolean lightsOn = glIsEnabled(GL_LIGHTING);
        if(lightsOn) glDisable(GL_LIGHTING);
        glColor3f(edgeColor[0],edgeColor[1],edgeColor[2]);
        glBegin(GL_LINE_LOOP);
        glVertex3f(visibleAreaBBox.leftBottom.x,visibleAreaBBox.leftBottom.y,visibleAreaVLimits[1]);
        glVertex3f(visibleAreaBBox.leftBottom.x,visibleAreaBBox.rightTop.y,visibleAreaVLimits[1]);
        glVertex3f(visibleAreaBBox.rightTop.x,visibleAreaBBox.rightTop.y,visibleAreaVLimits[1]);
        glVertex3f(visibleAreaBBox.rightTop.x,visibleAreaBBox.leftBottom.y,visibleAreaVLimits[1]);
        glEnd();
        if(lightsOn) glEnable(GL_LIGHTING);
      }
      // This is almost the same as SoChildList::traverse(), except that it
      //   checks render abort for each non-NULL child 

      nodeToRender = tiles+offsetLeftBottomVisibleArea;

      for(register unsigned int i=0;i<visibleAreaSize.y;i++) {
         for(register unsigned int j=0;j<visibleAreaSize.x;j++) {
            if(*nodeToRender != NULL && (*nodeToRender)->tile != NULL) {
               if (pc == SoAction::OFF_PATH && ! (*nodeToRender)->tile->affectsState()) {
                  nodeToRender++;
                  continue;
               }
               if (action->getCurPathCode() != SoAction::OFF_PATH ||
               (*nodeToRender)->tile->affectsState()) {
                  if (!action->abortNow()) {
                     //action->apply((SoNode*)(*nodeToRender)->tile);
                     ((SoNode*)(*nodeToRender)->tile)->GLRender(action);
                  }
                  else
                     SoCacheElement::invalidate(action->getState());
               }
               if (action->hasTerminated())
                  goto bailOutOffPathGLRender;
            }
            nodeToRender++;
         }
         nodeToRender+=nextRow;
      }
   bailOutOffPathGLRender:;
   }
*/
   void IvTiledBathymetry::useFastBoundingBox(bool useFast, float zmin, float zmax) {
      returnFastBBox=useFast;
      visibleAreaVLimits[0]=zmin;
      visibleAreaVLimits[1]=zmax;
   }

   void IvTiledBathymetry::getBoundingBox(SoGetBoundingBoxAction *action) {
      if(returnFastBBox) {
         SbBox3f bounds(visibleAreaBBox.leftBottom.x,visibleAreaBBox.leftBottom.y,visibleAreaVLimits[0],
            visibleAreaBBox.rightTop.x,visibleAreaBBox.rightTop.y,visibleAreaVLimits[1]);
         action->extendBy(bounds);
         SbVec3f center((visibleAreaBBox.leftBottom.x+visibleAreaBBox.rightTop.x)/2,
            (visibleAreaBBox.leftBottom.y+visibleAreaBBox.rightTop.y)/2,
            (visibleAreaVLimits[0]+visibleAreaVLimits[1])/2);
         action->setCenter(center,false);
         printf("IvTiledBathymetry::getBoundingBox() visibleAreaBBox=[%f,%f]->[%f,%f]\n",visibleAreaBBox.leftBottom.x,visibleAreaBBox.leftBottom.y,visibleAreaBBox.rightTop.x,visibleAreaBBox.rightTop.y);
      } 
      else {
         if(tiles != NULL) {
            MapTile::pointer *nodeToRender = tiles+offsetLeftBottomVisibleArea;
            unsigned int nextRow = cachedAreaSize.x-visibleAreaSize.x;
            for(register unsigned int i=0;i<visibleAreaSize.y;i++){
               for(register unsigned int j=0;j<visibleAreaSize.x;j++) {
                  if((*nodeToRender)->tile != NULL)
                     (*nodeToRender)->tile->getBoundingBox(action);
                  nodeToRender++;
               }
               nodeToRender+=nextRow;
            }
            SbBox3f bounds = action->getBoundingBox();
            visibleAreaVLimits[0]=bounds.getMin().getValue()[2];
            visibleAreaVLimits[1]=bounds.getMax().getValue()[2];         
         }
         else {
            action->extendBy(SbBox3f(0,0,0,0,0,0));
            action->setCenter(SbVec3f(0,0,0),false);
            visibleAreaVLimits[0]=visibleAreaVLimits[1]=0.0f;
         }
      }
   }

   void IvTiledBathymetry::write(SoWriteAction *action) {
      if(alternateRep.getValue() == NULL)
         alternateRep.setValue(new SoGroup);
      MapTile::pointer *nodeToRender = tiles+visibleArea.leftBottom.y*cachedAreaSize.x+visibleArea.leftBottom.x;
      unsigned int nextRow = cachedAreaSize.x-visibleAreaSize.x, i,j;
      for(i=0; i<visibleAreaSize.y; i++){
         for(j=0; j<visibleAreaSize.x; j++) {
            if((*nodeToRender)->tile != NULL) {
               ((SoGroup*)alternateRep.getValue())->addChild((*nodeToRender)->tile);
               (*nodeToRender)->tile->ref();
            }
            nodeToRender++;
         }
         nodeToRender+=nextRow;
      }
      SoGroup::write(action);
      ((SoGroup*)alternateRep.getValue())->removeAllChildren();
   }

   void IvTiledBathymetry::drawMissingTile(void* param, SoAction*) {
      static GLubyte pattern[] = {
         0x80,0x00,0x80,0x00,
         0x40,0x00,0x40,0x00,
         0x20,0x00,0x20,0x00,
         0x10,0x00,0x10,0x00,
         0x08,0x00,0x08,0x00,
         0x04,0x00,0x04,0x00,
         0x02,0x00,0x02,0x00,
         0x01,0x00,0x01,0x00,
         0x00,0x80,0x00,0x80,
         0x00,0x40,0x00,0x40,
         0x00,0x20,0x00,0x20,
         0x00,0x10,0x00,0x10,
         0x00,0x08,0x00,0x08,
         0x00,0x04,0x00,0x04,
         0x00,0x02,0x00,0x02,
         0x00,0x01,0x00,0x01,
         0x80,0x00,0x80,0x00,
         0x40,0x00,0x40,0x00,
         0x20,0x00,0x20,0x00,
         0x10,0x00,0x10,0x00,
         0x08,0x00,0x08,0x00,
         0x04,0x00,0x04,0x00,
         0x02,0x00,0x02,0x00,
         0x01,0x00,0x01,0x00,
         0x00,0x80,0x00,0x80,
         0x00,0x40,0x00,0x40,
         0x00,0x20,0x00,0x20,
         0x00,0x10,0x00,0x10,
         0x00,0x08,0x00,0x08,
         0x00,0x04,0x00,0x04,
         0x00,0x02,0x00,0x02,
         0x00,0x01,0x00,0x01 };
      GLboolean lightsOn = glIsEnabled(GL_LIGHTING);
      GLboolean stippleOn = glIsEnabled(GL_POLYGON_STIPPLE);
      if(lightsOn) glDisable(GL_LIGHTING);
      if(!stippleOn) glEnable(GL_POLYGON_STIPPLE);
      glColor3f(1.0f,1.0f,1.0f);
      glPolygonStipple(pattern);
      glRectf(((Rectangle2dFloat*)param)->leftBottom.x,((Rectangle2dFloat*)param)->leftBottom.y,
         ((Rectangle2dFloat*)param)->rightTop.x,((Rectangle2dFloat*)param)->rightTop.y);
      if(lightsOn) glEnable(GL_LIGHTING);
      if(!stippleOn) glDisable(GL_POLYGON_STIPPLE);
   }

   void IvTiledBathymetry::drawPendingTile(void* param, SoAction*) {
      GLboolean lightsOn = glIsEnabled(GL_LIGHTING);
      if(lightsOn) glDisable(GL_LIGHTING);
      glColor3f(1.0f,1.0f,1.0f);
      glRectf(((Rectangle2dFloat*)param)->leftBottom.x,((Rectangle2dFloat*)param)->leftBottom.y,
         ((Rectangle2dFloat*)param)->rightTop.x,((Rectangle2dFloat*)param)->rightTop.y);
      if(lightsOn) glEnable(GL_LIGHTING);
   }


   void IvTiledBathymetry::setColorscaleDepth(double mind, double maxd, int numC){
      mindepth=mind;
      maxdepth=maxd;
      numColors=numC;
      MapTile::pointer *mapTile = tiles,
         *lastTile = tiles+cachedAreaSize.x*cachedAreaSize.y;
      while(mapTile!=lastTile) {
         if((*mapTile)->tile!=NULL)
            setTileColorscaleDepth((*mapTile)->tile);
         mapTile++;
      }
   }

   void IvTiledBathymetry::setTileColorscaleDepth(SoNode *tile){
     if (tile==NULL) return;
     SoGroup *t=dynamic_cast<SoGroup*>(tile);
     if (t==NULL) return;
     SoIndexedFaceSet *face=NULL;
     SoCoordinate3 *coords=NULL;
     //find them in tile
     int i;
     SoNode *cur;
     SoType ftype=SoType::fromName("IndexedFaceSet");
     SoType ctype=SoType::fromName("Coordinate3");
     int num=t->getNumChildren();
     for (i=0;i<num;i++) {
       cur=t->getChild(i);
       if (cur->isOfType(ftype)) face=(SoIndexedFaceSet *)cur;
        else if (cur->isOfType(ctype)) coords=(SoCoordinate3 *)cur;
        else setTileColorscaleDepth(cur);

     }
   
     if ((face==NULL)||(coords==NULL)) return;
     num=face->coordIndex.getNum();
     int ind;
     double depth;
     double depthdiff=maxdepth-mindepth;
     if (depthdiff<=0) return;
     int *color=new int[num];
     double invdepthdiff=1.0/depthdiff;
     for (i=0;i<num;i++) {
       ind=face->coordIndex[i];
       if (ind < 0) color[i]=ind;
       else {
    depth=-coords->point[ind][2];
    int c=numColors-1;
    c=(int)((1-(depth-mindepth)*invdepthdiff)*(numColors-2)+0.5);
    if (c>(numColors-2)) c=numColors-2;
    if (depth<=mindepth) c=numColors-1;
    if (c<0) c=0;
    color[i]=c;
       }
     }
     face->materialIndex.deleteValues(0);
     face->materialIndex.setValues(0, num, color);
     delete[] color; 
   }
