#ifndef TRACKUIVIEW_H
#define TRACKUIVIEW_H

#include "track.h"
#include "UiMediator.h"

/*! A TrackUiView is the base class for all classes that work receiving notification changes from a track
 and sending the new data to the UI, via a mediator object that talks to the UI. */
   class TrackUiView: public Observer {
   protected:
   /// The track we are a view (in Model/View parlance, TrackUIView is the view, and track is the model).
      ObservableTrack* track;
   /// A reference to the object used to send data and communicate with the UI
      UiMediator& ui;
   /// This flag is only for implementation purposes. It is false when NEW_CONTACT has not been sent yet, and true afterwards, or till the contact is removed.
      bool newContactNotified;
   /*! This is the method that fills the ui member variable
       with all the information gathered from the track member variable.
      If you are going to redefine this method in a subclass, be sure
      to call TrackUiView::copyContactDataToMediator() at the beginning
      of your reimplementation. TrackUiView::copyContactDataToMediator()
      waits for the UI to have read the last data sent before writing the new one,
      and will fill the contactID with the OID (see OID class) of the TrackUIView model 
      (an ObservableTrack or one of its subclasses).
      Filling any other field needed is a responsibility of the subclass */
      virtual void copyContactDataToMediator();
   /*! notifyOfNewContact() is called at the end of TrackView::setModel().
     If there is something to do when a new model is set, do it by
     overriding notifyNewContact() in the subclass,
     and call TrackView::notifyOfNewContact() at the end of the subclass' reimplementation. */
      virtual void notifyOfNewContact();
   public :
      TrackUiView(UiMediator& mediator):track(NULL),ui(mediator),newContactNotified(false) { 
      }
   /// See Observer::update()
      virtual void update(void* aTrackEvent);
   /// See Observer::update()
      virtual void setModel(Observable* aDiscreteTrack);
      virtual ~TrackUiView();
   };

/*! This class extends TrackUiView to handle DiscreteTrackEvent-type events. */
class DiscreteTrackUiView: public TrackUiView {
  public:
   DiscreteTrackUiView(UiMediator& mediator):TrackUiView(mediator) { }
   /// See TrackUiView::update(). This method is redefined to deal with DiscreteTrackEvent events.
   virtual void update(void* aTrackEvent);
};

/*! This class takes data from a DiscreteTrack, assumes it is the ownship's track,
   and put the data in the UI. To tell an OwnshipTrackUiView which is the ownship's track,
   call TrackUiView::setModel() with the ownship track as parameter. */
   class OwnshipTrackUiView: public DiscreteTrackUiView {
   protected:
   /// See TrackUiView::copyContactDataToMediator() .
      virtual void copyContactDataToMediator();
   public:
      OwnshipTrackUiView(UiMediator& mediator):DiscreteTrackUiView(mediator) { 
      }
   };

/*! This class takes data from a DiscreteTrack, assumes it is a vehicle track,
   and put the data in the UI. To tell an VehicleTrackUiView which is the vehicle track to read from,
   call TrackUiView::setModel() with the vehicle track as parameter. */
   class VehicleTrackUiView: public DiscreteTrackUiView {
   protected:
   /// See TrackUiView::copyContactDataToMediator() .
      virtual void copyContactDataToMediator();
   /// See TrackUiView::notifyOfNewContact() .
      virtual void notifyOfNewContact();
   /// Time at which the contact was first seen (used to calculate contact's age)
      timeval startTime;
   public:
      VehicleTrackUiView(UiMediator& mediator):DiscreteTrackUiView(mediator) { 
      }
    /// If this is the current vehicle, propagates the change.See Observer::update()
      virtual void update(void* aTrackEvent);
   };

/*! This class takes data from a DiscreteTrack, assumes it is a weapon track,
   and put the data in the UI. To tell an WeaponTrackUiView which is the vehicle track to read from,
   call TrackUiView::setModel() with the weapon track as parameter. */
   class WeaponTrackUiView: public DiscreteTrackUiView {
   protected:
   /// See TrackUiView::copyContactDataToMediator() .
      virtual void copyContactDataToMediator();
      virtual void notifyOfNewContact();
   /// Time at which the contact was first seen (used to calculate contact's age)
      timeval startTime;
   public:
      WeaponTrackUiView(UiMediator& mediator):DiscreteTrackUiView(mediator) { 
      }
      /// If this is the current weapon, propagates the change.See Observer::update()
      virtual void update(void* aTrackEvent);
   };

/*! This class takes data from a DiscreteTrack, assumes it is a sensor track (an Area Of Uncertainty),
   and put the data in the UI. To tell an SensorTrackUiView which is the sensor track to read from,
   call TrackUiView::setModel() with the sensor track as parameter. */
   class SensorTrackUiView: public DiscreteTrackUiView {
   protected:
   /// See TrackUiView::copyContactDataToMediator() .
      virtual void copyContactDataToMediator();
   /// See TrackUiView::notifyOfNewContact() .
      virtual void notifyOfNewContact();
   /// Time at which the contact was first seen (used to calculate contact's age)
      timeval startTime;
   public:
      SensorTrackUiView(UiMediator& mediator):DiscreteTrackUiView(mediator) { 
      }
    /// If this is the current sensor, propagates the change.See Observer::update()
      virtual void update(void* aTrackEvent);
   };

/*! This class takes data from a DiscreteTrack, assumes it is a hypothetical track (a combo of other areas),
   and put the data in the UI. To tell a HypotheticalTrackUiView which is the sensor track to read from,
   call TrackUiView::setModel() with the sensor track as parameter. */
   class HypotheticalTrackUiView: public DiscreteTrackUiView {
   protected:
   /// See TrackUiView::copyContactDataToMediator() .
      virtual void copyContactDataToMediator();
   /// See TrackUiView::notifyOfNewContact() .
      virtual void notifyOfNewContact();
   /// Time at which the contact was first seen (used to calculate contact's age)
      timeval startTime;
   public:
      HypotheticalTrackUiView(UiMediator& mediator):DiscreteTrackUiView(mediator) { 
      }
    /// If this is the current hypothetical, propagates the change.See Observer::update()
      virtual void update(void* aTrackEvent);
   };


#endif
