#ifndef UI_MED_H
#define UI_MED_H

// 2002, Fernando Das-Neves, Virgina Tech UVAG Group //


#include <sys/time.h>
#include <string.h>
#include "SharedMemObj.h"
#include "coord.h"
#include "bezel/ui_constants.h"
#include "oid.h"

/// How many bytes it takes to store the image in the minimap
#define MAP_IMAGE_SIZE (3*MAP_IMAGE_X*MAP_IMAGE_Y)

/*! type and flags to indicate commands the UI can request to the main app */
namespace MainAppCommand {
   const int MAX_CONTACTNAME_LENGTH = 64;

   typedef ulong CommandFlag;

   const CommandFlag NONE                 = 0;
   const CommandFlag STEP_FORWARD_IN_TIME = 1;
   const CommandFlag STEP_BACK_IN_TIME    = 2;
   const CommandFlag STOP_MOVEMENT        = 4;
   const CommandFlag RESUME_MOVEMENT      = 8;
   const CommandFlag SET_COLOR_MAP       = 16;
   const CommandFlag SET_GREY_MAP        = 32;
   const CommandFlag SET_CAMERA          = 64;
   const CommandFlag STOP_FEEDER		= 128;
   const CommandFlag RESUME_FEEDER		= 256;
   const CommandFlag SELECT_TARGET		= 512;
   const CommandFlag SELECT_WEAPON		= 1024;     // 2^10
   const CommandFlag SNAPSHOT         		= 2048;     // 2^11
   const CommandFlag SET_BLUE_MAP        	= 4096;     // 2^12
   const CommandFlag SET_SAND_MAP        	= 4097;     // 2^12 +1
   const CommandFlag SHOW_INTERSECTING        	= 8192;
   const CommandFlag HIDE_INTERSECTING        	= 16384;
   const CommandFlag NAV_GRID                	= 32768;
   const CommandFlag RANGE_RINGS        	= 65536;
   const CommandFlag LON_LAT		= 131072;
   const CommandFlag HYPOTHESIS		= 262144;
   const CommandFlag HYPO_CONTACTS	= 524288;
   const CommandFlag SOLUTIONS		= 1048576;
   //const CommandFlag TARGET_SPEED	= 2097152;
   const CommandFlag REDUCEREGION	= 2097152;
   const CommandFlag VERYSHALLOW        = 4194304;
   const CommandFlag SURFACE_ON		= 4194305;
   const CommandFlag SURFACE_OFF	= 4194306;
   const CommandFlag SHALLOW		= 8388608;
   const CommandFlag MIDRANGE		= 16777216;
   const CommandFlag DEEP  		= 33554432;
   const CommandFlag SCROLL_VAL  	= 67108864;
   const CommandFlag GROWREGION  	= 134217728;
   const CommandFlag HIDE_GROWING	= 268435456;
   const CommandFlag FOLLOW_ON          = 536870912;
   const CommandFlag FOLLOW_OFF		= 1073741824;


     /*! LonLat contains the longitude and Latitude data to determine
      where the bottom map should be at the time.
  */
   struct LonLat{
	double lon_se, lon_nw, lat_se, lat_nw;
   };

   /*!Hypothesis is the user input best guess for the target
   	\param maxSpeed
	\param maxDepth
	\param minDepth
   */
   struct Hypothesis{
	double maxSpeed;
	double maxDepth;
	double minDepth;
   };

   ///hypothetical contact(s)
   struct HypoContact{
   	int numContact;
	char contactList[650][64];
        char regionName[80];
   };

   struct GrowRegion{
     int numSolution;
     char solutionList[650][64];
   };
   
   struct ReduceRegion{
     int numSolution;
     char solutionList[650][64];
   };
   
   struct Solution{
   	int numSolution;
	char solutionList[650][64];
   };
   ///TargetSpeed is represented by an integer.
   struct TargetSpeed{
	int targetSpeed;
   };

   ///ScrollVal is literally the value of the thing that zooms mini-map in and out.
   struct ScrollVal{float scroll_val;};

  /// A union encapsulating all possible types of arguments to send from the bezel to the main app
   typedef struct {
      union {
         unsigned int Uint;
         float Float;
         char Char;
         /*! whereToPositionCamera tells what object to center the camera on,
            and how to position the camera. The
            camera can be positioned as TOP_DOWN_CAMERA,
            TOP_DOWN_PERSPECTIVE_CAMERA, or CAMERA_CENTERED_ON.
            If whereToPositionCamera == CAMERA_CENTERED_ON, then
	    contactID must contain the OID of the object where to center the camera.
         */
         unsigned int whereToPositionCamera;
         char String[MAX_CONTACTNAME_LENGTH];
        };
        OID::Type contactID;
      } ArgType;

   const unsigned int TOP_DOWN_CAMERA = 0xFFFF;
   const unsigned int TOP_DOWN_PERSPECTIVE_CAMERA = 0xFFFE;
   const unsigned int CAMERA_CENTERED_ON = 0xFFFD;
   const unsigned int TOP_DOWN_CAMERA_CENTERED_ON = 0xFFFC;
   const unsigned int CAMERA_FIRST_PERSON_FROM = 0xFFFB;
   const unsigned int WINGMAN = 0xFFFA;
   const unsigned int CROSS = 0xFFEF;
}

/*! type and flags to indicate notifications from the main app to the bezel,
    and actions the main app can request to the bezel. */
namespace BezelCommand {
   /*! Constants to let the main program indicate the bezel what info
   has updated, and to let the bezel request the main program to quit.
    All this is done by setting bits in the actionFlag variable */
   typedef enum {
   	NONE =0,
   	QUIT =1,
   	READY_TO_EXIT =2,
   	NEW_CONTACT =4,
	REMOVE_CONTACT =8,
	NEW_CONTACT_POSITION =16,
   	RESENDING_CONTACT_POSITION = 32, // actual position is not new, just resent
   	MAP_IMAGE_UPDATED =128,
   	SELECT_CONTACT =256,
   CONTACT_DATA_AVAILABLE = NEW_CONTACT|REMOVE_CONTACT|NEW_CONTACT_POSITION|RESENDING_CONTACT_POSITION
   } ActionFlag;
	/// ContactData is a structure that holds all the information that can be passed between the main app and the UI
   typedef
      struct {
         /*! If this is a new contact, then 'name' contains the contact name,
           used to identify which contact the new data belongs to. In this
           case contactID also contains an object ID for the contact with
           that name in the main app. This object ID can be used by the bezel
           to refer to a contact when talking back to the main app.
           For all the other actions, like UPDATE_CONTACT or REMOVE_CONTACT,
           only contactID is valid, and name is ignored for efficiency.
         */
         char name[64];
         char solution[64];
      	/*! An object ID identifying the contact in the context of the main app.
      	    Every time the main app requests an action to the bezel,
      	    contactID will contain the object ID of the contact referred to by the action. */
         OID::Type contactID;
         /* data type (ownship, target, weapon) */
         ContactType contactType;
         /* contact information, like depth, speed, pos, etc */
         union {
            PreciseContact precise;
	    WeaponContact weapon;
	    ContactFromSensor fromSensor;
            HypotheticalContact hypothetical;
         };
      } ActionData;
}

/*! UiMediator is a class used to communicate between the main app and the bezel using shared memory.
    Although there is no reason for more than one instance of this class, there is no control of this rule in the code.
    UiMediator can only hold up to 1 command from the app to the bezel, and up to 1 command from
    the bezel to the main app. This means that:<BR>
    <LI>
    	<UL> The bezel must call isContactDataAvailable() before reading, and call
    	signalContactWasRead() when it read the new contact information.</UL>
    	<UL> The bezel can issue commands to the main app by calling giveOrder(), and the app must call signalOrderDone()
    	to acknowledge the order.</UL>
    </LI>
*/
   struct UiMediator: public ObjectInSharedMemory {
   public:
      enum { MAX_CONTACTNAME_LENGTH = 64 };

      // Data & commands from main app to bezel
      unsigned int actionFlag;
      BezelCommand::ActionData actionData;
      unsigned char image [MAP_IMAGE_SIZE]; // raw data of minimap image

      // Data & commands from bezel to main app
      MainAppCommand::CommandFlag command;
     //a bit of a hack since both actionFlag and command were sometimes overwritten before they were checked for quit.  plus quit operations are ORs that don't work well with current case statements
      MainAppCommand::ArgType commandArg;

      // datas collected from the bezel to be send over to main app
      MainAppCommand::LonLat lonLat;
      MainAppCommand::Hypothesis hypothesis;
      MainAppCommand::HypoContact hypoContact;
      MainAppCommand::Solution solution;
      MainAppCommand::TargetSpeed targetSpeed;
      MainAppCommand::ScrollVal scroll_val;
      MainAppCommand::GrowRegion growRegion;
      MainAppCommand::GrowRegion reduceRegion;

      OID::Type selectedTarget;
      OID::Type selectedWeapon;
   
      UiMediator();
   	// Generic command-setting methods
      bool isFlagSet(BezelCommand::ActionFlag flag) const {
         return (bool)(actionFlag & flag); }
      void clearFlag(BezelCommand::ActionFlag flag) {
         actionFlag &= ~flag; }
      void setFlag(BezelCommand::ActionFlag flag) {
         actionFlag |= flag; }
   	/// Tell the bezel the program is finishing
      void signalQuit() {
	 actionFlag |= BezelCommand::QUIT; }
   	/// Used by the bezel to ask if QUIT was requested
      bool readyToQuit() const {
         return (bool)(actionFlag & BezelCommand::QUIT); }
      /// Used  by the bezel to ask if a new minimap is available
      bool isNewMapAvailable() const {
         return isFlagSet(BezelCommand::MAP_IMAGE_UPDATED); }
   	/// Used by the main app to update the minimap data
      void updateMap(const unsigned char* newMap);
   	/// Used by the bezel to indicate it finished reading the minimap data
      void signalMapWasRead() {
         clearFlag(BezelCommand::MAP_IMAGE_UPDATED); }
   	/// Used by the bezel to check if new information about a contact is available
      bool isContactDataAvailable() const {
         return (bool)(actionFlag & BezelCommand::CONTACT_DATA_AVAILABLE); }
   	/// Used by the main app to indicate that it has posted new information about a contact
      void signalContactDataAvailable(BezelCommand::ActionFlag flag);
   	/// Used by the bezel to tell the main app it read the newest contact information the main app posted
      void signalContactDataWasRead();
   	/// Used by the main app to check if the bezel already read the last contact information the main app sent.
      bool contactDataWasRead() const {
         return (!(bool)(actionFlag & BezelCommand::CONTACT_DATA_AVAILABLE))||readyToQuit(); }
      /// Returns true if there is a command pending for the main app
      bool isCommandAvailable() const { return command != MainAppCommand::NONE; }
   	/// Used to set the parameters for a message from the bezel to the main app. Use it before calling UiMediator::giveOrder()
      void setArgumentForCommand(unsigned int arg) {
         commandArg.Uint = arg; }
      void setArgumentForCommand(float arg) {
         commandArg.Float = arg; }
      void setArgumentForCommand(char arg) {
         commandArg.Char = arg; }
   	/// request the main app to position the camera
      void setArgumentForCameraCommand(unsigned int view, OID::Type whereToCenterCamera);
   	/// Used by the bezel to signal the main app that the last requested order was completed
      void signalCommandDone() {
         command = BezelCommand::NONE; }
      /// send a command from the UI to the main app. Use setArgumentForCommand() to set the command's arguments
      void askMainApp(MainAppCommand::CommandFlag aCommand) { command = aCommand; }
      /// send a command from the UI to the main app, with an OID as argument
      void askMainApp(MainAppCommand::CommandFlag, OID::Type arg);
   	/// send a request for an action from the main app to the UI, with an (optional) OID as argument
      void askBezel(BezelCommand::ActionFlag, OID::Type arg = BAD_OID);
   };

#endif

