// This program takes an iv file and changes grayscale to color //
// Fernando Das-Neves, Virginia Tech, 2002                      //

#include <Inventor/SoInput.h>
#include <Inventor/SoOutput.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/SoDB.h>
#include <stdio.h>

/*! This class is an Inventor action that will change
    any SoBaseColor or Material node color from grayscale
    to a gradient of another color.
    This action will only work with graylevels, it will
    skip any color node that is not in grays */
class IvColorizeGrayscalesAction: public SoCallbackAction {
private:
    virtual void apply(SoNode*) { };
    virtual void apply(SoPath*) { };
    virtual void apply(const SoPathList &, SbBool) { };
protected:
    float hsvColor[3];
    static void colorizeSoMFColorField(SoMFColor& mfcolor, const float* newcolor);
    static SoCallbackAction::Response changeMaterialFromGrayscaleToColor(void* newColor,SoCallbackAction*, const SoNode* mat);
    static SoCallbackAction::Response changeBaseFromGrayscaleToColor(void* newColor,SoCallbackAction*, const SoNode* baseColor);
public:
    IvColorizeGrayscalesAction();
    virtual void applyHSVColor(SoNode *node, float h, float s, float v);
    virtual void applyRGBColor(SoNode *node, const SbColor& rgbNewColor);
};

void IvColorizeGrayscalesAction::colorizeSoMFColorField(SoMFColor& mfgrayscale, const float* hsvNewColor) {
  int i;
  float *color;
  SbColor *colors = mfgrayscale.startEditing();
  color = (float*)colors->getValue();
  /* Here we are using a cheap-and-fast trick to detect 
     when the current node is a node that we already processed, and 
     that happens to be referenced more than once in the scene graph.
     Since this algorithm generates only 3 possible types of HSV triplets:
     (hue,1,0), (hue,1,1), and (hue,1,0), and hue is always in the [0,1) range,
     any node with whose color has r=g=b is a node that we didn't process yet.
     Otherwise, we already transformed the gray levels in this node, 
     and so we skip the node. */
  if (color[0]== color[1] && color[1]== color[2]) {
    if (*color < 0.5) {
         colors->setHSVValue(hsvNewColor[0],1.0f,2*color[0]);
    } else {
         colors->setHSVValue(hsvNewColor[0],2-2*color[0],1.0f);
    }
    for(i=1; i<mfgrayscale.getNum(); i++) {
      colors++;
      color = (float*)colors->getValue();
      if (*color < 0.5) {
           colors->setHSVValue(hsvNewColor[0],1.0f,2*color[0]);
      } else {
         colors->setHSVValue(hsvNewColor[0],2-2*color[0],1.0f);
      }
    }
  }
  mfgrayscale.finishEditing();
}

SoCallbackAction::Response IvColorizeGrayscalesAction::changeMaterialFromGrayscaleToColor(void* hsvColor, SoCallbackAction*, const SoNode* mat) {
  colorizeSoMFColorField(((SoMaterial*)mat)->ambientColor,(float*)hsvColor);
  colorizeSoMFColorField(((SoMaterial*)mat)->diffuseColor,(float*)hsvColor);
  colorizeSoMFColorField(((SoMaterial*)mat)->specularColor,(float*)hsvColor);
  colorizeSoMFColorField(((SoMaterial*)mat)->emissiveColor,(float*)hsvColor);
  return SoCallbackAction::CONTINUE;
}

SoCallbackAction::Response IvColorizeGrayscalesAction::changeBaseFromGrayscaleToColor(void* hsvColor,SoCallbackAction*, const SoNode* baseColor) {
  colorizeSoMFColorField(((SoBaseColor*)baseColor)->rgb,(float*)hsvColor);
  return SoCallbackAction::CONTINUE;
}

IvColorizeGrayscalesAction::IvColorizeGrayscalesAction():SoCallbackAction() {
  addPostCallback(SoBaseColor::getClassTypeId(),changeBaseFromGrayscaleToColor,this->hsvColor);
  addPostCallback(SoMaterial::getClassTypeId(),changeMaterialFromGrayscaleToColor,this->hsvColor);
}

void IvColorizeGrayscalesAction::applyRGBColor(SoNode *node, const SbColor& rgbNewColor) {
  rgbNewColor.getHSVValue(hsvColor);
  if(hsvColor[0]>=1.0f)
	hsvColor[0] -= 1.0f; // guarantees hue is in the [0,1) range
  SoCallbackAction::apply(node);
}

void IvColorizeGrayscalesAction::applyHSVColor(SoNode *node, float h, float s, float v) {
  hsvColor[0] = h;
  if(hsvColor[0]>=1.0f)
	hsvColor[0] -= 1.0f; // guarantees hue is in the [0,1) range
  hsvColor[1] = s;
  hsvColor[2] = v;
  SoCallbackAction::apply(node);
}


void printUsage() {
  printf("USAGE: ivtocolor inventor_file output_name\n");
}

int main(int argc, char** argv) {
  SoDB::init();
  SbColor newColor;
  SoSeparator *root;  
  SoInput loader;
  SoWriteAction writer;
  IvColorizeGrayscalesAction traversingAction;

  float hsv[3];
  newColor.setValue(1,0.5,0);
  newColor.getHSVValue(hsv);
  printf("HSV = %f,%f,%f\n",hsv[0],hsv[1],hsv[2]);


  if(argc != 3) {
    printUsage();
    return 1;
  }
  
  if(loader.openFile(argv[1]) == FALSE) {
    printf("Couldn't open '%s'\n",argv[1]);
    return 2;
  }

  root = SoDB::readAll(&loader);
  loader.closeFile();
  if(root == NULL) {
    printf("Couldn't read a valid inventor object from file '%s'\n",argv[1]);
    return 3;
  }

  root->ref();
  traversingAction.applyHSVColor(root,hsv[0],hsv[1],hsv[2]);

  if(writer.getOutput()->openFile(argv[2]) == FALSE) {
    printf("Couldn't write result to '%s'\n",argv[1]);
    return 4;
  }
  writer.apply(root);
  writer.getOutput()->closeFile();
}
