/// 2002, Virginia Tech UVAG Group ///

#include "track.h"
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string>
#include <algorithm>
#include <functional>
#include "timeval_ops.h"



/***************** SingleObservable ****************/

   void SingleObservable::addView(Observer *anObserver, unsigned int bitMask, unsigned int eventsOfInterest) {
      if(observerInfo==NULL) 
         observerInfo = new ObserverInfo;
      observerInfo->interests = eventsOfInterest;
      observerInfo->userDefinedMask = bitMask;
      observerInfo->observer = anObserver;
   }

   void SingleObservable::notify(void* info) {
      if(observerInfo != NULL && observerInfo->interests & ((TrackEvent*)info)->eventType)
         observerInfo->observer->update(info);
   }

   void SingleObservable::removeView(Observer *anObserver) {
      if(observerInfo != NULL && anObserver == observerInfo->observer) {
         delete observerInfo;
         observerInfo = NULL;
      }
   }

   Observer* SingleObservable::findObserverWithMask(unsigned int bitMask, unsigned int* copyObjectBitMask) const {
      if(observerInfo != NULL && observerInfo->userDefinedMask & bitMask)
         return observerInfo->observer;
      else {
         if(copyObjectBitMask != NULL)
            *copyObjectBitMask = observerInfo->userDefinedMask;
         return NULL;	
      }
   }

   SingleObservable::~SingleObservable() {
      if(observerInfo != NULL) delete observerInfo;
   }


/***************** MultipleObservable **************/


   void MultipleObservable::addView(Observer *anObserver, unsigned int bitMask, unsigned int eventsOfInterest) {
      ObserverInfo * observerInfo = new ObserverInfo;
      observerInfo->interests = eventsOfInterest;
      observerInfo->userDefinedMask = bitMask;
      observerInfo->observer = anObserver;
      observers.push_back(observerInfo);
   }

   bool MultipleObservable::notifyEach(void* anObserver, void* info) {
      if( ((ObserverInfo*)anObserver)->interests & ((TrackEvent*)info)->eventType)
         ((ObserverInfo*)anObserver)->observer->update(info);
      return true;
   }

   void MultipleObservable::notify(void* info) {
      for_each(observers.begin(),observers.end(),bind2nd(ptr_fun(notifyEach),info) );
   }

   void MultipleObservable::removeView(Observer *anObserver) {
      ObserverList::iterator it = observers.begin();
      while(it != observers.end() && (((ObserverInfo*)*it)->observer != anObserver)) ++it;
      if(it != observers.end())
         observers.erase(it);
   }


   Observer* MultipleObservable::findObserverWithMask(unsigned int bitMask, unsigned int* copyObjectBitMask) const {
      ObserverList::const_iterator it = observers.begin();
      while(it != observers.end() && (((ObserverInfo*)*it)->userDefinedMask & bitMask) == 0) ++it;
      if(it == observers.end())
         return NULL;
      else {
         if(copyObjectBitMask != NULL)
            *copyObjectBitMask = ((ObserverInfo*)(*it))->userDefinedMask;
         return ((ObserverInfo*)(*it))->observer;	
      }	
   }


   bool MultipleObservable::deleteObserverInfo(void *observerInfo) {
      delete (ObserverInfo*)observerInfo;
      return true;
   }


   MultipleObservable::~MultipleObservable() {
      for_each(observers.begin(),observers.end(),ptr_fun(deleteObserverInfo));
   }


/******************* Timeline *******************/

   Timeline::Timeline():foreOrBackInTime(FORWARD_IN_TIME),lastTime(TIME_MIN),earliest(TIME_MAX),stopped(false) {
   }

   void Timeline::setTimeDirection(const TimeDirection td) {
      foreOrBackInTime = td;
   }

   void Timeline::addTrack(Track* aTrack) {
      tracks.push_back(aTrack);
      /* "fast forward" the events in the new track, till we get 
          to the closest point to the current time in the timeline. */
      timeval trackTime = aTrack->getTimeOfNextEvent(FORWARD_IN_TIME);
      if(trackTime != NO_EVENT && lastTime != NO_EVENT) {
         while(trackTime != NO_EVENT && trackTime <= lastTime) {
            aTrack->advanceToNextEvent(FORWARD_IN_TIME);
            trackTime = aTrack->getTimeOfNextEvent(FORWARD_IN_TIME);
         }
         if(aTrack->startTime() < earliest) {
            earliest = aTrack->startTime();
         }
      }
   }

   Track* Timeline::removeTrack(Track* aTrack) {
   
      TrackList::iterator it = find_if(tracks.begin(), tracks.end(), bind1st(equal_to<void*>(),(void*)aTrack) );
      if(it != tracks.end()) {
         tracks.erase(it);
         if(aTrack->startTime() == earliest) {
            earliest = TIME_MAX;
            for(it = tracks.begin(); it != tracks.end(); ++it)
               if(((Track*)*it)->startTime() < earliest) earliest = ((Track*)*it)->startTime();
         }
      } 
      else aTrack = NULL;
      return aTrack;
   }

   Track* Timeline::removeTrack(unsigned trackNo) {
      Track *theTrack;
      TrackList::iterator track_it;
      unsigned idx = 0;
      track_it = tracks.begin();
      while(idx < trackNo && track_it != tracks.end()) { ++track_it; ++idx; }
      if(track_it != tracks.end()) {
         theTrack = (ObservableTrack*)*track_it;
         tracks.erase(track_it);
         if(theTrack->startTime() == earliest) {
            earliest = TIME_MAX;
            for(TrackList::iterator track_it = tracks.begin(); track_it != tracks.end(); ++track_it)
               if( ((Track*)*track_it)->startTime() < earliest)
                  earliest = ((Track*)*track_it)->startTime();
         }
      } 
      else theTrack = NULL;
      return theTrack;
   }

/** Find what is next to play in every track, and play the more recent event,
    then ask every track that what's next.
    Returns true if more events are pending, or false otherwise */
   bool Timeline::playNextEvent() {
      bool moreEvents = false;
      TrackList::iterator track_it;
      timeval trackTime,
         closestTime = TIME_MAX;
      Track *track;
   
      if(stopped)
         return false;
   
      if(foreOrBackInTime == FORWARD_IN_TIME)
         closestTime = TIME_MAX;
      else
         closestTime = TIME_MIN;
   /* check if any track has changed, and find closest events. If all
     tracks answer NO_EVENT, then there is nothing to do here. */
      for(track_it = tracks.begin(); track_it != tracks.end(); ++track_it) {
         track = (Track*)(*track_it);
         trackTime = track->getTimeOfNextEvent(foreOrBackInTime);
         if(trackTime != NO_EVENT) {
            if(foreOrBackInTime == FORWARD_IN_TIME) {
               if(trackTime <= closestTime) {
                  closestTime = trackTime;
                  moreEvents = true;
               }
            } 
            else {
               if(trackTime >= closestTime) {
                  closestTime = trackTime;
                  moreEvents = true;
               }
            }
         }
      }
   
      if(moreEvents) {
      /* "play" all tracks up to current time. There may be more than one
       because there may be simultaneous events in different tracks */
         if(foreOrBackInTime == FORWARD_IN_TIME) {
            for(track_it = tracks.begin(); track_it != tracks.end(); ++track_it) {
               track = (Track*)(*track_it);
               trackTime = track->getTimeOfNextEvent(FORWARD_IN_TIME);
               if(trackTime != NO_EVENT && trackTime <= closestTime) {
               // tell track we want it to move forward
                  track->advanceToNextEvent(FORWARD_IN_TIME);
               }
            }
         } 
         else { // it's foreOrBackInTime == BACKWARDS_IN_TIME
            for(track_it = tracks.begin(); track_it != tracks.end(); ++track_it) {
               track = (Track*)(*track_it);
               trackTime = track->getTimeOfNextEvent(BACKWARD_IN_TIME);
               if(trackTime != NO_EVENT && trackTime >= closestTime) {
               // tell track we want it to move backwards
                  track->advanceToNextEvent(BACKWARD_IN_TIME);
               }
            }
         }
      }
      lastTime = closestTime;
      return moreEvents;
   }


   void Timeline::appendStepToTrack(ObservableTrack* aTrack, CoordAndTime* pos) {
      timeval nextEvent = aTrack->getTimeOfNextEvent(foreOrBackInTime);
      aTrack->appendStep(pos);
   // do we have a new earliest time for any track?
      if(earliest > pos->time)
         earliest = pos->time;
   /* if we're going forward, and we have already gone through all the steps
     in this track, then tell the track to move forward to show the new step */
      if(foreOrBackInTime == FORWARD_IN_TIME && !stopped) {
         if(nextEvent == NO_EVENT) {
         /* we could call this->playNextEvent() here, but it would
         go through all the tracks unnecessarily. Instead we only
         call advanceToNextEvent() for the track that changed. */
	   //aTrack->advanceToNextEvent(FORWARD_IN_TIME);
	   //if(pos->time > lastTime) lastTime = pos->time;
	   playNextEvent();
         }
      }
      //this may not be the best place for it, but
      //go through all the tracks and send them the new current time
      //if (GROW_REGIONS) {
         TrackTraceDecorator *track;
         TrackList::iterator track_it;
         for(track_it = tracks.begin(); track_it != tracks.end(); ++track_it) {
            track = (TrackTraceDecorator*)(*track_it);
            timeval tval(pos->time);
            CoordAndTime *next=track->grow(this,tval);
            if (next!=NULL)
               appendStepToTrack(track,next);
         }
	 //}
   
   }

   void Timeline::stop() {
      stopped = true;
   }

   void Timeline::rewind() {
      foreOrBackInTime = BACKWARD_IN_TIME;
      stopped = false;
      while(playNextEvent());

   }
   void Timeline::resume() {
      foreOrBackInTime = FORWARD_IN_TIME;
      stopped = false;
      while(playNextEvent());
   }

/**************** TrackTraceDecorator ****************/

/* Returns the time of current event after advancing,
   or NO_EVENT if there are no more events in that direction.
   IMPORTANT: When FORWARD_IN_TIME, the views of this decorator
   receive the update() call AFTER the "decorated" (contained) track has
   added a step.
   When BACKWARDS_IN_TIME, the views of this decorator receive the update()
   call BEFORE the "decorated" track removes a step, so the view can
   do whatever he has to do with the current position of the track. */
   const timeval& TrackTraceDecorator::advanceToNextEvent(const TimeDirection beforeOrAfter) {
      TrackEvent evt;
      if(beforeOrAfter == FORWARD_IN_TIME) {
         const timeval& time = track->advanceToNextEvent(beforeOrAfter);
         if(time != NO_EVENT) {
            evt.eventType = TrackEvent::TRACK_STEP_APPENDED;
            notify(&evt);
         }
         return time;
      } 
      else {
         CoordAndTime* coord = track->getCurrentCoordAndTime();
         if(coord != NULL && coord->time != NO_EVENT) {
            evt.eventType = TrackEvent::TRACK_STEP_REMOVED;
            notify(&evt);
            return track->advanceToNextEvent(beforeOrAfter);
         }
         else 
            return NO_EVENT;
      }
   }

/*************** ObservableTrack ******************/

void ObservableTrack::setThreatClass(const ThreatClass newThreatLevel) { 
  if(threatLevel != newThreatLevel) {
    TrackEvent evt;
    threatLevel = newThreatLevel;
    evt.eventType = TrackEvent::CONTACT_THREATCLASS_CHANGED;
    notify(&evt);
  }
}

   void ObservableTrack::notify(void* info) {
      ((TrackEvent*)info)->track = this;
      MultipleObservable::notify(info);
   }

   ObservableTrack::~ObservableTrack() {
      TrackEvent evt;
      evt.eventType = TrackEvent::TRACK_ABOUT_TO_DIE;
      notify(&evt);
   }

/*************** DiscreteTrack ******************/

   DiscreteTrack::DiscreteTrack():type(CONTACT_UNKNOWN),pastEndOfTrack(false) {
      currentStep = last = steps.end();
      endOfTrack = steps.end();
   }


   const timeval& DiscreteTrack::getTimeOfNextEvent(const TimeDirection beforeOrAfter) {
      CoordAndTimeList::iterator next = currentStep;
      if(beforeOrAfter == FORWARD_IN_TIME) {
         if(next == last) { // if current is the last element
            if(!pastEndOfTrack && endOfTrack != steps.end() && steps.size() > 0)      
               return FAKE_PAST_END_TIME;
            else
               return NO_EVENT;
         
         } 
         else if(currentStep == steps.end()) { // this only happens when the track is empty or it still has not advanced to the first step 
            next = steps.begin();
         } 
         else if (!pastEndOfTrack) {
            ++next;
         }
      } 
      else { // ... beforeOrAfter == BACKWARDS_IN_TIME
         if(currentStep == steps.end()) {
            return NO_EVENT;      
         }
         else if(next == steps.begin()) {
            if(!pastEndOfTrack && steps.size() > 0)
               return FAKE_PAST_BEGINNING_TIME;
            else
               return NO_EVENT;
         } 
         else if(!pastEndOfTrack) {
            --next;
         }
      }
      return ((CoordAndTime*)*next)->time;
   }


   const timeval& DiscreteTrack::advanceToNextEvent(const TimeDirection beforeOrAfter) {
      TrackEvent evt;
      if(beforeOrAfter == FORWARD_IN_TIME) {
         if(currentStep == last) {
            if(!pastEndOfTrack && endOfTrack != steps.end()) { // send PAST_END_OF_TRACK only if track has an end already
               // if we're here, is because we are being asked to move past beyond the last valid position
               pastEndOfTrack = true;
               evt.eventType = DiscreteTrackEvent::PAST_END_OF_TRACK;
               notify(&evt);
            }
            return NO_EVENT;
         } 
         else {
            if(pastEndOfTrack) { // this happens when we're "out of track", past behind the first valid step
               pastEndOfTrack = false; 
            } 
            else if(currentStep == steps.end()) { // this only happens once, when the track was empty, and this is the first move
               currentStep = steps.begin();
            } 
            else {
               ++currentStep;               
            }
            evt.eventType = TrackEvent::POS_CHANGED;
            notify(&evt);
            return ((CoordAndTime*)*currentStep)->time;                       
         }
      } 
      else { //...beforeOrAfter is BACKWARDS_IN_TIME
         if(currentStep == steps.begin() || currentStep == steps.end()) { // the beginning of the track is always defined
            if(!pastEndOfTrack) {
               pastEndOfTrack = true;
               evt.eventType = DiscreteTrackEvent::PAST_BEGINNING_OF_TRACK;
               notify(&evt);
            }
            return NO_EVENT;
         } 
         else if (pastEndOfTrack) { // this is true, means the track  has at least one step
            pastEndOfTrack = false;
         } 
         else {
            --currentStep;
         }
      }
      evt.eventType = TrackEvent::POS_CHANGED;
      notify(&evt);
      return ((CoordAndTime*)*currentStep)->time;
   }

   const timeval& DiscreteTrack::startTime() const {
      if(steps.empty())
         return NO_EVENT;
      else
         return ((CoordAndTime*)*(steps.begin()))->time;
   }

   const timeval& DiscreteTrack::finalTime() const {
      if(steps.empty())
         return NO_EVENT;
      else {
         return ((CoordAndTime*)steps.back())->time;
      }
   
   }


   CoordAndTime* DiscreteTrack::getCurrentCoordAndTime() const {
      if((CoordAndTimeList::const_iterator)currentStep != steps.end()) { 
         return (CoordAndTime*)*currentStep;
      } 
      else {
         return NULL; // this only happens when the track is empty or it still has not advanced to the first step 
      }
   }

   void DiscreteTrack::appendStep(CoordAndTime* step) {
     // if the track is closed, we won't allow to add another step - in fact, that's a bug!
      if(endOfTrack == steps.end()) {
         if(steps.size() == 0) {
            FAKE_PAST_BEGINNING_TIME = step->time;
            timeval dec;  dec.tv_sec = 0; dec.tv_usec=1;
            FAKE_PAST_BEGINNING_TIME -= dec;
         }
         steps.push_back(step);
         last = steps.end();
         --last;
       // tell observers that we have a new position (not necessarily the current one)
         TrackEvent evt(TrackEvent::TRACK_STEP_APPENDED);
         notify(&evt);
      }
   }

   void DiscreteTrack::closeTrack() {
      endOfTrack = last;
      if(steps.size() > 0) {
         FAKE_PAST_END_TIME = ((CoordAndTime*)*last)->time;
         timeval inc;  inc.tv_sec = 0; inc.tv_usec=1;
         FAKE_PAST_END_TIME += inc;
      }
   }

   void DiscreteTrack::reopenTrack() {
      endOfTrack = steps.end();
   }

// take the new current time and issue itself a new CoordAndTime in response
// do not do if out of range or there is already a step at or beyond the updated current time
   CoordAndTime *DiscreteTrack::grow(Timeline *timeline, timeval curtime) {
      if (endOfTrack!=steps.end()) 
         return NULL;
      timeval ftime = finalTime();
      if (ftime>=curtime) 
         return NULL;
      ContactType ctype=getContactType();
      CoordAndTime *c=(CoordAndTime*)*last;
      CoordAndTime *cnew;
      switch (ctype) {
         case CONTACT_UNKNOWN:
            cnew=new CoordAndTime(*(CoordAndTime*)c);
            break;
         case CONTACT_IS_VEHICLE:
         case CONTACT_IS_OWNSHIP:
            cnew=new PreciseContact(*(PreciseContact*)c);
            break;
         case CONTACT_IS_WEAPON:
            cnew=new WeaponContact(*(WeaponContact*)c);
            break;
         case CONTACT_FROM_SENSOR:
            cnew=new ContactFromSensor(*(ContactFromSensor*)c);
            break;
         case HYPOTHETICAL_CONTACT:
            cnew=NULL;//new HypotheticalContact(*(HypotheticalContact*)c);
            break;
         default:
            cnew=new CoordAndTime(*(CoordAndTime*)c);
            break;
      }
      if (cnew!=NULL)
         cnew->time=curtime;
      return cnew;
     //timeline->appendStepToTrack(this, cnew);
   }
