#include "trackview.h"
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMatrixTransform.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/SoInput.h>
#include <Inventor/actions/SoWriteAction.h>

#include "ASU/TargetMotionData.h"
#include "ASU/ContainmentRegion.h"
#include "ASU/MeshFromEnumeratedVolume.h"



#include "degree_conv.h"

#ifdef __linux__
  #include <values.h>
#else
  #include <limits.h>
#endif

#define NAVAL_TO_MATH(alpha) (-alpha+90)
#define KNOTS_TO_MS(kts) (kts*1852.0/3600.0)

#define DEFAULT_BEARING_ERROR 2.2
#define DEFAULT_MAX_RANGE 25000.0
#define DEFAULT_MIN_RANGE 10.0
#define DEFAULT_RANGE_ERROR_PERCENTAGE 0.1


/* A contant used when calling
   SingleObservable::addView() and MultipleObservable::addView()
   to indicate we won't define a user bit mask to tag the observer.
   Tagging the observers is useful to ask "give me all my observers
   of a certain type" */
#define NO_USER_BITMASK 0

#define MIN(a,b) (a>b?b:a)
#define MAX(a,b) (a>b?a:b)

/****************** Iv3dIconColorizer ******************/

   const float Iv3dIconColorizer::FRIENDLY_HSVCOLOR[3] = { 0.666f,    1.0f,    1.0f };
   const float Iv3dIconColorizer::HOSTILE_HSVCOLOR[3] = { 1.0f,    1.0f,    1.0f };
   const float Iv3dIconColorizer::UNKNOWN_THREAT_HSVCOLOR[3] = { 0.152f,    1.0f,    1.0f };
   const float Iv3dIconColorizer::OWNSHIP_RGBCOLOR[3] = { 0.2f   ,    0.2f, 1.0f };
   const float Iv3dIconColorizer::FRIENDLY_RGBCOLOR[3] = { 0.1f,    0.1f,    0.9f };
   const float Iv3dIconColorizer::HOSTILE_RGBCOLOR[3] = { 0.9f,    0.1f,    0.1f };
   const float Iv3dIconColorizer::UNKNOWN_THREAT_RGBCOLOR[3] = { 0.9f,    0.9f,    0.1f };
   const float Iv3dIconColorizer::WEAPON_RGBCOLOR[3]  = { 0.1   , 0.9f,    0.1 };


   void Iv3dIconColorizer::colorize3dIconAs(SoNode* node, ThreatClass threat) {
      switch(threat) {
         case HOSTILE:
            applyHSVColor(node,HOSTILE_HSVCOLOR[0],HOSTILE_HSVCOLOR[1],HOSTILE_HSVCOLOR[2]);
            break;
         case FRIENDLY:
            applyHSVColor(node,FRIENDLY_HSVCOLOR[0],FRIENDLY_HSVCOLOR[1],FRIENDLY_HSVCOLOR[2]);
            break;
         default:
            applyHSVColor(node,UNKNOWN_THREAT_HSVCOLOR[0],UNKNOWN_THREAT_HSVCOLOR[1],UNKNOWN_THREAT_HSVCOLOR[2]);
            break;
      }
   }

   void Iv3dIconColorizer::change3dIconColorTo(SoNode* node, ThreatClass threat) {
     switch(threat) {
         case FRIENDLY:
            changeColorHue(node,FRIENDLY_HSVCOLOR[0]);
            break;
         case HOSTILE:
            changeColorHue(node,HOSTILE_HSVCOLOR[0]);
            break;
         default:
            changeColorHue(node,UNKNOWN_THREAT_HSVCOLOR[0]);
            break;
      }
   }

   const float* Iv3dIconColorizer::getRGBColorForThreatClass(ThreatClass threat) const {
     switch(threat) {
         case FRIENDLY:
            return FRIENDLY_RGBCOLOR;
         case HOSTILE:
            return HOSTILE_RGBCOLOR;
         default:
            return UNKNOWN_THREAT_RGBCOLOR;
      }
   }

/****************** IvContactView ******************/

   SO_NODE_ABSTRACT_SOURCE(IvContactView);

   IvContactView::IvContactView():model(NULL),colorizer(NULL) {
      SO_NODE_CONSTRUCTOR(IvContactView);
      //adds the alternate representation field
      //if a program does not know how to render a node, it looks at the alternate rep field
      SO_NODE_ADD_FIELD(alternateRep,(NULL));
      SO_NODE_ADD_FIELD(outOfRange,(false));
      SO_NODE_ADD_FIELD(userVisible,(false));
      SO_NODE_ADD_FIELD(invisible_count,(0));
      if (SO_NODE_IS_FIRST_INSTANCE()) {
	SoSFBool *showIntersecting=(SoSFBool*)SoDB::createGlobalField("showIntersecting",SoType::fromName("SoSFBool"));
	showIntersecting->setValue(true);
      }
      whichChild = SO_SWITCH_NONE;
      which = new IvSwitchEngine;
      /*setting up the engine that determines whether or not this contactview will be seen*/
      which->showIntersection.connectFrom(SoDB::getGlobalField("showIntersecting"));
      which->intersectionCount.connectFrom(&invisible_count);
      which->outOfRange.connectFrom(&outOfRange);
      which->userVisible.connectFrom(&userVisible);
      whichChild.connectFrom(&which->which);
      grow=false;
      selfIntersect=false;
   }

   void IvContactView::write(SoWriteAction *action) {
     //update the alternate rep to contain a SoSwitch copy of this node
      if(alternateRep.getValue() == NULL) {
         SoSwitch *s=new SoSwitch(*this);
         alternateRep.setValue(s);
      }
      SoSwitch::write(action);
   }

   int IvContactView::getEventsOfInterest() const {
     return TrackEvent::POS_CHANGED|DiscreteTrackEvent::PAST_BEGINNING_OF_TRACK|DiscreteTrackEvent::PAST_END_OF_TRACK;
   }

   void IvContactView::setModel(Observable* aTrack, unsigned int userBitMask) {
      model = (DiscreteTrack*)aTrack;
      model->addView(this,userBitMask,getEventsOfInterest());
   }

   void IvContactView::initClass() {
   // Initialize type id variables
     SO_NODE_INIT_ABSTRACT_CLASS(IvContactView, SoSwitch, "Switch");
   }

   //!sets an override flag which allows the components of an intersection to be visible
   void IvContactView::setShowIntersecting(bool show) {
     SoSFBool *showIntersecting=(SoSFBool*)SoDB::getGlobalField("showIntersecting");
     if (show==showIntersecting->getValue()) return;
     showIntersecting->setValue(show);
   }

   //!keeps a count of objects which tell this object to not be visible
   void IvContactView::setVisible(bool vis) {
     int ic=invisible_count.getValue();
     if (vis) {
       ic=MAX(ic-1,0);
     } else {
       ic++;
     }
     invisible_count.setValue(ic);
   }

   // handles the hiding/showing of the contact icon
   void IvContactView::update(void *evt) {
      switch(((TrackEvent*)evt)->eventType) {
        case DiscreteTrackEvent::PAST_BEGINNING_OF_TRACK:
        case DiscreteTrackEvent::PAST_END_OF_TRACK:
	   outOfRange.setValue(true);
           break;
        default:
	  if (outOfRange.getValue())
	    outOfRange.setValue(false);
     }
   }

/************** IvPreciseContactView **************/

   IvPreciseContactView::IvPreciseContactView():marker(NULL) {
      SoTransformSeparator *sep = new SoTransformSeparator;
      addChild(sep);
      rotationX = new SoRotationXYZ;
      rotationY = new SoRotationXYZ;
      pos = new SoTranslation;
      sep->addChild(pos);
      sep->addChild(rotationX);
      sep->addChild(rotationY);
      scaleFactor[0]=scaleFactor[1]=scaleFactor[2]=1.0f;
   }

   void IvPreciseContactView::update(void *evt) {
      if(((TrackEvent*)evt)->eventType == TrackEvent::POS_CHANGED) {
         CoordAndTime *coord = model->getCurrentCoordAndTime();
         rotationX->axis.setValue(SoRotationXYZ::Z);
         rotationX->angle.setValue(-DEGREES_TO_RADIANS(coord->hpr[0]));
         rotationY->axis.setValue(SoRotationXYZ::X);
         rotationY->angle.setValue(DEGREES_TO_RADIANS(coord->hpr[1]));
         pos->translation.setValue(coord->xyz[0]*scaleFactor[0],coord->xyz[1]*scaleFactor[1],coord->xyz[2]*scaleFactor[2]);
      }
      //else if(((TrackEvent*)evt)->eventType == DiscreteTrackEvent::CONTACT_THREATCLASS_CHANGED) {
      if(colorizer && marker) {
	colorizer->change3dIconColorTo(marker, model->getThreatClass());
      }
	 //}
      // relay message to superclass
      IvContactView::update(evt);
   }

   void IvPreciseContactView::setMarker(SoNode* newMarker) {
      SoTransformSeparator *sep = (SoTransformSeparator*)getChild(0);
      if(marker)
	sep->replaceChild(sep->findChild(marker),newMarker);
      else
	sep->addChild(newMarker);
      if(colorizer) {
         colorizer->colorize3dIconAs(newMarker,model->getThreatClass());
      marker = newMarker;
      }
   }


   CContainmentRegion *IvPreciseContactView::getContainmentRegion() {
     return NULL; }

   void IvPreciseContactView::scaleBy(float newScaleFactorX, float newScaleFactorY, float newScaleFactorZ) {
     SbVec3f position=pos->translation.getValue();
     pos->translation.setValue(position[0]*newScaleFactorX/scaleFactor[0], position[1]*newScaleFactorY/scaleFactor[1], position[2]*newScaleFactorZ/scaleFactor[2]);
     scaleFactor[0]=newScaleFactorX;
     scaleFactor[1]=newScaleFactorY;
     scaleFactor[2]=newScaleFactorZ;
   }


/******************* IvAOUView *******************/

   /* These are class variables of the IvAOUView class */
   SoSeparator* IvAOUView::markerPOS = NULL;
   SoSeparator* IvAOUView::markerLOB = NULL;

   IvAOUView::IvAOUView(float d_max, float d_min):depth_min(d_min),depth_max(d_max) {

      poslobSwitch = new SoSwitch;
      addChild(poslobSwitch);

      SoSeparator *sepPOS = new SoSeparator;
      SoSeparator *sepLOB = new SoSeparator;
      SoSeparator *sepRegion = new SoSeparator;
      poslobSwitch->addChild(sepPOS);
      poslobSwitch->addChild(sepLOB);
      poslobSwitch->addChild(sepRegion);
      pos = new SoTranslation;
      region=NULL;
      curPos=curLob=NULL;

      posMaterial=new SoMaterial;
      lobMaterial=new SoMaterial;
      regionMaterial=new SoMaterial;
      setColor(1,1,1);
      colorizer=NULL;

      //POS stuff
      sepPOS->addChild(posMaterial);
      sepPOS->addChild(pos);

      rotationXY = new SoRotationXYZ;
      rotationXY->axis.setValue(SoRotationXYZ::Z);
      sepPOS->addChild(rotationXY);

      scale = new SoScale;
      sepPOS->addChild(scale);

   // add a 90 degree rotation to put the bottom of cylinder lying on the XY plane
      SoRotationXYZ *straigthen = new SoRotationXYZ;
      straigthen->axis.setValue(SoRotationXYZ::X);
      straigthen->angle.setValue(DEGREES_TO_RADIANS(90));
      sepPOS->addChild(straigthen);


   // 'marker', the AOU shape, is a class singleton, so we check if we already created it
      if(IvAOUView::markerPOS == NULL) {
         IvAOUView::markerPOS = new SoSeparator;
	 SoMaterialBinding *bind=new SoMaterialBinding;
	 bind->value=SoMaterialBinding::PER_PART_INDEXED;
         IvAOUView::markerPOS->addChild(bind);
         IvAOUView::markerPOS->addChild(new SoCylinder);
      }
      sepPOS->addChild(markerPOS);

      //LOB stuff
      sepLOB->addChild(lobMaterial);
      sepLOB->addChild(pos);
      points=new SoCoordinate3;
      sepLOB->addChild(points);
      float pointVals[8][3];
      points->point.setValues(0,8,pointVals);

      if(IvAOUView::markerLOB == NULL) {
         IvAOUView::markerLOB = new SoSeparator;
	 SoMaterialBinding *bind=new SoMaterialBinding;
	 bind->value=SoMaterialBinding::PER_PART_INDEXED;
         IvAOUView::markerLOB->addChild(bind);
	 SoIndexedFaceSet *lob=new SoIndexedFaceSet;
         IvAOUView::markerLOB->addChild(lob);
	 int lobOrder[] ={0,1,2,3,4, SO_END_FACE_INDEX, 5,6,7,8,9, SO_END_FACE_INDEX, 0,5,6,1, SO_END_FACE_INDEX, 1,6,7,2, SO_END_FACE_INDEX, 2,7,8,3, SO_END_FACE_INDEX, 3,8,9,4, SO_END_FACE_INDEX, 4,9,5,0, SO_END_FACE_INDEX };
	 lob->coordIndex.setValues(0,37,lobOrder);
	 int c[]={0,0,1,1,1,1,1};
	 lob->materialIndex.setValues(0,7,c);
      }
      sepLOB->addChild(markerLOB);

      //region stuff
      sepRegion->addChild(regionMaterial);
      sepRegion->addChild(pos);
      regionPoints=new SoCoordinate3;
      sepRegion->addChild(regionPoints);
      regionFaces=new SoIndexedFaceSet;
      sepRegion->addChild(regionFaces);


      scaleFactor[0]=scaleFactor[1]=scaleFactor[2]=1.0f;
  }

   void IvAOUView::setColor(float r, float g, float b) {
     posMaterial->ambientColor.set1Value(0,r,g,b);
     posMaterial->diffuseColor.set1Value(0,r,g,b);
     posMaterial->specularColor.set1Value(0,r,g,b);
     posMaterial->emissiveColor.set1Value(0,r,g,b);
     posMaterial->ambientColor.set1Value(1,r,g,b);
     posMaterial->diffuseColor.set1Value(1,r,g,b);
     posMaterial->specularColor.set1Value(1,r,g,b);
     posMaterial->emissiveColor.set1Value(1,r,g,b);
     posMaterial->ambientColor.set1Value(2,r,g,b);
     posMaterial->diffuseColor.set1Value(2,r,g,b);
     posMaterial->specularColor.set1Value(2,r,g,b);
     posMaterial->emissiveColor.set1Value(2,r,g,b);
     posMaterial->transparency.set1Value(0,0.7f);
     posMaterial->transparency.set1Value(1,0.4f);
     posMaterial->transparency.set1Value(2,0.4f);
     
     lobMaterial->ambientColor.set1Value(0,r,g,b);
     lobMaterial->diffuseColor.set1Value(0,r,g,b);
     lobMaterial->specularColor.set1Value(0,r,g,b);
     lobMaterial->emissiveColor.set1Value(0,r,g,b);
     lobMaterial->ambientColor.set1Value(1,r,g,b);
     lobMaterial->diffuseColor.set1Value(1,r,g,b);
     lobMaterial->specularColor.set1Value(1,r,g,b);
     lobMaterial->emissiveColor.set1Value(1,r,g,b);
     lobMaterial->transparency.set1Value(0,0.4f);
     lobMaterial->transparency.set1Value(1,0.7f);

     regionMaterial->ambientColor.set1Value(0,r,g,b);
     regionMaterial->diffuseColor.set1Value(0,r,g,b);
     regionMaterial->specularColor.set1Value(0,r,g,b);
     regionMaterial->emissiveColor.set1Value(0,r,g,b);
     regionMaterial->transparency.set1Value(0,0.7f);
   }

   void IvAOUView::setMarkerColorUsing(Iv3dIconColorizer *aColorizer) {
     colorizer = aColorizer; 
     if(model && colorizer) {
       const float *color = colorizer->getRGBColorForThreatClass(model->getThreatClass());
       setColor(color[0],color[1],color[2]);
     }
   }

   void IvAOUView::update(void *evt) {
      if(((TrackEvent*)evt)->eventType == TrackEvent::POS_CHANGED) {
	ContactFromSensor *coord = (ContactFromSensor*)model->getCurrentCoordAndTime();
	bool prevCoord=true;
	if (coord->type==LOB) { //LOB
	  if ((!curLob)||(curLob&&((curLob->bearing!=coord->bearing)||(curLob->bearing_err!=coord->bearing_err)||(curLob->range!=coord->range)||(curLob->range_err!=coord->range_err)))) {
	    prevCoord=false;
	    poslobSwitch->whichChild=1;
	    float bear = DEGREES_TO_RADIANS(NAVAL_TO_MATH(coord->bearing));
	    float minBear, maxBear, bearErr;
	    float minBearSin,minBearCos,maxBearSin,maxBearCos;
	    float bearSin, bearCos;
	    minBear=maxBear=bear;
	    if (coord->bearing_err>= 0.1) {
	      bearErr = DEGREES_TO_RADIANS(coord->bearing_err);
	    }
	    else {
	      bearErr = DEGREES_TO_RADIANS(DEFAULT_BEARING_ERROR);
	    }
	    minBear-=bearErr;
	    maxBear+=bearErr;
	    //minBearCos=cos(minBear);
	    //minBearSin=sin(minBear);
	    //maxBearCos=cos(maxBear);
	    //maxBearSin=sin(maxBear);
	    bearCos=cos(bear);
	    bearSin=sin(bear);
	    float errCos=cos(bearErr);
	    float errSin=sin(bearErr);
	    //Sin(X + Y) = Sin(X)Cos(Y) + Cos(X)Sin(Y)
	    //Sin(X - Y) = Sin(X)Cos(Y) - Cos(X)Sin(Y)
	    //Cos(X + Y) = Cos(X)Cos(Y) - Sin(X)Sin(Y)
	    //Cos(X - Y) = Cos(X)Cos(Y) + Sin(X)Sin(Y)
	    float coss=bearCos*errCos;
	    float sins=bearSin*errSin;
	    float sincos=bearSin*errCos;
	    float cossin=bearCos*errSin;
	    minBearCos=coss+sins;
	    maxBearCos=coss-sins;
	    minBearSin=sincos-cossin;
	    maxBearSin=sincos+cossin;
	    
	    float midRange, minRange, maxRange;
	    if (coord->range>= 0) {
	      midRange=coord->range;
	      if (coord->range_err >= 0) {
		minRange=midRange-coord->range_err;
		maxRange=midRange+coord->range_err;
	      }
	      else {
		minRange=midRange+midRange*DEFAULT_RANGE_ERROR_PERCENTAGE;
		maxRange=midRange-midRange*DEFAULT_RANGE_ERROR_PERCENTAGE;
	      }
	    }
	    else {
	      minRange=DEFAULT_MIN_RANGE;
	      maxRange=DEFAULT_MAX_RANGE;
	      midRange=(maxRange+minRange)*0.5;
	    }
	    minRange=METERS_TO_DEGREES(minRange);
	    maxRange=METERS_TO_DEGREES(maxRange);
	    midRange=METERS_TO_DEGREES(midRange);
	    double midx=midRange*bearCos;
	    double midy=midRange*bearSin;
	    coord->xyz[0]=midx+coord->os_lon;//midx and midy need to be converted to degrees
	    coord->xyz[1]=midy+coord->os_lat;
	    pos->translation.setValue(coord->xyz[0]*scaleFactor[0],coord->xyz[1]*scaleFactor[1],-depth_min*scaleFactor[2]);
	    
	    float pointVals[10][3];
	    pointVals[0][2]=pointVals[1][2]=pointVals[2][2]=pointVals[3][2]=pointVals[4][2]=0*scaleFactor[2];
	    pointVals[5][2]=pointVals[6][2]=pointVals[7][2]=pointVals[8][2]=pointVals[9][2]=(depth_min-depth_max)*scaleFactor[2];
	    pointVals[0][0]=pointVals[5][0]=(minBearCos*minRange-midx)*scaleFactor[0];
	    pointVals[0][1]=pointVals[5][1]=(minBearSin*minRange-midy)*scaleFactor[1];
	    
	    pointVals[1][0]=pointVals[6][0]=(maxBearCos*minRange-midx)*scaleFactor[0];
	    pointVals[1][1]=pointVals[6][1]=(maxBearSin*minRange-midy)*scaleFactor[1];
	    
	    pointVals[2][0]=pointVals[7][0]=(maxBearCos*maxRange-midx)*scaleFactor[0];
	    pointVals[2][1]=pointVals[7][1]=(maxBearSin*maxRange-midy)*scaleFactor[1];
	    
	    pointVals[3][0]=pointVals[8][0]=(bearCos*maxRange-midx)*scaleFactor[0];
	    pointVals[3][1]=pointVals[8][1]=(bearSin*maxRange-midy)*scaleFactor[1];
	    
	    pointVals[4][0]=pointVals[9][0]=(minBearCos*maxRange-midx)*scaleFactor[0];
	    pointVals[4][1]=pointVals[9][1]=(minBearSin*maxRange-midy)*scaleFactor[1];
	    points->point.deleteValues(0);
	    points->point.setValues(0,10,pointVals);
	  }
	  else {
	    coord->xyz[0]=curLob->xyz[0];
	    coord->xyz[1]=curLob->xyz[1];
	  }
	  curLob=coord;
	}
	else { //POS
	  if ((!curPos)||(curPos&&((curPos->semi_major_axis!=coord->semi_major_axis)||(curPos->semi_minor_axis!=coord->semi_minor_axis)||(curPos->ellipse_angle!=coord->ellipse_angle)||(curPos->xyz[0]!=coord->xyz[0])||(curPos->xyz[1]!=coord->xyz[1])||(curPos->xyz[2]!=coord->xyz[2])))) {
	    prevCoord=false;
	    poslobSwitch->whichChild=0;
	    scale->scaleFactor.setValue(METERS_TO_DEGREES(coord->semi_minor_axis)*scaleFactor[0], METERS_TO_DEGREES(coord->semi_major_axis)*scaleFactor[1],(depth_max-depth_min)*0.5*scaleFactor[2]);
	    rotationXY->angle.setValue(DEGREES_TO_RADIANS(-coord->ellipse_angle));
	    pos->translation.setValue(coord->xyz[0]*scaleFactor[0],coord->xyz[1]*scaleFactor[1],-(depth_max+depth_min)*0.5*scaleFactor[2]);
	  }
	  curPos=coord;
	}
	if (getGrow()&&(coord->time!=coord->startTime)&&(timevalToDouble(coord->time-coord->startTime)<=600)) {
	  if (region) delete region;
	  region=makeContainmentRegion();
	  //just grow current containment region for passed time
	  CTargetMotionData *theTarget = new CTargetMotionData(1, 20.0);
	  CContainmentRegion *temp = region->expandRegion(theTarget,(int)(timevalToDouble(coord->time)+0.5));
	  delete theTarget;
	  if (temp!=region) {
	    delete region;
	    region=temp;
	  }
	  addMeshFromRegion();
	  //show the mesh rep
	  poslobSwitch->whichChild=2;
	} else {
	  if (getSelfIntersect()) {
	    if (!prevCoord) {
	      //grow current region to new time and intersect it with new rep
	      CContainmentRegion *temp=makeContainmentRegion();
	      if (region!=NULL) {
		CTargetMotionData *theTarget = new CTargetMotionData(1, 20.0);
		CContainmentRegion *temp2 = region->expandRegion(theTarget,(int)(timevalToDouble(coord->time)+0.5));
		delete theTarget;
		if (region!=temp2) {
		  delete region;
		  region=NULL;
		}
		float colors[] = {1.0f, 0.0f, 0.0f};
		float trans = 0.8f;
		region=CContainmentRegion::intersectRegions(0,(int)(timevalToDouble(coord->time)+0.5), colors, trans,temp, temp2);
		addMeshFromRegion();
		poslobSwitch->whichChild=2;
	      } else {
		region=temp;
	      }
	    }
	  } else {
	    //no previous rep, just show regular rep
	    if (!prevCoord) {
	      if (region!=NULL) {
		delete region;
		region=NULL;
	      }
	      region=makeContainmentRegion();
	    }
	  }
	}
      }
      //else if(((TrackEvent*)evt)->eventType == DiscreteTrackEvent::CONTACT_THREATCLASS_CHANGED) {
	if(model && colorizer) {
	  const float* threatColor = colorizer->getRGBColorForThreatClass(model->getThreatClass());
	  setColor(threatColor[0],threatColor[1],threatColor[2]);
	}
	//}
      
      // relay message to superclass
      IvContactView::update(evt);
   }
/*
ASU

Feed data from ContactFromSensor into CContainmentRegion.

*/
   CContainmentRegion *IvAOUView::getContainmentRegion() {
     return region;
   }
//!creates points from feedervAOUView:
   CContainmentRegion *IvAOUView::makeContainmentRegion() {
     ContactFromSensor *coord = (ContactFromSensor*)model->getCurrentCoordAndTime();

     int (*tri_array)[3]=NULL;
     float (*pts)[3]=NULL;
     int numpts=0;
     int npts=0;
     int num_triangles=0;

     int i,j;

     if (coord->type==LOB) { //LOB
       numpts=5;
       npts=2*numpts;
       pts=new float[npts][3];
       for (i=0;i<npts;i++) {
	 pts[i][0]=(points->point[i].getValue()[0]+getPosition().getValue()[0])/scaleFactor[0];
	 pts[i][1]=(points->point[i].getValue()[1]+getPosition().getValue()[1])/scaleFactor[1];
	 if (i<npts/2)
	   pts[i][2]=depth_min;
	 else
	   pts[i][2]=depth_max;

       }

       //num tris and order
       num_triangles = 4*numpts-4;

       tri_array= new int[num_triangles][3];

       //This takes care of the first 4 tris

       for(i=0; i<(numpts-2); i++) {
	   tri_array[i][0] = 0;
		tri_array[i][1] = i+1;
		tri_array[i][2] = (i+2);
	}

       //This takes care of the next
       for(i=0; i<(numpts-2); i++) {
	 tri_array[i+(numpts-2)][0] = numpts;
	 tri_array[i+(numpts-2)][1] = i+(numpts)+1;
	 tri_array[i+(numpts-2)][2] = (i+2+(numpts));
       }

	//This takes care of the last
       for(i=0; i<(numpts); i++) {
	 tri_array[2*i+2*(numpts-2)][0] = i;
	 tri_array[2*i+2*(numpts-2)][1] = (i+1)%(numpts)+(numpts);
	 tri_array[2*i+2*(numpts-2)][2] = (i+1)%(numpts);

	 tri_array[2*i+2*(numpts-2)+1][0] = i;
	 tri_array[2*i+2*(numpts-2)+1][1] = (i+1)%(numpts)+(numpts);
	 tri_array[2*i+2*(numpts-2)+1][2] = i+(numpts);
       }

     } else if (coord->type==POS) { //POS
       float major=scale->scaleFactor.getValue()[1]/scaleFactor[1];
       float minor=scale->scaleFactor.getValue()[0]/scaleFactor[0];
       float b2=minor*minor;
       float a2=major*major;
       float centerx=getPosition().getValue()[0]/scaleFactor[0];
       float centery=getPosition().getValue()[1]/scaleFactor[1];
       int  numy=12;
       float x,y;
       numpts=4*numy;
       npts=2*numpts;
       pts=new float[npts][3];
       for (i=0;i<=numy;i++) {
	 if (i==0) {
	   pts[i][0]=minor;
	   pts[i][1]=0;
	   continue;
	 }
	 if (i==numy) {
	   pts[i][0]=0;
	   pts[i][1]=major;
	   continue;
	 }
	 y=major*(float)(((float)i)/((float)numy));
	 x=sqrt(b2-b2*y*y/a2);
	 pts[i][0]=x;
	 pts[i][1]=y;
       }
       for (i=0;i<numy;i++) {
	 pts[i][0]=pts[i][0];
	 pts[i][1]=pts[i][1];
	 pts[i][2]=0;
       }
       j=0;
       for (i=numy;i>0;i--) {
	 pts[j+numy][0]=-pts[i][0];
	 pts[j+numy][1]=pts[i][1];
	 pts[j+numy][2]=0;
	 j++;
       }
       for (i=0;i<numy;i++) {
	 pts[i+2*numy][0]=-pts[i][0];
	 pts[i+2*numy][1]=-pts[i][1];
	 pts[i+2*numy][2]=0;
       }
       j=0;
       for (i=numy;i>0;i--) {
	 pts[j+3*numy][0]=pts[i][0];
	 pts[j+3*numy][1]=-pts[i][1];
	 pts[j+3*numy][2]=0;
	 j++;
       }

       float angle=DEGREES_TO_RADIANS(coord->ellipse_angle);
       float co=cos(angle);
       float si=sin(angle);
       for (i=0;i<numpts;i++) {
	 x=co*pts[i][0]+si*pts[i][1];
	 y=co*pts[i][1]-si*pts[i][0];
	 pts[i][0]=x+centerx;
	 pts[i][1]=y+centery;
	 pts[i][2]=depth_min;
	 pts[numpts+i][0]=x+centerx;
	 pts[numpts+i][1]=y+centery;
	 pts[numpts+i][2]=depth_max;
       }

       num_triangles = 4*(numpts)-4;

       tri_array=new int[num_triangles][3];

	//This takes care of the first 4 tris

	for(i=0; i<(numpts-2); i++) {
	  tri_array[i][0] = 0;
	  tri_array[i][1] = i+1;
	  tri_array[i][2] = (i+2);
	}

	//This takes care of the next
	for(i=0; i<(numpts-2); i++) {
	  tri_array[i+(numpts-2)][0] = numpts;
	  tri_array[i+(numpts-2)][1] = i+(numpts)+1;
	  tri_array[i+(numpts-2)][2] = (i+2+(numpts));
	}

	//This takes care of the last
	for(i=0; i<(numpts); i++) {
	  tri_array[2*i+2*(numpts-2)][0] = i;
	  tri_array[2*i+2*(numpts-2)][1] = (i+1)%(numpts);
	  tri_array[2*i+2*(numpts-2)][2] = (i+1)%(numpts)+(numpts);

	  tri_array[2*i+2*(numpts-2)+1][0] = i;
	  tri_array[2*i+2*(numpts-2)+1][1] = (i+1)%(numpts)+(numpts);
	  tri_array[2*i+2*(numpts-2)+1][2] = i+(numpts);
	}
     }
     else {
       numpts=0;
       return NULL;
     }

     float colors[] = {1.0f, 0.0f, 0.0f};
     float trans = 0.8f;

     long int firstTime  = (int)(timevalToDouble(coord->startTime)+0.5);
     const char *name=getModel()->getName().c_str();
     char *id= new char[strlen(name)+1];
     strcpy(id,name);

     float minLat=90,maxLat=-90,minLon=180,maxLon=-180;
     float lat,lon;
     for (i=0;i<npts;i++) {
       lat=pts[i][1];
       lon=pts[i][0];
       if (lat<minLat) minLat=lat;
       if (lat>maxLat) maxLat=lat;
       if (lon<minLon) minLon=lon;
       if (lon>maxLon) maxLon=lon;
     }
     CContainmentRegion *firstVolume = new CContainmentRegion( 0,id, firstTime,minLat, maxLat,minLon, maxLon,250.0, colors, trans, 1, 0);

     //Convert the input data into an enumerated volume
     firstVolume->makeRegionFromMesh(pts,npts,tri_array,num_triangles);
     if (tri_array!=NULL)
       delete tri_array;
     if (pts!=NULL)
       delete pts;

     return firstVolume;

   }


//takes the ContainmentRegion region and adds its mesh to the mesh rep node.
   void IvAOUView::addMeshFromRegion() {
     if (region==NULL) return;
     int i;
     ContactFromSensor *coord = (ContactFromSensor*)model->getCurrentCoordAndTime();
     CMeshFromEnumeratedVolume *thirdMesh = new CMeshFromEnumeratedVolume(region);
     int numVerts = thirdMesh->getNumVertsInMesh();
     int numTriangles = thirdMesh->getNumTrianglesInMesh();

     if ((numVerts<=0)||(numTriangles<=0)) {
       delete region;
       region=NULL;
       return;
     }

     //Dynamically sized and allocated 2 dimensional array!
     float (*verticeList)[3] = new float[numVerts][3];
     int (*cornerList)[3] = new int[numTriangles][3];
     //End super array stuff.
     
     thirdMesh->getMesh(verticeList, cornerList);


     int *indices=new int[4*numTriangles];
     //Retrieves indexes, sorts them, and inserts Inventor flag.
     for (i=0;i<numTriangles;i++) {
       indices[4*i]=cornerList[i][0];
       indices[4*i+1]=cornerList[i][1];
       indices[4*i+2]=cornerList[i][2];
       indices[4*i+3]=SO_END_FACE_INDEX;
     }


     // this loop grabs the vertex list, scales the vertices, and translates them
     for (i=0; i<numVerts; i++) {
       verticeList[i][0]=verticeList[i][0]*scaleFactor[0]-pos->translation.getValue()[0];
       verticeList[i][1]=verticeList[i][1]*scaleFactor[1]-pos->translation.getValue()[1];
       verticeList[i][2]=-verticeList[i][2]*scaleFactor[2]-pos->translation.getValue()[2];
     }

     regionPoints->point.deleteValues(0);
     regionFaces->coordIndex.deleteValues(0);

     regionPoints->point.setValues(0,numVerts,verticeList);
     regionFaces->coordIndex.setValues(0,numTriangles*4,indices);

     delete thirdMesh;
     delete[] verticeList;
     delete[] cornerList;
     delete[] indices;
   }


   void IvAOUView::scaleBy(float newScaleFactorX, float newScaleFactorY, float newScaleFactorZ) {
     SbVec3f position=pos->translation.getValue();
     float scaleF[]={newScaleFactorX/scaleFactor[0], newScaleFactorY/scaleFactor[1], newScaleFactorZ/scaleFactor[2]};
     pos->translation.setValue(position[0]*scaleF[0], position[1]*scaleF[1], position[2]*scaleF[2]);
     if (model!=NULL) {
       ContactFromSensor *coord = (ContactFromSensor*)model->getCurrentCoordAndTime();
       if (coord!=NULL) {
	 if (coord->type==POS) {
	   position=scale->scaleFactor.getValue();
	   scale->scaleFactor.setValue(position[0]*scaleF[0], position[1]*scaleF[1], position[2]*scaleF[2]);
	 } else {
	   float p[10][3];
	   for (int i=0;i<10;i++) {
	     p[i][0]=points->point[i][0]*scaleF[0];
	     p[i][1]=points->point[i][1]*scaleF[1];
	     p[i][2]=points->point[i][2]*scaleF[2];
	   }
	   points->point.deleteValues(0);
	   points->point.setValues(0,10,p);
	 }
       }
       if (regionPoints!=NULL) {
	 int end=regionPoints->point.getNum();
	 if (end!=0) {
	   float (*p)[3]=new float[end][3];
	   for (int i=0;i<end;i++) {
	     p[i][0]=regionPoints->point[i][0]*scaleF[0];
	     p[i][1]=regionPoints->point[i][1]*scaleF[1];
	     p[i][2]=regionPoints->point[i][2]*scaleF[2];
	   }
	   regionPoints->point.deleteValues(0);
	   regionPoints->point.setValues(0,end,p);
	   delete[] p;
	 }
       }
     }
     scaleFactor[0]=newScaleFactorX;
     scaleFactor[1]=newScaleFactorY;
     scaleFactor[2]=newScaleFactorZ;
   }

/******************* IvHypotheticalView *******************/

   SoMaterial* IvHypotheticalView::material = NULL;

   IvHypotheticalView::IvHypotheticalView(NameToTrackMapping *mapping):map(mapping) {
     pos=new SoTranslation;
     points=new SoCoordinate3;
     faces=new SoIndexedFaceSet;
     SoSeparator *sep=new SoSeparator;
     depth_min=depth_max=0;
     if (material==NULL) {
       material = new SoMaterial;
       material->ambientColor.set1Value(0,1.0,0.3,1.0f);
       material->diffuseColor.set1Value(0,1.0,0.3,1.0f);
       material->specularColor.set1Value(0,1.0,0.3,1.0f);
       material->emissiveColor.set1Value(0,1.0,0.3,1.0f);
       /*       material->ambientColor.set1Value(1,0,0,1.0f);
		material->diffuseColor.set1Value(1,0,0,1.0f);
		material->specularColor.set1Value(1,0,0,1.0f);
		material->emissiveColor.set1Value(1,0,0,1.0f);
		material->transparency.set1Value(0,0.4f);
		material->transparency.set1Value(1,0.7f);
       */
       material->transparency.set1Value(0,0.7f);
     }
     SoMaterialBinding *bind=new SoMaterialBinding;
     //     bind->value=SoMaterialBinding::PER_PART_INDEXED;
     bind->value=SoMaterialBinding::OVERALL;
     sep->addChild(material);
     sep->addChild(bind);
     sep->addChild(pos);
     sep->addChild(points);
     sep->addChild(faces);
     addChild(sep);
     scaleFactor[0]=scaleFactor[1]=scaleFactor[2]=1.0f;
     cur=NULL;
   }

   void IvHypotheticalView::update(void *evt) {
     if(((TrackEvent*)evt)->eventType == TrackEvent::POS_CHANGED) {
       //check if update is a change in the components in this hypothetical
       if(((ObservableTrack*)((TrackEvent*)evt)->track)->getContactType() != HYPOTHETICAL_CONTACT) {
	 makeIntersection();
       } else {
	 HypotheticalContact *coord = (HypotheticalContact*)model->getCurrentCoordAndTime();
	 bool oldCoord=false;
	 if (cur) {
	   if (coord->numContacts==cur->numContacts) {
	     oldCoord=true;
	     for (int i=0;i<coord->numContacts;i++) {
	       if (strcmp(coord->contactList[i],cur->contactList[i])) oldCoord=false;
	     }
	   }
	 }
	 if (!oldCoord) {
	   removeAssoc();
	   for (int i=0;i<coord->numContacts;i++) {
	     makeAssoc(coord->contactList[i]);
	   }
	   makeIntersection();
	   IvContactView::update(evt);
	   if (outOfRange.getValue())
	     outOfRange.setValue(false);
	   cur=coord;
	 }
       }
     } else if  (((TrackEvent*)evt)->eventType == DiscreteTrackEvent::PAST_BEGINNING_OF_TRACK) {
       if(((ObservableTrack*)((TrackEvent*)evt)->track)->getContactType() == HYPOTHETICAL_CONTACT) {
	 removeAssoc();
	 outOfRange.setValue(true);
       }
     }
   }

//clear list of included contacts/unregister for their updates
   void IvHypotheticalView::removeAssoc() {
      AssocList::iterator it = assoc.begin();
      while(it != assoc.end()) {
	((ObservableTrack*)*it)->removeView(this);
	IvContactView *o=(IvContactView*)((ObservableTrack *)*it)->findObserverWithMask(OBSERVER_IS_CONTACTVIEW, NULL);
	if (o)
	  o->setVisible(true);
	it++;
      }
      assoc.clear();
   }

//!add to list of included contacts, look up its track, and register for its updates
   void IvHypotheticalView::makeAssoc(char *name) {
     if (strlen(name)<=0) return;
     string n(name);
     string s("");
     ObservableTrack *a=map->findTrackForContact(n,s);
     if (a==NULL) return;
     TrackTraceDecorator *d=dynamic_cast<TrackTraceDecorator *>(a);
     if (d!=NULL)
       a=d->getTrack();
     //assoc.push_back(a);
     //a->addView(this, NO_USER_BITMASK);
     IvContactView *o=(IvContactView*)a->findObserverWithMask(OBSERVER_IS_CONTACTVIEW, NULL);
     if ((o)&&(o!=this)) {
       assoc.push_back(a);
       a->addView(this, NO_USER_BITMASK);
       o->setVisible(false);
     }
   }

//!compose an intersection region from the included contacts.  Represented as an indexed face set.
   void IvHypotheticalView::makeIntersection() {

     float colors[] = {1.0f, 0.0f, 0.0f};
     float trans = 0.8f;
     CContainmentRegion *r1, *r2, *trigger;
     HypotheticalContact *coord = (HypotheticalContact*)model->getCurrentCoordAndTime();
     if (coord==NULL) return;
     points->point.deleteValues(0);
     faces->coordIndex.deleteValues(0);
     region=NULL;
     if (assoc.empty()) return;
     AssocList::iterator it = assoc.begin();
     IvContactView *o=(IvContactView*)((ObservableTrack*)*it)->findObserverWithMask(OBSERVER_IS_CONTACTVIEW, NULL);
     it++;
     r1=o->getContainmentRegion();
     trigger=r1;
     if (r1==NULL) return;
     while(it != assoc.end()) {
       IvContactView *o=(IvContactView*)((ObservableTrack*)*it)->findObserverWithMask(OBSERVER_IS_CONTACTVIEW, NULL);
       it++;
       r2=o->getContainmentRegion();
       if (r2==NULL) return;
       CContainmentRegion *r3 = CContainmentRegion::intersectRegions(0, (int)(timevalToDouble(coord->time)+0.5), colors, trans,r1, r2);
       if (r3==NULL) return;
       if ((r1!=NULL)&&(r1!=trigger))
	 delete r1;
       r1=r3;
     }
     if (r1==NULL) return;
     region=r1;
     addMeshFromRegion();

   }


//!takes the ContainmentRegion region and adds its mesh to the mesh rep node.
   void IvHypotheticalView::addMeshFromRegion() {
     int i;
     HypotheticalContact *coord = (HypotheticalContact*)model->getCurrentCoordAndTime();

     CMeshFromEnumeratedVolume *thirdMesh = new CMeshFromEnumeratedVolume(region);
     int numVerts = thirdMesh->getNumVertsInMesh();
     int numTriangles = thirdMesh->getNumTrianglesInMesh();
     
     if ((numVerts<=0)||(numTriangles<=0)) {
       region=NULL;
       return;
     }

     //Dynamically sized and allocated 2 dimensional array!
     float (*verticeList)[3] = new float[numVerts][3];
     int (*cornerList)[3] = new int[numTriangles][3];
     
     thirdMesh->getMesh(verticeList, cornerList);


     int *indices=new int[4*numTriangles];
     //Retrieves indexes, sorts them, and inserts Inventor flag.
     for (i=0;i<numTriangles;i++) {
       indices[4*i]=cornerList[i][0];
       indices[4*i+1]=cornerList[i][1];
       indices[4*i+2]=cornerList[i][2];
       indices[4*i+3]=SO_END_FACE_INDEX;
     }
     
     // find the x/y center of mass of the region.
     float com_x, com_y, sum_x, sum_y;
     

     for (i=0; i<numVerts; i++) {
       sum_x += verticeList[i][0];
       sum_y += verticeList[i][1];
     }
     
     com_x = sum_x/numVerts;
     com_y = sum_y/numVerts;
     
     // coord is a current coord and time variable
     coord->xyz[0]=com_x;
     coord->xyz[1]=com_y;
     coord->xyz[2]=0;
     
     // now that the center is found, use it as a frame of reference.
     pos->translation.setValue(coord->xyz[0]*scaleFactor[0], coord->xyz[1]*scaleFactor[1], coord->xyz[2]*scaleFactor[2]);
     
     // this loop grabs the vertex list, scales the vertices, and translates them
     for (i=0; i<numVerts; i++) {
       verticeList[i][0]=verticeList[i][0]*scaleFactor[0]-pos->translation.getValue()[0];
       verticeList[i][1]=verticeList[i][1]*scaleFactor[1]-pos->translation.getValue()[1];
       verticeList[i][2]=-verticeList[i][2]*scaleFactor[2]-pos->translation.getValue()[2];
     }
     
     points->point.setValues(0,numVerts,verticeList);
     faces->coordIndex.setValues(0,numTriangles*4,indices);

     delete thirdMesh;
     delete[] verticeList;
     delete[] cornerList;
     delete[] indices;
   }




   CContainmentRegion *IvHypotheticalView::getContainmentRegion() {
     return region;
   }

   CContainmentRegion *IvHypotheticalView::makeContainmentRegion() {
     return region;
   }

   void IvHypotheticalView::scaleBy(float newScaleFactorX, float newScaleFactorY, float newScaleFactorZ) {
     SbVec3f position=pos->translation.getValue();
     float scale[] = {newScaleFactorX/scaleFactor[0],newScaleFactorY/scaleFactor[1],newScaleFactorZ/scaleFactor[2]};
     int numpts=points->point.getNum();
     float (*pts)[3]=new float[numpts][3];
     for (int i=0;i<numpts;i++) {
       pts[i][0]=points->point[i][0]*scale[0];
       pts[i][1]=points->point[i][1]*scale[1];
       pts[i][2]=points->point[i][2]*scale[2];
     }
     points->point.deleteValues(0);
     points->point.setValues(0,numpts,pts);
     pos->translation.setValue(position[0]*scale[0], position[1]*scale[1], position[2]*scale[2]);
     scaleFactor[0]=newScaleFactorX;
     scaleFactor[1]=newScaleFactorY;
     scaleFactor[2]=newScaleFactorZ;
   }

/******************* IvTrackView *******************/

   IvTrackView::IvTrackView():dist_step(5.0),marker(NULL),colorizer(NULL) {
      whichChild=SO_SWITCH_ALL;
      posOfLastMarker.setValue(FLT_MIN,FLT_MIN,FLT_MIN);
      scaleFactor[0]=scaleFactor[1]=scaleFactor[2]=1.0f;
      // add a multiple transformation object; will contain 1 transform per marker position
      markerPositions = new SoMultipleCopy;
      // SoMultipleCopy adds a matrix at the beginning without asking. Remove it
      markerPositions->matrix.deleteValues(0,1);
      addChild(markerPositions);

      // add the line trace
      trace = new SoLineSet;
      trace->numVertices.setValue(0);
      SoVertexProperty * traceVertices = new SoVertexProperty;
      tracePoints = &traceVertices->vertex;
      traceColor = &traceVertices->orderedRGBA;
      setTraceColor(0xFF, 0xFF, 0xFF); // default trace color = white
      trace->vertexProperty = traceVertices;
      /* Next 2 lines are a temporary hack. SoLineSet turns some GL property
         off when the line has no normals,  and I still didn't find which one.
         Changing this property without having normals has the nasty side effect of screwing the colors
         in the HUD (Fernando). So, here we're adding normals. */
      traceVertices->normal.setValue(0,0,1);
      traceVertices->normalBinding = SoVertexProperty::OVERALL;
      addChild(trace);
   }

   void IvTrackView::setMarker(SoNode* aMarker) {
      marker = aMarker;
      marker->ref();
      if(colorizer && marker) {
         colorizer->colorize3dIconAs(marker,model->getThreatClass());
      }
   }


   void IvTrackView::setTraceColor(float r, float g, float b) {
      unsigned char red = (char)(r*255.0f), green = (char)(g*255.0f), blue = (char)(b*255.0f);
      traceColor->setValue((red << 24) | (green << 16) | (blue << 8) | 0xFF);
   }

   void IvTrackView::setTraceColorUsing(Iv3dIconColorizer *aColorizer) { 
      colorizer = aColorizer; 
      if(model && colorizer) {
        const float *color = colorizer->getRGBColorForThreatClass(model->getThreatClass());
        setTraceColor(color[0],color[1],color[2]);
      }
   }

   void IvTrackView::setModel(Observable* aTrack) {
      model = (TrackTraceDecorator*)aTrack;
      model->addView(this, OBSERVER_IS_TRACKVIEW, (TrackEvent::EventType)(TrackEvent::TRACK_STEP_APPENDED | TrackEvent::TRACK_STEP_REMOVED));
   }

   void IvTrackView::update(void *aTrackEvent) {
      float x,y,z, dist;
      CoordAndTime coord = *(model->getCurrentCoordAndTime());
      if(coord.time != NO_EVENT) { // means there is no current position
      // find distance between last marker and current position
         x = coord.xyz[0]-posOfLastMarker[0]/scaleFactor[0];
         y = coord.xyz[1]-posOfLastMarker[1]/scaleFactor[1];
         z = coord.xyz[2]-posOfLastMarker[2]/scaleFactor[2];
         dist = (float)sqrt(x*x+y*y+z*z);
         switch( ((TrackEvent*)aTrackEvent)->eventType) {
            case TrackEvent::TRACK_STEP_APPENDED:
               coord.xyz[0]*=scaleFactor[0];
               coord.xyz[1]*=scaleFactor[1];
               coord.xyz[2]*=scaleFactor[2];
               if(dist >= dist_step) {
                  posOfLastMarker = coord.xyz;
                  addMark(coord);
               }
               addLineSegment(&coord);
               break;
            case TrackEvent::TRACK_STEP_REMOVED:
               removeLastLineSegment();
               if(markerPositions->matrix.getNum() > 0) {
                  if(dist <= 0.005) {
                     removeLastMark();
                  }


                  // check again, we may have just removed the last marker
                  if(markerPositions->matrix.getNum() > 0) {
                  /* since we don't keep a list of the markers' past positions,
                     we get position of previous marker from position matrix */
                     const SbMatrix& m = markerPositions->matrix[markerPositions->matrix.getNum()-1];
                     posOfLastMarker.setValue(m[3][0],m[3][1],m[3][2]);
                  }
                  else {
                     posOfLastMarker.setValue(FLT_MIN,FLT_MIN,FLT_MIN);
                  }
               }
               break;
         }
      }
      //else if(((TrackEvent*)aTrackEvent)->eventType == DiscreteTrackEvent::CONTACT_THREATCLASS_CHANGED) {
         if(colorizer) {
            if(marker)
               colorizer->change3dIconColorTo(marker, model->getThreatClass());
            const float* threatColor = colorizer->getRGBColorForThreatClass(model->getThreatClass());
            setTraceColor(threatColor[0],threatColor[1],threatColor[2]);
         }
	 //}
   }

   void IvTrackView::addMark(const Coord& coord) {
     if (marker) {
       float pos[3];
       SbMatrix m, m2;
       SbVec3f axis(0,0,1);
       SbRotation r(axis,-DEGREES_TO_RADIANS(coord.hpr[0]));
       m.setRotate(r);
       axis.setValue(0,1,0);
       r.setValue(axis,DEGREES_TO_RADIANS(coord.hpr[1]));
       m2.setRotate(r);
       m.multRight(m2);
       axis.setValue(0,1,0);
       r.setValue(axis,DEGREES_TO_RADIANS(coord.hpr[2]));
       m2.setRotate(r);
       m.multRight(m2);
       pos[0]=coord.xyz[0];
       pos[1]=coord.xyz[1];
       pos[2]=coord.xyz[2];
       m2.setTranslate(pos);
       m.multRight(m2);
       if(markerPositions->matrix.getNum() == 0)
         markerPositions->addChild(marker);
       markerPositions->matrix.set1Value(markerPositions->matrix.getNum(),m);
     }
   }

   void IvTrackView::removeLastMark() {
      if(markerPositions->matrix.getNum() > 0) {
         markerPositions->matrix.deleteValues(markerPositions->matrix.getNum()-1,1);
         if(markerPositions->matrix.getNum() == 0)
            markerPositions->removeChild(0);
      }
   }

   void IvTrackView::addLineSegment(Coord* pos) {
      int pts = trace->numVertices[0];
      tracePoints->set1Value(pts,pos->xyz);
      trace->numVertices.setValue(pts+1);
   }

   void IvTrackView::removeLastLineSegment() {
      int pts = tracePoints->getNum()-1;
      if(pts >= 0) {
         tracePoints->deleteValues(pts,1);
         trace->numVertices.setValue(pts);
      }
   }


//need to scale the matrices 1 at a time
//otherwise if multiple threads attempt to modify them at the same time,
//a crash will occur.  this problem occurs when you use startEditing and finish Editing to reserve the markerPositions matrices
   void IvTrackView::scaleBy(float newScaleFactorX, float newScaleFactorY, float newScaleFactorZ) {
      SbMatrix matrix, new_matrix;
      int size=markerPositions->matrix.getNum();
      float factX,factY,factZ;
      factX=newScaleFactorX/scaleFactor[0];
      factY=newScaleFactorY/scaleFactor[1];
      factZ=newScaleFactorZ/scaleFactor[2];
      scaleFactor[0]=newScaleFactorX;
      scaleFactor[1]=newScaleFactorY;
      scaleFactor[2]=newScaleFactorZ;
      for(unsigned int i=0;i<size;i++) {
         matrix=markerPositions->matrix[i];
         new_matrix.setValue(matrix);
         new_matrix[3][0]*=factX;
         new_matrix[3][1]*=factY;
         new_matrix[3][2]*=factZ;
         markerPositions->matrix.set1Value(i,new_matrix);
      }
     //now scale the lines
      int pts = trace->numVertices[0];
      SbVec3f pt;
      for(unsigned int i=0;i<pts;i++) {
         pt=*(tracePoints->getValues(i));
         pt[0]*=factX;
         pt[1]*=factY;
         pt[2]*=factZ;
         tracePoints->set1Value(i,pt);
      }
   }

   IvTrackView::~IvTrackView() {
      if (marker) marker->unref();
   }


   /******************* IvWPNView *******************/

   IvWPNView::IvWPNView():marker(NULL) {
      sep = new SoTransformSeparator;
      addChild(sep);
      rotationX = new SoRotationXYZ;
      rotationY = new SoRotationXYZ;
      pos = new SoTranslation;
      sep->addChild(pos);
      sep->addChild(rotationX);
      sep->addChild(rotationY);

      SoSeparator *covSep = new SoSeparator;
      sep->addChild(covSep);

      covMaterial=new SoMaterial;
      setColor(0.5,0.5,0.5);
      covSep->addChild(covMaterial);

      covPts = new SoCoordinate3;
      covSep->addChild(covPts);
      int ptOrder[] = {0,1,5,SO_END_FACE_INDEX,1,2,6,5,SO_END_FACE_INDEX,0,1,2,3,SO_END_FACE_INDEX,0,5,6,4,SO_END_FACE_INDEX,0,3,4,SO_END_FACE_INDEX,3,4,6,2,SO_END_FACE_INDEX};
      SoIndexedFaceSet *cov = new SoIndexedFaceSet;
      covSep->addChild(cov);
      cov->coordIndex.setValues(0,28,ptOrder);
      scaleFactor[0]=scaleFactor[1]=scaleFactor[2]=1.0f;
   }

void IvWPNView::setColor(float r, float g, float b){
     covMaterial->ambientColor.set1Value(0,r,g,b);
     covMaterial->diffuseColor.set1Value(0,r,g,b);
     covMaterial->specularColor.set1Value(0,r,g,b);
     covMaterial->emissiveColor.set1Value(0,r,g,b);
     covMaterial->ambientColor.set1Value(1,r,g,b);
     covMaterial->diffuseColor.set1Value(1,r,g,b);
     covMaterial->specularColor.set1Value(1,r,g,b);
     covMaterial->emissiveColor.set1Value(1,r,g,b);
     covMaterial->transparency.set1Value(0,0.3f);
     covMaterial->transparency.set1Value(1,0.7f);
}

   void IvWPNView::update(void *evt) {
      if(((TrackEvent*)evt)->eventType == TrackEvent::POS_CHANGED) {
         CoordAndTime *coord = model->getCurrentCoordAndTime();
         rotationX->axis.setValue(SoRotationXYZ::Z);
         rotationX->angle.setValue(-DEGREES_TO_RADIANS(coord->hpr[0]));
         rotationY->axis.setValue(SoRotationXYZ::X);
         rotationY->angle.setValue(DEGREES_TO_RADIANS(coord->hpr[1]));
         pos->translation.setValue(coord->xyz[0]*scaleFactor[0],coord->xyz[1]*scaleFactor[1],coord->xyz[2]*scaleFactor[2]);
	float x, y, z,a;
	WeaponContact *wcoord = (WeaponContact *) coord;
	switch(wcoord->Mode){
	  case 0:
	  case 1:
	  case 2:
	  case 3:
	  case 4:
	  case 5:
	    	rangebear2xy(0.0,0.0, x,y);
		rangebear2xy(0.0,0.0, z,a);
	        break;
	  case 6:
	  	rangebear2xy(1500.0,36.0, x,y);
		rangebear2xy(1500.0,16.0, z,a);
  		//rangebear2xy(1500.0,5.0, z,a);
		break;
	  case 7: // COVER_SEARCH
	  	rangebear2xy(3000.0,36.0, x,y);
		rangebear2xy(3000.0,16.0, z,a);
		//rangebear2xy(1500.0,5.0, z,a);
		break;
	  case 10: // TERMINAL_HOMING
	    	rangebear2xy(1500.0,36.0, x,y);
		rangebear2xy(1500.0,16.0, z,a);
		//rangebear2xy(1500.0,5.0, z,a);
		break;
	  case 15: // FENCE_DEPLOYMNET
	        createFence();
		break;
	  case 88: // SHUTDOWN
           	if (findChild(fence)!=-1)
		  removeChild(fence);
		break;
	   case 99: 
	        rangebear2xy(15.0,36.0, x,y);
		rangebear2xy(15.0,16.0, z,a);
		break;	
	  default:
		rangebear2xy(0.0,0.0, x,y);
		rangebear2xy(0.0,0.0, z,a);
	        break;
	}

	x=METERS_TO_DEGREES(x)*scaleFactor[0];
	y=METERS_TO_DEGREES(y)*scaleFactor[1];
	if (z!=0.0f) {
	  double invz = 1.0/z;
	  float pts[7][3];
	  pts[0][0]=pts[0][1]=pts[0][2]=0.0;
	  pts[3][0]=x;
	  pts[3][1]=y;
	  pts[3][2]=-z;
	  pts[4][0]=-x;
	  pts[4][1]=y;
	  pts[4][2]=-z;
	  float depth =-coord->xyz[2];
	  if (z<depth){
	    pts[1][0]=pts[2][0]=x;
	    pts[1][1]=pts[2][1]=y;
	    pts[1][2]=pts[2][2]=z;
	    pts[5][0]=pts[6][0]=-x;
	    pts[5][1]=pts[6][1]=y;
	    pts[5][2]=pts[6][2]=z;
	  }
	  else{
	    pts[1][0]= depth*x*invz;
	    pts[1][1]= depth*y*invz;
	    pts[1][2]= depth;
	    pts[2][0]= x;
	    pts[2][1]=y;
	    pts[2][2]=depth;
	    pts[5][0]=-depth*x*invz;
	    pts[5][1]=depth*y*invz;
	    pts[5][2]=depth;
	    pts[6][0]= -x;
	    pts[6][1]=y;
	    pts[6][2]=depth;
	  }
	  for (int n=0;n<7;n++)
	    pts[n][2]*=scaleFactor[2];
	  covPts->point.deleteValues(0);
	  covPts->point.setValues(0,7,pts);
	}
	else {
	  float pts[7][3];
	  for (int n=0;n<7;n++)
	    pts[n][1]=pts[n][2]=pts[n][3]=0;
	  covPts->point.deleteValues(0);
	  covPts->point.setValues(0,7,pts);
	}

      }
      //else if(((TrackEvent*)evt)->eventType == DiscreteTrackEvent::CONTACT_THREATCLASS_CHANGED) {
      if(colorizer && marker) {
	colorizer->change3dIconColorTo(marker, model->getThreatClass());
      }
	 //}
      // relay message to superclass
      IvContactView::update(evt);
   }

   void IvWPNView::setMarker(SoNode* newMarker) {
      SoTransformSeparator *sep = (SoTransformSeparator*)getChild(0);
      if(marker)
	sep->replaceChild(sep->findChild(marker),newMarker);
      else
	sep->addChild(newMarker);
      if(colorizer) {
         colorizer->colorize3dIconAs(newMarker,model->getThreatClass());
      marker = newMarker;
      }
   }


   CContainmentRegion *IvWPNView::getContainmentRegion() {
     return NULL; }

   void IvWPNView::scaleBy(float newScaleFactorX, float newScaleFactorY, float newScaleFactorZ) {
     float scale[] = {newScaleFactorX/scaleFactor[0],newScaleFactorY/scaleFactor[1],newScaleFactorZ/scaleFactor[2]};
     SbVec3f position=pos->translation.getValue();
     pos->translation.setValue(position[0]*scale[0], position[1]*scale[1], position[2]*scale[2]);
     if (covPts->point.getNum()>=7) {
       float pts[7][3];
       for (int i=0;i<7;i++) {
	 pts[i][0]=covPts->point[i][0]*scale[0];
	 pts[i][1]=covPts->point[i][1]*scale[1];
	 pts[i][2]=covPts->point[i][2]*scale[2];
       }
       covPts->point.deleteValues(0);
       covPts->point.setValues(0,7,pts);
     }
     scaleFactor[0]=newScaleFactorX;
     scaleFactor[1]=newScaleFactorY;
     scaleFactor[2]=newScaleFactorZ;
   }


/*Function to convert range(meters) and bearing(degrees) to x and y in meters.
 //Weapon coverage description:
  // range(search) = 3000.0
  // range(terminal homing) = 1500.0
  //bearing(x-width) = +/- 36.0
  // bearing(z-depth) = +/- 16.0
*/
void IvWPNView::rangebear2xy(float range, float bearing, float& x, float& y){

    //	 Convert from degrees to radians;
    float pi = 3.1459;
    float theta = bearing * pi/180;

    float hypotenus;
    hypotenus = range / cos(theta);

    y = hypotenus * cos(theta);
    x = hypotenus * sin(theta);

}


void IvWPNView::createFence(){

      fence = new SoSeparator;
      addChild(fence);

      covMaterial=new SoMaterial;
      setColor(1.0,1.0,1.0);
      covMaterial->transparency.set1Value(0,0.7f);
      covMaterial->transparency.set1Value(1,1.0f);
      fence->addChild(covMaterial);

      SoTranslation *fencePos=new SoTranslation;
      fencePos->translation.setValue(pos->translation.getValue());
      fence->addChild(fencePos);

      fenceLength = 3000;
      fenceHeight = 500;
      fenceWidth = 10;
      SoScale *fenceScale = new SoScale;
      fenceScale->scaleFactor.setValue(fenceWidth,fenceLength,fenceHeight);
      SoCube *cov = new SoCube;

      // creating the one general fence side.
      SoGroup *side = new SoGroup;
      side->addChild(fenceScale);
      side->addChild(cov);

      // building back side of the fence
      SoSeparator *backSide = new SoSeparator;
      SoRotationXYZ *straigthen = new SoRotationXYZ;
      straigthen->axis.setValue(SoRotationXYZ::Z);
      straigthen->angle.setValue(DEGREES_TO_RADIANS(90));
      backSide->addChild(straigthen);
      backSide->addChild(side);


      // building  left side of the fence
      SoSeparator *leftSide = new SoSeparator;
      SoRotationXYZ *leftStraigthen = new SoRotationXYZ;
      leftStraigthen->axis.setValue(SoRotationXYZ::Z);
      leftStraigthen->angle.setValue(DEGREES_TO_RADIANS(180));

      SoTransform *leftTrans = new SoTransform;
      leftTrans->translation = SbVec3f(fenceLength, fenceLength, 0);

      leftSide->addChild(leftStraigthen);
      leftSide->addChild(leftTrans);
      leftSide->addChild(side);

      // building right side of the fence
      SoSeparator *rightSide = new SoSeparator;
      SoRotationXYZ *rightStraigthen = new SoRotationXYZ;
      rightStraigthen->axis.setValue(SoRotationXYZ::Z);
      rightStraigthen->angle.setValue(DEGREES_TO_RADIANS(180));

      SoTransform *rightTrans = new SoTransform;
      rightTrans->translation = SbVec3f(-fenceLength, fenceLength, 0);

      rightSide->addChild(rightStraigthen);
      rightSide->addChild(rightTrans);
      rightSide->addChild(side);

      // add all side of the fences
      fence->addChild(backSide);
      fence->addChild(leftSide);
      fence->addChild(rightSide);

}


