//EnumeratedVolume.h - Specifications for class CEnumeratedVolume, part of the NRL Project

#ifndef EnumeratedVolume_h
#define EnumeratedVolume_h

#include <vector>

typedef std::vector<int> INT_VECTOR;


/* =======================================================================================
	The EnumeratedVolume is a data construction used to represent a real world piece of ocean
		as a three dimensional grid of data points.  This is a base class to provide the basic
		information needed to define where the volume resides in the real world and how that
		same volume is translated into a numeric grid, the data space for computational
		manipulation.

	This software was developed as part of a grant by ONR 311, Office of Naval Research,
	800 North Quincy Street, Arlington, Virginia 22217, for development of an "Application
	of Volume Modeling Techniques to Dynamic Containment Regions for Naval Applications".
	Grant was awarded to Arizona State University, Department of Computer Science and Engineering,
	Dr. Greg Nielson and Gary Graf.


	Author: Gary Graf, Arizona State University.
	Version:  1.0b, 10 Apr 2002

  All rights reserved.
=========================================================================================*/

//!EnumeratedVolume
/*!
EnumeratedVolume is a data construction used to represent a real world piece of ocean as a three dimensional
grid of data points.  This is a base class to provide the basic information needed to define where the volume
resides in the real world and how that same volume is translated into a numeric grid, the data space for
computational manipulation.
*/
class CEnumeratedVolume
{
/***************************  Public class varibles **************************************/
public:
		/*!This structure is a rectangle for defining a boundary or the limits of influence of
		  a volume. Coordinates are specified as degrees of latitude and longitude.
		  */
	struct FloatRect {
		float minLat;
		float minLon;
		float maxLat;
		float maxLon;
	};

		/*!The origin or creation time of the associated volume.  The value represents time in
		seconds from an unspecified starting time.
		*/
	long int timeReference;

		/*!This variable will hold the net size of the volumeData vector.  It's value will be
		//equal to the product of the three volumeDimension[] elements.
		*/
	long int vectorSize;

		/*!This is the vector that records all of the data relating to the enumerated space.
		//  The vector index relates to a location in a virtual three dimensional array of
		//  the volume's grid points.  Those vector indices relate to the order of the data
		//  by the relationship:
		//       vector index = [lonGridPt + latGridPt*(maxLonGridSize+1) +
		//                       depthGridPt*(maxLonGridSize+1)*(maxLatGridSize+1)].
		//   Note too that maxLonGridSize + 1 = volumeDimension[0] and
		//                maxLatGridSize + 1 = volumeDimension[1].
		*/
	INT_VECTOR volumeData;

		//!Flag to note if voulume should be displayed.
	bool displayVol;

		/*!Color of region when redered. The value of regnColor[0] is the red component,
		  regnColor[1] the green component, and regnColor[2] is the blue component.
		  */
	float regnColor[3];

		//!Transparancy of region when rendered
	float transparancy;

/***************************  Protected class varibles ***********************************/
protected:
		/*!A tag to accurately track which volume is which.
		*/
	int volumeID;

		/*!The dimension of the enumerated volume reflected by the number of data points.
		  This will be equivalent to maxExtent divided by the allowedRangeError after unit
		  conversion.  The [0] ordinal is the number of points in the direction of
		  increasing longitude (X). The [1] ordinal is the number of points in the direction of
		  increasing latitude (Y). The [2] ordinal is the number of points in the direction of
		  increasing depth (Z).
		*/
	int volumeDimension[3];

		/*!Maximum area the volume covers in the real world coordinate system, degrees of latitude and
		//	  longitude. The values of minX and minY represent the origin of the volume (with
		//   zero, the surface, being the third coordinate).  Unless otherwise specified, all
		//   volumes are presumed to extend to 30,000 ft (10,000 yd) in depth.
		*/
	FloatRect maxExtent;

		/*!Since they are frequently used, precompute and store the midpoint of the region's latitude
		//  and the span of the latitude and longitude
		*/
	float centerLat;
	float spanLat, spanLon;

		/*!The maximum allowed range error (in yards) used for determining the bounds of the
		//   enumerated volume.  The spacing between grid points (after conversion) is equal
		//   to this error value.
		*/
	float allowedRangeError;

		/*!Once the extent and allowedRangeError are determined, this value will provide the
		//  conversion factors to convert latitude to a grid position.  The offset component
		//  is the maxExtent.minLat.
		*/
	float latToGridScale;

		/*!Once the extent and allowedRangeError are determined, this value will provide the
		//  conversion factors to convert longitude to a grid position.  The offset component
		//  is the maxExtent.minLon.
		*/
	float lonToGridScale;


/***************************  Public  methods ********************************************/
public:
		//Constructors and destructors

	/*!
		The default constructor establishes an enumerated volume of approximately, 200 kyds x
		200 kyds, 500 yd spacing, origin at 25 deg N, 25 deg E, with no ID number.
	*/
	CEnumeratedVolume();


	/*!
		This constructor will take a min latitude and longitude along with the required grid
		dimensions to compute the final extent of the region.  It is primarily used in the
		construction of resultant containment regions following the expansion of another region.

		@param newID - An integer which is a unique identifier for the containment region.
		@param refTime - A long integer which represents the time when the given dimensions of the
				volume data for the region are valid.
		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param latDim - An integer which specifies how many grid cell the region should occupy
				in latitude.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param lonDim - An integer which specifies how many grid cell the region should occupy
				in longitude.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
		@param colors  - A pointer to a three element float array specifying the color to be used 
				in rendering this particular containment region.  The nature of the color space 
				used and the spread of values are not specified.  The information contained in 
				in the colors[] array are only stored to be later passed to the rendering engine.
		@param trans  - A float specifying the transparancy (or opacity) of the containment region
				when rendered.  The nature of the transparancy and the spread of values are not 
				specified.  The information contained is only stored to be later passed to the 
				rendering engine.
	*/
	CEnumeratedVolume(int, long int, float, float, float, float, float, float*, float);


	/*!
		This constructor will take a min latitude and longitude along with the required grid
		dimensions to compute the final extent of the region.  It is used in the
		construction of expansion regions. 

		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param latDim - An integer which specifies how many grid cell the region should occupy
				in latitude.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param lonDim - An integer which specifies how many grid cell the region should occupy
				in longitude.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
	*/
	CEnumeratedVolume(float, int, float, int, float);


	/*!
		This is the fully specified constructor.

		@param newID - An integer which is a unique identifier for the containment region.
		@param refTime - A long integer which represents the time when the given dimensions of the
				volume data for the region are valid.
		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param mxLat  - A float which is the maximum latitude boundary of the region.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param mxLon  - A float which is the maximum longitude boundary of the region.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
		@param colors  - A pointer to a three element float array specifying the color to be used 
				in rendering this particular containment region.  The nature of the color space 
				used and the spread of values are not specified.  The information contained in 
				in the colors[] array are only stored to be later passed to the rendering engine.
		@param trans  - A float specifying the transparancy (or opacity) of the containment region
				when rendered.  The nature of the transparancy and the spread of values are not 
				specified.  The information contained is only stored to be later passed to the 
				rendering engine.
	*/
	CEnumeratedVolume(int, long int, float, int,   float, int,   float, float*, float);


	//!Destructor
	~CEnumeratedVolume();

	

		//General Methods



	/*!
		Method will zero all of the data for the enumerated volume but leave the data
			structure of the underlying vector intact.
	*/
	void  zeroVolumeData();


	/*!This method streamlines the repetitive calculation of the vector offset for a set of lon,
		lat & depth grid coordinates.  Effectively translates the 3D array (grid[lon][lat][depth]) 
		into the equivalent linear vector position.

		@param ln - An integer which is the X grid index (longitude) of the volume data for the region.
		@param lt - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param dp - An integer which is the Z grid index (depth) of the volume data for the region.
		@param midOff - A long integer which is the product of the X and Y grid indices for the
								volume data for the region.  Precomputed and passed to avoid repetition.
		@returns - A long integer which is the vector offset equivalent to grid[i][j][k].
	*/	
	long int getOffset(int, int, int, long int);


	/*!
		Simple accessor will allow access to the volume ID number.

		@returns - An integer which is a unique identifier for the enumerated volume.
	*/
	int  getVolumeID();


	/*!
		Simple manipulator will allow changing the volume ID number.

		@param ID - An integer which is a unique identifier for the enumerated volume.
	*/
	void setVolumeID(int ID);


	/*!
		Simple accessor will allow access to the data in the FloatRect structure which holds data
			about the size and real world coordinates of the enumerated volume. Setting values
			is only allowed as the object is instantiated.

		@returns - A FloatRect object which holds data about the size and real world coordinates
			of the enumerated volume.
	*/
	FloatRect getExtent();


	/*!
		Simple accessor will return the latitude of the midpoint of the region. Setting values is
			only allowed when the object as instantiated.

		@returns - A float which is the latitude of the midpoint of the region.
	*/
	float getCenterLat();


	/*!
		Simple accessor will return the difference between the maximum and mininum latitudes of
			the region. Setting values is only allowed when the object as instantiated.

		@returns - A float which is the difference between the maximum and mininum latitude.
	*/
	float getSpanLat();


	/*!
		Simple accessor will return the difference between the maximum and mininum longitudes of
			the region. Setting values is only allowed when the object as instantiated.

		@returns - A float which is the difference between the maximum and mininum longitude.
	*/
	float getSpanLon();


	/*!
		Simple accessor will allow access to the size of the underlying grid structure of
			the enumerated volume. Setting values is only allowed when the object as instantiated.

		@retuns - A pointer to a three element integer array, the compones of which hold the
				size of the grid in dimensions of [longitude][latitude][depth].
	*/
	int* getGridDimensions();


	/*!
		Simple accessor will return the maximum allowable range error which this volume is using
			to detrimine grid spacing.  Setting values is only allowed when the object as instantiated.

		@returns - A float which is the maximum allowable range error of the volume.
	*/
	float getAllowedRangeError();


	/*!
		Method is to allow quick conversion from a real world coordinate to the corresponding
			grid point along the latitude axis of the enumerated volume.  To improve speed
			performance, error checking is not done.  If the passed latitude is outside the area
			defined by the enumerated volume, the returned grid coordinates could be negative or
			a value greater than the actual dimensions of the grid structure.

		@param lat - A float representing a latitude.
		@returns - An integer which is the latitude element of the grid index for the point
			which corresponds closest to the passed latitude.
	*/
	int   getGridFromLat(float lat);


	/*!
		Method is to allow quick conversion from grid point along the latitude axis of the
			enumerated volume to the corresponding real world coordinate.

		@param grid - An integer representing a grid index in the latitude dirction.
		@returns - A float which is the corresponding latitude of the passed grid index.
	*/
	float getLatFromGrid(int grid);


	/*!
		Method is to allow quick conversion from a real world coordinate to the corresponding
			grid point along the longitude axis of the enumerated volume.  To improve speed
			performance, error checking is not done.  If the passed longitude is outside the area
			defined by the enumerated volume, the returned grid coordinates could be negative or
			a value greater than the actual dimensions of the grid structure.

		@param lon - A float representing a longitude.
		@returns - An integer which is the longitude element of the grid index for the point
			which corresponds closest to the passed longitude.
	*/
	int   getGridFromLon(float lon);


	/*!
		Method is to allow quick conversion from grid point along the longitude axis of the
			enumerated volume to the corresponding real world coordinate.

		@param grid - An integer representing a grid index in the longitude dirction.
		@returns - A float which is the corresponding longitude of the passed grid index.
	*/
	float getLonFromGrid(int grid);


	/*!
		Method is to allow quick conversion from a real world coordinate to the corresponding
			grid point along the depth axis of the enumerated volume.  To improve speed
			performance, error checking is not done.  If the passed depth is outside the area
			defined by the enumerated volume, the returned grid coordinates could be negative or
			a value greater than the actual dimensions of the grid structure.
		NOTE:  Depth values are in feet (not yards).

		@param depth - A float representing a depth.
		@returns - An integer which is the depth element of the grid index for the point
			which corresponds closest to the passed depth.
	*/
	int   getGridFromDepth(float dpth);


	/*!
		Method is to allow quick conversion from grid point along the depth axis of the
			enumerated volume to the corresponding real world coordinate.
		NOTE:  Depth values are in feet (not yards).

		@param grid - An integer representing a grid index in the depth dirction.
		@returns - A float which is the corresponding depth of the passed grid index.
	*/
	float getDepthFromGrid(int grid);


	/*!
		Simple manipulator to set the time reference related to the volume.

		@param time - A long integer which is the time reference related to the volume.
	*/
	void setReferenceTime(long int);


	/*!
		Simple accessor will return the time reference related to the volume.

		@returns - A long integer which is the time reference related to the volume.
	*/
	long int getReferenceTime();


	/*!
		Simple manipulator to set the color that the volume is to be rendered with.  The
			passed parameter is assumed to be a three element array, but no specific limits
			are placed on the color space to be used or on the range of values passed.

		@param newColor - A pointer to float array which is the color that the volume is to be
				rendered with.
	*/
	void setColor(float*);


	/*!
		Simple accessor will return the color that the volume is to be rendered with.  The
			value is assumed to be a three element array, but no specific limits
			are placed on the color space to be used or on the range of values returned.

		@returns - A pointer to float array which is the color that the volume is to be
				rendered with.
	*/
	float* getColor();


	/*!
		Simple manipulator to set the transparancy (or opacity) that the volume is to be rendered
			with.  There are no specific limits placed on the range of values passed.

		@param trans - A float which indicates the degree of transparancy (or opacity) that the
				volume is to be rendered with.
	*/
	void setTransparancy(float);


	/*!
		Simple accessor will return the transparancy (or opacity) that the volume is to be rendered
			with.  There are no specific limits placed on the range of values returned.

		@returns - A float which indicates the degree of transparancy (or opacity) that the
				volume is to be rendered with.
	*/
	float getTransparancy();

/***************************  Protected methods ******************************************/
protected:

	/*!
	The method will take the dimensions of the volume, as determined by the maximum extent of
		the latitude and longitude, and convert the real world coordinate space to a enumerated
		grid with the minimum latitude and longitude values at zero in the grid.
	WARNING:  The algorithm will fail if the latitude inputs exceed 90 degres.  For
	  use in polar regions, use a transverse coordinate system or adjust algorithm accordingly.
	 The method will also compute the conversion factors to directly relate a point of latitude
	  or longitude to their respective grid coordinates.
	*/
	void computeVolumeDimensions();

};	//end class CEnumeratedVolume specifications
#endif

