//ExpansionTemplate.cpp - Body of associated class.

#include "ExpansionTemplate.h"
#include <math.h>   

/* =======================================================================================
	The ExpansionTemplate is used to develop a template which is overlaid on a containment
	region.  The sum of the values of the expansion template and the original containment
	region then define the scope of the final expanded containment.

	This software was developed as part of a grant by ONR 311, Office of Naval Research, 
	800 North Quincy Street, Arlington, Virginia 22217, for development of an "Application 
	of Volume Modeling Techniques to Dynamic Containment Regions for Naval Applications".
	Grant was awarded to Arizona State University, PRISM Laboratory, Department of Computer 
	Science and Engineering, Dr. Greg Nielson and Gary Graf.


	Author: Gary Graf, PRISM Laboratory, Arizona State University.
	Version:  1.0, 10 Sep 2002

  Copyright 2002, all rights reserved.
  
	Notes:
=========================================================================================*/

		//A set of constants to enumerate the options for how the region should be expanded.
//TODO:  Add more than spherical options.
	int const static SPHERICAL = 0;
	

		//A constant to multiply times degrees to get radians.
	float const static DEG_TO_RAD = 0.017453292f;
		//A constant to convert yards to nautical miles.  Note that one nautical mile (nm) is
		//	 also equal to one degree of latitude.
	float const static YARDS_PER_NM = 2000.0f;
		//A constant to convert minutes of latitude/longitude to degrees.
	float const static MIN_PER_DEGREE = 60.0f;

/**************************************************************************************
---------------------------------------------------------------------------------------
		Constructors / Destructors
---------------------------------------------------------------------------------------
**************************************************************************************/

													
	/*
		This constructor will take a min latitude and longitude along with the required grid
		dimensions to compute the final extent of the region.  It is used in the
		construction of simple spherical expansion templates.  

		@param mnLat  - A float which is the minimum latitude boundary of the region.
		@param latDim - An integer which specifies how many grid cell the region should occupy
				in latitude.
		@param mnLon  - A float which is the minimum longitude boundary of the region.
		@param lonDim - An integer which specifies how many grid cell the region should occupy
				in longitude.
		@param rngErr  - A float specifying the maximum allowed range error used in the rendering
				and analysis of the region.
		@param target - An pointer to a CTargetMotionData object which provides the necessary
				information on target speed and motion needed to develop an expansion volume.
		@param timeHr - A float which is the ellapsed time to calculate the expansion.
		@param type - An enumerated integer which identifies how the exansion is developed.  The
				enumerated options are noted in documentation with the class constants.
	*/
CExpansionTemplate::CExpansionTemplate(float mnLat, int latDim, float mnLon, int lonDim, 
													float rngErr, CTargetMotionData *target, 
													float timeHr, int type) : 
						  CEnumeratedVolume(mnLat, latDim, mnLon, lonDim, rngErr){

	assocTarget = target;
	elapsedTimeHr = timeHr;

	regionCenter[0] = volumeDimension[0] / 2;
	regionCenter[1] = volumeDimension[1] / 2;
	regionCenter[2] = volumeDimension[2] / 2;

		//Adjust expansion volume to have an odd number of grid points.
	adjustExpansionVolume();

		//Compute growth as a fraction of a degree.
	float maxDiffNM = target->maxSpeed * elapsedTimeHr;

//TODO:  Add more than spherical expansion options.


		//Create the internal expansion shape for the region.
	if(type == SPHERICAL){
		fillSphericalRegion(maxDiffNM);
	}	//end if

	findSurfacePoints();
	findArcsAndQuads();
}	//end constructor CExpansionTemplate(float, int, float, int, float, CTargetMotionData, 
	//												 float, int)


	//Destructor
CExpansionTemplate::~CExpansionTemplate(){
		//Empty point list
	if(!surfaceList.empty())
		surfaceList.clear();

		//Empty arc lists
	if(!arcsNE.empty())
		arcsNE.clear();
	if(!arcsNW.empty())
		arcsNW.clear();
	if(!arcsSE.empty())
		arcsSE.clear();
	if(!arcsSW.empty())
		arcsSW.clear();

	if(!arcsUE.empty())
		arcsUE.clear();
	if(!arcsUW.empty())
		arcsUW.clear();
	if(!arcsDE.empty())
		arcsDE.clear();
	if(!arcsDW.empty())
		arcsDW.clear();

	if(!arcsUN.empty())
		arcsUN.clear();
	if(!arcsUS.empty())
		arcsUS.clear();
	if(!arcsDN.empty())
		arcsDN.clear();
	if(!arcsDS.empty())
		arcsDS.clear();

		//Empty quadrant lists
	if(!quadsUNE.empty())
		quadsUNE.clear();
	if(!quadsUNW.empty())
		quadsUNW.clear();
	if(!quadsUSE.empty())
		quadsUSE.clear();
	if(!quadsUSW.empty())
		quadsUSW.clear();

	if(!quadsDNE.empty())
		quadsDNE.clear();
	if(!quadsDNW.empty())
		quadsDNW.clear();
	if(!quadsDSE.empty())
		quadsDSE.clear();
	if(!quadsDSW.empty())
		quadsDSW.clear();
}	//end destructor ~CExpansionTemplate()



/**************************************************************************************
---------------------------------------------------------------------------------------
		Private methods
---------------------------------------------------------------------------------------
**************************************************************************************/


	/*
	  Method takes a volume and ensures that it has the odd number of grid points in each
		dimension required to correctly allign the center of the expansion template with
		the containment volume which is being expanded.  Geographic coordinates and grid
		coordinates are both adjusted as necessary.
	*/
void CExpansionTemplate::adjustExpansionVolume(){
		//First you must take the existing volume and ensure that the grid
		//  alignment is odd so it can be properly centered over the original
		//  region to find the expanded region.

		//Do latitude first since you need the new center latitude to later
		//  compute the correct longitudinal dimensions.
	if((volumeDimension[1] % 2) == 0){
		volumeDimension[1]++;
		spanLat = (volumeDimension[1] - 1) * allowedRangeError / MIN_PER_DEGREE / YARDS_PER_NM;
		maxExtent.maxLat = maxExtent.minLat + spanLat;
		centerLat = spanLat / 2 + maxExtent.minLat;
		regionCenter[1] = (float)(volumeDimension[1] - 1) / 2.0;
		latToGridScale = (volumeDimension[1] - 1) / spanLat;
	}	//end if

			//Now for the longitude component
	if((volumeDimension[0] % 2) == 0){
		volumeDimension[0]++;
		spanLon = (volumeDimension[0] - 1) * allowedRangeError / cos(centerLat * DEG_TO_RAD) / MIN_PER_DEGREE / YARDS_PER_NM;
		maxExtent.maxLon = maxExtent.minLon + spanLon;
		regionCenter[0] = (float)(volumeDimension[0] - 1) / 2.0;
		lonToGridScale = (volumeDimension[0] - 1) / spanLon; 
	}	//end if
}	//end adjustExpansionVolume()


	/*
	  This method will take the expansion template and fill its data values with a spherical 
			cloud of points.

		@param maxDist - A float representing the radius of the spherical data within the template,
					expressed as a measure of nautical miles.
	*/
void CExpansionTemplate::fillSphericalRegion(float maxDist){

	float radius = maxDist * YARDS_PER_NM / allowedRangeError;

		//Since the expansion is spherical, make the orthangonal radius values all the same
	expDistNorth = (int)(radius + 0.5);
	expDistSouth = expDistNorth;
	expDistEast  = expDistNorth;
	expDistWest  = expDistNorth;
	expDistUp    = expDistNorth;
	expDistDown  = expDistNorth;

		//Gerneate a sphere within the expansion template and update the vector
	long int lonOffset, latOffset, depthOffset;
	double x, y, z, sphereValue;
	double ySqr, zSqr; 
	double radiusSqr = radius * radius + 1;

		//Clear the existing data from the vector to add new data.
	volumeData.clear();

	long int midOffset = volumeDimension[1] * volumeDimension[0];

	for(int k=0; k < volumeDimension[2]; ++k){
		z = k - regionCenter[2];
		zSqr = z * z;
		depthOffset = k * midOffset;
		for(int j=0; j < volumeDimension[1]; ++j){
			y = j - regionCenter[1];
			ySqr = y * y;
			latOffset = j * volumeDimension[0] + depthOffset;
			for(int i=0; i < volumeDimension[0]; ++i){
				x = i - regionCenter[0];
				sphereValue = x*x + ySqr + zSqr - radiusSqr;
				lonOffset = i + latOffset;
				if(sphereValue <= 0)
					volumeData.push_back(1);	
				else
					volumeData.push_back(0);	
			}	//end for i
		}	//end for j
	}	//end for k
}	//end fillSphericalRegion()


	/*
	  This method will search the volume for points which lie on the surface, including
	    all points which lie on the boundary of the volume.  A surface point is defined as
		 a location where the data is greater than zero and at least one point adjacent, not
		 counting diagonals, is equal to zero.  When it finds a surface point
	    it stores the corresponding vector and array indices in the appropriate surface list.  
		 The method automatically flags boundary points as surface points and then tests the
		 interior points.
	*/	
void CExpansionTemplate::findSurfacePoints(){
		//Clear any potentially old data
		//Empty point list
	if(!surfaceList.empty())
		surfaceList.clear();


	long int lonOffset, latOffset, depthOffset;
	long int midOffset = volumeDimension[1] * volumeDimension[0];

	for(int k=0; k < volumeDimension[2]; ++k){
		depthOffset = k * midOffset;
		for(int j=0; j < volumeDimension[1]; ++j){
			latOffset = j * volumeDimension[0] + depthOffset;
			for(int i=0; i < volumeDimension[0]; ++i){
				lonOffset = i + latOffset;
				if(volumeData[lonOffset] > 0)
						//Quickly eliminate boundary points then test interior points
					if(k == 0)
						addSurfacePointToList(lonOffset, i, j, k);
					else if(j == 0)
						addSurfacePointToList(lonOffset, i, j, k);
					else if(i == 0)
						addSurfacePointToList(lonOffset, i, j, k);
					else if(k == volumeDimension[2] - 1)
						addSurfacePointToList(lonOffset, i, j, k);
					else if(j == volumeDimension[1] - 1)
						addSurfacePointToList(lonOffset, i, j, k);
					else if(i == volumeDimension[0] - 1)
						addSurfacePointToList(lonOffset, i, j, k);
					else if(pointIsOnSurface(i, j, k, midOffset))
						addSurfacePointToList(lonOffset, i, j, k);
			}	//end for i
		}	//end for j
	}	//end for k
}	//end findSurfacePoints()



	/*
	  This method takes all the surface points and sorts them into arcs and quads.  All of the
	    surface points of the templante may be catagorized based on their relationship to the
		 ordinal planes which cut through the template.  That catagorization is later useful in
		 determining how the points of the template relate to the points in the grid of the 
		 containment region.

	  A face point is a point which lies on the surface of the expansion template and at the 
	    intersection of two ordinal planes.  Since the expansion template is defined by a bounding
		 rectangle, those face points are the surface points closest to the center of each face
		 of the bounding rectangle (but may or may not be on the surface of the rectangle).  With
		 six faces to the bounding rectange, there are six face points.  Face points are determined
		 in the method "addSurfacePointToList(long int, int, int, int)" for efficiency and not
		 handled in this method.  With only six points to be concerned with, they are not sorted
		 with respect to a sepcific orientation.
		  
	  An arc is defined as the sequece of points making up a line on the surface of the template 
	    which also lies in an ordinal plane cutting through the center of the template.  The 
		 points are further subdivided into which quadrant of the intersecting plane they lie in.  
		 The position of the point is then stored separately in a list appropriate to that arcs 
		 orientation.  With three planes cutting the volume and four quadrants per plane, there are 
		 12 possible arcs defined.  The sorting is relative to the position in the real world as
		 a two element combination of north, south, east, west, up and down.

	  A quad is defined as a set of points which line on the surface of the template.  The points
	    are subdivided into quadrants of the bounding rectangle relative to the position with
		 respect to the three possible ordinal planes cutting through the region.  With three
		 planes there are eight possible quadrants.  The sorting is relative to the position in the 
		 real world as a three element combination of north, south, east, west, up and down.

	  Note that face points are also the terminal ends of arcs and so those points show in both
	    lists.  Similarly, arcs are the terminal edges of quads and so show in both lists.  The
		 duplication is done to ensure that when mapping out a new expansion region that no gaps
		 are inadvertently generated by the inability to plot the appropriate boundary point.
	*/	
void CExpansionTemplate::findArcsAndQuads(){
		//Clear any potentially old data. (Faces are cleared and set in addSurfacePointToList() ).
		//Empty arc lists
	if(!arcsNE.empty())
		arcsNE.clear();
	if(!arcsNW.empty())
		arcsNW.clear();
	if(!arcsSE.empty())
		arcsSE.clear();
	if(!arcsSW.empty())
		arcsSW.clear();

	if(!arcsUE.empty())
		arcsUE.clear();
	if(!arcsUW.empty())
		arcsUW.clear();
	if(!arcsDE.empty())
		arcsDE.clear();
	if(!arcsDW.empty())
		arcsDW.clear();

	if(!arcsUN.empty())
		arcsUN.clear();
	if(!arcsUS.empty())
		arcsUS.clear();
	if(!arcsDN.empty())
		arcsDN.clear();
	if(!arcsDS.empty())
		arcsDS.clear();

		//Empty quadrant lists
	if(!quadsUNE.empty())
		quadsUNE.clear();
	if(!quadsUNW.empty())
		quadsUNW.clear();
	if(!quadsUSE.empty())
		quadsUSE.clear();
	if(!quadsUSW.empty())
		quadsUSW.clear();

	if(!quadsDNE.empty())
		quadsDNE.clear();
	if(!quadsDNW.empty())
		quadsDNW.clear();
	if(!quadsDSE.empty())
		quadsDSE.clear();
	if(!quadsDSW.empty())
		quadsDSW.clear();


	LIST_INDEX_VECTOR::iterator surfPtIterator;
	ListIndices nextIndex;
	long int vecPos = 0;
	int ptEW = 0;
	int ptNS = 0;
	int ptUD = 0;

		//Walk through all the surface points and sort them into arcs and quadrants.
	for (surfPtIterator = surfaceList.begin(); surfPtIterator != surfaceList.end(); ++surfPtIterator){
		nextIndex = (*surfPtIterator);
		ptEW = nextIndex.lonIndex;
		ptNS = nextIndex.latIndex;
		ptUD = nextIndex.depthIndex;

			//Test for arcs.  Arcs are determined by those points which line along one ordinal axis.
		if(ptEW == regionCenter[0]){	      //Point is on an arc centered in the EW plane. (Arcs UN, US, DN, DS).
			if(ptUD <= regionCenter[2]){		//Arcs UN, US
				if(ptNS > regionCenter[1]){   //Arc  UN
					arcsUN.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants  
					quadsUNE.push_back(nextIndex);
					quadsUNW.push_back(nextIndex);
				}	//end if
				else{                         //Arc  US
					arcsUS.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants  
					quadsUSE.push_back(nextIndex);
					quadsUSW.push_back(nextIndex);
				}	//end else
			}	//end if
			if(ptUD >= regionCenter[2]){		//Arcs DN, DS
				if(ptNS > regionCenter[1]){   //Arc  DN
					arcsDN.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants  
					quadsDNE.push_back(nextIndex);
					quadsDNW.push_back(nextIndex);
				}	//end if
				else{                         //Arc  DS
					arcsDS.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants  
					quadsDSE.push_back(nextIndex);
					quadsDSW.push_back(nextIndex);
				}	//end else
			}	//end if
		}	//end if

		else if(ptNS == regionCenter[1]){	//Point is on an arc centered in the NS plane. (Arcs UE, UW, DE, DW).
			if(ptEW <= regionCenter[0]){		//Arcs UW, DW
				if(ptUD < regionCenter[2]){   //Arc  UW
					arcsUW.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants   
					quadsUNW.push_back(nextIndex);
					quadsUSW.push_back(nextIndex);
				}	//end if
				else{                         //Arc  DW
					arcsDW.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants   
					quadsDNW.push_back(nextIndex);
					quadsDSW.push_back(nextIndex);
				}	//end else
			}	//end if
			if(ptEW >= regionCenter[0]){		//Arcs UE, DE
				if(ptUD < regionCenter[2]){   //Arc  UE
					arcsUE.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants  
					quadsUNE.push_back(nextIndex);
					quadsUSE.push_back(nextIndex);
				}	//end if
				else{                         //Arc  DE
					arcsDE.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants   
					quadsDNE.push_back(nextIndex);
					quadsDSE.push_back(nextIndex);
				}	//end else
			}	//end if
		}	//end else if

		else if(ptUD == regionCenter[2]){   //Point is on an arc centered in the UD plane. (Arcs NE, NW, SE, SW).
			if(ptNS <= regionCenter[1]){		//Arcs SE or SW
				if(ptEW > regionCenter[0]){   //Arc  SE
					arcsSE.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants  
					quadsUSE.push_back(nextIndex);
					quadsDSE.push_back(nextIndex);
				}	//end if
				else{                         //Arc  SW
					arcsSW.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants  
					quadsUSW.push_back(nextIndex);
					quadsDSW.push_back(nextIndex);
				}	//end else
			}	//end if
			if(ptNS >= regionCenter[1]){		//Arcs NE or NW
				if(ptEW > regionCenter[0]){   //Arc  NE
					arcsNE.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants  
					quadsUNE.push_back(nextIndex);
					quadsDNE.push_back(nextIndex);
				}	//end if
				else{                         //Arc  NW
					arcsNW.push_back(nextIndex);
						//Every point on an arc is also in two adjacent quadrants  
					quadsUNW.push_back(nextIndex);
					quadsDNW.push_back(nextIndex);
				}	//end else
			}	//end if
		}	//end else if


			//Point is not on an ordinal plane which means it is in a quadrant.  Now find which 
			//  quad it is in.
		else{
				//EAST side
			if(ptEW > regionCenter[0]){			//Point is in UNE, USE, DNE, or DSE quadrants
				if(ptNS > regionCenter[1]){		//Point is in UNE, or DNE quadrants
					if(ptUD > regionCenter[2])		//Point is in DNE quadrant
						quadsDNE.push_back(nextIndex);
					else									//Point is in UNE quadrant
						quadsUNE.push_back(nextIndex);
				}	//end if
				else{										//Point is in USE, or DSE quadrants
					if(ptUD > regionCenter[2])		//Point is in DSE quadrant
						quadsDSE.push_back(nextIndex);
					else									//Point is in USE quadrant
						quadsUSE.push_back(nextIndex);
				}	//end else
			}	//end if

				//WEST side
			else{											//Point is in UNW, USW, DNW, or DSW quadrants
				if(ptNS > regionCenter[1]){		//Point is in UNW, or DNW quadrants
					if(ptUD > regionCenter[2])		//Point is in DNW quadrant
						quadsDNW.push_back(nextIndex);
					else									//Point is in UNW quadrant
						quadsUNW.push_back(nextIndex);
				}	//end if
				else{										//Point is in USW or DSW quadrants
					if(ptUD > regionCenter[2])		//Point is in DSW quadrant
						quadsDSW.push_back(nextIndex);
					else									//Point is in USW quadrant
						quadsUSW.push_back(nextIndex);
				}	//end else
			}	//end else
		}	//end else
	}	//end for

		//Save memory and empty point list
	if(!surfaceList.empty())
		surfaceList.clear();
}	//end findArcsAndQuads()



	/*	
	  This method will test a point which is internal to the volume to ascertain if the point 
	    is on the surface.  A non-zero point is considered to be on the surface if any adjacent 
		 grid point (not counting diagonals) are zero.

		@param i - An integer which is the X grid index (longitude) of the volume data for the region.
		@param j - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param k - An integer which is the Z grid index (depth) of the volume data for the region.
		@param midOff - A long integer which is the product of the X and Y grid indices for the
								volume data for the region.  Precomputed and passed to avoid repetition.
		@returns - A boolean which is true if the point is on the surface, false otherwise.
	*/	
bool CExpansionTemplate::pointIsOnSurface(int i, int j, int k, long int midOff){
	if(volumeData[getOffset(i + 1, j, k, midOff)] == 0)
		return true;
	else if(volumeData[getOffset(i - 1, j, k, midOff)] == 0)
		return true;
	else if(volumeData[getOffset(i, j + 1, k, midOff)] == 0)
		return true;
	else if(volumeData[getOffset(i, j - 1, k, midOff)] == 0)
		return true;
	else if(volumeData[getOffset(i, j, k + 1, midOff)] == 0)
		return true;
	else if(volumeData[getOffset(i, j, k - 1, midOff)] == 0)
		return true;

	return false;
}	//end pointIsOnSurface()



	/*
		Takes a surface point and adds it to the point list.  Points are not catagorized with
			respect to facing.

		@param vecIndex - A long integer which is the vector offset equivalent to grid[i][j][k].
		@param lnIndex - An integer which is the X grid index (longitude) of the volume data for the region.
		@param ltIndex - An integer which is the Y grid index (latitude) of the volume data for the region.
		@param depIndex - An integer which is the Z grid index (depth) of the volume data for the region.
	*/
void CExpansionTemplate::addSurfacePointToList(long int vecIndex, int lnIndex, 
															  int ltIndex, int depIndex){
		//Format and add basic data
	ListIndices newIndex;
		newIndex.vectorIndex = vecIndex;
		newIndex.lonIndex = lnIndex;
		newIndex.latIndex = ltIndex;
		newIndex.depthIndex = depIndex;

	surfaceList.push_back(newIndex);
}	//end addSurfacePointToList()

//	end class ExpansionTemplate

