//MeshFromEnumeratedVolume.h - Specification for the associated class.


#ifndef MeshFromEnumeratedVolume_h
#define MeshFromEnumeratedVolume_h

#include "ContainmentRegion.h"

/* =======================================================================================
	The MeshFromEnumeratedVolume is used to convert a enumerated volume as encapsulated as a
	CContainmentRegion and translate the surface of that volume into a triangular mesh.  The
	list of vertices and surface triangles is then available for use by standard graphics
	programs.

	This software was developed as part of a grant by ONR 311, Office of Naval Research, 
	800 North Quincy Street, Arlington, Virginia 22217, for development of an "Application 
	of Volume Modeling Techniques to Dynamic Containment Regions for Naval Applications".
	Grant was awarded to Arizona State University, Department of Computer Science and Engineering,
	Dr. Greg Nielson and Gary Graf.


	Author: Gary Graf, Arizona State University.
	Version:  1.0b, 7 July 2002

  All rights reserved.
  
	Notes:
=========================================================================================*/

/*!The MeshFromEnumeratedVolume is used to convert a enumerated volume as encapsulated as a
	CContainmentRegion and translate the surface of that volume into a triangular mesh.  The
	list of vertices and surface triangles is then available for use by standard graphics
	programs.
*/

class CMeshFromEnumeratedVolume
{
/***************************  Public class varibles **************************************/	
public:
 
	
/***************************  Protected class varibles ***********************************/ 
protected:

		//!The original enumerated volume which requires conversion.
	CContainmentRegion *region;

		/*!This structure is to hold the various array indices, including a computed
		  vector index, for the containment region.  As each surface, or other special
		  point in the volume is located, the indices are loaded into this structure
		  so they can be store, en masse, in a vector (e.g., list of edge points).
		*/
	struct LocalMeshData {
		float (*localVerticeList)[3];
		int localNumVerts;
		int (*localTriangleList)[3];
		int localNumTriangles;
	};

		//!Temporary variable used throughout to create data to store in the vertice vector
	LocalMeshData localMesh;

		/*!Define a vector to hold the list of points which define the outside surface
		  of the containment region.
		  */
	typedef std::vector<LocalMeshData> LIST_LOCAL_MESH;

		/*!Vectors which will hold the list of all the points which lie on the outside surface
		  of the containment region and whose points on a flat face, shelf, or corner, respectively.
		*/
	LIST_LOCAL_MESH localPointMeshList;
	LIST_LOCAL_MESH localArcMeshList;
	LIST_LOCAL_MESH localQuadMeshList;

	//!The arrays and size index for the final list of vertices in the triangular mesh produced.
	float (*finalVerticeList)[3];
	int finalVertCount;

		//!The arrays and size index for the final list of triangles in the mesh produced.
	int (*finalCornerList)[3];
	int finalTriangleCount;

	int  maxDepth;
	int* regionDim;
	long int sizeOfGridArea;


CContainmentRegion::LIST_INDEX_VECTOR::iterator regionSurfIterator;
CContainmentRegion::ListIndices nextRegionSurfPt;


/***************************  Public  methods ********************************************/
public:

		//Constructors and destructors


		/*!
			The default constructor takes the containment region to convert and calls the
				appropriate methods to initiate conversion.

			@param target - A pointer to the CContainmentRegion which is to be translated
				back into a triangular mesh.
		*/
	CMeshFromEnumeratedVolume(CContainmentRegion *region);


		//!Destructor
	~CMeshFromEnumeratedVolume();



		/*!
			This method provides a vehicle to retrieve the arrays of vertices and
				triangles created from the class.  Fixed structures are not usable size the
				size of the dataset is unknown beforehand.  The user passes pointers to
				the method which then gets them to point to the correct data structures.

			@param vList - A pointer to an array of n x 3 vertices.  The value n refers to the
					total number of vertices in the containment mesh, and the three refers to
					the value of the longitude/latitude/depth.
			@param triList - A pointer to an array of m x 3 triangles.  The value m refers to the
					total number of triangles in the containment mesh, and the three refers to
					the value of the index for each corner vertice.  The index relates directly
					to the number (n) of the vertice passed in the vList[n][3] pointer.
		*/
	void getMesh(float (*)[3], int (*)[3]);


		/*!
			A simple accessor.  I returns the number (n) of vertices in the containment mesh.  It
				is needed to correctly size the pointer to the array of vertices.

			@returns - An integer reflecting the size of the vertice array.
		*/
	int getNumVertsInMesh();


		/*!
			A simple accessor.  I returns the number (m) of triangles in the containment mesh.  It
				is needed to correctly size the pointer to the array of triangles.

			@returns - An integer reflecting the size of the triangle array.
		*/
	int getNumTrianglesInMesh();



/***************************  Private methods ********************************************/

	/*
		Several temporary data structures are created to effect the data conversion.  This 
			method clears all the structures and disposes of unused memory.  It is called in
			both initialization and destruction.
	*/
	void clearLocalLists();

	void basicSurfaceMesh(float, float, float);
	void recordSurfaceMeshData(int grid[4][3], float, float, float);

	int testBentCorner();
	void recordBentSurfaceMeshData(int grid[4][3], int altGrid[4][3], float, float, float);

	void basicEdgeMesh(float, float, float);
	void recordEdgeMeshData(int grid[8][3], float, float, float);

	void sharpEdgeMesh(float, float, float);
	void recordSharpEdgeMeshData(int grid[10][3], float, float, float);

	void basicCornerMesh(float, float, float);
	void recordCornerMeshData(int grid[12][3], float, float, float);

	void sharpCornerMesh(float, float, float);
	void recordSharpCornerMeshData(int grid[14][3], float, float, float);

	void testForHolePatch(int, int, float, float, float);
	void recordHolePatchMeshData(int grid[3][3], float, float, float);

	void blendMeshes();


/***************************  Protected methods ******************************************/
protected:

};	//end class CMeshFromEnumeratedVolume specifications
#endif

