//TargetMotionData.cpp - Body of associated class.

#include "TargetMotionData.h"


/* =======================================================================================
	The TargetMotionData is used to store information about the speed and direction of the target
	associated with a specific contaiment region.  This information is needed to ascertain how
	the contaiment region should grow over time.

	This software was developed as part of a grant by ONR 311, Office of Naval Research, 
	800 North Quincy Street, Arlington, Virginia 22217, for development of an "Application 
	of Volume Modeling Techniques to Dynamic Containment Regions for Naval Applications".
	Grant was awarded to Arizona State University, PRISM Laboratory, Department of Computer 
	Science and Engineering, Dr. Greg Nielson and Gary Graf.


	Author: Gary Graf, PRISM Laboratory, Arizona State University.
	Version:  1.0, 10 Sep 2002

  Copyright 2002, all rights reserved.
  
	Notes:
=========================================================================================*/

		//A constant to multiply times degrees to get radians.
	float const static DEG_TO_RAD = 0.017453292f;

		//Constants to specify how the target motion is to be treated.
	int const static FIXED = 0;
	int const static DIRECTIONAL = 1;


/**************************************************************************************
---------------------------------------------------------------------------------------
		Constructors / Destructors
---------------------------------------------------------------------------------------
**************************************************************************************/

		/*
			The default constructor establishes a target with an arbitrary  speed of 10 kts
				and no know direction.
		*/
	CTargetMotionData::CTargetMotionData(){
		targetID = 0;
		expansionType = FIXED;
		course = 0.0;
		courseDeviation = 180;
		maxSpeed = 10;
		minSpeed = 10;
	}	//end constructor CTargetMotionData()


		/*
			Constructor allows for the development of a simple target which will use spherical expansion
				only at the specified speed.

			@param ID  - An integer uniquely identifying this target from all others.
			@param spd - A float representing the expected target speed.
		*/
	CTargetMotionData::CTargetMotionData(int ID, float spd){
		targetID = ID;
		expansionType = FIXED;
		course = 0.0;
		courseDeviation = 180;

		if(spd < 0)
			maxSpeed = 30;	//Arbitrary growth rate
		else
			maxSpeed = spd;

		minSpeed = maxSpeed;
	}	//end constructor CTargetMotionData()


		/*
			This constructor fully populates the target information to allow for the development
				of directional or nonisotropic growth of the containment volume.

			@param ID  - An integer uniquely identifying this target from all others.
			@param type  - An enumerated integer identifying how target motion is to be
						treated.  See the documetation with the class constants (above) for the
						options of the enumeration.
			@param cse - A float representing the estimated target course.
			@param dev - A float representing the estimate of the deviation of course from
						the base course.
			@param maxSp - A float representing the maximum expected target speed.
			@param minSp - A float representing the minimum expected target speed.
		*/
	CTargetMotionData::CTargetMotionData(int ID, int type, float cse, float dev, 
												    float maxSp, float minSp){
		targetID = ID;

		if((type == FIXED) || (type == DIRECTIONAL))
			expansionType = type;
		else
			expansionType = FIXED;


		if((cse < 0) || (cse > 360)){
			course = 0.0;
			expansionType = FIXED;
		}	//end if
		else
			course = cse;


		if((dev < 0) || (dev > 180))
			courseDeviation = 180;
		else
			courseDeviation = dev;


		if(maxSp < 0)
			maxSpeed = 30;	//Arbitrary growth rate
		else
			maxSpeed = maxSp;

		if((minSp < 0) || (minSp > maxSp))
			minSpeed = 0;
		else
			minSpeed = minSp;
	}	//end constructor CTargetMotionData(int, int, float, float, float, float)


	
		//Destructor
	CTargetMotionData::~CTargetMotionData(){
	}	//end destructor ~CTargetMotionData()


/**************************************************************************************
---------------------------------------------------------------------------------------
		Protected methods
---------------------------------------------------------------------------------------
**************************************************************************************/


/**************************************************************************************
---------------------------------------------------------------------------------------
		Public methods - General
---------------------------------------------------------------------------------------
**************************************************************************************/


/**************************************************************************************
---------------------------------------------------------------------------------------
		Public methods - Accessors & Manipulators
---------------------------------------------------------------------------------------
**************************************************************************************/

