//TargetMotionData.h - Specification for the associated class.


#ifndef TargetMotionData_h
#define TargetMotionData_h


/* =======================================================================================
	The TargetMotionData is used to store information about the speed and direction of the target
	associated with a specific contaiment region.  This information is needed to ascertain how
	the contaiment region should grow over time.

	This software was developed as part of a grant by ONR 311, Office of Naval Research, 
	800 North Quincy Street, Arlington, Virginia 22217, for development of an "Application 
	of Volume Modeling Techniques to Dynamic Containment Regions for Naval Applications".
	Grant was awarded to Arizona State University, Department of Computer Science and Engineering,
	Dr. Greg Nielson and Gary Graf.


	Author: Gary Graf, Arizona State University.
	Version:  1.0b, 14 Apr 2002

  All rights reserved.
  
	Notes:
=========================================================================================*/

/*!The TargetMotionData is used to store information about the speed and direction of the target
	associated with a specific contaiment region.  This information is needed to ascertain how
	the contaiment region should grow over time.
*/

class CTargetMotionData
{
/***************************  Public class varibles **************************************/	
public:
		//!A tag to accurately track which targe belongs to which containment region.
	int targetID;
		/*!Only two choices:  <b>FIXED</b> implies that the containment region will grow in a spherical
		  fashion at the target's maximum speed.  <b>DIRECTIONAL</b> presumes that some information is
		  known about the target couse and speed and the additional data is used to develop an
		  ellipsoid expansion.
		*/
	int expansionType;

		//!Estimated target course, in degrees, if known.  <b>Used only with DIRECTIONAL.</b>
	float course;

		/*!The plus/minus error, in degrees, off the given course that the target may acutally be
		  in.  Used to develop ellipsoid type expansions.  Values range from 0 to 180 as the
		  error is applied to both sides of the base course.  Used only with <b>DIRECTIONAL</B>.
		*/
	float courseDeviation;

		/*!The targets estimated maximum speed, in knots.  If used with <b>FIXED</b> is the speed of
		  expansion.  If used with <b>DIRECTIONAL</b>, it is used to develop the shape of the ellipsoid.
		  */
	float maxSpeed;

		/*!The targets estimated minimum speed, in knots. An aid to develop an ellipsoidal expansion
		  volume.  Used only with <b>DIRECTIONAL</b>.
		*/
	float minSpeed;

/***************************  Public  methods ********************************************/
public:

		//Constructors and destructors


		/*!
			The default constructor establishes a target with an arbitrary  speed of 10 kts
				and no know direction.
		*/
	CTargetMotionData();


		/*!
			Constructor allows for the development of a simple target which will use spherical expansion
				only at the specified speed.

			@param ID  - An integer uniquely identifying this target from all others.
			@param spd - A float representing the expected target speed.
		*/
	CTargetMotionData(int, float);
	

		/*!
			This constructor fully populates the target information to allow for the development
				of directional or nonisotropic growth of the containment volume.

			@param ID  - An integer uniquely identifying this target from all others.
			@param type  - An enumerated integer identifying how target motion is to be
						treated.  See the documetation with the class constants (above) for the
						options of the enumeration.
			@param cse - A float representing the estimated target course.
			@param dev - A float representing the estimate of the deviation of course from
						the base course.
			@param maxSp - A float representing the maximum expected target speed.
			@param minSp - A float representing the minimum expected target speed.
		*/
	CTargetMotionData(int, int, float, float, float, float);


		//!Destructor
	~CTargetMotionData();

};	//end class CTargetMotionData specifications
#endif

