#ifndef GEOM2D_H
#define GEOM2D_H
#include <Inventor/SbLinear.h>

#define min(x,y) (x<y?x:y)
#define max(x,y) (x>y?x:y)

/*!Point2D returns any x,y and can be used for lat/long*/
template<class CoordType> 
   struct Point2D {
      typedef CoordType Coord;
      Coord x;
      Coord y;
      //Point2D returns any x,y and can be used for lat/long
      Point2D() { 
      }
      Point2D(const Coord xcoord, const Coord ycoord):x(xcoord),y(ycoord) {
      }
      void setValue(const Coord newX, const Coord newY) {
         x=newX; y=newY; }
      void moveBy(const Coord xofs, const Coord yofs) {
         x+=xofs;
         y+=yofs;
      }
      bool operator!=(const Point2D& pt) const {
         return x!=pt.x || y!=pt.y;
      }
   };

   typedef Point2D<float> Point2dFloat;
   typedef Point2D<int> Point2dInt;
   typedef Point2D<unsigned int> Point2dUint;

   /*!Rectangle2D inherits from Point2D*/
template<class Point2DType>
   class Rectangle2D {
   public:
   /*!*/
      Point2DType leftBottom;  /* lower left corner of the box */
      Point2DType rightTop; /* upper right corner of the box */
   
      Rectangle2D() {
      }
      Rectangle2D(typename Point2DType::Coord left,typename Point2DType::Coord bottom, typename Point2DType::Coord right, typename Point2DType::Coord top) :leftBottom(left,bottom),rightTop(right,top)  {
      }
   
      // a valid rectangle is a rectangle where leftBottom <= rightTop
      bool isValidRectangle() const { 
         return leftBottom.x <= rightTop.x && leftBottom.y <= rightTop.y; }
   
      void makeInvalidRectangle() {
         leftBottom.x = 1;
         leftBottom.y = 1;
         rightTop.x = 0;
         rightTop.y = 0; 
      }
   
      void setValue(typename Point2DType::Coord left, typename Point2DType::Coord bottom, typename Point2DType::Coord right, typename Point2DType::Coord top) {
         leftBottom.x = left; leftBottom.y = bottom;
         rightTop.x = right; rightTop.y = top;
      }
      inline bool contains(typename Point2DType::Coord x, typename Point2DType::Coord y) const {
         return leftBottom.x <= x && leftBottom.y <= y && rightTop.x >= x && rightTop.y >= y;
      }
      inline bool contains(const SbVec3f& pt) const {
         return contains(pt[0],pt[1]);
      }
      bool isect(const Rectangle2D& input, Rectangle2D& result) const {
         if(this->overlaps(input)) {
            result.leftBottom.x = max(leftBottom.x,input.leftBottom.x);
            result.leftBottom.y = max(leftBottom.y,input.leftBottom.y);
            result.rightTop.x = min(rightTop.x,input.rightTop.x);
            result.rightTop.y = min(rightTop.y,input.rightTop.y);
            return true;
         }
         result.makeInvalidRectangle();
         return false;
      };
   
      bool diff(const Rectangle2D& input, Rectangle2D bands[4])  const {
         Rectangle2D common;
         if(this->isect(input,common)) {
            if(common != *this) {
             // There are at most 4 bands: top, left, right, and bottom
             // do we have a top band?
               if(rightTop.y > input.rightTop.y) {
                  bands[0].setValue(min(leftBottom.x,common.leftBottom.x),common.rightTop.y,
                     max(rightTop.x, common.rightTop.x), rightTop.y);
               } 
               else bands[0].makeInvalidRectangle();
            
             // do we have a left band?
               if(leftBottom.x < common.leftBottom.x) {
                  bands[1].setValue(leftBottom.x,max(leftBottom.y,common.leftBottom.y),
                     common.leftBottom.x,min(rightTop.y,common.rightTop.y));
               } 
               else bands[1].makeInvalidRectangle();
            
             // do we have a right band?
               if(rightTop.x > common.rightTop.x) {
                  bands[2].setValue(common.rightTop.x,max(leftBottom.y,common.leftBottom.y),
                     rightTop.x,min(rightTop.y,common.rightTop.y));
               } 
               else bands[2].makeInvalidRectangle();
            
             // do we have a bottom band?
               if(leftBottom.y < common.leftBottom.y) {
                  bands[3].setValue(min(leftBottom.x, common.leftBottom.x), leftBottom.y,
                     max(rightTop.x, common.rightTop.x), common.leftBottom.y);
               } 
               else bands[3].makeInvalidRectangle();
               return true;
            } 
            else {
               bands[0].makeInvalidRectangle();bands[1].makeInvalidRectangle();bands[2].makeInvalidRectangle();bands[3].makeInvalidRectangle();
               return false;
            }  
         } 
         else {
            bands[0]=*this; bands[1].makeInvalidRectangle();bands[2].makeInvalidRectangle();bands[3].makeInvalidRectangle();
            return false;
         }
      }
   
      bool gridDiff(const Rectangle2D& input, Rectangle2D bands[4])  const {
         Rectangle2D common;
         if(this->isect(input,common)) {
            if(common != *this) {
             // There are at most 4 bands: top, left, right, and bottom
             // do we have a top band?
               if(rightTop.y > input.rightTop.y) {
                  bands[0].setValue(min(leftBottom.x,common.leftBottom.x),common.rightTop.y+1,
                     max(rightTop.x, common.rightTop.x), rightTop.y);
               } 
               else bands[0].makeInvalidRectangle();
            
             // do we have a left band?
               if(leftBottom.x < common.leftBottom.x) {
                  bands[1].setValue(leftBottom.x,max(leftBottom.y,common.leftBottom.y),
                     common.leftBottom.x-1,min(rightTop.y,common.rightTop.y));
               } 
               else bands[1].makeInvalidRectangle();
            
             // do we have a right band?
               if(rightTop.x > common.rightTop.x) {
                  bands[2].setValue(common.rightTop.x+1,max(leftBottom.y,common.leftBottom.y),
                     rightTop.x,min(rightTop.y,common.rightTop.y));
               } 
               else bands[2].makeInvalidRectangle();
            
             // do we have a bottom band?
               if(leftBottom.y < common.leftBottom.y) {
                  bands[3].setValue(min(leftBottom.x, common.leftBottom.x), leftBottom.y,
                     max(rightTop.x, common.rightTop.x), common.leftBottom.y-1);
               } 
               else bands[3].makeInvalidRectangle();
               return true;
            } 
            else {
               bands[0].makeInvalidRectangle();bands[1].makeInvalidRectangle();bands[2].makeInvalidRectangle();bands[3].makeInvalidRectangle();
               return false;
            }  
         } 
         else {
            bands[0]=*this; bands[1].makeInvalidRectangle();bands[2].makeInvalidRectangle();bands[3].makeInvalidRectangle();
            return false;
         }
      }
   
      /*! Returns true if the two rectangles overlap, false if they don't */
      bool overlaps(const Rectangle2D& rectangle2) const {
         return !(rectangle2.leftBottom.x > rightTop.x ||
            rectangle2.rightTop.x < leftBottom.x ||
            rectangle2.leftBottom.y > rightTop.y ||
            rectangle2.rightTop.y < leftBottom.y);
      }
   
      Rectangle2D& operator=(const SbBox3f& abox) {
         float dummyz; abox.getBounds(leftBottom.x,leftBottom.y,dummyz,rightTop.x,rightTop.y,dummyz);
         return *this; }
   
      void moveBy(typename Point2DType::Coord x, typename Point2DType::Coord y) {
         leftBottom.moveBy(x,y);
         rightTop.moveBy(x,y);
      }  
   
      bool operator!=(const Rectangle2D& rect) const {
         return leftBottom!=rect.leftBottom || rightTop!=rect.rightTop;
      }
   };


   typedef Rectangle2D<Point2dFloat> Rectangle2dFloat;
   typedef Rectangle2D<Point2dInt> Rectangle2dInt;
   typedef Rectangle2D<Point2dUint> Rectangle2dUint;


/* Here we undefine the min() and max() macros to avoid colliding 
   with other functions or macros with the same name  */
#undef min
#undef max

#endif

