#ifndef IVCONTACTLABELER2D_H
#define IVCONTACTLABELER2D_H

/// 2002, Fernando Das-Neves, Virginia Tech UVAG Group ///

#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoSubNode.h>
#include <Inventor/fields/SoSFMatrix.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/Xt/SoXtRenderArea.h>

#include <map>
#include "observer.h"
#include "trackview.h"
#include "DiverseHUD.h"

/// 2002, Fernando Das-Neves, Virginia Tech UVAG Group ///


/*! IvContactLabeler2D is designed to be a child node of an instance of the IvHUD class.
    It is able to take a group of IvContactViews, and
    display their names in 2D on top. An IvContactLabeler2D will set
    itself as an observer of the IvContactView's observed object
    (which we know is an ObservableTrack, because that's 
    what a IvContactView's model must be). In this way, when the track
    is destroyed, the IvContactLabeler2D will be automatically notified 
    and will stop displaying the track name. In other words,
    once you add an IvContactView, you can leave all the handling of the
    labels to this object.
    Because an IvContactLabeler2D will be a child of an instance of IvHUD
    (which changes OpenGL's modelview transform), any transformation that
    affects the position of the IvContactViews icons will not affect 
    where the labels are drawn. The transform field in this class allows 
    to apply the same transformations that affect the icon position to 
    the position contact labels are drawn. To do this, set the transform
    field (a 4x4 matrix) to point to the SoSFMatrix that represents
    all accumulated transforms affecting the IvContactView's positions. */
   class IvContactLabeler2D: public SoCallback, public Observer {
      SO_NODE_HEADER(IvContactLabeler2D);
   protected:
   /// IvContactLabeler2D keeps one instance of this struct for every label to display
      typedef 
         struct {
            IvContactView *view;
            unsigned int labelLengthInPixels;
            float labelColor[3];
         } LabelData; 
   /// label background color
      float bgColor[3];
   /// Associates the IvContactView's model with an instance of LabelData. Used to identify which track has been destroyed
      typedef std::map<void*,void*> ContactList;
      ContactList contacts;
   /// The HUD this object is part of. Used to get modelview, proj matrix and viewport to get screen coords
      DiverseHUD* hud;
   /// setModel() associates a label with a track (theModel)
      virtual void setModel(Observable *theModel) { 
      }
   /// drawLabels() renders labels to screen, done manuals so that the labels are placed correctly.
      static void drawLabels(void*, SoAction*);
   public:
   /*! This Inventor field represents a 4x4 matrix that is applied to 
      the contact's position before projecting to screen 
      to find where to draw them */
      SoSFMatrix transform;
      static void initClass();
      IvContactLabeler2D();
      IvContactLabeler2D(DiverseHUD* aHUD);
      void useHUD(DiverseHUD* aHUD);
   /// Use this method to add an IvContactView to the list of objects to label, and optionally give a color for the label
      void displayLabelForContactView(IvContactView* acv, float labelRed =1.0f, float labelGreen = 1.0f, float labelBlue = 0.0f);
   /// This method will be called by an IvContactView's model, you don't need to call it directly.
      virtual void update(void *info);
   /// Set label background color
      void setBgColor(float r, float g, float b) { 
         bgColor[0]=r; bgColor[1]=g; bgColor[2]=b; }
   };

#endif
