#ifndef BATHYMETRY_H
   #define BATHYMETRY_H

   #include "Geom2D.h"
   #include <Inventor/nodes/SoSubNode.h>
   #include <Inventor/SbBox.h>
   #include <Inventor/SbLinear.h>
   #include <Inventor/fields/SoSFNode.h>
   #include <Inventor/nodes/SoGroup.h>
   #include <limits.h>

/*!
       If BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP is defined, then some extra code is
       implemented to be sure that the inventor objects go to the right
       MapTile objects, even if the MapTiles are being shifted.
       This is implemented by making a copy of the MapTile pointers in an array,
       and then using the copy to find where to assign the inventor object.
       If BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP is undefined, then shiftMapBy() must
       cancel the loading of all pending tiles before shifting anything.
       For small tiles sets (like 5x5) and tiles that don't take much time to load,
       defining BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP is probably overkill. */
/*
   #define BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP 1
*/

   typedef unsigned long TILE_KEY_NUMBER;
   #define MAX_TILE_KEY_NUMBER UINT_MAX
   #define MIN_TILE_KEY_NUMBER 0

   /*!TileKeyRange specifies and holds the beginning and ending keys of the tile*/
   struct TileKeyRange {
      TILE_KEY_NUMBER begin;
      TILE_KEY_NUMBER end;
   };

   class Box3dFloat: public SbBox3f {
   public:
      Box3dFloat() {
      }
      Box3dFloat(float xmin, float ymin, float zmin, float xmax, float ymax, float zmax):SbBox3f(xmin,ymin,zmin,xmax,ymax,zmax) { 
      }
      void setValue(float xmin, float ymin, float zmin, float xmax, float ymax, float zmax) { 
         setBounds(xmin,ymin,zmin,xmax,ymax,zmax); }
      void moveBy(float x, float y) {
         SbVec3f& min = getMin(), max = getMax(); 
         min[0]+=x; min[1]+=y;
         max[0]+=x; max[1]+=y;
      }
      bool overlaps(const Rectangle2dFloat& aRectangle);
      void moveBy(float x, float y, float z) { 
         SbVec3f& min = getMin(), max = getMax(); 
         min[0]+=x; min[1]+=y; min[2]+=z;
         max[0]+=x; max[1]+=y; max[2]+=z;
      }
      Box3dFloat& operator=(const SbBox3f& aBbox) { 
         setBounds(aBbox.getMin(),aBbox.getMax()); return *this; }
      SbVec3f& getLeftBottom() { return getMin(); }
      SbVec3f& getRightTop() { return getMax(); }
   };


   ///MapTile holds the content of each cell in a bathymetry grid
   struct MapTile {
      ///Inventor object containing the tile geometry
      SoNode* tile;
      /// tile's bounding box
      Box3dFloat bbox;
      typedef MapTile* pointer;
      void setBoundingBox(float xmin,float ymin, float zmin, float xmax, float ymax, float zmax) {
         bbox.setBounds(xmin,ymin,zmin,xmax,ymax,zmax); }
      MapTile():tile(NULL) {}
   };


   // forward declaration, this class is only used internally by IvTiledBathymetry
   class BathymetryTileLoaderThread;

   /*!
       This class implements a bathymetry made by rectangular cells arranged in a rectangular shape.
       Each cell may contains an inventor object, called a tile.
       The bathymetry can hold an area made of a certain number of tile rows and columns. A subset of this
       area is the "visible area", the area that is drawn. The visible area works as a window, showing
       a section of the tileset at that makes the bathymetry at a time.
       The cells around the visible area work  as a cache, called "the outskirt", that will keep tiles
       even when they're not in the visible area,  the map shifts. The purpose of the outskirt is to avoid
       having th load tiles from disk when the movement is mostly cicular.
       By changing the size of the outskirt, you can balance how many tiles to keep in memory versus how many
       tiles to load from  memory. To create an IvTiledBathymetry, call any of the constructors. If you use
       the default constructor IvTiledBathymetry() then you must call initInstance() afterwards to initialize
       the visible area and outskirt siz, among other things.  To fill the visible area with tiles, call
       fillVisibleAreaWithTiles(). To move set of tiles in  the visible area one or more row or colum
       call shiftMapBy(). If a cell in visible area is empty (has no tile) after shifting, the tile will be loaded.
   */
   class IvTiledBathymetry: public SoGroup {
      SO_NODE_HEADER(IvTiledBathymetry);
      /// Array of cached (in-memory) tiles. Includes the visible tiles + the band of tiles (outskirt) that we keep all around the visible area.
      MapTile::pointer *tiles;
   	//  Array used when shifting the map to keep tiles that would be overwritten
   	MapTile::pointer *backTileBuffer;
      /*! rectangle  [(row,col),(row,col)] inside the \c tiles array that defines the visible area of the map
         (when tiles is looked as an rectangle of size (chachedAreaSixe.x,chachedAreaSixe.y) */
      Rectangle2dUint visibleArea;
      /// number of tile rows and columns that are visible -- this is the size of the rectangle in the visibleArea member variable
      Point2dUint visibleAreaSize;
      /// (tile col,tile row)  of the left bottom tile in the visible area of the map. Useful to quickly calculate which tiles we need, when the map if shifted
      Point2dUint leftBottomVisibleTileCoord;
      /*! offsetLeftBottomVisibleArea the offset of the left bottom corner of the visible area inside the tiles array.
          It is always equal to leftBottomVisibleTileCoord.y*cachedAreaSize.x+leftBottomVisibleTileCoord.x */
      unsigned int offsetLeftBottomVisibleArea;
      /// Bounding box of the visible area, in whatever units the tiles are using (this is NOT number of tiles)
      Rectangle2dFloat visibleAreaBBox;
      /// maximum and minimum z coordinates for the bathymetry - only needed if returnFastBBox is true.
      float visibleAreaVLimits[2];
      /*! true if getBoundingBox() returns a bbox made of visibleAreaBBox and visibleAreaVLimits;
          otherwise getBoundingBox() will calculate the bounding box from the actual geometry of the tiles.
          It is false by default. */
      bool returnFastBBox;
      /// how many extra rows  to keep at the top and bottom, and columns to the left and right of of the visible area
      Point2dUint outskirtSize;
      /// number of tile rows and columns that fit in *tiles (the cached map area). These are equal or greater than nRowsVisibleArea and  nColsVisibleArea
      Point2dUint cachedAreaSize;
      /// Total number of tile rows and columns that fit in 'worldMapBBox', according to 'tileSize'
      Point2dUint worldMapSize;
      /// Size of each tile in the map. All tiles are asummed to be of exactly the same size, except at the map borders
      Point2dFloat tileSize;
      /// Bounding box of the world - no tile is outside of this box.
      Rectangle2dFloat worldMapBBox;
      /// Tells whether to draw a rectangle around the visible area or not
      bool drawEdge;
      /// color used to draw the edge around the visible are - white by default
      float edgeColor[3];
      /// sets minimum depths
      double mindepth;
      /// sets maximum depths
      double maxdepth;
      /// sets number of colors
      int numColors;
   public:
      /// Initializes this class. IvTiledBathymetry::initClass() must be called before any instance is created.
      static void initClass();

      typedef void(*TileNotifierCallback)(MapTile::pointer, void* extraArg);
   protected:
      SoSFNode alternateRep;
      TileNotifierCallback loadNotifyCallback;
      void *loadNotifierArg;
      /// Removes all Inventor objects representing tiles inside the \c tiles array.
      void clearTileCache();
      /// This is a callback used internally to get a notification every time a new tile was loaded
      static void AnotherTileIsReady(void* myself, const TileKeyRange& range, unsigned int currentTileKey, bool tileWasLoaded);
      /*! This function transfoms a [col,row] pair from world tile coordinates to a [col,row] coordinate in the part of the map that is cached.
			  Returns true if the coordinate falls inside the cached area, or false otherwise. */
      bool worldColRowToCachedColRow(const Point2dUint& worldColRow, Point2dUint& relativeColRow);
#ifdef BATHYMETRY_KEEPLOADING_WHILE_SHIFTINGMAP
      static void DiscardMapTileArray(void*, const TileKeyRange&, void* theMapTileArray);
#endif
      static void drawMissingTile(void* param, SoAction*);
      static void drawPendingTile(void* param, SoAction*);
   public:
      /// Thread used to load tiles asynchronously, encapsulated in an object
      BathymetryTileLoaderThread* workerThread;
      /// Default constructor. The program MUST also call initInstance() before using this object.
      IvTiledBathymetry();
      /*! initInstance initializes all the values needed to load and handle the tiles, and starts the loading thread. It only needs to be explicitly called when the default constructor was used.
          \param mapArea  Bounding box of the map, in the same coordinate system of the points in a tile. No tile must be outside of this box.
                          The bottom left of this box is used to calculate the tile number, so it is specially importan that it is accurate.
          \param theTileSize 	(length,height) of a tile, in the same units of the points in a tile geometry.<B>IMPORTANT:<I>All tiles but the ones at the edges must be of the same size</I></B>.
          \param visibleAreaSize	(lenght,height) of the number of tiles that will be visible in the map, expressed in number of tiles. Minimum is (1,1).
          \param leftSlack,bottomSlack	If the bottom row or left column of tiles doesn't have the width or height of a full tile, then use this parameters to tell what is their width or length. The class will handle by itself the case in which the top row or right column are smaller than a tile.
          \param plusOutskirtRows       Number of rows and columns around the visible area (in tile units) to use as "cache" for tiles. IvTiledBathymetry will keep an area around the visible area that, althought not visible, will keep tiles when the map is shifted, so for example turning around 180 degrees will not force the bathymetry to load the tiles again. IvTiledBathymetry keeps  plusOutskirtRows to the left,  plusOutskirtRows to the right, plusOutskirtCols at the top, and plusOutskirtCols tiles at the bottom.
	  \param plusOutskirtCols	Number of rows and columns around the visible area (in tile units) to use as "cache" for tiles. IvTiledBathymetry will keep an area around the visible area that, althought not visible, will keep tiles when the map is shifted, so for example turning around 180 degrees will not force the bathymetry to load the tiles again. IvTiledBathymetry keeps  plusOutskirtRows to the left,  plusOutskirtRows to the right, plusOutskirtCols at the top, and plusOutskirtCols tiles at the bottom.
          \param tileFilenamePrefix	All the tiles are asummed to be named with an (optional) prefix, followed by the tile number, and the extension ".iv". This is the optional prefix in the tile filename.

      */
      void initInstance(const Rectangle2dFloat& mapArea, const Point2dFloat& theTileSize, const Point2dUint& visibleAreaSize,
      		float leftSlack=0.0f, float bottomSlack=0.0f, unsigned int plusOutskirtRows=0, unsigned int plusOutskirtCols=0,
		const char* tileFilenamePrefix=NULL);

      /*!IvTiledBathymetry initializes values needed to load and handle the tiles
         \param mapArea  Bounding box of the map, in the same coordinate system of the points in a tile. No tile must be outside of this box.
                          The bottom left of this box is used to calculate the tile number, so it is specially importan that it is accurate.
	 \param row
	 \param col
	 \param tileFilenamePrefix is the initialization of the name, tile number, location, and contains extension *iv.
      */
      IvTiledBathymetry(const Rectangle2dFloat& mapArea,unsigned int row=3, unsigned int col=3, const char* tileFilenamePrefix=NULL);

      IvTiledBathymetry(const Rectangle2dFloat& totalArea, const Point2dFloat& theTileSize,
      		unsigned int visibleRows, unsigned int visibleCols,
      		unsigned int plusOutskirtRows=1, unsigned int plusOutskirtCols=1,
      		float leftSlack=0.0f, float bottomSlack=0.0f, const char* tileFilenamePrefix=NULL);
      /*! Sets a callback to be called every time a new tile is loaded to fill a void in the visible area:
          \param	cb A pointer to a function of type IvTiledBathymetry::TileNotifierCallback.
          \param arg	This pointer is passed as the last parameter to the callback in the cb argument
      */
      void setTileLoadNotifyCallback(TileNotifierCallback cb, void* arg) {
         loadNotifyCallback = cb; loadNotifierArg = arg; }
      /// Sets or changes the expected size (width, length) of a tile. Note the class "believes" in this size, will use it to calculate many things, and will not check it against the size of the inventor objects in each tile.
      void setTileSize(const Point2dFloat& theTileSize);
      /// Returns the size (width, length) of a tile.
      const Point2dFloat& getTileSize() const {
         return tileSize; }
      /// Returns the total area covered by the whole map, not just the visible area.
      const Rectangle2dFloat& getMapArea() const {
         return worldMapBBox; }
      /// Returns the area covered by the visible area.
      const Rectangle2dFloat& getVisibleArea() const {
         return visibleAreaBBox; }
      /*! Resizes the map to hold at most (howManyRows+2*plusOutskirtRows)*(howManyCols+2*plusOutskirtCols) tiles.
        the visibleArea will be set to a size of [howManyVisibleRows,howManyVisibleCols],
        and the map will also cache at most 'plusOutskirtRows' at the top and bottom of the visible area,
        and 'plusOutskirtCols' to the left and right of the visible area.
      */
      void resizeTo(unsigned int howManyVisibleRows,unsigned int howManyVisibleCols, unsigned int howManyOutskirtRows, unsigned int howManyOutskirtCols);
      /*!  \overload resizeTo */
      void resizeTo(unsigned int howManyVisibleRows,unsigned int howManyVisibleCols);
      /* Returns a pointer to the MapTile at the beginning of a row in the cached area. */
      MapTile::pointer* operator[](unsigned int row) {
         return tiles+cachedAreaSize.x*row; }
      /// Shift the the visible area by a number of rows and columns. In order to fill the map first, call filltVisibleAreaWithTiles().
      void shiftMapBy(int no_rows, int no_cols);
      /*! Fills the map with tiles, and sets the size of the visible area.
          \param area	the (x,y) corners the corners limiting the area to show.
          \param outskirtRows,outskirtCols	the number of rows and columns to keep around the visible area, to accelerate loading when the map is shifted.
      */
      void fillVisibleAreaWithTiles(const Rectangle2dFloat& area, unsigned int outskirtRows=1, unsigned int outskirtCols=1);
      /* Returns a point with the (column number, row number) of the tile at the given coordinate.
         \param pt	the (x,y) coordinate of a point to translate to a tile (row,column) in the map. */
      Point2dUint worldPointToWorldColRow(const Point2dFloat& pt) const;
      /// Returns the tile number corresponding to the tile in a certain (column,row) in the map.
      TILE_KEY_NUMBER worldColRowToWorldMapTile(const Point2dUint& pos) const;
      /// Returns the (column, row) in the map corresponding to a certain tile number.
      TILE_KEY_NUMBER worldColRowToWorldMapTile(unsigned int col, unsigned int row) const;
      Point2dUint worldMapTileToWorldColRow(TILE_KEY_NUMBER tile_number) const;
      /*! This function will tell the bathymetry to use an estimation of the bounding box, 
          based on the tile size and requested area, instead of actually querying the tiles for their bounding boxes and merging them.
         \param useFast	true if object can use the estimation, false if must use the tile geomrtries.
         \param zmin,zmax if useFast is true, then zmin and zmax must have 
                the minimum and maximum expected z values, since they it is not possible to estimate them from the area requested or shown, without querying the tile geometries.
      */
      void useFastBoundingBox(bool useFast, float zmin=0, float zmax=0);
      /// Tells the bathymetry to draw a rectangle around the area occupied by the tiles. This is useful if the tiles take some time to load.
      void drawEdgeAroundVisibleArea(bool yesOrNo) { drawEdge = yesOrNo; }
      /// Tells which color to use to draw an edge around the area occupied by the tiles.
      void setEdgeColor(float r, float g, float b) { edgeColor[0]=r; edgeColor[1]=g; edgeColor[2]=b; }
      ~IvTiledBathymetry();
      void setColorscaleDepth(double mindepth, double maxdepth, int numColors);
      void setTileColorscaleDepth(SoNode *tile);
SoEXTENDER public:
	//virtual void GLRender(SoGLRenderAction *action);
      virtual void getBoundingBox(SoGetBoundingBoxAction *action);
      virtual void write(SoWriteAction *action);
   };

#endif

