#ifndef _TRACK_H
#define _TRACK_H

/// 2002, Virginia Tech UVAG Group

#include <fstream>
#include <string>
#include "observer.h"
#include <limits.h>
#include <sys/time.h>
#include "VoidPtrList.h"
#include "timeval_ops.h"
#include "coord.h"
#include "ThreatType.h"



/* This is a bit mask used to mark a certain track observer
   as being an instance of IvContactView. We need this later
   to find the the object used in the scene graph to
   represent a track's current position,
   from all the observers attached to the track. */
#define OBSERVER_IS_CONTACTVIEW 0x1
#define OBSERVER_IS_TRACKVIEW 0x2

   using namespace std;

   class Track; // just a forward declaration
   class Timeline;
/*! TrackEvent is a base (and concrete) class for ObservableTracks
  to communicate with their observers via events when something
  interesting happened */
   class TrackEvent {
   public:
      typedef enum {
      /// This event is sent when the threat class of the contact (hostile, friendly, neutral) has changed
      CONTACT_THREATCLASS_CHANGED = 1,
      /// a new step is appended to the end of the track
      TRACK_STEP_APPENDED = 2,
      /// the last step in the track is removed
      TRACK_STEP_REMOVED = 4,
      /// the Track current position is changed
      POS_CHANGED = 8,
      /// The Track object is about to be destroyed
      TRACK_ABOUT_TO_DIE = 16,
      /// This event means that somebody requested the current position of the track to be resent to the views. This NOT a notification of new position.
      RESENDING_TRACK_POS = 32,
      /// This is the first number that can be used as an application event, that an app may want the track views to notice
      TRACK_USR_EVENT=64,
      /// general notifications, telling 'something happened' to the track
      TRACK_NOTIFY = UINT_MAX
      } EventType;
      unsigned int eventType;
   /*! This is a pointer to the track where the event originated */
      Track *track;
      TrackEvent() {
      }
      TrackEvent(EventType type):eventType(type) {
      }
   };


   /*! These are events that only the TrackEvent class sends to its observers.
       These events notify of the special case that a DiscreteTrack can move
       just past its starting and ending position, and move back to a valid value.
       See DiscreteTrack for a rationale. */
   struct DiscreteTrackEvent: public TrackEvent {
      typedef enum {
       /*! This event means the track was requested to move beyond the beginning of the track.
          The usual response from a view is either to become invalid, or to hide. itself.
          Note when this event is received, the track doesn't have a current position.  */
      PAST_BEGINNING_OF_TRACK = TRACK_USR_EVENT,
       /*! This event means the track was requested to move beyond the end of the track.
          The usual response from a view is either to become invalid, or to hide. itself.
          Note when this event is received, the track doesn't have a current position.
          This event is received only if the track was 'closed', that is, the current
          last position is the definitive last position to be ever received for that track.  */
      PAST_END_OF_TRACK = (TRACK_USR_EVENT<<1)
      } EventType;
      DiscreteTrackEvent():TrackEvent() {
      }
      DiscreteTrackEvent(unsigned int eventType):TrackEvent((TrackEvent::EventType)eventType) {
      }
   };

/*! SingleObservable is an Observable that can hold and send event notifications to at most 1 Observer */
   class SingleObservable: public Observable {
   protected:
      struct ObserverInfo {
         unsigned int userDefinedMask;
         unsigned int  interests;
         Observer* observer;
      };
      ObserverInfo * observerInfo;
   public:
      /*! addView(Observer *anObserver) adds a view with bit mask set to 0 (zero), that will receive all events sent by the observable.
          A mask is a set of bits whose meaning is defined by the user.
          Its purpose is to let the user tag the observers and find them later
          by their tag. */
      virtual void addView(Observer *anObserver)  {
         addView(anObserver,0,UINT_MAX); }
      /*! addView(Observer *anObserver, unsigned int bitMaskValue, unsigned int eventsOfInterest) adds a view with bit mask set to
         'bitMaskValue'. A mask is a set of bits whose meaning is defined by the user. Its purpose is
          to let the user tag the observers and find them later by their tag. */
      virtual void addView(Observer *anObserver, unsigned int bitMaskValue, unsigned int eventsOfInterest);
      /*! notify() indicates the observable has been modified and calls the update function of the observers*/
      virtual void notify(void *info);
      /*! removeView(Observer *anObserver) stops updating a viewer by passing a observer. */
      virtual void removeView(Observer *anObserver);
      virtual ~SingleObservable();
    /*! Returns the observer if (observer_mask AND bitMask) != 0,
          or NULL if none is found, or there is no observer attached.
         If the return value != NULL, and  copyTheOSelectedObjectBitMask != NULL,
         then *copyTheOSelectedObjectBitMask  containts
         the bit mask of the returned observer. */
      Observer* findObserverWithMask(unsigned int bitMask, unsigned int* copyTheSelectedObjectBitMask) const;
   };

/*! A MultipleObservable is an Observable that can hold and notify many Observers */
   class MultipleObservable: public Observable {
   protected:
   /*!ObserverInfo lets the user tag the observers and find them later by their tag.
   	\param userDefinedMask is the mask is a set of bits whose meaning is defined by the user.
   \param interests
   */
      struct ObserverInfo {
         // the mask is a set of bits whose meaning is defined by the user. Its purpose is to let the user tag the observers and find them later by their tag.
         unsigned int userDefinedMask;
         unsigned int interests;
         Observer* observer;
      };
      typedef VoidPtrList ObserverList;
      ObserverList observers;
      static bool notifyEach(void* observer, void* info);
      static bool deleteObserverInfo(void* observerInfo);
   public:
      /*! Adds a view with bit mask set to 0 (zero). A mask is a set of bits
          whose meaning is defined by the user. Its purpose is
          to let the user tag the observers and find them later
          by their tag. */
      virtual void addView(Observer *anObserver)   {
         addView(anObserver,0,UINT_MAX); }
      /*! Adds a view with bit mask set to 'bitMaskValue'. A mask is a set of bits
          whose meaning is defined by the user. Its purpose is
          to let the user tag the observers and find them later
          by their tag.
          By defaults the observer will receive all events sent by the observable,
          unless the parameter 'eventsOfInterest' is set. */
      virtual void addView(Observer *anObserver, unsigned int bitMaskValue, unsigned int eventsOfInterest = UINT_MAX);
      /*! notify() indicates the observable has been modified and calls the update function of the observers*/
      virtual void notify(void *info);
      /*! removeView(Observer *anObserver) stops updating a viewer by passing a observer. */
      virtual void removeView(Observer *anObserver);
      unsigned int numberOfObservers() const {
         return observers.size(); }
      /*! Returns the first observer such that (observer_mask AND bitMask) != 0,
          or NULL if none is found, or there are no observers attached.
         if the return value != NULL, and  copyTheOSelectedObjectBitMask != NULL,
         then *copyTheOSelectedObjectBitMask  containts
         the bit mask of the returned observer. */
      Observer* findObserverWithMask(unsigned int bitMask, unsigned int* copyTheOSelectedObjectBitMask) const;
      virtual ~MultipleObservable();
   };


/* Some constants to indicate particular time events */
/*! NO_EVENT is a timeval constant that means "there is no time assigned to this event" */
   const timeval NO_EVENT = BAD_TIME; // This constant means there are no more events in the track


/*! This type hold the valid hints that needs to be passed to a Track object, when asked about the next event */
   typedef enum { BACKWARD_IN_TIME=-1, FORWARD_IN_TIME=1 } TimeDirection;

/*! Abstract base class for Tracks. A track is an discrete sequence of coordinates ordered
    by time. A track has to be able at least to:
    <UL>
     <LI> Tell its starting time
     <LI> Tell its final time, if any (otherwise should return TIME_MAX)
     <LI> Be notified of the current time by Track::advanceToNextEvent()
     <LI> Tell the time of the next event by Track::getTimeOfNextEvent() , given the current time.
     <LI> Tell the current time and position with Track::getCurrentCoordAndTime()
    </UL>
   In order to get the time of next event or time or to advance time, the Track asks
   needs a hint of type TimeDirection, that tells wether you are asking for the
   event that is before or after the current event in time. In this way you
   can query the Track in both time directions. */
   class Track {
   public:
   /*! getTimeOfNextEvent() returns the time of the event that is before/after the last event played.
    This means that getTimeOfNextEvent() WILL NOT return the time of the event that
    is after/before any arbitrary time. It will return NO_EVENT if there is
    no next event in the direction indicated by \a beforeOrAfter */
      virtual const timeval& getTimeOfNextEvent(const TimeDirection beforeOrAfter) =0;
   /*! advanceToNextEvent() will change the current track event one step forward or backwards
      (depending on the \a beforeOrAfter parameter). It must return NO_EVENT if
      there are no more events in the direction indicated by \a beforeOrAfter */
      virtual const timeval& advanceToNextEvent(const TimeDirection beforeOrAfter) =0;
   
   /*! startTime() is the starting time for each step.*/
      virtual const timeval& startTime() const =0;
   /*! finalTime() is the termination time for each step*/
      virtual const timeval& finalTime() const =0;
   /*! getCurrentCoordAndTime*/
      virtual CoordAndTime* getCurrentCoordAndTime() const =0;
   };




/*! The abstract class ObservableTrack is a track that will call view->update() every time
   the current step or the number of steps changes,
   to tell all its observers about the change.
   An ObservableTrack also supports a name to identify them easily,
   and can describe what kind of contact (Vehicle, weapon, etc) it represents. */
   class ObservableTrack: public Track, public MultipleObservable {
   protected:
      Category vehicleType;
      ThreatClass threatLevel;
   public:
      ObservableTrack():vehicleType(CATEGORY_UNKNOWN),threatLevel(UNKNOWN) { }
      virtual const timeval& getTimeOfNextEvent(const TimeDirection beforeOrAfter) =0;
      virtual const timeval& advanceToNextEvent(const TimeDirection beforeOrAfter) =0;
      virtual const timeval& startTime() const =0;
      virtual const timeval& finalTime() const =0;
      virtual CoordAndTime* getCurrentCoordAndTime() const =0;
      virtual void setContactType(const ContactType newType) =0;
      virtual ContactType getContactType() const =0;
   /*! Adds a step at the end of the Track.
   <B>IMPORTANT:</B> The \a ct parameter is assumed to point to an object that can be
   owned by the track. This means the track object assumes the object won't be destroyed by
   any other object other than the track object itself. This also means the a CoordAndTime
   object cannot be shared among tracks; each one must have its own copy passed as parameter. */
      virtual void appendStep(CoordAndTime* ct) =0;
      virtual const string& getName() const =0;
      virtual void setName(const string& newName) =0;
      virtual const string& getSolution() const =0;
      virtual void setSolution(const string& newSolution) =0;
      virtual void setContactCategory(const Category newVehicleType) { vehicleType = newVehicleType; }
      virtual Category getContactCategory() const { return vehicleType; }
      // Sets the threat class for the contact, and send the CONTACT_THREATCLASS_CHANGED event to all its observers if the new threat class is different from the current one.
      virtual void setThreatClass(const ThreatClass newThreatLevel);
      virtual ThreatClass getThreatClass() const { return threatLevel; }
   /*! notify() assumes info is an instance of TrackEvent, fills the 'track' field of the
      TrackEvent, and calls notify() of the Observable superclass. */
      virtual void notify(void *info);
   /*! call notify() with a TrackEvent containint 'event' in the TrackEvent::eventType field. */
      virtual void changed(TrackEvent::EventType event) { 
         TrackEvent evt(event); notify(&evt); }
      //grow is meaningless right now to Tracks that are not discrete tracks
      virtual CoordAndTime *grow(Timeline *timeline, timeval curtime) {
         return NULL;};
      /*! Destructor */
      virtual ~ObservableTrack();
   };

/*! DiscreteTrack is an implementation of the interface described by the ObservableTrack class.
   It also allows the track's current position to move right beyond the last step, or
   right before the first one. The utility of this case is that then handling
   forward/backward in time is symmetric: a track moves from "nowhere" to the
   first position when FORWARD_IN_TIME, and from "nowhere" to the last position
   when BACKWARD_IN_TIME.
   When a track moves "nowhere", it will notify its observers that it moved
   DiscreteTrackEvent::PAST_BEGINNING_OF_TRACK or DiscreteTrackEvent::PAST_END_OF_TRACK, and when it moves from "nowhere"
   to a valid position again, it will send DiscreteTrackEvent::INSIDE_TRACK_AGAIN.  */
   class DiscreteTrack: public ObservableTrack {
   protected:      
      typedef VoidPtrList CoordAndTimeList;
   /*! holds the steps that make this track, it will hold pointers to CoordAndTime  */
      CoordAndTimeList steps;
   /*! Current and last position in this track, stored as iterators of the \a steps list */
      CoordAndTimeList::iterator currentStep, last;
   /*! last position of a track, when a track is 'closed' (no more event appending is accepted).
       Otherwise endOfTrack == steps.end(). */
      CoordAndTimeList::iterator endOfTrack;
   /*! name of the track, usually the name of the contact */
      string name;
   /*! name of the solution */
      string solution;
   /*! What kind of contact (ownship, vehicle, weapon, etc) is this track's */
      ContactType type;
   /*! a flag used to indicate the current track position is "out of boundaries", that is, 
       before the first event or after the last event (the latter only if the track is closed).
       By marking a track position is "out of boundaries", we can mark the we went back before the first event,
       or after the last event, without having to store an invalid position in currentStep. */
      bool pastEndOfTrack;
      timeval FAKE_PAST_BEGINNING_TIME, FAKE_PAST_END_TIME;
   public:
      DiscreteTrack();
      virtual const timeval& getTimeOfNextEvent(const TimeDirection beforeOrAfter);
      virtual const timeval& advanceToNextEvent(const TimeDirection beforeOrAfter);
      virtual const timeval& startTime() const;
      virtual const timeval& finalTime() const;
      virtual CoordAndTime* getCurrentCoordAndTime() const;
   /*! Adds a step at the end of the Track.
   <B>IMPORTANT:</B> The \a ct parameter is assumed to point to an object that can be
   owned by the track.
   See Track::appendStep() for more details.
   */
      virtual void appendStep(CoordAndTime* ct);
      ///getName() returns the string ID of the track.
      virtual const string& getName() const {
         return name; }
      ///setName() assigns a new name to a track
      virtual void setName(const string& newName) {
         name = newName; }
      ///getSolution() returns the string solution ID of the track.
      virtual const string& getSolution() const {
         return solution; }
      ///setSolution() assigns a new solution name to a track
      virtual void setSolution(const string& newSolution) {
         solution = newSolution; }
      ///setContactType() assigns which type of contact to assign to a track
      virtual void setContactType(const ContactType newType) {
         type = newType; }
      ///getContactType() returns track contact type
      virtual ContactType getContactType() const {
         return type; }
      /// closeTrack() marks the track as closed (with a known, definite ending) in the current track position. A closed track does not allow to append positions. */
      void closeTrack();
      /// reopenTrack() reopens a track that was closed before by calling DiscreteTrack::closeTrack().
      void reopenTrack();
      ///grow() issues a new CoordAndTime for itself with updated time
      virtual CoordAndTime *grow(Timeline *timeline, timeval curtime);
   };


/*! A TrackTraceDecorator is a decorator for a DiscreteTrack; it takes care of
   the "trace", marking the previous positions of the DiscreteTrack.
    A TrackTraceDecorator holds a track, and behaves as one. In fact, after doing what it needs,
    it will pass all methods to the contained track.
    A TrackTraceDecorator's job is to talk to its observers, by notifying them of
    steps added AFTER the contained track  moved to its new position (so they know
    the current track position), and to notify them BEFORE an event is removed
    from the contained track, so they can react over the position the contained
    track is going to remove.
    For more information on the Decorator Design Pattern, read "Design Patterns" by Gamma et al, or
    go to http://www.castle-cadenza.demon.co.uk/decorate.htm . */
   class TrackTraceDecorator: public ObservableTrack {
   /// This is the 'decorated' track, the one the TrackTraceDecorator 'decorates', and relays any method call to
      ObservableTrack *track;
   public:
      TrackTraceDecorator(ObservableTrack* aTrack):track(aTrack) {
      }
      ObservableTrack *getTrack() const {
         return track;}
      virtual const timeval& getTimeOfNextEvent(const TimeDirection beforeOrAfter) {
         return track->getTimeOfNextEvent(beforeOrAfter); }
      virtual const timeval& advanceToNextEvent(const TimeDirection beforeOrAfter);
      virtual const timeval& startTime() const {
         return track->startTime(); }
      virtual const timeval& finalTime() const {
         return track->finalTime(); }
      virtual CoordAndTime* getCurrentCoordAndTime() const {
         return track->getCurrentCoordAndTime(); }
      virtual void setThreatClass(const ThreatClass newThreatLevel) {track->setThreatClass(newThreatLevel);};
      virtual ThreatClass getThreatClass() const { return track->getThreatClass(); };
   /*! Adds a step at the end of the Track.
   <B>IMPORTANT:</B> The \a ct parameter is assumed to point to an object that can be
   owned by the track object passed in the constructor of this class.
   See Track::appendStep() for more details.
   */
      virtual void appendStep(CoordAndTime* ct) {
         track->appendStep(ct); }
      ///getName() returns the string ID of the track.
      virtual const string& getName() const {
         return track->getName(); }
      ///setName() assigns a new name to a track
      virtual void setName(const string& newName) {
         track->setName(newName); }
      ///getSolution() returns the string solution ID of the track.
      virtual const string& getSolution() const {
         return track->getSolution(); }
      ///setSolution() assigns a new solution name to a track
      virtual void setSolution(const string& newSolution) {
         track->setSolution(newSolution); }
      ///setContactType() assigns which type of contact to assign to a track
      virtual void setContactType(const ContactType newType) {
         track->setContactType(newType); }
      ///getContactType() returns track contact type
      virtual ContactType getContactType() const {
         return track->getContactType(); }
      /*!grow() gets passed along to track*/
      virtual CoordAndTime *grow(Timeline *timeline, timeval curtime) {
         return track->grow(timeline,curtime); }
   };


/*! A Timeline is a track coordinator. Given multiple tracks, and an external clock,
   a Timeline will manage to notify every track when is time to do something, by
   calling Track::advanceToNextEvent(timeDirection). A Track finds out when to tell
   a track to do something because it asks the track what is next to do, by calling
   Track::getTimeOfNextEvent().
*/
   class Timeline {
   /// time of last event played
      timeval lastTime,
      /// earliest time any of the tracks needs to do something
      earliest;
      typedef VoidPtrList TrackList;
   /// Keeps all the tracks, we'll put Track* inside
      TrackList tracks;
   /// Remembers which time direction (fore or back) the timelines is moving
      TimeDirection foreOrBackInTime;
   /// stopped is true if we should ignore calls to playNextEvent()
      bool stopped;
   public:
      Timeline();
   /// Assign a track to this timeline
      void addTrack(Track* aTrack);
   /// remove a track from the Timeline and return it
      Track* removeTrack(unsigned int trackNo);
      Track* removeTrack(Track* aTrack);
   /// number of tracks in Timeline
      unsigned int size() const { 
         return tracks.size(); }
   /// Plays everything up to timeOfNextEvent
      bool playNextEvent();
   /// Restart the timeline from the beginning
      void rewind();// { 
      //         lastTime = earliest; }
   /// returns the earliest time any of the tracks need to do something
      const timeval& startTime() const { 
         return earliest; }
      void setTimeDirection(const TimeDirection td);
      const timeval& getCurrentTimeOnTimeline() const { 
         return lastTime; }
      TimeDirection timeDirection() const { 
         return foreOrBackInTime; }
   /*! Adds a step at the end of the Track.
   <B>IMPORTANT:</B> The \a ct parameter is assumed to point to an object that can be
   owned by the track object passed as parameter.
   See Track::appendStep() for more details.
   */
      void appendStepToTrack(ObservableTrack* aTrack, CoordAndTime* pos);
   /// stops showing the update of the timeline. playNextEvent() will be ignored until resume() isc called  */
      void stop();
   /// resumes showing the update of the timeline, and goes up to the end of the timeline
      void resume();
   /// returns true if the timeline is stopped (see Timeline::stop() and Timeline::resume() )
      bool isStopped() const {
         return stopped; }
   };



#endif
