#include "DisplayConfigTree.h"
#include <dxml/XmlLexer.h>
#include <dxml/XmlNode.h>
#include <displaymgr/DisplayConfig.h>
#include <tracer/Tracer.h>
#include <qfiledialog.h>
#include <qmessagebox.h>
#include <stdarg.h>
#include <stdio.h>

char* format( char *s, const char *fmt, ... )
{
    va_list args;
    va_start(args, fmt);
    vsprintf( s, fmt, args );
    va_end(args);
    return s;
}

DataListViewItem::DataListViewItem( QListView* parent, Data* data )
	:QListViewItem( parent, "All Displays", "Root" )
{
	m_data = data;
	if ( m_data )
		m_data->setItem( this );
}

DataListViewItem::DataListViewItem( DataListViewItem* parent, Data* data )
	:QListViewItem( parent, data->type().c_str(), data->getName().c_str() )
{
	m_data = data;
	if ( m_data )
		m_data->setItem( this );
}

DataListViewItem::~DataListViewItem()
{
	if ( m_data )
		delete m_data;
}

Data::Data( Data* parent, const string& name )
{
	m_parent = parent;
	m_name = name;
}

DisplayData::DisplayData( Data* parent, const string& name )
    :Data( parent, name )
{
	m_filename = "";
	m_threaded = false;
	m_windownamenum = 1;
}

DisplayData::~DisplayData()
{
	vector<WindowData*>::iterator itr;
	for ( itr = m_windows.begin(); itr != m_windows.end(); itr++ )
	{
		(*itr)->getItem()->setData( NULL );
		delete (*itr);
	}
}

WindowData::WindowData( Data* parent, const string& name )
    :Data( parent, name )
{
	m_viewportnamenum = 1;
    m_left = 0;
    m_bottom = 0;
    m_width = 640;
    m_height = 480;
    m_xscreen = 0;
    m_quadbuffer = false;
    m_border = true;
	m_fullscreen = false;
	m_sharedcontextwindow = "";
}

WindowData::~WindowData()
{
	vector<ViewportData*>::iterator itr;
	for ( itr = m_viewports.begin(); itr != m_viewports.end(); itr++ )
	{
		(*itr)->getItem()->setData( NULL );
		delete (*itr);
	}
}

ViewportData::ViewportData( Data* parent, const string& name )
    :Data( parent, name )
{
    m_left = 0;
    m_bottom = 0;
    m_width = 640;
    m_height = 480;
	m_camera = new CameraData;
}

ViewportData::~ViewportData()
{
	if ( m_camera )
		delete m_camera;
}

CameraData::CameraData()
{
    m_trackername = "headtracker";
    m_eye = MONO;
	m_projectiontype = SCREEN;
    m_perspective = new PerspectiveData;
	m_screen = new ScreenData;
}

CameraData::~CameraData()
{
	if ( m_perspective )
		delete m_perspective;
	if ( m_screen )
		delete m_screen;
}

PerspectiveData::PerspectiveData()
{
    m_fovx = 35.0;
    m_fovy = 35.0;
}

ScreenData::ScreenData()
{
	m_name = "Screen";
    m_centerx = 0.0;
    m_centery = 1.0;
    m_centerz = 0.0;
    m_heading = 0.0;
    m_pitch = 0.0;
    m_roll = 0.0;
    m_width = 1.0;
    m_height = 1.0;
    m_depth = 1.0;
}

DisplayConfigTree::DisplayConfigTree( QWidget* parent )
    :QListView( parent )
{
	Tracer::printTrace( true );
	Tracer::printDebug( 1 );
	Tracer::printError( 1 );
	Tracer::printWarn( 1 );
	Tracer::printStatus( 1 );
    addColumn( "Display Components" );
	setColumnWidth( 0, 150 );
    addColumn( "Name" );
	setColumnWidth( 1, 150 );
    setRootIsDecorated( true );
    m_treeroot = new DataListViewItem( this, new Data( NULL ) );
	setSelected( m_treeroot, true );
    connect( this, SIGNAL(selectionChanged( QListViewItem* )),
			parent, SLOT( updateSelection( QListViewItem* )) );
	connect( this, SIGNAL(rightButtonPressed ( QListViewItem *, const QPoint &, int )),
			parent, SLOT(itemContextMenu( QListViewItem *, const QPoint &, int )) );

}

void DisplayConfigTree::newDisplay()
{
    DisplayData* display = new DisplayData( NULL );
	DataListViewItem* displayitem = new DataListViewItem( m_treeroot, display );
	ensureItemVisible( displayitem );
	m_displays.push_back( display );
	setCurrentItem( displayitem );
    WindowData* window = new WindowData( display );
	DataListViewItem* windowitem = new DataListViewItem( displayitem, window );
	ensureItemVisible( windowitem );
	display->addWindow( window );
    ViewportData* viewport = new ViewportData( window );
	DataListViewItem* viewportitem = new DataListViewItem( windowitem, viewport );
	ensureItemVisible( viewportitem );
	window->addViewport( viewport );
	setCurrentItem( displayitem );
	display->setChanged( true );
}

void DisplayConfigTree::openDisplay()
{
	QString filename( QFileDialog::getOpenFileName( QString::null, "Displays (*.xml)", this, "opendialog", "Open display configuration file" ) );

	DisplayConfig config;
	if ( config.open( filename.latin1() ) )
		return;

	if ( config.displays.size() < 1 )
		return;

	vector<DisplayProps*>::iterator displayitr = config.displays.begin();
	DisplayData* display = new DisplayData( NULL );
	display->setName( (*displayitr)->name );
	display->setThreaded( (*displayitr)->threaded );
	DataListViewItem* displayitem = new DataListViewItem( m_treeroot, display );
	ensureItemVisible( displayitem );
	setCurrentItem( displayitem );

    vector<WindowProps*>::iterator windowitr = (*displayitr)->windows.begin();
	while ( windowitr != (*displayitr)->windows.end() )
	{
		WindowData* window = new WindowData( display );
		display->addWindow( window );
		window->setName( (*windowitr)->name );
		window->setLeft( (*windowitr)->left );
		window->setBottom( (*windowitr)->bottom );
		window->setWidth( (*windowitr)->width );
		window->setHeight( (*windowitr)->height );
		window->setXScreen( (*windowitr)->xscreen );
		window->setQuadBuffer( (*windowitr)->quadbuffer );
		window->setBorder( (*windowitr)->border );
		window->setFullScreen( (*windowitr)->fullscreen );
		window->setSharedContextWindow( (*windowitr)->sharedcontextwindow );
		DataListViewItem* windowitem = new DataListViewItem( displayitem, window );

		vector<ViewportProps*>::iterator viewportitr = (*windowitr)->viewports.begin();
		while( viewportitr != (*windowitr)->viewports.end() )
		{
			ViewportData* viewport = new ViewportData( window );
			window->addViewport( viewport );
			viewport->setName( (*viewportitr)->name );
			viewport->setLeft( (*viewportitr)->left );
			viewport->setBottom( (*viewportitr)->bottom );
			viewport->setWidth( (*viewportitr)->width );
			viewport->setHeight( (*viewportitr)->height );
			DataListViewItem* viewportitem = new DataListViewItem( windowitem, viewport );
            if ( viewportitem )
            {
                CameraProps* cameraprop = (*viewportitr)->camera;
                CameraData* camera = viewport->getCamera();
                camera->setTrackerName( cameraprop->tracker );
                camera->setProjectionType( cameraprop->type );
                camera->setEye( cameraprop->eye );
                ScreenProps* screenprop = cameraprop->screen;
                if ( screenprop )
                {
                    ScreenData* screen = camera->getScreen();
                    screen->setCenterX( screenprop->centerx );
                    screen->setCenterY( screenprop->centery );
                    screen->setCenterZ( screenprop->centerz );
                    screen->setHeading( screenprop->normx );
                    screen->setPitch( screenprop->normy );
                    screen->setRoll( screenprop->normz );
                    screen->setWidth( screenprop->width );
                    screen->setHeight( screenprop->height );
                    screen->setDepth( screenprop->depth );
                }
                PerspectiveProps* perspectiveprop = cameraprop->perspective;
                if ( perspectiveprop )
                {
                    PerspectiveData* perspective = camera->getPerspective();
                    perspective->setFOVX( perspectiveprop->fovx );
                    perspective->setFOVY( perspectiveprop->fovy );
                }
            }
			viewportitr++;
		}
		windowitr++;
	}
	display->setFileName( filename.latin1() );
	display->setChanged( false );
}

void DisplayConfigTree::saveDisplay( const string& filename )
{
	DataListViewItem* item = (DataListViewItem*)currentItem();
	Data* data = item->getData();
	while ( data->type() != "DISPLAY" )
	{
		if ( data->type() == "ROOT" )
			return;
		data = data->getParent();
	}
	DisplayData* display = (DisplayData*)data;

	if ( !display->isChanged() )
		return;

	if ( filename == "" && display->getFileName() == "" )
	{
		saveAsDisplay();
		return;
	}

	char temp[32];
	XmlNode* rootnode = new XmlNode( "document" );
	rootnode->setParameter( "type", "DIVERSE Display Config" );
	XmlNode* displaynode = new XmlNode( "display" );
	rootnode->appendNode( displaynode );
	displaynode->setParameter( "name", display->getName() );
	displaynode->setParameter( "threaded", display->getThreaded() ? "true" : "false" );
	vector<WindowData*> windows = display->getWindows();
	vector<WindowData*>::iterator windowitr = windows.begin();
	while ( windowitr != windows.end() )
	{
		XmlNode* windownode = new XmlNode( "window" );
		displaynode->appendNode( windownode );
		windownode->setParameter( "name", (*windowitr)->getName().c_str() );
		windownode->setParameter( "left", format( temp, "%d", (*windowitr)->getLeft() ) );
		windownode->setParameter( "bottom", format( temp, "%d", (*windowitr)->getBottom() ) );
		windownode->setParameter( "width", format( temp, "%d", (*windowitr)->getWidth() ) );
		windownode->setParameter( "height", format( temp, "%d", (*windowitr)->getHeight() ) );
		windownode->setParameter( "xscreen", format( temp, "%d", (*windowitr)->getXScreen() ) );
		windownode->setParameter( "quadbuffer", (*windowitr)->getQuadBuffer() ? "true" : "false" );
		windownode->setParameter( "border", (*windowitr)->getBorder() ? "true" : "false" );
		windownode->setParameter( "fullscreen", (*windowitr)->getFullScreen() ? "true" : "false" );
		vector<ViewportData*> viewports = (*windowitr)->getViewports();
		vector<ViewportData*>::iterator viewportitr = viewports.begin();
		while ( viewportitr != viewports.end() )
		{
			XmlNode* viewportnode = new XmlNode( "viewport" );
			windownode->appendNode( viewportnode );
			viewportnode->setParameter( "name", (*viewportitr)->getName() );
			viewportnode->setParameter( "left", format( temp, "%d", (*viewportitr)->getLeft() ) );
			viewportnode->setParameter( "bottom", format( temp, "%d", (*viewportitr)->getBottom() ) );
			viewportnode->setParameter( "width", format( temp, "%d", (*viewportitr)->getWidth() ) );
			viewportnode->setParameter( "height", format( temp, "%d", (*viewportitr)->getHeight() ) );
			CameraData* camera = (*viewportitr)->getCamera();
			XmlNode* cameranode = new XmlNode( "camera" );
			viewportnode->appendNode( cameranode );
			cameranode->setParameter( "tracker", camera->getTrackerName().c_str() );
			string eyename;
			switch ( camera->getEye() )
			{
			case CameraData::MONO:
					eyename = "MONO";
					break;
				case CameraData::LEFT:
					eyename = "LEFT";
					break;
				case CameraData::RIGHT:
					eyename = "RIGHT";
					break;
				case CameraData::STEREO:
					eyename = "STEREO";
					break;
			}
			cameranode->setParameter( "eye", eyename );
			cameranode->setParameter( "type", camera->getProjectionType() ? "PERSPECTIVE" : "SCREEN" );
			if ( camera->getProjectionType() == CameraData::SCREEN )
			{
				ScreenData* screen = camera->getScreen();
				XmlNode* screennode = new XmlNode( "screen" );
				cameranode->appendNode( screennode );
				screennode->setParameter( "name", screen->getName().c_str() );
				screennode->setParameter( "centerx", format( temp, "%10.8f", screen->getCenterX() ) );
				screennode->setParameter( "centery", format( temp, "%10.8f", screen->getCenterY() ) );
				screennode->setParameter( "centerz", format( temp, "%10.8f", screen->getCenterZ() ) );
				screennode->setParameter( "normx", format( temp, "%10.8f", screen->getHeading() ) );
				screennode->setParameter( "normy", format( temp, "%10.8f", screen->getPitch() ) );
				screennode->setParameter( "normz", format( temp, "%10.8f", screen->getRoll() ) );
				screennode->setParameter( "width", format( temp, "%10.8f", screen->getWidth() ) );
				screennode->setParameter( "height", format( temp, "%10.8f", screen->getHeight() ) );
				screennode->setParameter( "depth", format( temp, "%10.8f", screen->getDepth() ) );
			}
			else if ( camera->getProjectionType() == CameraData::PERSPECTIVE )
			{
				PerspectiveData* perspective = camera->getPerspective();
				XmlNode* perspectivenode = new XmlNode( "perspective" );
				cameranode->appendNode( perspectivenode );
				perspectivenode->setParameter( "fovx", format( temp, "%10.8f", perspective->getFOVX() ) );
				perspectivenode->setParameter( "fovy", format( temp, "%10.8f", perspective->getFOVY() ) );
			}
			viewportitr++;
		}
		windowitr++;
	}

	ofstream file;
	if ( filename == "" )
		file.open( display->getFileName().c_str() );
	else
		file.open( filename.c_str() );


    if( !file.good() )
    {
        Tracer::warn( "Config file %s could not be created.", filename );
//		delete rootnode;// this crashes in windows???
		return;
	}

	rootnode->write( file, 0 );
	file << "\n";

	file.close();

	display->setFileName( filename );
	display->setChanged( false );
//	delete rootnode;// this crashes in windows???
}

void DisplayConfigTree::saveAsDisplay()
{
	QString filename( QFileDialog::getSaveFileName( QString::null, "Displays (*.xml)", this, "savedialog", "Save display configuration file" ) );
	if ( filename.isNull() )
		return;
	saveDisplay( filename.latin1() );
}

void DisplayConfigTree::closeDisplay()
{
	DataListViewItem* item = (DataListViewItem*)currentItem();
	Data* data = item->getData();
	while ( data->type() != "DISPLAY" )
	{
		if ( data->type() == "ROOT" )
			return;
		data = data->getParent();
	}
	DisplayData* display = (DisplayData*)data;
	if ( display->isChanged() )
	{
		int button = QMessageBox::information( this, "Save changes",
				"Changes have been made to this document\n"
				"Do you want to save it before exiting?",
				"&Save", "&Don't Save", "&Cancel", 0, 2 );
		if ( button == 2 )
			return;
		if ( button == 0 )
			saveDisplay( display->getFileName() );
	}
	setCurrentItem( m_treeroot );
	delete display->getItem();
}

void DisplayConfigTree::closeAllDisplays()
{
	DataListViewItem* item = (DataListViewItem*)m_treeroot->firstChild();
	while ( item )
	{
		setCurrentItem( item );
		if ( ((DisplayData*)item->getData())->isChanged() )
		{
		}
		delete item;
		item = (DataListViewItem*)m_treeroot->firstChild();
	}
	setCurrentItem( m_treeroot );
}

void DisplayConfigTree::newWindow()
{
	DataListViewItem* item = (DataListViewItem*)currentItem();
	if ( item->getData()->type() != "DISPLAY" )
		return;
	char temp[128];
    WindowData* window = new WindowData( item->getData(), format( temp,
			"Window%d", ((DisplayData*)item->getData())->getWindowNameNum() ) );
	DataListViewItem* windowitem = new DataListViewItem( item, window );
	ensureItemVisible( windowitem );
	((DisplayData*)item->getData())->addWindow( window );
    ViewportData* viewport = new ViewportData( window );
	DataListViewItem* viewportitem = new DataListViewItem( windowitem, viewport );
	ensureItemVisible( viewportitem );
	window->addViewport( viewport );
	((DisplayData*)item->getData())->setChanged( true );
}

void DisplayConfigTree::removeWindow()
{
	DataListViewItem* item = (DataListViewItem*)currentItem();
	if ( item->getData()->type() != "WINDOW" )
		return;
    DisplayData* display = (DisplayData*)((DataListViewItem*)item->parent())->getData();
	display->removeWindow( (WindowData*)item->getData() );
    setCurrentItem( item->parent() );
	delete item;
}

void DisplayConfigTree::newViewport()
{
	DataListViewItem* item = (DataListViewItem*)currentItem();
	if ( item->getData()->type() != "WINDOW" )
		return;
	char temp[128];
    ViewportData* viewport = new ViewportData( item->getData(), format( temp,
			"Viewport%d", ((WindowData*)item->getData())->getViewportNameNum() ) );
	DataListViewItem* newitem = new DataListViewItem( item, viewport );
	ensureItemVisible( newitem );
	((WindowData*)item->getData())->addViewport( viewport );
	((DisplayData*)((DataListViewItem*)item->parent())->getData())->setChanged( true );
}

void DisplayConfigTree::removeViewport()
{
	DataListViewItem* item = (DataListViewItem*)currentItem();
	if ( item->getData()->type() != "VIEWPORT" )
		return;
    WindowData* window = (WindowData*)((DataListViewItem*)item->parent())->getData();
	window->removeViewport( (ViewportData*)item->getData() );
    setCurrentItem( item->parent() );
	delete item;
}
