#include"DGL.h"
#include<iostream>
#include<string>
#include<vector>
#include<signal.h>

#ifdef WIN32
#include <windows.h>
#else
#include<dirent.h>
#endif /* WIN32 */
#include<GL/gl.h>

#ifndef __sgi
 #include<cstdio>
#else
 #include<stdio.h>
#endif

#include"DGLAugmentInterface.h"
#include<diverseconfig.h>

#include<navmgr/NavManager.h>
#include<displaymgr/DisplayManager.h>
#include<pluginmgr/PluginManager.h>
#include<servicemgr/GServiceManager.h>
#include<dmath/dMatrix.h>
#include<tracer/Tracer.h>
#include<dreg/Registry.h>
#include<dreg/Entry.h>
#include<dutil/ContextVar.h>

using namespace std;

#ifdef WIN32
#define DSEPARATOR ';'
#else
#define DSEPARATOR ':'
#endif

bool                DGL::m_initialized = false;
bool                DGL::m_configured  = false;
DGL*                DGL::m_app         = NULL;

vector<DGLAugmentInterface*> DGL::m_augments;


//=============================================================================
/// @brief Initializes the application and creates the one and only DGL object.
/// @param argc - number of command line arguments.
/// @param argv - the aurguments.
//=============================================================================
void DGL::init( int argc, char** argv )
{
    if( m_initialized ) return;
    Tracer::init();
    Registry::init();
    EventManager::init();
    NavManager::init();
    DisplayManager::init();
    if( !m_initialized ) new DGL( argc, argv );
    DisplayManager::addDrawCallback( m_app );
}

//=============================================================================
/// @brief Default constructor. Calls all configuration functions.
/// @param argc - the number of arguments to dgl.
/// @param argv - the arguments passed to dgl.
//=============================================================================
DGL::DGL( int argc, char** argv )
    :GService( "DglService" )
{
    string priority = 
        Registry::getValue( "/DIVERSE/DGL/ServicePriority", "9" );
    
    setPriority( atoi( priority.c_str() ), false );
    GServiceManager::addService( this );

    m_initialized = true;
    m_app         = this;
}

//=============================================================================
/// @brief Returns a pointer to the one and only DGL app 
/// @return a pointer to the only DGL app
//=============================================================================
DGL* DGL::instance()
{
    return m_app;
}

//=============================================================================
/// @brief Registers a DGL augment with the application so that its callbacks will
/// be called.
/// @param augment - a pointer to the augment to add.
//=============================================================================
void DGL::addAugment( DGLAugmentInterface* augment )
{
    m_augments.push_back( augment );
}

//=============================================================================
/// @brief Unloads the specified augment from DGL.
/// @param augment - a pointer to the augment to remove.
//=============================================================================
void DGL::removeAugment( DGLAugmentInterface* augment )
{
    vector<DGLAugmentInterface*>::iterator itr = m_augments.begin();
    while( itr != m_augments.end() )
    {
        if( *itr == augment )
        {
            m_augments.erase( itr );
            return;
        }
        itr++;
    }
}

//=============================================================================
/// @brief Runs preConfig for all augments.
///
/// Calls Init if DGL has not yet been initialized.
//=============================================================================
void DGL::preConfig()
{
    Tracer::debug(7, "DGL::preConfig()");
    init(); // if not initialized, initialize
    //========================================================
    // Load the DSO Augments PreConfig() methods
    //========================================================
    for( unsigned int i = 0; i < m_augments.size(); i++ )
        m_augments[i]->preConfig();
}

//=============================================================================
/// @brief Runs postConfig for all augments.
///
/// Calls Init if DGL has not yet been initialized.
//=============================================================================
void DGL::postConfig()
{
    //========================================================
    // Load the DSO Augments PostConfig() methods
    //========================================================
    for( unsigned int i = 0; i < m_augments.size(); i++ )
        m_augments[i]->postConfig();
    m_configured = true;
}

//=============================================================================
/// @brief Culls the display. Not implemented.
//=============================================================================
void DGL::cull( DCamera* )
{
}

//=============================================================================
/// @brief Clears the display. Not implemented.
//=============================================================================
void DGL::clear( DCamera* )
{
}

//=============================================================================
/// @brief This function does all the OpenGL drawing and calls the draw function
/// for all registered DGLAugments.
///
/// Because this function could be called by multiple threads it is a const 
/// function.
//=============================================================================
void DGL::draw( DCamera* )
{
	unsigned int i;
    static ContextVar<bool> inited(false);
    if ( !inited.isDefined() )
        inited = false;
    if ( !*inited )
    {
        DGLCONTEXT context = inited.getContext();
        DGLCONTEXT parent = DisplayManager::getSharedContextParent( context );
        for( i = 0; i < m_augments.size(); i++ )
        {
            if ( parent == context || parent == NULL )
                m_augments[i]->initContextLists();
            m_augments[i]->initContext();
            inited = true;
        }
    }
    for( i = 0; i < m_augments.size(); i++ )
    {
        m_augments[i]->preNavDraw();
    }
    //=========================
    // Do navigation
    //=========================
    glPushMatrix();
    glMultMatrixf(NavManager::getMatrix().ptr());
    for( i = 0; i < m_augments.size(); i++ )
    {
        glPushMatrix();
	/*
        glPushAttrib(GL_CURRENT_BIT|GL_DEPTH_BUFFER_BIT |
                     GL_ENABLE_BIT|GL_ACCUM_BUFFER_BIT |
                     GL_COLOR_BUFFER_BIT |
                     GL_TRANSFORM_BIT |
                     GL_TEXTURE_BIT |
                     GL_STENCIL_BUFFER_BIT |
                     GL_SCISSOR_BIT |
                     GL_POLYGON_STIPPLE_BIT |
                     GL_POLYGON_BIT |
                     GL_POINT_BIT |
                     GL_PIXEL_MODE_BIT |
                     GL_LIST_BIT |
                     GL_LINE_BIT |
                     GL_LIGHTING_BIT |
                     GL_HINT_BIT 
                     );
	*/
 
        m_augments[i]->draw();
        glPopMatrix();
    }
    for( i = 0; i < m_augments.size(); i++ )
    {
        glPushMatrix();
        m_augments[i]->postdraw();
        glPopMatrix();
    }
    glPopMatrix();
}

//=============================================================================
/// @brief This function does all of the "heavy lifting" between frames and is
/// where all code that may make changes is called.
//=============================================================================
void DGL::postProcess()
{
    for( unsigned int i = 0; i < m_augments.size(); i++ )
    {
        m_augments[i]->postProcess();
    }
}

//=============================================================================
/// @brief This function does all of the "heavy lifting" between frames and is
/// where all code that may make changes is called.
//=============================================================================
void DGL::preProcess()
{
    for( unsigned int i = 0; i < m_augments.size(); i++ )
    {
        m_augments[i]->preProcess();
    }
}
