/***************************************************************************
                          DVtkRenderer.cpp  - Version 0.3.2
                             -------------------
    begin                : Fri Apr 26 2002
    copyright            : (C) 2002 by Michael Shore
    email                : mshore@vt.edu
 ***************************************************************************/

/***************************************************************************
 *
 * Portions of this code modified from the Visualization Toolkit.
 * The VTK code is covered under the following copyright:
 *
 * 	Copyright (c) 1993-2002 Ken Martin, Will Schroeder, Bill Lorensen 
 *		All rights reserved.
 *		See Copyright.txt or http://www.kitware.com/Copyright.htm for details.
 *
 *		This software is distributed WITHOUT ANY WARRANTY; without even 
 *   	the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 *		PURPOSE.  See the above copyright notice for more information.
 *
 ***************************************************************************/

#include "DVtkRenderer.h"
#include <GL/gl.h>

#include <vtk/vtkProp.h>
#include <vtk/vtkMapper.h>

#include "tracer/Tracer.h"

void MyUpdateGeometry(void *ren);

//Method to create a new renderer
DVtkRenderer *DVtkRenderer::New(int argc, char** argv)
{
	return new DVtkRenderer(argc,argv);
}

//Add actor to the renderer
void DVtkRenderer::AddActor(vtkProp *p)
{
	static bool global_rendering = false;

	if (p->IsA("vtkActor") && !global_rendering)
	{
		((vtkActor*)p)->GetMapper()->GlobalImmediateModeRenderingOn();
		global_rendering = true;
	}

	vtkRenderer::AddActor(p);
}


//Method to rotate scene along X axis
void DVtkRenderer::RotateSceneX(float deg)
{
	degX = deg;
}

//The render method calls DGL methods, which render the scene via callbacks
void DVtkRenderer::Render()
{
   frame();
}


//Constructor
DVtkRenderer::DVtkRenderer(int argc, char** argv)
:DglApp(argc,argv)
{

   start();

	degX = 0;
   
   //Thread Safety
   pthread_mutex_init(&mutex, NULL);

}


//Destructor
DVtkRenderer::~DVtkRenderer()
{
	//Thread Saftey
   pthread_mutex_destroy(&mutex);
}


//Callback for OpenGL Initialization
void DVtkRenderer::initContext()
{
	
   // A bunch of GL calls to initialize lighting, materials, etc.
   // Most of this is stolen from earlier versions of vtk4CAVE, I think.
   glClearColor(this->Background[0],this->Background[1],this->Background[2],1.0);
   glClearDepth(1.);
   GLfloat mat_specular[]={1.0,1.0,1.0,1.0};
   GLfloat lt_specular[]={1.0,1.0,1.0,1.0};
   GLfloat lt_diffuse[]={0.8,0.8,0.8,1.0};
   GLfloat lt_position0[]={0,0,1,1};
   GLfloat lt_position1[]={-1,-1,1,1};
   GLfloat mat_shininess[]={50.0};
   glDisable(GL_CULL_FACE);
   glClearColor(this->Background[0],this->Background[1],this->Background[2],1.0);
   glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
   glShadeModel(GL_SMOOTH);
   glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,mat_specular);
   glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,mat_shininess);
   glLightfv(GL_LIGHT0,GL_SPECULAR,lt_specular);
   glLightfv(GL_LIGHT0,GL_DIFFUSE,lt_diffuse);
   glLightfv(GL_LIGHT0,GL_POSITION,lt_position0);
   glLightfv(GL_LIGHT1,GL_SPECULAR,lt_specular);
   glLightfv(GL_LIGHT1,GL_DIFFUSE,lt_diffuse);
   glLightfv(GL_LIGHT1,GL_POSITION,lt_position1);
   glLightModelf(GL_LIGHT_MODEL_TWO_SIDE,1.0);
   glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   glEnable(GL_LIGHT1);
   glEnable(GL_DEPTH_TEST);
   glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
   glEnable(GL_COLOR_MATERIAL);
   glLineWidth(1.0);

   //enable alpha / blending
   glEnable(GL_ALPHA_TEST);
   glAlphaFunc(GL_LESS,0);
   glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	
}


void DVtkRenderer::preProcess()
{
	/***********************************************
	 * This method contains code modified from     *
	 * part of the Render() method of vtkRenderer. *
	 * *********************************************/

	int      i;
   vtkProp  *aProp;
   
   //set background color
   glClearColor(this->Background[0],this->Background[1],this->Background[2],1.0);

   //create PropArray for storing props to be rendered
   if ( this->Props->GetNumberOfItems() > 0 )
   {
		this->PropArray = new vtkProp *[this->Props->GetNumberOfItems()];
	}
   else
   {
		this->PropArray = NULL;
   }

   //fill prop array
   this->PropArrayCount = 0;
   for ( i = 0, this->Props->InitTraversal(); (aProp = this->Props->GetNextProp());i++ )
   {
		if ( aProp->GetVisibility() )
		{
	   	this->PropArray[this->PropArrayCount++] = aProp;
		}
	}
}


//Callback to display graphics
void DVtkRenderer::draw() const
{
   /***********************************************
	 * This method contains code modified from     *
	 * part of the Render() method of vtkRenderer. *
	 * *********************************************/
	
	//if there are no props, give error message
   if ( this->PropArrayCount == 0 )
   {
		vtkDebugMacro( << "There are no visible props!" );
   }
   //otherwise, render the props
   else
   {
		//render props
		glPushMatrix();

		if (degX != 0)
			glRotatef(degX,1,0,0);

		MyUpdateGeometry((void *)this); //friend function to get around const problem

		glPopMatrix();
   }

}

//Friend function to get around problem of trying to call UpdateGeometry in a const method.
void MyUpdateGeometry(void *ren)
{
	DVtkRenderer *renderer = (DVtkRenderer*) ren;
      
	//UpdateGeometry does not appear to be thread safe.
	//Lock the renderer with a mutex while it renders 
	//to allow multi-threaded displays to render.
   renderer->lock();
	
		renderer->UpdateGeometry(); //Render actors
		
   renderer->unlock(); //After renderering, unlock so next thread can render.
}



