#include "DglApp.h"
#include<string>
#include<eventmgr/EventManager.h>
#include<navmgr/NavManager.h>
#include<tracer/Tracer.h>
#include<dreg/Registry.h>
#include<dreg/Entry.h>
#include<servicemgr/GServiceManager.h>
#include<displaymgr/DisplayManager.h>
#include<pluginmgr/PluginManager.h>
#include<dgl/dglcore/DGL.h>
#include<diverseconfig.h>
using namespace std;

//=============================================================================
// Function Name: Constructor
/// @brief Constructor. Initializes the DglApp.
/// @param argc - the number of command line arguments.
/// @param argv - the command line arguments.
/// @param cfgfile - the configuration file to be used for registry settings.
/// @param init - Always passes false. Not implemented.
//=============================================================================
DglApp::DglApp( int argc, char** argv, const string& cfgfile, bool init )
    :DiverseApp(argc, argv,  cfgfile, false )
{
    m_setup = false;
    setupRegistry();
    if( init ) initModules();
}

//=============================================================================
/// @brief Initializes the various DIVERSE managers.
//=============================================================================
void DglApp::initModules()
{
    DiverseApp::initModules();
    DGL::init();
    DGL::addAugment(this);
    
    //Set the class up to get all mouse, keyboard, wand, and tracker events
    EventManager::addMouseListener(this);
    EventManager::addKeyListener(this);
    EventManager::addWandListener(this);
    string HSegment = Registry::getValue( "/DGL/HeadTracker", "headtracker" );
    EventManager::addTrackListener( HSegment, this);
}

//=============================================================================
/// @brief Sets up the registry using the specified configuration file.
/// @param filename - the configuration file to be used for registry settings.
//=============================================================================
void DglApp::setupRegistry( const string& filename )
{
    if( m_setup ) return;
    DGL::setupRegistry( filename );
    //=================================================
    // If DIVERSE_PLUGINS is set then we don't want the 
    // normal default, however if it isn't set then load
    // the DGLDesktopGroup as default
    //=================================================
    char* plg_env = getenv("DIVERSE_PLUGINS");
	if (plg_env == NULL)
	{
	}
    else
    {
        Tracer::debug( 5, "DIVERSE_PLUGINS is set"
                          "setting DGLDesktopGroup default to no" );
    }
    
    m_setup = true;
}

//=============================================================================
// Function Name: Destructor
/// @brief Destructor.
//=============================================================================
DglApp::~DglApp()
{
}

//=============================================================================
/// @brief Set the interocular distance for the application.
/// @param iod - the interocular distance.
//=============================================================================
void DglApp::setIOD( const float& iod )
{
    DisplayManager::setIOD( iod );
}

//=============================================================================
/// @brief Get the interocular distance.
/// @return the interocular distance.
//=============================================================================
float DglApp::getIOD()const
{
    return DisplayManager::getIOD();
}

//=============================================================================
/// @brief Set the near clip plane for the application.
/// @param nclip - the near clip plane.
//=============================================================================
void DglApp::setNearClip( const float& nclip )
{
    DisplayManager::setNearClip( nclip );
}

//=============================================================================
/// @brief Get the near clip plane.
/// @return the near clip plane.
//=============================================================================
float DglApp::getNearClip()const
{
    return DisplayManager::getNearClip();
}

//=============================================================================
/// @brief Set the far clip plane.
/// @param fclip - the far clip plane.
//=============================================================================
void DglApp::setFarClip( const float& fclip )
{
    DisplayManager::setFarClip( fclip );
}

//=============================================================================
/// @brief Get the far clip plane.
/// @return the far clip plane.
//=============================================================================
float DglApp::getFarClip()const
{
    return DisplayManager::getFarClip();
}

//=============================================================================
/// @brief Add an augment to the list of augments to be called by the application.
/// @param arg - the augment to be added.
//=============================================================================
void DglApp::addAugment( DGLAugmentInterface* arg )
{
    DGL::addAugment( arg );
}

//=============================================================================
/// @brief Remove an augment from the list of augments to be called by the application.
/// @param arg - the augment to be removed.
//=============================================================================
void DglApp::removeAugment( DGLAugmentInterface* arg )
{
    DGL::removeAugment( arg );
}
