#include"DWindow.h"
#include"DViewport.h"
#include<tracer/Tracer.h>
#include<Producer/RenderSurface>

//=============================================================================
/// @brief Constructor.
///
/// Constructs a window based on the WindowProps object, the DWindow
/// takes responsibility for the config and will delete it in the destructor.
/// @param cfg - the configuration for the window
//=============================================================================
DWindow::DWindow( WindowProps* cfg )
{
    m_config = cfg;
    if( !m_config )
    {
        Tracer::error( "DWindow::DWindow( WindowProps* cfg = NULL )\n"
                       "\t\tcfg may not be null!" );
        exit(1);
    }
    VisualChooser *chooser = new VisualChooser();
    chooser->setSimpleConfiguration();
    chooser->addExtendedAttribute( VisualChooser::DepthSize, 24 );
    
    
    if( cfg->quadbuffer )
    {
        cout<<"QUAD BUFFER!!!!!\n";
        chooser->useStereo();
    }   
    m_surface = new RenderSurface();
    m_surface->setWindowName( cfg->name );
    m_surface->setScreenNum( cfg->xscreen );
    m_surface->useBorder( cfg->border );
    m_surface->setVisualChooser( chooser );
//    m_surface->fullScreen( cfg->fullscreen );
    m_surface->setWindowRectangle( cfg->left, cfg->bottom, 
                                   cfg->width, cfg->height, 
                                   cfg->resizeable );
    
    for( unsigned int i = 0; i < cfg->viewports.size(); i++ )
    {
        DViewport* vp = new DViewport( cfg->viewports[i], this );
        m_viewports.push_back(vp);
    }
    m_input    = new DMKCallback;
    m_keymouse = new KeyboardMouse(m_surface);
    m_keymouse->setCallback(m_input);
    m_input->setRS(m_surface);
    m_input->setKM(m_keymouse);
}

//=============================================================================
/// @brief Destructor.
///
/// Deletes the WindowProps data passed in the constructor.
//=============================================================================
DWindow::~DWindow()
{
    delete m_config;
}

//=============================================================================
/// @brief Get the name of the window.
/// @return a string containing the name of the window.
//=============================================================================
string DWindow::getWindowName()const
{
    return m_config->name;
}

//=============================================================================
/// @brief Set the name of the window.
///
/// The name must be unique for all children of the parent DDisplay object.
/// @param name - a string containg the name of the window.
//=============================================================================
void DWindow::setWindowName( const string& name )
{
    m_config->name = name;
    m_surface->setWindowName( name );
}

//=============================================================================
/// @brief Set the geometry of the window.
/// @param geom - a Rect containing the left and bottom coordinate and the width
/// and height of the window.
//=============================================================================
void DWindow::setWindowGeometry( const Rect& geom )
{
    m_surface->setWindowRectangle( geom.x(), geom.y(), geom.width(), geom.height() );
    m_config->left   = geom.x();
    m_config->bottom = geom.y();
    m_config->width  = geom.width();
    m_config->height = geom.height();
}

//=============================================================================
/// @brief Get the geometry of the window.
/// @return a Rect containing the left and bottom coordinate and the width
/// and height of the window.
//=============================================================================
Rect DWindow::getWindowGeometry()const
{   
    return Rect(m_config->left, m_config->bottom, 
                m_config->width, m_config->height);
}

//=============================================================================
/// @brief Get a list of all viewports associated with the window.
/// @return a vector containing pointers to each of the DViewports.
//=============================================================================
vector<DViewport*> DWindow::getViewports()const
{
    return m_viewports;
}

//=============================================================================
/// @brief Get the border property.
/// @return true if the window has a border, else false.
//=============================================================================
bool DWindow::hasBorder()const
{
    return m_config->border;
}

//=============================================================================
/// @brief Set the state of the border ( no effect in Windows ).
/// @param state - true if the window should have borders, else false.
//=============================================================================
void DWindow::setBorder( bool state )
{
    m_config->border = state;
    m_surface->useBorder(state);
}

//=============================================================================
/// @brief Called for each viewport during the preProcess stage.
//=============================================================================
void DWindow::preFrame()
{
    vector<DViewport*> vpts = getViewports();
    for( unsigned int i = 0; i < vpts.size(); i++ )
        vpts[i]->preFrame();
}

//=============================================================================
/// @brief Called for each viewport during the process stage.
//=============================================================================
void DWindow::frame()
{
    vector<DViewport*> vpts = getViewports();
    for( unsigned int i = 0; i < vpts.size(); i++ )
        vpts[i]->frame();
}

//=============================================================================
/// @brief Called for each viewport during the postProcess stage.
//=============================================================================
void DWindow::postFrame()
{
    m_keymouse->update( *m_input, false );
    vector<DViewport*> vpts = getViewports();
    for( unsigned int i = 0; i < vpts.size(); i++ )
        vpts[i]->postFrame();
}

//=============================================================================
/// @brief Called for each camera during the sync stage.
//=============================================================================
void DWindow::sync()
{
    vector<DViewport*> vpts = getViewports();
    for( unsigned int i = 0; i < vpts.size(); i++ )
        vpts[i]->sync();
}
