#include"DisplayConfig.h"
#include<dxml/XmlNode.h>
#include<dxml/XmlParser.h>
#include<tracer/Tracer.h>
#include<stdlib.h>
#include<stdio.h>
#ifdef WIN32
#include <windows.h>
#endif /* WIN32 */

//=============================================================================
/// @brief Constructor. Does nothing.
//=============================================================================
DisplayConfig::DisplayConfig()
{
}

//=============================================================================
/// @brief Opens and parses the specified configuration file.
/// @param cfgfile - the name of the file to open.
/// @return true if no errors occurred, else false.
//=============================================================================
bool DisplayConfig::open( const string& cfgfile )
{
    m_parser = new XmlParser( cfgfile );
    XmlNode* root = m_parser->getRoot();
    if( !root )
    {
        Tracer::debug(5, "Couldn't find an XML root\n" );
        return false;
    }
    return parse( root );
}

//=============================================================================
/// @brief Looks through the given node for dispay children.
/// @param root - the root node to look through.
/// @return false if no errors occured, else true.
//=============================================================================
bool DisplayConfig::parse( XmlNode* root )
{
    bool error = false;
    list<XmlNode*> children = root->getChildren();
    list<XmlNode*>::iterator itr = children.begin();

    while( itr != children.end() )
    {
        if( (*itr)->getTag() == "display" )
        {
            DisplayProps* dsp = new DisplayProps;
            bool err = dsp->parse( *itr );
            error = err || error;
            if( !err )
                displays.push_back( dsp );
            else 
            {
                Tracer::error( "Error parsing display" );
                delete dsp;
            }
        }
        else
        {
            Tracer::warn( 5, "Unknown tag \"%s\" in display section",
                              (*itr)->getTag().c_str() );
        }
        itr++;
    }
    return error;
}

//=============================================================================
/// @brief Parses the given XML node for display tags.
/// @param root - the root of the display section.
/// @return false if no errors, else true.
//=============================================================================
bool DisplayProps::parse( XmlNode* root )
{
    bool error = false;

    if( root->hasParameter( "name" ) )
        name = root->getParameter( "name" );
    else
    {
        Tracer::warn( 5, "display tag missing name property");
        name = "unnamed";
    }
    if( root->hasParameter( "threaded" ) )
    {
        threaded = root->getParameter("threaded") == "true";
    }
    else
    {
        Tracer::warn( 5, "display tag missing threaded property."
                         "\n\tDefaulting to no threads." );
        threaded = false;
    }
    
    list<XmlNode*> children = root->getChildren();
    list<XmlNode*>::iterator itr = children.begin();

    while( itr != children.end() )
    {
        if( (*itr)->getTag() == "window" )
        {
            WindowProps* win = new WindowProps;
            bool err = win->parse( *itr );
            error = err || error;
            if( !err )
                windows.push_back( win );
            else 
            {
                delete win;
                win = NULL;
                Tracer::error( "Error parsing window tag" );
            }
        }
        else
        {
            Tracer::warn( 5, "Unknown tag \"%s\" in display window's section",
                              (*itr)->getTag().c_str() );
        }
        itr++;
    }
    return error;
}

//=============================================================================
/// @brief Parses the given XML node for window tags.
/// @param root - the root of the window section.
/// @return false if no errors, else true.
//=============================================================================
bool WindowProps::parse( XmlNode* root )
{
    bool error = false;

    if( root->hasParameter( "name" ) )
        name = root->getParameter( "name" );
    else
    {
        Tracer::warn( 5, "display tag missing name property");
        name = "unnamed";
    }
    if( root->hasParameter( "sharedcontextwindow" ) )
        sharedcontextwindow = root->getParameter( "sharedcontextwindow" );
    else
    {
        Tracer::warn( 5, "display tag missing sharedcontextwindow property");
        sharedcontextwindow = "";
    }
    if( root->hasParameter( "width" ) )
        width = atoi( root->getParameter("width").c_str() );
    else
    {
        Tracer::error( 5, "window tag missing width property.");
        return true;
    }
    if( root->hasParameter( "height" ) )
        height = atoi( root->getParameter("height").c_str() );
    else
    {
        Tracer::error( 5, "window tag missing height property.");
        return true;
    }
    if( root->hasParameter( "xscreen" ) )
        xscreen = atoi( root->getParameter("xscreen").c_str() );
    else
    {
        Tracer::error( 5, "window tag missing xscreen property.");
        return true;
    }
    if( root->hasParameter( "left" ) )
        left = atoi( root->getParameter("left").c_str() );
    else
    {
        Tracer::error( 5, "window tag missing left property.");
        return true;
    }
    if( root->hasParameter( "bottom" ) )
	{
        bottom = atoi( root->getParameter("bottom").c_str() );
#ifdef WIN32
		int screenheight = GetSystemMetrics( SM_CYFULLSCREEN );
		bottom = screenheight - ( height + bottom ) + 20;
#endif
	}
    else
    {
        Tracer::error( 5, "window tag missing bottom property.");
        return true;
    }
    if( root->hasParameter( "quadbuffer" ) )
        quadbuffer = root->getParameter("quadbuffer") == "true";
    else
    {
        Tracer::warn( 5, "window tag missing quadbuffer property."
                         "Defaulting to false.");
        quadbuffer = false;
    }
    if( root->hasParameter( "border" ) )
        border = root->getParameter("border") == "true";
    else
    {
        Tracer::warn( 5, "window tag missing border property."
                         "Defaulting to false.");
        border = false;
    }
    if( root->hasParameter( "fullscreen" ) )
        fullscreen = root->getParameter("fullscreen") == "true";
    else
    {
        Tracer::warn( 5, "window tag missing full screen property."
                         "Defaulting to false.");
        fullscreen = false;
    }

    list<XmlNode*> children = root->getChildren();
    list<XmlNode*>::iterator itr = children.begin();

    while( itr != children.end() )
    {
        if( (*itr)->getTag() == "viewport" )
        {
            ViewportProps* vp = new ViewportProps;
            bool err = vp->parse( *itr );
            error = err || error;
            if( !err )
                viewports.push_back( vp );
            else 
            {
                delete vp;
                vp = NULL;
                Tracer::error( "Error parsing viewport tag" );
                return true;
            }
        }
        else
        {
            Tracer::warn( 5, "Unknown tag \"%s\" in display window's section",
                              (*itr)->getTag().c_str() );
        }
        itr++;
    }
    return error;
}

//=============================================================================
/// @brief Parses the given XML node for viewport tags.
/// @param root - the root of the viewport section.
/// @return false if no errors, else true.
//=============================================================================
bool ViewportProps::parse( XmlNode* root )
{
    bool error = false;

    if( root->hasParameter( "name" ) )
        name = root->getParameter( "name" );
    else
    {
        Tracer::warn( 5, "viewport tag missing name property");
        name = "unnamed";
    }
    if( root->hasParameter( "width" ) )
        width = atoi( root->getParameter("width").c_str() );
    else
    {
        Tracer::error( 5, "viewport tag missing width property.");
        return true;
    }
    if( root->hasParameter( "height" ) )
        height = atoi( root->getParameter("height").c_str() );
    else
    {
        Tracer::error( 5, "viewport tag missing height property.");
        return true;
    }
    if( root->hasParameter( "left" ) )
        left = atoi( root->getParameter("left").c_str() );
    else
    {
        Tracer::error( 5, "viewport tag missing left property.");
        return true;
    }
    if( root->hasParameter( "bottom" ) )
        bottom = atoi( root->getParameter("bottom").c_str() );
    else
    {
        Tracer::error( 5, "viewport tag missing bottom property.");
        return true;
    }
    
    list<XmlNode*> children = root->getChildren();
    list<XmlNode*>::iterator itr = children.begin();

    while( itr != children.end() )
    {
        if( (*itr)->getTag() == "camera" )
        {
            camera = new CameraProps;
            bool err = camera->parse( *itr );
            error = err || error;
            if( err )
            {
                Tracer::error( "Error parsing camera tag\n" );
                delete camera;
                camera = NULL;
            }
            return error;
        }
        else
        {
            Tracer::warn( 5, "Unknown tag \"%s\" in window's section",
                              (*itr)->getTag().c_str() );
        }
        itr++;
    }
    return error;
}

//=============================================================================
/// @brief Get the field of view for the width of the viewport. Only used with
/// PERSPECTIVE screen type.
/// @return the field of view of the width of the viewport.
//=============================================================================
float CameraProps::getFovX()
{
	return  perspective->fovx;
}

//=============================================================================
/// @brief Get the field of view of the height of the viewport. Only used with
/// PERSPECTIVE screen type.
/// @return the field of view of the height of the viewport.
//=============================================================================
float CameraProps::getFovY()
{
	return  perspective->fovy;
}

//=============================================================================
/// @brief Parses the given XML node for camera tags.
/// @param root - the root of the camera section.
/// @return false if no errors, else true.
//=============================================================================
bool CameraProps::parse( XmlNode* root )
{
    bool error = false;
    if( root->hasParameter( "tracker" ) )
        tracker = root->getParameter( "tracker" );
    else
    {
        Tracer::warn( 5, "viewport tag missing tracker property.\n"
                         "\tDefaulting to none.");
        tracker = "none";
    }
    string eyestr = root->getParameter( "eye" );
    if ( eyestr == "MONO" )
		eye = MONO;
	else if( eyestr == "LEFT" )
        eye = LEFT;
    else if( eyestr == "RIGHT" )
        eye = RIGHT;
    else if( eyestr == "STEREO" )
        eye = STEREO;
    else
    {
        Tracer::warn( 5, "Bad value for camera's eye parameter: %s\n",
                         "\tExpected MONO, LEFT, RIGHT or STEREO\n"
                         "\tdefaulting to MONO", eyestr.c_str());
        eye = MONO;
    }
    string typestr = root->getParameter( "type" );
    if( typestr == "SCREEN" )
        type = SCREEN;
    else if( typestr == "PERSPECTIVE" )
        type = PERSPECTIVE;
    else
    {
        Tracer::warn( 5, "Bad value for camera's type parameter: %s\n",
                         "\tExpected SCREEN or PERSPECTIVE\n"
                         "\tdefaulting to PERSPECTIVE", typestr.c_str());
        type = PERSPECTIVE;
    }
    list<XmlNode*> children = root->getChildren();
    list<XmlNode*>::iterator itr = children.begin();
    while( itr != children.end() )
    {   
        if( type == SCREEN )
        {
            Tracer::debug( "TYPE == SCREEN");
            if( (*itr)->getTag() != "screen" )
            {
                Tracer::warn( 5, "found non-screen tag \"%s\" under camera",
                                  (*itr)->getTag().c_str() );
            }
            else
            {
                screen = new ScreenProps;
                bool err = screen->parse( *itr );
                if( err )
                {
                    Tracer::error( "Error parsing Screen Props" );
                    delete screen;
                    screen = NULL;
                }
                return error || err;
            }
        }
        else  // type ==PERSPECTIVE 
        {
            if( (*itr)->getTag() != "perspective" )
            {
                Tracer::warn( 5, "found non-projection tag \"%s\" under camera",
                                  (*itr)->getTag().c_str() );
            }
            else
            {
                perspective = new PerspectiveProps;
                bool err = perspective->parse( *itr );
                error = err || error;
                if( err )
                {
                    Tracer::error( "Error parsing projection" );
                    delete perspective;
                    perspective = NULL;
                } 
                return error;
            }
        }
        itr++;
    }
    return error;
}

//=============================================================================
/// @brief Parses the given XML node for screen tags.
/// @param root - the root of the screen section.
/// @return false if no errors, else true.
//=============================================================================
bool ScreenProps::parse( XmlNode* root )
{
    if( root->hasParameter( "name" ) )
        name = root->getParameter( "name" );
    else
    {
        Tracer::warn( 5, "screen tag missing name property");
        name = "unnamed";
    }
    
    if( root->hasParameter( "centerx" ) )
    {
        string val = root->getParameter( "centerx" );
        sscanf( val.c_str(), "%lf", &centerx );
    }
    else
    {
        Tracer::warn( 5, "screen tag missing centerx property");
        centerx = 0.0;
    }
    if( root->hasParameter( "centery" ) )
    {
        string val = root->getParameter( "centery" );
        sscanf( val.c_str(), "%lf", &centery );
    }
    else
    {
        Tracer::warn( 5, "screen tag missing centery property");
        centery = 0.0;
    }
    if( root->hasParameter( "centerz" ) )
    {
        string val = root->getParameter( "centerz" );
        sscanf( val.c_str(), "%lf", &centerz );
    }
    else
    {
        Tracer::warn( 5, "screen tag missing centerz property");
        centerz = 0.0;
    }
    if( root->hasParameter( "normx" ) )
    {
        string val = root->getParameter( "normx" );
        sscanf( val.c_str(), "%lf", &normx );
    }
    else
    {
        Tracer::warn( 5, "screen tag missing normx property");
        normx = 0.0;
    }
    if( root->hasParameter( "normy" ) )
    {
        string val = root->getParameter( "normy" );
        sscanf( val.c_str(), "%lf", &normy );
    }
    else
    {
        Tracer::warn( 5, "screen tag missing normy property");
        normy = 0.0;
    }
    if( root->hasParameter( "normz" ) )
    {
        string val = root->getParameter( "normz" );
        sscanf( val.c_str(), "%lf", &normz );
    }
    else
    {
        Tracer::warn( 5, "screen tag missing normz property");
        normz = 0.0;
    }
    if( root->hasParameter( "width" ) )
    {
        string val = root->getParameter( "width" );
        sscanf( val.c_str(), "%lf", &width );
    }
    else
    {
        Tracer::warn( 5, "screen tag missing width property");
        width = 1;
    }
    if( root->hasParameter( "height" ) )
    {
        string val = root->getParameter( "height" );
        sscanf( val.c_str(), "%lf", &height );
    }
    else
    {
        Tracer::warn( 5, "screen tag missing height property");
        height = 1;
    }
    if( root->hasParameter( "depth" ) )
    {
        string val = root->getParameter( "depth" );
        sscanf( val.c_str(), "%lf", &depth );
    }
    else
    {
        Tracer::warn( 5, "screen tag missing depth property");
        depth = 1;
    }
    return false;
}   

//=============================================================================
/// @brief Parses the given XML node for perspective tags.
/// @param root - the root of the perspective section.
/// @return false if no errors, else true.
//=============================================================================
bool PerspectiveProps::parse( XmlNode* root )
{
    if( root->hasParameter( "fovy" ) )
    {
        string val = root->getParameter( "fovy" );
        sscanf( val.c_str(), "%f", &fovy );
    }
    else
    {
        Tracer::warn( 5, "projection tag missing fovy property, defaulting to 45");
        fovy = 45;
    }
    if( root->hasParameter( "fovx" ) )
    {
        string val = root->getParameter( "fovx" );
        sscanf( val.c_str(), "%f", &fovx );
    }
    else
    {
        Tracer::warn( 5, "projection tag missing fovx property, defaulting to 45");
        fovx = 45;
    }
    return false;  // no errors
}   
