#ifndef _DISPLAYCONFIG_H_
#define _DISPLAYCONFIG_H_
#include<string>
#include<vector>

#ifdef WIN32
#ifdef DISPLAYMGR_EXPORTS
#define DISPLAYMGR_API __declspec(dllexport)
#else
#define DISPLAYMGR_API __declspec(dllimport)
#endif /* DISPLAYMGR_EXPORTS */
#else
#define DISPLAYMGR_API
#endif /* WIN32 */

using namespace std;

class XmlNode;
class XmlParser;


//=============================================================================
/// @class ScreenProps
/// @author Daniel Larimer
/// @brief Encapsulates a screen's properties.
///
/// Center is the coordinate representing where the eye is looking from the origin.
/// The norm coordinate represents the angles ( in degrees ) of rotation about
/// each major axis. The width, depth, and height are the number of meters for each
/// of the sides of the unit cube.
///
/// All positions and dimensions are in meters. All angles are in degrees.
//=============================================================================
class DISPLAYMGR_API ScreenProps
{
    public:
        bool parse( XmlNode* root );
        
        string name;
        double centerx;
        double centery;
        double centerz;
        double normx;
        double normy;
        double normz;
        double width;
        double height;
        double depth;  
};

//=============================================================================
/// @class PerspectiveProps
/// @author Daniel Larimer
/// @brief Encapsulates a screen's properties.
///
/// Simpler configuration which allows you to pass the FOV or field of view
/// for the width and height of your viewport.
//=============================================================================
class DISPLAYMGR_API PerspectiveProps
{
    public:
        bool parse( XmlNode* root );
        /// Field of view in degrees
        float fovx;
        float fovy;
};

//=============================================================================
/// @class CameraProps
/// @author Daniel Larimer
/// @brief Encapsulates a camera's properties.
///
/// A camera can either be used with a SCREEN or PERSPECTIVE configuration.
/// The respective property class must be defined and the associated type set.
/// The tracker property is the name of the tracker device to be used with this display.
/// The eye property must be MONO, LEFT, RIGHT, or STEREO.
///
/// MONO - standard monoscopic display.
/// LEFT - the view from the left eye.
/// RIGHT - the view from the right eye.
/// STEREO - a stereoscopic view.
//=============================================================================
class DISPLAYMGR_API CameraProps
{
    public:
        CameraProps() { perspective= NULL;
                        screen     = NULL; }
        enum EYE { MONO, LEFT, RIGHT, STEREO };
        enum TYPE{ SCREEN, PERSPECTIVE };
        
        bool parse( XmlNode* root );
	float getFovX();
	float getFovY();
        
        string           tracker;
        int              eye;
        int              type;
        PerspectiveProps* perspective;
        ScreenProps*     screen;
};

//=============================================================================
/// @class ViewportProps
/// @author Daniel Larimer
/// @brief Encapsulates a viewport's properties.
///
/// The name must be unique for the window. left, bottom, width, and height are
/// used to determine the size and position of the viewport within the window.
//=============================================================================
class DISPLAYMGR_API ViewportProps
{
    public:
        bool parse( XmlNode* root );
        
        string       name;
        int          width;
        int          height;
        int          left;
        int          bottom;
        CameraProps* camera;
};

//=============================================================================
/// @class WindowProps
/// @author Daniel Larimer
/// @brief Encapsulates a window's properties.
///
/// The name of each window of a display must be unique within that display.
/// The left, bottom, width, and height position and size the display. Quadbuffer
/// implies that the graphics hardware supports stereo. Border tells whether a
/// window border is desired ( in Windows the windows will always have borders
/// May change ina future version ). fullscreen is a request for a full screen window.
/// resizeable tells whether you want the window to be resizable. xscreen is the
/// screen number that the window should appear in.
//=============================================================================
class DISPLAYMGR_API WindowProps
{
    public:
        bool parse( XmlNode* root );

        string                 name;
		string                 sharedcontextwindow;
        int                    width;
        int                    height;
        int                    left;
        int                    bottom;
        bool                   quadbuffer;
        bool                   border;
        bool                   fullscreen;
        bool                   resizeable;
        int                    xscreen;
        vector<ViewportProps*> viewports;
};

//=============================================================================
/// @class DisplayProps
/// @author Daniel Larimer
/// @brief Encapsulates a display's properties.
///
/// The display must have a name. threaded tells whether the display should be
/// threaded ( does not function properly on all systems ).
//=============================================================================
class DISPLAYMGR_API DisplayProps
{
    public:
        bool parse( XmlNode* root );

        string               name;
        bool                 threaded;
        vector<WindowProps*> windows;
};

//=============================================================================
/// @class DisplayConfig
/// @author Daniel Larimer
/// @brief Encapsulates the configuration from the display xml file.
//=============================================================================
class DISPLAYMGR_API DisplayConfig
{
    public:
        DisplayConfig();
        bool open( const string& cfgfile );
        bool parse( XmlNode* root );

        vector<DisplayProps*> displays;

    private:
        XmlParser*  m_parser;
};

#endif
