#include <iostream>
#include <map>
#include <string>
#include <vector>
#include <dmath/Rect.h>
using namespace std;
#include "DisplayManager.h"
#include "DisplayCallback.h"
#include <tracer/Tracer.h>
#include <dreg/Registry.h>
#include <diverseconfig.h>
#include <servicemgr/GServiceManager.h>
#include "DDisplay.h"

vector<DisplayCallback*>   DisplayManager::m_cullcbs;
vector<DisplayCallback*>   DisplayManager::m_clearcbs;
vector<DisplayCallback*>   DisplayManager::m_drawcbs;
vector<DisplayCallback*>   DisplayManager::m_overlaycbs;
vector<DDisplay*>          DisplayManager::m_displays;
map<DGLCONTEXT,DGLCONTEXT> DisplayManager::m_sharedcontexts;
DisplayConfig              DisplayManager::m_config;
DisplayManager*            DisplayManager::m_self = NULL;
bool                       DisplayManager::m_initialized = false;
bool                       DisplayManager::m_setup = false;
float                      DisplayManager::m_nclip = 0.1;
float                      DisplayManager::m_fclip = 5000;
float                      DisplayManager::m_iod = 0.03;
float                      DisplayManager::m_mpu = 1;

//=============================================================================
/// @brief Constructor. Reads parses the config file and builds the displays
/// @param cfg - the name of the config file, if not set the value of
/// /DIVERSE/DisplayManager/ConfigFile is used from the registry,
/// which is set to ENVIRONMENT_VAR by default and therefore the
/// DIVERSE_DISPLAY_CONFIG environment variable is used.
//=============================================================================
DisplayManager::DisplayManager( const string& cfg )
    :GService( "DisplayManager" )
{
    string priority = 
        Registry::getValue( "/DIVERSE/DisplayManager/ServicePriority", "10" );
    
    setPriority( atoi( priority.c_str() ), false );
    GServiceManager::addService( this );

    string ignore = Registry::getValue("/DIVERSE/DisplayManager/NoDisplays", "false", "false or true" );
    if( ignore != "true" )
    {
        string configfile = cfg;
        if( configfile == "" )
        {
            configfile = Registry::getValue("/DIVERSE/DisplayManager/ConfigFile" );
        }
        char* cfgfilename = getenv("DIVERSE_DISPLAY_CONFIG");
        if (cfgfilename != NULL)
            configfile = string(cfgfilename);
        
        if( m_config.open( configfile ) )
        {
            Tracer::error( "Couldn't open display configuration file: %s",
                            configfile.c_str() );
            return;
        }
        for( unsigned int i = 0; i < m_config.displays.size(); i++ )
        {
            m_displays.push_back( new DDisplay( m_config.displays[i] ) );
        }
    }
}

//=============================================================================
/// @brief Initializes the DisplayManager.
/// @param cfg - the configuration file to be used for the registry settings.
//=============================================================================
void DisplayManager::init( const string& cfg )
{
    if( m_initialized ) return;
    setupRegistry();
    m_self = new DisplayManager( cfg );
    m_initialized = true;
}

//=============================================================================
/// @brief Sets up the registry for the displaymgr.
/// @param cfg - the configuration file to be used for the registry settings.
//=============================================================================
void DisplayManager::setupRegistry( const string& cfg )
{
    if( m_setup ) return;
    Registry::init(cfg);

    Registry::getValue("/DIVERSE/DisplayManager", "none", 
                       "Display manager settings go under this tag" );
    string DefaultDisplay = string(INSTALLDIR) + "/etc/" + "desktopdisplay.xml";
    Registry::getValue("/DIVERSE/DisplayManager/NoDisplays", "false", "false or true" );
    Registry::getValue("/DIVERSE/DisplayManager/ConfigFile", DefaultDisplay ); 
    Registry::getValue("/DIVERSE/DisplayManager/", DefaultDisplay ); 
    m_setup = true;
}

//=============================================================================
/// @brief Called before config() which allows plugins to modify the displays
/// before the graphics context is realized.
//=============================================================================
void DisplayManager::preConfig()
{
    for( unsigned int i = 0; i < m_displays.size(); i++ )
    {
        m_displays[i]->preConfig();    
    }
}


//=============================================================================
/// @brief Called to create the displays and initialize the OpenGL context of
/// each display.
//=============================================================================
void DisplayManager::config()
{
    Tracer t("DisplayManager::config()");
    for( unsigned int i = 0; i < m_displays.size(); i++ )
    {
        m_displays[i]->config();    
    }
}

//=============================================================================
/// @brief Called after all of the displays have been created and realized.
//=============================================================================
void DisplayManager::postConfig()
{
    for( unsigned int i = 0; i < m_displays.size(); i++ )
        m_displays[i]->postConfig();    
}

//=============================================================================
/// @brief Called once per frame before threads fork to render the scene.
//=============================================================================
void DisplayManager::preProcess()
{
    for( unsigned int i = 0; i < m_displays.size(); i++ )
        m_displays[i]->preFrame();    
}

//=============================================================================
/// @brief Called to clear, cull, draw, and then sync all displays
//=============================================================================
void DisplayManager::process()
{
    for( unsigned int i = 0; i < m_displays.size(); i++ )
    {
        m_displays[i]->frame();    
//        m_displays[i]->advance();
//        m_displays[i]->sync();    
    }
}

//=============================================================================
/// @brief Called while various threads render the different views of the scene.
//=============================================================================
void DisplayManager::postProcess()
{
    for( unsigned int i = 0; i < m_displays.size(); i++ )
    {
        m_displays[i]->postFrame();    
    }
}

//=============================================================================
/// @brief Waits for all threads to complete.
//=============================================================================
void DisplayManager::sync()
{
//    for( unsigned int i = 0; i < m_displays.size(); i++ )
//        m_displays[i]->sync();    
}

//=============================================================================
/// @brief Returns a list of the displays.
/// @return the current displays.
//=============================================================================
vector<DDisplay*> DisplayManager::getDisplays()
{
    return m_displays;
}

//=============================================================================
/// @brief Get the number of displays in the system.
/// @return the current number displays.
//=============================================================================
int DisplayManager::numDisplays()
{
    return m_displays.size();
}

//=============================================================================
/// @brief Add a child context to a parent.
///
/// This function does NOT set up sharing of contexts. The contexts must have
/// set up to share. However, this function must be used if sharing has been
/// initiated so that DIVERSE knows that it should look at the parent context
/// for display lists and texture names rather than creating new lists and
/// textures for the child context.
/// @param child - the child context ( HGLRC for Windows, GLXContext for XWindows ).
/// @param parent - the child context ( HGLRC for Windows, GLXContext for XWindows ).
//=============================================================================
bool DisplayManager::addSharedContext( DGLCONTEXT child, DGLCONTEXT parent )
{
    map<DGLCONTEXT,DGLCONTEXT>::iterator childitr = m_sharedcontexts.find( child );
    if ( childitr == m_sharedcontexts.end() )
    {
        m_sharedcontexts[child] = parent;
        return true;
    }
    else
    {
        Tracer::warn( "Child context is already shared" );
        return false;
    }
}

//=============================================================================
/// @brief Get the shared context parent for the specified context.
/// @param child - the context for which to find the parent.
/// @return the parent context, or NULL if there is not a parent.
//=============================================================================
DGLCONTEXT DisplayManager::getSharedContextParent( DGLCONTEXT child )
{
    map<DGLCONTEXT,DGLCONTEXT>::iterator childitr;
    for ( childitr = m_sharedcontexts.begin();childitr != m_sharedcontexts.end();childitr++)
    {
        if ( childitr->first == child )
            return childitr->second;
    }
    Tracer::warn( 6, "Child context is not shared or is nonexistent" );
    return NULL;
}

//=============================================================================
/// @brief Get all of the children for the specified context.
/// @param parent - the context for which to find the children.
/// @return a vector containing a list of all children. The vector may be empty.
//=============================================================================
vector<DGLCONTEXT> DisplayManager::getSharedContextChildren( DGLCONTEXT parent )
{
    vector<DGLCONTEXT> children;
    map<DGLCONTEXT,DGLCONTEXT>::iterator childitr;
    for ( childitr = m_sharedcontexts.begin();childitr != m_sharedcontexts.end();childitr++ )
    {
        if ( childitr->second == parent )
            children.push_back( childitr->first );
    }
    return children;
}

//=============================================================================
/// @brief Remove a child from the list of shared contexts.
/// @param child - the child context to remove from sharing.
/// @return true if the context was successfully removed, else returns false.
//=============================================================================
bool DisplayManager::removeSharedContext( DGLCONTEXT child )
{
    map<DGLCONTEXT,DGLCONTEXT>::iterator childitr = m_sharedcontexts.find( child );
    if ( childitr != m_sharedcontexts.end() )
    {
        m_sharedcontexts.erase( childitr );
        return true;
    }
    Tracer::warn( "Child context cannot be removed - does not exist" );
    return false;
}

//=============================================================================
/// @brief Adds the display call back to a list that will get called by each
/// viewport in every display every frame.
/// @param cb - the call back to remove.
//=============================================================================
void DisplayManager::removeDrawCallback( DisplayCallback* cb )
{
    vector<DisplayCallback*>::iterator itr = m_drawcbs.begin();
    while( itr != m_drawcbs.end() )
    {
        if( *itr == cb )
        {
            m_drawcbs.erase(itr);
            return;
        }
        itr++;
    }
}

//=============================================================================
/// @brief Adds the display call back to a list that will get called by each
/// viewport in every display every frame.
/// @param cb - the call back to remove.
//=============================================================================
void DisplayManager::removeOverlayCallback( DisplayCallback* cb )
{
    vector<DisplayCallback*>::iterator itr = m_overlaycbs.begin();
    while( itr != m_overlaycbs.end() )
    {
        if( *itr == cb )
        {
            m_overlaycbs.erase(itr);
            return;
        }
        itr++;
    }
}

//=============================================================================
/// @brief Adds the display call back to a list that will get called by each
/// viewport in every display every frame.
/// @param cb - the call back to remove.
//=============================================================================
void DisplayManager::removeCullCallback( DisplayCallback* cb )
{
    vector<DisplayCallback*>::iterator itr = m_cullcbs.begin();
    while( itr != m_cullcbs.end() )
    {
        if( *itr == cb )
        {
            m_cullcbs.erase(itr);
            return;
        }
        itr++;
    }
}

//=============================================================================
/// @brief Adds the display call back to a list that will get called by each
/// viewport in every display every frame.
/// @param cb - the call back to remove.
//=============================================================================
void DisplayManager::removeClearCallback( DisplayCallback* cb )
{
    vector<DisplayCallback*>::iterator itr = m_clearcbs.begin();
    while( itr != m_clearcbs.end() )
    {
        if( *itr == cb )
        {
            m_clearcbs.erase(itr);
            return;
        }
        itr++;
    }
}

//=============================================================================
/// @brief Adds the display call back to a list that will get called by each
/// viewport in every display every frame.
/// @param cb - the call back to add.
//=============================================================================
void DisplayManager::addDrawCallback( DisplayCallback* cb )
{
    m_drawcbs.push_back(cb);
}

//=============================================================================
/// @brief Adds the display call back to a list that will get called by each
/// viewport in every display every frame.
/// @param cb - the call back to add.
//=============================================================================
void DisplayManager::addOverlayCallback( DisplayCallback* cb )
{
    m_overlaycbs.push_back(cb);
}

//=============================================================================
/// @brief Adds the display call back to a list that will get called by each
/// viewport in every display every frame
/// @param cb - the call back to add
//=============================================================================
void DisplayManager::addCullCallback( DisplayCallback* cb )
{
    m_cullcbs.push_back(cb);
}

//=============================================================================
/// @brief Adds the display call back to a list that will get called by each
/// viewport in every display every frame
/// @param cb - the call back to add
//=============================================================================
void DisplayManager::addClearCallback( DisplayCallback* cb )
{
    m_clearcbs.push_back(cb);
}

//=============================================================================
/// @brief Spews OpenGL for all registered callbacks.
/// @param cam - the description of the view to be drawn.
//=============================================================================
void DisplayManager::draw( DCamera* cam )
{
    for( unsigned int i = 0; i < m_drawcbs.size(); i++ )
    {
        m_drawcbs[i]->draw(cam);
    }
}

//=============================================================================
/// @brief Draws the overlays.
//=============================================================================
void DisplayManager::overlay()
{
    for( unsigned int i = 0; i < m_overlaycbs.size(); i++ )
    {
        m_overlaycbs[i]->overlay();
    }
}

//=============================================================================
/// @brief Calls clear for all registered call backs.
/// @param cam - the description of the view to be cleared.
//=============================================================================
void DisplayManager::clear( DCamera* cam )
{
    for( unsigned int i = 0; i < m_clearcbs.size(); i++ )
    {
        m_clearcbs[i]->clear(cam);
    }
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );
}

//=============================================================================
/// @brief Calls cull for all registered call backs.
/// @param cam - the description of the view to be culled.
//=============================================================================
void DisplayManager::cull( DCamera* cam )
{
    for( unsigned int i = 0; i < m_cullcbs.size(); i++ )
    {
        m_cullcbs[i]->cull(cam);
    }
}

//=============================================================================
/// @brief Scales a vector in meters to DIVERSE units.
/// @param world - the vector to be scaled in meters.
/// @return a vector converted to DIVERSE units.
//=============================================================================
dVec3 DisplayManager::worldToUnit( const dVec3& world )
{
    return world * getUnitsPerMeter();
}

//=============================================================================
/// @brief Scales a vector in DIVERSE units to meters.
/// @param unit - the vector to be scaled in DIVERSE units.
/// @return a vector converted to meters.
//=============================================================================
dVec3 DisplayManager::unitToWorld( const dVec3& unit )
{
    return unit * getMetersPerUnit();
}
