#ifndef _DISPLAYMANAGER_H_
#define _DISPLAYMANAGER_H_
#include<vector>
#include<map>
#include<string>
#include<displaymgr/DisplayConfig.h>
#include<servicemgr/GService.h>
#include<dmath/dVec3.h>
#include<Producer/Camera>

#ifdef WIN32
    #include <windows.h>
#else
    #include <GL/glx.h>
#endif

using namespace std;

#ifdef WIN32
    #ifdef DISPLAYMGR_EXPORTS
        #define DISPLAYMGR_API __declspec(dllexport)
    #else
        #define DISPLAYMGR_API __declspec(dllimport)
    #endif /* DISPLAYMGR_EXPORTS */
#else
    #define DISPLAYMGR_API
#endif /* WIN32 */

#ifdef WIN32
    #define DGLCONTEXT HGLRC
#else
    #define DGLCONTEXT GLXContext
#endif

class DDisplay;
class DisplayCallback;
class DCamera;

//=============================================================================
/// @class DisplayManager
/// @brief The DisplayManager manages Open Producer displays used by DIVERSE.
///
/// Most likely it will not be necessary to use any of the displaymgr classes
/// directly. The DisplayManager uses display configuration files to build a
/// layout of the displays requested. All of the necessary information to use
/// Open Producer displays can be accessed through the display configuration
/// files.
///
/// However, should the need arise, the DisplayManager, like most other DIVERSE
/// managers, is a static object initialized by the core application. It must
/// be initialized prior to calling any of its functions.
///
/// During the process stage of the application loop, there are several callback
/// functions possibly called for each display. There are functions to add
/// callbacks for each of these functions.
/// clear - clears the frame buffers for the display's context.
/// cull - performs any required culling of the display.
/// draw - draw to the frame.
/// overlay - draw to the frame over what already exists in the frame buffer.
///
/// Calibration functions are available which can be used to match the display
/// size with the DIVERSE unit cube. Units per meter is the number unit cube
/// lengths per meter. The meters per unit is the measurement of meters per
/// unit cube length. The near clip plane is the point closest to the eye that
/// the image will be displayed. The far clip plane is the farthest distance
/// from the eye that the image will be displayed. The image will be clipped at
/// both the near and the far clip planes. The near clipping plane affects the
/// resolution of the depth buffer and should be placed as far from the eye as
/// possible. The depth buffer is set to 24 bits in DIVERSE. If you find objects
/// are not showing up properly in regard to their depth, you may wish to increase
/// the value of the near clipping plane to improve the depth resolution. IOD is
/// the inter ocular distance or distance between the eyes. This is used for in
/// calculations using stereoscopic views.
///
/// The DisplayManager manages the displays and all display callbacks. Display
/// callbacks are derived from the DisplayCallback class and should use the proper
/// add callback function to be included in the processing loop. To stop the
/// callback from being called in the loop, use the appropriate remove callback
/// function.
//=============================================================================
class DISPLAYMGR_API DisplayManager : public GService
{
    public:
        static void init(const string& cfg = "");
        
        static void draw ( DCamera* cam );
        static void overlay();
        static void clear( DCamera* cam );
        static void cull ( DCamera* cam );

        static void addDrawCallback ( DisplayCallback* cb );
        static void addOverlayCallback ( DisplayCallback* cb );
        static void addCullCallback ( DisplayCallback* cb );
        static void addClearCallback( DisplayCallback* cb );
        static void removeDrawCallback ( DisplayCallback* cb );
        static void removeOverlayCallback ( DisplayCallback* cb );
        static void removeCullCallback ( DisplayCallback* cb );
        static void removeClearCallback( DisplayCallback* cb );
       
        static void setupRegistry( const string& cfg = "diverse.xml" );

        static vector<DDisplay*>  getDisplays();
        static int                numDisplays();

        static bool               addSharedContext( DGLCONTEXT child, 
                                                    DGLCONTEXT parent );
        static DGLCONTEXT         getSharedContextParent( DGLCONTEXT child );
        static vector<DGLCONTEXT> getSharedContextChildren( DGLCONTEXT parent );
        static bool               removeSharedContext( DGLCONTEXT child );

        void preConfig();
        void config();
        void postConfig();

        //=================================================================
        /// Used to calibrate the display for world coordinates based on
        /// the DIVERSE unit cube.
        //=================================================================
		/// @brief Set the units per meter.
		/// @param upm - the units per meter.
        static void  setUnitsPerMeter( const float upm ){ m_mpu = 1.0/upm; }
		/// @brief Set the meters per unit.
		/// @param mpu - the meters per unit.
        static void  setMetersPerUnit( const float mpu ){ m_mpu   = mpu;   }
		/// @brief Set the near clip plane.
		/// @param clip - the distance to the near clip plane. Must be positive.
        static void  setNearClip( const float& clip )   { m_nclip = clip;  }
		/// @brief Set the far clip plane.
		/// @param clip - the distance to the far clip plane. Must be positive.
        static void  setFarClip ( const float& clip )   { m_fclip = clip;  }
		/// @brief Set the inter ocular distance
		/// @param iod - inter ocular distance ( distance between the eyes ).
        static void  setIOD     ( const float& iod  )   { m_iod   = iod;   }

		/// @brief Get the meters per unit.
		/// @return the meters per unit.
        static float getMetersPerUnit()               { return m_mpu;     }
		/// @brief Get the units per meter.
		/// @return the units per meter.
        static float getUnitsPerMeter()               { return 1.0/m_mpu; }
		/// @brief Get the near clip plane distance.
		/// @return the near clip plane distance.
        static float getNearClip()                    { return m_nclip;   }
		/// @brief Get the far clip plane distance.
		/// @return the far clip plane distance.
        static float getFarClip()                     { return m_fclip;   }
		/// @brief Get the inter ocular distance.
		/// @return the inter occular distance.
        static float getIOD()                         { return m_iod;     }

        static dVec3 worldToUnit( const dVec3& vec );
        static dVec3 unitToWorld( const dVec3& vec );
        
        /// @brief Calls preFrame for all displays.
        void preFrame() { preProcess(); }
        void preProcess();
        
        /// @brief Calls frame for each display.
        void frame() { process(); }
        void process();
        
        /// @brief Calls postFrame for each display.
        void postFrame() { postProcess(); }
        void postProcess();
        
        void sync();
    
    private:
        DisplayManager(const string& cfg = "");
        static bool                         m_initialized;
        static bool                         m_setup;
        static vector<DDisplay*>            m_displays;
        static map<DGLCONTEXT,DGLCONTEXT>   m_sharedcontexts;
        static DisplayConfig                m_config;
        static vector<DisplayCallback*>     m_clearcbs;
        static vector<DisplayCallback*>     m_drawcbs;
        static vector<DisplayCallback*>     m_cullcbs;
        static vector<DisplayCallback*>     m_overlaycbs;
        static DisplayManager*              m_self;
        static float                        m_nclip;
        static float                        m_fclip;
        static float                        m_iod;
        static float                        m_mpu; // meters per unit
};
#endif
