#include"PerspectiveHandler.h"
#include"DisplayManager.h"
#include"DisplayConfig.h"
#include<tracer/Tracer.h>
#include<dgl/dglcore/DGL.h>
#include<GL/glu.h>
#include<eventmgr/devents.h>
#include<eventmgr/EventManager.h>
#include"DCamera.h"
using namespace Producer;

//=============================================================================
/// @brief Constructor.
/// @param cfg - the configuration for the perspective handler.
//=============================================================================
PerspectiveHandler::PerspectiveHandler( PerspectiveProps* cfg )
{
	if (cfg == NULL)
		Tracer::error("Warning, display config has a problem, please rebuild it\n");
    m_config = cfg;
}

//=============================================================================
/// @brief Clears the scene.
/// @param cam - the Producer camera to clear
//=============================================================================
void PerspectiveHandler::clear( Camera& cam )
{
    return;
#ifdef WIN32
    SceneHandler::clear(cam);
#else
    Producer::Camera::SceneHandler::clear(cam);
#endif
    int x, y;
    unsigned int w, h;
    cam.getProjectionRectangle( x, y, w, h );
    float* VMatrix = (float*)cam.getViewMatrix();
    float* PAMatrix = (float*)cam.getPositionAndAttitudeMatrix();
    if (VMatrix == NULL || PAMatrix == NULL)
    {
	Tracer::error("Perspective handler conversion problem\n");
    }
    DCamera dcam( Rect(x,y,w,h), VMatrix, PAMatrix);
                               
//    glDrawBuffer(GL_BACK);
//    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glClearColor( 0, 0, 0, 0 );
    DisplayManager::clear(&dcam);
}

//=============================================================================
/// @brief Draws the DGL scene by calling DGL::Draw() which in turn calls
/// all of the DGLAugment->Draw() functions.
/// @param cam - the Producer camera viewing the scene
//=============================================================================
void PerspectiveHandler::draw( Camera& cam )
{
	if (m_config == NULL)
		Tracer::error("Bad display config, please check file\n");
    int x, y;
    unsigned int w, h;
    cam.getProjectionRectangle( x, y, w, h );
    glMatrixMode( GL_PROJECTION );
        glLoadIdentity();
	if (m_config != NULL)
        gluPerspective( m_config->fovx, m_config->fovx/m_config->fovy,
                        DisplayManager::getNearClip(), DisplayManager::getFarClip() );
    
    // TBD: Optimize this
    DTrackerEvent* headtracker = 
                    EventManager::getTrackerState("headtracker");
    
    dMatrix headmat;
    headmat.makeIdentity();
    // rotate about 0, 0, 0 and then translate to head pos
    dVec3 headpos(0,0,0);
    dVec3 headdir(0,1,0);
    if( headtracker )
    {
        headpos = headtracker->getPosition();// = DisplayManager::getHeadPos();
        headdir = headtracker->getDirection(); 
        dMatrix euler;
        euler.makeRotate( DEG2RAD(headdir.h()), dVec3( 0, 0, -1 ),
                          DEG2RAD(headdir.p()+90), dVec3( -1, 0, 0 ),
                          DEG2RAD(headdir.r()), dVec3( 0, -1, 0 ) );
        // rotate about 0, 0, 0 and then translate to head pos
        headmat = dMatrix::translate( -headpos ) * euler;
    }

    glMatrixMode( GL_MODELVIEW );
        glLoadIdentity();
        gluLookAt( 0, 0, 0, 0, 1, 0, 0, 0, 1 );
        glRotatef( 90, 1, 0, 0 );
        glMultMatrixf( headmat.ptr() );

    dMatrix projmat;
    if (m_config != NULL)
    {
    projmat.makePerspective( m_config->fovx, m_config->fovx/m_config->fovy,
                                DisplayManager::getNearClip(), DisplayManager::getFarClip() );
    }
    DCamera dcam( Rect( x, y, w, h ), projmat.ptr(), headmat.ptr() );

    DisplayManager::cull(&dcam);
    DisplayManager::draw(&dcam);
}

//=============================================================================
/// @brief Sets up the projection, culls and draws the scene.
/// @param cam - the Producer camera for which to draw the frame.
/// @return always returns true.
//=============================================================================
bool PerspectiveHandler::frame( Camera& cam )
{
    cam.getRenderSurface()->makeCurrent();
    
//    cam.clear();
    glMatrixMode( GL_PROJECTION );
    //int x, y;
    int w = cam.getRenderSurface()->getWindowWidth();
	int h = cam.getRenderSurface()->getWindowHeight();
//    cam.getRenderSurface()->getWindowRectangle( x, y, w, h );
    glViewport( 0, 0, w, h );
    glScissor( 0, 0, w, h );
//    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT |GL_STENCIL_BUFFER_BIT );
    
    // TBD: Optimize this
    DTrackerEvent* headtracker = 
                    EventManager::getTrackerState("headtracker");
    
    dMatrix headmat;
    headmat.makeIdentity();
    // rotate about 0, 0, 0 and then translate to head pos
    dVec3 headpos(0,0,0);
    dVec3 headdir(0,1,0);
    if( headtracker )
    {
        headpos = headtracker->getPosition();// = DGL::getHeadPos();
        headdir = headtracker->getDirection(); 
        dMatrix euler;
        euler.makeRotate( DEG2RAD(headdir.h()), dVec3( 0, 0, -1 ),
                          DEG2RAD(headdir.p()), dVec3( -1, 0, 0 ),
                          DEG2RAD(headdir.r()), dVec3( 0, -1, 0 ) );
        // rotate about 0, 0, 0 and then translate to head pos
        headmat = dMatrix::translate( -headpos ) * euler;
    }

    glMatrixMode( GL_PROJECTION );
        glLoadIdentity();
//        glViewport( 0, 0, w, h );
	if (m_config != NULL)
	{
        gluPerspective( m_config->fovx, float(w)/h, 
                        DisplayManager::getNearClip(), 
                        DisplayManager::getFarClip() );
	}
//        cout<<"Viewport( 0, 0, "<<w<<", "<<h<<" )\r";
//        cam.setPerspectiveRectangle( 0, 0, w, h );

    glMatrixMode( GL_MODELVIEW );
        glLoadIdentity();
        glRotatef( -90.0f, 1.0f, 0.0f, 0.0f );
        glMultMatrixf( headmat.ptr() );

    dMatrix viewmat;
    if (m_config !=NULL)
    {
	    viewmat.makePerspective( m_config->fovx, float(w)/h,
                                        DisplayManager::getNearClip(), 
                                        DisplayManager::getFarClip() );
    }

    headmat = headmat * dMatrix::rotate( -M_PI/2.0, 1, 0, 0 );
    DCamera dcam( Rect( 0, 0, w, h ),
                 viewmat.ptr(), headmat.ptr() ); 

    DisplayManager::clear(&dcam);
    DisplayManager::cull(&dcam);
    DisplayManager::draw(&dcam);
    DisplayManager::overlay();
    return true;
}

//=============================================================================
/// @brief Culls the scene.
/// @param cam - the Producer camera viewing the scene.
//=============================================================================
void PerspectiveHandler::cull( Camera& cam )
{
    int x, y;
    unsigned int w, h;
    cam.getProjectionRectangle( x, y, w, h );
    float* VMatrix = (float*)cam.getViewMatrix();
    float* PAMatrix = (float*)cam.getPositionAndAttitudeMatrix();
    if (VMatrix == NULL || PAMatrix == NULL)
    {
	Tracer::error("Perspective handler conversion problem\n");
    }
    DCamera dcam( Rect(x,y,w,h), VMatrix, PAMatrix);
                                 
    DisplayManager::cull( &dcam ); 
}
