#include"ScreenHandler.h"
#include<dgl/dglcore/DGL.h>
#include<GL/glu.h>
#include<dmath/dVec3.h>
#include<dmath/dMatrix.h>
#include<dmath/Rect.h>
#include<eventmgr/devents.h>
#include<dreg/Registry.h>
//#include<Producer/Matrix>

#include"DCamera.h"
#include"DisplayConfig.h"
#include"DisplayManager.h"

//=============================================================================
/// @brief Constructor.
/// @param cfg - the configuration for the screen handler.
/// @param view - the index of the view.
/// @param tracker - the name of the tracker device.
//=============================================================================
ScreenHandler::ScreenHandler( ScreenProps* cfg, int view, const string& tracker ) 
{
    m_config = cfg;
    m_view = view;
    m_head = tracker;
    m_pos_orig = DisplayManager::worldToUnit(dVec3( cfg->centerx, cfg->centery, cfg->centerz));
    setPosition( m_pos_orig, dVec3( cfg->normx, cfg->normy, cfg->normz) );
}

//=============================================================================
/// @brief Sets the position of the screen.
/// @param pos - the position of the center of the screen.
/// @param ang - the normal to the screen.
//=============================================================================
void ScreenHandler::setPosition( const dVec3& pos, const dVec3& ang )
{
    m_rot_mat = dMatrix::rotate( DEG2RAD(-ang.h()), dVec3( 0, 0, 1 ),
                                 DEG2RAD(-ang.p()), dVec3( 1, 0, 0 ),
                                 DEG2RAD(-ang.r()), dVec3( 0, 1, 0 ) );
    m_inv_rot_mat = dMatrix::inverse(m_rot_mat);
    //===================================
    // Find Normal
    //===================================
    dVec4 n(0,1,0,0);
    m_norm = m_rot_mat * n;
    m_norm.normalize();
   
    //cout<<"norm = "<<m_norm.x()<<", "<<m_norm.y()<<", "<<m_norm.z()<<endl;
    //===================================
    // Find Screen left, right, top, bottom
    //===================================
    m_pos = m_inv_rot_mat * dVec4( pos.x(), pos.y(), pos.z(), 0 );
    //cout<<"pos = "<<m_pos.x()<<", "<<m_pos.y()<<", "<<m_pos.z()<<endl;
    
    m_screen_left   = m_pos.x() - 1;
    m_screen_right  = m_pos.x() + 1;
    m_screen_top    = m_pos.z() + 1;
    m_screen_bottom = m_pos.z() - 1;

    //==================================
    // Find the up vector
    //==================================
    m_up = dVec4( 0, 0, 1, 0 );
    m_up = m_rot_mat * m_up;
    m_up.normalize();

    //cout<<"up = "<<m_up.x()<<", "<<m_up.y()<<", "<<m_up.z()<<endl;

    //cout<<"===========================================\n";
}

//=============================================================================
/// @brief Used to set the aspect ration of the screen.  Default is 1.  
/// @param height_over_width - the height of the screen divided by width.
//=============================================================================
void ScreenHandler::setAspectRatio( float height_over_width )
{
    m_aspect = height_over_width;
    m_screen_top    = m_pos.z() + height_over_width;
    m_screen_bottom = m_pos.z() - height_over_width;
}

//=============================================================================
/// @brief Sets up the Projection and Model View matrix for the screen and
/// then renders in stereo, if specified
/// @param cam - the Producer camera used for the screen.
/// @return always true - we want to override Camera::frame()
//=============================================================================
bool ScreenHandler::frame( Camera& cam )
{
    cam.getRenderSurface()->makeCurrent();
    int x, y;
    unsigned int w, h;
    cam.getRenderSurface()->getWindowRectangle( x, y, w, h );
   // glViewport( 0, 0, w, h );
    
    float* VMatrix = (float*)cam.getViewMatrix();
    float* PAMatrix = (float*)cam.getPositionAndAttitudeMatrix();
    if (VMatrix == NULL || PAMatrix == NULL)
    {
    	Tracer::error("ScreenHandler had conversion problem");
    }
    DCamera dcam( Rect(0,0,w,h), VMatrix, PAMatrix);
                                 
    //=================================
    // Clear the buffers
    //=================================
    clear(cam);
    
    //==========================================
    // Derive the two eye position vector
    //==========================================
    DTrackerEvent* headstate = EventManager::getTrackerState(m_head);
    dVec3 headdir;
    if( headstate )
        headdir = headstate->getDirection();
    headdir.p() += 90;

    // This may slow things down, maybe move it someplace else
    dMatrix rotmat;
    rotmat.makeRotate( DEG2RAD(headdir.h()), dVec3( 0, 0, 1 ),
                       DEG2RAD(headdir.p()), dVec3( 1, 0, 0 ),
                       DEG2RAD(headdir.r()), dVec3( 0, 1, 0 ) );
    headdir = rotmat * dVec3( 0,1,0 );
    headdir.normalize();

    
    m_eyedelta = headdir ^ dVec3( 0., 0., 1. ); 
                        // view direction cross up vector
    m_eyedelta.normalize();
    
    m_eyedelta.x() *= DisplayManager::getIOD()/2.0;
    m_eyedelta.y() *= DisplayManager::getIOD()/2.0;
    m_eyedelta.z() *= DisplayManager::getIOD()/2.0;

    if( m_view == CameraProps::STEREO )
    {
        glDrawBuffer(GL_BACK_LEFT);
        glClear(  GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
        frame( CameraProps::LEFT, cam );
        glDrawBuffer(GL_BACK_RIGHT);
        glClear(  GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
        frame( CameraProps::RIGHT, cam );
    }
    else{
        glDrawBuffer( GL_BACK );
        frame( m_view, cam );
    }
    return true;
}

//=============================================================================
/// @brief Clear the scene.
/// @param cam - the Producer camera used for the screen.
//=============================================================================
void ScreenHandler::clear( Camera& cam )
{
    //===========================================
    // Clear Left and Right Buffers
    //===========================================
    if( m_view == CameraProps::STEREO )
    {
        glDrawBuffer(GL_BACK_LEFT);
        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
        glDrawBuffer(GL_BACK_RIGHT);
        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    }else
    {
        glDrawBuffer(GL_BACK);
        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    }
}


//=============================================================================
/// @brief Draws sets the frustum for the left or right eye and draws
/// assumes that the proper buffer has been selected
/// @param view - one of the values from the enum view.
/// @param cam - the Producer camera used for the screen.
//=============================================================================
void ScreenHandler::frame( int view, Camera& cam )
{
    int x, y;
    unsigned int w, h;
    cam.getRenderSurface()->getWindowRectangle( x, y, w, h );

    dVec3 headpos;
    DTrackerEvent* headevt = EventManager::getTrackerState(m_head);
    if( headevt )
        headpos = headevt->getPosition();

//cout<<"eyedelta.x = "<<m_eyedelta.x()<<"  eyedelta.y = "<<m_eyedelta.y()<<endl;
    //=============================
    // Calculate Position
    //=============================
    dVec3 eye;
    if( view == CameraProps::MONO )
    {
        //cout<<"LEFT_EYE\n";
        eye = headpos;
    }
    else if( view == CameraProps::LEFT )
    {
        //cout<<"LEFT_EYE\n";
        eye = headpos + m_eyedelta;
    }
    else if( view == CameraProps::RIGHT )
    {
        //cout<<"RIGHT EYE\n";
        eye = headpos - m_eyedelta;
    }
        
    // Translate eye to face screen
    dVec3 eyetx = m_inv_rot_mat * eye;
    
    dVec3 v = m_pos_orig - eye;
    // Distance from eye to screen
    float D = v.x()* m_norm.x() + v.y()*m_norm.y() + v.z()*m_norm.z();
    
    // Calculate "scale"
    float s = 0.001 / (D);
    
    float left   = (m_screen_left   - eyetx.x() ) * s;
    float right  = (m_screen_right  - eyetx.x() ) * s;
    float top    = (m_screen_top    - eyetx.z() ) * s;
    float bottom = (m_screen_bottom - eyetx.z() ) * s;
    
    glMatrixMode( GL_PROJECTION );
        glLoadIdentity();
        glFrustum( left, right, bottom, top,
                   0.001, DisplayManager::getFarClip() );
    
    glMatrixMode( GL_MODELVIEW );
        glLoadIdentity();
        gluLookAt( 
               eye.x(), eye.y(), eye.z(),
               eye.x() + m_norm.x(),
               eye.y() + m_norm.y(),
               eye.z() + m_norm.z(),
               m_up.x(), m_up.y(), m_up.z() );
    dMatrix proj;
    proj.makeFrustum( 
               left, right, bottom, top,
               0.001, DisplayManager::getFarClip() );
        
    dMatrix viewmat;
    viewmat.makeLookAt( 
               eye,
               dVec3(eye.x() + m_norm.x(),
               eye.y() + m_norm.y(),
               eye.z() + m_norm.z()),
               dVec3(m_up.x(), m_up.y(), m_up.z()) );

    DCamera dcam( Rect(0,0,w,h), proj.ptr(), 
                                 viewmat.ptr() );
               
    //=================================
    // Give DIVERSE plugins a chance to cull
    //=================================
    DisplayManager::cull(&dcam);
    DisplayManager::draw(&dcam);
    DisplayManager::overlay();
}
