//=============================================================================
// ScreenHandler.h
//   - This file defines the DGL screen class
//
//   Created 5/12/03 By Daniel Larimer
//=============================================================================
#ifndef _DGLSCREEN_H_
#define _DGLSCREEN_H_
#include<dmath/dVec3.h>
#include<dmath/dVec4.h>
#include<Producer/Camera>
#include<dmath/dMatrix.h>
#include<eventmgr/EventManager.h>

#ifdef WIN32
#ifdef DISPLAYMGR_EXPORTS
#define DISPLAYMGR_API __declspec(dllexport)
#else
#define DISPLAYMGR_API __declspec(dllimport)
#endif /* DISPLAYMGR_EXPORTS */
#else
#define DISPLAYMGR_API
#endif /* WIN32 */

using namespace Producer;
class ScreenProps;

//=============================================================================
/// @class ScreenHandler
/// @brief Sets up a camera and window for screen based virtual environments.
///
/// Creates two cameras, one for each eye, and a window that they will be drawn
/// onto. If the screen is mono, only one camera is used.  
//=============================================================================
class DISPLAYMGR_API ScreenHandler : public Camera::SceneHandler 
{
    public:
		/// @brief The type of view or eye.
        enum VIEW
		{
			/// standard monoscopic display.
			MONO,
			/// the view from the left eye.
			MONO_LEFT,
			/// a stereoscopic view.
			STEREO,
			/// the view from the right eye.
			MONO_RIGHT
		};
        ScreenHandler( ScreenProps* cfg, int view, const string& headname );

        void setPosition( const dVec3& pos, const dVec3& norm );
        void setAspectRatio( float height_over_width );
		/// @brief Get the aspect ratio of the screen.
		/// @return the aspect ratio.
        float getAspectRatio()const { return m_aspect; }

        //=============================================================
        // Overloaded so we can control the projection and model view
        // and then render in stereo
        //=============================================================
        virtual bool frame( Camera& cam );
        virtual void clear( Camera& cam );
		/// @brief Not implemented.
		/// @return always returns false.
        virtual bool useAutoView(){ return false; }

        /// @brief not used because frame takes care of it for us
        virtual void draw( Camera& ){};
        
    private:
        void frame( int view, Camera& cam );
        string               m_head;
        dMatrix              m_rot_mat;
        dMatrix              m_inv_rot_mat;
        dVec4                m_pos;
        dVec3                m_pos_orig;
        dVec3                m_eyedelta;
        dVec4                m_up;
        dVec4                m_norm;
        int                  m_view;
        float                m_screen_left;
        float                m_screen_right;
        float                m_screen_top;
        float                m_screen_bottom;
        float                m_aspect;
        ScreenProps*         m_config;
};

#endif
