#include"GVector.h"
#include<math.h>


//=============================================================================
//  Function Name: GVector::GVector 
/// @brief Constructs a GVector from an origin to the tip
/// @param origin - the origin of the vector
/// @param tip    - the tip of the vector
//=============================================================================
GVector::GVector( const dVec3& origin, const dVec3& tip )
{
    m_origin     = origin;
    m_direction  = tip - origin;
}

//=============================================================================
//  Function Name: GVector::GVector 
/// @brief Constructs a GVector with the origin at (0,0,0) and the 
/// direction specified by dx, dy and dz.
/// @param dx - the x coordinate of the direction vector.
/// @param dy - the y coordinate of the direction vector.
/// @param dz - the z coordinate of the direction vector.
//=============================================================================
GVector::GVector( float dx, float dy, float dz )
    :m_origin( 0, 0, 0 ), m_direction( dx, dy, dz )
{
}

//=============================================================================
//  Function Name: GVector::GVector
/// @brief Copy constructor.
/// @param copy - the GVector being copied. 
//=============================================================================
GVector::GVector( const GVector& copy )
    :m_origin( copy.m_origin ), m_direction( copy.m_direction )
{
}

//=============================================================================
//  Function Name: GetTip
/// @brief Calculates the location of the tip of the vector and returns it as
/// a point.
/// @return the tip of the vector.
//=============================================================================
dVec3 GVector::GetTip()const
{
    return m_origin + m_direction;
}

//=============================================================================
//  Function Name: GetOrigin
/// @brief returns the origin of the vector.
/// @return the origin of the vector.
//=============================================================================
dVec3 GVector::GetOrigin()const
{
    return m_origin;
}

//=============================================================================
//  Function Name: SetTip
/// @brief Sets the tip of the vector to a new location leaving the origin where
/// it was.  Net result is changing the magnitude and direction.
/// @param tip - the new tip for the vector.
//=============================================================================
void GVector::SetTip( const dVec3& tip )
{
    m_direction = tip - m_origin;
}

//=============================================================================
//  Function Name: SetOrigin
/// @brief Sets the origin of the vector but keeps the direction the same.
/// @param origin - the new origin for the vector.
//=============================================================================
void GVector::SetOrigin( const dVec3& origin )
{
    m_origin = origin;
}

//=============================================================================
//  Function Name: GetMagnitude
/// @brief Calculates and returns the magnitude of the vector.
/// @return the magnitude.
//=============================================================================
float GVector::GetMagnitude()const
{
    return sqrt(x()*x() + y()*y() + z()*z());
}

//=============================================================================
//  Function Name: GetMagSquared
/// @brief Calculates and returns the magnitude of the vector.
/// @return the magnitude.
//=============================================================================
float GVector::GetMagSquared()const
{
    return x()*x() + y()*y() + z()*z();
}

//=============================================================================
//  Function Name: SetMagnitude
/// @brief Sets a new magnitude for the vector.
/// @param mag - the new magnitude.
//=============================================================================
void GVector::SetMagnitude( const float& mag )
{
    float current_mag = GetMagnitude();
    if( current_mag == mag ) return;

    m_direction.x() /= current_mag;
    m_direction.y() /= current_mag;
    m_direction.z() /= current_mag;

    if( mag == 1 ) return;
    
    m_direction.x() *= mag;
    m_direction.y() *= mag;
    m_direction.z() *= mag;
}

//=============================================================================
//  Function Name: ToUnit
/// @brief Returns the unit vector, leaves this vector unchanged.
/// @return the unit vector in this direction.
//=============================================================================
GVector GVector::ToUnit()const
{
    GVector tmp(*this);
    tmp.SetMagnitude( 1 );
    return tmp;
}

//=============================================================================
//  Function Name: operator* ( float scale )
/// @brief Scales the vector by a scalar.
/// @param scale - the amount to scale by.
/// @return the vector scaled by scale.
//=============================================================================
GVector GVector::operator* ( const float &scale )const
{
    GVector tmp(*this);
    tmp.x() *= scale;
    tmp.y() *= scale;
    tmp.z() *= scale;
    return tmp;
}

//=============================================================================
//  Function Name: operator*=
/// @brief Scales this vector by the specified amount.
/// @param scale - the amount to scale by.
/// @return a reference to this vector.
//=============================================================================
GVector& GVector::operator*=( const float &scale )
{
    x() *= scale;
    y() *= scale;
    z() *= scale;
    return *this;
}

//=============================================================================
//  Function Name: operator*( const GVector& )
/// @brief Returns the dot product.
/// @param vec - the vector to dot with.
/// @return the dot product.
//=============================================================================
float GVector::operator*( const GVector& vec )const
{
    return m_direction * vec.m_direction;
}

//=============================================================================
//  Function Name: operator*( const GVector& )
/// @brief Returns the cross product.
/// @param vec - the vector to cross with.
/// @return the cross product.
//=============================================================================
GVector GVector::operator^( const GVector& vec )const
{
    GVector tmp(*this);
    tmp.m_direction = tmp.m_direction ^ vec.m_direction;
    return tmp;
}

//=============================================================================
//  Function Name: operator*( const GVector& )
/// @brief Returns the cross product.
/// @param vec - the vector to cross with.
/// @return the cross product.
//=============================================================================
GVector& GVector::operator^=( const GVector& vec )
{
    m_direction = m_direction ^ vec.m_direction;
    return *this;
}

//=============================================================================
//  Function Name:operator=( const GVector vec )
/// @brief Assignment operator.
/// @param vec - the vector to assign to this.
/// @return a reference to this vector.
//=============================================================================
GVector& GVector::operator=( const GVector& vec )
{
    m_origin    = vec.m_origin;
    m_direction = vec.m_direction;
    return *this;
}

//=============================================================================
//  Function Name: operator+( const GVector& )
/// @brief Returns the sum of the two vectors.
/// @param vec - the vector to add with.
/// @return the sum of the vectors.
//=============================================================================
GVector GVector::operator+( const GVector& vec )const
{
    GVector tmp(*this);
    tmp.m_direction += vec.m_direction;
    return tmp;
}

//=============================================================================
//  Function Name: operator+=( const GVector& )
/// @brief Returns the sum of the two vectors.
/// @param vec - the vector to add to this one.
/// @return a reference to this vector.
//=============================================================================
GVector& GVector::operator+=( const GVector& vec )
{
    m_direction += vec.m_direction;
    return *this;
}

//=============================================================================
//  Function Name: operator-( const GVector& )
/// @brief Returns the sum of the two vectors.
/// @param vec - the vector to subtract with.
/// @return the sum of the vectors.
//=============================================================================
GVector GVector::operator-( const GVector& vec )const
{
    GVector tmp(*this);
    tmp.m_direction -= vec.m_direction;
    return tmp;
}

//=============================================================================
//  Function Name: operator-=( const GVector& )
/// @brief Returns the sum of the two vectors.
/// @param vec - the vector to cross with.
/// @return a reference to this vector.
//=============================================================================
GVector& GVector::operator-=( const GVector& vec )
{
    m_direction -= vec.m_direction;
    return *this;
}

//=============================================================================
//  Function Name: operator<<(ostream& os, const GVector& v )
/// @fn ostream& operator<<(ostream& os, const GVector& v )
/// @brief Print out the vector using standard ostream output ( i.e. cout ).
//=============================================================================
ostream& operator<<(ostream& os, const GVector& v )
{
    dVec3 o = v.GetOrigin();
    os<<"( "<<o.x()<<", "<<o.y()<<", "<<o.z()<<" ) -> ( "
              <<v.x()<<", "<<v.y()<<", "<<v.z()<<" )";
    return os;
}
