//=============================================================================
// Rect.h
// - This class defines a simple rectangle defined by top left corner and 
// a width, height
//
// Created 9/21/02 By Daniel Larimer modified by Andrew A. Ray
//=============================================================================
#ifndef _RECT_H_
#define _RECT_H_

//=============================================================================
/// @class Rect
/// @brief Encapsulates a minimal set of properties for rectangles used by DIVERSE.
///
/// The Rect class is defined by the upper left coordinate and its width and
/// height of the rectangle. The class only supports integer arguments.
/// x() will return the left coordinate. y() will return
/// the upper coordinate. right() and bottom() are calulated based on the upper left
/// coordinate and the width and height respectively.
///
/// As DIVERSE is uses a lower left coordinate as the origin of the display,
/// this class can be used similarly. The Rect would then be defined using the
/// lower left coordinate and the bottom() function would then return the upper
/// coordinate.
//=============================================================================
class Rect
{
	public:
		/// @brief Constructor.
		/// @param x - left coordinate of the rectangle.
		/// @param y - upper coordinate of the rectangle.
		/// @param w - width of the rectangle.
		/// @param h - height of the rectangle.
		Rect( int x = 0, int y = 0, int w = 0, int h = 0 )
		{
			m_x = x;
			m_y = y;
			m_w = w;
			m_h = h;
		}
		/// @brief Sets the left coordinate.
		/// @param x - left coordinate of the rectangle.
        void setx(const int& x){m_x = x;};
		/// @brief Sets the upper coordinate.
		/// @param y - upper coordinate of the rectangle.
        void sety(const int& y){m_y = y;};
		/// @brief Sets the width.
		/// @param w - width of the rectangle.
        void setw(const int& w){m_w = w;};
		/// @brief Sets the height.
		/// @param h - height of the rectangle.
        void seth(const int& h){m_h = h;};
		/// @brief Get the left coordinate of the rectangle.
		/// @return the left coordinate of the rectangle.
		int x()const { return m_x; }
		/// @brief Get the upper coordinate of the rectangle.
		/// @return the upper coordinate of the rectangle.
		int y()const { return m_y; }
		/// @brief Get the width of the rectangle.
		/// @return the width of the rectangle.
		int width()const { return m_w; }
		/// @brief Get the height of the rectangle.
		/// @return the height of the rectangle.
		int height()const { return m_h; }
		/// @brief Get the right coordinate of the rectangle.
		/// @return the right coordinate of the rectangle.
		int right()const { return m_x + m_w; }
		/// @brief Get the bottom coordinate of the rectangle.
		/// @return the bottom coordinate of the rectangle.
		int bottom()const { return m_y + m_h; }

		/// @brief determines whether the point is within the rectangle.
		/// @param x - the x coordinate of the point to test.
		/// @param y - the y coordinate of the point to test.
		/// @return true if the point is within the rectangle, else returns false.
		bool intersects( int x, int y )const
		{
			return (x > m_x) && (x < (m_x+m_w)) && (y > m_y) && (y < (m_y+m_h));
		}

	private:
		int m_x;
		int m_y;
		int m_w;
		int m_h;
};

#endif

