/* -*-c++-*- OpenSceneGraph - Copyright (C) 1998-2003 Robert Osfield
 *
 * This library is open source and may be redistributed and/or modified under  
 * the terms of the OpenSceneGraph Public License (OSGPL) version 0.0 or 
 * (at your option) any later version.  The full license is in LICENSE file
 * included with this distribution, and on the openscenegraph.org website.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * OpenSceneGraph Public License for more details.
*/
#include "dMatrix.h"
#include "dQuat.h"
#include "dmath.h"


#define SET_ROW(row, v1, v2, v3, v4 )    \
    _mat[(row)][0] = (v1); \
    _mat[(row)][1] = (v2); \
    _mat[(row)][2] = (v3); \
    _mat[(row)][3] = (v4);

#define INNER_PRODUCT(a,b,r,c) \
     ((a)._mat[r][0] * (b)._mat[0][c]) \
    +((a)._mat[r][1] * (b)._mat[1][c]) \
    +((a)._mat[r][2] * (b)._mat[2][c]) \
    +((a)._mat[r][3] * (b)._mat[3][c])


//=============================================================================
/// @brief Default constructor.
//=============================================================================
dMatrix::dMatrix()
{
    makeIdentity();
}

//=============================================================================
/// @brief Copy constructor.
/// @param other - the matrix whose values are to be copied to this one.
//=============================================================================
dMatrix::dMatrix( const dMatrix& other)
{
    set( (const float *) other._mat );
}

//=============================================================================
/// @brief Constructor. Set to the values referenced by the pointer.
/// @param def - a float pointer to 16 contiguous floats.
//=============================================================================
dMatrix::dMatrix( const float * const def )
{    
    set( def ); 
}

//=============================================================================
/// @brief Constructor. Set to the values specified.
//=============================================================================
dMatrix::dMatrix( float a00, float a01, float a02, float a03,
                float a10, float a11, float a12, float a13,
                float a20, float a21, float a22, float a23,
                float a30, float a31, float a32, float a33)
{
    SET_ROW(0, a00, a01, a02, a03 )
    SET_ROW(1, a10, a11, a12, a13 )
    SET_ROW(2, a20, a21, a22, a23 )
    SET_ROW(3, a30, a31, a32, a33 )
}

//=============================================================================
/// @brief Set this matrix to the values specified.
//=============================================================================
void dMatrix::set( float a00, float a01, float a02, float a03,
                  float a10, float a11, float a12, float a13,
                  float a20, float a21, float a22, float a23,
                  float a30, float a31, float a32, float a33)
{
    SET_ROW(0, a00, a01, a02, a03 )
    SET_ROW(1, a10, a11, a12, a13 )
    SET_ROW(2, a20, a21, a22, a23 )
    SET_ROW(3, a30, a31, a32, a33 )
}

//=============================================================================
/// @brief Set the translation to the values specified.
/// @param tx, ty, tz - the x, y, z values to be set.
//=============================================================================
void dMatrix::setTrans( float tx, float ty, float tz )
{
    _mat[3][0] = tx;
    _mat[3][1] = ty;
    _mat[3][2] = tz;
}


//=============================================================================
/// @brief Set the translation to the values specified.
/// @param v - a vector containing the values to be set.
//=============================================================================
void dMatrix::setTrans( const dVec3& v )
{
    _mat[3][0] = v[0];
    _mat[3][1] = v[1];
    _mat[3][2] = v[2];
}

//=============================================================================
/// @brief Set the matrix to identity.
//=============================================================================
void dMatrix::makeIdentity()
{
    SET_ROW(0,    1, 0, 0, 0 )
    SET_ROW(1,    0, 1, 0, 0 )
    SET_ROW(2,    0, 0, 1, 0 )
    SET_ROW(3,    0, 0, 0, 1 )
}

//=============================================================================
/// @brief Set the matrix to identity and multiply by scale factors.
/// @param v - a vector containing the scale vectors.
//=============================================================================
void dMatrix::makeScale( const dVec3& v )
{
    makeScale(v[0], v[1], v[2] );
}

//=============================================================================
/// @brief Set the matrix to identity and multiply by scale factors.
/// @param x, y, z - the scale factors for each axis.
//=============================================================================
void dMatrix::makeScale( float x, float y, float z )
{
    SET_ROW(0,    x, 0, 0, 0 )
    SET_ROW(1,    0, y, 0, 0 )
    SET_ROW(2,    0, 0, z, 0 )
    SET_ROW(3,    0, 0, 0, 1 )
}

//=============================================================================
/// @brief Set the matrix to identity and add the translation.
/// @param v - a vector containing the translation.
//=============================================================================
void dMatrix::makeTranslate( const dVec3& v )
{
    makeTranslate( v[0], v[1], v[2] );
}

//=============================================================================
/// @brief Set the matrix to identity and add the translation.
/// @param x, y, z - the translation coordinates.
//=============================================================================
void dMatrix::makeTranslate( float x, float y, float z )
{
    SET_ROW(0,    1, 0, 0, 0 )
    SET_ROW(1,    0, 1, 0, 0 )
    SET_ROW(2,    0, 0, 1, 0 )
    SET_ROW(3,    x, y, z, 1 )
}

//=============================================================================
/// @brief Set the matrix to identity and rotate from one vector to another.
/// @param from - the start vector.
/// @param to - the end vector.
//=============================================================================
void dMatrix::makeRotate( const dVec3& from, const dVec3& to )
{
    dQuat quat;
    quat.makeRotate(from,to);
    quat.get(*this);
}

//=============================================================================
/// @brief Set the matrix to identity and rotate an angle about an axis.
/// @param angle - the rotation angle (radians).
/// @param axis - a vector containing the axis of rotation.
//=============================================================================
void dMatrix::makeRotate( float angle, const dVec3& axis )
{
    dQuat quat;
    quat.makeRotate( angle, axis);
    quat.get(*this);
}

//=============================================================================
/// @brief Set the matrix to identity and rotate an angle about an axis.
/// @param angle - the rotation angle (radians).
/// @param x, y, z - the values of the axis of rotation.
//=============================================================================
void dMatrix::makeRotate( float angle, float x, float y, float z ) 
{
    dQuat quat;
    quat.makeRotate( angle, x, y, z);
    quat.get(*this);
}

//=============================================================================
/// @brief Set the matrix to identity and rotate an angle about an axis.
/// @param q - the quaternion representing the rotation.
//=============================================================================
void dMatrix::makeRotate( const dQuat& q )
{
    q.get(*this);    
}

//=============================================================================
/// @brief Set the matrix to identity and rotate an angle about an axis.
/// @param angle1, angle2, angle3 - the rotation angles (radians).
/// @param axis1, axis2, axis3 - the respective axes of rotation.
//=============================================================================
void dMatrix::makeRotate( float angle1, const dVec3& axis1, 
                         float angle2, const dVec3& axis2,
                         float angle3, const dVec3& axis3)
{
    dQuat quat;
    quat.makeRotate(angle1, axis1, 
                    angle2, axis2,
                    angle3, axis3);
    quat.get(*this);
}

//=============================================================================
/// @brief Multiply 2 matrices.
/// @param lhs - the left matrix.
/// @param rhs - the right matrix.
//=============================================================================
void dMatrix::mult( const dMatrix& lhs, const dMatrix& rhs )
{   
    if (&lhs==this)
    {
        postMult(rhs);
        return;
    }
    if (&rhs==this)
    {
        preMult(lhs);
        return;
    }

// PRECONDITION: We assume neither &lhs nor &rhs == this
// if it did, use preMult or postMult instead
    _mat[0][0] = INNER_PRODUCT(lhs, rhs, 0, 0);
    _mat[0][1] = INNER_PRODUCT(lhs, rhs, 0, 1);
    _mat[0][2] = INNER_PRODUCT(lhs, rhs, 0, 2);
    _mat[0][3] = INNER_PRODUCT(lhs, rhs, 0, 3);
    _mat[1][0] = INNER_PRODUCT(lhs, rhs, 1, 0);
    _mat[1][1] = INNER_PRODUCT(lhs, rhs, 1, 1);
    _mat[1][2] = INNER_PRODUCT(lhs, rhs, 1, 2);
    _mat[1][3] = INNER_PRODUCT(lhs, rhs, 1, 3);
    _mat[2][0] = INNER_PRODUCT(lhs, rhs, 2, 0);
    _mat[2][1] = INNER_PRODUCT(lhs, rhs, 2, 1);
    _mat[2][2] = INNER_PRODUCT(lhs, rhs, 2, 2);
    _mat[2][3] = INNER_PRODUCT(lhs, rhs, 2, 3);
    _mat[3][0] = INNER_PRODUCT(lhs, rhs, 3, 0);
    _mat[3][1] = INNER_PRODUCT(lhs, rhs, 3, 1);
    _mat[3][2] = INNER_PRODUCT(lhs, rhs, 3, 2);
    _mat[3][3] = INNER_PRODUCT(lhs, rhs, 3, 3);
}

//=============================================================================
/// @brief Pre-multiply this matrix with another.
/// @param other - the other matrix to multiply.
//=============================================================================
void dMatrix::preMult( const dMatrix& other )
{
    // brute force method requiring a copy
    //dMatrix tmp(other* *this);
    // *this = tmp;

    // more efficient method just use a float[4] for temporary storage.
    float t[4];
    for(int col=0; col<4; ++col) {
        t[0] = INNER_PRODUCT( other, *this, 0, col );
        t[1] = INNER_PRODUCT( other, *this, 1, col );
        t[2] = INNER_PRODUCT( other, *this, 2, col );
        t[3] = INNER_PRODUCT( other, *this, 3, col );
        _mat[0][col] = t[0];
        _mat[1][col] = t[1];
        _mat[2][col] = t[2];
        _mat[3][col] = t[3];
    }

}

//=============================================================================
/// @brief Post-multiply this matrix with another.
/// @param other - the other matrix to multiply.
//=============================================================================
void dMatrix::postMult( const dMatrix& other )
{
    // brute force method requiring a copy
    //dMatrix tmp(*this * other);
    // *this = tmp;

    // more efficient method just use a float[4] for temporary storage.
    float t[4];
    for(int row=0; row<4; ++row)
    {
        t[0] = INNER_PRODUCT( *this, other, row, 0 );
        t[1] = INNER_PRODUCT( *this, other, row, 1 );
        t[2] = INNER_PRODUCT( *this, other, row, 2 );
        t[3] = INNER_PRODUCT( *this, other, row, 3 );
        SET_ROW(row, t[0], t[1], t[2], t[3] )
    }
}

#undef INNER_PRODUCT


template <class T>
inline T SGL_ABS(T a)
{
   return (a >= 0 ? a : -a);
}

#ifndef SGL_SWAP
#define SGL_SWAP(a,b,temp) ((temp)=(a),(a)=(b),(b)=(temp))
#endif

//=============================================================================
/// @brief Invert the matrix.
/// @param mat - the matrix to invert.
/// @return true if the matrix was inverted, false if it is non-invertible.
//=============================================================================
bool dMatrix::invert( const dMatrix& mat )
{
    if (&mat==this) {
       dMatrix tm(mat);
       return invert(tm);
    }

    unsigned int indxc[4], indxr[4], ipiv[4];
    unsigned int i,j,k,l,ll;
    unsigned int icol = 0;
    unsigned int irow = 0;
    float temp, pivinv, dum, big;

    // copy in place this may be unnecessary
    *this = mat;

    for (j=0; j<4; j++) ipiv[j]=0;

    for(i=0;i<4;i++)
    {
       big=(float)0.0;
       for (j=0; j<4; j++)
          if (ipiv[j] != 1)
             for (k=0; k<4; k++)
             {
                if (ipiv[k] == 0)
                {
                   if (SGL_ABS(operator()(j,k)) >= big)
                   {
                      big = SGL_ABS(operator()(j,k));
                      irow=j;
                      icol=k;
                   }
                }
                else if (ipiv[k] > 1)
                   return false;
             }
       ++(ipiv[icol]);
       if (irow != icol)
          for (l=0; l<4; l++) SGL_SWAP(operator()(irow,l),
                                       operator()(icol,l),
                                       temp);

       indxr[i]=irow;
       indxc[i]=icol;
       if (operator()(icol,icol) == 0)
          return false;

       pivinv = 1.0/operator()(icol,icol);
       operator()(icol,icol) = 1;
       for (l=0; l<4; l++) operator()(icol,l) *= pivinv;
       for (ll=0; ll<4; ll++)
          if (ll != icol)
          {
             dum=operator()(ll,icol);
             operator()(ll,icol) = 0;
             for (l=0; l<4; l++) operator()(ll,l) -= operator()(icol,l)*dum;
          }
    }
    for (int lx=4; lx>0; --lx)
    {
       if (indxr[lx-1] != indxc[lx-1])
          for (k=0; k<4; k++) SGL_SWAP(operator()(k,indxr[lx-1]),
                                       operator()(k,indxc[lx-1]),temp);
    }

    return true;
}

//=============================================================================
/// @brief Set to an orthographic projection. See glOrtho for further details.
/// @param left - the left clipping plane.
/// @param right - the right clipping plane.
/// @param bottom - the bottom clipping plane.
/// @param top - the top clipping plane.
/// @param zNear - the near clipping plane.
/// @param zFar - the far clipping plane.
//=============================================================================
void dMatrix::makeOrtho(double left, double right,
                       double bottom, double top,
                       double zNear, double zFar)
{
    // note transpose of dMatrix wr.t OpenGL documentation, since the OSG use post multiplication rather than pre.
    double tx = -(right+left)/(right-left);
    double ty = -(top+bottom)/(top-bottom);
    double tz = -(zFar+zNear)/(zFar-zNear);
    SET_ROW(0, 2.0f/(right-left),              0.0f,               0.0f, 0.0f )
    SET_ROW(1,              0.0f, 2.0f/(top-bottom),               0.0f, 0.0f )
    SET_ROW(2,              0.0f,              0.0f, -2.0f/(zFar-zNear), 0.0f )
    SET_ROW(3,                tx,                ty,                 tz, 1.0f )
}

//=============================================================================
/// @brief Set to a frustum projection. See glFrustum for further details.
/// @param left - the left clipping plane.
/// @param right - the right clipping plane.
/// @param bottom - the bottom clipping plane.
/// @param top - the top clipping plane.
/// @param zNear - the near clipping plane.
/// @param zFar - the far clipping plane.
//=============================================================================
void dMatrix::makeFrustum(double left, double right,
                         double bottom, double top,
                         double zNear, double zFar)
{
    // note transpose of dMatrix wr.t OpenGL documentation, since the OSG use post multiplication rather than pre.
    double A = (right+left)/(right-left);
    double B = (top+bottom)/(top-bottom);
    double C = -(zFar+zNear)/(zFar-zNear);
    double D = -2.0*zFar*zNear/(zFar-zNear);
    SET_ROW(0, 2.0f*zNear/(right-left),                    0.0f, 0.0f,  0.0f )
    SET_ROW(1,                    0.0f, 2.0f*zNear/(top-bottom), 0.0f,  0.0f )
    SET_ROW(2,                       A,                       B,    C, -1.0f )
    SET_ROW(3,                    0.0f,                   0.0f,     D,  0.0f )
}


//=============================================================================
/// @brief Set to a symmetrical perspective projection, See gluPerspective for further details.
/// @param fovy - the angle of the field of view between the top and bottom of the window/display.
/// @param aspectRatio - the aspect ratio between the width and height of the window/display.
/// @param zNear - the near clipping plane.
/// @param zFar - the far clipping plane.
//=============================================================================
void dMatrix::makePerspective(double fovy,double aspectRatio,
                             double zNear, double zFar)
{
    // calculate the appropriate left, right etc.
    double tan_fovy = tan(DegreesToRadians(fovy*0.5));
    double right  =  tan_fovy * aspectRatio * zNear;
    double left   = -right;
    double top    =  tan_fovy * zNear;
    double bottom =  -top;
    makeFrustum(left,right,bottom,top,zNear,zFar);
}

//=============================================================================
/// @brief Set to the lookat or camera matrix.
/// @param eye - the location of the eye.
/// @param center - a point where the eye is looking.
/// @param up - the up vector.
//=============================================================================
void dMatrix::makeLookAt(const dVec3& eye,const dVec3& center,const dVec3& up)
{
    dVec3 f(center-eye);
    f.normalize();
    dVec3 s(f^up);
    s.normalize();
    dVec3 u(s^f);
    u.normalize();

    set(
        s[0],     u[0],     -f[0],     0.0f,
        s[1],     u[1],     -f[1],     0.0f,
        s[2],     u[2],     -f[2],     0.0f,
        0.0f,     0.0f,     0.0f,      1.0f);

    preMult(dMatrix::translate(-eye));
}

//=============================================================================
/// @brief Get the quaternion representing the current matrix.
/// @return a dQuat representing the current matrix.
//=============================================================================
dQuat dMatrix::getQuat()
{
	float x,y,z,w;

  float trace = _mat[0][0] + _mat[1][1] + _mat[2][2] + 1.f;
  
  if(trace > .001) 
    {
      float s = 0.5f/sqrt(trace);
      w = 0.25f/s;
      x = (_mat[2][1] - _mat[1][2])*s;
      y = (_mat[0][2] - _mat[2][0])*s;
      z = (_mat[1][0] - _mat[0][1])*s;
    } 
  else 
    {
      if (_mat[0][0]>_mat[1][1] && _mat[0][0]>_mat[2][2]) 
        {
          float s = 2.f*sqrt(1.f + _mat[0][0] - _mat[1][1] - _mat[2][2]);
          x = 0.25f*s;
          y = (_mat[0][1] + _mat[1][0])/s;
          z = (_mat[0][2] + _mat[2][0])/s;
          w = (_mat[2][1] - _mat[1][2])/s;
        } 
      else if (_mat[1][1] > _mat[2][2]) 
        {
          float s = 2.f*sqrt(1.f + _mat[1][1] - _mat[0][0] - _mat[2][2]);
          x = (_mat[0][1] + _mat[1][0])/s;
          y = 0.25f*s;
          z = (_mat[1][2] + _mat[2][1])/s;
          w = (_mat[0][2] - _mat[2][0])/s;
        } 
      else 
        {
          float s = 2.f*sqrtf(1.0f + _mat[2][2] - _mat[0][0] - _mat[1][1]);
          x = (_mat[0][2] + _mat[2][0])/s;
          y = (_mat[1][2] + _mat[2][1])/s;
          z = 0.25f*s;
          w = (_mat[1][0] - _mat[0][1])/s;
        }
    }

  float len = sqrt(((x)*(x) + (y)*(y) + (z)*(z) + (w)*(w)));

  (x) /= len;
  (y) /= len;
  (z) /= len;
  (w) /= len;
  return dQuat(x,y,z,w);
}

//=============================================================================
/// @brief Get a vector containing the rotation about each axes in degrees.
/// @return a vector containing the rotation about each axes in degrees.
//=============================================================================
dVec3 dMatrix::getHPR()
{
    dMatrix mat( *this );
    mat( 3, 0 ) = 0.0f;
    mat( 3, 1 ) = 0.0f;
    mat( 3, 2 ) = 0.0f;
    dVec3 hpr;
    dVec3 diry = mat * dVec3( 0.0f, 1.0f, 0.0f ), dirz = mat * dVec3( 0.0f, 0.0f, 1.0f );
    diry.normalize();
    dirz.normalize();
    if ( ABS( diry.x() ) < 1.0e-6 && ABS( diry.y() ) < 1.0e-6 )
    {
        hpr.h() = 0.0f;
        hpr.p() = 90.0f * diry.z();
    }
    else
    {
        hpr.h() = RAD2DEG( atan2( diry.x(), diry.y() ) );
        float length = sqrt( diry.x() * diry.x() + diry.y() * diry.y() );
        if ( length < 1.0e-6 )
            hpr.p() = RAD2DEG( atan2( diry.z(), length ) );
        else
            hpr.p() = 90.0f * diry.z();
    }
    dVec3 temp = diry ^ dVec3( 0.0f, 0.0f, 1.0f );//x'' = y' X z ( cross product )
    temp.normalize();
    dVec3 temp2 = temp ^ diry;//z'' = x'' X y' ( cross product )
    temp.normalize();
    float dotprod = temp2 * dirz;//dot product
    if ( dotprod > 1.0f )
        dotprod = 1.0f;
    else if ( dotprod < -1.0f )
        dotprod = -1.0f;

    hpr.r() = RAD2DEG( acos( dotprod ) );

    return hpr;
}

#undef SET_ROW
