/* -*-c++-*- OpenSceneGraph - Copyright (C) 1998-2003 Robert Osfield
 *
 * This library is open source and may be redistributed and/or modified under  
 * the terms of the OpenSceneGraph Public License (DIVERSEPL) version 0.0 or 
 * (at your option) any later version.  The full license is in LICENSE file
 * included with this distribution, and on the openscenegraph.org website.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * OpenSceneGraph Public License for more details.
*/

#ifndef DIVERSE_MATRIX
#define DIVERSE_MATRIX 1

#include <dmath/dVec3.h>
#include <dmath/dVec4.h>

#include <string.h>
#include <dmath/osdep.h>

#include <iostream>
#include <algorithm>

#define DEG2RAD(x)  ((float) ((x)*M_PI/180.0f))
#define RAD2DEG(x)  ((float) ((x)*180.0f/M_PI))
#define ABS(x)   (((x) > 0.0f)?(x):(-(x)))

#ifdef WIN32
#ifdef DMATH_EXPORTS
#define DMATH_API __declspec(dllexport)
#else
#define DMATH_API __declspec(dllimport)
#endif /* DMATH_EXPORTS */
#else
#define DMATH_API
#endif /* WIN32 */

class dQuat;

//=============================================================================
/// @class dMatrix
/// @brief A general purpose 4 x 4 matrix class.
//=============================================================================
class DMATH_API dMatrix
{

    public:

        dMatrix();
        dMatrix( const dMatrix& other);
        explicit dMatrix( float const * const def );
        dMatrix( float a00, float a01, float a02, float a03,
            float a10, float a11, float a12, float a13,
            float a20, float a21, float a22, float a23,
            float a30, float a31, float a32, float a33);

		/// @brief Destructor.
        ~dMatrix() {}

		/// @brief Compare this matrix and the specified matrix.
		/// @param m - the matrix to be compared with this matrix.
		/// @return 0 if the matrices are identical, else a non-zero value.
        int compare(const dMatrix& m) const { return memcmp(_mat,m._mat,sizeof(_mat)); }

		/// @brief Less than relational operator.
        bool operator < (const dMatrix& m) const { return compare(m)<0; }
		/// @brief Equal relational operator.
        bool operator == (const dMatrix& m) const { return compare(m)==0; }
		/// @brief Not equal relational operator.
        bool operator != (const dMatrix& m) const { return compare(m)!=0; }

		/// @brief Get a reference to the value at specified row and column.
		/// @param row - the index of the row.
		/// @param col - the index of the column.
		/// @return a reference to the value at the specified  row and column.
        inline float& operator()(int row, int col) { return _mat[row][col]; }
		/// @brief Get the value at specified row and column.
		/// @param row - the index of the row.
		/// @param col - the index of the column.
		/// @return the value at the specified  row and column.
        inline float operator()(int row, int col) const { return _mat[row][col]; }

		/// @brief Determine if the dMatrix is valid.
		/// @return true if matrix is valid, else returns false.
        inline bool valid() const { return !isNaN(); }
		/// @brief Returns true if any of the float values forming the matrix
		/// are invalid, else returns false.
        inline bool isNaN() const { return ::isNaN(_mat[0][0]) || ::isNaN(_mat[0][1]) || ::isNaN(_mat[0][2]) || ::isNaN(_mat[0][3]) ||
                                           ::isNaN(_mat[1][0]) || ::isNaN(_mat[1][1]) || ::isNaN(_mat[1][2]) || ::isNaN(_mat[1][3]) ||
                                           ::isNaN(_mat[2][0]) || ::isNaN(_mat[2][1]) || ::isNaN(_mat[2][2]) || ::isNaN(_mat[2][3]) ||
                                           ::isNaN(_mat[3][0]) || ::isNaN(_mat[3][1]) || ::isNaN(_mat[3][2]) || ::isNaN(_mat[3][3]); }

		/// @brief Equal relational operator.
        inline dMatrix& operator = (const dMatrix& other)
        {
            if( &other == this ) return *this;
            std::copy((float*)other._mat,(float*)other._mat+16,(float*)(_mat));
            return *this;
        }

        /// @brief Set this matrix to the value of another matrix.
		/// @param other - the matrix whose values are to be copied to this one.
        inline void set(const dMatrix& other)
        {
            std::copy((float*)other._mat,(float*)other._mat+16,(float*)(_mat));
        }

        /// @brief Set this matrix to the values from a pointer to an array of 16 floats.
		/// @param ptr - a float pointer to 16 contiguous floats.
        inline void set(float const * const ptr)
        {
            std::copy(ptr,ptr+16,(float*)(_mat));
        }
        
        void set( float a00, float a01, float a02, float a03,
                  float a10, float a11, float a12, float a13,
                  float a20, float a21, float a22, float a23,
                  float a30, float a31, float a32, float a33);

		/// @brief Get a float pointer to the values of the matrix.
		/// @return a float pointer to the values of the matrix
        float * ptr() { return (float *)_mat; }
		/// @brief Get a const float pointer to the values of the matrix.
		/// @return a const float pointer to the values of the matrix
        float * ptr() const { return (float *)_mat; }

        void makeIdentity();
        
        void makeScale( const dVec3& );
        void makeScale( float, float, float );
        
        void makeTranslate( const dVec3& );
        void makeTranslate( float, float, float );
        
        //=====================================================================
        // Function Name: makeEuler
        /// @brief Rotates heading around y(up), pitch around x(right), and roll
        /// around z(back)
        /// <pre>
        ///                               +y              
        ///                                |
        ///                                |___ +x
        ///                               /
        ///                             +z
        /// </pre>
        /// @param v - the head/pitch/roll vector in radians
        //=====================================================================
        void makeEulerYXZ( const dVec3& v )
        {
            makeRotate( v.h(), dVec3( 0, 1, 0 ),
                        v.p(), dVec3( 1, 0, 0 ),
                        v.r(), dVec3( 0, 0, 1 ) );
        }
        //=====================================================================
        // Function Name: makeEuler
        /// @brief Rotates heading around z(up), pitch around x(right), and roll
        /// around y(front)
        /// <pre>
        ///                               +y              
        ///                                |
        ///                                |___ +x
        ///                               /
        ///                             +z
        /// </pre>
        /// @param v - the head/pitch/roll vector in radians
        //=====================================================================
        void makeEulerZXY( const dVec3& v )
        {
            makeRotate( v.h(), dVec3( 0, 0, 1 ),
                        v.p(), dVec3( 1, 0, 0 ),
                        v.r(), dVec3( 0, 1, 0 ) );
        }
        
        void makeRotate( const dVec3& from, const dVec3& to );
        void makeRotate( float angle, const dVec3& axis );
        void makeRotate( float angle, float x, float y, float z );
        void makeRotate( const dQuat& );
        void makeRotate( float angle1, const dVec3& axis1, 
                         float angle2, const dVec3& axis2,
                         float angle3, const dVec3& axis3);

        
        
        void makeOrtho(double left, double right,
                       double bottom, double top,
                       double zNear, double zFar);

        /// @brief Set to a 2D orthographic projection. See glOrtho2D for further details.
        inline void makeOrtho2D(double left, double right,
                                double bottom, double top)
        {
            makeOrtho(left,right,bottom,top,-1.0,1.0);
        }

        void makeFrustum(double left, double right,
                         double bottom, double top,
                         double zNear, double zFar);

        void makePerspective(double fovy,double aspectRatio,
                             double zNear, double zFar);

        /** Set to the position and orientation as per a camera, using the same convention as gluLookAt. */
        void makeLookAt(const dVec3& eye,const dVec3& center,const dVec3& up);
        

        bool invert( const dMatrix& );

        //basic utility functions to create new matrices
        inline static dMatrix identity( void );
        inline static dMatrix scale( const dVec3& sv);
        inline static dMatrix scale( float sx, float sy, float sz);
        inline static dMatrix translate( const dVec3& dv);
        inline static dMatrix translate( float x, float y, float z);
        inline static dMatrix rotate( const dVec3& from, const dVec3& to);
        inline static dMatrix rotate( float angle, float x, float y, float z);
        inline static dMatrix rotate( float angle, const dVec3& axis);
        inline static dMatrix rotate( float angle1, const dVec3& axis1, 
                                     float angle2, const dVec3& axis2,
                                     float angle3, const dVec3& axis3);
        inline static dMatrix rotate( const dQuat& quat);
        inline static dMatrix inverse( const dMatrix& matrix);
        
        /** Create a orthographic projection. See glOrtho for further details.*/
        inline static dMatrix ortho(double left, double right,
                                   double bottom, double top,
                                   double zNear, double zFar);

        /** Create a 2D orthographic projection. See glOrtho for further details.*/
        inline static dMatrix ortho2D(double left, double right,
                                     double bottom, double top);

        /** Create a perspective projection. See glFrustum for further details.*/
        inline static dMatrix frustum(double left, double right,
                                     double bottom, double top,
                                     double zNear, double zFar);

        /** Create a symmetrical perspective projection, See gluPerspective for further details.
          * Aspect ratio is defined as width/height.*/
        inline static dMatrix perspective(double fovy,double aspectRatio,
                                         double zNear, double zFar);

        /** Create the position and orientation as per a camera, using the same convention as gluLookAt. */
        inline static dMatrix lookAt(const dVec3& eye,const dVec3& center,const dVec3& up);




        inline dVec3 preMult( const dVec3& v ) const;
        inline dVec3 postMult( const dVec3& v ) const;
        inline dVec3 operator* ( const dVec3& v ) const;
        inline dVec4 preMult( const dVec4& v ) const;
        inline dVec4 postMult( const dVec4& v ) const;
        inline dVec4 operator* ( const dVec4& v ) const;

        void setTrans( float tx, float ty, float tz );
        void setTrans( const dVec3& v );
		/// @brief Get the translation values.
		/// @return a vector containing the translation values.
        inline dVec3 getTrans() const { return dVec3(_mat[3][0],_mat[3][1],_mat[3][2]); } 
        
		/// @brief Get the scale values.
		/// @return a vector containing the scale values.
        inline dVec3 getScale() const { return dVec3(_mat[0][0],_mat[1][1],_mat[2][2]); }
        
        /** apply apply an 3x3 transform of v*M[0..2,0..2]  */
        inline static dVec3 transform3x3(const dVec3& v,const dMatrix& m);
        /** apply apply an 3x3 transform of M[0..2,0..2]*v  */
        inline static dVec3 transform3x3(const dMatrix& m,const dVec3& v);


        // basic dMatrix multiplication, our workhorse methods.
        void mult( const dMatrix&, const dMatrix& );
        void preMult( const dMatrix& );
        void postMult( const dMatrix& );

		//Quaternion accessor functions
		dQuat getQuat();

		/// @brief Unary multiply opertator.
        inline void operator *= ( const dMatrix& other ) 
        {    if( this == &other ) {
                dMatrix temp(other);
                postMult( temp );
            }
            else postMult( other ); 
        }

		/// @brief Binary multiply operator.
        inline dMatrix operator * ( const dMatrix &m ) const
		{
			dMatrix r;
				r.mult(*this,m);
			return  r;
		}

		dVec3 getHPR();

// temporarily commented out while waiting for a more generic implementation
// of dMatrixProduct proxy class.
//         // Helper class to optimize product expressions somewhat
//         class dMatrixProduct {
//         public:
//             const dMatrix& A;
//             const dMatrix& B;
// 
//             dMatrixProduct( const dMatrix& lhs, const dMatrix& rhs ) : A(lhs), B(rhs) {}
//         };
// 
//         inline dMatrixProduct operator * ( const dMatrix& other ) const    
//             {    return dMatrixProduct(*this, other); }
// 
//         inline void operator = ( const dMatrixProduct& p ) 
//             {
//                 if( this == &(p.A)) postMult(p.B);
//                 else if( this == &(p.B)) preMult(p.A);
//                 else mult( p.A, p.B );
//             }
// 
//         dMatrix( const dMatrixProduct& p ) //allows implicit evaluation of the product
//             {    mult( p.A, p.B ); }

    protected:
        float _mat[4][4];

};

//static utility methods
/// @brief Get an identity matrix.
/// @return an identity matrix.
inline dMatrix dMatrix::identity(void)
{
    dMatrix m;
    m.makeIdentity();
    return m;
}

/// @brief Get a matrix with the specified scale factors.
/// @param sx - the x scale by which to multiply the matrix.
/// @param sy - the y scale by which to multiply the matrix.
/// @param sz - the z scale by which to multiply the matrix.
/// @return a matrix with the specified scale factors.
inline dMatrix dMatrix::scale(float sx, float sy, float sz)
{
    dMatrix m;
    m.makeScale(sx,sy,sz);
    return m;
}

/// @brief Get a matrix with the specified scale factors.
/// @param v - a vector containing the scale value for each axis.
/// @return a matrix with the specified scale factors.
inline dMatrix dMatrix::scale(const dVec3& v )
{
    return scale(v.x(), v.y(), v.z() );
}

/// @brief Get a matrix with the specified translation.
/// @param tx - the x coordinate.
/// @param ty - the y coordinate.
/// @param tz - the z coordinate.
/// @return a matrix with the specified translation.
inline dMatrix dMatrix::translate(float tx, float ty, float tz)
{
    dMatrix m;
    m.makeTranslate(tx,ty,tz);
    return m;
}

/// @brief Get a matrix with the specified translation.
/// @param v - a vector containing the translation coordinates.
/// @return a matrix with the specified translation.
inline dMatrix dMatrix::translate(const dVec3& v )
{
    return translate(v.x(), v.y(), v.z() );
}

/// @brief Get a matrix with the specified quaternion.
/// @param q - a quaternion.
/// @return a matrix with the rotation specifued by the quaternion.
inline dMatrix dMatrix::rotate( const dQuat& q )
{
    dMatrix m;
    m.makeRotate( q );
    return m;
}

/// @brief Get a matrix with the specified rotation.
/// @param angle - the angle of rotation in radians.
/// @param x - the x value of the axis of rotation.
/// @param y - the y value of the axis of rotation.
/// @param z - the z value of the axis of rotation.
/// @return a matrix with the specified rotation.
inline dMatrix dMatrix::rotate(float angle, float x, float y, float z )
{
    dMatrix m;
    m.makeRotate(angle,x,y,z);
    return m;
}

/// @brief Get a matrix with the specified rotation.
/// @param angle - the angle of rotation in radians.
/// @param axis - a vector containing the axis of rotation.
/// @return a matrix with the specified rotation.
inline dMatrix dMatrix::rotate(float angle, const dVec3& axis )
{
    dMatrix m;
    m.makeRotate(angle,axis);
    return m;
}

/// @brief Get a matrix with the specified rotation.
/// @param angle1 - the angle of rotation about axis1 in radians.
/// @param axis1 - a vector containing the axis of rotation.
/// @param angle2 - the angle of rotation about axis2 in radians.
/// @param axis2 - a vector containing the axis of rotation.
/// @param angle3 - the angle of rotation about axis3 in radians.
/// @param axis3 - a vector containing the axis of rotation.
/// @return a matrix with the specified rotation.
inline dMatrix dMatrix::rotate( float angle1, const dVec3& axis1, 
                              float angle2, const dVec3& axis2,
                              float angle3, const dVec3& axis3)
{
    dMatrix m;
    m.makeRotate(angle1,axis1,angle2,axis2,angle3,axis3);
    return m;
}

/// @brief Get a matrix with the specified rotation.
/// @param from - the start vector.
/// @param to - the end vector.
/// @return a matrix with the specified rotation.
inline dMatrix dMatrix::rotate(const dVec3& from, const dVec3& to )
{
    dMatrix m;
    m.makeRotate(from,to);
    return m;
}

/// @brief Get the inverse of the matrix.
/// @param matrix - the matrix to invert.
/// @return the inverted matrix.
inline dMatrix dMatrix::inverse( const dMatrix& matrix)
{
    dMatrix m;
    m.invert(matrix);
    return m;
}

/// @brief Get the orthographic projection matrix.
/// @param left - the left clipping plane.
/// @param right - the right clipping plane.
/// @param bottom - the bottom clipping plane.
/// @param top - the top clipping plane.
/// @param zNear - the near clipping plane.
/// @param zFar - the far clipping plane.
/// @return the orthographic projection matrix.
inline dMatrix dMatrix::ortho(double left, double right,
                            double bottom, double top,
                            double zNear, double zFar)
{
    dMatrix m;
    m.makeOrtho(left,right,bottom,top,zNear,zFar);
    return m;
}

/// @brief Get the 2D orthographic projection matrix.
/// @param left - the left clipping plane.
/// @param right - the right clipping plane.
/// @param bottom - the bottom clipping plane.
/// @param top - the top clipping plane.
/// @return the 2D orthographic projection matrix.
inline dMatrix dMatrix::ortho2D(double left, double right,
                              double bottom, double top)
{
    dMatrix m;
    m.makeOrtho2D(left,right,bottom,top);
    return m;
}

/// @brief Get the frustum projection matrix.
/// @param left - the left clipping plane.
/// @param right - the right clipping plane.
/// @param bottom - the bottom clipping plane.
/// @param top - the top clipping plane.
/// @param zNear - the near clipping plane.
/// @param zFar - the far clipping plane.
/// @return the frustum projection matrix.
inline dMatrix dMatrix::frustum(double left, double right,
                              double bottom, double top,
                              double zNear, double zFar)
{
    dMatrix m;
    m.makeFrustum(left,right,bottom,top,zNear,zFar);
    return m;
}

/// @brief Get the perspective projection matrix.
/// @param fovy - the angle of the field of view between the top and bottom of the window/display.
/// @param aspectRatio - the aspect ratio between the width and height of the window/display.
/// @param zNear - the near clipping plane.
/// @param zFar - the far clipping plane.
/// @return the perspective projection matrix.
inline dMatrix dMatrix::perspective(double fovy,double aspectRatio,
                                  double zNear, double zFar)
{
    dMatrix m;
    m.makePerspective(fovy,aspectRatio,zNear,zFar);
    return m;
}

/// @brief Get the lookat or camera matrix.
/// @param eye - the location of the eye.
/// @param center - a point where the eye is looking.
/// @param up - the up vector.
/// @return the lookat or camera matrix.
inline dMatrix dMatrix::lookAt(const dVec3& eye,const dVec3& center,const dVec3& up)
{
    dMatrix m;
    m.makeLookAt(eye,center,up);
    return m;
}

/// @brief Post-multiply a matrix with a dVec3.
/// @param v - the vector to post-multiply.
/// @return the resulting dVec3 vector.
inline dVec3 dMatrix::postMult( const dVec3& v ) const
{
    float d = 1.0f/(_mat[3][0]*v.x()+_mat[3][1]*v.y()+_mat[3][2]*v.z()+_mat[3][3]) ;
    return dVec3( (_mat[0][0]*v.x() + _mat[0][1]*v.y() + _mat[0][2]*v.z() + _mat[0][3])*d,
        (_mat[1][0]*v.x() + _mat[1][1]*v.y() + _mat[1][2]*v.z() + _mat[1][3])*d,
        (_mat[2][0]*v.x() + _mat[2][1]*v.y() + _mat[2][2]*v.z() + _mat[2][3])*d) ;
}

/// @brief Pre-multiply a matrix with a dVec3.
/// @param v - the vector to pre-multiply.
/// @return the resulting dVec3 vector.
inline dVec3 dMatrix::preMult( const dVec3& v ) const
{
    float d = 1.0f/(_mat[0][3]*v.x()+_mat[1][3]*v.y()+_mat[2][3]*v.z()+_mat[3][3]) ;
    return dVec3( (_mat[0][0]*v.x() + _mat[1][0]*v.y() + _mat[2][0]*v.z() + _mat[3][0])*d,
        (_mat[0][1]*v.x() + _mat[1][1]*v.y() + _mat[2][1]*v.z() + _mat[3][1])*d,
        (_mat[0][2]*v.x() + _mat[1][2]*v.y() + _mat[2][2]*v.z() + _mat[3][2])*d);
}

/// @brief Post-multiply a matrix with a dVec3.
/// @param v - the vector to post-multiply.
/// @return the resulting dVec3 vector.
inline dVec4 dMatrix::postMult( const dVec4& v ) const
{
    return dVec4( (_mat[0][0]*v.x() + _mat[0][1]*v.y() + _mat[0][2]*v.z() + _mat[0][3]*v.w()),
        (_mat[1][0]*v.x() + _mat[1][1]*v.y() + _mat[1][2]*v.z() + _mat[1][3]*v.w()),
        (_mat[2][0]*v.x() + _mat[2][1]*v.y() + _mat[2][2]*v.z() + _mat[2][3]*v.w()),
        (_mat[3][0]*v.x() + _mat[3][1]*v.y() + _mat[3][2]*v.z() + _mat[3][3]*v.w())) ;
}

/// @brief Pre-multiply a matrix with a dVec3.
/// @param v - the vector to pre-multiply.
/// @return the resulting dVec3 vector.
inline dVec4 dMatrix::preMult( const dVec4& v ) const
{
    return dVec4( (_mat[0][0]*v.x() + _mat[1][0]*v.y() + _mat[2][0]*v.z() + _mat[3][0]*v.w()),
        (_mat[0][1]*v.x() + _mat[1][1]*v.y() + _mat[2][1]*v.z() + _mat[3][1]*v.w()),
        (_mat[0][2]*v.x() + _mat[1][2]*v.y() + _mat[2][2]*v.z() + _mat[3][2]*v.w()),
        (_mat[0][3]*v.x() + _mat[1][3]*v.y() + _mat[2][3]*v.z() + _mat[3][3]*v.w()));
}

/// @brief Multiply a vector by the 3x3 sub matrix.
/// @param v - the vector to be transformed.
/// @param m - the dMatrix to be used as the transformation matrix.
/// @return the resulting dVec3 vector.
inline dVec3 dMatrix::transform3x3(const dVec3& v,const dMatrix& m)
{
    return dVec3( (m._mat[0][0]*v.x() + m._mat[1][0]*v.y() + m._mat[2][0]*v.z()),
                 (m._mat[0][1]*v.x() + m._mat[1][1]*v.y() + m._mat[2][1]*v.z()),
                 (m._mat[0][2]*v.x() + m._mat[1][2]*v.y() + m._mat[2][2]*v.z()));
}

/// @brief Multiply the 3x3 sub matrix by a vector.
/// @param m - the dMatrix to be used as the transformation matrix.
/// @param v - the vector to be transformed.
/// @return the resulting dVec3 vector.
inline dVec3 dMatrix::transform3x3(const dMatrix& m,const dVec3& v)
{
    return dVec3( (m._mat[0][0]*v.x() + m._mat[0][1]*v.y() + m._mat[0][2]*v.z()),
                 (m._mat[1][0]*v.x() + m._mat[1][1]*v.y() + m._mat[1][2]*v.z()),
                 (m._mat[2][0]*v.x() + m._mat[2][1]*v.y() + m._mat[2][2]*v.z()) ) ;
}

/// @brief Multiply operator to multiply a dVec3 by a dMatrix.
inline dVec3 operator* (const dVec3& v, const dMatrix& m )
{
	return m.preMult(v);
}

/// @brief Multiply operator to multiply a dVec4 by a dMatrix.
inline dVec4 operator* (const dVec4& v, const dMatrix& m )
{
	return m.preMult(v);
}

/// @brief Post-multiply operator to multiply a dVec3 by a dMatrix.
inline dVec3 dMatrix::operator* (const dVec3& v) const
{
	return postMult(v);
}

/// @brief Pre-multiply operator to multiply a dVec4 by a dMatrix.
inline dVec4 dMatrix::operator* (const dVec4& v) const
{
	return postMult(v);
}

/// @brief Print out the vector using standard ostream output ( i.e. cout ).
inline std::ostream& operator<< (std::ostream& os, const dMatrix& m )
{
    os << "{"<<std::endl;
    for(int row=0; row<4; ++row) {
        os << "\t";
        for(int col=0; col<4; ++col)
            os << m(row,col) << " ";
        os << std::endl;
    }
    os << "}" << std::endl;
    return os;
}

#endif
