/* -*-c++-*- OpenSceneGraph - Copyright (C) 1998-2003 Robert Osfield 
 *
 * This library is open source and may be redistributed and/or modified under  
 * the terms of the OpenSceneGraph Public License (OSGPL) version 0.0 or 
 * (at your option) any later version.  The full license is in LICENSE file
 * included with this distribution, and on the openscenegraph.org website.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * OpenSceneGraph Public License for more details.
*/

#ifndef OSG_QUAT
#define OSG_QUAT 1

#include <dmath/dVec3.h>
#include <dmath/dVec4.h>
#include <dmath/dMatrix.h>

//=============================================================================
/// @class dQuat
/// @brief A quaternion class.
///
/// It can be used to represent an orientation in 3D space. The class uses dVec4
/// to store the quaternion. They are always ordered x, y, z, and w such that indices
/// 0 - 3 reference the same data.
//=============================================================================
class DMATH_API dQuat
{

    public:

     /* ----------------------------------------------------------
				DATA MEMBERS
        The only data member is a
	dVec4 which holds the elements

	In other words, dmath:dQuat is composed of an dmath::dVec4
	The dmath::dQuat aggregates an dmath::dVec4

	These seem to be different jargon for the same thing :-)
	---------------------------------------------------------- */

		dVec4 _fv;	// a four-vector

		/// @brief Default constructor.
		inline dQuat(): _fv(0.0f,0.0f,0.0f,1.0f) {}
		/// @brief Constructor.
		/// @param x - the x value of the quaternion.
		/// @param y - the y value of the quaternion.
		/// @param z - the z value of the quaternion.
		/// @param w - the w value of the quaternion.
		inline dQuat( float x, float y, float z, float w ):  _fv(x,y,z,w) {}
		/// @brief Constructor.
		/// @param v - a vector containing the parameters for the quaternion.
		inline dQuat( const dVec4& v ): _fv(v) {}
		/// @brief Constructor.
		/// @param angle - the angle of rotation about axis in radians.
		/// @param axis - a vector containing the axis of rotation.
        inline dQuat( float angle, const dVec3& axis)
        {
            makeRotate(angle,axis);
        }
		/// @brief Constructor.
		/// @param angle1 - the angle of rotation about axis1 in radians.
		/// @param axis1 - a vector containing the axis of rotation.
		/// @param angle2 - the angle of rotation about axis2 in radians.
		/// @param axis2 - a vector containing the axis of rotation.
		/// @param angle3 - the angle of rotation about axis3 in radians.
		/// @param axis3 - a vector containing the axis of rotation.
        inline dQuat( float angle1, const dVec3& axis1, 
                     float angle2, const dVec3& axis2,
                     float angle3, const dVec3& axis3)
        {
            makeRotate(angle1,axis1,angle2,axis2,angle3,axis3);
        }

     /* ----------------------------------
	   Methods to access data members
	---------------------------------- */
		/// @brief Get the quaternion as a dVec4.
		/// @return the quaternion as a dVec4.
		inline dVec4& asdVec4()
		{
			return _fv;
		}

		/// @brief Get the quaternion as a const dVec4.
		/// @return the quaternion as a const dVec4.
		inline const dVec4& asdVec4() const
		{
			return _fv;
		}

		/// @brief Get the quaternion as a dVec3.
		/// @return the quaternion as a dVec3.
		inline const dVec3 asdVec3() const
		{
			return dVec3(_fv[0], _fv[1], _fv[2]);
		}

		/// @brief Set the values of the quaternion.
		/// @param x - the x value.
		/// @param y - the y value.
		/// @param z - the z value.
		/// @param w - the w value.
        inline void set(float x, float y, float z, float w)
        {
            _fv.set(x,y,z,w);
        }

		/// @brief Set the values of the quaternion.
		/// @param v - a dVec4 containing the values for the quaternion..
        inline void set(const dVec4& v)
        {
            _fv = v;
        }

		/// @brief Get a reference to the value at the specified index.
        inline float& operator [] (int i) { return _fv[i]; }
		/// @brief Get the value at the specified index.
        inline float  operator [] (int i) const { return _fv[i]; }

		/// @brief Get a reference to the value at the specified index.
        inline float& x() { return _fv[0]; }
		/// @brief Get a reference to the value at the specified index.
        inline float& y() { return _fv[1]; }
		/// @brief Get a reference to the value at the specified index.
        inline float& z() { return _fv[2]; }
		/// @brief Get a reference to the value at the specified index.
        inline float& w() { return _fv[3]; }

		/// @brief Get the value at the specified index.
        inline float x() const { return _fv[0]; }
		/// @brief Get the value at the specified index.
        inline float y() const { return _fv[1]; }
		/// @brief Get the value at the specified index.
        inline float z() const { return _fv[2]; }
		/// @brief Get the value at the specified index.
        inline float w() const { return _fv[3]; }

    	/// @brief Determine if the dQuat represents a zero rotation.
		/// @return true if the dQuat represents a zero rotation, and therefore can be ignored in computations.
    	bool zeroRotation() const { return _fv[0]==0.0f && _fv[1]==0.0f && _fv[2]==0.0f && _fv[3]==1.0f; } 


     /* ------------------------------------------------------------- 
	 		  BASIC ARITHMETIC METHODS		    
	Implemented in terms of dVec4s.  Some dVec4 operators, e.g.
	operator* are not appropriate for quaternions (as
	mathematical objects) so they are implemented differently.
	Also define methods for conjugate and the multiplicative inverse.		    
	------------------------------------------------------------- */
		/// @brief Multiply by scalar .
        inline const dQuat operator * (float rhs) const
        {
            return dQuat(_fv*rhs);
        }

		/// @brief Unary multiply by scalar.
        inline dQuat& operator *= (float rhs)
        {
            _fv*=rhs;
            return *this;		// enable nesting
        }

		/// @brief Binary multiply.
		inline const dQuat operator*(const dQuat& rhs) const
		{
			return dQuat( rhs._fv[3]*_fv[0] + rhs._fv[0]*_fv[3] + rhs._fv[1]*_fv[2] - rhs._fv[2]*_fv[1],
				 rhs._fv[3]*_fv[1] - rhs._fv[0]*_fv[2] + rhs._fv[1]*_fv[3] + rhs._fv[2]*_fv[0],
				 rhs._fv[3]*_fv[2] + rhs._fv[0]*_fv[1] - rhs._fv[1]*_fv[0] + rhs._fv[2]*_fv[3],
				 rhs._fv[3]*_fv[3] - rhs._fv[0]*_fv[0] - rhs._fv[1]*_fv[1] - rhs._fv[2]*_fv[2] );
		}

		/// @brief Unary multiply.
		inline dQuat& operator*=(const dQuat& rhs)
		{
			float x = rhs._fv[3]*_fv[0] + rhs._fv[0]*_fv[3] + rhs._fv[1]*_fv[2] - rhs._fv[2]*_fv[1];
			float y = rhs._fv[3]*_fv[1] - rhs._fv[0]*_fv[2] + rhs._fv[1]*_fv[3] + rhs._fv[2]*_fv[0];
			float z = rhs._fv[3]*_fv[2] + rhs._fv[0]*_fv[1] - rhs._fv[1]*_fv[0] + rhs._fv[2]*_fv[3];
			_fv[3]   = rhs._fv[3]*_fv[3] - rhs._fv[0]*_fv[0] - rhs._fv[1]*_fv[1] - rhs._fv[2]*_fv[2];

			_fv[2] = z;
			_fv[1] = y;
			_fv[0] = x;

			return (*this);			// enable nesting
		}

		/// @brief Divide by scalar.
        inline const dQuat operator / (float rhs) const
        {
            return dQuat(_fv/rhs);
        }

		/// @brief Unary divide by scalar.
        inline dQuat& operator /= (float rhs)
        {
            _fv/=rhs;
            return *this;
        }

		/// @brief Binary divide.
		inline const dQuat operator/(const dQuat& denom) const
		{
			return ( (*this) * denom.inverse() );
		}

		/// @brief Unary divide.
		inline dQuat& operator/=(const dQuat& denom)
		{
			(*this) = (*this) * denom.inverse();
			return (*this);			// enable nesting
		}

		/// @brief Binary addition.
        inline const dQuat operator + (const dQuat& rhs) const
        {
            return dQuat( _fv + rhs._fv );
        }

		/// @brief Unary addition.
        inline dQuat& operator += (const dQuat& rhs)
        {
            _fv += rhs._fv;
            return *this;			// enable nesting
        }

        /// @brief Binary subtraction.
        inline const dQuat operator - (const dQuat& rhs) const
        {
            return dQuat( _fv - rhs._fv );
        }

        /// @brief Unary subtraction.
        inline dQuat& operator -= (const dQuat& rhs)
        {
            _fv-=rhs._fv;
            return *this;			// enable nesting
        }

        /// @brief Negation operator - returns the negative of the quaternion.
	    /// Basically just calls operator - () on the dVec4 */
        inline const dQuat operator - () const
        {
            return dQuat ( -_fv );
        }

        /// @brief Length of the quaternion = sqrt( vec . vec )
        float length() const
        {
            return _fv.length();
        }

        /// @brief Length of the quaternion = vec . vec
        float length2() const
        {
            return _fv.length2();
        }

        /// @brief Conjugate of the quaternion.
        inline dQuat conj () const
        { 
             return dQuat( -_fv[0], -_fv[1], -_fv[2], _fv[3] );
        }

        /// @brief Multiplicative inverse method: q^(-1) = q^*/(q.q^*)
        inline const dQuat inverse () const
        {
             return conj() / length2();
         }

      /* -------------------------------------------------------- 
	 		  METHODS RELATED TO ROTATIONS
	Set a quaternion which will perform a rotation of an
	angle around the axis given by the vector (x,y,z).
	Should be written to also accept an angle and a dVec3?

	Define Spherical Linear interpolation method also

	Not inlined - see the dQuat.cpp file for implementation
	-------------------------------------------------------- */
        void makeRotate( float angle, 
                          float x, float y, float z );
        void makeRotate ( float angle, const dVec3& vec );

        void makeRotate ( float angle1, const dVec3& axis1, 
                          float angle2, const dVec3& axis2,
                          float angle3, const dVec3& axis3);

        /** Make a rotation dQuat which will rotate vec1 to vec2.
            Generally take adot product to get the angle between these
            and then use a cross product to get the rotation axis
            Watch out for the two special cases of when the vectors
            are co-incident or opposite in direction.*/
        void makeRotate( const dVec3& vec1, const dVec3& vec2 );
    
        /** Return the angle and vector components represented by the quaternion.*/
        void getRotate ( float& angle, float& x, float& y, float& z ) const;
        /** Return the angle and vector represented by the quaternion.*/
        void getRotate ( float& angle, dVec3& vec ) const;

        /** Spherical Linear Interpolation.
	    As t goes from 0 to 1, the dQuat object goes from "from" to "to". */
        void slerp   ( float t, const dQuat& from, const dQuat& to);
        
        /** Set quaternion to be equivalent to specified matrix.*/
        void set( const dMatrix& m );

        /** Get the equivalent matrix for this quaternion.*/
        void get( dMatrix& m ) const;

        /** Get the equivalent matrix for this quaternion.*/
        dMatrix getMatrix() const
        {
            dMatrix matrix;
            get(matrix);
            return matrix;
        }

		/// @brief Print out the vector using standard ostream output ( i.e. cout ).
        friend inline std::ostream& operator << (std::ostream& output, const dQuat& vec);

};	// end of class prototype

inline std::ostream& operator << (std::ostream& output, const dQuat& vec)
{
    output << vec._fv[0] << " "
	   << vec._fv[1] << " "
	   << vec._fv[2] << " "
	   << vec._fv[3];
    return output; 	// to enable cascading
}

#endif 
