/* -*-c++-*- OpenSceneGraph - Copyright (C) 1998-2003 Robert Osfield 
 *
 * This library is open source and may be redistributed and/or modified under  
 * the terms of the OpenSceneGraph Public License (DIVERSEPL) version 0.0 or 
 * (at your option) any later version.  The full license is in LICENSE file
 * included with this distribution, and on the openscenegraph.org website.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * OpenSceneGraph Public License for more details.
*/

#ifndef DIVERSE_VEC3
#define DIVERSE_VEC3 1

#include <dmath/dmath.h>
#include <dmath/osdep.h>

#include <iostream>
#include <string.h>


/*  General purpose float triple for use as vertices, vectors and normals.
    Provides general maths operations from addition through to cross products.
    No support yet added for float * dVec3 - is it necessary?
    Need to define a non-member non-friend operator*  etc.
   			     dVec3 * float is okay
*/
//=============================================================================
/// @class dVec3
/// @brief General purpose float triple for use as vertices, vectors and normals.
///
/// Provides general maths operations from addition through to cross products.
//=============================================================================
class dVec3
{
    public:

		/// @brief Default constructor.
        dVec3() { _v[0]=0.0f; _v[1]=0.0f; _v[2]=0.0f;}
		/// @brief Constructor. Takes 3 floats as parameters.
        dVec3(float x,float y,float z) { _v[0]=x; _v[1]=y; _v[2]=z; }

        /// @brief Returns the number of bytes needed to pack the dVec3.
        int packsize()const{ return 3 * sizeof(float); }
        /// @brief Packs the vec into buffer .
        int pack  ( char* buffer )const{ memcpy( buffer, _v, packsize() ); return packsize(); }
        /// @brief Sets the value of vec based on buffer.
        int unpack(char* buffer ){ memcpy( _v, buffer, packsize() ); return packsize(); }

        float _v[3];

		/// @brief Equal relational operator.
		inline bool operator == (const dVec3& v) const { return _v[0]==v._v[0] && _v[1]==v._v[1] && _v[2]==v._v[2]; }
        /// @brief Not equal raltional operator.
        inline bool operator != (const dVec3& v) const { return _v[0]!=v._v[0] || _v[1]!=v._v[1] || _v[2]!=v._v[2]; }
		/// @brief Less than relational operator.
		inline bool operator <  (const dVec3& v) const
        {
            if (_v[0]<v._v[0]) return true;
            else if (_v[0]>v._v[0]) return false;
            else if (_v[1]<v._v[1]) return true;
            else if (_v[1]>v._v[1]) return false;
            else return (_v[2]<v._v[2]);
        }

		/// @brief Get a pointer to the array of 3 floats of the vector.
        inline float* ptr() { return _v; }
		/// @brief Get a const pointer to the array of 3 floats of the vector.
        inline const float* ptr() const { return _v; }

		/// @brief Set the values of the vector
        inline void set( float x, float y, float z)
        {
            _v[0]=x; _v[1]=y; _v[2]=z;
        }

		/// @brief Get a reference to the value at the specified index.
        inline float& operator [] (int i) { return _v[i]; }
		/// @brief Get the value at the specified index.
        inline float operator [] (int i) const { return _v[i]; }

		/// @brief Get a reference to the value at the specified index.
        inline float& x() { return _v[0]; }
		/// @brief Get a reference to the value at the specified index.
        inline float& y() { return _v[1]; }
		/// @brief Get a reference to the value at the specified index.
        inline float& z() { return _v[2]; }

		/// @brief Get the value at the specified index.
        inline float x() const { return _v[0]; }
		/// @brief Get the value at the specified index.
        inline float y() const { return _v[1]; }
		/// @brief Get the value at the specified index.
        inline float z() const { return _v[2]; }
		/// @brief Get a reference to the value at the specified index.
        inline float& h() { return _v[0]; }
		/// @brief Get a reference to the value at the specified index.
        inline float& p() { return _v[1]; }
		/// @brief Get a reference to the value at the specified index.
        inline float& r() { return _v[2]; }

		/// @brief Get the value at the specified index.
        inline float h() const { return _v[0]; }
		/// @brief Get the value at the specified index.
        inline float p() const { return _v[1]; }
		/// @brief Get the value at the specified index.
        inline float r() const { return _v[2]; }

		/// @brief Determine if the dVec4 is valid.
        inline bool valid() const { return !isNaN(); }
		/// @brief Returns true if any of the float values forming the vector
		/// are invalid, else returns false.
        inline bool isNaN() const { return ::isNaN(_v[0]) || ::isNaN(_v[1]) || ::isNaN(_v[2]); }

        /// @brief The dot product.
        inline float operator * (const dVec3& rhs) const
        {
            return _v[0]*rhs._v[0]+_v[1]*rhs._v[1]+_v[2]*rhs._v[2];
        }

        /// @brief The cross product.
        inline const dVec3 operator ^ (const dVec3& rhs) const
        {
            return dVec3(_v[1]*rhs._v[2]-_v[2]*rhs._v[1],
                _v[2]*rhs._v[0]-_v[0]*rhs._v[2] ,
                _v[0]*rhs._v[1]-_v[1]*rhs._v[0]);
        }

        /// @brief Multiply by scalar.
        inline const dVec3 operator * (float rhs) const
        {
            return dVec3(_v[0]*rhs, _v[1]*rhs, _v[2]*rhs);
        }

        /// @brief Unary multiply by scalar.
        inline dVec3& operator *= (float rhs)
        {
            _v[0]*=rhs;
            _v[1]*=rhs;
            _v[2]*=rhs;
            return *this;
        }

        /// @brief Divide by scalar.
        inline const dVec3 operator / (float rhs) const
        {
            return dVec3(_v[0]/rhs, _v[1]/rhs, _v[2]/rhs);
        }

        /// @brief Unary divide by scalar.
        inline dVec3& operator /= (float rhs)
        {
            _v[0]/=rhs;
            _v[1]/=rhs;
            _v[2]/=rhs;
            return *this;
        }

        /// @brief Binary vector add.
        inline const dVec3 operator + (const dVec3& rhs) const
        {
            return dVec3(_v[0]+rhs._v[0], _v[1]+rhs._v[1], _v[2]+rhs._v[2]);
        }

        /// @brief Unary vector add.
		///
		/// Slightly more efficient because no temporary intermediate object.
        inline dVec3& operator += (const dVec3& rhs)
        {
            _v[0] += rhs._v[0];
            _v[1] += rhs._v[1];
            _v[2] += rhs._v[2];
            return *this;
        }

        /// @brief Binary vector subtract.
        inline const dVec3 operator - (const dVec3& rhs) const
        {
            return dVec3(_v[0]-rhs._v[0], _v[1]-rhs._v[1], _v[2]-rhs._v[2]);
        }

        /// @brief Unary vector subtract.
        inline dVec3& operator -= (const dVec3& rhs)
        {
            _v[0]-=rhs._v[0];
            _v[1]-=rhs._v[1];
            _v[2]-=rhs._v[2];
            return *this;
        }

        /// @brief Negation operator.  Returns the negative of the dVec3.
        inline const dVec3 operator - () const
        {
            return dVec3 (-_v[0], -_v[1], -_v[2]);
        }

        /// @brief Length of the vector = sqrt( vec . vec )
        inline float length() const
        {
            return sqrtf( _v[0]*_v[0] + _v[1]*_v[1] + _v[2]*_v[2] );
        }

        /// @brief Length squared of the vector = vec . vec
        inline float length2() const
        {
            return _v[0]*_v[0] + _v[1]*_v[1] + _v[2]*_v[2];
        }

        /// @brief Normalize the vector so that it has length unity
        /// returns the previous length of the vector.
        inline float normalize()
        {
            float norm = dVec3::length();
            if (norm>0.0f)
            {
                _v[0] /= norm;
                _v[1] /= norm;
                _v[2] /= norm;
            }                
            return( norm );
        }

		/// @brief Print out the vector using standard ostream output ( i.e. cout ).
		friend inline std::ostream& operator << (std::ostream& output, const dVec3& vec);

};	// end of class dVec3

inline std::ostream& operator << (std::ostream& output, const dVec3& vec)
{
    output << vec._v[0] << " "
           << vec._v[1] << " "
           << vec._v[2];
    return output; 	// to enable cascading
}

const dVec3 X_AXIS(1.0f,0.0f,0.0f);
const dVec3 Y_AXIS(0.0f,1.0f,0.0f);
const dVec3 Z_AXIS(0.0f,0.0f,1.0f);


#endif
