/* -*-c++-*- OpenSceneGraph - Copyright (C) 1998-2003 Robert Osfield 
 *
 * This library is open source and may be redistributed and/or modified under  
 * the terms of the OpenSceneGraph Public License (DIVERSEPL) version 0.0 or 
 * (at your option) any later version.  The full license is in LICENSE file
 * included with this distribution, and on the openscenegraph.org website.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * OpenSceneGraph Public License for more details.
*/

#ifndef DIVERSE_VEC4
#define DIVERSE_VEC4 1

#include <dmath/dVec3.h>

#include <iostream>

/*  General purpose float quad, uses include representation
    of colour coordinates.
    No support yet added for float * dVec4 - is it necessary?
    Need to define a non-member non-friend operator*  etc.
   			     dVec4 * float is okay
*/
//=============================================================================
/// @class dVec4
/// @brief General purpose float quad, uses include representation
/// of colour coordinates.
//=============================================================================
class dVec4
{
    public:

	// Methods are defined here so that they are implicitly inlined

		/// @brief Default constructor.
        dVec4() { _v[0]=0.0f; _v[1]=0.0f; _v[2]=0.0f; _v[3]=0.0f;}
        
		/// @brief Constructor. Takes 4 floats as parameters.
        dVec4(float x, float y, float z, float w)
        {
            _v[0]=x;
            _v[1]=y;
            _v[2]=z;
            _v[3]=w;
        }

		/// @brief Constructor. Takes a vector and a float as parameters.
        dVec4(const dVec3& v3,float w)
        {
            _v[0]=v3[0];
            _v[1]=v3[1];
            _v[2]=v3[2];
            _v[3]=w;
        }
            

        float _v[4];

		/// @brief Equal relational operator.
        inline bool operator == (const dVec4& v) const { return _v[0]==v._v[0] && _v[1]==v._v[1] && _v[2]==v._v[2] && _v[3]==v._v[3]; }

		/// @brief Not equal relational operator.
        inline bool operator != (const dVec4& v) const { return _v[0]!=v._v[0] || _v[1]!=v._v[1] || _v[2]!=v._v[2] || _v[3]!=v._v[3]; }

		/// @brief Less than relational operator.
        inline bool operator <  (const dVec4& v) const
        {
            if (_v[0]<v._v[0]) return true;
            else if (_v[0]>v._v[0]) return false;
            else if (_v[1]<v._v[1]) return true;
            else if (_v[1]>v._v[1]) return false;
            else if (_v[2]<v._v[2]) return true;
            else if (_v[2]>v._v[2]) return false;
            else return (_v[3]<v._v[3]);
        }

		/// @brief Get a pointer to the array of 4 floats of the vector.
        inline float* ptr() { return _v; }
		/// @brief Get a const pointer to the array of 4 floats of the vector.
        inline const float* ptr() const { return _v; }

		/// @brief Set the values of the vector
        inline void set( float x, float y, float z, float w)
        {
            _v[0]=x; _v[1]=y; _v[2]=z; _v[3]=w;
        }

		/// @brief Get a reference to the value at the specified index.
        inline float& operator [] (unsigned int i) { return _v[i]; }
		/// @brief Get the value at the specified index.
        inline float  operator [] (unsigned int i) const { return _v[i]; }

		/// @brief Get a reference to the value at the specified index.
        inline float& x() { return _v[0]; }
		/// @brief Get a reference to the value at the specified index.
        inline float& y() { return _v[1]; }
		/// @brief Get a reference to the value at the specified index.
        inline float& z() { return _v[2]; }
		/// @brief Get a reference to the value at the specified index.
        inline float& w() { return _v[3]; }

		/// @brief Get the value at the specified index.
        inline float x() const { return _v[0]; }
		/// @brief Get the value at the specified index.
        inline float y() const { return _v[1]; }
		/// @brief Get the value at the specified index.
        inline float z() const { return _v[2]; }
		/// @brief Get the value at the specified index.
        inline float w() const { return _v[3]; }

		/// @brief Get a long value representing the color in ABGR format.
        inline unsigned long asABGR() const
        {
            return (unsigned long)clampTo((_v[0]*255.0f),0.0f,255.0f)<<24 |
                   (unsigned long)clampTo((_v[1]*255.0f),0.0f,255.0f)<<16 |
                   (unsigned long)clampTo((_v[2]*255.0f),0.0f,255.0f)<<8  |
                   (unsigned long)clampTo((_v[3]*255.0f),0.0f,255.0f);
        }

		/// @brief Get a long value representing the color in RGBA format.
        inline unsigned long asRGBA() const
        {
            return (unsigned long)clampTo((_v[3]*255.0f),0.0f,255.0f)<<24 |
                   (unsigned long)clampTo((_v[2]*255.0f),0.0f,255.0f)<<16 |
                   (unsigned long)clampTo((_v[1]*255.0f),0.0f,255.0f)<<8  |
                   (unsigned long)clampTo((_v[0]*255.0f),0.0f,255.0f);
        }

		/// @brief Determine if the dVec4 is valid.
        inline bool valid() const { return !isNaN(); }
		/// @brief Returns true if any of the float values forming the vector
		/// are invalid, else returns false.
        inline bool isNaN() const { return ::isNaN(_v[0]) || ::isNaN(_v[1]) || ::isNaN(_v[2]) || ::isNaN(_v[3]); }

        /// @brief The dot product.
        inline float operator * (const dVec4& rhs) const
        {
            return _v[0]*rhs._v[0]+
	           _v[1]*rhs._v[1]+
	           _v[2]*rhs._v[2]+
	           _v[3]*rhs._v[3] ;
        }

        /// @brief Multiply by scalar.
        inline dVec4 operator * (float rhs) const
        {
            return dVec4(_v[0]*rhs, _v[1]*rhs, _v[2]*rhs, _v[3]*rhs);
        }

        /// @brief Unary multiply by scalar.
        inline dVec4& operator *= (float rhs)
        {
            _v[0]*=rhs;
            _v[1]*=rhs;
            _v[2]*=rhs;
            _v[3]*=rhs;
            return *this;
        }

        /// @brief Divide by scalar.
        inline dVec4 operator / (float rhs) const
        {
            return dVec4(_v[0]/rhs, _v[1]/rhs, _v[2]/rhs, _v[3]/rhs);
        }

        /// @brief Unary divide by scalar.
        inline dVec4& operator /= (float rhs)
        {
            _v[0]/=rhs;
            _v[1]/=rhs;
            _v[2]/=rhs;
            _v[3]/=rhs;
            return *this;
        }

        /// @brief Binary vector add.
        inline dVec4 operator + (const dVec4& rhs) const
        {
            return dVec4(_v[0]+rhs._v[0], _v[1]+rhs._v[1],
		        _v[2]+rhs._v[2], _v[3]+rhs._v[3]);
        }

        /// @brief Unary vector add.
		///
		/// Slightly more efficient because no temporary intermediate object.
        inline dVec4& operator += (const dVec4& rhs)
        {
            _v[0] += rhs._v[0];
            _v[1] += rhs._v[1];
            _v[2] += rhs._v[2];
            _v[3] += rhs._v[3];
            return *this;
        }

        /// @brief Binary vector subtract.
        inline dVec4 operator - (const dVec4& rhs) const
        {
            return dVec4(_v[0]-rhs._v[0], _v[1]-rhs._v[1],
		        _v[2]-rhs._v[2], _v[3]-rhs._v[3] );
        }

        /// @brief Unary vector subtract.
        inline dVec4& operator -= (const dVec4& rhs)
        {
            _v[0]-=rhs._v[0];
            _v[1]-=rhs._v[1];
            _v[2]-=rhs._v[2];
            _v[3]-=rhs._v[3];
            return *this;
        }

        /// @brief Negation operator.  Returns the negative of the dVec4.
        inline const dVec4 operator - () const
        {
            return dVec4 (-_v[0], -_v[1], -_v[2], -_v[3]);
        }

        /// @brief Length of the vector = sqrt( vec . vec )
        inline float length() const
        {
            return sqrtf( _v[0]*_v[0] + _v[1]*_v[1] + _v[2]*_v[2] + _v[3]*_v[3]);
        }

        /// @brief Length squared of the vector = vec . vec
        inline float length2() const
        {
            return _v[0]*_v[0] + _v[1]*_v[1] + _v[2]*_v[2] + _v[3]*_v[3];
        }

        /// @brief Normalize the vector so that it has length unity
        /// returns the previous length of the vector.
        inline float normalize()
        {
            float norm = dVec4::length();
            _v[0] /= norm;
            _v[1] /= norm;
            _v[2] /= norm;
            _v[3] /= norm;
            return( norm );
        }

		/// @brief Print out the vector using standard ostream output ( i.e. cout ).
        friend inline std::ostream& operator << (std::ostream& output, const dVec4& vec)
        {
	    output << vec._v[0] << " "
                   << vec._v[1] << " "
                   << vec._v[2] << " "
                   << vec._v[3];
	    return output; 	// to enable cascading
	}

};	// end of class dVec4


/// @brief Compute the dot product of a (dVec3,1.0) and a dVec4.
inline float operator * (const dVec3& lhs,const dVec4& rhs)
{
    return lhs[0]*rhs[0]+lhs[1]*rhs[1]+lhs[2]*rhs[2]+rhs[3];
}

/// @brief Compute the dot product of a dVec4 and a (dVec3,1.0).
inline float operator * (const dVec4& lhs,const dVec3& rhs)
{
    return lhs[0]*rhs[0]+lhs[1]*rhs[1]+lhs[2]*rhs[2]+lhs[3];
}

#endif
