#include"Entry.h"
#include<tracer/Tracer.h>
#include<dxml/Token.h>
#include<iostream>
using namespace std;

//=============================================================================
/// @brief Constructor.
/// @param key - the registry key to use for this entry
/// @param val - the value of the key
/// @param cmt - the comment of the entry
//=============================================================================
Entry::Entry( const string& key, const string& val, const string& cmt )
{
    m_key = key;
    m_val = val;
    m_comment = cmt;
}

//=============================================================================
/// @brief Retrieves the value for the specified path or uses the value passed
/// if the path and value does not exist.
/// @param path - the registry node from which to get the value.
/// @param defval - the default value if the path doesn't exist in the registry.
/// @param cmt - the comment to use if the path doesn't exist in the registry.
/// @return a string containing the value, or "ERROR" if the node could not be found..
//=============================================================================
string Entry::getValue( const string& path, const string& defval,
                        const string& cmt ) 
{
//    Tracer::debug("getValue( %s, %s)", path.c_str(), defval.c_str() );
    if( path.size() == 0 )
    {
        if( m_val == "__UNDEFINED__" )
            m_val = defval;
        if( m_comment == "__UNDEFINED__" )
            m_comment = cmt;
        return m_val;
    }
    string key;
    unsigned int i;
    for( i = 0; i < path.size(); i++ )
    {   
        if( path[i] != '/' )
            key += path[i];
        else
        {
            i++;
            break;
        }
    }
    map<string, Entry*>::iterator itr = m_subentries.find(key);
    if( itr == m_subentries.end() )
    {
        if( i == path.size() )
        {
            Entry* ent = new Entry( key, defval, cmt );
            m_subentries[key] = ent;
            return defval;
        }
        else
        {
           Entry* ent = new Entry( key );
           m_subentries[key] = ent;
           return ent->getValue( path.substr( i, path.size() ), defval, cmt );
        }
    }
    else
    {
        return itr->second->getValue( path.substr( i, path.size() ), defval, cmt );
    }
    
    return "ERROR";
}

//=============================================================================
/// @brief Retrieves the comment for the specified path.
/// @param path - the registry node from which to get the comment.
/// @param defval - the default comment if the path doesn't exist in the registry.
/// @return a string containing the comment, or "ERROR" if the node could not be found..
//=============================================================================
string Entry::getComment( const string& path, const string& defval )
{
//    Tracer::debug("getComment( %s, %s)", path.c_str(), defval.c_str() );
    if( path.size() == 0 )
    {
        if( m_val == "__UNDEFINED__" )
            m_val = defval;
        return m_val;
    }
    string key;
    unsigned int i;
    for( i = 0; i < path.size(); i++ )
    {   
        if( path[i] != '/' )
            key += path[i];
        else
        {
            i++;
            break;
        }
    }
    map<string, Entry*>::iterator itr = m_subentries.find(key);
    if( itr == m_subentries.end() )
    {
        if( i == path.size() )
        {
            Entry* ent = new Entry( key, defval );
            m_subentries[key] = ent;
            return defval;
        }
        else
        {
           Entry* ent = new Entry( key );
           m_subentries[key] = ent;
           return ent->getComment( path.substr( i, path.size() ), defval );
        }
    }
    else
    {
        return itr->second->getComment( path.substr( i, path.size() ), defval );
    }
    
    return "ERROR";
}

//=============================================================================
/// @brief Retrieves the Entry existing at the specified registry path.
/// @param path - The path of the entry to return .
/// @return the entry at the specified path, or NULL if the Entry could not be found.
//=============================================================================
Entry* Entry::getEntry( const string& path )
{
    //Tracer::debug( "getEntry( %s )", path.c_str() );
    if( path.size() == 0 )
        return this;
    
    string key;
    unsigned int i;
    for( i = 0; i < path.size(); i++ )
    {   
        if( path[i] != '/' )
            key += path[i];
        else
        {
            i++;
            break;
        }
    }
    map<string, Entry*>::iterator itr = m_subentries.find(key);
    if( itr == m_subentries.end() )
        return NULL;
    else
    {
        return itr->second->getEntry( path.substr( i, path.size() ) );
    }
    
    return NULL;
}

//=============================================================================
/// @brief Sets the value for the specified path.
/// @param path - the registry node for which to set the value.
/// @param defval - the value to be set.
//=============================================================================
void Entry::setValue( const string& path, const string& defval )
{
//    Tracer::Debug("SetValue( %s, %s)", path.c_str(), defval.c_str() );
    if( path.size() == 0 )
    {
        m_val = defval;
        return;
    }
    
    string key;
    unsigned int i;
    for( i = 0; i < path.size(); i++ )
    {   
        if( path[i] != '/' )
            key += path[i];
        else
        {
            i++;
            break;
        }
    }
    map<string, Entry*>::iterator itr = m_subentries.find(key);
    if( itr == m_subentries.end() )
    {
        if( i == path.size() )
        {
            Entry* ent = new Entry( key, defval );
            m_subentries[key] = ent;
            return;
        }
        else
        {
           Entry* ent = new Entry( key );
           m_subentries[key] = ent;
           ent->setValue( path.substr( i, path.size() ), defval );
           return;
        }
    }
    else
    {
        itr->second->setValue( path.substr( i, path.size() ), defval );
        return;
    }
}

//=============================================================================
/// @brief Sets the comment for the specified path.
/// @param path - the registry node for which to set the comment.
/// @param defval - the comment to be set.
//=============================================================================
void Entry::setComment( const string& path, const string& defval )
{
//    Tracer::Debug("SetComment( %s, %s)", path.c_str(), defval.c_str() );
    if( path.size() == 0 )
    {
        m_val = defval;
        return;
    }
    
    string key;
    unsigned int i;
    for( i = 0; i < path.size(); i++ )
    {   
        if( path[i] != '/' )
            key += path[i];
        else
        {
            i++;
            break;
        }
    }
    map<string, Entry*>::iterator itr = m_subentries.find(key);
    if( itr == m_subentries.end() )
    {
        if( i == path.size() )
        {
            Entry* ent = new Entry( key, defval );
            m_subentries[key] = ent;
            return;
        }
        else
        {
           Entry* ent = new Entry( key );
           m_subentries[key] = ent;
           ent->setComment( path.substr( i, path.size() ), defval );
           return;
        }
    }
    else
    {
        itr->second->setComment( path.substr( i, path.size() ), defval );
        return;
    }
}

//=============================================================================
/// @brief Saves the registry to the given output file stream
/// @param out - The output file stream to save the registry to
//=============================================================================
void Entry::save( ofstream& out )
{
    m_Save( out, 0 );
}

//=============================================================================
/// @brief The private save function.
/// @param out - the output file stream to save.
/// @param depth - the depth of the tree to output.
//=============================================================================
void Entry::m_Save( ofstream& out, int depth )
{
	int i;
    for( i = 0; i < depth; i++ )
        out<<"  ";
    out<<"<"<<m_key;
    if( m_val != "__UNDEFINED__" )
        out<<" val=\""<<m_val<<"\"";
    if( m_comment != "__UNDEFINED__" )
        out<<" comment=\""<<m_comment<<"\"";
    if( m_subentries.size() == 0 )
    {
        out<<"/>\n";
        return;
    }
    out<<">\n";
    map<string, Entry*>::iterator itr = m_subentries.begin();
    while( itr != m_subentries.end() )
    {
        itr->second->m_Save( out, depth+1 );
        itr++;
    }
    for( i = 0; i < depth; i++ )
        out<<"  ";
    out<<"</"<<m_key<<">\n";
}

//=============================================================================
/// @brief Read an entry from a lexer.
/// @param lex - the RegLexer from which to read.
/// @return true if successful, otherwise false.
//=============================================================================
bool Entry::read( RegLexer& lex )
{
    lex.pushFlushToken( Token( RegLexer::LEX_OTAG, "<" ) );
    lex.flush();
    Token* tok = lex.peak();
   
    //================================================
    // If flush didn't return a < then
    // we hit some other flush token and should return
    //================================================
    if( tok->type != RegLexer::LEX_OTAG )
    {
        lex.popFlushToken( Token( RegLexer::LEX_OTAG, "<") );
        return false;
    }
    // Pop the "<"
    delete lex.pop();

    // We have now entered a Entry item, flush set is now
    // val, ">", "</"
    lex.pushFlushToken( Token( RegLexer::LEX_KEYWORD, "val" ) );
    lex.pushFlushToken( Token( RegLexer::LEX_KEYWORD, "comment" ) );
    lex.pushFlushToken( Token( RegLexer::LEX_CTAG, ">" ) );
    lex.pushFlushToken( Token( RegLexer::LEX_CATAG, "/>" ) );
    lex.pushFlushToken( Token( RegLexer::LEX_OCTAG, "</" ) );

    // Check for the LEX_ID
    tok = lex.peak();
    if( lex.peak()->type != RegLexer::LEX_ID )
    {
        Tracer::error( "ID expected, but found %s", tok->source.c_str() );
        m_key = lex.peak()->source;
    }
    m_key = tok->source;
    delete lex.pop(); // ID

    // Flush until we find val
    lex.flush();
    lex.popFlushToken( Token( RegLexer::LEX_KEYWORD, "val" ) );
    if( lex.peak()->lexem != "val" )
    {
        // val wasn't found, warn the user and continue
//        Tracer::Warn("No val attribute for tag %s", m_key.c_str() );
        m_val = "__UNDEFINED__";
    }
    else // val was found
    {
        lex.pop(); // pop val
        tok = lex.peak();
        if( tok->type != RegLexer::LEX_EQ )
        {
            Tracer::error( "Expected \"=\" after val" );
        }
        else 
        {
            lex.pop(); // "="
            tok = lex.peak();
            if( tok->type != RegLexer::LEX_STR )
            {
                Tracer::error( "String Expected, but found %s", tok->source.c_str() );
            }
            else
            {
                m_val = tok->source;      
                lex.pop(); // the value of val
            }
        }
    }
    // Flush until we find comment 
    lex.flush();
    lex.popFlushToken( Token( RegLexer::LEX_KEYWORD, "comment" ) );
    if( lex.peak()->lexem != "comment" )
    {
        // val wasn't found, warn the user and continue
//        Tracer::Warn("No val attribute for tag %s", m_key.c_str() );
        m_comment = "__UNDEFINED__";
    }
    else // comment was found
    {
        lex.pop(); // pop comment 
        tok = lex.peak();
        if( tok->type != RegLexer::LEX_EQ )
        {
            Tracer::error( "Expected \"=\" after comment" );
        }
        else 
        {
            lex.pop(); // "="
            tok = lex.peak();
            if( tok->type != RegLexer::LEX_STR )
            {
                Tracer::error( "String Expected, but found %s", tok->source.c_str() );
            }
            else
            {
                m_comment = tok->source;      
                lex.pop(); // the value of val
            }
        }
    }
    // look for the close of the tag
    lex.flush();
    if( lex.peak()->type == RegLexer::LEX_CTAG )
    {
        lex.popFlushToken( Token( RegLexer::LEX_CTAG, ">" ) ); // we found the close
        lex.pop(); // pop >

        // Load all the sub entries possible
        bool subent = true;
        while( subent )
        {
            Entry* ent = new Entry("", "");
            subent = ent->read( lex );
            if( subent )
            {
                map<string, Entry*>::iterator itr = m_subentries.find(ent->key() );
                if( itr == m_subentries.end() )
                    m_subentries[ent->key()] = ent;
                else
                {
                    Tracer::warn( "Duplicate Key: %s - Ignoring", ent->key().c_str() );
                    delete ent;
                }
            }
        }
    }
    else if( lex.peak()->type == RegLexer::LEX_CATAG )
    {
        lex.pop(); // />
        lex.popFlushToken( Token( RegLexer::LEX_CATAG, "/>" ) ); // found />
        lex.popFlushToken( Token( RegLexer::LEX_CTAG, ">" ) );// not looking for >
        lex.popFlushToken( Token( RegLexer::LEX_OCTAG, "</" ) ); // not looking for</ 
        return true;
    }
    // flush until </
    lex.flush();
    if( lex.peak()->type != RegLexer::LEX_OCTAG )
    {
        Tracer::error( "Expected </, but found %s", lex.peak()->source.c_str() );
        return false;
    }
    lex.popFlushToken( Token( RegLexer::LEX_OCTAG, "</" ) ); // found </ 
    lex.pushFlushToken( Token( RegLexer::LEX_CTAG, ">" ) ); // we are looking for >
    lex.pop(); // pop </
    if( lex.peak()->source != m_key )
    {
        Tracer::error( "Wrong close tag found: %s - expected %s",
                        lex.peak()->source.c_str(), m_key.c_str() );
        lex.popFlushToken( Token( RegLexer::LEX_CTAG, ">" ) );
        return false;
    }
    lex.pop(); // the key
    lex.flush(); // flush until >
    if( lex.peak()->type != RegLexer::LEX_CTAG )
    {
        lex.popFlushToken( Token(RegLexer::LEX_CTAG, ">") ); // we didn't find >, return false
        return false;
    }
    lex.pop(); // >
    return true;
}

//=============================================================================
/// @brief Remove a registry node from an entry .
/// @param path - the registry node to be removed.
//=============================================================================
void Entry::remove( const string& path )
{
    Tracer::debug("Entry::remove( %s )", path.c_str() );
    map<string, Entry*>::iterator itr = m_subentries.find( path );
    if( itr == m_subentries.end() )
    {
        string key;
        unsigned int i;
        for( i = 0; i < path.size(); i++ )
        {   
            if( path[i] != '/' )
                key += path[i];
            else
            {
                i++;
                break;
            }
        }
        itr = m_subentries.find(key);
        if( itr == m_subentries.end() )
        {
            Tracer::warn( "Entry::remove( %s ) - no such entry",
                           (path.c_str()+i) );
            return;
        }
        else
        {
            if( path.size() > i )
                itr->second->remove( path.c_str()+i );
            return;
        }

    }
    Tracer::debug("\tRemoved %s ", itr->first.c_str() );
    delete itr->second;
    m_subentries.erase(itr);
}
