#include"RegLexer.h"
#include<map>
#include<vector>
#include<algorithm>
#include<iostream>
#include<fstream>
#include<ctype.h>
#include<dxml/Token.h>
#include<dxml/fsm.h>
#include<tracer/Tracer.h>

using namespace std;

#ifdef WIN32
#define DSEPARATOR ';'
#else
#define DSEPARATOR ':'
#endif

//=============================================================================
/// @brief Constructor.
/// @param fname - the filename to read.
//=============================================================================
RegLexer::RegLexer( const string& fname )
    :m_file(fname.c_str()), m_type_str( 13 )
{
    m_fname = fname;
    m_cur_line = 0; 
    m_start = LEX_INIT;
    m_done  = LEX_DONE;
    m_last  = -1;
    m_printdebug = false;

    m_keywords.push_back( "val" );
    m_keywords.push_back( "comment" );
    
    m_type_str[ LEX_DONE      ] = "DONE";
    m_type_str[ LEX_INIT      ] = "INIT";
    m_type_str[ LEX_ID        ] = "LEX_ID";
    m_type_str[ LEX_STR       ] = "LEX_STR";
    m_type_str[ LEX_EQ        ] = "LEX_EQ";
    m_type_str[ LEX_CTAG      ] = "LEX_CTAG";
    m_type_str[ LEX_OTAG      ] = "LEX_OTAG";
    m_type_str[ LEX_OCTAG     ] = "LEX_OCTAG";
    m_type_str[ LEX_CATAG     ] = "LEX_CATAG";
    m_type_str[ LEX_KEYWORD   ] = "LEX_KEYWORD"; 
    initFSM();

    if( !m_file.good() )
    {
        Tracer::warn( "Config file %s not found.", m_fname.c_str() );
        m_tokens.push_back( new Token( -2, "EOF", m_cur_line, 0 ) );
        m_done = -1;
    }
    readLine();
}


//=============================================================================
/// @brief Destructor. Deletes all tokens.
//=============================================================================
RegLexer::~RegLexer( )
{
    return;
    list<Token*>::iterator itr = m_tokens.begin();
    while( itr != m_tokens.end() )
        delete *itr;
    m_tokens.clear();
    m_file.close();
}


//=============================================================================
/// @brief Reads a line of input and generates tokens.
//=============================================================================
void RegLexer::readLine()
{
    if( m_start == -1 || m_done == -1 ) return;
    getline( m_file, m_line );
    if( m_printdebug ) cout<<m_line<<endl;

    if( m_file.eof() )
    {
        m_tokens.push_back( new Token( -2, "EOF", m_cur_line, 0 ) );
        return;
    }
    m_line += '\n';
    m_cur_line++;

    m_last = m_fsm->getCurrentState();
    if( m_last == -1 ) m_fsm->gotoState( m_start );

    int next_state;
    int action;
    int start_col = 0;  // The starting col of the token
    string lexem;   // the lexem we are building
    
    for( unsigned int i = 0; i < m_line.size(); i++ )
    {
        if( -1 == m_fsm->getNextState( tolower(m_line[i]), &next_state, &action ) )
        {
            cout<<"Line: "<<m_cur_line<<":"<<(start_col+1)<<" - unknown Token: \'"<<m_line[i]<<"\' found\n";
            m_fsm->gotoState( m_start ); // Start Over if error
            lexem = "";
            start_col = i;
            i--;
            continue;
        }
        switch( action )
        {
            case PUSH_BACK:
                i--;
                break;
            case APPEND:
                lexem += m_line[i];
                break;
            case SKIP:
                break;
        }
        if( next_state == m_done )
        {
            if( m_last == LEX_ID )
                if( find( m_keywords.begin(), m_keywords.end(), lexem ) != m_keywords.end() )
                    m_last = LEX_KEYWORD;
            Token* tok = new Token( m_last, lexem, m_cur_line, start_col+1 );
            m_tokens.push_back( tok );
            m_fsm->gotoState( m_start );
            start_col = i+1;
            lexem = "";
            continue;
        }
        m_last = next_state;
        if( m_last == -1 )
        {
            cout<<"Line: "<<m_cur_line<<":"<<(start_col+1)<<" - unknown Token: \'"<<(lexem + m_line[i])<<"\' found\n";
            m_fsm->gotoState( m_start ); // Start Over if error
            lexem = "";
            start_col = i;
        }
    }
}

//=============================================================================
/// @brief Adds a new token to the flush list, when flush is called it will
/// go through the list of tokens.
/// @param tok - the token to add to the list to check.
//=============================================================================
void RegLexer::pushFlushToken( Token tok )
{
    m_flushset.push_front( tok );
}

//=============================================================================
/// @brief Searches through the flush set for the given token and removes the first
/// token that matches.
/// @param tok - the token to remove.
//=============================================================================
void RegLexer::popFlushToken( Token tok )
{
    list<Token>::iterator i = m_flushset.begin();
    while( i != m_flushset.end() )
    {
        if( *i == tok )
        {
            m_flushset.erase( i );
            return;
        }
        i++;
    }
}

//=============================================================================
/// @brief This function flushes the tokens until it finds one in the flushset
/// and returns false if the first token is not in the flush set.
/// @return false if first token is not in flush set.
//=============================================================================
bool RegLexer::flush()
{
    if( peak()->type == -2 ){
        Tracer::debug( 9, "RegLexer::flush() - EOF REACHED" );
        return false;
    }
    if( !test( *peak() ) )
    {
        Token* tok = peak();
        Tracer::error( "Unexpected Token: %s\n\t\t %s %d:%d", 
                        tok->source.c_str(), m_fname.c_str(), tok->line, tok->col );
        pop();
        flush();
        return false;
    }
    return true;
}

//=============================================================================
/// @brief Check if the front token is in the flush set.
/// @param tok - the Token to test.
/// @return true if first token is in the flush set.
//=============================================================================
bool RegLexer::test( Token tok )
{
    for( list<Token>::iterator i = m_flushset.begin(); i != m_flushset.end(); i++ )
    {
        if( tok == *i )
            return true;
    }
    return false;
}

//=============================================================================
/// @brief Returns the current token.
/// @return a pointer to the token at the front of the list.
//=============================================================================
Token* RegLexer::peak()
{
    while( m_tokens.size() == 0 )
        readLine();
    return m_tokens.front();
}


//=============================================================================
/// @brief Returns the current token and pop's it off the stack.
/// @return the current token.
//=============================================================================
Token* RegLexer::pop()
{
    Token* tok = peak();
    m_tokens.pop_front();
//    Tracer::Debug( "pop( %d, %s )", tok->type, tok->source.c_str() );
    return tok;
}

//=============================================================================
/// @brief Puts a token back on the front of the list. This is used if you 
/// popped a token that you didn't want.
/// @param tok - a pointer to the token to put back
//=============================================================================
void RegLexer::pushToken( Token* tok )
{
    m_tokens.push_front( tok );
}


//=============================================================================
/// @brief Initializes the FSM.
//=============================================================================
void RegLexer::initFSM()
{

    m_fsm = new FSM();
    
    for( int i = 0; i <= 14; i++ )
        m_fsm->addState( i );

    // Initial State
    m_fsm->addTransition( LEX_INIT, Transition( KEY( 'a', 'z' ), LEX_ID, RegLexer::APPEND ) );
        // Lex ID / Keyword
        m_fsm->addTransition( LEX_ID, Transition( KEY( 'a', 'z' ), LEX_ID, RegLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( 'A', 'Z' ), LEX_ID, RegLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( '0', '9' ), LEX_ID, RegLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( DSEPARATOR ), LEX_ID, RegLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( '_' ), LEX_ID, RegLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( '$' ), LEX_ID, RegLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( '.' ), LEX_ID, RegLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( '@' ), LEX_ID, RegLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY(), LEX_DONE, RegLexer::PUSH_BACK ) ); 
                    // EVERYTHING ELSE
                
    m_fsm->addTransition( LEX_INIT, Transition( KEY( '<' ), LEX_OTAG, RegLexer::APPEND ) );
        // Lex Open 
        m_fsm->addTransition( LEX_OTAG, Transition( KEY( '/' ), LEX_OCTAG, RegLexer::APPEND ) );
        m_fsm->addTransition( LEX_OTAG, Transition( KEY(), LEX_DONE, RegLexer::PUSH_BACK ) );
        
            // Lex Open Close Tag
            m_fsm->addTransition( LEX_OCTAG, Transition( KEY(), LEX_DONE, RegLexer::PUSH_BACK ) );
            
    m_fsm->addTransition( LEX_INIT, Transition( KEY( '>' ), LEX_CTAG, RegLexer::APPEND ) );
        // Close Tag
        m_fsm->addTransition( LEX_CTAG, Transition( KEY(), LEX_DONE, RegLexer::PUSH_BACK ) );
    
        
    
    m_fsm->addTransition( LEX_INIT, Transition( KEY( '=' ), LEX_EQ, RegLexer::APPEND ) );
        // Lex = 
        m_fsm->addTransition( LEX_EQ, Transition( KEY(), LEX_DONE, RegLexer::PUSH_BACK ) );
    
    
    m_fsm->addTransition( LEX_INIT, Transition( KEY( '\"' ), 12, RegLexer::SKIP ) );
        // Lex String Const 
        m_fsm->addTransition( 12, Transition( KEY('\"'), LEX_ERROR, RegLexer::PUSH_BACK ) );
        m_fsm->addTransition( 12, Transition( KEY('\n'), LEX_ERROR, RegLexer::PUSH_BACK ) );
        m_fsm->addTransition( 12, Transition( KEY(), 13, RegLexer::APPEND ) );
            // building string
            m_fsm->addTransition( 13, Transition( KEY( '\"' ), LEX_STR, RegLexer::SKIP ));
            m_fsm->addTransition( 13, Transition( KEY('\n'), LEX_ERROR, RegLexer::PUSH_BACK ) );
            m_fsm->addTransition( 13, Transition( KEY(), 13, RegLexer::APPEND ));
                m_fsm->addTransition( LEX_STR, Transition( KEY(), LEX_DONE, RegLexer::PUSH_BACK ) );
    
    m_fsm->addTransition( LEX_INIT, Transition( KEY( '/' ), 14, RegLexer::APPEND ) );
        // Close Tag
        m_fsm->addTransition( 14, Transition( KEY('>'), LEX_CATAG, RegLexer::APPEND ) );
                m_fsm->addTransition( LEX_CATAG, Transition( KEY(), LEX_DONE, RegLexer::PUSH_BACK ) );
        
    m_fsm->addTransition( LEX_INIT, Transition( KEY(' '), LEX_INIT, RegLexer::SKIP ) );
    m_fsm->addTransition( LEX_INIT, Transition( KEY('\t'), LEX_INIT, RegLexer::SKIP ) );
    m_fsm->addTransition( LEX_INIT, Transition( KEY('\n'), LEX_INIT, RegLexer::SKIP ) );
    m_fsm->addTransition( LEX_INIT, Transition( KEY(), LEX_ERROR, RegLexer::SKIP ) );
}
