#include"Registry.h"
#include"RegLexer.h"
#include<tracer/Tracer.h>
#include<dxml/Token.h>

string Registry::m_appfile;
bool   Registry::m_initialized  = false;
Entry* Registry::m_root         = NULL;

//=============================================================================
/// @brief Constructor.
/// @param filename - the name of the configuration file to load.
//=============================================================================
Registry::Registry( const string& filename )
{
    init();
}

//=============================================================================
/// @brief Initializes the statis Registry object.
/// @param filename - the name of the configuration file to load.
//=============================================================================
void Registry::init( const string& filename )
{
    if( m_initialized ) return;
    m_initialized = true;
    m_appfile = filename;

    m_root = new Entry( "Registry", filename );

    //=====================================
    // Parse the config file
    //=====================================
    RegLexer lex( m_appfile );
    lex.pushFlushToken( Token(-2, "EOF" ) );    
    if( !m_root->read( lex ) )
    {
        Tracer::debug( 5, "Error reading config file" );
    }
}

//=============================================================================
/// @brief Saves the registry.
//=============================================================================
void Registry::save()
{
    Tracer t("Registry::Save()");
    ofstream out(m_appfile.c_str());
    m_root->save( out );
    out.close();
}

//=============================================================================
/// @brief Saves the registry as the specified file name.
/// @param fname - the name of the file to save the registry to.
//=============================================================================
void Registry::saveAs( const string& fname )
{
    ofstream out(fname.c_str());
    m_root->save( out );
    out.close();
}

//=============================================================================
/// @brief Retrieves the value for the specified path or uses the value passed
/// if the path and value does not exist.
/// @param path - the registry node from which to get the value.
/// @param val - the default value if the path doesn't exist in the registry.
/// @param cmt - the comment to use if the path doesn't exist in the registry.
/// @return a string containing the value.
//=============================================================================
string Registry::getValue( const string& path, const string& val,
                           const string& cmt )
{
    if( path[0] == '/' )
        return m_root->getValue( (path.c_str()+1), val, cmt );
    return m_root->getValue( path, val, cmt );
}

//=============================================================================
/// @brief Retrieves the comment for the specified path.
/// @param path - the registry node from which to get the comment.
/// @return a string containing the comment.
//=============================================================================
string Registry::getComment( const string& path )
{
    if( path[0] == '/' )
        return m_root->getComment( (path.c_str()+1) );
    return m_root->getComment( path );
}

//=============================================================================
/// @brief Sets the value for the specified path.
/// @param path - the registry node for which to set the value.
/// @param val - the value to be set.
//=============================================================================
void Registry::setValue( const string& path, const string& val )
{
    if( path[0] == '/' )
        m_root->setValue( (path.c_str()+1), val );
    else
        m_root->setValue( path, val );
}

//=============================================================================
/// @brief Sets the comment for the specified path.
/// @param path - the registry node for which to set the comment.
/// @param cmt - the comment to be set.
//=============================================================================
void Registry::setComment( const string& path, const string& cmt )
{
    if( path[0] == '/' )
        m_root->setComment( (path.c_str()+1), cmt );
    else
        m_root->setComment( path, cmt );
}

//=============================================================================
/// @brief Remove the specified path from the registry.
/// @param path - the registry node to remove.
//=============================================================================
void Registry::remove( const string& path )
{
    if( path[0] == '/' )
        m_root->remove( (path.c_str()+1) );
    else
        m_root->remove( path );
}

//=============================================================================
/// @brief Retrieve the Entry for the specified registry node.
/// @param path - thethe registry node for which to retrieve the Entry.
/// @return a pointer to the Entry for the specified path.
//=============================================================================
Entry* Registry::getEntry( const string& path )
{
    if( path.size() == 0 ) return m_root;
    if( path[0] == '/' ) return m_root->getEntry( (path.c_str()+1) );
    return m_root->getEntry( path );
}
