#ifndef _DGLCONTEXT_H_
#define _DGLCONTEXT_H_
#include <map>
#include <vector>
#include <tracer/Tracer.h>
#include <iostream>
#include <displaymgr/DisplayManager.h>
#ifdef WIN32
#include <windows.h>
#else
#include <GL/glx.h>
#endif

#ifdef WIN32
#ifdef DISPLAYMGR_EXPORTS
#define DISPLAYMGR_API __declspec(dllexport)
#else
#define DISPLAYMGR_API __declspec(dllimport)
#endif /* DISPLAYMGR_EXPORTS */
#else
#define DISPLAYMGR_API
#endif /* WIN32 */

#ifdef WIN32
#define DGLCONTEXT HGLRC
#else
#define DGLCONTEXT GLXContext
#endif

#ifdef WIN32
#define GETCURRENTCONTEXT wglGetCurrentContext
#else
#define GETCURRENTCONTEXT glXGetCurrentContext
#endif

using namespace std;

//=============================================================================
/// @class ContextVar dutil/ContextVar.h
/// @author Patrick Shinpaugh
/// @brief A template class which stores OpenGL context sensitive variables for
/// each context.
///
/// The ContextVar template class is a useful template for variables which may
/// have different values across multiple contexts within an application. It
/// uses WGL and GLX function calls to determine the current context and pass
/// or store the correct value for the correct context.
//=============================================================================
template<class cvtype>
class ContextVar
{
    public:
        ContextVar();
        ContextVar( const cvtype& val );

        cvtype  get()const;
        cvtype  set( const cvtype& val );

        cvtype& operator*();   // ContextVar<int> x;     *x == x.get()
        cvtype& operator=( const cvtype& val );
        typename map<DGLCONTEXT, cvtype>::iterator begin(){ return m_value.begin(); }
        typename map<DGLCONTEXT, cvtype>::iterator end(){ return m_value.end();  }

        bool isDefined()const;
        DGLCONTEXT getContext() { return GETCURRENTCONTEXT(); };
        int size();

    private:
        cvtype m_default;
        map<DGLCONTEXT, cvtype> m_value;
};

/// @brief Default constructor for ContextVar.
///
/// The initial value of the context sensitive variable is set to 0 or NULL
template<class cvtype>
ContextVar<cvtype>::ContextVar()
{
    m_default = 0;
}

/// @brief Overloaded constructor for ContextVar.
///
/// This overloaded constructor allows the initial/default value to be set.
/// @param val - Value to be set of the template type originally set
template<class cvtype>
ContextVar<cvtype>::ContextVar( const cvtype& val )
{
    m_default = 0;
    DGLCONTEXT context = GETCURRENTCONTEXT();
    if ( context )
        m_value[context] = val;
    else
        Tracer::warn(6, "ContextVar::ContextVar - Context does not exist" );
}

/// @brief Gets the value associated with the current context
/// @return the value associated with the current context
template<class cvtype>
cvtype ContextVar<cvtype>::get()const
{
    DGLCONTEXT context = GETCURRENTCONTEXT();
    if( !context )
    {
        Tracer::warn("ContextVar::get - No known contexts" );
        return m_default;
    }
    
    if( m_value.find(context) == m_value.end() )
    {
         Tracer::warn("ContextVar::get() - No Value for Context %d", 
                      context );
        
         DGLCONTEXT parent = 
                    DisplayManager::getSharedContextParent( context );
         if ( parent )
         {
             if( m_value.find(parent) == m_value.end() )
             {
                 Tracer::warn("ContextVar::get() - "
                              "No Value for Context parent %d", 
                               parent );
                 return m_default;
             }
             return m_value.find(parent)->second;
         }
    }
    return m_value.find(context)->second;
}

/// @brief Sets the value of the current context
/// @param val - A reference to the value to set for the current context
/// @return the value set for the current context
template<class cvtype>
cvtype ContextVar<cvtype>::set( const cvtype& val )
{
    DGLCONTEXT context = GETCURRENTCONTEXT();
    if ( context )
    {
        m_value[context] = val;
        return m_value[context];
    }
    else
        Tracer::warn("ContextVar::set - No known contexts" );
    return m_default;
}

/// @brief Returns a reference to the value associated with the current context
/// @return a reference to the current context such that the value can be manipulated directly
template<class cvtype>
cvtype& ContextVar<cvtype>::operator*()
{
    DGLCONTEXT context = GETCURRENTCONTEXT();
    if ( context )
    {
        typename map<DGLCONTEXT, cvtype>::iterator itr = m_value.find(context);
        if ( itr != m_value.end() )
            return itr->second;
        else
        {
            DGLCONTEXT parent = DisplayManager::getSharedContextParent( context );
            if ( parent )
                return m_value[parent];
        }
    }
    return m_value[context];
}

/// @brief Can be used in conjunction with the reference operator * to set the value for the current context
/// @param val - A reference to the value to set for the current context
/// @return a reference to the value set for the current context
template<class cvtype>
cvtype& ContextVar<cvtype>::operator=( const cvtype& val )
{
    DGLCONTEXT context = GETCURRENTCONTEXT();
    if ( context )
        return m_value[context] = val;
    else
        Tracer::warn("ContextVar::operator= - No known contexts" );
    return m_default;
}

/// @brief Determines whether a value has been set for the current context
/// @return true if the variable is set for this context, returns false if it is not
template<class cvtype>
bool ContextVar<cvtype>::isDefined()const
{
    DGLCONTEXT context = GETCURRENTCONTEXT();

    if ( context )
    {
        typename map<DGLCONTEXT, cvtype>::const_iterator itr = m_value.find(context);
        if ( itr != m_value.end() )
            return true;
    }
    Tracer::warn("ContextVar::isDefined - No known contexts" );
    return false;
}

/// @brief Determines the number of contexts for which the variable has been set
/// @return the number of contexts for which the variable has been set
template<class cvtype>
int ContextVar<cvtype>::size()
{
    return m_value.size();
}

#endif
