#ifndef _DTIME_H_
#define _DTIME_H_

#ifdef WIN32
#include <time.h>
#include <sys/timeb.h>
#include <winsock2.h>

#ifdef DUTIL_EXPORTS
#define DUTIL_API __declspec(dllexport)
#else
#define DUTIL_API __declspec(dllimport)
#endif /* DUTIL_EXPORTS */

#ifdef __cplusplus
extern "C" {
#endif

/// @brief Wrapper function to emulate gettimeofday UNIX/LINUX function.
/// @param t - the current system time is passed through the timeval struct.
inline DUTIL_API int gettimeofday( timeval* t, int )
{
	_timeb wintime;
	_ftime( &wintime );
	t->tv_sec = wintime.time;
	t->tv_usec = wintime.millitm * 1000;
	return 0;
};

#ifdef __cplusplus
}
#endif

#else
#define DUTIL_API
#endif /* WIN32 */

//=============================================================================
/// @class DTime
/// @brief A class for use as a timer or to hold time information.
///
/// To use DTime as a timer, instantiate a DTime object. Use the start function
/// to start the timer which will contan the current time in microseconds since
/// January 1, 1970. The elapsed function will return the time elapsed since the 
/// timer was started. The reset function will return the time elapsed since the
/// timer was started and reset the timer with the current time.
///
/// DTime can also be used for time, though it only supports hours, minutes,
/// seconds, milliseconds, and microseconds. Common arithmetic operations are
/// also available.
//=============================================================================
class DUTIL_API DTime
{
    public:
        DTime( int usec = 0 );
        DTime( int sec, int usec );
        DTime( int min, int sec, int usec );
        DTime( int hr, int min, int sec, int usec );
        static DTime current();

        void  start();
        DTime reset();
        DTime elapsed()const;

		/// @brief Returns the number of hours since January 1, 1970 or since the
		/// last start or reset.
		/// @return a float representing the number of hours.
        float hour()const   { return minute()/60.; }
		/// @brief Returns the number of minutes since January 1, 1970 or since the
		/// last start or reset.
		/// @return a float representing the number of minutes.
        float minute()const { return sec()/60.;    }
		/// @brief Returns the number of seconds since January 1, 1970 or since the
		/// last start or reset.
		/// @return a float representing the number of seconds.
        float sec()const    { return msec()/1000.; }
		/// @brief Returns the number of milliseconds since January 1, 1970 or since the
		/// last start or reset.
		/// @return a float representing the number of milliseconds.
        float msec()const   { return m_usec/1000.; }
		/// @brief Returns the number of microseconds since January 1, 1970 or since the
		/// last start or reset.
		/// @return a float representing the number of microseconds.
        int   usec()const   { return m_usec; }

		/// @brief subtraction operator.
		/// @return difference of the values of the DTime objects
        DTime  operator- (const DTime& b ) { return DTime( m_usec - b.m_usec ); }
		/// @brief subtraction assignment operator.
		/// @return difference of the values of two DTime objects
        DTime& operator-=(const DTime& b ) { m_usec -= b.m_usec; return *this;  }
		/// @brief addition operator.
		/// @return sum of the values of two DTime objects
        DTime  operator+ (const DTime& b ) { return DTime( m_usec + b.m_usec ); }
		/// @brief addition assignment operator.
		/// @return sum of the values of two DTime objects
        DTime& operator+=(const DTime& b ) { m_usec += b.m_usec; return *this;  }
		/// @brief greater than relational operator.
		/// @return true if this time is greater than time b, else returns false.
        bool operator> (const DTime& b ) { return m_usec >  b.m_usec; }
		/// @brief greater than or equal relational operator.
		/// @return true if this time is greater than or equal to time b, else returns false.
        bool operator>=(const DTime& b ) { return m_usec >= b.m_usec; }
		/// @brief less than relational operator.
		/// @return true if this time is less than time b, else returns false.
        bool operator< (const DTime& b ) { return m_usec <  b.m_usec; }
		/// @brief less than or equal relational operator.
		/// @return true if this time is less than or equal to time b, else returns false.
        bool operator<=(const DTime& b ) { return m_usec <= b.m_usec; }
		/// @brief equal relational operator.
		/// @return true if this time is equal to time b, else returns false.
        bool operator==(const DTime& b ) { return m_usec == b.m_usec; }
		/// @brief not equal relational operator.
		/// @return true if this time is not equal to time b, else returns false.
        bool operator!=(const DTime& b ) { return m_usec != b.m_usec; }

    private:
		int   m_usec; // micro-sec from Jan 1, 1970
        int   m_start;
};

#endif
