#include"XmlLexer.h"
#include<map>
#include<vector>
#include<algorithm>
#include<iostream>
#include<fstream>
#include<ctype.h>
#include"Token.h"
#include"fsm.h"
#include<tracer/Tracer.h>

using namespace std;

#ifdef WIN32
#define DSEPARATOR ';'
#else
#define DSEPARATOR ':'
#endif

//=============================================================================
/// @brief Constructor.
/// @param fname - the filename to read.
//=============================================================================
XmlLexer::XmlLexer( const string& fname )
    :m_file(fname.c_str()), m_type_str( 18 )
{
    m_fname = fname;
    m_cur_line = 0; 
    m_start = LEX_INIT;
    m_done  = LEX_DONE;
    m_last  = -1;
    m_printdebug = false;

    m_type_str[ LEX_DONE   ] = "DONE";
    m_type_str[ LEX_INIT   ] = "INIT";
    m_type_str[ LEX_ID     ] = "LEX_ID";
    m_type_str[ LEX_STR    ] = "LEX_STR";
    m_type_str[ LEX_EQ     ] = "LEX_EQ";
    m_type_str[ LEX_CTAG   ] = "LEX_CTAG";
    m_type_str[ LEX_OTAG   ] = "LEX_OTAG";
    m_type_str[ LEX_OCTAG  ] = "LEX_OCTAG";
    m_type_str[ LEX_CATAG  ] = "LEX_CATAG";
    m_type_str[ LEX_TEXT   ] = "LEX_TEXT"; 
    m_type_str[ 13         ] = "INTERMEDIATE1";
    m_type_str[ 14         ] = "INTERMEDIATE2";
    m_type_str[ 15         ] = "INTERMEDIATE3";
    m_type_str[ LEX_NOID   ] = "LEX_NOID";
    m_type_str[ LEX_1SPACE ] = "LEX_1SPACE";
    initFSM();

    if( !m_file.good() )
    {
        Tracer::warn( "Config file %s not found.", m_fname.c_str() );
        m_tokens.push_back( new Token( -2, "EOF", m_cur_line, 0 ) );
        m_done = -1;
    }
    readLine();
}

//=============================================================================
/// @brief Get the string for the specified type.
/// @param type - the type being requested.
/// @return a string represening the type specified.
//=============================================================================
string XmlLexer::getTypeStr( int type )const
{
    if( type < 0 ) return "ERROR";
    return m_type_str[type];
}


//=============================================================================
/// @brief Destructor. Deletes all tokens.
//=============================================================================
XmlLexer::~XmlLexer( )
{
    return;
    list<Token*>::iterator itr = m_tokens.begin();
    while( itr != m_tokens.end() )
        delete *itr;
    m_tokens.clear();
    m_file.close();
}


//=============================================================================
/// @brief Reads a line of input and generates tokens.
//=============================================================================
void XmlLexer::readLine()
{
    if( m_start == -1 || m_done == -1 ) return;
    getline( m_file, m_line );
    if( m_printdebug ) cout<<m_line<<endl;

    if( m_file.eof() )
    {
        m_tokens.push_back( new Token( -2, "EOF", m_cur_line, 0 ) );
        return;
    }
    m_line += '\n';
    m_cur_line++;


    int next_state;
    int action;
    int start_col = 0;  // The starting col of the token
//    string lexem;   // the lexem we are building
    
    for( unsigned int i = 0; i < m_line.size(); i++ )
    {
        m_last = m_fsm->getCurrentState();
        if( m_last == -1 )
            m_fsm->gotoState( m_start );
        
        if( -1 == m_fsm->getNextState( m_line[i], &next_state, &action ) )
        {
            cout<<"Line: "<<m_cur_line<<":"<<(start_col+1)<<" - unknown Token: \'"<<m_line[i]<<"\' found\n";
            m_fsm->gotoState( m_start ); // Start Over if error
            m_lexem = "";
            start_col = i;
            i--;
            continue;
        }
       
        if( m_printdebug )
            cout<<"prev state = "<<getTypeStr(m_last)<<"  next state = "<<getTypeStr(next_state)<<" action = ";
        switch( action )
        {
            case PUSH_BACK:
                i--;
                //cout<<"PUSH_BACK";
                break;
            case APPEND:
                m_lexem += m_line[i];
               // cout<<"APPEND ["<<i<<"] = \'"<<m_line[i]<<"\'";
                break;
            case SKIP:
               // cout<<"SKIP";
                break;
            case APPEND_SPACE:
               // cout<<"APPEND_SPACE";
                m_lexem += " ";
                break;
        }
//        cout<<" state: "<<next_state<<" Lexem: "<<m_lexem<<endl;
        if( next_state == m_done || next_state == LEX_NOID )
        {
//            cout<<"finish token: last state= "<<getTypeStr(m_last)<<" lexem: "<<m_lexem<<endl;
            // if the last state was LEX_NOID then we had no text
            if( m_last == LEX_NOID )
            {
                m_lexem = "";
                start_col = i+1;
                m_fsm->gotoState( m_start );
                // start_col doesn't change
                // lexem doesn't change
                //cout<<"No LEX_TEXT was found, therefore no token\n";
                continue;
            }
            Token* tok = new Token( m_last, m_lexem, m_cur_line, start_col+1 );
            m_tokens.push_back( tok );
            if( next_state == m_done )
                m_fsm->gotoState( m_start );
            start_col = i+1;
            m_lexem = "";
            continue;
        }
        m_last = next_state;
        if( m_last == -1 )
        {
            cout<<"Line: "<<m_cur_line<<":"<<(start_col+1)<<" - unknown Token: \'"<<(m_lexem + m_line[i])<<"\' found\n";
            m_fsm->gotoState( m_start ); // Start Over if error
            m_lexem = "";
            start_col = i;
        }
    }
}

//=============================================================================
/// @brief Adds a new token to the flush list, when flush is called it will
/// go through the list of tokens.
/// @param tok - the token to add to the list to check when flushing.
//=============================================================================
void XmlLexer::pushFlushToken( Token tok )
{
    m_flushset.push_front( tok );
}

//=============================================================================
/// @brief Searches through the flush set for the given token and removes the first
/// token that matches.
/// @param tok - the token to remove.
//=============================================================================
void XmlLexer::popFlushToken( Token tok )
{
    list<Token>::iterator i = m_flushset.begin();
    while( i != m_flushset.end() )
    {
        if( *i == tok )
        {
            m_flushset.erase( i );
            return;
        }
        i++;
    }
}

//=============================================================================
/// @brief This function flushes the tokens until it finds one in the flushset
/// and returns false if the first token is not in the flush set.
/// @return false if first token is not in flush set.
//=============================================================================
bool XmlLexer::flush()
{
    if( peak()->type == -2 ){
        Tracer::debug( 9, "XmlLexer::flush() - EOF REACHED" );
        return false;
    }
    if( !test( *peak() ) )
    {
        Token* tok = peak();
        Tracer::error( "Unexpected Token: type:%s lexem:%s:\n\t\t %s %d:%d", 
                        getTypeStr(tok->type).c_str(),
                        tok->source.c_str(), m_fname.c_str(), tok->line, tok->col );
//        for( list<Token>::iterator i = m_flushset.begin(); i != m_flushset.end(); i++ )
//        {
//            cout<<"Flush Set: "<<i->lexem<<"  "<<m_type_str[i->type]<<endl;
//        }
        pop();
        flush();
        return false;
    }
    return true;
}

//=============================================================================
/// @brief Check if the front token is in the flush set.
/// @return true if first token is in the flush set.
//=============================================================================
bool XmlLexer::test( Token tok )
{
    for( list<Token>::iterator i = m_flushset.begin(); i != m_flushset.end(); i++ )
    {
        if( tok == *i )
            return true;
    }
    return false;
}

//=============================================================================
/// @brief Returns the current token.
/// @return a poitner to the token at the front of the list.
//=============================================================================
Token* XmlLexer::peak()
{
    while( m_tokens.size() == 0 )
        readLine();
    return m_tokens.front();
}

//=============================================================================
/// @brief Returns the current token and pop's it off the stack.
/// @return the current token.
//=============================================================================
Token* XmlLexer::pop()
{
    Token* tok = peak();
    m_tokens.pop_front();
//    Tracer::debug( "pop( %s, %s )", m_type_str[tok->type].c_str(),
//                                    tok->lexem.c_str() );
    return tok;
}

//=============================================================================
/// @brief Puts a token back on the front of the list. This is used if you 
/// popped a token that you didn't want.
/// @param tok - a pointer to the token to put back.
//=============================================================================
void XmlLexer::pushToken( Token* tok )
{
    m_tokens.push_front( tok );
}

//=============================================================================
/// @brief Initializes the FSM
//=============================================================================
void XmlLexer::initFSM()
{

    m_fsm = new FSM();
    
    for( int i = 0; i <= 18; i++ )
        m_fsm->addState( i );

    m_fsm->addTransition( LEX_INIT, Transition( KEY('\n'), LEX_INIT, XmlLexer::SKIP ) );
    m_fsm->addTransition( LEX_INIT, Transition( KEY(' '), LEX_INIT, XmlLexer::SKIP ) );
    m_fsm->addTransition( LEX_INIT, Transition( KEY('\t'), LEX_INIT, XmlLexer::SKIP ) );

    // Initial State
    m_fsm->addTransition( LEX_INIT, Transition( KEY( 'a', 'z' ), LEX_ID, XmlLexer::APPEND ) );
    m_fsm->addTransition( LEX_INIT, Transition( KEY( 'A', 'Z' ), LEX_ID, XmlLexer::APPEND ) );
        // Lex ID / Keyword
        m_fsm->addTransition( LEX_ID, Transition( KEY( 'a', 'z' ), LEX_ID, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( 'A', 'Z' ), LEX_ID, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( '0', '9' ), LEX_ID, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( DSEPARATOR ), LEX_ID, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( '_' ), LEX_ID, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( '$' ), LEX_ID, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( '.' ), LEX_ID, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY( '@' ), LEX_ID, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_ID, Transition( KEY(), LEX_DONE, XmlLexer::PUSH_BACK ) ); 
                    // EVERYTHING ELSE
                
    m_fsm->addTransition( LEX_INIT, Transition( KEY( '<' ), LEX_OTAG, XmlLexer::APPEND ) );
        // Lex Open 
        m_fsm->addTransition( LEX_OTAG, Transition( KEY( '/' ), LEX_OCTAG, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_OTAG, Transition( KEY(), LEX_DONE, XmlLexer::PUSH_BACK ) );
        
            // Lex Open Close Tag
            m_fsm->addTransition( LEX_OCTAG, Transition( KEY(), LEX_DONE, XmlLexer::PUSH_BACK ) );
            
    // Close Tag >
    m_fsm->addTransition( LEX_INIT, Transition( KEY( '>' ), LEX_CTAG, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_CTAG, Transition( KEY(), LEX_NOID, XmlLexer::PUSH_BACK ) );
   
    // Lex = 
    m_fsm->addTransition( LEX_INIT, Transition( KEY( '=' ), LEX_EQ, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_EQ, Transition( KEY(), LEX_DONE, XmlLexer::PUSH_BACK ) );
    
    
    m_fsm->addTransition( LEX_INIT, Transition( KEY( '\"' ), 13, XmlLexer::SKIP ) );
        // Lex String Const 
        m_fsm->addTransition( 13, Transition( KEY('\"'), LEX_STR, XmlLexer::SKIP ) );
        m_fsm->addTransition( 13, Transition( KEY('\n'), LEX_ERROR, XmlLexer::PUSH_BACK ) );
        m_fsm->addTransition( 13, Transition( KEY(), 13, XmlLexer::APPEND ) );
            m_fsm->addTransition( LEX_STR, Transition( KEY(), LEX_DONE, XmlLexer::PUSH_BACK ) );
    
    m_fsm->addTransition( LEX_INIT, Transition( KEY( '/' ), 15, XmlLexer::APPEND ) );
        // Close Tag
        m_fsm->addTransition( 15, Transition( KEY('>'), LEX_CATAG, XmlLexer::APPEND ) );
                m_fsm->addTransition( LEX_CATAG, Transition( KEY(), LEX_NOID, XmlLexer::PUSH_BACK ) );
        
    // support XML text tokens, anything until an open tag '<'  or new line or space
    m_fsm->addTransition( LEX_NOID, Transition( KEY(), LEX_TEXT, XmlLexer::PUSH_BACK ) );
        m_fsm->addTransition( LEX_TEXT, Transition( KEY('<'), LEX_DONE, XmlLexer::PUSH_BACK ) );
        m_fsm->addTransition( LEX_TEXT, Transition( KEY(' '), LEX_1SPACE, XmlLexer::APPEND ) );
        m_fsm->addTransition( LEX_TEXT, Transition( KEY('\t'), LEX_1SPACE, XmlLexer::APPEND_SPACE ) );
        m_fsm->addTransition( LEX_TEXT, Transition( KEY('\n'), LEX_1SPACE, XmlLexer::APPEND_SPACE ) );
            m_fsm->addTransition( LEX_1SPACE, Transition( KEY(' '), LEX_1SPACE, XmlLexer::SKIP ) );
            m_fsm->addTransition( LEX_1SPACE, Transition( KEY('\t'), LEX_1SPACE, XmlLexer::SKIP ) );
            m_fsm->addTransition( LEX_1SPACE, Transition( KEY('\n'), LEX_1SPACE, XmlLexer::SKIP ) );
            m_fsm->addTransition( LEX_1SPACE, Transition( KEY(), LEX_TEXT, XmlLexer::PUSH_BACK ) );
        m_fsm->addTransition( LEX_TEXT, Transition( KEY(), LEX_TEXT, XmlLexer::APPEND ) );

    m_fsm->addTransition( LEX_INIT, Transition( KEY(), LEX_ERROR, XmlLexer::SKIP ) );
}
