//=============================================================================
// XmlLexer.h
//   - This class provides a general purpoes FSM based lexer that uses a 
//   lookup table.  It reads one line at a time and provides functions for 
//   getting the current token, popping a token and pushing a token back 
//
// Created 9/19/02 By Daniel Larimer
//=============================================================================
#ifndef _REGLEXER_H_
#define _REGLEXER_H_
#include<list>
#include<string>
#include<fstream>
#include<vector>
#include<list>
#include<string>

#ifdef WIN32
#ifdef DXML_EXPORTS
#define DXML_API __declspec(dllexport)
#else
#define DXML_API __declspec(dllimport)
#endif /* DXML_EXPORTS */
#else
#define DXML_API
#endif /* WIN32 */

struct Token;
class FSM;
using namespace std;

//=============================================================================
/// @class XmlLexer
/// @author Daniel Larimer
/// @brief The lexer for XML.
///
/// This class provides a general purpoes FSM based lexer that uses a 
/// lookup table.  It reads one line at a time and provides functions for 
/// getting the current token, popping a token and pushing a token back .
//=============================================================================
class DXML_API XmlLexer
{
    public:
        XmlLexer( const string& fname );
        virtual ~XmlLexer();

		/// brief Tokens used with the XmlLexer.
        enum REG_TOKENS
		{
			/// An error occurred
			LEX_ERROR    = -1,
			/// The lexing process is complete.
            LEX_DONE     = 0,
			/// Initialization.
			LEX_INIT     = 1,
			/// Standard characters a-z, A-Z, 0-9
			LEX_ID       = 2,
			/// A string in quotes ( "..." )
			LEX_STR      = 3,
			/// An equal symbol ( = )
			LEX_EQ       = 4,
			/// A less than symbol ( < )
			LEX_OTAG     = 8,
			/// A greater than symbol ( > )
			LEX_CTAG     = 9,
			/// A C tag representing the end of an entry with sub-entries ( </ )
			LEX_OCTAG    = 10,
			/// An OC tag representing the the end of an entry without sub-entries ( /> )
			LEX_CATAG    = 11,
			/// A keyword or value
			LEX_TEXT     = 12,
			/// state after first "
			LEX_1QUOTE   = 13,
			/// state for building string
			LEX_BSTR     = 14,
			/// state after / while we look for >
			LEX_1SLASH   = 15,
			/// state after > where we are looking for text, but no ID
			LEX_NOID     = 16,
			/// state after ' ' that will skip every additional space
			LEX_1SPACE   = 17,
		};
                          
        enum ACTIONS
		{
			/// Append the token to previous token.
			APPEND,
			/// Add the token to the token list.
			PUSH_BACK,
			/// Skip the token, ignoring it.
			SKIP,
			/// Append a space to the previous token.
			APPEND_SPACE
		};

        virtual Token* peak(); 
        
        virtual Token* pop();
        void   pushToken( Token* tok );

        bool flush();
        bool test( Token t );
        void pushFlushToken( Token t );
        void popFlushToken( Token t );

        //==================================================
        // Initializes Transition Matrix from a string
        // description that can be either hard coded or 
        // read from a file
        //==================================================
		/// @brief Set the FSM for the RegLexer.
		/// @param fsm - the FSM to be used with the RegLexer.
        void setFSM( FSM* fsm ){ m_fsm = fsm; }
		/// @brief Set the debug state for the RegLexer.
		/// @param state - set to true if debugging, else set to false.
        void setDebugState( bool state) { m_printdebug = state; }
		/// @brief Set the start state of the RegLexer.
		/// @param state - the start state of the RegLexer.
        void setStartState( int state ) { m_start = state; }
		/// @brief Set the completion state of the RegLexer.
		/// @param state - the completion state of the RegLexer.
        void setDoneState( int state ) { m_done  = state; }
        string getTypeStr( int type )const; 

    protected:
        void readLine();
        void initFSM();
        
        list<Token*> m_tokens;   // List of tokens
        int          m_cur_line; // The current line
        string       m_lexem;
        string       m_line;
        string       m_fname;
        ifstream     m_file;
        FSM          *m_fsm;

        int          m_start;
        int          m_done;
        int          m_last;
        
        vector<string> m_type_str;
        list<Token>    m_flushset;

        bool         m_printdebug;
};

#endif
