#include"XmlNode.h"
#include <tracer/Tracer.h>

//=============================================================================
/// @brief Constructor that takes the name of the XML tag.
/// @param tagname - the name of the tag <tagname/>.
//=============================================================================
XmlNode::XmlNode( const string& tagname )
{
    m_tagname = tagname;
}

//=============================================================================
/// @brief Destructor. Deletes all child nodes.
//=============================================================================
XmlNode::~XmlNode()
{
    list<XmlNode*>::iterator itr = m_children.begin();
    while( itr != m_children.end() )
    {
		Tracer::debug( "Parent - %s    Child - %s", 
                        this->m_tagname.c_str(), 
                        (*itr)->m_tagname.c_str() );
        delete *itr;
        itr++;
    }
}

//=============================================================================
/// @brief Returns the name of this tag.
/// @return the tag name.
//=============================================================================
string XmlNode::getTag()const
{
    return m_tagname;
}

//=============================================================================
/// @brief Sets the tagname for this node.
/// @param tagname - the name of the tag.
//=============================================================================
void XmlNode::setTag( const string& tagname )
{
    m_tagname = tagname;
}

//=============================================================================
/// @brief Returns true if this node contains text.
/// @return true if the node contains text, otherwise false.
//=============================================================================
bool XmlNode::isTextNode()const
{
    return m_tagname=="";
}

//=============================================================================
/// @brief Gets the text if this is a text node, ie: no tag name.
/// @return the text of the node if it is a text node.
//=============================================================================
string XmlNode::getText()const
{
    return m_text;
}

//=============================================================================
/// @brief Sets the value of the text for a text node, ie: a tag w/out a name 
/// @param txt - the value to set the text to
//=============================================================================
void XmlNode::setText( const string& txt )
{
    m_text = txt;
}

//=============================================================================
/// @brief Determines if this node has a parameter by the specified name.
/// @param param - the name of the parameter to look for.
/// @return true if the parameter exists, else returns false.
//=============================================================================
bool XmlNode::hasParameter( const string& param )const
{
    return m_params.find(param) != m_params.end();
}

//=============================================================================
/// @brief Returns the value of the specified parameter.
/// @param param - the name of the parameter for which to get the value.
/// @return the value of specified parameter.
//=============================================================================
string XmlNode::getParameter( const string& param )const
{
    return m_params.find(param)->second;
}   

//=============================================================================
/// @brief Sets the value of the specified parameter.
/// @param param - the name of the parameter to set.
/// @param val - the value to be set for the specified parameter.
//=============================================================================
void  XmlNode::setParameter( const string& param, const string& val )
{
    m_params[param] = val;
}

//=============================================================================
/// @brief Adds the node to the end of the children list.
/// @param node - the node to append.
//=============================================================================
void XmlNode::appendNode( XmlNode* node )
{
    node->setParent( this );
    m_children.push_back( node );
}

//=============================================================================
/// @brief Adds the node to the front of the children list.
/// @param node - the node to append.
//=============================================================================
void XmlNode::prependNode( XmlNode* node )
{
    node->setParent( this );
    m_children.push_front( node );
}

//=============================================================================
/// @brief Writes the XML node to the specified out stream
/// @param out - the stream to print to
/// @param indent - the number of spaces to indent
//=============================================================================
void XmlNode::write( ostream& out, int indent )
{
    printIndent( out, indent );
    if( isTextNode() )
    {
        out<<m_text;
        return;
    }
    out<<"<"<<m_tagname<<" ";
    map<string, string>::iterator itr = m_params.begin();
    while( itr != m_params.end() )
    {
        out<<itr->first<<"=\""<<itr->second<<"\" ";
        itr++;
    }
    if( m_children.size() == 0 )
        out<<"/>";
    else
    {
        out<<">\n";
        list<XmlNode*>::iterator citr = m_children.begin();
        while( citr != m_children.end() )
        {
            (*citr)->write( out, indent+1 );
            out<<endl;
            citr++;
        }
        printIndent( out, indent );
        out<<"</"<<m_tagname<<">";
    }
}

//=============================================================================
/// @brief Used to print an indent when writing the tree to a file.
/// @param out - the output file stream to which the tree is being saved.
/// @param indent - the number of indents to use. An indent is 2 spaces.
//=============================================================================
void XmlNode::printIndent( ostream& out, int indent )
{
    while( indent > 0 )
    {
        out<<"  ";
        --indent;
    }
}
