#include"XmlParser.h"
#include"XmlNode.h"
#include"XmlLexer.h"
#include"Token.h"
#include<tracer/Tracer.h>
#include<iostream>
using namespace std;

//=============================================================================
/// @brief Constructor. Builds and parses the specified file.
/// @param fname - the file to parse
//=============================================================================
XmlParser::XmlParser( const string& fname )
{
    m_lex = NULL;
    m_root  = NULL;
    m_filename = fname;
    if( m_filename != "" )
        parse( m_filename );
}

//=============================================================================
/// @brief Destructor. Deletes the parsed tree.
//=============================================================================
XmlParser::~XmlParser()
{
    if( m_root ) delete m_root;
}

//=============================================================================
/// @brief Parses the file into separate nodes
/// @param fname - the name of the file to be parsed.
/// @return the root node of the tree.
//=============================================================================
XmlNode* XmlParser::parse( const string& fname )
{
    if( m_root  ) delete m_root;
    if( m_lex ) delete m_lex;
    m_lex = new XmlLexer( fname );
    m_root = parsehelper();
    return m_root;
}

//=============================================================================
/// @brief Parses one node from \<tag\> to \</tag\>.
///
/// grammar:                                                 <BR>
///             NODE => TEXT                                 <BR>
///                  => TAG                                  <BR>
///          TAG => \<ID PARAMLIST\> CHILDREN \</ID\>        <BR>
///          TAG => \<ID PARAMLIST/\>                        <BR>
///          PARAMLIST => ID=STR PARAMLIST                   <BR>
///                    => eps                                <BR>
///           CHILDREN => NODE CHILDREN                      <BR>
///                    => eps                                <BR>
/// @return the node parsed
//=============================================================================
XmlNode* XmlParser::parsehelper()
{
    // Push flush tokens
    m_lex->pushFlushToken( Token( XmlLexer::LEX_OTAG ) );
    m_lex->pushFlushToken( Token( XmlLexer::LEX_TEXT ) );
    m_lex->pushFlushToken( Token( XmlLexer::LEX_ID ) ); // also text
    m_lex->flush();

    //==================================================
    // Did we flush to an open tag, or a TEXT block?
    //==================================================
    Token* tok = m_lex->peak();
    if( tok->type == XmlLexer::LEX_TEXT )
    {
        m_lex->popFlushToken( Token( XmlLexer::LEX_TEXT ) );
        //============================================
        // If we got a single space ignore it, no text
        //============================================
        if( tok->lexem != " " )
        {
            //============================================
            // Create a text node
            //============================================
            XmlNode* txt = new XmlNode();
            txt->setText( tok->lexem );
            delete m_lex->pop();
            
            m_lex->popFlushToken( Token( XmlLexer::LEX_OTAG ) ); // <
            return txt;  // we found a text node
        }
        delete m_lex->pop();
        tok = NULL;
    }

    // We aren't looking for a text node anymore, pop it
    m_lex->popFlushToken( Token( XmlLexer::LEX_TEXT ) );
    m_lex->flush();
    tok = m_lex->peak();
    
    //===================================================
    // If flush didn't return < then we hit some other
    // flush token and should return
    //===================================================
    if( tok->type != XmlLexer::LEX_OTAG )
    {
        m_lex->popFlushToken( Token(0,"<") );
        Tracer::warn( 9, "XmlParser - failed to find open tag \'<\'" );
        Tracer::warn( 9, "\t Found %s -\'%s\' instead.\n"
                         "\t\t Line: %d Column: %d", 
                          m_lex->getTypeStr(tok->type).c_str(),
                          tok->lexem.c_str(),
                          tok->line, tok->col);
        return NULL;
    }
    // pop the < token
    delete m_lex->pop(); 

    // we have now entered the tag, add ID to the flush set
    m_lex->pushFlushToken( Token( XmlLexer::LEX_ID    ) ); // any id
    m_lex->flush();
    
    // check for the tag name
    tok = m_lex->peak();
    if( tok->type != XmlLexer::LEX_ID )
    {
        Tracer::warn( 5, "ID expected, but found %s", tok->source.c_str() );
        m_lex->popFlushToken( Token( XmlLexer::LEX_ID ) );
        return NULL;
    }

    // create our new node
    XmlNode* node = new XmlNode( tok->lexem );

    // pop the tag name ID token
    delete m_lex->pop();
   
    // we now have a node, now we must look for a close tag
    // but are also looking for param ID's so we won't pop that yet
    m_lex->pushFlushToken( Token( XmlLexer::LEX_CTAG  ) ); // >
    m_lex->pushFlushToken( Token( XmlLexer::LEX_CATAG ) ); // />
    m_lex->pushFlushToken( Token( XmlLexer::LEX_OCTAG ) ); // </
    m_lex->flush();
    tok = m_lex->peak();
    
    // see if we found a parameter
    while( tok->type == XmlLexer::LEX_ID )
    {
        string param = tok->lexem;
        delete m_lex->pop();

        // we are now looking for a = 
        m_lex->pushFlushToken( Token( XmlLexer::LEX_EQ ) );
        m_lex->flush();
        // if we didn't find it then we aren't looking for it anymore
        m_lex->popFlushToken( Token( XmlLexer::LEX_EQ ) );
        
        tok = m_lex->peak();
        if( tok->type != XmlLexer::LEX_EQ )
        {
            Tracer::warn( "Found parameter with out assignment, Token: %s Line: %d:%d",
                          tok->lexem.c_str(), tok->line, tok->col);
            break;
        }
        // we found a = so pop it and look for the string
        delete m_lex->pop();
        
        // look for the string
        m_lex->pushFlushToken( Token( XmlLexer::LEX_STR ) );
        m_lex->flush();
        m_lex->popFlushToken( Token( XmlLexer::LEX_STR ) );
        
        tok = m_lex->peak();
        if( tok->type == XmlLexer::LEX_STR )
        {
            node->setParameter( param, tok->lexem );
            delete m_lex->pop();
        }
        // flush to the next ID or token
        m_lex->flush();
        tok = m_lex->peak();
    }
    // we are no longer looking for param ids
    m_lex->popFlushToken( Token( XmlLexer::LEX_ID ) );
    
    m_lex->flush();
    tok = m_lex->peak();
    
    // if we don't have /> now, we aren't looking for it
    m_lex->popFlushToken( Token( XmlLexer::LEX_CATAG ) ); // />

    // see if we are done with the node
    if( tok->type == XmlLexer::LEX_CATAG ) // />
    {
        delete m_lex->pop();
        m_lex->popFlushToken( Token( XmlLexer::LEX_CTAG  ) ); // >
        m_lex->popFlushToken( Token( XmlLexer::LEX_OCTAG ) ); // </
        return node;
    }
    
    m_lex->popFlushToken( Token( XmlLexer::LEX_CTAG ) ); 
    if( tok->type == XmlLexer::LEX_CTAG ) // >  then we may have children
    {
        // we aren't looking for > anymore
        delete m_lex->pop(); // >
        m_lex->pushFlushToken( XmlLexer::LEX_TEXT );
        m_lex->flush();
        m_lex->popFlushToken( XmlLexer::LEX_TEXT );
        tok = m_lex->peak();
        while( tok->type != XmlLexer::LEX_OCTAG ) // != </ 
        {
            XmlNode* child = parsehelper();  // read child node
            if( child ) node->appendNode( child ); // add child
            else break; // no more children TBD: add more inteligence, 
                        // we might be at the start of another node
            tok = m_lex->peak();
        }
    }
    m_lex->flush();
    
    // if we don't have </ now then we aren't looking for it
    // because there was some sort of error
    m_lex->popFlushToken( Token( XmlLexer::LEX_OCTAG ) ); 
    
    tok = m_lex->peak();
    if( tok->type != XmlLexer::LEX_OCTAG )
    {
        Tracer::warn( 5, "Found %s, but looking for </", 
                        m_lex->getTypeStr(tok->type).c_str() );
        Tracer::warn( 5, "Couldn't find close tag for %s", 
                      node->getTag().c_str() );
        delete node;
        return NULL; 
    }
    // pop the </
    delete m_lex->pop(); 

    // look for the close tag id
    m_lex->pushFlushToken( Token( XmlLexer::LEX_ID ) );
    m_lex->flush();
    m_lex->popFlushToken( Token( XmlLexer::LEX_ID ) );
    
    tok = m_lex->peak();
    if( tok->type != XmlLexer::LEX_ID || tok->lexem != node->getTag() )
    {
        Tracer::error( "</ must be followed by the tag name: %s",
                        node->getTag().c_str() );
        delete node;
        return NULL;
    }
    delete m_lex->pop(); // pop the close tag name
    m_lex->pushFlushToken( Token( XmlLexer::LEX_CTAG ) );
    m_lex->flush();
    m_lex->popFlushToken( Token( XmlLexer::LEX_CTAG ) );
    tok = m_lex->peak();

    if( tok->type != XmlLexer::LEX_CTAG )
    {
        Tracer::error( "</%s is missing closing >", node->getTag().c_str() );
        delete node;
        return NULL;
    }
    delete m_lex->pop(); // delete /> we found it, but don't need it any more
    return node; // success.. we parsed a node
}
