//=============================================================================
// fsm.h
// - This is the implimentation of the finite state machine, it uses a map 
// of maps to jump from state to state based on an input  
//
// Created 9/20/02 By Daniel Larimer
//=============================================================================
#ifndef _FSM_H_
#define _FSM_H_
#include<map>
#include<vector>
#include<ctype.h>

#ifdef WIN32
#ifdef DXML_EXPORTS
#define DXML_API __declspec(dllexport)
#else
#define DXML_API __declspec(dllimport)
#endif /* DXML_EXPORTS */
#else
#define DXML_API
#endif /* WIN32 */

using namespace std;

//==================================================
/// @struct KEY
/// @brief This struct represents a column KEY for the 
/// fsm table.
///
/// This KEY can be a single character
/// or a range of characters (i.e. a-z or 0-9). It also
/// provides overloaded operators that allow it to
/// recognize when a character is between the range
//==================================================
struct DXML_API KEY
{
	/// @brief Constructor. KEY includes only the single character specified.
	/// @param val - the single character to use as the KEY.
    KEY( unsigned char val )
    {
        m_min = m_max = val;
    }
	/// @brief Constructor. KEY includes all characters from 0 to 255.
    KEY()
    {
        m_min = 0; m_max = 255;
    }
	/// @brief Constructor. KEY includes the range of characters specified.
	/// @param min - the minimum character in the range of characters to use as the KEY.
	/// @param max - the maximum character in the range of characters to use as the KEY.
    KEY( unsigned char min, unsigned char max )
    {
        m_min = min;
        m_max = max;
    }

    // Defines a range of characters 
    unsigned char m_min;
    unsigned char m_max;

	/// @brief equal relational operator.
    bool operator==( const unsigned char& ch )
    {
        return ch >= m_min && ch <= m_max;
    }
	/// @brief less than relational operator
    bool operator <( const unsigned char& ch )
    {
        return m_min < ch;
    }
	/// @brief greater than relational operator.
    bool operator >( const unsigned char& ch )
    {
        return m_min < ch;
    }
};

//===================================================
// @struct Transition
// @brief This struct is used to represent the next state
// and corresponding action to take.
//===================================================
struct DXML_API Transition 
{
	/// @brief Constructor.
	/// @param k - the KEY for transition.
	/// @param next - the next state.
	/// @param act - the action to take.
    Transition( KEY k, int next, int act )
    {
        m_next_state = next;
        m_action = act;
        m_key = k;
    }
    KEY m_key;
    int m_next_state;
    int m_action;
};

//===============================================================
/// @class FSM
/// @brief This is the finite state machine.
///
/// It has two main functions
/// and two setup functions. GotoState forces a switch to 
/// a particular state, and GetNextState takes the next input
/// character and uses it to determine the next state and action.
///
/// A sparse table is used where each state has fields only for
/// the Transitions that it expects and does not fill every possible
/// entry in the table.
//================================================================
class DXML_API FSM
{
    public:
        FSM();
        int  gotoState( int state );
        int  getNextState( char next_ch, int* nextstate, int* action ); 

        int  addState( int state );
         
        void addTransition( int state, Transition trans );

		/// @brief Get the current state of the finite state machine.
		/// @return the current state of the finite state machine.
        int  getCurrentState(){ return m_state->first; }
        
    private:
        map< int, vector<Transition> >           m_table;
        map< int, vector<Transition> >::iterator m_state;
};

#endif
