//=============================================================================
/// @EventListener
/// @brief An interface to any object that wants to receive events
/// @author Daniel Larimer
//=============================================================================
#ifndef _EVENTLISTENER_H_
#define _EVENTLISTENER_H_
#include<eventmgr/EventManager.h>
class DEvent;
class DMouseEvent;
class DWandEvent;
class DKeyEvent;
class DTrackerEvent;

//=============================================================================
/// @class EventListener
/// @brief Interface for event listening.
///
/// Plugins should be derived from EventListener if they need to respond to events.
/// The derived class must call the static EventManager add listener functions and
/// add itself as a listener to each event type that it needs.
///
/// Example:
///
/// EventManager::addMouseListener( this );
///
/// The derived class should handle each on event function for which it was added as
/// a listener.
//=============================================================================
class EventListener
{
    public:
        virtual ~EventListener(){}

		/// @brief Handle any event that passes through the EventManager.
		/// @param evt - the DEvent to be handled.
        /// @return true if the event was handled, otherwise false
        virtual bool onAnyEvent( DEvent* evt )           { return false; }
		/// @brief Handle mouse events that pass through the EventManager.
		/// @param evt - the DMouseEvent to be handled.
        /// @return true if the event was handled, otherwise false
        virtual bool onMouseEvent( DMouseEvent* evt )    { return false; }
		/// @brief Handle key events that pass through the EventManager.
		/// @param evt - the DKeyEvent to be handled.
        /// @return true if the event was handled, otherwise false
        virtual bool onKeyEvent( DKeyEvent* evt )        { return false; }
		/// @brief Handle wand events that pass through the EventManager.
		/// @param evt - the DWandEvent to be handled.
        /// @return true if the event was handled, otherwise false
        virtual bool onWandEvent( DWandEvent* evt )      { return false; }
		/// @brief Handle tracker events that pass through the EventManager.
		/// @param evt - the DTrackerEvent to be handled.
        /// @return true if the event was handled, otherwise false
        virtual bool onTrackerEvent( DTrackerEvent* evt ){ return false; }
		/// @brief Handle other events that pass through the EventManager.
		/// @param evt - the DEvent to be handled.
        /// @return true if the event was handled, otherwise false
        virtual bool onOtherEvent( DEvent* evt )         { return false; }

		/// @brief Sends all mouse events to this listener first.
		/// @return true if the mouse was available for grabbing
        virtual bool grabMouse()   { return EventManager::grabMouse( this ); }
		/// @brief Sends all keyboard events to this listener first.
		/// @return true if the key was available for grabbing
        virtual bool grabKeyboard(){ return EventManager::grabKeyboard( this ); }
		/// @brief Sends all wand events to this listener first.
		/// @return true if the wand was available for grabbing
        virtual bool grabWand()    { return EventManager::grabWand( this ); }
		/// @brief Sends all events from the specified tracker to this listener
		/// @return true if the tracker was available for grabbing
        virtual bool grabTracker(const string& tname)
                          { return EventManager::grabTracker( tname, this ); }
		/// @brief Sends all events of type etype to this listener
		/// @return true if the event type is available for grabing
        virtual bool grabOther(const string& etype)
                          { return EventManager::grabOther( etype, this ); }

		/// @brief Releases a grab by this listener for all mouse events, though it will
		/// still receive the events.
		/// @return true if the mouse was available for releaseing
        virtual bool releaseMouse()   { return EventManager::releaseMouse( this ); }
		/// @brief Releases a grab by this listener for all keyboard events, though it will
		/// still receive the events.
		/// @return true if the keyboard was available for releasing.
        virtual bool releaseKeyboard(){ return EventManager::releaseKeyboard( this ); }
		/// @brief Releases a grab by this listener for all wand events, though it will
		/// still receive the events.
		/// @return true if the wand was available for releasing.
        virtual bool releaseWand()    { return EventManager::releaseWand( this ); }
		/// @brief Releases a grab by this listener for all tracker events of the given
		/// name, though it will still receive the events.
		/// @return true if the specified tracker was available for releasing.
        virtual bool releaseTracker(const string& tname)
                          { return EventManager::releaseTracker( tname, this ); }
		/// @brief Releases a grab by this listener for all specified other events, though
		/// it will still receive the events.
		/// @param etype - a string containing the name of the type of event.
		/// @return true if the specified other event was available for releasing.
        virtual bool releaseOther(const string& etype)
                          { return EventManager::releaseOther( etype, this ); }
    private:

};

#endif
