#include"EventManager.h"
#include<tracer/Tracer.h>
#include<dutil/DTime.h>
#include"devents.h"
#include"EventListener.h"
#include"DKeyboard.h"
#include<dreg/Registry.h>
#include<algorithm>
#include<servicemgr/GServiceManager.h>

EventManager*   EventManager::m_self = NULL;
bool            EventManager::m_setup = false;

//=============================================================================
/// @brief Initializes the event manager.
//=============================================================================
void EventManager::init()
{
    if( m_self ) return;
    GServiceManager::init();
    m_self = new EventManager();
    m_self->setupRegistry();
    m_self->m_mousegrab = NULL;
    m_self->m_keygrab = NULL;
    m_self->m_wandgrab = NULL;
    m_self->m_mouse = NULL;
    m_self->m_wand  = NULL;
    m_self->m_keyboard = new DKeyboard();
    m_self->m_mousestate = new DMouseEvent( 0 );
    Tracer::stat( 5, "Event Manager Initialized");
}

//=============================================================================
/// @brief Constructor.
//=============================================================================
EventManager::EventManager()
    :GService( "EventManager" )
{
    string priority = 
        Registry::getValue( "/DIVERSE/EventManager/ServicePriority", "7" );
    
    setPriority( atoi( priority.c_str() ), false );
    GServiceManager::addService( this );
}

//=============================================================================
/// @brief Destructor.
//=============================================================================
EventManager::~EventManager()
{
	delete m_self->m_keyboard;
	delete m_self->m_mousestate;
	delete m_self->m_wandstate;
}

//=============================================================================
/// @brief Returns a pointer to the only instane of the event manager
/// @return a pointer to the Event Manager
//=============================================================================
EventManager* EventManager::instance()
{
    return m_self;
}   

//=============================================================================
/// @brief adds the EventListener object as a listener to all events.
/// @param l - a pointer to the event listener
//=============================================================================
void EventManager::addAllListener( EventListener* l )
{
    m_self->m_alllist.push_back(l);
}

//=============================================================================
/// @brief removes the EventListener object as a listener to all events.
/// @param l - a pointer to the event listener
//=============================================================================
void EventManager::removeAllListener( EventListener* l )
{
	vector<EventListener*>::iterator eraser = 
                    find(m_self->m_alllist.begin(), m_self->m_alllist.end(), l);
	if (eraser != m_self->m_alllist.end())
		m_self->m_alllist.erase(eraser);
}

//=============================================================================
/// @brief Removes the EventListener object as a listener to mouse events.
/// @param l - a pointer to the event listener
//=============================================================================
void EventManager::removeMouseListener( EventListener* l )
{
	vector<EventListener*>::iterator eraser = find(m_self->m_mouselist.begin(), m_self->m_mouselist.end(), l);
	if (eraser != m_self->m_mouselist.end())
		m_self->m_mouselist.erase(eraser);
}

//=============================================================================
/// @brief Removes the EventListener object as a listener to keyboard events.
/// @param l - a pointer to the event listener
//=============================================================================
void EventManager::removeKeyListener( EventListener* l )
{
	vector<EventListener*>::iterator eraser = find(m_self->m_keylist.begin(), m_self->m_keylist.end(), l);
	if (eraser != m_self->m_keylist.end())
		m_self->m_keylist.erase(eraser);
}

//=============================================================================
/// @brief Removes the EventListener object as a listener to wand events.
/// @param l - a pointer to the event listener
//=============================================================================
void EventManager::removeWandListener( EventListener* l )
{
	vector<EventListener*>::iterator eraser = find(m_self->m_wandlist.begin(), m_self->m_wandlist.end(), l);
	if (eraser != m_self->m_wandlist.end())
		m_self->m_wandlist.erase(eraser);
}

//=============================================================================
/// @brief Removes the EventListener object as a listener to a named tracker's events.
/// @param tname - the name of the tracker.
/// @param l - a pointer to the event listener.
//=============================================================================
void EventManager::removeTrackListener(const string& tname,  EventListener* l )
{
	map<string, vector<EventListener*> >::iterator itr = m_self->m_tracklist.find(tname);
	vector<EventListener*> Hope2 = itr->second;
	vector<EventListener*>::iterator blah= find(Hope2.begin(), Hope2.end(), l);
	if (blah != Hope2.end())
		Hope2.erase(blah);
}

//=============================================================================
/// @brief Removes the EventListener object as a listener to named other events.
/// @param etype - the type name of the custom event.
/// @param l - a pointer to the event listener
//=============================================================================
void EventManager::removeOtherListener(const string& etype,  EventListener* l )
{
	map<string, vector<EventListener*> >::iterator itr = m_self->m_otherlist.find(etype);
	vector<EventListener*> Hope2 = itr->second;
	vector<EventListener*>::iterator blah= find(Hope2.begin(), Hope2.end(), l);
	if (blah != Hope2.end())
		Hope2.erase(blah);
}

//=============================================================================
/// @brief Adds the listener to the mouse event call back list, does not not 
/// check to see if listener is already registered.
/// @param l - a pointer to the listener
//=============================================================================
void EventManager::addMouseListener( EventListener* l )
{
    m_self->m_mouselist.push_back(l);    
}

//=============================================================================
/// @brief Adds the listener to the keyboard event call back list, does not not 
/// check to see if listener is already registered.
/// @param l - a pointer to the listener
//=============================================================================
void EventManager::addKeyListener( EventListener* l )
{
    m_self->m_keylist.push_back(l);    
}

//=============================================================================
/// @brief Adds the listener to the wand event call back list, does not not 
/// check to see if listener is already registered.
/// @param l - a pointer to the listener
//=============================================================================
void EventManager::addWandListener( EventListener* l )
{
    m_self->m_wandlist.push_back(l);    
}

//=============================================================================
/// @brief Adds the listener to the tracker event callback list for the given
/// tracker name.
/// @param tname - the name of the tracker
/// @param l - a pointer to the listener
//=============================================================================
void EventManager::addTrackListener( const string& tname, EventListener* l )
{
    m_self->m_tracklist[tname].push_back(l);
}

//=============================================================================
/// @brief Registers to receive other event types
/// @param etype - the type() of the event
/// @param l - a pointer to the listener
//=============================================================================
void EventManager::addOtherListener( const string& etype, EventListener* l )
{
    m_self->m_otherlist[etype].push_back(l);
}

//=============================================================================
/// @brief Adds an event to the event queue that will be processed at the next
/// processEvents() call
/// @param evt - the event to add
//=============================================================================
void EventManager::pushEvent( DEvent* evt )
{
    evt->setWhen( DTime::current() );
    m_self->m_equeue.push_back( evt );
    if( evt->type() == "MOUSE" )
    {
        // if we have a previous mouse event, delete it
        if( m_self->m_mouse )
			delete m_self->m_mouse;
        // set the current state
        m_self->m_mouse = (DMouseEvent*)evt->clone();
		// set the current state
        m_self->m_mousestate->setPosition( ((DMouseEvent*)evt)->x(), ((DMouseEvent*)evt)->y() );
		if ( ((DMouseEvent*)evt)->action() == DMouseEvent::PRESS )
		{
			for ( int i=0;i<3;i++ )
			{
				if ( ((DMouseEvent*)evt)->getButton( i ) )
					m_self->m_mousestate->setButton( i, true );
			}
		}
		else if ( ((DMouseEvent*)evt)->action() == DMouseEvent::RELEASE )
		{
			for ( int i=0;i<3;i++ )
			{
				if ( ((DMouseEvent*)evt)->getButton( i ) )
					m_self->m_mousestate->setButton( i, false );
			}
		}
    }
    else if( evt->type() == "KEYBOARD" )
    {
        // track the new keyboard state
        m_self->m_keyboard->handleEvent( (DKeyEvent*)evt );
    }
    else if( evt->type() == "WAND" )
    {
        // if we have a previous wand event, delete it
        if( m_self->m_wand )
			delete m_self->m_wand;
        // set the current wand event
        m_self->m_wand = (DWandEvent*)evt->clone();
		// if we do not have a current state, create one
		if ( !m_self->m_wandstate )
		{
			m_self->m_wandstate = new DWandEvent( 0 );
			m_self->m_wandstate->setNumButtons( ((DWandEvent*)evt)->getNumButtons() );
		}
		m_self->m_wandstate->setJoystickPos( ((DWandEvent*)evt)->getJoystickX(), ((DWandEvent*)evt)->getJoystickY() );
		m_self->m_wandstate->setPosition( ((DWandEvent*)evt)->getPosition() );
		m_self->m_wandstate->setDirection( ((DWandEvent*)evt)->getDirection() );
		vector<bool> buttons = ((DWandEvent*)evt)->getButtons();
		for (unsigned  int i=0;i<buttons.size();i++ )
		{
			if ( buttons[i] )
				m_self->m_wandstate->setButton( i, true );
			else
				m_self->m_wandstate->setButton( i, false);
		}
    }
    else if( evt->type() == "TRACKER" )
    {
        DTrackerEvent* te = (DTrackerEvent*)evt;
        // check to see if we have a prev tracker state
        map<string, DTrackerEvent*>::iterator curstate =
                              m_self->m_tracker.find( te->getName() );
        // if we do, delete it
        if( curstate != m_self->m_tracker.end() )
            delete m_self->m_tracker[te->getName()];
        // set the state for this tracker
        m_self->m_tracker[te->getName()] = (DTrackerEvent*)evt->clone();
//        cout<<"m_tracker["<<te->getName()<<"] = evt->clone();\n";
    }

//    ORIGINAL CODE - JUST IN CASE!!!!!

/*    evt->setWhen( DTime::current() );
    m_self->m_equeue.push_back( evt );
    if( evt->type() == "MOUSE" )
    {
        // if we have a previous state, delete it
        if( m_self->m_mouse ) delete m_self->m_mouse;
        // set the current state
        m_self->m_mouse = (DMouseEvent*)evt->clone();
        m_self->m_mousestate->SetPosition( ((DMouseEvent*)evt)->x(), ((DMouseEvent*)evt)->y() );
        for ( int i=0;i<3;i++ )
        {
            if ( ((DMouseEvent*)evt)->GetButton( i ) )
                m_self->m_mousestate->SetButton( i, !((DMouseEvent*)evt)->GetButton( i ) );
        }
    }
    else if( evt->type() == "KEYBOARD" )
    {
        // track the new keyboard state
        m_self->m_keyboard->handleEvent( (DKeyEvent*)evt );
    }
    else if( evt->type() == "WAND" )
    {
        // if we have a previous wand state, delete it
        if( m_self->m_wand ) delete m_self->m_wand;
        // set the current wand state
        m_self->m_wand = (DWandEvent*)evt->clone();
    }
    else if( evt->type() == "TRACKER" )
    {
        DTrackerEvent* te = (DTrackerEvent*)evt;
        // check to see if we have a prev tracker state
        map<string, DTrackerEvent*>::iterator curstate =
                              m_self->m_tracker.find( te->getName() );
        // if we do, delete it
        if( curstate != m_self->m_tracker.end() )
            delete m_self->m_tracker[te->getName()];
        // set the state for this tracker
        m_self->m_tracker[te->getName()] = (DTrackerEvent*)evt->clone();
//        cout<<"m_tracker["<<te->getName()<<"] = evt->clone();\n";
    }*/
}

//=============================================================================
/// @brief Sends all mouse events to this listener first.
/// @param l - the listener to give the mouse
/// @return true if the mouse was available for grabbing
//=============================================================================
bool EventManager::grabMouse( EventListener* l )
{
    if( m_self->m_mousegrab ) return false;
    m_self->m_mousegrab = l;
    return true;
}

//=============================================================================
/// @brief Sends all keyboard events to this listener first.
/// @param l - the listener to give the keyboad
/// @return true if the key was available for grabbing
//=============================================================================
bool EventManager::grabKeyboard( EventListener* l )
{
    if( m_self->m_keygrab ) return false;
    m_self->m_keygrab = l;
    return true;
}

//=============================================================================
/// @brief Sends all wand events to this listener first.
/// @param l - the listener to give the wand
/// @return true if the wand was available for grabbing
//=============================================================================
bool EventManager::grabWand( EventListener* l )
{
    if( m_self->m_wandgrab ) return false;
    m_self->m_wandgrab = l;
    return true;
}

//=============================================================================
/// @brief Sends all events from the specified tracker to this listener
/// @param tname - the name of the tracker
/// @param l - a pointer to the event listener
/// @return true if the tracker was available for grabbing
//=============================================================================
bool EventManager::grabTracker( const string& tname, EventListener* l )
{
    map<string, EventListener*>::iterator itr = m_self->m_trackgrab.find(tname);
    if( itr != m_self->m_trackgrab.end() ) return false;
    m_self->m_trackgrab[tname] = l;
    return true;
}

//=============================================================================
/// @brief Sends all events of type etype to this listener
/// @param etype - the type of event to grab, can not be mouse, keyboard,
/// tracker or wand
/// @param l - a pointer to the event listener
/// @return true if the event type is available for grabing
//=============================================================================
bool EventManager::grabOther( const string& etype, EventListener* l )
{
    map<string, EventListener*>::iterator itr = m_self->m_othergrab.find(etype);
    if( itr != m_self->m_othergrab.end() ) return false;
    m_self->m_othergrab[etype] = l;
    return true;
}

//=============================================================================
/// @brief Releases a grab by this listener for all mouse events, though it will
/// still receive the events.
/// @param l - the listener to release the mouse
/// @return true if the mouse was available for releaseing
//=============================================================================
bool EventManager::releaseMouse( EventListener* l )
{
    if( m_self->m_mousegrab != l ) return false;
    m_self->m_mousegrab = NULL;
    return true;
}

//=============================================================================
/// @brief Releases a grab by this listener for all keyboard events, though it will
/// still receive the events.
/// @param l - the listener to release the keyboard.
/// @return true if the keyboard was available for releasing.
//=============================================================================
bool EventManager::releaseKeyboard( EventListener* l )
{
    if( m_self->m_keygrab != l ) return false;
    m_self->m_keygrab = NULL;
    return true;
}

//=============================================================================
/// @brief Releases a grab by this listener for all wand events, though it will
/// still receive the events.
/// @param l - the listener to release the wand.
/// @return true if the wand was available for releasing.
//=============================================================================
bool EventManager::releaseWand( EventListener* l )
{
    if( m_self->m_wandgrab != l ) return false;
    m_self->m_wandgrab = NULL;
    return true;
}

//=============================================================================
/// @brief Releases a grab by this listener for all tracker events of the given
/// name, though it will still receive the events.
/// @param tname - the name of the tracker
/// @param l - the listener to release the specified tracker.
/// @return true if the specified tracker was available for releasing.
//=============================================================================
bool EventManager::releaseTracker( const string& tname, EventListener* l )
{
    map<string, EventListener*>::iterator itr = m_self->m_trackgrab.find(tname);
    if( itr == m_self->m_trackgrab.end() ) return false;
    m_self->m_trackgrab.erase( itr );
    return true;
}

//=============================================================================
/// @brief Releases a grab by this listener for all specified other events, though
/// it will still receive the events.
/// @param etype - the type of event to grab, can not be mouse, keyboard,
/// tracker or wand
/// @param l - the listener to release the specified other events.
/// @return true if the specified other event was available for releasing.
//=============================================================================
bool EventManager::releaseOther( const string& etype, EventListener* l )
{
    map<string, EventListener*>::iterator itr = m_self->m_othergrab.find(etype);
    if( itr == m_self->m_othergrab.end() ) return false;
    m_self->m_othergrab.erase( itr );
    return true;
}

//=============================================================================
/// @brief Goes through all events in the queue and calls the appropriate listeners
/// until either the event queue is empty or time expires
/// @param maxtime stop processing events after this amount of time, -1 to 
///         process all events;
//=============================================================================
void EventManager::processEvents( const DTime& maxtime )
{
    // ce = current event
    DEvent* ce = NULL;
    DTime timer = DTime::current();
    timer.start();
    
    // while the queue is not empty and time has not expired
    while( ((maxtime.usec() < 0) || (timer.elapsed() < maxtime)) 
            &&  (m_self->m_equeue.size() > 0) )
    {
        ce = m_self->m_equeue.front();
        for( unsigned int i = 0; i < m_self->m_alllist.size(); i++ )
            m_self->m_alllist[i]->onAnyEvent( ce );
        
        if( ce->type() == "MOUSE" )
        {
            // if we have a previous state, delete it
            if( m_self->m_mouse ) delete m_self->m_mouse;
            // set the current state
            m_self->m_mouse = (DMouseEvent*)ce->clone();
            // call the listeners
            if( m_self->m_mousegrab && !m_self->m_mousegrab->onMouseEvent( (DMouseEvent*)ce ) )
            {
                for( unsigned int x = 0; x < m_self->m_mouselist.size(); x++ )
                {
                    m_self->m_mouselist[x]->onMouseEvent((DMouseEvent*)ce);
                }
            }
            else
            {
                for( unsigned int x = 0; x < m_self->m_mouselist.size(); x++ )
                {
                    m_self->m_mouselist[x]->onMouseEvent((DMouseEvent*)ce);
                }

            }
        }
        else if( ce->type() == "KEYBOARD" )
        {
            // track the new keyboard state
            m_self->m_keyboard->handleEvent( (DKeyEvent*)ce );
            // call the listeners
            if( m_self->m_keygrab && !m_self->m_keygrab->onKeyEvent( (DKeyEvent*)ce ) )
            {
                for( unsigned int x = 0; x < m_self->m_keylist.size(); x++ )
                {
                    m_self->m_keylist[x]->onKeyEvent((DKeyEvent*)ce);
                }
            }
            else
            {
                for( unsigned int x = 0; x < m_self->m_keylist.size(); x++ )
                {
                    m_self->m_keylist[x]->onKeyEvent((DKeyEvent*)ce);
                }
            }
        }
        else if( ce->type() == "WAND" )
        {
            // if we have a previous wand state, delete it
            if( m_self->m_wand ) delete m_self->m_wand;
            // set the current wand state
            m_self->m_wand = (DWandEvent*)ce->clone();
            // call the listeners
            if( m_self->m_wandgrab && !m_self->m_wandgrab->onWandEvent( (DWandEvent*)ce ) )
            {
                for( unsigned int x = 0; x < m_self->m_wandlist.size(); x++ )
                {
                    m_self->m_wandlist[x]->onWandEvent((DWandEvent*)ce);
                }
            }
            else
            {
                for( unsigned int x = 0; x < m_self->m_wandlist.size(); x++ )
                {
                    m_self->m_wandlist[x]->onWandEvent((DWandEvent*)ce);
                }
            }
        }
        else if( ce->type() == "TRACKER" )
        {
            DTrackerEvent* te = (DTrackerEvent*)ce;
            // check to see if we have a prev tracker state
            map<string, DTrackerEvent*>::iterator curstate = 
                                  m_self->m_tracker.find( te->getName() );
            // if we do, delete it
            if( curstate != m_self->m_tracker.end() )
                delete m_self->m_tracker[te->getName()];
            // set the state for this tracker
            m_self->m_tracker[te->getName()] = (DTrackerEvent*)ce->clone();
	    

            // see if anyone has grabbed this tracker
            map<string, EventListener*>::iterator itr =
                                        m_self->m_trackgrab.find(te->getName() );
            if( itr != m_self->m_trackgrab.end() && !itr->second->onTrackerEvent( te ) )
            {
                map<string, vector<EventListener*> >::iterator tlist =
                                        m_self->m_tracklist.find(te->getName() );
                if( tlist != m_self->m_tracklist.end() )
                {
                    for( unsigned int x = 0; x < tlist->second.size(); x++ )
                    {
                        tlist->second[x]->onTrackerEvent((DTrackerEvent*)ce);
                    }
                }
            }
            else
            {
                map<string, vector<EventListener*> >::iterator tlist =
                                        m_self->m_tracklist.find(te->getName());
                if( tlist != m_self->m_tracklist.end() )
                {
                    for( unsigned int x = 0; x < tlist->second.size(); x++ )
                    {
                        tlist->second[x]->onTrackerEvent((DTrackerEvent*)ce);
                    }
                }
            }
        }
        else
        {
            map<string, EventListener*>::iterator itr = 
                                      m_self->m_othergrab.find(ce->type() );
            if( itr != m_self->m_othergrab.end() && !itr->second->onOtherEvent( ce ) )
            {
                map<string, vector<EventListener*> >::iterator tlist =
                                        m_self->m_otherlist.find(ce->type() );
                if( tlist != m_self->m_otherlist.end() )
                {
                    for( unsigned int x = 0; x < tlist->second.size(); x++ )
                    {
                        tlist->second[x]->onOtherEvent(ce);
                    }
                }
                else
                {
                    for( unsigned int x = 0; x < tlist->second.size(); x++ )
                    {
                        tlist->second[x]->onOtherEvent(ce);
                    }
                }

            }
        }
        delete ce;
        m_self->m_equeue.pop_front();
    }
}

//=============================================================================
/// @brief Get the state of a named tracker.
/// @return a pointer to a DTrackerEvent representing the current state of a
/// named tracker.
//=============================================================================
DTrackerEvent* EventManager::getTrackerState( const string& tname )
{
    map<string, DTrackerEvent*>::iterator itr = m_self->m_tracker.find(tname);
    if( itr != m_self->m_tracker.end() )
        return itr->second;
    return NULL;
}

//=============================================================================
/// @brief Set all default registry values here
/// @param cfg - the registry config file to use
//=============================================================================
void EventManager::setupRegistry( const string& cfg )
{
    if( m_setup ) return;
    Registry::init(cfg);
    // add all settings here
    m_setup = true;
}
