#ifndef _EVENTMANAGER_H_
#define _EVENTMANAGER_H_
#include<vector>
#include<list>
#include<map>
#include<string>
#include<dutil/DTime.h>
#include<servicemgr/GService.h>
using namespace std;

#ifdef WIN32
#ifdef EVENTMGR_EXPORTS
#define EVENTMGR_API __declspec(dllexport)
#else
#define EVENTMGR_API __declspec(dllimport)
#endif /* EVENTMGR_EXPORTS */
#else
#define EVENTMGR_API
#endif /* WIN32 */

class EventListener;
class DEvent;
class DMouseEvent;
class DWandEvent;
class DTrackerEvent;
class DKeyboard;

//=============================================================================
/// @class EventManager
/// @brief The EventManager distributes events to other parts of the application.
///
/// It is a static class which can only be instantiated once per application. To
/// instantiate EventManager, call the function Init() which will simply return
/// if the object is already instantiated.
///
/// All of the events managed by the EventManager are defined in the devents.h file.
/// The tracker events are named as there could be additional tracker devices
/// which are implemented by programmers. The default head tracker name is "headtracker".
/// Any custom tracker events should use or be derived from the DTrackerEvent class.
/// The other event type is associated with any custom event type which should be 
/// derived from DEvent which is not a tracker event. These should also be named.
///
/// To receive events, an add listener function should be called for the events
/// that need to be received. The add listener functions are addMouseListener,
/// addKeyListener, addWandListener, addTrackListener, and addOtherListener.
///
/// To stop listening for those events, call the associated remove listener
/// functions. The remove listener functions are removeMouseListener,
/// removeKeyListener, removeWandListener, removeTrackListener, and removeOtherListener.
///
/// Events are handled in the preProcess stage of the application. Neither
/// the preProcess function nor the ProcessEvents function should be called
/// directly.
///
/// Use the pushEvent function to push events onto the event stack if you intend
/// to create events yourself. As an example, this is used in the caveDeviceSimulator.
///
/// To grab an event type so that it is called first for an object before another
/// object can see the event, use the grab event functions. The grab event functions
/// are grabMouse, grabKeyboard, grabWand, grabTracker, and grabOther.
///
/// To allow other objects to see an event type after an event type has been grabbed,
/// use the release event functions. The release event functions are releaseMouse,
/// releaseKeyboard, releaseWand, releaseTracker, and releaseOther.
///
/// To get the current state of any event type, use the get state functions. The
/// get state event functions are getKeyboardState, getMouseState, getWandState,
/// and getTrackerState.
//=============================================================================
class EVENTMGR_API EventManager : public GService
{
    public:
        static void init();

        static EventManager* instance();

        static void addMouseListener( EventListener* l );
        static void addKeyListener  ( EventListener* l );
        static void addWandListener ( EventListener* l );
        static void addTrackListener( const string& tname, EventListener* l );
        static void addOtherListener( const string& etype, EventListener* l );
        /// Adds a call back that is called for any event
        static void addAllListener( EventListener* l );

        static void removeMouseListener( EventListener* l );
        static void removeKeyListener  ( EventListener* l );
        static void removeWandListener ( EventListener* l );
        static void removeTrackListener( const string& tname, EventListener* l );
        static void removeOtherListener( const string& etype, EventListener* l );
        /// Removes the callback for any event
        static void removeAllListener( EventListener* l );

        static void pushEvent( DEvent* evt );

		/// @brief Events are handled in the preProcess stage of the application.
		/// This function should not be called directly.
        virtual void preProcess() { processEvents(); }
        static void processEvents(const DTime& maxtime = DTime(-1));

        static bool grabMouse( EventListener* l );
        static bool grabKeyboard( EventListener* l );
        static bool grabWand( EventListener* l );
        static bool grabTracker( const string& tname, EventListener* l );
        static bool grabOther( const string& etype, EventListener* l );

        static bool releaseMouse( EventListener* l );
        static bool releaseKeyboard( EventListener* l );
        static bool releaseWand( EventListener* l );
        static bool releaseTracker( const string& tname, EventListener* l );
        static bool releaseOther( const string& etype, EventListener* l );

		/// @brief Get the state of the keyboard.
		/// @return a pointer to the DKeyboard object representing the current state of
		/// the keyboard.
        static DKeyboard*       getKeyboardState(){return m_self->m_keyboard;};
		/// @brief Get the state of the mouse.
		/// @return a pointer to a DMouseEvent representing the current state of the mouse.
        static DMouseEvent*     getMouseState(){ return m_self->m_mousestate; }
		/// @brief Get the state of the wand.
		/// @return a pointer to a DWandEvent representing the current state of the wand
		/// tracker.
        static DWandEvent*      getWandState() { return m_self->m_wandstate;  } 
        static DTrackerEvent*   getTrackerState(const string& tname);
       

        /// Returns a map of tracker to value information that contains
        /// the state of every tracker for which at least 1 event has
        /// been received
        static map<string, DTrackerEvent*> getTrackers() 
                                                { return m_self->m_tracker; }
        static void setupRegistry( const string& cfgfile = "diverse.xml" );
    private:
        EventManager();
		~EventManager();

        DMouseEvent*                m_mousestate;
        DMouseEvent*                m_mouse;
        DWandEvent*                 m_wandstate;
        DWandEvent*                 m_wand;
        map<string, DTrackerEvent*> m_tracker;
        DKeyboard*                  m_keyboard;
         

        EventListener*              m_mousegrab;
        EventListener*              m_keygrab;
        EventListener*              m_wandgrab;
        map<string, EventListener*> m_othergrab;
        map<string, EventListener*> m_trackgrab;


        vector<EventListener*>               m_mouselist;
        vector<EventListener*>               m_keylist;
        map<string, vector<EventListener*> > m_tracklist;
        vector<EventListener*>               m_wandlist;
        vector<EventListener*>               m_alllist;

        map<string, vector<EventListener*> >    m_otherlist;

        list<DEvent*>                        m_equeue;

        static bool                         m_setup;
        static EventManager*                m_self;
};

#endif
