#include"devents.h"
#include<dutil/byteswap.h>
#ifdef WIN32
#include <winsock2.h>
#else
#include<netinet/in.h>
#endif /* WIN 32 */

//=============================================
// Register the Events w/ the Factory on load
//=============================================
RegisterEvent<DTrackerEvent> dtrackerevent;
RegisterEvent<DWandEvent>    dwandevent;
RegisterEvent<DMouseEvent>   dmouseevent;
RegisterEvent<DKeyEvent>     dkeyevent;

//=======================================================
/// @brief Copy constructor.
/// @param evt - the DEvent being copied.
//=======================================================
DEvent::DEvent( const DEvent& evt )
{
    m_source = evt.m_source;
    m_when   = evt.m_when;
}

//=======================================================
/// @brief returns the packsize of the event.
/// @return the number of bytes required to pack the event.
//=======================================================
int DEvent::packsize()const
{
    int size = 0;
    size += type().size() + 1; // event type
    size += m_source.size() + 1; // source
    size += sizeof(int); // time
    return size; 
}

//=======================================================
/// @brief packs the event into the buffer.
/// @param buffer - the buffer into which the event will
/// be packed. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes packed into the buffer.
//=======================================================
int DEvent::pack( char* buffer )const
{
    char* tmp = buffer;
     // pack the type
    *tmp = char(type().size());
    tmp += sizeof(char);
    memcpy( tmp, type().c_str(), type().size() );
    tmp += type().size();
    
    // pack the source
    *tmp = char(m_source.size());
    tmp += sizeof(char);
    memcpy( tmp, m_source.c_str(), m_source.size() );
    tmp += m_source.size();

    // pack the time
    int usec = htonl(m_when.usec());
    memcpy( tmp, &usec, sizeof(int) );
    tmp += sizeof(int);
    
    return tmp - buffer;
}

//=======================================================
/// @brief unpacks the buffer into the event.
/// @param buffer - the buffer from which the event will
/// be unpacked. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes unpacked from the buffer.
//=======================================================
int DEvent::unpack( char* buffer )
{
    char* tmp = buffer;

    char size = *tmp;
    tmp += sizeof(char);
    m_unpacktype = string( tmp, size );
    tmp += size;
    
    size = *tmp;
    tmp += sizeof(char);
    m_source = string( tmp, size );
    tmp += size;

    int usec;
    memcpy( &usec, tmp, sizeof(usec) );
    usec = ntohl(usec);
    m_when = DTime( usec );
    tmp += sizeof(int);
    
    return tmp - buffer;
}

//=======================================================
/// @brief Copy constructor.
/// @param cpy - the DTrackerEvent being copied.
//=======================================================
DTrackerEvent::DTrackerEvent( const DTrackerEvent& cpy )
    :DEvent((DEvent)cpy)
{
    m_name = cpy.m_name;
    m_pos  = cpy.m_pos;
    m_dir  = cpy.m_dir;
}

//=======================================================
/// @brief clones the DTrackerEvent.
/// @return pointer to a DEvent where the cloned DTrackerEvent
/// resides.
//=======================================================
DEvent* DTrackerEvent::clone()const
{
    return new DTrackerEvent(*this);
}

//=======================================================
/// @brief returns the packsize of the event.
/// @return the number of bytes required to pack the event.
//=======================================================
int DTrackerEvent::packsize()const
{
    int size = DEvent::packsize();

    size += sizeof(char) + m_name.size();
    size += m_pos.packsize();
    size += m_dir.packsize();
    return size;
}

//=======================================================
/// @brief packs the event into the buffer.
/// @param buffer - the buffer into which the event will
/// be packed. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes packed into the buffer.
//=======================================================
int DTrackerEvent::pack( char* buffer )const
{
    char* tmp = buffer;
    tmp += DEvent::pack(tmp);

    // pack name
    *tmp = m_name.size();
    tmp += sizeof(char);
    memcpy( tmp, m_name.c_str(), m_name.size() );
    tmp += m_name.size();

    tmp += m_pos.pack(tmp);
    tmp += m_dir.pack(tmp);

    return tmp - buffer;
}

//=======================================================
/// @brief unpacks the buffer into the event.
/// @param buffer - the buffer from which the event will
/// be unpacked. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes unpacked from the buffer.
//=======================================================
int DTrackerEvent::unpack( char* buffer )
{
    char* tmp = buffer;
    tmp += DEvent::unpack(buffer);
    
    char size = *tmp;
    tmp += sizeof(char);
    m_name = string( tmp, size );
    tmp += size;

    tmp += m_pos.unpack(tmp);
    tmp += m_dir.unpack(tmp);

    return tmp - buffer;
}

//=======================================================
/// @brief Copy constructor.
/// @param evt - the DWandEvent being copied.
//=======================================================
DWandEvent::DWandEvent( const DWandEvent& evt )
    :DTrackerEvent((DTrackerEvent)evt)
{
    m_buttons = evt.m_buttons;
    m_action = evt.m_action;
    m_x = evt.m_x;
    m_y = evt.m_y;
}

//=======================================================
/// @brief Sets the number of buttons available for the wand
/// @param buttonNum - the number of buttons the wand tracker
/// has available.
//=======================================================
void DWandEvent::setNumButtons(int buttonNum)
{
	for (int i=0; i< buttonNum; i++)
	{
		m_buttons.push_back(false);
	}
}

//=======================================================
/// @brief returns the packsize of the event.
/// @return the number of bytes required to pack the event.
//=======================================================
int DWandEvent::packsize()const
{
    int size = DTrackerEvent::packsize();
    size += sizeof(char); // num buttons
    size += m_buttons.size() * sizeof(char); // buttons
    size += 2 * sizeof(float); //m_x & m_y
    size += sizeof(m_action);
    return size;
}

//=======================================================
/// @brief packs the event into the buffer.
/// @param buffer - the buffer into which the event will
/// be packed. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes packed into the buffer.
//=======================================================
int DWandEvent::pack( char* buffer )const
{
    char* tmp = buffer;
    tmp += DTrackerEvent::pack(tmp);
    
    *tmp = char(m_buttons.size()); tmp++;
    for( unsigned int i = 0; i < m_buttons.size(); i++ )
        *tmp = char(m_buttons[i]); tmp++;

    float tfloat = m_x;
    host2std(&tfloat, sizeof(tfloat) );
    memcpy( tmp, &tfloat, sizeof(m_x) );
    tmp += sizeof(m_x);
    tfloat = m_y;
    host2std(&tfloat, sizeof(tfloat) );
    memcpy( tmp, &tfloat, sizeof(m_y) );
    tmp += sizeof(m_y);

    int act = htonl(m_action);
    memcpy( tmp, &act, sizeof(int) );
    tmp += sizeof(int);

    return tmp - buffer;
}

//=======================================================
/// @brief unpacks the buffer into the event.
/// @param buffer - the buffer from which the event will
/// be unpacked. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes unpacked from the buffer.
//=======================================================
int DWandEvent::unpack( char* buffer )
{
    char* tmp = buffer;
    tmp += DTrackerEvent::unpack(tmp);

    m_buttons.clear();
    char size = *tmp; tmp++;
    for( int i = 0; i < size; i++ )
    {
        m_buttons.push_back( bool(*tmp) ); 
        tmp++;
    }
    
    memcpy( &m_x, tmp, sizeof(m_x) );
    tmp += sizeof(m_x);
    memcpy( &m_y, tmp, sizeof(m_y) );
    tmp += sizeof(m_y);
    
    memcpy( &m_action, tmp, sizeof(m_action) );
    m_action = ntohl(m_action);
    tmp += sizeof(m_action);
    return tmp - buffer;
}

//=======================================================
/// @brief clones the DWandEvent.
/// @return pointer to a DEvent where the cloned DWandEvent
/// resides.
//=======================================================
DEvent* DWandEvent::clone()const
{
    return new DWandEvent(*this);
}

//=======================================================
/// @brief Copy constructor.
/// @param evt - the DMouseEvent being copied.
//=======================================================
DMouseEvent::DMouseEvent( const DMouseEvent& evt )
    :DEvent( (DEvent)evt )
{
    memcpy( m_buttons, evt.m_buttons, 3*sizeof(bool) );
    m_action = evt.m_action;
    m_x      = evt.m_x;
    m_y      = evt.m_y;
}

//=======================================================
/// @brief clones the DMouseEvent.
/// @return pointer to a DEvent where the cloned DMouseEvent
/// resides.
//=======================================================
DEvent* DMouseEvent::clone()const
{
    return new DMouseEvent(*this);
}

//=======================================================
/// @brief returns the packsize of the event.
/// @return the number of bytes required to pack the event.
//=======================================================
int DMouseEvent::packsize()const
{
    int size = DEvent::packsize();
    size += 3 * sizeof(char); // buttons
    size += 2 * sizeof(m_x); //m_x & m_y
    size += sizeof(m_action);
    return size;
}

//=======================================================
/// @brief packs the event into the buffer.
/// @param buffer - the buffer into which the event will
/// be packed. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes packed into the buffer.
//=======================================================
int DMouseEvent::pack( char* buffer )const
{
    char* tmp = buffer;
    tmp += DEvent::pack(tmp);

    *tmp = char(m_buttons[0]); tmp++;
    *tmp = char(m_buttons[1]); tmp++;
    *tmp = char(m_buttons[2]); tmp++;
    
    int tint = htonl(m_x);
    memcpy( tmp, &tint, sizeof(m_x) );
    tmp += sizeof(m_x);

    tint = htonl(m_y);
    memcpy( tmp, &tint, sizeof(m_y) );
    tmp += sizeof(m_y);

    tint = htonl(m_action);
    memcpy( tmp, &tint, sizeof(m_y) );
    tmp += sizeof(int);

    return tmp - buffer;
}

//=======================================================
/// @brief unpacks the buffer into the event.
/// @param buffer - the buffer from which the event will
/// be unpacked. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes unpacked from the buffer.
//=======================================================
int DMouseEvent::unpack( char* buffer )
{
    char* tmp = buffer;
    tmp += DEvent::unpack(tmp);

    m_buttons[0] = bool(*tmp); tmp++;
    m_buttons[1] = bool(*tmp); tmp++;
    m_buttons[2] = bool(*tmp); tmp++;

    memcpy( &m_x, tmp, sizeof(m_x) );
    m_x = ntohl(m_x);
    tmp += sizeof(m_x);
    memcpy( &m_y, tmp, sizeof(m_y) );
    m_y = ntohl(m_y);
    tmp += sizeof(m_y);

    memcpy( &m_action, tmp, sizeof(m_action) );
    m_action = ntohl(m_action);
    tmp += sizeof(m_action);

    return tmp - buffer;
}

//=======================================================
/// @brief Copy constructor.
/// @param evt - the DKeyEvent being copied.
//=======================================================
DKeyEvent::DKeyEvent( const DKeyEvent& evt )
    :DEvent( (DEvent)evt )
{
    m_action = evt.m_action;
    m_ascii = evt.m_ascii;
    m_key = evt.m_key;
}

//=======================================================
/// @brief clones the DKeyEvent.
/// @return pointer to a DEvent where the cloned DKeyEvent
/// resides.
//=======================================================
DEvent* DKeyEvent::clone()const
{
    return new DKeyEvent(*this);
}

//=======================================================
/// @brief returns the packsize of the event.
/// @return the number of bytes required to pack the event.
//=======================================================
int DKeyEvent::packsize()const
{
    return DEvent::packsize() + 3*sizeof(int);
}

//=======================================================
/// @brief packs the event into the buffer.
/// @param buffer - the buffer into which the event will
/// be packed. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes packed into the buffer.
//=======================================================
int DKeyEvent::pack( char* buffer )const
{
    char* tmp = buffer;
    tmp += DEvent::pack(tmp);

    int tint = htonl(m_ascii);
    memcpy( tmp, &tint, sizeof(int) );
    tmp += sizeof(int);
    
    tint = m_key;
    memcpy( tmp, &tint, sizeof(int) );
    tmp += sizeof(int);
    
    tint = m_action;
    memcpy( tmp, &tint, sizeof(int) );
    tmp += sizeof(int);

    return tmp - buffer;
}

//=======================================================
/// @brief unpacks the buffer into the event.
/// @param buffer - the buffer from which the event will
/// be unpacked. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes unpacked from the buffer.
//=======================================================
int DKeyEvent::unpack( char* buffer )
{
    char* tmp = buffer;
    tmp += DEvent::unpack(tmp);

    memcpy( &m_ascii, tmp, sizeof(int) );
    m_ascii = ntohl(m_ascii);
    tmp += sizeof(int);
    
    memcpy( &m_key, tmp, sizeof(int) );
    m_key = ntohl(m_key);
    tmp += sizeof(int);
    
    memcpy( &m_action, tmp, sizeof(int) );
    m_action = ntohl(m_action);
    tmp += sizeof(int);

    return tmp - buffer;
}
