//=============================================================================
// DEvent.h
// This file contains the class definitions for all DGL events.  These events
// are then passed to DGL displays and augments.
//
// Created 5/25/03 By Daniel Larimer
//=============================================================================
#ifndef _DEVENT_H_
#define _DEVENT_H_
#include<string>
#include<vector>
#include<dutil/DTime.h>
#include<tracer/Tracer.h>
#include<dmath/dVec3.h>
#include<eventmgr/EventFactory.h>
using namespace std;

#ifdef WIN32
#ifdef EVENTMGR_EXPORTS
#define EVENTMGR_API __declspec(dllexport)
#else
#define EVENTMGR_API __declspec(dllimport)
#endif /* EVENTMGR_EXPORTS */
#else
#define EVENTMGR_API
#endif /* WIN32 */

//==========================================================================
/// @class DEvent
/// @brief DEvent is the base class for all events used in DIVERSE.
///
/// The DEvent class contains utility functions for events. The events can be
/// named so that the source is known and applications or libraries which both
/// listen and post events can ignore the events they generated.
//==========================================================================
class EVENTMGR_API DEvent
{
    public:
		/// @brief Constructor
        DEvent(){ m_source = "local"; };
		/// @brief Copy constructor
        DEvent( const DEvent& evt);
        
        /// @brief Used to determine the type of event when it is unpacked.
		/// @return A string containing the type of event.
        virtual string type()const { return m_unpacktype; }
		/// @brief Destructor
        virtual ~DEvent(){}

		/// @brief clones the DEvent.
		/// @return pointer to a DEvent where the cloned DEvent
		/// resides.
        virtual DEvent* clone()const { return new DEvent(*this); }
        virtual int     packsize()const;
        virtual int     pack( char* buffer )const;
        virtual int     unpack( char* buffer );

		/// @brief Sets the name used to identify the source of the event.
		/// @param s - a string containing the name of the source.
        void setSource( const string& s ) { m_source = s;     }
		/// @brief Sets the time the event occurred.
		/// @param t - a DTime object containing the time of the event.
        void setWhen( const DTime& t )    { m_when = t;       }

		/// @brief Returns the time the event occurred.
        /// @return the time the event occured.
        DTime  when()const      { return m_when; }

		/// @brief Returns the name of the source of the event.
        /// @return a string containing the source of the event.
        string getSource()const { return m_source; }
        
    private:
        string m_unpacktype;
        string m_source;
        DTime  m_when;
};

//==========================================================================
/// @class DTrackerEvent
/// @brief Encapsulates the properties used for tracker events.
///
/// Tracker events contain the position and orientation of a device. Tracker
/// events are spawned when the tracker moves from its current position.
///
/// Each DTrackerEvent has a name associated with it so multiple trackable
/// devices can be used.
//==========================================================================
class EVENTMGR_API DTrackerEvent : public DEvent
{
    public:
		/// @brief Constructor
        /// @param name - the name of the tracker, head, wand etc..
        /// @param pos - the position of the tracker
        /// @param dir - the direction in heading, pitch, rool
        DTrackerEvent( const string& name, 
                       const dVec3& pos, const dVec3& dir)
        {
            m_name = name;
            m_pos  = pos;
            m_dir  = dir;
        }
		/// @brief Constructor
        DTrackerEvent(){}
        DTrackerEvent( const DTrackerEvent& cpy);
        DEvent* clone()const;

        /// @brief Used to determine the type of event.
		/// @return A string containing the type of event.
        string type()const { return "TRACKER"; }
        
        virtual int     packsize()const;
        virtual int     pack( char* buffer )const;
        virtual int     unpack( char* buffer );
        
		/// @brief Get the name of the tracker.
        /// @return the name of the tracker generating the event.
        string  getName()const { return m_name; }
		/// @brief get the position of the tracker
        /// @return the position as vector containing x, y, z.
        dVec3 getPosition()const { return m_pos; }
		/// @brief Get the orientation of the tracker.
        /// @return the direction as vector containing heading, pitch, and roll.
        dVec3 getDirection()const{ return m_dir; }
		/// @brief Set the position of the tracker.
		/// @param pos - the new tracker position.
		void setPosition( dVec3 pos ) { m_pos = pos; };
		/// @brief Set the orientation of the tracker.
		/// @param dir - the new tracker orientation.
		void setDirection( dVec3 dir ) { m_dir = dir; };

    protected:
        string   m_name;
        dVec3    m_pos;
        dVec3    m_dir;
};

//==========================================================================
/// @class DWandEvent
/// @brief Encapsulates the properties for a wand tracker device.
///
/// Wand events contain, in addition to DTrackerEvent information, joystick
/// and button information about a wand tracker device.
///
/// There are four actions associated with a DWandEvent. The action
/// determines the meaning of the event.
///
/// Actions:
///
/// BUTTONPRESS   - a joystick button was pressed. The state of the button will
/// be true for the button which was pressed. All other buttons will be false.
/// Only the button information is guaranteed to be valid.
///
/// BUTTONRELEASE - a joystick button was released. The state of the button 
/// will be true for the button which was released. All other buttons will be
/// false. Only the button information is guaranteed to be valid.
///
/// JOYSTICKMOVE  - the joystick was moved from its current position. Only the
/// joystick position is guaranteed to be valid.
///
/// TRACK         - the wand tracker has moved and has a new position and/or
/// orientation. Only the position and orientation are guaranteed to be
/// accurate.
///
//==========================================================================
class EVENTMGR_API DWandEvent : public DTrackerEvent
{
    public:
		/// @brief The action represented by the DWandEvent.
        enum ACTION
		{
			/// - a joystick button was pressed.
			BUTTONPRESS = 1,
			/// - a joystick button was released.
			BUTTONRELEASE = 2,
			/// - the joystick was moved from its last position.
			JOYSTICKMOVE = 4,
			/// - the wand tracker was moved from its last location.
			TRACK = 8
		};
		/// @brief Constructor
        /// @param action - the event action
        /// @param name - the name of the tracker, head, wand etc.
        /// @param pos - the position of the tracker
        /// @param dir - the direction in heading, pitch, roll
        DWandEvent( int action = 0,
                    const string& name = "none",
                    const dVec3 pos = dVec3(),
                    const dVec3 dir = dVec3())
            :DTrackerEvent( name, pos, dir )
        {
            m_x = m_y = 0;
            m_action = action; 
        }
        DWandEvent( const DWandEvent& evt );
        DEvent* clone()const;
        virtual int     packsize()const;
        virtual int     pack( char* buffer )const;
        virtual int     unpack( char* buffer );

        /// @brief Used to determine the type of event.
		/// @return A string containing the type of event.
        string type()const { return "WAND"; }
		/// @brief Get the action associated with the event
        /// @return the action taken, see DWandEvent::ACTION 
        int action()const { return m_action; }
 
		/// @brief Set the joystick position.
		/// @param x - the x coordinate for the new joystick position.
		/// @param y - the y coordinate for the new joystick position.
        void setJoystickPos( float x, float y )
        {
            m_x = x; m_y = y;
        }
		/// @brief Set the state of the button.
        /// @param state - a boolean value determining the state of the button
        /// @param button - The button value to set
		/// for a particular action.
        void setButton( int button, bool state)
        {
			if ((unsigned int)button < m_buttons.size())
				m_buttons[button] = state;
        }

		void setNumButtons(int buttonNum);
		/// @brief Get the number of wand buttons.
		/// @return the number of wand buttons.
		int  getNumButtons(){return m_buttons.size();};

		/// @brief Gets the actual state of the button.
		/// @param button - the ID of the button to be checked.
		/// @return a boolean value determining the state of the button for
		/// a particular action.
		bool getButton( unsigned int button )
		{
			if (button < m_buttons.size())
				return m_buttons[button];
			else
				Tracer::error("Trying to obtain a button that does not exist in a wand event");
			return false;
		}
		/// @brief Get the x position of the joystick.
        /// @return the x position of the joystick.
        float  getJoystickX()const { return m_x; }
		/// @brief Get the y position of the joystick.
        /// @return the y position of the joystick.
        float  getJoystickY()const { return m_y; }

		/// @brief Get a vector containing the state of the buttons for a particular action
        vector<bool> getButtons()const { return m_buttons; }

    private:
        vector<bool> m_buttons;
        int          m_action;
        float        m_x;
        float        m_y;
};

//==========================================================================
/// @class DMouseEvent
/// @brief Encapsulates the properties for mouse devices.
///
/// Mouse events contain a mouse position and button information for a mouse
/// device.
///
/// There are three actions associated with a DMouseEvent. The action
/// determines the meaning of the event.
///
/// Action:
///
/// PRESS   - the mouse button was pressed. During a PRESS action, the button
/// state will be true if the button was pressed and caused this event.
///
/// RELEASE - the mouse button was released. During a RELEASE action, the
/// button state will be true if the button was released and caused this event.
///
/// MOVE    - the mouse was moved while a mouse button was pressed. Passive
/// mouse movement is not currently supported.
///
/// Each action has an associated position representing where the action took
/// place.
///
/// You can get the current state of each event type by calling the static 
/// getState functions of the EventManager.
//==========================================================================
class EVENTMGR_API DMouseEvent : public DEvent
{
    public:
		/// @brief The action represented by the DMouseEvent.
        enum ACTION
		{
			/// - the mouse button was pressed.
			PRESS = 1,
			/// - the mouse button was released.
			RELEASE = 2,
			/// - the mouse was moved while a mouse button was pressed.
			MOVE = 4
		};
		/// @brief The mouse button which was eiether pressed or released.
        enum BUTTON
		{
			/// - the left mouse button.
			LEFT,
			/// - the middle mouse button or scroll wheel.
			MIDDLE,
			/// - the right mouse button.
			RIGHT
		};
		/// @brief ButtonState is not currently implemented, and may not be.
        enum ButtonState {		// mouse/keyboard state values
        NoButton	= 0x0000,
        LeftButton	= 0x0001,
        RightButton	= 0x0002,
        MidButton	= 0x0004,
        MouseButtonMask = 0x00ff,
        ShiftButton	= 0x0100,
        ControlButton   = 0x0200,
        AltButton	= 0x0400,
        MetaButton	= 0x0800,
        KeyButtonMask	= 0x0fff,
        Keypad		= 0x4000
        };
		/// @brief Constructor
		/// @param action - the action to be associated with the DMouseEvent.
		/// Default is PRESSED.
        DMouseEvent( int action = 0)
        {   
            m_x = m_y = 0;
            m_buttons[0] = m_buttons[1] = m_buttons[2] = false;
            m_action = action; 
        }
        DMouseEvent( const DMouseEvent& evt );
        DEvent* clone()const;
        virtual int     packsize()const;
        virtual int     pack( char* buffer )const;
        virtual int     unpack( char* buffer );
        
        /// @brief Used to determine the type of event.
		/// @return A string containing the type of event.
        string type()const   { return "MOUSE";    }
		/// @brief Get the action associated with the event
        /// @return the action taken, see DWandEvent::ACTION 
        int action()const { return m_action; }

		/// @brief Set the position of the mouse event.
		/// @param x - the x coordinate of the mouse position.
		/// @param y - the y coordinate of the mouse position.
        void setPosition( int x, int y )
        {
            m_x = x; m_y = y;
        }
		/// @brief Sets the state for the button
		/// @param button - an int representing the button whose state is to
		/// be changed.
		/// @param state - a bool representing the state of the button
		/// relative to the action.
        void setButton( int button, bool state)
        {
            m_buttons[button] = state;
        }
		/// @brief get the state of the button for a particular action.
		/// @param button - the ID of the button for which to check the state.
		/// @return the state of the button for the particular action.
        bool getButton( int button ){ return m_buttons[button]; }
		/// @brief Get the x coordinate of the mouse position.
		/// @return the x coordinate of the mouse position.
        int  x()const { return m_x; }
		/// @brief Get the y coordinate of the mouse position.
		/// @return the y coordinate of the mouse position.
        int  y()const { return m_y; }

    private:
        bool    m_buttons[3];
        int     m_action;
        int     m_x;
        int     m_y;
};

//==========================================================================
/// @class DKeyEvent
/// @brief Encapsulates properties for keyboard events.
///
/// Keyboard events contain keys that were pressed on the keyboard. Keys
/// pressed which have an associated ascii character code will contain both
/// an ascii code and a Key code. Otherwise the event will conatin only a Key
/// code and the ascii code will be set to 0.
///
/// There are three actions associated with a DKeyEvent. The action
/// determines the meaning of the event.
///
/// Actions:
///
/// UP     - the key been released ( having been previously pressed ).
///
/// DOWN   - the key has been pressed.
///
/// REPEAT - the key was pressed and has been held down.
//==========================================================================
class EVENTMGR_API DKeyEvent : public DEvent
{
    public:
		/// @brief The action represented by the DKeyEvent.
        enum Action
		{
			/// - the key was pressed and has now been released.
			UP = 0,
			/// - the key has been pressed.
			DOWN = 1,
			/// - the key was pressed and has been held down.
			REPEAT = 2
		};
		/// @brief Constructor
		/// @param action - the action as defined in enum Action.
		/// @param ascii - the ascii character code, if defined for the key.
		/// @param key - the key code as defined in enum Key.
        DKeyEvent( int action = 0, int ascii = 0, int key = 0)
        {
            m_action = action;
            m_ascii = ascii;
            m_key = key;
        };
        DKeyEvent( const DKeyEvent& evt );
        DEvent* clone()const;
        virtual int     packsize()const;
        virtual int     pack( char* buffer )const;
        virtual int     unpack( char* buffer );
        /// @brief Set the key code for the DKeyEvent.
		/// @param key - the key code as defined in enum Key.
        void SetKey  ( int key   ) { m_key   = key;   }
		/// @brief Set the ascii code for the DKeyEvent.
		/// @param ascii - the ascii character code
        void SetAscii( int ascii ) { m_ascii = ascii; }
        /// @brief Used to determine the type of event.
		/// @return A string containing the type of event.
        string type()const   { return "KEYBOARD"; }
		/// @brief Returns the ascii character code associated with the event.
		/// @return an int representing the ascii character code.
        int ascii()const  { return m_ascii;  }
		/// @brief Returns the key code associated with the event.
		/// @return an int representing the key code as defined in enum Key.
        int key()const    { return m_key;    }
		/// @brief Returns the action associated with the event.
		/// @return an int representing the action as defined in enum Action.
        int action()const { return m_action; }

    private:
        int m_ascii;
        int m_key;
        int m_action;
    
    public:
		/// @brief ButtonState is not currently implemented.
		enum ButtonState {				// mouse/keyboard state values
			NoButton	= 0x0000,
			LeftButton	= 0x0001,
			RightButton	= 0x0002,
			MidButton	= 0x0004,
			MouseButtonMask = 0x00ff,
			ShiftButton	= 0x0100,
			ControlButton   = 0x0200,
			AltButton	= 0x0400,
			MetaButton	= 0x0800,
			KeyButtonMask	= 0x0fff,
			Keypad		= 0x4000
		};
		/// @brief The key codes used by DIVERSE. ( Borrowed from Qt )
		enum Key {
			Key_Escape = 0x1000,		// misc keys
			Key_Tab = 0x1001,
			Key_Backtab = 0x1002, Key_BackTab = Key_Backtab,
			Key_Backspace = 0x1003, Key_BackSpace = Key_Backspace,
			Key_Return = 0x1004,
			Key_Enter = 0x1005,
			Key_Insert = 0x1006,
			Key_Delete = 0x1007,
			Key_Pause = 0x1008,
			Key_Print = 0x1009,
			Key_SysReq = 0x100a,
			Key_Home = 0x1010,		// cursor movement
			Key_End = 0x1011,
			Key_Left = 0x1012,
			Key_Up = 0x1013,
			Key_Right = 0x1014,
			Key_Down = 0x1015,
			Key_Prior = 0x1016, Key_PageUp = Key_Prior,
			Key_Next = 0x1017, Key_PageDown = Key_Next,
			Key_Shift = 0x1020,		// modifiers
			Key_Control = 0x1021,
			Key_Meta = 0x1022,
			Key_Alt = 0x1023,
			Key_CapsLock = 0x1024,
			Key_NumLock = 0x1025,
			Key_ScrollLock = 0x1026,
			Key_F1 = 0x1030,		// function keys
			Key_F2 = 0x1031,
			Key_F3 = 0x1032,
			Key_F4 = 0x1033,
			Key_F5 = 0x1034,
			Key_F6 = 0x1035,
			Key_F7 = 0x1036,
			Key_F8 = 0x1037,
			Key_F9 = 0x1038,
			Key_F10 = 0x1039,
			Key_F11 = 0x103a,
			Key_F12 = 0x103b,
			Key_F13 = 0x103c,
			Key_F14 = 0x103d,
			Key_F15 = 0x103e,
			Key_F16 = 0x103f,
			Key_F17 = 0x1040,
			Key_F18 = 0x1041,
			Key_F19 = 0x1042,
			Key_F20 = 0x1043,
			Key_F21 = 0x1044,
			Key_F22 = 0x1045,
			Key_F23 = 0x1046,
			Key_F24 = 0x1047,
			Key_F25 = 0x1048,		// F25 .. F35 only on X11
			Key_F26 = 0x1049,
			Key_F27 = 0x104a,
			Key_F28 = 0x104b,
			Key_F29 = 0x104c,
			Key_F30 = 0x104d,
			Key_F31 = 0x104e,
			Key_F32 = 0x104f,
			Key_F33 = 0x1050,
			Key_F34 = 0x1051,
			Key_F35 = 0x1052,
			Key_Super_L = 0x1053, 		// extra keys
			Key_Super_R = 0x1054,
			Key_Menu = 0x1055,
			Key_Hyper_L = 0x1056,
			Key_Hyper_R = 0x1057,
			Key_Help = 0x1058,
			Key_Direction_L = 0x1059,
			Key_Direction_R = 0x1060,
			Key_Space = 0x20,		// 7 bit printable ASCII
			Key_Any = Key_Space,
			Key_Exclam = 0x21,
			Key_QuoteDbl = 0x22,
			Key_NumberSign = 0x23,
			Key_Dollar = 0x24,
			Key_Percent = 0x25,
			Key_Ampersand = 0x26,
			Key_Apostrophe = 0x27,
			Key_ParenLeft = 0x28,
			Key_ParenRight = 0x29,
			Key_Asterisk = 0x2a,
			Key_Plus = 0x2b,
			Key_Comma = 0x2c,
			Key_Minus = 0x2d,
			Key_Period = 0x2e,
			Key_Slash = 0x2f,
			Key_0 = 0x30,
			Key_1 = 0x31,
			Key_2 = 0x32,
			Key_3 = 0x33,
			Key_4 = 0x34,
			Key_5 = 0x35,
			Key_6 = 0x36,
			Key_7 = 0x37,
			Key_8 = 0x38,
			Key_9 = 0x39,
			Key_Colon = 0x3a,
			Key_Semicolon = 0x3b,
			Key_Less = 0x3c,
			Key_Equal = 0x3d,
			Key_Greater = 0x3e,
			Key_Question = 0x3f,
			Key_At = 0x40,
			Key_A = 0x41,
			Key_B = 0x42,
			Key_C = 0x43,
			Key_D = 0x44,
			Key_E = 0x45,
			Key_F = 0x46,
			Key_G = 0x47,
			Key_H = 0x48,
			Key_I = 0x49,
			Key_J = 0x4a,
			Key_K = 0x4b,
			Key_L = 0x4c,
			Key_M = 0x4d,
			Key_N = 0x4e,
			Key_O = 0x4f,
			Key_P = 0x50,
			Key_Q = 0x51,
			Key_R = 0x52,
			Key_S = 0x53,
			Key_T = 0x54,
			Key_U = 0x55,
			Key_V = 0x56,
			Key_W = 0x57,
			Key_X = 0x58,
			Key_Y = 0x59,
			Key_Z = 0x5a,
			Key_BracketLeft = 0x5b,
			Key_Backslash = 0x5c,
			Key_BracketRight = 0x5d,
			Key_AsciiCircum = 0x5e,
			Key_Underscore = 0x5f,
			Key_QuoteLeft = 0x60,
			Key_BraceLeft = 0x7b,
			Key_Bar = 0x7c,
			Key_BraceRight = 0x7d,
			Key_AsciiTilde = 0x7e,

			// Latin 1 codes adapted from X: keysymdef.h,v 1.21 94/08/28 16:17:06

			Key_nobreakspace = 0x0a0,
			Key_exclamdown = 0x0a1,
			Key_cent = 0x0a2,
			Key_sterling = 0x0a3,
			Key_currency = 0x0a4,
			Key_yen = 0x0a5,
			Key_brokenbar = 0x0a6,
			Key_section = 0x0a7,
			Key_diaeresis = 0x0a8,
			Key_copyright = 0x0a9,
			Key_ordfeminine = 0x0aa,
			Key_guillemotleft = 0x0ab,	// left angle quotation mark
			Key_notsign = 0x0ac,
			Key_hyphen = 0x0ad,
			Key_registered = 0x0ae,
			Key_macron = 0x0af,
			Key_degree = 0x0b0,
			Key_plusminus = 0x0b1,
			Key_twosuperior = 0x0b2,
			Key_threesuperior = 0x0b3,
			Key_acute = 0x0b4,
			Key_mu = 0x0b5,
			Key_paragraph = 0x0b6,
			Key_periodcentered = 0x0b7,
			Key_cedilla = 0x0b8,
			Key_onesuperior = 0x0b9,
			Key_masculine = 0x0ba,
			Key_guillemotright = 0x0bb,	// right angle quotation mark
			Key_onequarter = 0x0bc,
			Key_onehalf = 0x0bd,
			Key_threequarters = 0x0be,
			Key_questiondown = 0x0bf,
			Key_Agrave = 0x0c0,
			Key_Aacute = 0x0c1,
			Key_Acircumflex = 0x0c2,
			Key_Atilde = 0x0c3,
			Key_Adiaeresis = 0x0c4,
			Key_Aring = 0x0c5,
			Key_AE = 0x0c6,
			Key_Ccedilla = 0x0c7,
			Key_Egrave = 0x0c8,
			Key_Eacute = 0x0c9,
			Key_Ecircumflex = 0x0ca,
			Key_Ediaeresis = 0x0cb,
			Key_Igrave = 0x0cc,
			Key_Iacute = 0x0cd,
			Key_Icircumflex = 0x0ce,
			Key_Idiaeresis = 0x0cf,
			Key_ETH = 0x0d0,
			Key_Ntilde = 0x0d1,
			Key_Ograve = 0x0d2,
			Key_Oacute = 0x0d3,
			Key_Ocircumflex = 0x0d4,
			Key_Otilde = 0x0d5,
			Key_Odiaeresis = 0x0d6,
			Key_multiply = 0x0d7,
			Key_Ooblique = 0x0d8,
			Key_Ugrave = 0x0d9,
			Key_Uacute = 0x0da,
			Key_Ucircumflex = 0x0db,
			Key_Udiaeresis = 0x0dc,
			Key_Yacute = 0x0dd,
			Key_THORN = 0x0de,
			Key_ssharp = 0x0df,
			Key_agrave = 0x0e0,
			Key_aacute = 0x0e1,
			Key_acircumflex = 0x0e2,
			Key_atilde = 0x0e3,
			Key_adiaeresis = 0x0e4,
			Key_aring = 0x0e5,
			Key_ae = 0x0e6,
			Key_ccedilla = 0x0e7,
			Key_egrave = 0x0e8,
			Key_eacute = 0x0e9,
			Key_ecircumflex = 0x0ea,
			Key_ediaeresis = 0x0eb,
			Key_igrave = 0x0ec,
			Key_iacute = 0x0ed,
			Key_icircumflex = 0x0ee,
			Key_idiaeresis = 0x0ef,
			Key_eth = 0x0f0,
			Key_ntilde = 0x0f1,
			Key_ograve = 0x0f2,
			Key_oacute = 0x0f3,
			Key_ocircumflex = 0x0f4,
			Key_otilde = 0x0f5,
			Key_odiaeresis = 0x0f6,
			Key_division = 0x0f7,
			Key_oslash = 0x0f8,
			Key_ugrave = 0x0f9,
			Key_uacute = 0x0fa,
			Key_ucircumflex = 0x0fb,
			Key_udiaeresis = 0x0fc,
			Key_yacute = 0x0fd,
			Key_thorn = 0x0fe,
			Key_ydiaeresis = 0x0ff,

			Key_unknown = 0xffff
		};
};
#endif
