#include <tracer/Tracer.h>
#include "FontManager.h"
#include "dglfont.h"

using namespace std;

FontManager* FONTMANAGER_API FontManager::m_fontmgr = NULL;
#ifdef WIN32
map<int, ContextVar<dglfont*> > FontManager::m_fontlist;
#else
map<string, ContextVar<dglfont*> > FontManager::m_fontlist;
#endif /* WIN32 */

//=============================================================================
/// @brief Constructor.
//=============================================================================
FontManager::FontManager()
{
    Tracer::debug( "Font Manager created" );
}

//=============================================================================
/// @brief Destructor.
//=============================================================================
FontManager::~FontManager()
{
#ifdef WIN32
    map<int, ContextVar<dglfont*> >::iterator fontlistitr;
#else
    XCloseDisplay( m_fontmgr->m_display );
    map<string, ContextVar<dglfont*> >::iterator fontlistitr;
#endif /* WIN32 */
    for (fontlistitr = m_fontlist.begin(); fontlistitr != m_fontlist.end(); fontlistitr++)
    {
        map<DGLCONTEXT, dglfont*>::iterator itr;
        for ( itr = fontlistitr->second.begin(); itr != fontlistitr->second.end(); itr++ )
            delete itr->second;
    }
    Tracer::debug( "Font Manager destroyed" );
}

#ifdef WIN32
//=============================================================================
/// @brief (Windows specific) Initializes the FontManager for use by the application.
///
/// Only one FontManager exists per application and can only be created using the
/// init function.
/// @param hWnd  - handle to any window. Default value is NULL
/// if none is provided.
//=============================================================================
void FontManager::init( HWND hWnd )
#else
//=============================================================================
/// @brief (XWindows specific) Initializes the FontManager for use by the application.
///
/// Only one FontManager exists per application and can only be created using the
/// init function.
/// @param dpy  - pointer to a display. Default value is NULL
/// if none is provided.
//=============================================================================
void FontManager::init( Display* dpy )
#endif
{
#ifdef WIN32
    if ( m_fontmgr )
		return;

	m_fontmgr = new FontManager();
	m_fontmgr->m_hwnd = hWnd;
    if ( !*m_fontmgr->m_hwnd )
    {
        Tracer::error( "Font Manager - unable to determine HWND for current window" );
        delete m_fontmgr;
    }
/*	m_fontmgr->m_hdc = GetWindowDC( *m_fontmgr->m_hwnd );
    if ( !*m_fontmgr->m_hdc )
    {
        Tracer::error( "Font Manager - unable to determine HDC for current window" );
        delete m_fontmgr;
    }*/
#else
    if ( m_fontmgr )
        return;

    m_fontmgr = new FontManager();
    if ( !dpy )
        m_fontmgr->m_display = XOpenDisplay( NULL );
    else
        m_fontmgr->m_display = dpy;

    if ( !m_fontmgr->m_display )
    {
        Tracer::error( "Font Manager - unable to connect to display" );
        delete m_fontmgr;
    }
#endif /* WIN32 */

    Tracer::debug( "Font Manager initialized" );
}

#ifdef WIN32
//=============================================================================
/// @brief (Windows specific) Request a bitmap font be created for use by OpenGL.
///
/// Will return a pointer to the font or NULL if the font could not be created.
/// @param stockfont  - enumerated value for the font requested.
/// @param start - the first character requested.
/// @param count - the number of characters requested.
//=============================================================================
dglfont* FontManager::requestFont( int stockfont, unsigned char start, unsigned char count )
{
    map<int, ContextVar<dglfont*> >::iterator fontlistitr = m_fontmgr->m_fontlist.find( stockfont );
    if ( fontlistitr == m_fontmgr->m_fontlist.end() || !fontlistitr->second.get() )
    {
        dglfont* newfont = new dglfont( wglGetCurrentDC(), stockfont, start, count );
        if ( !newfont)
        {
            Tracer::error( "FontManager: dglfont could not be created" );
            return NULL;
        }
        m_fontmgr->m_fontlist[stockfont] = newfont;
        return newfont;
    }
    return fontlistitr->second.get();
}

#else

//=============================================================================
/// @brief (X Windows specific) Request a bitmap font be created for use by OpenGL.
///
/// Will return a pointer to the font or NULL if the font could not be created.
/// @param fontname  - a char pointer to the name of the font
/// requested.
/// @param start - the first character requested.
/// @param count - the number of characters requested.
//=============================================================================
dglfont* FontManager::requestFont( char* fontname, unsigned char start, 
                                                    unsigned char count )
{
    string font = fontname;
    map<string, ContextVar<dglfont*> >::iterator fontlistitr = 
                                        m_fontmgr->m_fontlist.find( fontname );
    if ( fontlistitr == m_fontmgr->m_fontlist.end() || 
                            !fontlistitr->second.get() )
    {
        dglfont* newfont = new dglfont( m_fontmgr->m_display, 
                                        fontname, start, count );
        if ( !newfont)
        {
            Tracer::error( "FontManager: dglfont %s could not be created", 
                            fontname );
            return NULL;
        }
        m_fontmgr->m_fontlist[fontname] = newfont;
        return newfont;
    }
    return fontlistitr->second.get();
}
#endif /* WIN32 */

//=============================================================================
/// @brief Delete the specified font. Not implemented.
/// @return always false.
//=============================================================================
bool FontManager::deleteFont( dglfont* font )
{
    return false;
}
