#include <tracer/Tracer.h>
#ifdef WIN32
#include <windows.h>
#endif /* WIN32 */
#include <GL/gl.h>
#include <GL/glu.h>
#ifndef WIN32
#include <GL/glx.h>
#endif /* !WIN32 */
#include "dglfont.h"

#ifdef WIN32
//=============================================================================
/// @brief Constructor (Windows specific).
/// @param hDC (Windows specific) - handle to the windows drawing context.
/// @param stockfont (Windows specific) - enumerated value for the font requested.
/// @param start - the first character requested.
/// @param count - the number of characters requested.
//=============================================================================
dglfont::dglfont( HDC hDC, int stockfont, unsigned char start, unsigned char count )
#else
//=============================================================================
/// @brief Constructor (X Windows specific).
/// @param dpy (X Windows specific) - a pointer to the display.
/// @param fontname (X Windows specific) - a char pointer to the name of the font
/// requested.
/// @param start - the first character requested.
/// @param count - the number of characters requested.
//=============================================================================
dglfont::dglfont( Display* dpy, char* fontname, unsigned char start, unsigned char count )
#endif /* WIN32 */
{
#ifdef WIN32
	m_stockfont = stockfont;
	switch( stockfont )
	{
	case ANSI_FIXED_FONT:
		m_fontname = "ANSI_FIXED_FONT";
		break;
	case ANSI_VAR_FONT:
		m_fontname = "ANSI_VAR_FONT";
		break;
	case DEVICE_DEFAULT_FONT:
		m_fontname = "DEVICE_DEFAULT_FONT";
		break;
	case DEFAULT_GUI_FONT:
		m_fontname = "DEFAULT_GUI_FONT";
		break;
	case OEM_FIXED_FONT:
		m_fontname = "OEM_FIXED_FONT";
		break;
	case SYSTEM_FONT:
		m_fontname = "SYSTEM_FONT";
		break;
	case SYSTEM_FIXED_FONT:
		m_fontname = "SYSTEM_FIXED_FONT";
		break;
	}
#else
    m_fontname = fontname;
#endif
    m_start = start;
    m_count = count;
#ifdef WIN32
	m_hfont = SelectObject( hDC, GetStockObject( stockfont ) );
	if ( !m_hfont )
    {
        Tracer::error( "dglfont: font - %s could not be loaded - font not created.", m_fontname );
        delete this;
        return;
    }
#else
    m_xfont = XLoadQueryFont( dpy, fontname );
    if ( !m_xfont )
    {
        Tracer::error( "dglfont: font - %s could not be loaded - font not created.", m_fontname );
        delete this;
        return;
    }
#endif /* WIN32 */

    m_fontbase = glGenLists( count );
    if ( !glIsList(m_fontbase) )
    {
        Tracer::error( "dglfont: out of display lists - font not created." );
        delete this;
        return;
    }
#ifdef WIN32
	wglUseFontBitmaps( hDC, start, count, m_fontbase );
	TEXTMETRIC tm;
	GetTextMetrics( hDC, &tm );
    m_columnwidth = tm.tmMaxCharWidth;
    m_rowheight = tm.tmAscent;
#else
    glXUseXFont( m_xfont->fid, start, count, m_fontbase );
    m_columnwidth = m_xfont->max_bounds.rbearing - m_xfont->min_bounds.lbearing;
    m_rowheight = m_xfont->ascent;
#endif /* WIN32 */
    Tracer::debug( "dglfont: font - %s created", m_fontname );
}

//=============================================================================
/// @brief Destructor.
//=============================================================================
dglfont::~dglfont()
{
    Tracer::debug( "dglfont destroyed" );
}

//=============================================================================
/// @brief Displays the input string.
///
/// Displays the input string starting at the current raster position, if it is
/// a valid raster position.
/// @param output - pointer to const char containing the array of characters to
/// be output.
//=============================================================================
void dglfont::dglPrint( const char* output )
{
    glPushAttrib( GL_LIST_BIT );
    glListBase( m_fontbase );
    glCallLists( strlen( output ), GL_UNSIGNED_BYTE, (GLubyte *)output );
    glPopAttrib();
} 

//=============================================================================
/// Displays the input string at the requested raster position using int coordinates,
/// if it is a valid raster position.
/// @param output - pointer to const char containing the array of characters to be
/// output.
/// @param x - the x coordinate of the requested raster position.
/// @param y - the y coordinate of the requested raster position.
/// @param z - the z coordinate of the requested raster position.
//=============================================================================
void dglfont::dglPrintPos( const char* output, int x, int y, int z )
{
    glRasterPos3i( x, y, z );
    dglPrint( output );
}

//=============================================================================
/// Displays the input string at the requested raster position using float coordinates,
/// if it is a valid raster position.
/// \param output - pointer to const char containing the array of characters to be output.
/// \param x - the x coordinate of the requested raster position.
/// \param y - the y coordinate of the requested raster position.
/// \param z - the z coordinate of the requested raster position.
//=============================================================================
void dglfont::dglPrintPos( const char* output, float x, float y, float z )
{
    glRasterPos3f( x, y, z );
    dglPrint( output );
}
