#include"DiverseApp.h"
#include<eventmgr/EventManager.h>
#include<navmgr/NavManager.h>
#include<tracer/Tracer.h>
#include<dreg/Registry.h>
#include<dreg/Entry.h>
#include<servicemgr/GServiceManager.h>
#include<diverseconfig.h>
#include<dutil/CommandArgs.h>
#include<pluginmgr/PluginManager.h>
#include<displaymgr/DisplayManager.h>
#include <dutil/ColonList.h>
using namespace std;

#ifdef WIN32
#define DSEPARATOR ';'
#else
#define DSEPARATOR ':'
#endif

//=============================================================================
// Function Name: DiverseApp
/// @brief Constructor.
///
/// Creates a new Diverse application and initializes the default registry
/// settings and initializes the following modules in the specified order.<br>
/// Tracer - Sets up the debug printing. <br> 
/// EventManager - Sets up event subsystem.<br>
/// GServiceManager - Loads and starts all services.<br>
/// DisplayManager - Loads all displays.<br>
/// @param argc - The number of arguments passed to the DIVERSE application.
/// @param argv - The command line arguments passed to the DIVERSE application.
/// @param cfgfile - The config file to use for registry settings.
/// @param init - Flag for initalizing modules.
//=============================================================================
DiverseApp::DiverseApp( int argc, char** argv, const string& cfgfile, bool init )
{
    CommandArgs::init( argc, argv );
    m_setup = false;
    setupRegistry( cfgfile );
    if( init ) initModules();
    
}

//=============================================================================
/// @brief Sets up the registry using the specified configuration file.
/// @param cfgfile - the configuration file to be used for registry settings.
//=============================================================================
void DiverseApp::setupRegistry( const string& cfgfile )
{
    if( m_setup )
    {
        cout<<"WARNING - DiverseApp::setupRegistry called more than once\n";
        return;
    }
    Registry::init( cfgfile );
    Registry::getValue( "/DIVERSE/NavManager/navigations/MouseNav","yes" );
    Registry::getValue( "/DIVERSE/NavManager/navigations/mouseLookNav","no" );
    Registry::getValue( "/DIVERSE/NavManager/navigations/DtkNav", "no" );
    Registry::getValue( "/DIVERSE/NavManager/navigations/wandJoystickNav","no" );
    Registry::getValue( "/DIVERSE/NavManager/navigations/wandJoystickHPRNav","no" );
    Registry::getValue( "/DIVERSE/NavManager/navigations/wandJoystickRollNav","no" );
    Registry::getValue( "/DIVERSE/NavManager/navigations/wandJoystickPitchNav","no" );
    
    Registry::getValue( "/DIVERSE/GServiceManager/services/EventService", "5" );
    Registry::getValue( "/DIVERSE/GServiceManager/services/NavService", "6" );
    Registry::getValue( "/DIVERSE/GServiceManager/services/DisplayService", "7" );
    Registry::getValue( "/DIVERSE/GServiceManager/services/dtkwt","-4" );
    Registry::getValue( "/DIVERSE/GServiceManager/services/dtkht","-4" );
    if( WITH_QT == "yes" )
    {
        Registry::getValue( "/DIVERSE/GServiceManager/services/QtService", "4" );
    }
	else
	{
        Registry::getValue( "/DIVERSE/GServiceManager/services/QtService", "-4" );
	}
    
    char* plg_env = getenv("DIVERSE_PLUGINS");
	if (plg_env != NULL)
	{
		ColonList* plist = new ColonList(string(plg_env));
		for (int i=0; i< plist->size(); i++)
		{
            cout<<"add item to the registry: "<<plist->getItem(i)<<endl;
            // Add them to the registry to get loaded w/ the other plugins
            Registry::setValue("/DIVERSE/PluginManager/plugins/"+
                                plist->getItem(i), "yes" );
		} 
        Tracer::stat( 5, "Adding environment variable DIVERSE_PLUGINS(%s) to the Registry",
                       plg_env );
	}
    PluginManager::setupRegistry();
    EventManager::setupRegistry();
    NavManager::setupRegistry();
    GServiceManager::setupRegistry();
    DisplayManager::setupRegistry();
    m_setup = true;
}

//=============================================================================
/// @brief Initializes the various DIVERSE managers.
//=============================================================================
void DiverseApp::initModules()
{
    Tracer::init();
    EventManager::init();
    NavManager::init();
    GServiceManager::init();
    DisplayManager::init();
    PluginManager::init();
    Tracer::stat(6, "DIVERSE modules loaded");
}

//=============================================================================
/// @brief Destructor.
//=============================================================================
DiverseApp::~DiverseApp()
{
}

//=============================================================================
// Function Name: config()
/// @brief Must be called before postConifg.  This is where the displays are created.
/// This happens automatically if you are using configApp
//=============================================================================
void DiverseApp::config()
{
    GServiceManager::config();
}

//=============================================================================
// Function Name: preProcess
/// @brief Must be called before process() and is called automatically by
/// processApp.
//=============================================================================
void DiverseApp::preProcess()
{
    GServiceManager::preProcess();
}

//=============================================================================
// Function Name: postProcess
/// @brief Must be called after process() and is called automatically by
/// processApp.
//=============================================================================
void DiverseApp::postProcess()
{
    GServiceManager::postProcess();
}

//=============================================================================
// Function Name: preConfig
/// @brief Must be called before config() and is called automatically by configApp.
//=============================================================================
void DiverseApp::preConfig()
{
    GServiceManager::preConfig();
}

//=============================================================================
// Function Name: postConfig
/// @brief Must be called after config() and is called automatically by
/// configApp.
//=============================================================================
void DiverseApp::postConfig()
{
    GServiceManager::postConfig();
}

//=============================================================================
// Function Name: config
/// @brief Calls all the configuration functions.
//=============================================================================
void DiverseApp::configApp()
{
    GServiceManager::configApp();    
}

//=============================================================================
// Function Name: run
/// @brief Enters the main event loop.
//=============================================================================
void DiverseApp::run()
{
    GServiceManager::run();
    GServiceManager::exit();
    PluginManager::exit();
}

//=============================================================================
// Function Name: frame
/// @brief Calls the following functions on all services in order:
/// <ol>   <li> sync() </li>
///        <li> preProcess() </li>
///        <li> process() </li>
///        <li> postProcess() </li> </ol>
//=============================================================================
void DiverseApp::frame()
{
    GServiceManager::frame();
}

//=============================================================================
// Function Name: isRunning
/// @brief Returns true after start() has been called and until quit() is called.
//=============================================================================
bool DiverseApp::isRunning()
{
    return GServiceManager::isRunning();
}

//=============================================================================
// Function Name: start
/// @brief Sets the service manager's state as running.
//=============================================================================
void DiverseApp::start()
{
    GServiceManager::start();
}

//=============================================================================
// Function Name: quit
/// @brief Sets GServiceManager::isRunning() to false which will cause
/// the service manager to exit at the end of the current frame.
//=============================================================================
void DiverseApp::quit()
{
    GServiceManager::quit();
}

//=============================================================================
/// @brief Override this function to handle keyboard events.
/// @param evt - the keyboard event to be processed.
//=============================================================================
bool DiverseApp::onKeyEvent(DKeyEvent* evt)
{
	return false;
}

//=============================================================================
/// @brief Override this function to handle mouse events.
/// @param evt - the mouse event to be processed.
//=============================================================================
bool DiverseApp::onMouseEvent(DMouseEvent* evt)
{
	return false;
}

//=============================================================================
/// @brief Override this function to handle tracker events.
/// @param evt - the tracker event to be processed.
//=============================================================================
bool DiverseApp::onTrackerEvent(DTrackerEvent* evt)
{
	return false;
}

//=============================================================================
/// @brief Override this function to handle wand events.
/// @param evt - the wand event to be processed.
//=============================================================================
bool DiverseApp::onWandEvent(DWandEvent* evt)
{
	return false;
}
