#ifndef _NAVINTERFACE_H_
#define _NAVINTERFACE_H_
#include<dmath/dMatrix.h>

//=============================================================================
/// @class NavInterface
/// @brief Interface for navigations.
///
/// The NavInterface is the base class for all navigations used in DIVERSE. All
/// navigations must be derived from NavInterface and reimplement all of the
/// virtual functions except getPivot which can be ignored if the navigation
/// does not have a set or easily determined pivot axis.
///
/// Please view MouseNav, MouseLookNav, wandJoystickNav, etc for examples of
/// implementing a custom navigation using DIVERSE.
//=============================================================================
class NavInterface
{
    public:
		/// @brief Return the current speed factor used by the navigation.
		///
		/// IMPORTANT - Must be defined by the derived class.
        /// @return the current speed factor
        virtual float getSpeedFactor()const = 0;
		/// @brief Return the current velocity calculated by the navigation. Must
		/// be defined by the derived class.
        /// @return the current velocity as a vector
		virtual dVec3 getVelocity()const = 0;

		/// @brief Return the current navigation matrix calculated by the navigation.
		///
		/// IMPORTANT - Must be defined by the derived class.
		/// @return the current navigation matrix.
        virtual dMatrix getNavMatrix()const = 0;

		/// @brief Called during the process stage of the application. It is at this
		/// stage that the navigation matrix should be calculated based upon the
		/// input devices or simulators you use.
		///
		/// IMPORTANT - Must be defined by the derived class.
        virtual void process() = 0;
		/// @brief Called to start the navigation. When started, isRunning() should
		/// return true.
		///
		/// IMPORTANT - Must be defined by the derived class.
        virtual void start() = 0;
		/// @brief Called to stop the navigation. When stopped, isRunning() should
		/// return false.
		///
		/// IMPORTANT - Must be defined by the derived class.
        virtual void stop() = 0;

		/// @brief Returns the status of the navigation.
		///
		/// IMPORTANT - Must be defined by the derived class.
        /// @return true if the navigation has been started, false if it is stopped.
        virtual bool isRunning()const = 0;

		/// @brief Set the navigation matrix for the navigation.
		///
		/// IMPORTANT - Must be defined by the derived class.
		/// @param mat - the navigation matrix to be used by this navigation.
        virtual void setNavMatrix( const dMatrix& mat ) = 0;
        /// @brief Sets the speed factor to be used for translational velocity calculations.
		///
		/// IMPORTANT - Must be defined by the derived class.
		/// @param speedfactor - the speed factor to be used by the navigation.
        virtual void setSpeedFactor( float speedfactor ) = 0;
        /// @brief Sets the rotation factor to be used for rate of rotation calculations.
		///
		/// IMPORTANT - Must be defined by the derived class.
		/// @param rotfactor - the rotate factor to be used by the navigation.
        virtual void setRotateFactor( float rotfactor ) = 0;
		/// @brief Returns the pivot axis about which rotations occur. Some navigations may
		/// have more than 1 or infinite pivot axes. It is NOT necessary to implement
		/// this function, which will return a zero vector by default.
		/// @return a vector containing the pivot axis for the navigation
		virtual dVec3 getPivot() { return dVec3(); };

		/// @brief Get the name of the navigation. A unique navigation name should be
		/// returned by this function. If the name is not unique, only one of the
		/// navigations of the same name will load.
		///
		/// IMPORTANT - Must be defined by the derived class.
        /// @return the name the NavManager will use to identify this navigation
        virtual string getNavigationName()const = 0;
    private:
};

#endif
