#include"NavManager.h"
#include<eventmgr/EventManager.h>
#include<eventmgr/devents.h>
#include"NavInterface.h"
#include<dreg/Registry.h>
#include<tracer/Tracer.h>
#include<diverseconfig.h>
#include<pluginmgr/PluginManager.h>
#include<eventmgr/keyConverter.h>
#include<servicemgr/GServiceManager.h>

NavManager*     NavManager::m_self = NULL;
bool            NavManager::m_setup = false;
int             NavManager::m_double_speed_key = 0;
int             NavManager::m_half_speed_key = 0;

//=============================================================================
/// @brief Initializes NavManager. Should be called before any other NavManager
/// function.
//=============================================================================
void NavManager::init()
{
    if( m_self ) return;
    EventManager::init();
    m_self = new NavManager();
    m_self->setupRegistry();
    
    m_self->m_speedfactor = 
            atoi( Registry::getValue( "/DIVERSE/NavManager/defaultspeed" ).c_str() );
    m_self->m_rfactor = atoi( 
        Registry::getValue( "/DIVERSE/NavManager/defaultrfactor" ).c_str() );

    m_double_speed_key = keyConverter::translateStringToKey( 
                            Registry::getValue("/DIVERSE/NavManager/double_speed_key") );
    m_half_speed_key   = keyConverter::translateStringToKey( 
                            Registry::getValue("/DIVERSE/NavManager/half_speed_key") );

    //==================================
    // LOAD navigations from config file
    //==================================
    Entry* loadent = Registry::getEntry( "/DIVERSE/NavManager/navigations" );
    map<string, Entry*> eplugins = loadent->getEntries();
    map<string, Entry*>::iterator itr = eplugins.begin();
    while( itr != eplugins.end() )
    {
        if( itr->second->getValue( "", "yes" ) == "yes" )
            load( itr->first );
        itr++;
    }
    EventManager::addKeyListener( m_self );
    Tracer::stat( 3, "Navigation Manager Initialized");
}

//=============================================================================
/// @brief Constructor. Call init to instantiate the NavManager static object.
//=============================================================================
NavManager::NavManager()
    :GService( "NavManager" )
{
    string priority = 
        Registry::getValue( "/DIVERSE/NavManager/ServicePriority", "8" );
    
    setPriority( atoi( priority.c_str() ), false );
    GServiceManager::addService( this );
}

//=============================================================================
/// @brief Sets up the registry.
/// @param cfg - the name of a registry file.
//=============================================================================
void NavManager::setupRegistry( const string& cfg )
{
    Registry::init(cfg);
    if( m_setup )return;
    Registry::getValue( "/DIVERSE/NavManager/defaultspeed", "2.0", "Multiplied by navigation translation speed" );
    Registry::getValue( "/DIVERSE/NavManager/defaultrfactor", "1.0", "No Units" );
    Registry::getValue("/DIVERSE/NavManager/navigations", "load", 
                       "The list of navigations to load" );
    Registry::getValue("/DIVERSE/NavManager/double_speed_key", "Key_F" );
    Registry::getValue("/DIVERSE/NavManager/half_speed_key", "Key_D" );

    //================================
    // Get the current plugin path
    //================================
    string ppath = Registry::getValue("/DIVERSE/PluginManager/pluginpath" );
    ColonList path(ppath);
    //================================
    // Get the specified service path
    //================================
    string srvpath = Registry::getValue("/DIVERSE/NavManager/navpath", 
                        DEFAULT_NAV_PATH,
                        "A Colon separated path");
    ColonList spath(srvpath);
    //================================
    // Merge Them
    //================================
    for( int i = 0; i < spath.size(); i++ )
        path.addItem(spath[i], true);
    
    //================================
    // Set the new pluginpath
    //================================
    Registry::setValue("/DIVERSE/PluginManager/pluginpath", path.getString() );
    m_setup = true;
}

//=============================================================================
/// @brief Gives the current navigation a chance to update
//=============================================================================
void NavManager::process()
{
    if( !m_self )
        Tracer::error("Call NavManager::init()");
    if( !m_self->m_running ) return;
    if( m_self->m_curnav == m_self->m_navlist.end() ) 
    {
        return;
//      m_self->m_curnav = m_self->m_navlist.begin();
    }
    m_self->m_curnav->second->process();
    m_self->m_navmat =  m_self->m_curnav->second->getNavMatrix() /** dMatrix::rotate( DEG2RAD(-90), 1, 0, 0 ) */;
    m_self->m_speedfactor = m_self->m_curnav->second->getSpeedFactor();
}

//=============================================================================
/// @brief Returns the load status of the specified navigation.
/// @param navname - the name of the navigation
/// @return load state, true if loaded, false if not
//=============================================================================
bool NavManager::isLoaded( const string& navname )
{   
    return ( m_self->m_navlist.find(navname) != m_self->m_navlist.end() );
}

//=============================================================================
/// @brief loads the specified navigation if it exists in the search path.
/// @param navigationname - the name of the navigation.
/// @return true if the navigation was loaded successfully, false if it was not
/// loaded.
//=============================================================================
bool NavManager::load( const string& navigationname )
{
#ifdef DEBUG
    if( !m_self )
        Tracer::error("Call NavManager::init()");
#endif
    PluginManager::loadPlugin( navigationname );
    if( m_self->m_navlist.find(navigationname) != m_self->m_navlist.end() )
        return true;
    
    Tracer::error( "Couldn't load navigation \"%s\"", navigationname.c_str() );
    return false;
}

//=============================================================================
/// @brief unloads the specified navigation if it is loaded.
/// @param navigationname - the name of the navigation.
/// @return true if the navigation was unloaded successfully, else returns false.
//=============================================================================
bool NavManager::unload( const string& navigationname )
{
    map<string, NavInterface*>::iterator itr = m_self->m_navlist.find( navigationname );
    if( itr == m_self->m_navlist.end() )
    {
        Tracer::error( "Unload navigation - unknown navigation %s", navigationname.c_str() );
        return false;
    }
    Tracer::debug( 3, "Attempting to unload navigation: %s", navigationname.c_str() );
    if( PluginManager::unloadPlugin( navigationname ) )
        removeNavigation( navigationname );
    if( m_self->m_navlist.find( navigationname ) == m_self->m_navlist.end() )
    {
        nextNavigation();
        Tracer::debug( 3, "Unloaded %s successfully", navigationname.c_str() );
        return true;
    }
    Tracer::error( "Couldn't unload navigation \"%s\"", navigationname.c_str() );
	return false;
}

//=============================================================================
/// @brief Adds a new navigation to the call list, it doesn't start executing until
/// start("navigationname") is called.
/// @param navigation - the pointer to the navigation to add.
/// @return true if the navigation was added successfully, else returns false.
//=============================================================================
bool NavManager::addNavigation( NavInterface* navigation )
{
#ifdef DEBUG
    if( !m_self )
        Tracer::error("Call NavManager::init()");
#endif
    map<string, NavInterface*>::iterator itr = 
                        m_self->m_navlist.find( navigation->getNavigationName() );
    if( itr != m_self->m_navlist.end() )
    {
        Tracer::error( "Navigation with name \"%s\" already registered.", 
                        itr->first.c_str() );
        return false;
    }
    m_self->m_navlist[navigation->getNavigationName()] = navigation;
    startNavigation();
    //If we are the first navigation then we need to do some initalization
    if (m_self->m_navlist.size() == 1)
    {
        m_self->m_curnav = m_self->m_navlist.begin();
        m_self->m_curnav->second->start();
    }
    Tracer::stat( 2, "Loaded %s navigation.", navigation->getNavigationName().c_str() );
    return true;
}

//=============================================================================
/// @brief Removes the navigation from the list.
/// @param name - the name of the navigation to remove.
/// @return true if the navigation was removed successfully, else returns false.
//=============================================================================
bool NavManager::removeNavigation( const string& name )
{
#ifdef DEBUG
    if( !m_self )
        Tracer::error("Call NavManager::init()");
#endif
    map<string, NavInterface*>::iterator itr = 
                        m_self->m_navlist.find( name );
    if( itr == m_self->m_navlist.end() )
    {
        Tracer::warn("Couldn't remove unknown navigation \"%s\".", name.c_str() );
        return false;
    }
    m_self->m_navlist.erase( itr );
    Tracer::stat( 2, "Removed navigation \"%s\"", name.c_str() );
    return true;
}

//=============================================================================
/// @brief Sets the current navigation based on name, if the navigation isn't
/// loaded then it returns NULL, otherwise a pointer to the new navigation.
/// @param navname - the name of the navigiaton.
/// @return a pointer to the navigation, or NULL if the navigation could not be
/// set.
//=============================================================================
NavInterface* NavManager::setNavigation( const string& navname )
{
    map<string, NavInterface*>::iterator itr
                                = m_self->m_navlist.find(navname);
    if ( itr == m_self->m_navlist.end() )
    {
        Tracer::warn( 1, "NavManager::setNavigation( %s )\n"
                         "\t\tCouldn't find navigation \"%s\"",
                         navname.c_str(), navname.c_str() );
        return NULL;
    }
    if( getCurrentNav() ) getCurrentNav()->stop();
    m_self->m_curnav= itr;

    getCurrentNav()->setNavMatrix( m_self->m_navmat );
    getCurrentNav()->setSpeedFactor( getSpeedFactor() );
    getCurrentNav()->setRotateFactor( getRotateFactor() );
    getCurrentNav()->start();

    Tracer::stat( 2, "Set navigation to: %s", 
                      getCurrentNav()->getNavigationName().c_str() );

    return itr->second;
}

//=============================================================================
/// @brief Get a pointer to the current navigation.
/// @return a pointer to the current navigation.
//=============================================================================
NavInterface* NavManager::getCurrentNav()
{
#ifdef DEBUG
    if( !m_self )
        Tracer::error("Call NavManager::init()");
#endif
    if( m_self->m_curnav == m_self->m_navlist.end() )
        return NULL;
    return m_self->m_curnav->second;
}

//=============================================================================
/// @brief Set the current navigation to be the next navigation in the list.
/// @return a pointer to the new current navigation.
//=============================================================================
NavInterface* NavManager::nextNavigation()
{
#ifdef DEBUG
    if( !m_self )
        Tracer::error("Call NavManager::init()");
#endif

    if (m_self->m_navlist.size() == 0)
    {
        m_self->m_curnav = m_self->m_navlist.begin();
        Tracer::warn("Trying to switch navigations when none are loaded\n");
        return NULL;
    }

    if (m_self->m_navlist.size() == 1)
    {
        m_self->m_curnav = m_self->m_navlist.begin();
        return getCurrentNav();
    }
    
    if( getCurrentNav() )
        getCurrentNav()->stop();
    
    m_self->m_curnav++;
    if( !getCurrentNav() )
        m_self->m_curnav = m_self->m_navlist.begin();

    getCurrentNav()->setNavMatrix( m_self->m_navmat );
    getCurrentNav()->setSpeedFactor( getSpeedFactor() );
    getCurrentNav()->setRotateFactor( getRotateFactor() );
    getCurrentNav()->start();

    Tracer::stat( 2, "Next navigation: %s",
                      getCurrentNav()->getNavigationName().c_str() );
    return getCurrentNav();
}

//=============================================================================
/// @brief Set the current navigation to be the previous navigation in the list.
/// @return a pointer to the new current navigation.
//=============================================================================
NavInterface* NavManager::prevNavigation()
{
#ifdef DEBUG
    if( !m_self )
        Tracer::error("Call NavManager::init()");
#endif
   if( m_self->m_navlist.size() < 2 )
       return getCurrentNav();

   if( getCurrentNav() )
       getCurrentNav()->stop();

   if( m_self->m_curnav == m_self->m_navlist.begin() )
       m_self->m_curnav = m_self->m_navlist.end();
   m_self->m_curnav--;

   if( getCurrentNav() )
   {
        getCurrentNav()->setNavMatrix( m_self->m_navmat );
        getCurrentNav()->setSpeedFactor( getSpeedFactor() );
        getCurrentNav()->setRotateFactor( getRotateFactor() );
        getCurrentNav()->start();
   }
   return getCurrentNav();
}

//=============================================================================
/// @brief Start the navigation manager.
//=============================================================================
void NavManager::startNavigation()
{
#ifdef DEBUG
    if( !m_self )
        Tracer::error("Call NavManager::init()");
#endif
    m_self->m_running = true;
}

//=============================================================================
/// @brief Stop the navigation manager.
//=============================================================================
void NavManager::stopNavigation()
{
#ifdef DEBUG
    if( !m_self )
        Tracer::error("Call NavManager::init()");
#endif
    m_self->m_running = false;
}

//=============================================================================
/// @brief Handle key events so that for 'p', the next navigation is made current.
/// Should not be called directly.
//=============================================================================
bool NavManager::onKeyEvent( DKeyEvent* evt )
{
    if( evt->action() == DKeyEvent::DOWN && evt->ascii() == 'p' )
        nextNavigation();
    return false;
}

//=============================================================================
/// @brief Returns the current navigation transformation matrix.
/// @return the current navigation transformation matrix.
//=============================================================================
dMatrix NavManager::getMatrix()
{
    // Convert to DIVERSE cords, Z up, Y forwad, X right
    return m_self->m_navmat;// * dMatrix::rotate( DEG2RAD(-90), 1, 0, 0 );
}

//=============================================================================
/// @brief Set the navigation transformation matrix.
/// @param mat - the new navigation matrix.
//=============================================================================
void NavManager::setMatrix( const dMatrix& mat )
{
    m_self->m_navmat = mat;// * dMatrix::rotate( DEG2RAD(90), 1, 0, 0 );
    if( getCurrentNav() ) getCurrentNav()->setNavMatrix( m_self->m_navmat );
}

//=============================================================================
/// @brief Get the current speed factor.
/// @return the current speed factor.
//=============================================================================
float NavManager::getSpeedFactor()
{
#ifdef DEBUG
    if( !m_self )
        Tracer::error("Call NavManager::init()");
#endif
    return m_self->m_speedfactor;
}

//=============================================================================
/// @brief The factor by which all rotational speed should be multiplied. If
/// a navigation would normaly rotate at 10 deg per sec it should rotate
/// by 10 * getRotateFactor() deg per sec.
/// @return - the current rotate factor.
//=============================================================================
float NavManager::getRotateFactor( )
{
    return m_self->m_rfactor;
}

//=============================================================================
/// @brief Sets the factor by which all navigations should multiply their 
/// rotational speed.
//=============================================================================
void NavManager::setRotateFactor( float rfactor )
{
    m_self->m_rfactor = rfactor;
    if( getCurrentNav() ) getCurrentNav()->setRotateFactor( rfactor );
}

//=============================================================================
/// @brief Sets the speed factor
/// @param speedfactor - the speed factor
//=============================================================================
void NavManager::setSpeedFactor( float speedfactor )
{
    m_self->m_speedfactor = speedfactor;
    if( getCurrentNav() ) getCurrentNav()->setSpeedFactor( speedfactor );
}

//=============================================================================
/// @brief Get the current velocity as a vector.
/// @return the current velocity as a vector.
//=============================================================================
dVec3 NavManager::getVelocity()
{
    if( getCurrentNav() )
		return getCurrentNav()->getVelocity();
	return dVec3();
}
