#ifndef _NAVMANAGER_H_
#define _NAVMANAGER_H_
#include<map>
#include<string>
#include<dmath/dVec3.h>
#include<dmath/dMatrix.h>
#include<eventmgr/EventListener.h>
#include<servicemgr/GService.h>

#ifdef WIN32
#ifdef NAVMGR_EXPORTS
#define NAVMGR_API __declspec(dllexport)
#else
#define NAVMGR_API __declspec(dllimport)
#endif
#else
#define NAVMGR_API
#endif /* WIN32 */

using namespace std;
class NavInterface;

//=============================================================================
/// @class NavManager
/// @author Daniel Larimer
/// @brief Maintains navigation properties and interacts with navigation plugins.
///
/// NavManager is a static class and there can be only one NavManager object.
/// It maintains the navigation matrix which is used by DIVERSE to
/// determine the current position and orientation within the DIVERSE world.
/// The navigation matrix is typically updated by a navigation as translation
/// or rotation occurs but it can also be set directly using setMatrix. You can
/// get a copy of the matrix using getMatrix. The speed factor is used to determine
/// actual translation speed dependent upon the navigation being used. The rotate
/// factor is used to determine the actual rate of rotation dependent upon the
/// navigation being used. When a navigation becomes current, the current navigation
/// matrix, speed factor, and rotation factor will be set for the new navigation.
///
/// NavManager has several utility functions for manipulating navigations. You can
/// determine if a navigation is loaded using the isLoaded function. Navigations can
/// be loaded and unloaded, When a navigation is first initialized, it should call
/// addNavigation to register itself with NavManager. When the navigation becomes
/// current, startNavigation is called. removeNavigation can be used to remove it
/// from the list of available navigations, though it will remain loaded.
///
/// A global keyboard command was set up for 'p', such that it will toggle to the
/// next navigation if one exists.
//=============================================================================
class NAVMGR_API NavManager : public GService,  public EventListener
{
    public:
        static void init();
        void process();

        static void          setMatrix( const dMatrix& vec );
        static dMatrix       getMatrix();
        
        static void          setSpeedFactor( float speedfactor );
        static float         getSpeedFactor();
        static void          setRotateFactor( float rfactor );
        static float         getRotateFactor();

		static dVec3         getVelocity();

        static NavInterface* getCurrentNav();
        static NavInterface* nextNavigation();
        static NavInterface* prevNavigation();
        static NavInterface* setNavigation( const string& navname );

        static bool isLoaded        ( const string& navigation );
        static bool load            ( const string& navigation );
        static bool unload          ( const string& navigation );
        static bool addNavigation   ( NavInterface* navigation );
        static bool startNavigation ( const string& name );
        static bool removeNavigation( const string& name );

        static void startNavigation();
        static void stopNavigation();

        virtual bool onKeyEvent( DKeyEvent* evt );
        
        static void setupRegistry( const string& cfgfile = "diverse.xml" );
        
    private:
        NavManager();

        static bool                          m_setup;
        static NavManager*                   m_self;
        map<string, NavInterface*>           m_navlist;
        map<string, NavInterface*>::iterator m_curnav;
        bool                                 m_running;
        static int                           m_double_speed_key;
        static int                           m_half_speed_key;

        dMatrix m_navmat;
        float   m_speedfactor;
        float   m_rfactor;
};

#endif
