#include"POMessageFactory.h"
#include"POMessage.h"
#include<tracer/Tracer.h>

map<string, POMessage*>*  POMessageFactory::m_messages = NULL;

//=============================================================================
/// @brief Initializes the POMessageFactory static object.
//=============================================================================
void POMessageFactory::init()
{
    POMessageFactory factory; // make sure it is loaded first
    if( factory.m_messages == NULL ){
        factory.m_messages = new map<string, POMessage*>;
    }
}

//=============================================================================
// Function Name: Register
/// @brief Used to add a new message type to the factory.
///
/// Msg is a pointer to a message that the factory will keep around and use to
/// create new messages as needed.
/// @param msg - the message to use for a factory.
/// @return true if no conflicts, false if type is registered.
//=============================================================================
bool POMessageFactory::Register( POMessage* msg )
{
    Tracer::debug( 5, "POMessageFactory::Register( %s )",
                      msg->messageType().c_str() );
    init();
    if( m_messages->find( msg->messageType() ) != m_messages->end() )
    {   
        Tracer::error( 2, "Message type conflict, \"%s\" is already"
                          "registered with the factory.", 
                          msg->messageType().c_str() );
        return false;
    }
    (*m_messages)[ msg->messageType() ] = msg;
    return true;
}


//=============================================================================
// Function Name: UnRegister
/// @brief Removes a message type from the factory.
///
/// Useful if you are loading and unloading DSOs.
/// @param type - the message type.
/// @return true if message could be removed, false if unknown type.
//=============================================================================
bool POMessageFactory::UnRegister( const string& type )
{
    map<string, POMessage*>::iterator itr = m_messages->find(type);
    if( itr == m_messages->end() )
    {
    //    cerr<<"Couldn't remove "<<type<<" message from factory.\n";
        return false;
    }
    delete itr->second;
    m_messages->erase( itr );
    return true;
}

//=============================================================================
// Function Name: BuildFromBuffer
/// @brief Creates a new message based on a packed buffer.
/// @param buffer - the packed buffer.
/// @return the message.
//=============================================================================
POMessage* POMessageFactory::BuildFromBuffer( char* buffer )
{
    POMessage msg;
    msg.unpack(buffer);
    POMessage* newmsg = Build( msg.getPackedType() );
    newmsg->unpack(buffer);
    return newmsg;
}

//=============================================================================
// Function Name: Build
/// @brief Builds a new message based on type.
/// @param type - the MessageType.
/// @return a copy of the message.
//=============================================================================
POMessage* POMessageFactory::Build( const string& type )
{
    map<string, POMessage*>::iterator itr = m_messages->find(type);
    if( itr == m_messages->end() )
    {
        Tracer::error( 2, "POMessageFactory::Build( %s )\n"
                          "\t\tCouldn\'t create message, unknown type.",
                       type.c_str() );
        return NULL;
    }
    return itr->second->clone(); 
}
