#include"POSmallFileMsg.h"
#include<tracer/Tracer.h>
#include<qfile.h> // Qt is required by pony express anyway

// registers the message when the library is loaded
RegisterMessage<POSmallFileMsg> posmallfilemsg;

//=============================================================================
/// @brief Constructor.
/// @param srcname - the name of the source file.
/// @param destfile - the name of the destination file.
//=============================================================================
POSmallFileMsg::POSmallFileMsg( const string& srcname,
                                const string& destfile)
    :POMessage()
{
    m_filedata = NULL;
    m_filesize = 0;
    loadFile( srcname );
    m_destfile = destfile;
}

//=============================================================================
/// @brief Copy constructor.
/// @param cpy - the POSmallFileMsg object to be copied.
//=============================================================================
POSmallFileMsg::POSmallFileMsg( const POSmallFileMsg& cpy )
    :POMessage( cpy )
{
    m_filesize = cpy.m_filesize;
    m_filename = cpy.m_filename;
    m_destfile = cpy.m_destfile;
    if( m_filesize.val() > 0 )
    {
        m_filedata = new char[m_filesize.val()];
        memcpy( m_filedata, cpy.m_filedata, m_filesize.val() );
    }
    else
        m_filedata = NULL;
}

//=============================================================================
/// @brief Destructor.
//=============================================================================
POSmallFileMsg::~POSmallFileMsg()
{
    if( m_filedata ) delete[] m_filedata;
}

//=============================================================================
/// @brief Clone the POSmallFileMsg.
/// @return a POMessage pointer of the cloned POSmallFileMsg. 
//=============================================================================
POMessage* POSmallFileMsg::clone()const
{
    return (POMessage*) new POSmallFileMsg( *this );
}   

//=============================================================================
/// @brief Loads the specified file into the POSmallFileMsg object.
/// @param fname - a string containing the name of the file to load.
/// @return true if the file was loaded successfully, else false.
//=============================================================================
bool POSmallFileMsg::loadFile( const string& fname )
{
    if( m_filedata ) delete[] m_filedata;
    m_filesize = 0;
    
    if( fname == "" )return false;

    if( !QFile::exists( fname.c_str() ) )
    {
        Tracer::error( "POSmallFileMsg::loadFile( %s )\n"
                       "\t\t- File doesn\'t exist." );
        return false;
    }
    QFile tmpfile( fname.c_str() );
    m_filesize = tmpfile.size();
    m_filedata = new char[m_filesize];
    tmpfile.readBlock( m_filedata, m_filesize );
    return true;
}

//=============================================================================
/// @brief Returns the pack size of the event.
/// @return the number of bytes required to pack the event.
//=============================================================================
int POSmallFileMsg::packsize()const
{
    // always remember to pack the base class data
    int size = POMessage::packsize();
    size += m_filename.packsize(); // pack the file name
    size += m_destfile.packsize();
    size += m_filesize.packsize(); 
    size += m_filesize.val(); // the size of the file
    return size;
}

//=============================================================================
/// @brief Packs the event into the buffer.
/// @param buffer - the buffer into which the event will
/// be packed. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes packed into the buffer.
//=============================================================================
int POSmallFileMsg::pack( char* buffer )const
{
    char* tmp = buffer; // don't change the starting point
    tmp += POMessage::pack(tmp); // pack the base class 
                                 // and increment pointer 
    tmp += m_filename.pack(tmp); // pack the name
    tmp += m_destfile.pack(tmp); // the destination file
    tmp += m_filesize.pack(tmp); // pack the file size
     
    // return the number of bytes packed if filesize is 0
    if( m_filesize.val() <= 0 ) return tmp - buffer;
    
    // copy the file to the packed buffer
    memcpy( tmp, m_filedata, m_filesize.val() );
    tmp += m_filesize.val();
    return tmp - buffer; // tmp - buffer will be what you packed
}

//=============================================================================
/// @brief Unpacks the buffer into the event.
/// @param buffer - the buffer from which the event will
/// be unpacked. This is used for sending and receiving
/// events as messages.
/// @return the number of bytes unpacked from the buffer.
//=============================================================================
int POSmallFileMsg::unpack( char* buffer )
{
    if( m_filedata ) delete[] m_filedata;
    m_filedata = NULL;
    
    char* tmp = buffer;
    tmp += POMessage::unpack(tmp);  // very easy to unpack
    tmp += m_filename.unpack(tmp);  // classes that implement
    tmp += m_destfile.unpack(tmp);
    tmp += m_filesize.unpack(tmp);  // this interface
    
    // if filesize is 0 there will be no file data
    if( m_filesize <= 0 ) return tmp - buffer; 
    
    // create the buffer to store the data
    m_filedata = new char[m_filesize];
    // copy the data
    memcpy( m_filedata, tmp, m_filesize );
    // increment the pointer
    tmp += m_filesize;
    
    return tmp - buffer; // the number of bytes unpacked
}
