#include"PostOffice.h"
#include"POClient.h"
#include"POServer.h"
#include"POBox.h"
#include"POMessage.h"
#include"POBoxListMsg.h"
#include<iostream>
#include<qdatetime.h>
#include<qapplication.h>
#include<tracer/Tracer.h>
#include<dutil/DTime.h>
using namespace std;


map<string, PostOffice*>    PostOffice::m_postoffices;

//=============================================================================
// Function Name: PostOffice
/// @brief Constructor.
/// @param name - the name of the post office.
//=============================================================================
PostOffice::PostOffice( const string& name )
:m_poname(name)
{
    if( m_postoffices.find(name) != m_postoffices.end() )
    {
        Tracer::warn("PostOffice::PostOffice( %s ) has already been created",
                     name.c_str() );
    }
    else
    {
        PostOffice::m_postoffices[name] = this;
    }
}

//=============================================================================
/// @brief Destructor.
//=============================================================================
PostOffice::~PostOffice()
{
    cout<<"POstOffice;:~PostOffice() - CLEAN UP NEEDED\n";
}

//=============================================================================
// Function Name: getPOBox
/// @brief Returns a pointer to the specified POBox.
/// @param name - the name for the POBox.
/// @return a pointer to the new POBox.
//=============================================================================
POBox* PostOffice::getPOBox( const string& name )const
{
    cout<<"PostOffice::getPOBox( "<<name<<" )\n";
    map<string, POBox*>::const_iterator itr = m_localboxes.find(name);
    if( itr == m_localboxes.end() )
    {
        Tracer::error( "PostOffice::getPOBox() - No POBox with the name \"%s\"",
                        name.c_str() );
        return NULL;
    }
    return itr->second;
}

//=============================================================================
// Function Name: registerPOBox
/// @brief Creates a new post office box and updates all connections.
/// @param name - the name for the POBox.
/// @return a pointer to the new POBox.
//=============================================================================
POBox* PostOffice::registerPOBox( const string& name )
{
    map<string, POBox*>::iterator itr = m_localboxes.find(name);
    if( itr != m_localboxes.end() )
    {
        Tracer::error( "POBox with the name \"%s\" already registered",
                        name.c_str() );
        return NULL;
    }
    POBox* box = new POBox( name, this );
    m_localboxes[name] = box;

    POBoxListMsg msg( POBoxListMsg::NEWBOXES );
    msg.setSource( POAddress( m_poname, "PostOffice" ) );
    msg.setDestination( POAddress( "", "PostOffice" ) );
    msg.addBox( name );

    map<string, POClient*>::iterator cl = m_clients.begin();
    while( cl != m_clients.end() )
    {
        cl->second->sendMessage( &msg );        
        cl++;
    }
    emit newPOBox( m_poname, name );
    return box;
}

//=============================================================================
/// @brief Creates a new client POBox.
/// @param cname - the name of the commune server.
/// @param boxname - the name of the POBox to create.
//=============================================================================
void PostOffice::newClientPOBox( const string& cname, const string& boxname )
{
    emit newPOBox( m_poname, cname, boxname );
}

//=============================================================================
// Function Name: ClosePOBox
/// @brief Removes the box from the post office.
///
/// This should only be called by ~POBox() when you delete it.
/// @param boxname - the name of the box to remove.
/// @param dodelete - true if you want the post office box to be deleted as well.
/// @return true if successful, otherwise false.
//=============================================================================
bool PostOffice::closePOBox( const string& boxname, bool dodelete )
{
    map<string, POBox*>::iterator itr = m_localboxes.find(boxname);
    if( itr == m_localboxes.end() )
    {
        Tracer::error("PostOffice::closePOBox()\n"
                      "\t\tPOBox with the name \"%s\" isn't registered", 
                      boxname.c_str() );
        return false;
    }

    POBoxListMsg msg( POBoxListMsg::CLOSEBOXES );
    msg.setSource( POAddress( boxname, "PostOffice" ) );
    msg.setDestination( POAddress( "", "PostOffice" ) );
    msg.addBox( boxname );

    map<string, POClient*>::iterator cl = m_clients.begin();
    while( cl != m_clients.end() )
    {
        cl->second->sendMessage( &msg );        
        cl++;
    }
    if( dodelete ) delete itr->second; 
    m_localboxes.erase( itr );
    return false;
}

//=============================================================================
// Function Name: ConnectClient
/// @brief This function is called when a POServer gets a new connection. The
/// connection is connected by sending it a list of POBoxes in this post office.
/// @param client - the new client.
//=============================================================================
void PostOffice::connectClient( POClient* client )
{
    Tracer::stat(7, "PostOffice::connectClient(...)");
    POBoxListMsg msg( POBoxListMsg::NEWBOXES );
    msg.setSource( POAddress( m_poname, "PostOffice" ) );
    msg.setDestination( POAddress( "", "PostOffice" ) );
    
    map<string, POBox*>::iterator itr = m_localboxes.begin();
    while( itr != m_localboxes.end() )
    {
        msg.addBox( itr->first );
        itr++;
    }

    client->sendMessage( &msg );
    
    //=========================================================
    // We will add the client to our list once it has been
    // connected and we have received the first POBoxListMsg
    // containing the clients name
    //=========================================================
    connect( client, SIGNAL( connectionEstablished( POClient* ) ),
             this,   SLOT  ( addConnectedClient( POClient* ) ) );
    connect( client, SIGNAL( connectionLost( POClient* ) ),
             this, SLOT( closeConnection( POClient* ) ) );
}   

//=============================================================================
// Function Name: AddConnectedClient
/// @brief Takes a connected client and adds it to the client list m_clients.
/// @param client - the client to add
//=============================================================================
void PostOffice::addConnectedClient( POClient* client )
{
    Tracer::debug( 5, "PostOffice::addConnectedClient(...)" );
    map<string, POClient*>::iterator itr = m_clients.find( client->getClientName() );
    if( itr != m_clients.end() )
    {
        Tracer::error("DUPLICATE CLIENT NAME: %s", client->getClientName().c_str() );
//        POErrorMsg error( "DUPLICATE CLIENT NAME" );
//        error.setSource( POAddress( name, "PostOffice" ) );
//        error.setDestination( POAddress( "", "PostOffice" ) );
        return; 
    }
    m_clients[client->getClientName()] = client;
    connect( client, SIGNAL( receivedMessage( const POMessage* ) ),
             this, SLOT( deliverMessage( const POMessage* ) ) );

    emit newClient( m_poname, client->getClientName() );
}

//=============================================================================
// Function Name: SendMessage
/// @brief Identifies where the message needs to go and sends it.
///
/// Assumes that the to/from fields have been filled out by the POBox.
/// @param msg - a pointer to the message
/// @return true on success, false on failure
//=============================================================================
bool PostOffice::sendMessage( const POMessage* msg )
{
    POAddress adr = msg->getDestination();
    if( adr.getPO() == m_poname )
        return deliverMessage( msg );
    
    map<string, POClient*>::iterator itr = m_clients.find( adr.getPO() );
    if( itr == m_clients.end() )
    {
        Tracer::warn( "Post Office \"%s\" not found", adr.getPO().c_str() );
        return false;
    }
    return itr->second->sendMessage( adr, msg );
}

//=============================================================================
// Function Name: sendMessage
/// @brief Sends a message to the specified address.
///
/// The message is sent from the specified pobox in this PostOffice to the
/// specified address and returns true if the destination exists.
/// @param msg - the message to send.
/// @param from - the local POBox name.
/// @param to - the destination address.
/// @return true if destination exists.
//=============================================================================
bool PostOffice::sendMessage( POMessage* msg, const string& from,
                              const POAddress& to )
{
    msg->setSource( POAddress( m_poname, from ) );
    msg->setDestination( to );
    return sendMessage( msg );
}

//=============================================================================
/// @brief Deliver a message.
/// @param msg - the message to be delivered.
/// @return true if successful, else false.
//=============================================================================
bool PostOffice::deliverMessage( const POMessage* msg )
{
    POAddress adr = msg->getDestination();
    map<string, POBox*>::iterator itr = m_localboxes.find( adr.getBox() );
    if( itr == m_localboxes.end() )
    {
        Tracer::error( 5,"POBox %s not found at Post Office %s",
                       adr.getBox().c_str(), adr.getPO().c_str() );
        return false;
    }
    itr->second->addMessage( msg->clone() );
    return true;
}



//=============================================================================
// Function Name: ConnectTo
/// @brief Creates a new client and attempts to connect to the specified server.
/// @param host - the hostname/ip of the server.
/// @param port - the port to use.
/// @param block - true to block, false not to block.
/// @return the name of the remote post office.
//=============================================================================
string PostOffice::connectTo( const string& host, int port, bool block )
{
    Tracer::stat( 5, "PostOffice::connectTo( %s:%d )", host.c_str(), port );
    POClient* client = new POClient( host, port );
    connectClient( client );
    if( !block ) return "NOT BLOCKING";

    DTime t;
    t.start();
    while( t.elapsed() < DTime( 10, 0 ) && 
           client->getState() != POClient::CONNECTED )
    {
        qApp->processEvents();
    }
    if( client->getState() != POClient::CONNECTED )
    {
        Tracer::warn( 3, "PostOffice::connectTo( %s:%d ) timed out", host.c_str(), port );
    }
    return client->getClientName();
}

//=============================================================================
// Function Name: ServicePort
/// @brief Creates a new POServer that listens for connections on the specified
/// port.
/// @param  port - the port to listen on.
/// @return true if successful, else false.
//=============================================================================
bool PostOffice::servicePort( int port )
{
    Tracer::stat( 5, "Opening port %d", port );
    map<int, POServer*>::iterator itr = m_servers.find(port);
    if( itr != m_servers.end() )
    {
        Tracer::warn( 5, "Port %d is already open for connections.", port );
        return true;
    }
    QHostAddress hadr;
    hadr.setAddress("127.0.0.1");
    POServer* server = new POServer( hadr, port );
    if( !server->ok() )
    {
        Tracer::warn("Failed to open port %d", port );
        delete server;
        return false;
    }
    connect( server, SIGNAL( newClient( POClient* ) ),
             this, SLOT( connectClient( POClient* ) ) );
    m_servers[port] = server;
    return true;
}

//=============================================================================
// Function Name: ClosePort
/// @brief Closes the server on a given port.
/// @param port - the port to close.
//=============================================================================
void PostOffice::closePort( int port )
{
    map<int, POServer*>::iterator itr = m_servers.find(port);
    if( itr == m_servers.end() )
    {
        Tracer::warn("PostOffice::closePort( %d ) - Post office wasn't servicing port", port);
        return;
    }
    delete itr->second;
    m_servers.erase( itr );
}

//=============================================================================
// Function Name: CloseConnection
/// @brief Closes a connection and removes it from the client list.
/// @param client - the client to close.
//=============================================================================
void PostOffice::closeConnection( POClient* client )
{
    map<string, POClient*>::iterator itr = 
                m_clients.find( client->getClientName() );
    if( itr == m_clients.end() )
    {
        Tracer::warn(3, "PostOffice::CloseConnection() - Unknown Client" );
        return;
    }
    if( client != itr->second )
    {
        Tracer::error("PostOffice::CloseConnection(%s) - client pointer mismatch\n"
                      "\texpected %d, but found %d which has the name: %s",
                       client->getClientName().c_str(),
                       (int)itr->second, (int)client,
                       itr->first.c_str() );
        exit(1);
        return;
    }
    emit clientClosed( getName(), client->getClientName() );
    delete client;
    m_clients.erase(itr);
}

//=============================================================================
// Function Name: popMessage
/// @brief This function allows you to pop a message from the specified POBox.
/// @param box_name - the name of the POBox to pop the message from.
/// @return a pointer to the oldest message in the box, NULL if no message found.
//=============================================================================
const POMessage* PostOffice::popMessage( const string& box_name )
{
    map<string, POBox*>::iterator itr = m_localboxes.find(box_name);
    if( itr == m_localboxes.end() )
    {
        Tracer::warn( 5, "PostOffice::popMessage( %s ) - Unknown POBox" );
        return NULL;
    }
    return itr->second->popMessage();
}

//=============================================================================
// Function Name: hasMessage
/// @brief Checks a POBox for messages.
/// @param box_name - the name of the box to check.
/// @return true if there are messages, otherwise false.
//=============================================================================
bool PostOffice::hasMessage( const string& box_name )
{
    map<string, POBox*>::iterator itr = m_localboxes.find(box_name);
    if( itr == m_localboxes.end() )
    {
        Tracer::warn( 5, "PostOffice::popMessage( %s ) - Unknown POBox" );
        return false;
    }
    return itr->second->hasMessages();
}

//=============================================================================
// Function Name: getAllPostOffices
/// @brief This function returns a pointer to all post offices running within
/// the current process.
/// @return a map of Post Office Name to PostOffice pointer.
//=============================================================================
map<string, PostOffice*> PostOffice::getAllPostOffices()
{
    return m_postoffices;
}

//=============================================================================
/// @brief Get the PostOffice.
/// @param po - the name of the PostOffice.
/// @return a pointer to the PostOffice.
//=============================================================================
PostOffice* PostOffice::getPostOffice( const string& po )
{
    cout<<"getPostOffice("<<po<<")\n";
    return m_postoffices[po];
}

//=============================================================================
// Function Name: getAllClients
/// @brief Get a map of POClient names to pointers.
/// @return a map of all connected clients, this should not be used by
/// non-api developers.
//=============================================================================
map<string, POClient*> PostOffice::getAllClients()const
{
    return m_clients;
}

//=============================================================================
/// @brief Get the POClient with the specified name.
/// @param cname - the name of the client.
/// @return a pointer to the POClient.
//=============================================================================
POClient* PostOffice::getClient( const string& cname )const
{
    map<string, POClient*>::const_iterator itr = m_clients.find( cname );
    if( itr != m_clients.end() )
        return itr->second;
    Tracer::warn( 5, "PostOffice::getClient( %s ) - Unknown client name", cname.c_str() );
    return NULL;
}
