#ifndef _PLUGIN_H_
#define _PLUGIN_H_
#include<string>
#include<pluginmgr/PluginManager.h>
#include<tracer/Tracer.h>

using namespace std;

#ifdef WIN32
#include <windows.h>
#ifdef PLUGINMGR_EXPORTS
#define PLUGINMGR_API __declspec(dllexport)
#else
#define PLUGINMGR_API __declspec(dllimport)
#endif /* PLUGINMGR_EXPORTS */
#else
#define PLUGINMGR_API
#endif /* WIN32 */

///============================================================================
/// @def PLGHANDLE
/// @brief A handle to the plugin.
///
/// Windows:    HMODULE
/// All others: void*
///============================================================================
#ifdef WIN32
#define PLGHANDLE HMODULE
#else
#define PLGHANDLE void*
#endif

///============================================================================
/// @class Plugin
/// @author Daniel Larimer
/// @brief Base class for plugins, provides a way to dynamically load/unload
/// classes.
///
/// All plugins should be derived from Plugin. The macro REGISTER_PLUGIN(PLUG)
/// should be defined such that PLUG is the name of the derived class to be loaded.
/// The derived class should pass the name of the plugin to the Plugin constructor
/// through its own constructor, and set a description of the plugin using the
/// setDescription function. You may wish to override the default load and unload
/// functions. If so, call Plugin::load or Plugin::unload from your function or
/// the plugin may not function properly.
///============================================================================
class PLUGINMGR_API Plugin
{
    public:
        Plugin( const string& name, const string& des = "None");
		/// @brief Destructor.
        virtual ~Plugin(){};

        virtual void load();
        virtual void unload();
        virtual string getName()const;
        virtual string getDescription()const;
        virtual void   setDescription( const string& desc );
        virtual bool   isLoaded()const;

        void   setFileName( const string& filename );
        string getFileName()const;
        void   setHandle( PLGHANDLE handle );
        PLGHANDLE getHandle()const;

    private:
        string m_name;
        string m_description;
        bool   m_loaded;
		PLGHANDLE m_handle;
        string m_file;
};
#ifdef _DARWIN_
#define REGISTER_PLUGIN(PLUG) \
extern "C" \
{\
    void _init() \
    { \
        PluginManager::registerPlugin( new PLUG() ); \
    } \
    void* loadplugin() \
    { \
        return new PLUG(); \
    } \
}
#elif defined WIN32
#define REGISTER_PLUGIN(PLUG) \
extern "C" __declspec(dllexport) \
    Plugin* loadplugin(void); \
	__declspec(dllexport) \
    Plugin* loadplugin(void) \
	{ \
        return new PLUG(); \
    }
#else
#define REGISTER_PLUGIN(PLUG) \
extern "C" \
{\
    void* loadplugin() \
    { \
        return new PLUG(); \
    } \
}

#endif
#endif
