#include"PluginManager.h"
#include"Plugin.h"
#include<tracer/Tracer.h>
#include<iostream>
#include<fstream>
#include<dreg/Registry.h>
#include<diverseconfig.h>
#include<commander/CommandOut.h>
#include<commander/CommandManager.h>

#ifdef _LINUX_
#include"dlfcn.h"
#endif

#ifdef _DARWIN_
#import<mach-o/dyld.h>
#endif

#ifdef _IRIX64_
#include"dlfcn.h"
#endif


using namespace std;

map<string, Plugin*> PluginManager::m_plugins;
ColonList            PluginManager::m_path;
bool                 PluginManager::m_setup = false;
bool                 PluginManager::m_initialized = false;
PluginManager*       PluginManager::m_instance = NULL;

//=============================================================================
/// @brief Initializes the PluginManager by reading the registry to see what
/// plugins need to be loaded and loads them.
//=============================================================================
void PluginManager::init()
{
    if( m_initialized )return;
    setupRegistry();


    //==================================
    // LOAD navigations from config file
    //==================================
    Entry* loadent = Registry::getEntry( "/DIVERSE/PluginManager/plugins" );
    map<string, Entry*> eplugins = loadent->getEntries();
    map<string, Entry*>::iterator itr = eplugins.begin();
    Tracer::stat( 5, "================= Registry Plugins ===============" );
    Tracer::stat( 5, "Plugin Name\t\tLoaded" );
    Tracer::stat( 5, "==================================================" );
    if( itr == eplugins.end() ) Tracer::stat( 5, "none" );
    while( itr != eplugins.end() )
    {
        if( itr->first.size() > 12 )
            Tracer::stat( 5, "%s\t%s", itr->first.c_str(), 
                                 itr->second->getValue("", "yes" ).c_str() );
        else
            Tracer::stat( 5, "%s\t\t%s", itr->first.c_str(), 
                                 itr->second->getValue("", "yes" ).c_str() );
        if( itr->second->getValue( "", "yes" ) == "yes" )
        {
            loadPlugin( itr->first );
        }
        itr++;
    }
    Tracer::stat( 5, "==================================================" );
    m_initialized = true;
    m_instance = new PluginManager();
}

//=============================================================================
/// @brief Constructor.
//=============================================================================
PluginManager::PluginManager()
{
    CommandManager.addCommandHandler( "pluginmgr", this );
}

//=============================================================================
/// @brief Exits the PluginManager and unloads all loaded plugins.
//=============================================================================
void PluginManager::exit()
{
    map<string, Plugin*>::iterator itr = m_plugins.begin();
    while( itr != m_plugins.end() )
    {
        itr->second->unload();
        itr++;
    }
}

//=============================================================================
/// @brief Sets up the registry for the pluginmgr.
/// @param cfg - the configuration file to be used for the registry settings.
//=============================================================================
void PluginManager::setupRegistry( const string& cfg )
{
    if( m_setup ) return;
    Registry::init(cfg);

    Registry::getValue( "/DIVERSE/PluginManager/pluginpath",
                        DEFAULT_PLUGIN_PATH, 
                        "The colon separated path searched for plugins" );
    Registry::getValue( "/DIVERSE/PluginManager/plugins", "load",
                        "Contains the list of plugins to load" );       
    
    m_setup = true;
}

//=============================================================================
/// @brief Opens a shared library and gets a pointer to the plugin.
///
/// The plugin is not initialized until loadPlugin is called.
/// @param filename - the name of the shared lib file to open.
/// @param out - a pointer to a CommandOut object.
/// @return the plugin name, or "NO FILE" if no plugin was found in the library.
/// On Mac OS X this function returns "SUCCESS" instead of the plugin name due
/// to limitations of dynamic loading on Mac OS X.
//=============================================================================
string PluginManager::openPlugin( const string& filename, CommandOut* out )
{
    unsigned int i;
    Tracer t("PluginManager::openPlugin( "+filename+" )", true);
    string ppath = Registry::getValue("/DIVERSE/PluginManager/pluginpath");
    ColonList path(ppath);
    for( i = 0; int(i) < path.size(); i++ )
    {
        addSearchDirectory( path[i] );
    }
    string plugin = ""; 
    vector<string> m_paths = m_path.getList();
    for( i = 0; i < m_paths.size(); i++ )
    {
         Tracer::debug(8, "Trying path: %s", m_paths[i].c_str() );
         plugin = m_OpenPlugin( m_paths[i] +"/" + filename, out );
         if( plugin != "NO_FILE" )
         {
            return plugin;
         }
    }
    
    if( plugin == "NO_FILE" )
    {
        Tracer::error( "PluginManager::OpenPlugin( %s )\n"
                       "\tcouldn't find %s in search path",
                       filename.c_str(), filename.c_str() );
        Tracer::error( "Search path: " );
        for( unsigned int i = 0; i < m_paths.size(); i++ )
            Tracer::error( "\t\t%s", m_paths[i].c_str() );
        if( out )
        {
            out->print( "PluginManager::OpenPlugin( %s )\n"
                           "\tcouldn't find %s in search path\n",
                           filename.c_str(), filename.c_str() );
            out->print( "Search path: \n" );
            for( unsigned int i = 0; i < m_paths.size(); i++ )
                out->print( "\t\t%s\n", m_paths[i].c_str() );
        }
        return "NO_FILE";
    }
    return "ERROR";
}

//=============================================================================
/// @brief A private function that loads a plugin based on a given file name.
///
/// Once the DSO has been loaded a plugin will register itself w/ the
/// plugin manager because it will impliment a _init() function.
/// @param filename - the name of the DSO file to load.
/// @param out - a pointer to a CommandOut object.
/// @return "NO FILE" on an error and "SUCCESS" when it loads.
//=============================================================================
string PluginManager::m_OpenPlugin( const string& filename, CommandOut* out )
{
    Tracer::debug(8, "PluginManager::m_OpenPlugin(%s)", filename.c_str() );
    
    //===================================
    // Check to see if the file is there
    //===================================
    ifstream in(filename.c_str());
    if( !in.good() )
    {   
        in.close();
        return "NO_FILE";
    }
    in.close();
    //===================================

//==========================================================
// Mac OS X Implimentation
//==========================================================
#ifdef _DARWIN_
    NSObjectFileImage img;
    NSModule          handle;
    NSSymbol          sym;

    int rtn = NSCreateObjectFileImageFromFile( filename.c_str(), &img );
    if( rtn != NSObjectFileImageSuccess )
    {
        Tracer::error( "Could not load bundle file %s", filename.c_str() );
        if( out ) out->print( "Could not load bundle file %s\n", filename.c_str() );
        return "ERROR";
    }
    
    handle  = NSLinkModule( img, filename.c_str(), NSLINKMODULE_OPTION_PRIVATE );
    sym     = NSLookupSymbolInModule( handle, "_loadplugin" );
    if( sym == NULL )
    {
        Tracer::error( "Could not find symbol: loadplugin() in plugin"
                       " file %s.", filename.c_str() );
        if( out )
        {
            out->print( "Could not find symbol: loadplugin() in plugin"
                        " file %s.\n", filename.c_str() );
        }
        return "ERROR";
    }

    Plugin* (*loadplugin)(void);
    Plugin* plug;
    
    loadplugin = (Plugin*(*)())NSAddressOfSymbol(sym);
    plug = loadplugin();
    plug->setHandle( handle );
    plug->setFileName( filename );
    if( m_plugins.find( plug->getName() ) != m_plugins.end() )
    {
        Tracer::error( "Already loaded plugin with name: %s", 
                        plug->getName().c_str() );
        if( out ) out->print( "Already loaded plugin with name: %s\n", 
                               plug->getName().c_str() );
        delete plug;
        return "ERROR";
    }
    Tracer::stat( 5, "Loaded Plugin: %s", plug->getName().c_str() );
    m_plugins[plug->getName()] = plug;
    
    return plug->getName();
#endif

//==============================================================
// Linux & Irix Implimentation
//==============================================================
#ifndef WIN32
#ifndef _DARWIN_
    void* handle;
    handle = dlopen( filename.c_str(), RTLD_NOW );
    if( !handle )
    {
        char* err = dlerror();
        Tracer::error("PluginManager::m_OpenPlugin(%s): \n\t%s", 
                          filename.c_str(), err );
        if( out ) out->print("PluginManager::m_OpenPlugin(%s): \n\t%s\n", 
                              filename.c_str(), err );
        return "ERROR";
    }
    
    char* error;
    Plugin* (*loadplugin)(void);
    loadplugin = (Plugin*(*)())dlsym(handle, "loadplugin");
    if( (error = dlerror()) != NULL )
    {
        Tracer::error("Couldn't find loadplugin() in file: %s\n\t%s",
                       filename.c_str(),
                       error );
        if( out ) out->print("Couldn't find loadplugin() in file: %s\n\t%s\n",
                              filename.c_str(),
                              error );
        dlclose(handle);
        return "ERROR";
    }
    Plugin* plug = (*loadplugin)();
    plug->setHandle( handle );
    plug->setFileName( filename );
    
    string name = plug->getName();
    if( m_plugins.find(name) != m_plugins.end() )
    {
        Tracer::error( "Already loaded plugin with name: %s", 
                        filename.c_str() );
        if( out ) out->print( "Already loaded plugin with name: %s\n", 
                               filename.c_str() );
        delete plug;
        dlclose(handle);
        return name;
    }
    Tracer::debug( 6, "PluginManager::OpenPlugin()\n"
                  "\t\tFound plugin %s in file %s", plug->getName().c_str(),
                  filename.c_str() );
    m_plugins[plug->getName()] = plug;
    return plug->getName();
#endif
#endif

//===========================================================
// WINDOWS implimentation
//===========================================================
#ifdef WIN32
    HMODULE handle;
    LPVOID lpMsgBuf;
    
    handle = LoadLibrary(filename.c_str());
    if( !handle )
    {
        FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
            NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            (LPTSTR) &lpMsgBuf, 0, NULL);
        cerr<<"Failed to open file: "<<filename<<endl;
        cerr<<(char*)lpMsgBuf<<endl;
        LocalFree(lpMsgBuf);
        return "NO FILE";
    }
    Plugin* (*loadplugin)(void);
    loadplugin = (Plugin*(*)())GetProcAddress(handle, "loadplugin");
    if( loadplugin == NULL )
    {
         FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
            NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            (LPTSTR) &lpMsgBuf, 0, NULL);
        cerr<<"no load plugin\n";
        cerr<<(char*)lpMsgBuf<<endl;
        LocalFree(lpMsgBuf);
        FreeLibrary(handle);
        return "NO LOADPLUGIN";
    }
    Plugin* plug = (*loadplugin)();
    plug->setHandle( handle );
    plug->setFileName( filename );
    
    string name = plug->getName();
    if( m_plugins.find(name) != m_plugins.end() )
    {
        Tracer::error( "Already loaded plugin with name: %s", 
                        filename.c_str() );
        delete plug;
        FreeLibrary(handle);
        return name;
    }
    Tracer::debug( 6, "PluginManager::OpenPlugin()\n"
                  "\t\tFound plugin %s in file %s", plug->getName().c_str(),
                  filename.c_str() );
    m_plugins[plug->getName()] = plug;
    return plug->getName();
#endif
    
}

//=============================================================================
/// @brief Add a directory for where to search for plugins.
/// @param path - a string containing the directory path.
//=============================================================================
void PluginManager::addSearchDirectory( const string& path )
{
    m_path.addItem(path, true);
}

//=============================================================================
/// @brief Remove a directory from the search path list.
/// @param path - a string containing the directory path.
//=============================================================================
void PluginManager::removeSearchDirectory( const string& path )
{
    m_path.removeItem(path);
}

//=============================================================================
/// @brief Get a list of search directories.
/// @return a vector containing the list of search directories.
//=============================================================================
vector<string> PluginManager::getSearchDirectories()
{
    return m_path.getList();
}

//=============================================================================
/// @brief Set the path list using a colon separated list.
/// @param path - a string containing a colon separated list of search directories.
/// For Windows, the list should be separated by semicolons.
//=============================================================================
void PluginManager::setPath( const string& path )
{
    m_path = ColonList( path );
}

//=============================================================================
/// @brief Get a string containing the search directories.
/// @return a string containing a colon separated list of the search directories.
/// For Windows, the list should be separated by semicolons.
//=============================================================================
string PluginManager::getPath()
{
    return m_path.getString();
}

//=============================================================================
/// @brief Loads the plugin w/ the given plugin name.
///
/// If no plugin by that name is found then it will try to load pluginname.plug
/// and then load the plugin. Once loaded it returns a pointer to the plugin,
/// or NULL if it couldn't load the plugin.
/// @param pluginname - the name of the plugin, not the file name.
/// @param out - a pointer to a CommandOut object.
/// @return a pointer to the loaded plugin.
//=============================================================================
Plugin* PluginManager::loadPlugin( const string& pluginname, CommandOut* out )
{
    map<string, Plugin*>::iterator itr = m_plugins.find( pluginname );
    if( itr == m_plugins.end() )
    {
        #ifdef WIN32
        #ifdef _DEBUG
        Tracer::debug( 8, "Looking for plugin: %s", (pluginname+".plugd").c_str() );
        #else
        Tracer::debug( 8, "Looking for plugin: %s", (pluginname+".plug").c_str() );
        #endif
        #else
        Tracer::debug( 8, "Looking for plugin: %s", (pluginname+".plug").c_str() );
        #endif
       
        string pname;
        #if defined WIN32
        #ifdef _DEBUG
        pname = openPlugin( pluginname + ".plugd", out );
        #else
        pname = openPlugin( pluginname + ".plug", out );
        #endif
        #else
        pname = openPlugin( pluginname + ".plug", out );
        #endif
        itr = m_plugins.find( pname );
    }
    if( itr == m_plugins.end() )
    {
        Tracer::error("PluginManager::loadPlugin() - Couldn't load %s\n"
                      "\tbecause the plugin didn't register with\n"
                      "\tthe PluginManager", pluginname.c_str());
        if( out )              
           out->print("PluginManager::loadPlugin() - Couldn't load %s\n"
                      "\tbecause the plugin didn't register with\n"
                      "\tthe PluginManager\n", pluginname.c_str());
        return NULL;
    }
    if( itr->second->isLoaded() ) 
    {
        return itr->second;
    }    
    Tracer::debug( 6, "Loading Plugin( %s )", pluginname.c_str() );
    itr->second->load();
    return itr->second;
}

//=============================================================================
/// @brief Unloads a plugin and closes the shared library.
/// @param pluginname - the name of the plugin to close.
/// @return true if the plugin unloaded successfully, else false.
//=============================================================================
bool PluginManager::closePlugin( const string& pluginname )
{
    map<string, Plugin*>::iterator itr = m_plugins.find( pluginname );
    if( itr == m_plugins.end() )
    {
        Tracer::error( "ClosePlugin() - Unknown Plugin: %s", pluginname.c_str() );
        return false;
    }
    if( itr->second->isLoaded() ) itr->second->unload();

#ifdef WIN32
    HMODULE handle = itr->second->getHandle();
    delete itr->second;
    FreeLibrary(handle);
#endif
#ifndef WIN32
#ifndef _DARWIN_
    void* handle = itr->second->getHandle();
    delete itr->second;
    dlclose( handle );
#endif
#endif
#ifdef _DARWIN_
    Tracer::warn( 3, "PluginManager::closePlugin( %s )\n"
                     "\t Doesn't unload on Mac OS X.", pluginname.c_str() );
#endif
    m_plugins.erase( itr );
    Tracer::debug( 3, "Unloaded %s successfully", pluginname.c_str() );
    return true;
}

//=============================================================================
/// @brief Unloads the specified plugin.
/// @param pluginname - the name of the plugin to unload.
/// @return true if plugin was unloaded successfully, else returns false.
//=============================================================================
bool PluginManager::unloadPlugin( const string& pluginname )
{
    map<string, Plugin*>::iterator itr = m_plugins.find( pluginname );
    if( itr == m_plugins.end() )
    {
        Tracer::debug( "UnLoadPlugin() - Unknown Plugin: %s", pluginname.c_str() );
        return false;
    }
    if( itr->second->isLoaded() )
        itr->second->unload();
    return true;
}

//=============================================================================
/// @brief Returns a pointer to the plugin if it exists, whether or not it is 
/// loaded.
/// @param pluginname - the name of the plugin.
/// @return a pointer to the plugin or NULL if it doesn't exist.
//=============================================================================
Plugin* PluginManager::getPlugin( const string& pluginname )
{
    map<string, Plugin*>::iterator itr = m_plugins.find( pluginname );
    if( itr == m_plugins.end() )
    {
        cerr<<"getPlugin() - Unknown Plugin: "<<pluginname<<endl;
        return NULL;
    }
    return itr->second;
}

//=============================================================================
/// @brief Get the list of loaded plugins.
/// @return a map of plugin name to pointer to the plugin.
//=============================================================================
map<string, Plugin*> PluginManager::getPlugins()
{
    return m_plugins;
}

//=============================================================================
/// @brief Adds the plugin to the list of loaded plugins.
/// @param plug - the plugin to load.
//=============================================================================
void PluginManager::registerPlugin( Plugin* plug )
{
    Tracer::debug( "RegisterPlugin( %s )", plug->getName().c_str() );
    m_plugins[plug->getName()] = plug;
}

//=============================================================================
/// @brief Executes a Commander command, allowing Commander to load and unload
/// plugins dynamically.
/// @param cmd - the command to execute.
/// @return a pointer to a CommandOut object.
//=============================================================================
CommandOut* PluginManager::execute( const Command& cmd )
{
    CommandOut* out = new CommandOut( cmd );
    vector<string> args = cmd.getArgs();
    if( args.size() <= 1 )
    {
        out->print( "Usage: \n"
                    "\tpluginmgr load PLUGINNAME ...\tloads 1 or more plugins\n"
                    "\tpluginmgr unload PLUGINNAME ...\tunloads 1 or more plugins \n"
                    "\tpluginmgr open PLUGINFILE ...\topens 1 or more plugin files\n"
                    "\tpluginmgr get path\t\t\treturns the plugin search path\n"
                    "\tpluginmgr get pluginlist\t\ta colon separated list of\n"
                    "\t\t\t\t\t  plugin data with one plugin per line\n");
        return out;
    }
    if( args[1] == "load" )
    {
        if( args.size() <= 2 )
        {
            out->print( "Usage: pluginmgr load [pluginname]...\n" );
            return out;
        }
        for( unsigned int i = 2; i < args.size(); i++ )
            if( loadPlugin( args[i], out ) )
                out->print( "Plugin \"%s\" loaded\n", args[i].c_str() );
            else
                out->print( "Error loading plugin \"%s\"\n", args[i].c_str() );
        return out;
    }
    else if( args[1] == "unload" )
    {
        if( args.size() <= 2 )
        {
            out->print( "Usage: pluginmgr unload [pluginname] ...\n" );
            return out;
        }
        for( unsigned int i = 2; i < args.size(); i++ )
        {
            cout<<"in unload loop!\n";
            if( unloadPlugin( args[i] ) )
                out->print( "Plugin \"%s\" unloaded\n", args[i].c_str() );
            else
                out->print( "Error unloading plugin \"%s\"\n", args[i].c_str() );
        }
        return out;
    }
    else if( args[1] == "open" )
    {
        if( args.size() <= 2 )
        {
            out->print( "Usage: pluginmgr open [filename] ...\n" );
            return out;
        }
        for( unsigned int i = 2; i < args.size(); i++ )
        {   
            string pname = openPlugin( args[i], out);
            if( pname == "ERROR" )
                out->print( "Open Plugin: An error occured loading \"%s\".", 
                            args[i].c_str() );
            else if( pname == "NO_FILE" )
                out->print( "Open Plugin: File \"%s\" not found in plugin path.", 
                            args[i].c_str() );
            else
                out->print( "Open Plugin: \"%s\" found in file %s\n", 
                            pname.c_str(),
                            args[i].c_str() );
        }
        return out;
    }else if( args[1] == "get" )
    {
        if( args[2] == "path" ) 
        {
            out->print( m_path.getString().c_str() );
            return out;
        }
        if( args[2] == "pluginlist" )
        {
            map<string, Plugin*>::iterator itr = m_plugins.begin();
            while( itr != m_plugins.end() )
            {
                out->print( itr->first.c_str() );
                if( itr->second->isLoaded() )
                    out->print(":loaded:");
                else
                    out->print(":not loaded:");
                out->print( "%s\n", itr->second->getDescription().c_str() );
                itr++;
            }
            return out;
        }
    }
    
    out->print( "Unknown option \"%s\" for command \"pluginmgr\".\n", args[1].c_str() );
    return out;
}
