//=============================================
// EventMonitorWidget.h
// Author: Patrick Shinpaugh
// Date: July 1, 2003
//=============================================
#include <dmath/dVec3.h>
#include <dreg/Registry.h>
#include <dreg/Entry.h>
#include <eventmgr/EventManager.h>
#include <eventmgr/EventFactory.h>
#include <revent/REventClient.h>
#include "EventMonitorWidget.h"
#include <qpushbutton.h>
#include <stdarg.h>
#include <navmgr/NavManager.h>
#include <tracer/Tracer.h>
#include <qtimer.h>
#include <qtabwidget.h>

static char* float9( char *s, float f );
static char* format( char *s, const char *fmt, ... );

char* float9( char *s, float f )
{
    float a;
    a = f < 0 ? -f : f;

    if (a==0.0f)
        sprintf(s,"%-+8.6f",f) ;
    else if (a>=100000000.f || a<=.0001f)
        sprintf(s,"%-+8.2e",f) ;
    else if (a<10.f)
        sprintf(s,"%-+8.6f",f) ;
    else if (a<100.f)
        sprintf(s,"%-+8.5f",f) ;
    else if (a<1000.f)
        sprintf(s,"%-+8.4f",f) ;
    else if (a<10000.f)
        sprintf(s,"%-+8.3f",f) ;
    else if (a<100000.f)
        sprintf(s,"%-+8.2f",f) ;
    else if (a<1000000.f)
        sprintf(s,"%-+8.1f",f) ;
    else if (a<10000000.f)
        sprintf(s,"%-+8.0f",f) ;
    else
        sprintf(s,"%-+8.0f",f) ;
    return s;
}

static char* format( char *s, const char *fmt, ... )
{
    va_list args;
    va_start(args, fmt);
    vsprintf( s, fmt, args );
    va_end(args);
    return s;
}

RemoteEventCheckBox::RemoteEventCheckBox( QWidget *parent, const char* name )
:QCheckBox( parent, name )
{
}

void RemoteEventCheckBox::eventImport( bool state )
{
    emit importEvent( name(), state );
}

void RemoteEventCheckBox::eventExport( bool state )
{
    emit exportEvent( name(), state );
}

EventMonitorWidget::EventMonitorWidget( QWidget* parent )
    :QWidget( parent )
{
	int i;
    char temp[20];

    m_HSegment = Registry::getValue( "/DGL/HeadTracker", "headtracker" );

    EventManager::addMouseListener( this );
    EventManager::addKeyListener( this );
    EventManager::addWandListener( this );
    EventManager::addTrackListener( "headtracker", this );

    QBoxLayout* layout = new QBoxLayout( this, QBoxLayout::LeftToRight, 0, -1, "layout" );
    QTabWidget* tabwidget = new QTabWidget( this, "tabwidget" );
    layout->addWidget( tabwidget );
    QWidget* monitorwidget = new QWidget( tabwidget, "monitorwidget" );
    QWidget* importexportwidget = new QWidget( tabwidget, "importexportwidget" );
    tabwidget->addTab( monitorwidget, "Event Monitor" );
    tabwidget->addTab( importexportwidget, "Remote Event Control" );

    //Event Monitor WIdget
    QGridLayout* grid = new QGridLayout( monitorwidget, 9, 12, 0, -1, "eventmonitorgrid" );
    for ( i=0;i<7;i++ )
        for( int j=0;j<12;j++ )
        {
            m_labels[i][j] = new QLabel( monitorwidget, format( temp, "label_%d_%d", i, j ) );
            if ( i == 0 || ( j > 1 && j < 12 ) )
                m_labels[i][j]->setAlignment( AlignHCenter );
            else if ( j == 1 )
                m_labels[i][j]->setAlignment( AlignRight );
            else
                m_labels[i][j]->setAlignment( AlignLeft );
            if ( i > 0 && j > 1 )
                m_labels[i][j]->setText( "--" );
            grid->addWidget( m_labels[i][j], i + 1, j );
        }

    for ( i=0;i<4;i++ )
        m_titles[i] = new QLabel( monitorwidget, format( temp, "title_%d", i ) );
    m_titles[0]->setText( "Event" );
    m_titles[0]->setAlignment( AlignHCenter );
    grid->addMultiCellWidget( m_titles[0], 0, 0, 0, 1 );
    m_titles[1]->setText( "Position" );
    m_titles[1]->setAlignment( AlignHCenter );
    grid->addMultiCellWidget( m_titles[1], 0, 0, 2, 4 );
    m_titles[2]->setText( "Direction" );
    m_titles[2]->setAlignment( AlignHCenter );
    grid->addMultiCellWidget( m_titles[2], 0, 0, 5, 7 );
    m_titles[3]->setText( "Buttons" );
    m_titles[3]->setAlignment( AlignHCenter );
    grid->addMultiCellWidget( m_titles[3], 0, 0, 8, 11 );
    
    m_labels[0][0]->setText( "Type" );
    m_labels[0][1]->setText( "Sub-Type" );
    m_labels[0][2]->setText( "X" );
    m_labels[0][3]->setText( "Y" );
    m_labels[0][4]->setText( "Z" );
    m_labels[0][5]->setText( "H" );
    m_labels[0][6]->setText( "P" );
    m_labels[0][7]->setText( "R" );
    m_labels[0][8]->setText( "L" );
    m_labels[0][9]->setText( "R" );
    m_labels[0][10]->setText( "M" );
    m_labels[0][11]->setText( "4" );

    grid->setColStretch( 0, 30 );
    grid->setColStretch( 1, 30 );
    grid->setColStretch( 2, 30 );
    grid->setColStretch( 3, 30 );
    grid->setColStretch( 4, 30 );
    grid->setColStretch( 5, 30 );
    grid->setColStretch( 6, 30 );
    grid->setColStretch( 7, 30 );
    grid->setColStretch( 8, 20 );
    grid->setColStretch( 9, 20 );
    grid->setColStretch( 10, 20 );
    grid->setColStretch( 11, 20 );

    m_labels[1][0]->setText( "Head" );
    m_labels[1][1]->setText( "TRACK" );
    m_labels[2][0]->setText( "Wand" );
    m_labels[3][0]->setText( "Wand" );
    m_labels[4][0]->setText( "Wand" );
    m_labels[2][1]->setText( "TRACK" );
    m_labels[3][1]->setText( "JOYSTICK" );
    m_labels[4][1]->setText( "BUTTON" );
    m_labels[5][0]->setText( "Mouse" );
    m_labels[5][1]->setText( "STATE" );
    m_labels[6][0]->setText( "LOCATION" );

    for ( i=0;i<3;i++ )
        m_keys[i] = new QLabel( monitorwidget, format( temp, "keylabel_%d", i ) );
    grid->addWidget( m_keys[0], 8, 0 );
    grid->addWidget( m_keys[1], 8, 1 );
    grid->addMultiCellWidget( m_keys[2], 8, 8, 2, 11 );
    m_keys[0]->setText( "Key" );
    m_keys[1]->setText( "STATE" );
    m_keys[2]->setText( "--" );
    this->setFixedSize( QSize( 600, 220 ) );

    DMouseEvent *mousestate = EventManager::getMouseState();
    for ( i=0;i<3;i++ )
    {
        if ( mousestate )
            m_mousebuttons[i] = mousestate->getButton( i );
        else
            m_mousebuttons[i] = false;
    }

    DWandEvent *wandstate = EventManager::getWandState();
    for ( i=0;i<4;i++ )
    {
        if ( wandstate )
            m_wandbuttons[i] = wandstate->getButton( i );
        else
            m_wandbuttons[i] = false;
    }

    DKeyboard *keyboardstate = EventManager::getKeyboardState();
    if ( keyboardstate )
        m_asciimap = keyboardstate->getAsciiKeys();

    QTimer* timer = new QTimer( this, "timer" );
    timer->start( 50 );
    connect( timer, SIGNAL( timeout() ), this, SLOT(location()));

    //Remote Event Control
    grid = new QGridLayout( importexportwidget, 6, 3, 0, -1, "remoteeventcontrolgrid" );
    QLabel *titles[3];
    titles[0] = new QLabel( "Event Type", importexportwidget, "title_remote_0" );
    titles[1] = new QLabel( "Import", importexportwidget, "title_remote_1" );
    titles[2] = new QLabel( "Export", importexportwidget, "title_remote_2" );
    grid->addWidget( titles[0], 0, 0, Qt::AlignHCenter | Qt::AlignVCenter );
    grid->addWidget( titles[1], 0, 1, Qt::AlignHCenter | Qt::AlignVCenter );
    grid->addWidget( titles[2], 0, 2, Qt::AlignHCenter | Qt::AlignVCenter );

    map<string,DEvent*> events = EventFactory::getEventMap();
    map<string,DEvent*>::iterator itr;
    int index = 1;
    for ( itr = events.begin(); itr != events.end(); itr++ )
    {
        QLabel *label = new QLabel( importexportwidget );
        label->setText( itr->first.c_str() );
        grid->addWidget( label, index, 0, Qt::AlignHCenter | Qt::AlignVCenter );
        for ( i=1;i<3;i++ )
        {
            RemoteEventCheckBox* events = new RemoteEventCheckBox( importexportwidget, 
                    format( temp, "checkbox_%d", i ) );
            events->setName( itr->first.c_str() );
            grid->addWidget( events, index, i, Qt::AlignHCenter | Qt::AlignVCenter );
            connect( events, SIGNAL(toggled( bool )), events, i == 1 ? SLOT(eventImport( bool )) : 
                    SLOT(eventExport( bool )) );
            connect( events, i == 1 ? SIGNAL(importEvent( const char*, bool )) : 
                    SIGNAL(exportEvent( const char*, bool )), this,
                    i == 1 ? SLOT(eventImport( const char*, bool )) : SLOT(eventExport( const char*, bool )) );
        }
        index++;
    }
}

EventMonitorWidget::~EventMonitorWidget()
{
    EventManager::removeMouseListener( this );
    EventManager::removeKeyListener( this );
    EventManager::removeWandListener( this );
    EventManager::removeTrackListener( "headtracker", this );
}

bool EventMonitorWidget::onMouseEvent( DMouseEvent* evt )
{
    int row;
    char temp[20];

    row = 5;

    m_labels[row][2]->setText( format( temp, "%d", evt->x() ) );
    m_labels[row][3]->setText( format( temp, "%d", evt->y() ) );

    for ( int i=0;i<3;i++ )
    {
        if ( evt->getButton( i ) )
            if ( evt->action() == DMouseEvent::PRESS )
                m_mousebuttons[i] = true;
            else if ( evt->action() == DMouseEvent::RELEASE )
                m_mousebuttons[i] = false;
        m_labels[row][8 + i]->setText( m_mousebuttons[i] ? "X" : "--" );
    }

    location();
    return false;
}

bool EventMonitorWidget::onKeyEvent( DKeyEvent* evt )
{
    map<char,KeyState>::iterator itr;
    switch ( evt->action() )
    {
        case DKeyEvent::UP:
            itr = m_asciimap.find(evt->ascii());
            m_asciimap.erase( m_asciimap.find(evt->ascii()) );
            break;
        case DKeyEvent::DOWN:
            m_asciimap[evt->ascii()] = KeyState(evt->key(), evt->ascii(), evt->action() );
            break;
    }
    string pressed, released;

    for ( itr = m_asciimap.begin(); itr != m_asciimap.end(); itr++ )
    {
        if ( itr->second.state )
        {
            char* temp = new char[2];
            temp[0] = itr->second.ascii;
            temp[1] = '\0';
            pressed += string(temp);
            delete temp;
        }
    }
    if ( !pressed.size() )
        pressed = "--";
    m_keys[2]->setText( pressed.c_str() );
    location();
    return false;
}

bool EventMonitorWidget::onWandEvent( DWandEvent* evt )
{
    int row;
    char temp[20];
    
    row = 2;

    switch( evt->action() )
    {
        case DWandEvent::TRACK:
            row = 2;
            break;
        case DWandEvent::JOYSTICKMOVE:
            row = 3;
            break;
        case DWandEvent::BUTTONPRESS:
        case DWandEvent::BUTTONRELEASE:
            row = 4;
            break;
    }
    dVec3 pos = evt->getPosition(), dir = evt->getDirection();

    if ( evt->action() == DWandEvent::JOYSTICKMOVE || evt->action() == DWandEvent::TRACK )
    {
        m_labels[row][2]->setText( float9( temp, evt->getJoystickX() ) );
        m_labels[row][3]->setText( float9( temp, evt->getJoystickY() ) );
    }
    if ( evt->action() == DWandEvent::TRACK )
    {
        m_labels[row][2]->setText( float9( temp, pos.x() ) );
        m_labels[row][3]->setText( float9( temp, pos.y() ) );
        m_labels[row][4]->setText( float9( temp, pos.z() ) );
        m_labels[row][5]->setText( format( temp, "%+-3.f", dir.h() ) );
        m_labels[row][6]->setText( format( temp, "%+-3.f", dir.p() ) );
        m_labels[row][7]->setText( format( temp, "%+-3.f", dir.r() ) );
    }
    else if ( evt->action() == DWandEvent::BUTTONPRESS || evt->action() == DWandEvent::BUTTONRELEASE )
    {
        for (int i=0;i<4;i++)
        {
            if ( evt->getButton( i ) )
                if ( evt->action() == DWandEvent::BUTTONPRESS )
                    m_wandbuttons[i] = true;
                else if ( evt->action() == DWandEvent::BUTTONRELEASE )
                    m_wandbuttons[i] = false;
            m_labels[row][8 + i]->setText( m_wandbuttons[i] ? "X" : "--" );
        }
    }

    location();
    return false;
}

bool EventMonitorWidget::onTrackerEvent( DTrackerEvent* evt )
{
    int row;
    char temp[20];
    dVec3 pos = evt->getPosition(), dir = evt->getDirection();
    if ( evt->getName() == m_HSegment )
        row = 1;
    else
        return false;
    m_labels[row][2]->setText( float9( temp, pos.x() ) );
    m_labels[row][3]->setText( float9( temp, pos.y() ) );
    m_labels[row][4]->setText( float9( temp, pos.z() ) );
    m_labels[row][5]->setText( format( temp, "%+-3.f", dir.h() ) );
    m_labels[row][6]->setText( format( temp, "%+-3.f", dir.p() ) );
    m_labels[row][7]->setText( format( temp, "%+-3.f", dir.r() ) );

    location();
    return false;
}

bool EventMonitorWidget::onOtherEvent( const string& etype, DEvent* evt )
{
    location();
    return false;
}

void EventMonitorWidget::location()
{
    char temp[20];
    int row = 6;

    dMatrix navmat = NavManager::getMatrix();
    dVec3 dir = navmat.getHPR();
    dVec3 pos = navmat.getTrans();

    m_labels[row][2]->setText( format( temp, "%7.2f", pos.x() ) );
    m_labels[row][3]->setText( format( temp, "%7.2f", pos.y() ) );
    m_labels[row][4]->setText( format( temp, "%7.2f", pos.z() ) );
    m_labels[row][5]->setText( format( temp, "%7.3f", dir.h() ) );
    m_labels[row][6]->setText( format( temp, "%7.3f", dir.p() ) );
    m_labels[row][7]->setText( format( temp, "%7.3f", dir.r() ) );
}

void EventMonitorWidget::eventImport( const char* name, bool state )
{
    Tracer::debug( "Import remote events for %s - %s", name, state ? "ON" : "OFF" );
    REventClient::instance()->importEvent( name, state );
}

void EventMonitorWidget::eventExport( const char* name, bool state )
{
    Tracer::debug( "Export remote events for %s - %s", name, state ? "ON" : "OFF" );
    REventClient::instance()->exportEvent( name, state );
}
