#include"QDWidget.h"
#include<dmath/dVec3.h>
#include<dmath/dMatrix.h>
#include<dgl/dglcore/DGL.h>
#include<eventmgr/devents.h>
#include<eventmgr/EventManager.h>
#include<displaymgr/DisplayManager.h>
#include<displaymgr/DCamera.h>
#include<dreg/Registry.h>
#include<iostream>
#include<qevent.h>
#include<tracer/Tracer.h>
using namespace std;

//=============================================================================
// Function Name: QDWidget
/// @brief Constructor.
/// @param parent - the parent widget.
//=============================================================================
QDWidget::QDWidget( QWidget* parent )
    :QGLWidget( parent )
{
    m_left   = - 1;
    m_right  =   1;
    m_top    =   1;
    m_bottom = - 1;
    m_fov    = atoi( 
                Registry::getValue("/DIVERSE/DWidgets/QDWidget/field_of_view",
                                        "35", 
                                        "The field of view in degrees"
                                        ).c_str() );

    m_framenum  = 0;
    m_displayid = 0;
    
    setFocusPolicy(StrongFocus);

    m_headtracker_name = 
        Registry::getValue("/DIVERSE/HeadTracker", "headtracker" );
}

//=============================================================================
// Function Name: paintGL
/// @brief Sets up the frustum and model view and then displays the scene.
//=============================================================================
void QDWidget::paintGL()
{
    if( !isVisible() ) return;

    // TBD: Optimize this
    DTrackerEvent* headtracker = 
                    EventManager::getTrackerState(m_headtracker_name);
    
    dMatrix headmat;
    headmat.makeIdentity();
    // rotate about 0, 0, 0 and then translate to head pos
    dVec3 headpos(0,0,0);
    dVec3 headdir(0,1,0);
    if( headtracker )
    {
        headpos = headtracker->getPosition();// = DGL::getHeadPos();
        headdir = headtracker->getDirection();
        dMatrix euler;
        euler.makeRotate( DEG2RAD(headdir.h()), dVec3( 0, 0, -1 ),
                          DEG2RAD(headdir.p()), dVec3( -1, 0, 0 ),
                          DEG2RAD(headdir.r()), dVec3( 0, -1, 0 ) );
        // rotate about 0, 0, 0 and then translate to head pos
        headmat = dMatrix::translate( -headpos ) * euler;
    }

    glMatrixMode( GL_MODELVIEW );
        glLoadIdentity();
        glRotatef( -90.0f, 1.0f, 0.0f, 0.0f );
        glMultMatrixf( headmat.ptr() );

    dMatrix viewmat;
    viewmat.makePerspective( m_fov, float(width())/height(),
                                        DisplayManager::getNearClip(),
                                        DisplayManager::getFarClip() );

    headmat = headmat * dMatrix::rotate( -M_PI/2.0, 1, 0, 0 );
    DCamera cam( Rect( 0, 0, width(), height() ),
                 viewmat.ptr(), headmat.ptr() ); 
    m_framenum++;
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );

    DisplayManager::clear(&cam);
    DisplayManager::cull(&cam);
    DisplayManager::draw(&cam);
    DisplayManager::overlay();
}

//=============================================================================
/// @brief Resize the OpenGL viewport when the window is resized.
/// @param width - the width of the viewport.
/// @param height - the height of the viewport.
//=============================================================================
void QDWidget::resizeGL( int width, int height )
{
    glViewport( 0, 0, width, height );
    glScissor( 0, 0, width, height );
    glMatrixMode( GL_PROJECTION );
        glLoadIdentity();
        gluPerspective( m_fov, float(width)/height, 
                        DisplayManager::getNearClip(), 
                        DisplayManager::getFarClip() );
}

//=============================================================================
/// @brief Handle Qt mouse button press events, create the event and push it
/// onto the event stack.
/// @param evt - the QMouseEvent to be converted to a DIVERSE event.
//=============================================================================
void QDWidget::mousePressEvent( QMouseEvent* evt )
{
    DMouseEvent* devt = new DMouseEvent(DMouseEvent::PRESS);
    devt->setPosition( evt->pos().x(), evt->pos().y() );
    if( evt->button() & Qt::LeftButton )
        devt->setButton( DMouseEvent::LEFT, true );
    if( evt->button() & Qt::RightButton )
        devt->setButton( DMouseEvent::RIGHT, true );
    if( evt->button() & Qt::MidButton )
        devt->setButton( DMouseEvent::MIDDLE, true );
    EventManager::pushEvent( devt );
}

//=============================================================================
/// @brief Handle Qt mouse move events, create the event and push it
/// onto the event stack.
/// @param evt - the QMouseEvent to be converted to a DIVERSE event..
//=============================================================================
void QDWidget::mouseMoveEvent( QMouseEvent* evt )
{
    DMouseEvent* devt = new DMouseEvent(DMouseEvent::MOVE);
    devt->setPosition( evt->pos().x(), evt->pos().y() );
    if( evt->button() & Qt::LeftButton )
        devt->setButton( DMouseEvent::LEFT, true );
    if( evt->button() & Qt::RightButton )
        devt->setButton( DMouseEvent::RIGHT, true );
    if( evt->button() & Qt::MidButton )
        devt->setButton( DMouseEvent::MIDDLE, true );
    EventManager::pushEvent( devt );
}

//=============================================================================
/// @brief Handle Qt mouse button release events, create the event and push it
/// onto the event stack.
/// @param evt - the QMouseEvent to be converted to a DIVERSE event..
//=============================================================================
void QDWidget::mouseReleaseEvent( QMouseEvent* evt )
{
    DMouseEvent* devt = new DMouseEvent(DMouseEvent::RELEASE);
    devt->setPosition( evt->pos().x(), evt->pos().y() );
    if( evt->button() & Qt::LeftButton )
        devt->setButton( DMouseEvent::LEFT, true );
    if( evt->button() & Qt::RightButton )
        devt->setButton( DMouseEvent::RIGHT, true);
    if( evt->button() & Qt::MidButton )
        devt->setButton( DMouseEvent::MIDDLE, true );
    EventManager::pushEvent( devt );
}

//=============================================================================
/// @brief Handle Qt key press events, create the event and push it
/// onto the event stack.
/// @param evt - the QKeyEvent to be converted to a DIVERSE event.
//=============================================================================
void QDWidget::keyPressEvent( QKeyEvent* evt )
{
    DKeyEvent* devt = new DKeyEvent( DKeyEvent::DOWN, 
                                evt->ascii(), evt->key() );
    EventManager::pushEvent( devt );
}

//=============================================================================
/// @brief Handle Qt key release events, create the event and push it
/// onto the event stack.
/// @param evt - the QKeyEvent to be converted to a DIVERSE event.
//=============================================================================
void QDWidget::keyReleaseEvent( QKeyEvent* evt )
{
    DKeyEvent* devt = new DKeyEvent( DKeyEvent::UP, 
                                evt->ascii(), evt->key() );
    EventManager::pushEvent( devt );
}
