//=============================================
// caveDeviceSimulatorWidget.h
// Author: Patrick Shinpaugh
// Date: July 3, 2003
//=============================================
#ifndef _CAVEDEVICESIMULATORWIDGET_H_
#define _CAVEDEVICESIMULATORWIDGET_H_
#include <dmath/dVec3.h>
#include <qwidget.h>
#include <qslider.h>
#include <qpoint.h>
#include <qpushbutton.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qcheckbox.h>
#include <qcanvas.h>
#include <eventmgr/EventListener.h>

#ifdef WIN32
#ifdef QTUTIL_EXPORTS
#define QTUTIL_API __declspec(dllexport)
#else
#define QTUTIL_API __declspec(dllimport)
#endif /* QTUTIL_EXPORTS */
#else
#define QTUTIL_API
#endif /* WIN32 */

//=============================================================================
// caveDeviceSimulator provides an interface ot simulate wand tracker events.
//
// Please see the user documentation for details on interacting with this tool.
//=============================================================================


//=============================================================================
// A struct containing a position and orientation vector.
//=============================================================================
struct Coord
{
    dVec3 pos;
    dVec3 dir;
};

//=============================================================================
// Provides a button class which needed additional functionality for its tasks.
//=============================================================================
class QTUTIL_API CaveDevSimButton:public QPushButton
{
    Q_OBJECT
    public:
        CaveDevSimButton( const QString & text, QWidget * parent, const char * name = 0 );
        CaveDevSimButton( int num, const QString & text, QWidget * parent, const char * name = 0 );
        CaveDevSimButton( int headwand, int num, const QString & text, QWidget * parent, const char * name = 0 );
        
    public slots:
        void resetSlider();
        void initSlider();
        void resetButtons();
        void initButtons();
        void updateButtons( bool on );
    
    signals:
        void buttonToggled( int num, bool on );
        void reset( int headwand, int num );
        void init( int headwand, int num );
        
    private:
        int m_headwand;
        int m_num;
};

//=============================================================================
// Provides a label class which needed additional functionality for its tasks.
//=============================================================================
class QTUTIL_API CaveDevSimLabel:public QLabel
{
    Q_OBJECT
    public:
        CaveDevSimLabel( QWidget * parent, const char * name = 0, WFlags f = 0 );

    public slots:
        void sliderToXYZ( int val );
        void sliderToHPR( int val );
};

//=============================================================================
// Provides a canvas view class which needed additional functionality for its tasks.
//=============================================================================
class QTUTIL_API CaveDevSimCanvasView:public QCanvasView
{
    Q_OBJECT
    public:
        CaveDevSimCanvasView( QCanvas * canvas, QWidget * parent = 0, const char * name = 0, WFlags f = 0 );
		void setJoystickCanvasPassed( bool passed ) { m_joystickcanvaspassed = passed; };
		bool getJoystickCanvasPassed() { return m_joystickcanvaspassed; };

    public slots:
        void updateXPosition( const QString &text );
        void updateYPosition( const QString &text );
        void updatePosition( int x, int y );
        void updateSpringStick( bool val );
        
    protected:
        void contentsMouseMoveEvent( QMouseEvent* evt );
        void contentsMousePressEvent( QMouseEvent* evt );
        void contentsMouseReleaseEvent( QMouseEvent* evt );
    
    signals:
        void joystickChanged( float x, float y );
        void joystickChanged();
        
    private:
        QCanvas* m_canvas;
        QCanvasLine* m_joysticklines[2];
        bool m_springstickval;
		bool m_joystickcanvaspassed;
};

//=============================================================================
// The main widget which holds all of the other widgets required for the
// caveDeviceSimulator.
//=============================================================================
class QTUTIL_API caveDeviceSimulatorWidget:public QWidget, public EventListener
{
    Q_OBJECT
    public:
        caveDeviceSimulatorWidget( QWidget* parent );
        ~caveDeviceSimulatorWidget();
        
        bool onTrackerEvent( DTrackerEvent* evt );
        bool onWandEvent( DWandEvent* evt );

    private slots:
        void resetAll();
        void initAll();
        void updateMem();
        void updateWidget();
        void help();
        void resetSlider( int headwand, int num );
        void initSlider( int headwand, int num );
        void resetJoystick();
        void initJoystick();
        void resetButton( int headwand, int num );
        void initButton( int headwand, int num );
        void updateHeadTrackEvent( int val );
        void updateWandTrackEvent( int val );
        void updateWandJoystickckEvent();
        void updateWandButtonEvent( int num, bool on );
        void updateJoystickLineEdit( float x, float y );

    signals:
        void widgetClosed();

    private:
		static int m_nextindex;
		int m_index;
		string m_sourcename;
        string m_HSegment;
        bool m_headeventpassed;
        bool m_wandeventpassed;

        QSlider* m_sliders[2][6];
        CaveDevSimLabel* m_trackvals[2][6];
        CaveDevSimButton* m_buttons[4];
        QLineEdit* m_joysticklineedit[2];
        QCheckBox* m_springstick;
        CaveDevSimCanvasView* m_joystickview;

        Coord m_headval;
        Coord m_wandval;
        float m_joystickval[2];
        bool m_buttonsval[4];
        
        Coord m_headinitval;
        Coord m_wandinitval;
        float m_joystickinitval[2];
        bool m_buttonsinitval[4];

        Coord m_headresetval;
        Coord m_wandresetval;
        float m_joystickresetval[2];
        bool m_buttonsresetval[4];
};

#endif
