#include"EventServer.h"
#include<pexpr/PostOffice.h>
#include<pexpr/POBox.h>
#include<pexpr/POAddress.h>
#include<pexpr/POTextMsg.h>
#include"EventMsg.h"
#include<eventmgr/devents.h>
#include"REventClient.h"
#include<qstring.h>
#include<dutil/ColonList.h>
#include<dreg/Registry.h>
#include<qapplication.h>
#include<servicemgr/GServiceManager.h>
#ifndef WIN32
#include <unistd.h>
#endif

//=============================================================================
/// @brief Constructor.
//=============================================================================
EventServer::EventServer()
    :GService("EventServer")
{
    EventManager::addAllListener( this );
    GServiceManager::addService(this);


    char* hostname = new char[255];
    gethostname( hostname, 255 );
    //getenv( "HOSTNAME" );
//    if( hostname == NULL )
//    {
//        Tracer::error("HOSTNAME == NULL");
//        hostname = "HOSTNAME";
//    }
    m_po = new PostOffice( hostname+string("_EventServer") );

    string servicept = Registry::getValue( "/DIVERSE/EventServer/port", "9300" );
    m_servicept = atoi(servicept.c_str());
    m_po->servicePort( m_servicept );
    m_box = m_po->registerPOBox( "REVENT_SERVER" );
    delete [] hostname;
}

//=============================================================================
/// @brief Handles the postProcess stage of the application cycle for EventServer.
//=============================================================================
void EventServer::postProcess( )
{
    while( m_box->hasMessages(true) ) // check now, limit latency
    {
        const POMessage* msg = m_box->popMessage();
        handleMessage( msg );
        delete msg;
    }
}

//=============================================================================
/// @brief Handles the specified message.
/// @param msg - the name of the POMessage to handle.
//=============================================================================
void EventServer::handleMessage( const POMessage* msg )
{
    if( msg->messageType() == "POTextMsg" )
    {
        ColonList l(((const POTextMsg*)msg)->getText() );
        if( l.size() > 0 )
        {
            if( l[0] == "connect" )
                addConnection( msg->getSource(), l );
            else if( l[0] == "disconnect" )
                removeConnection( msg->getSource(), l );
            else if( l[0] == "rconnect" )
                addRemoteConnection( l );
            else if( l[0] == "rdisconnect" )
                removeRemoteConnection( l ); 
        }
    }
    if( msg->messageType() == "EventMsg" )
    {
        handleEventMsg( ((const EventMsg*)msg) );
    }
}

//=============================================================================
// Function Name: addRemoteConnection
/// @brief Takes a colon list of the format:
///         rconnect:hostname:EVENTS
///     and sends the following message to hostname:
///         connect:EVENTS
///     
/// This will cause the remote EventServer to send us all events of that
/// type and we then forward it to everyone connected to us.
/// @param cl - colon list containing the specified format rconnect:hostname:EVENTS.
//=============================================================================
void EventServer::addRemoteConnection( const ColonList& cl )
{
    //================================================
    // If we aren't connected to that host, connect 
    //================================================
    if( m_hostmap.find(cl[1]) == m_hostmap.end() )
    {
        string host_po = m_po->connectTo( cl[1], m_servicept );
        m_hostmap[cl[1]] = host_po;
    }
   
    //========================================================
    // Register to receive those event types from that server
    //========================================================
    POAddress cadr( m_hostmap[cl[1]], "EVENT_SERVER" );
    string revents("connect:");
    for( int x = 2; x < cl.size(); x++ )
        revents += cl[x] + ":"; 
    POTextMsg msg(revents);
    m_box->sendMessage( &msg, cadr );

    // Send our events of that type to them?  let them decide..
    return;
    
    //================================================
    // Copy the event types
/*    for( int i = 1; i < evts.size(); i++ )
        m_exports[evts[i]].push_back( POAddress(host_po, "REVENT_SERVER") );
    
    POAddress cadr( m_hostmap[cl[1]], "EVENT_SERVER" );
    string revents("connect:");
    for( int x = 2; x < cl.size(); x++ )
        revents += cl[x] + ":"; 
    POTextMsg msg(revents);
    m_box->sendMessage( &msg, cadr );
*/
}

//=============================================================================
// Function Name: removeRemoteConnection
/// @brief Takes a command as a colon list in the following format:
///     rdisconnect:hostname:EVENTTYPE:+
///     
/// And then sends a message to hostname to stop receiving all
/// events of the specified types
/// @param cl - colon list containing the specified format rdisconnect:hostname:EVENTTYPE:+.
//=============================================================================
void EventServer::removeRemoteConnection( const ColonList& cl )
{
    POAddress cadr( m_hostmap[cl[1]], "EVENT_SERVER" );
    string revents("disconnect:");
    for( int x = 2; x < cl.size(); x++ )
        revents += cl[x] + ":"; 
    POTextMsg msg(revents);
    m_box->sendMessage( &msg, cadr );
    
    // stop sending them events?  let them decide...
}

//=============================================================================
/// @brief Add a connection.
/// @param adr - the POAddress object containing the address information.
/// @param evts - colon list containing the event information.
//=============================================================================
void EventServer::addConnection( const POAddress& adr, const ColonList& evts )
{
    for( int i = 1; i < evts.size(); i++ )
        m_exports[evts[i]].push_back( adr ); // TODO: MAKE UNIQUE
}

//=============================================================================
/// @brief Remove a connection.
/// @param adr - the POAddress object containing the address information.
/// @param evts - colon list containing the event information.
//=============================================================================
void EventServer::removeConnection( const POAddress& adr, 
                                     const ColonList& evts )
{
    for( int i = 1; i < evts.size(); i++ )
    {
        vector<POAddress> vec = m_exports[evts[i]];
        vector<POAddress>::iterator itr = vec.begin();
        while( itr != vec.end() )
        {
            if( *itr == adr )
            {
                vec.erase(itr);
                m_exports[evts[i]] = vec;
                break;
            }
            itr++;
        }
    }
}

//=============================================================================
/// @brief Handles events by passing them to each POBox as messages.
/// @param evt - the event that is being processed.
/// @return always return false.
//=============================================================================
bool EventServer::onAnyEvent( DEvent* evt )
{
    ColonList l( evt->getSource() );
    if( l[0] == "EventServer" )
    {
        return false; // avoid circular loops
    }
    //======================================================
    // Forward Event to all connections for that event type
    //======================================================
    EventMsg cpy(evt);
    vector<POAddress> adrs = m_exports[evt->type()];
    for( unsigned int i = 0; i < adrs.size(); i++ )
    {
        m_box->sendMessage( &cpy, adrs[i], true ); // send now, limit latency
    }
    return false;
}

//=============================================================================
/// @brief Handle an event message.
/// @param msg - the message to be handled.
//=============================================================================
void EventServer::handleEventMsg( const EventMsg* msg )
{
    if( !msg->getEvent() )
    {
        Tracer::error("EventServer::handleEventMsg() received NULL event");
        return;
    }
    DEvent* evt = msg->getEvent()->clone();
    evt->setSource( "EventServer:" + evt->getSource() );
    EventManager::pushEvent( evt );

    //======================================================
    // Forward Event to all connections for that event type
    //======================================================
    EventMsg* cpy = (EventMsg*)msg->clone();
    vector<POAddress> adrs = m_exports[msg->getEvent()->type()];
    for( unsigned int i = 0; i < adrs.size(); i++ )
    {
        //cpy->getEvent()->setWhen( DTime::current() );
        if( adrs[i] != msg->getSource() ) // Don't send it back to the source
            m_box->sendMessage( cpy, adrs[i], true ); // send now, limit latency
    }
    delete cpy;
}
