#include"REventClient.h"
#include<pexpr/PostOffice.h>
#include<pexpr/POBox.h>
#include<pexpr/POAddress.h>
#include<pexpr/POTextMsg.h>
#include<dreg/Registry.h>
#include<dutil/ColonList.h>
#include<dutil/CommandArgs.h>
#include<servicemgr/GServiceManager.h>
#include<eventmgr/EventManager.h>
#include<eventmgr/devents.h>
#include"EventMsg.h"
#include<qstring.h>
#ifdef WIN32
#define getpid GetCurrentProcessId
#else
#include<unistd.h>
#endif
#include<qapplication.h>

REventClient*       REventClient::m_instance = NULL;
PostOffice*         REventClient::m_po  = NULL;
POBox*              REventClient::m_box = NULL;
POAddress           REventClient::m_server;
map<string, bool>   REventClient::m_exports;
map<string, bool>   REventClient::m_imports;
int                 REventClient::m_servicept;

//=============================================================================
/// @brief Constructor.
//=============================================================================
REventClient::REventClient()
    :GService("EventClient")
{
	int i;
    EventManager::addAllListener( this );
    GServiceManager::addService(this);

    char* hostname = new char[255];
    gethostname( hostname, 255 ); //getenv("HOSTNAME");

    m_po = new PostOffice( 
            string("REventClient_") 
            + QString::number(getpid()).latin1() +
            "_" + string( getenv( "HOSTNAME" ) ) );
    m_box = m_po->registerPOBox( "REventClient" );
    
    string servicept = Registry::getValue( "/DIVERSE/EventServer/port", "9300" );
    string host = Registry::getValue( "/DIVERSE/EventServer/host", "localhost" );
    m_servicept = atoi(servicept.c_str());

    //======================================================
    // Check the command line args to override registry
    //======================================================
    CommandArgs::addUsage( "-event-host", "localhost|...", 
    "The host name of the computer running the Event Server, defaults to localhost", false );

    CommandArgs::addUsage( "-event-host-port", "9300", 
                           "The port used by the Event Server", false );

    string chost = CommandArgs::getValue( "-event-host" );
    string cport = CommandArgs::getValue( "-event-host-port" );

    if( chost != "" ) host = chost;
    if( cport != "" ) m_servicept = atoi(cport.c_str());
    

    string srvname = m_po->connectTo( host, m_servicept);
    m_server = POAddress(srvname, "REVENT_SERVER");

    if( srvname == "undefined" )
    {
        Tracer::error( "REventClient couldn't connect to server\n" );
        return;
    }
    
    string importevts = Registry::getValue("/DIVERSE/GServiceManager/EventClient/importevents",
                                           "TRACKER:WAND:KEYBOARD:MOUSE" );

    ColonList imports(importevts);
    for( i = 0; i < imports.size(); i++ )
        importEvent( imports[i], true );

    string exportevts = Registry::getValue("/DIVERSE/GServiceManager/EventClient/exportevents",
                                           "TRACKER:WAND:KEYBOARD:MOUSE" );
    ColonList exports(exportevts);
    for( i = 0; i < exports.size(); i++ )
        exportEvent( exports[i], true );
    delete[] hostname;
}

//=============================================================================
/// @brief Handles each message during the postProcess stage of the application.
//=============================================================================
void REventClient::postProcess()
{
    while( m_box->hasMessages(true) )
    {
        const POMessage* msg = m_box->popMessage();
        handleMessage( msg );
        delete msg;
    }
}

//=============================================================================
/// @brief Handles the message if it is an event message.
/// @param msg - the POMessage to be handled.
//=============================================================================
void REventClient::handleMessage( const POMessage* msg )
{
    if( msg->messageType() == "EventMsg" )
    {
        handleEventMsg( ((const EventMsg*)msg) );
    }
}

//=============================================================================
/// @brief Handles events by passing them to each POBox as messages.
/// @param evt - the event that is being processed.
/// @return always return false.
//=============================================================================
bool REventClient::onAnyEvent( DEvent* evt )
{
    ColonList l( evt->getSource() );
    if( l[0] == "REventClient" )
    {
        return false; // avoid circular loops
    }
    if( m_exports[evt->type()] )
    {
        EventMsg msg(evt);
        m_box->sendMessage( &msg, m_server, true ); 
    }
    return false;
}

//=============================================================================
/// @brief Tells the current process to export events to the local Event Daemon.
/// @param type - the type of event.
/// @param state - true to export the event type, else false .
//=============================================================================
void REventClient::exportEvent( const string& type, bool state )
{
    m_exports[type] = state;
}

//=============================================================================
/// @brief Tells the current process to import events from the local event Daemon.
/// @param type - the type of event.
/// @param state - true to import the event type, else false .
//=============================================================================
void REventClient::importEvent( const string& type, bool state )
{
    m_imports[type] = state;

    string action;
    if( state ) action = "connect";
    else action = "disconnect";
    
    string cmd( action +":" + type );
    
    POTextMsg msg( cmd );
    m_box->sendMessage( &msg, m_server );
}

//=============================================================================
/// @brief Sends a message to the EventServer telling it to 
/// import events of the specified type from the specified host.
/// @param type - the type of event.
/// @param host - the name of the host machine.
/// @param state - true to export the event type, else false .
//=============================================================================
void REventClient::importRemoteEvents( const string& type,
                                       const string& host, bool state )
{
    string action;
    if( state ) action = "rconnect";
    else action = "rdisconnect";
    
    string cmd( action +":" + host + ":" + type );
    POTextMsg msg( cmd );
    m_box->sendMessage( &msg, m_server );
}

//=============================================================================
/// @brief Sends a message to the EventServer telling it to export
/// events to the specified host.
/// @param type - the type of event.
/// @param host - the name of the host machine.
/// @param state - true to export the event type, else false .
//=============================================================================
void REventClient::exportRemoteEvents( const string& type,
                                       const string& host, bool state )
{
    Tracer::warn("REventClient::exportRemoteEvents(...) isn't implemented");
}

//=============================================================================
/// @brief Handles event messages. Do not call directly, call handleMessage instead. 
/// @param msg - the EventMsg to be handled.
//=============================================================================
void REventClient::handleEventMsg( const EventMsg* msg )
{
    if( m_imports[msg->getEvent()->type()] == false )
    {
        Tracer::warn( "Remote Event Client received an unrequested\n"
                      "\t\tevent type \"%s\".", msg->getEvent()->type().c_str() );
        return;
    }
    DEvent* evt = msg->getEvent()->clone();
    //cout<<"Delay: "<<-(evt->when().sec() - DTime::current().sec())<<"s\n";
    evt->setSource( "REventClient:" + evt->getSource() ); // mark the source 
    EventManager::pushEvent( evt );
}

//=============================================================================
/// @brief Get a pointer to the REventClient instance. Creates the instance if
/// it doesn not already exist.
/// @return a pointer to the REventClient instance.
//=============================================================================
REventClient* REventClient::instance()
{
    if( m_instance ) return m_instance;
    m_instance = new REventClient();
    return m_instance;
}
