#ifndef _GSERVICE_H_
#define _GSERVICE_H_
#include<string>
#include<dutil/DTime.h>
#include<tracer/Tracer.h>
using namespace std;

#ifdef WIN32
#ifdef SERVICEMGR_EXPORTS
#define SERVICEMGR_API __declspec(dllexport)
#else
#define SERVICEMGR_API __declspec(dllimport)
#endif /* SERVICEMGR_EXPORTS */
#else
#define SERVICEMGR_API
#endif /* WIN32 */

//=============================================================================
/// @class GService 
/// @author Daniel Larimer
/// @brief Base class for all services run by the ServiceManager.
///
/// Any plugin which will be run as a service should be derived from GService. 
/// The derived service must have a unique name or it may not load, or cause 
/// another service not to load. A service will be called for each stage
/// of the application cycle, but you need only override the stages you need
/// to handle.
///
/// The priority of the service can be set such that it will be called in a
/// certain order relative to the other services called. The default priority
/// is 1 if it is not set. Services with a priority of 0 will be called last.
/// Services with the same priority will be called based on when they were
/// initialized compared to each other. If the priority is set to 0 then it is
/// never started and will not run. If it had already started, it will stop running.
///
/// Services must be started in order for them to be called by the application.
/// They can be stopped and restarted if necessary using the start and stop
/// function calls. Stopping the service will prevent it from being called in
/// the processing stages of the application.
//=============================================================================
class SERVICEMGR_API GService 
{
    public:
        GService( const string& name );
		/// @brief Destructor.
        virtual ~GService(){}

        virtual void preConfig();
        virtual void config();
        virtual void postConfig();

		/// @brief Called during the sync stage.
        virtual void sync(){};
		/// @brief Called during the preProcess stage.
        virtual void preProcess(){};
		/// @brief Called during the process stage.
        virtual void process(){};
		/// @brief Called during the postProcess stage.
        virtual void postProcess();

        virtual bool start();
        virtual bool stop();
        virtual bool isRunning()const;
        
        string  getServiceName()const;
        
        virtual void setPriority( int p, bool tell_servicemgr = true );
		/// @brief Get the priority value of the service.
		/// @return the priority value of the service.
        virtual int  getPriority()const   { return m_priority; }
    private:
		/// @brief Private constructor.
        GService(){}
        string m_servicename;
        bool   m_running;
        int    m_priority;
};

#endif
