#include"GServiceManager.h"
#include"GService.h"
#include<signal.h>
#include<tracer/Tracer.h>
#include<pluginmgr/PluginManager.h>
#include<commander/CommandManager.h>
#include<commander/CommandOut.h>
#include<dreg/Registry.h>
#include<diverseconfig.h>
#include<stdio.h>
#ifdef WIN32
#define snprintf _snprintf
#else
#include<unistd.h>
#endif
#include<iostream>
using namespace std;

GServiceManager* GServiceManager::m_self = NULL;
bool	         GServiceManager::m_config = false;
bool	         GServiceManager::m_setup = false;
unsigned int     GServiceManager::m_curservice = 0;
int              GServiceManager::m_sleeptime = 1;
bool             GServiceManager::m_running = false;

void DiverseSignalHandler( int sig )
{
    static int count = 0;
    if( sig == 2 )
        count++;
    if( count == 3 )
    {
        Tracer::error("GServiceManager won't quit! exit(1)");
        exit(1);
    }
    Tracer::debug( 5, "GServiceManager - Caught Signal: %d"
                   "\t\tQuitting", sig );
    GServiceManager::quit();
}

//=============================================================================
/// @brief Initializes the service manager, which loads all the services
/// specified in the registry config file.
//=============================================================================
void GServiceManager::init()
{
    if( m_self ) return;
    setupRegistry();
    Tracer::init();
    
    signal( SIGINT, DiverseSignalHandler );
    m_self = new GServiceManager();
    CommandManager.addCommandHandler( "quit", m_self );
    CommandManager.addCommandHandler( "servicemgr", m_self );

    string sleeptime = Registry::getValue("DIVERSE/GServiceManager/SleepTime");
    m_self->m_sleeptime = atoi( sleeptime.c_str() );
    m_self->m_running = false;

}

//=============================================================================
/// @brief Executes a Commander command.
/// @param cmd - the Command to execute.
/// @return a CommandOut.
//=============================================================================
CommandOut* GServiceManager::execute( const Command& cmd )
{
    CommandOut* out = new CommandOut( cmd );
    vector<string> args = cmd.getArgs();
    if( args[0] == "quit" )
    {
        out->print( "Calling GServiceManager::quit()" );
        quit();
        return out;
    }
    else if( args[0] == "servicemgr" )
    {
        if( args.size() == 1 || args[1] == "--help" || args[1] == "-h") 
        {
            out->print( "Usage:\n"
                        "  servicemgr getServices\n"
                        "\t- Lists 1 service per line\n"
                        "\t    with the format of SERVICENAME:PRIORITY\n\n" 
                        "  servicemgr setServicePriority SERVICENAME PRIORITY\n"
                        "\t - Sets the priority of the specified service\n\n"
                        "  servicemgr start SERVICENAME\n"
                        "\t - Starts the service SERVICENAME\n\n"
                        "  servicemgr stop SERVICENAME\n"
                        "\t - Stops the service SERVICENAME\n\n"
                        );
            return out;
        }
        if( args[1] == "getServices" )
        {
            map<string, GService*>::iterator itr = m_self->m_services.begin();
            char tmp[5];
            while( itr != m_self->m_services.end() )
            {
                out->print( "%s:", itr->first.c_str() );
                snprintf( tmp, 5,"%d\n", itr->second->getPriority() );
                out->print( tmp );
                itr++;
            }
            return out;
        }
        if( args[1] == "setServicePriority" )
        {
            if( args.size() < 4 ) 
            {
                out->print( "Usage: servicemgr setServicePriority SERVICENAME PRIORITY\n" );
                return out;
            }
            map<string, GService*>::iterator itr = m_self->m_services.find(args[2]);
            if( itr == m_self->m_services.end() )
            {
                out->print( "servicemgr: setServicePriority: Unknown service \"%s\"",
                            args[2].c_str() );
                return out;
            }
            itr->second->setPriority( atoi( args[3].c_str() ) );
            out->print( "Priority of service \"%s\" set to %d\n", 
                        args[2].c_str(), atoi(args[3].c_str()) );
            return out;
        }
        if( args[1] == "start" )
        {
            if( args.size() < 3 ) 
            {
                out->print( "Usage: servicemgr start SERVICENAME\n" );
                return out;
            }
            map<string, GService*>::iterator itr = m_self->m_services.find(args[2]);
            if( itr == m_self->m_services.end() )
            {
                out->print( "servicemgr: start: Unknown service \"%s\"",
                            args[2].c_str() );
                return out;
            }
            itr->second->start();
            out->print( "Priority of service \"%s\" set to %d\n", 
                        args[2].c_str(), itr->second->getPriority() );
            return out;
        }
        if( args[1] == "stop" )
        {
            if( args.size() < 3 ) 
            {
                out->print( "Usage: servicemgr stop SERVICENAME\n" );
                return out;
            }
            map<string, GService*>::iterator itr = m_self->m_services.find(args[2]);
            if( itr == m_self->m_services.end() )
            {
                out->print( "servicemgr: stop: Unknown service \"%s\"",
                            args[2].c_str() );
                return out;
            }
            itr->second->stop();
            out->print( "Priority of service \"%s\" set to %d\n", 
                        args[2].c_str(), itr->second->getPriority() );
            return out;
        }
        out->print( "GServiceManager: servicemgr: Unknown command \"%s\"", 
                     cmd.getArgs()[1].c_str() );
    }
    
    out->print( "GServiceManager: Unknown command \"%s\"", 
                 cmd.getArgs()[0].c_str() );
    return out;
}

//=============================================================================
/// @brief Sets up the registry for the servicemgr.
/// @param cfgfile - the configuration file to be used for the registry settings.
//=============================================================================
void GServiceManager::setupRegistry( const string& cfgfile )
{
    if( m_setup ) return;
    Registry::init( cfgfile );
    PluginManager::setupRegistry( cfgfile );

    //====================================================
    // Get the current plugin path
    //====================================================
    string ppath = Registry::getValue("/DIVERSE/PluginManager/pluginpath" );
    ColonList path(ppath);

    //====================================================
    // Get the specified service path
    //====================================================
    string srvpath = Registry::getValue("/DIVERSE/GServiceManager/servicepath", 
                        DEFAULT_SERVICE_PATH,
                        "A Colon separated path");
    ColonList spath(srvpath);
    
    //====================================================
    // Merge Them
    //====================================================
    for( int i = 0; i < spath.size(); i++ )
        path.addItem(spath[i], true);
    
    //====================================================
    // Set the new pluginpath
    //====================================================
    Registry::setValue("/DIVERSE/PluginManager/pluginpath", path.getString() );
    
    Registry::getValue("DIVERSE/GServiceManager/SleepTime", "1");
    m_setup = true;
}

//=============================================================================
/// @brief Gets the next service from the service list.
/// @return a pointer to the next service.
//=============================================================================
GService* GServiceManager::getNextService()
{
    if( m_curservice >= m_self->m_ordered.size() )
        return NULL;
    return m_self->m_ordered[m_curservice++];
}

//=============================================================================
/// @brief Unloads all services and cleans up the service manager.
/// Not currently implemented.
//=============================================================================
void GServiceManager::exit()
{
}

//=============================================================================
/// @brief Adds a new service to the call list, it doesn't start executing until
/// start("servicename") is called.
/// @param service - the pointer to the service to add.
//=============================================================================
bool GServiceManager::addService( GService* service )
{
    //===============================================================
    // Check to see if the service has already been loaded
    //===============================================================
    map<string, GService*>::iterator sitr = 
                        m_self->m_services.find( service->getServiceName() );
    if( sitr != m_self->m_services.end() )
    {
        Tracer::warn( 1, "Service with name \"%s\" already registered.", 
                        sitr->first.c_str() );
        return false;
    }

    //============================================================
    // Add the service to the list of known services
    //============================================================
    m_self->m_services[service->getServiceName()] = service;
//    Tracer::stat( 2, "Loaded %s service.", service->getServiceName().c_str() );

    //============================================================
    // Insert the service into the call list based on its priority
    // if it has a priority > 0 otherwise return
    //============================================================
    if( service->getPriority() < 0 ) return true;

    vector<GService*>::iterator itr = m_self->m_ordered.begin();
    while( itr != m_self->m_ordered.end() )
    {
        if( (*itr)->getPriority() > service->getPriority() )
        {
            m_self->m_ordered.insert( itr, service );
            return true;
        }
        itr++;
    }
    // Add it to the end if it wasn't inserted before
    m_self->m_ordered.push_back( service );
    return true;
}

//=============================================================================
/// @brief Goes through all known services and creates the call list.
///
/// It will include all services with a priority > 0 and list them in 
/// order from highest priority, 1 to lowest priority infinity ( maximum signed
/// int value for your system ).
//=============================================================================
void GServiceManager::updateSchedule()
{
    vector<GService*> temp;
    map<string, GService*>::iterator itr = m_self->m_services.begin();
    while( itr != m_self->m_services.end() )
    {
        //============================================================
        // Skip Services of lower priority
        //============================================================
        if( itr->second->getPriority() < 0 )
        {
            itr++;
            continue;
        }
        
        //============================================================
        /// Insert the service into the call list based on its priority
        //============================================================
        vector<GService*>::iterator oitr = temp.begin();
        while( oitr != temp.end() )
        {
            if( (*oitr)->getPriority() > itr->second->getPriority() )
            {
                
                temp.insert( oitr, itr->second );
                goto continue_outer_loop;
            }
            oitr++;
        }
        // Add it to the end if it wasn't inserted before
        temp.push_back( itr->second );
        continue_outer_loop:
        itr++;
    }

    // Assign the sorted service list to the call order list
    m_self->m_ordered = temp;
}

//=============================================================================
/// @brief Removes the service from the list.
/// @param name - the name of the service to remove.
/// @return true if the service was removed successfully, else returns false.
//=============================================================================
bool GServiceManager::removeService( const string& name )
{
#ifdef DEBUG
    if( !m_self )
    { 
        Tracer::error( "GServiceManager::init() needs to be"
                       " called before GServiceManager::removeService()");
        return false;
    }
#endif        
    map<string, GService*>::iterator itr = 
                        m_self->m_services.find( name );
    if( itr == m_self->m_services.end() )
    {
        Tracer::warn( 1, "Couldn't remove unknown service \"%s\".", name.c_str() );
        return false;
    }
    m_self->m_services.erase( itr );
    updateSchedule();
    
    Tracer::stat( 1, "Removed service \"%s\"", name.c_str() );
    return true;
}

//=============================================================================
/// @brief Calls preConfig for each service.
//=============================================================================
void GServiceManager::preConfig()
{
    if( m_config )return;
    Tracer::stat( 5, "==================================================");
    Tracer::stat( 5, "Service Name\t\tPriority\tRunning");
    Tracer::stat( 5, "==================================================");
    m_curservice = 0;
    GService* serv = NULL;
    while( (serv = getNextService()) )
    {
        if( serv->isRunning() )
        {
            if( serv->getServiceName().size() >= 12 )
                Tracer::stat( 5, "%s\t\t%d\t\t  yes", 
                               serv->getServiceName().c_str(),
                               serv->getPriority());
            else
                Tracer::stat( 5, "%s\t\t\t%d\t\t  yes", 
                               serv->getServiceName().c_str(),
                               serv->getPriority());
            serv->preConfig();
        }
        else
        {
            Tracer::stat( 5, "%s\t\t%d\t\t    no",
                           serv->getServiceName().c_str(),
                           serv->getPriority());
        }
    }
    Tracer::stat( 5, "==================================================");
    Tracer::stat( 3, "Pre-Config Complete");
}

//=============================================================================
/// @brief Calls config for each service.
//=============================================================================
void GServiceManager::config()
{
    if( m_config )return;
    m_curservice = 0;
    GService* serv = NULL;
    while( (serv = getNextService()) )
        if( serv->isRunning() ) serv->config();
    Tracer::stat( 3, "Config Complete");
}

//=============================================================================
/// @brief Calls postConfig for each service.
//=============================================================================
void GServiceManager::postConfig()
{
    if( m_config )return;
    m_curservice = 0;
    GService* serv = NULL;
    while( (serv = getNextService()) )
        if( serv->isRunning() ) serv->postConfig();
    Tracer::stat( 3, "postConfig Complete");
    m_config = true;
}

//=============================================================================
/// @brief Calls the configuration stages on all of the services
///
/// This calls the following three functions for each service:
/// preconfig
/// config
/// postconfig
//=============================================================================
void GServiceManager::configApp()
{
#ifdef DEBUG
    if( !m_self )
    { 
        Tracer::error( "GServiceManager::init() needs to be"
                       " called before GServiceManager::config()");
        return;
    }
#endif        
    if( m_config ) return;
    preConfig();
    config();
    postConfig();
}

//=============================================================================
/// @brief Calls sync for each service.
//=============================================================================
void GServiceManager::sync()
{
    m_curservice = 0;
    GService* serv = NULL;
    while( (serv = getNextService()) )
        if( serv->isRunning() ) serv->sync();
}

//=============================================================================
/// @brief Calls preProcess for each service.
//=============================================================================
void GServiceManager::preProcess()
{
    m_curservice = 0;
    GService* serv = NULL;
    while( (serv = getNextService()) )
        if( serv->isRunning() ) serv->preProcess();
}

//=============================================================================
/// @brief Calls process for each service.
//=============================================================================
void GServiceManager::process()
{
    m_curservice = 0;
    GService* serv = NULL;
    while( (serv = getNextService()) )
        if( serv->isRunning() ) serv->process();
    if( m_sleeptime > 0 )
#ifdef WIN32
		Sleep( m_sleeptime / 1000 );
#else
		usleep( m_sleeptime );
#endif /* WIN32 */
}

//=============================================================================
/// @brief Calls postProcess for each service.
//=============================================================================
void GServiceManager::postProcess()
{
    m_curservice = 0;
    GService* serv = NULL;
    while( (serv = getNextService()) )
        if( serv->isRunning() ) serv->postProcess();
}

//=============================================================================
/// @brief Execute one iteration of the main application loop.
///
/// The following 4 functions are called in order:
/// sync
/// preProcess
/// process
/// postProcess
//=============================================================================
void GServiceManager::frame()
{
    sync();
    preProcess();
    process();
    postProcess();
}

//=============================================================================
/// @brief Runs in an infinite loop until GServiceManager::quit() is called.
///
/// This runs the system, the user has no control over the run loop.
/// It will configure the application if you have not configured it.
/// This function calls the low level functions in the order described below.
///
/// The following are called once following intialization of the application:
/// preconfig()
/// config()
/// postconfig()
//
/// The following are called in a loop in the following order after configuration:
/// sync()
/// preProcess()
/// process()
/// postProcess()
//=============================================================================
void GServiceManager::run()
{
#ifdef DEBUG
    if( !m_self )
    { 
        Tracer::error( "GServiceManager::init() needs to be"
                       " called before GServiceManager::quit()");
        return;
    }
#endif        
    configApp();
    Tracer::stat( 1, "Service Manager Running");
    m_self->m_running = true;
    while( m_self->m_running )
        frame();
}

//=============================================================================
/// @brief Exits the run() loop.
//=============================================================================
void GServiceManager::quit()
{
#ifdef DEBUG
    if( !m_self )
    { 
        Tracer::error( "GServiceManager::init() needs to be"
                       " called before GServiceManager::quit()");
        return;
    }
#endif        
    m_self->m_running = false;
}

//=============================================================================
/// @brief Starts the specified service.
/// @param name - the name of the service to start.
/// @return true if the service was started successfully, else returns false.
//=============================================================================
bool GServiceManager::startService( const string& name )
{   
#ifdef DEBUG
    if( !m_self )
    { 
        Tracer::error( "GServiceManager::init() needs to be"
                       " called before GServiceManager::startService()");
        return false;
    }
#endif        
    map<string, GService*>::iterator itr = 
                        m_self->m_services.find( name );
    if( itr == m_self->m_services.end() )
    {
        Tracer::error("Couldn't start unknown service \"%s\".", name.c_str() );
        return false;
    }
    Tracer::stat( 1, "Starting service \"%s\"", name.c_str() );
    return itr->second->start();
}

//=============================================================================
/// @brief Stops the specified service.
/// @param name - the name of the service to stop.
/// @return true if the service was stopped successfully, else returns false.
//=============================================================================
bool GServiceManager::stopService( const string& name )
{   
#ifdef DEBUG
    if( !m_self )
    { 
        Tracer::error( "GServiceManager::init() needs to be"
                       " called before GServiceManager::stopService()");
        return false;
    }
#endif        
    map<string, GService*>::iterator itr = 
                        m_self->m_services.find( name );
    if( itr == m_self->m_services.end() )
    {
        Tracer::warn("Couldn't stop unknown service \"%s\".", name.c_str() );
        return false;
    }
    Tracer::stat( 1, "Stopping service \"%s\"", name.c_str() );
    return itr->second->stop();
}

//=============================================================================
/// @brief Get a pointer to the specified service.
/// @param name - the name of the service for which to retrieve a pointer.
/// @return a pointer to the requested service, or NULL if the service could
/// not be found.
//=============================================================================
GService* GServiceManager::getService( const string& name )
{
#ifdef DEBUG
    if( !m_self )
    { 
        Tracer::error( "GServiceManager::init() needs to be"
                       " called before GServiceManager::getService()");
        return NULL;
    }
#endif        
    map<string, GService*>::iterator itr = m_self->m_services.find(name);
    if( itr == m_self->m_services.end() )
    {
        Tracer::warn( 5,  "GServiceManager::getService( %s )\n"
                      "\t\tUnknown service: \"%s\" ", name.c_str(),
                      name.c_str());
        return NULL;
    }
    return itr->second;
}

//=============================================================================
/// @brief Return the state of the service manager.
/// @return true if the service manager is running, else false.
//=============================================================================
bool GServiceManager::isRunning()
{
    return m_self->m_running;
}

//=============================================================================
/// @brief Starts the service manager.
///
/// Use this if you want to manually control your own applications
/// and do not want to manually call config for each service.
//=============================================================================
void GServiceManager::start()
{
    configApp();
    m_self->m_running = true;
}

//=============================================================================
/// @brief Stops all services that aren't in the list and and sets the priority
/// based on the index in the vector.
/// @param order - a vector containing the names of the services in the order
/// to be set.
//=============================================================================
void GServiceManager::setServiceOrder( const vector<string>& order )
{
    for( map<string, GService*>::iterator itr = m_self->m_services.begin();
         itr != m_self->m_services.end(); itr++ )
    {
        itr->second->setPriority( -1, false);
    }
    for( unsigned int i = 0; i < order.size(); i++ )
    {
        map<string, GService*>::iterator itr = 
                                        m_self->m_services.find(order[i]);
        if( itr == m_self->m_services.end() )
        {
            Tracer::warn( 4, "GServiceManager::setServiceOrder(..)\n"
                             "\t Unknown service %s", order[i].c_str() );
            continue;
        }
        itr->second->setPriority(i, false);
    }
    updateSchedule();
}
