#ifndef _GSERVICEMANAGER_H_
#define _GSERVICEMANAGER_H_
#include<commander/CommandHandler.h>
#include<string>
#include<vector>
#include<map>
using namespace std;

#ifdef WIN32
#ifdef SERVICEMGR_EXPORTS
#define SERVICEMGR_API __declspec(dllexport)
#else
#define SERVICEMGR_API __declspec(dllimport)
#endif /* SERVICEMGR_EXPORTS */
#else
#define SERVICEMGR_API
#endif /* WIN32 */

class GService;

//=============================================================================
/// @class GServiceManager
/// @author Daniel Larimer / Andrew A. Ray
/// @brief Manages the various services used by DIVERSE.
///
/// A static object which must be initialized before any functions are called.
/// Manages different DIVERSE services like DGL, DOSG, and DPF which need
/// pre/post frame process callbacks in additon to frame call backs.  
/// Uses the PluginManager to load additional services. These
/// services are run in a round-robin fashion in a tight loop. You can use the
/// start() function instead of run() if you want more control over the application
/// stages.
///
/// The application stages are discussed below.
/// The following are called once following intialization of the application:
/// preConfig() - prior to window and OpenGL context initialization.
/// config() - initializes the window and the OpenGL context.
/// postConfig() - configuration required after the window and context have been created.
///
/// The following are called in a loop in the following order after configuration:
/// sync() - syncronizes the displays.
/// preProcess() - processing necessary prior to drawing the frame.
/// process() - processing for drawing the frame.
/// postProcess() - processing required after drawing the frame.
///
/// Future work: Eventually services may run as seperate threads splitting up
/// the work between different processors.
//=============================================================================
class SERVICEMGR_API GServiceManager : public CommandHandler
{
    public:
        static void init();
        static void exit(); 

        static void run();

        static void configApp();
        
        static void start();
        
        static void frame();
    
        //Now have all of the low level parts of the system
        //You MUST call these in the specific order at the proper
        //times or the system will NOT work in the proper time
        ///This preconfigures the system
        static void preConfig();
        
        ///This configs the system
        static void config();
        
        ///This postconfigs the system
        static void postConfig();
        
        ///Make sure you call start before you start your loop with
        ///sync and the rest of the main event loop functions
        //This syncs all the displays, in case you need extra data
        static void sync();
        
        //Pre process
        static void preProcess();
        ///This runs one frame
        static void process();
        //Post process
        static void postProcess();
        ///This quits the application
        static void quit();

        ///This will tell you if the application is running
        static bool isRunning();
        
        ///Add a service
        static bool addService   ( GService* service );
        
        ///Start a specific service
        static bool startService ( const string& name );
        
        ///Stop a service
        static bool stopService  ( const string& name );
        
        ///Remove a service
        static bool removeService( const string& name );
        
        //Get a specific service from the service manager
        static GService* getService(const string& name);

        static void setupRegistry( const string& cfg = "diverse.xml" );
        
        static void              updateSchedule();
        static void              setServiceOrder( const vector<string>& order );
        static vector<GService*> getServiceOrder()
                                        { return m_self->m_ordered; }

        CommandOut* execute( const Command& cmd );
    private:
        //==============================================
        // Used for scheduling only, internal functions
        //==============================================
        static unsigned int     m_curservice;
        static GService*        getNextService();
        //==============================================
        
        GServiceManager(){};
        static GServiceManager* m_self;
        static bool             m_config;
        static bool             m_setup;

        /// The time to sleep each frame
        static int              m_sleeptime;
        
        static bool             m_running;
        vector<GService*>       m_ordered;
        map<string, GService*>  m_services;
};

#endif
