#include"CClient.h"
#include<sys/types.h>
#ifdef WIN32
#define getpid GetCurrentProcessId
#else
#include<unistd.h>
#endif
#include<qstring.h>
#include<pexpr/PostOffice.h>
#include<pexpr/POTextMsg.h>
#include<pexpr/POErrorMsg.h>
#include<tracer/Tracer.h>
#include"CSSegmentListMsg.h"
#include"CSCreateSegmentMsg.h"
#include<dutil/DTime.h>
#include<pexpr/POBox.h>
#include<qapplication.h>
#include"ShMem.h"
#include<iostream>

//=============================================================================
/// @brief Constructor.
/// @param cname - the name of the client.
/// @param po - the PostOffice to which the client connects.
/// @param server - the address of the server.
//=============================================================================
CClient::CClient( const string& cname, PostOffice* po, const POAddress& server )
{
    m_po = po;
    m_serveradr = server;
    m_mbox = m_po->registerPOBox( cname );
    m_serveradr = POAddress( "CommuneServer", "CommuneServer" );
    CSSegmentListMsg msg;
    m_mbox->sendMessage( &msg, m_serveradr );
}

//=============================================================================
/// @brief Constructor.
/// @param cname - the name of the client.
/// @param hostname - the name of the host.
/// @param port - the port which should be used for shared memory.
//=============================================================================
CClient::CClient( const string& cname, const string& hostname, int port )
{
    int pid = getpid();
    m_po = new PostOffice( ("CommuneClient_" + QString::number(pid)).latin1() );
    m_mbox = m_po->registerPOBox( cname );
    m_po->connectTo( hostname, port, true/*block*/ );
    m_serveradr = POAddress( "CommuneServer", "CommuneServer" );
    CSSegmentListMsg msg;
    m_mbox->sendMessage( &msg, m_serveradr );

    connect( m_mbox, SIGNAL( messageArrived() ),
             this,   SLOT  ( handleMessage() ) );
}

//=============================================================================
// Function Name: createSegment
/// @brief Creates a new shared memory segment of the specified size and name
/// @param name - the name of the memory segment.
/// @param size - the size of the segment to create.
/// @param usec_timeout - the number of milliseconds to wait before timing out.
/// @return true if the segment could be created.
//=============================================================================
bool CClient::createSegment( const string& name, int size, int usec_timeout )
{
    Tracer t("CClient::createSegment()");
    if( getSegment( name, false, usec_timeout ) ) 
    {
        Tracer::warn("CClient::createSegment( %s, %d )\n"
                     "\tSegment Exists", name.c_str(), size );
        return false;
    }
    CSCreateSegmentMsg msg( name, size );
    m_mbox->sendMessage( &msg, m_serveradr );
    return true;
}

//=============================================================================
/// @brief Removes a previously created shared memory segment.
/// @param name - the name of the memory segment.
/// @param usec_timeout - the number of milliseconds to wait before timing out.
/// @return true if the segment was removed.
//=============================================================================
bool CClient::removeSegment( const string& name, int usec_timeout )
{
    POTextMsg discon("DISCONNECT");
    
    map<string, ShMem*>::iterator itr = m_segs.find(name);
    if( itr != m_segs.end() )
    {
 //       ShMem->closeSegment();
        delete itr->second;
    }
    return false;
}

//=============================================================================
// Function Name: getSegment
/// @brief Returns a pointer to the memory segment, and optionally creates
/// a new memory segment if it doesn't exist.
/// @param name - the name of the segment.
/// @param create - true if you want to create the segment if it doesn't exist.
/// @param usec_timeout - the amount of time to wait for a response.
/// @return a pointer to the shared memory segment or NULL on error or if no
/// segment exists.
//=============================================================================
ShMem* CClient::getSegment( const string& name, bool create, int usec_timeout )
{
    Tracer t("CClient::getSegment()");
    DTime time;
    time.start();
    
    map<string, ShMem*>::iterator itr = m_segs.find(name);
    if( itr != m_segs.end() )
        return (ShMem*)itr->second;
    while( time.elapsed().usec() < usec_timeout )
    {
        for( unsigned int i = 0; i < m_segnames.size(); i++ )
        {
            if( m_segnames[i] == name )
            {
                 POAddress segadr( m_serveradr.getPO(), name );
                 m_segs[name] = new ShMem( m_po, segadr );
                 return m_segs[name];
            }
        }
        if( create )
        {
            createSegment( name, 0, usec_timeout - time.elapsed().usec() );
            return getSegment(name, false, usec_timeout  );
        }
        qApp->processEvents();
    }
    Tracer::debug( "Couldn't find segment in list, returning NULL\n");
    return NULL;
}

//=============================================================================
/// @brief Get a list of the names of the segments.
/// @return a vector containing the list.
//=============================================================================
vector<string> CClient::getSegmentNames()
{
    return m_segnames;
}

//=============================================================================
/// @brief Get a map of connected segments.
/// @return a map of shared memory names to the shared memory segments.
//=============================================================================
map<string, ShMem*> CClient::getConnectedSegments()
{
    return m_segs;
}

//=============================================================================
/// @brief Handles all messages of type "MSWriteMsg".
//=============================================================================
void CClient::handleMessage()
{
    const POMessage* msg = NULL;
    while( m_mbox->hasMessages() )
    {
        msg = m_mbox->popMessage();
        if( msg->messageType() == "CSSegmentListMsg" )
            handleSegmentListMsg( (CSSegmentListMsg*)msg );
        delete msg;
    }
}

//=============================================================================
/// @brief Handles all list segment messages.
/// @param msg - the list message to be handled.
//=============================================================================
void CClient::handleSegmentListMsg( CSSegmentListMsg* msg )
{
    if( msg->getCommand() == CSSegmentListMsg::STATUS )
    {
        m_segnames = msg->getSegments();
    }
    else if( msg->getCommand() == CSSegmentListMsg::NEWSEGMENTS )
    {
        vector<string> newsegs = msg->getSegments();
        for( unsigned int i = newsegs.size(); i < newsegs.size(); i++ )
            m_segnames.push_back( newsegs[i] );
    }
    else
    {
        Tracer::error( "CClient::handleSegmentListMsg( "
                       "CSSegmentListMsg::DELETESEGMENTS )"
                       "\n\t Hasn't been implimented" );
    }
}

//=============================================================================
// Function Name: mapSegment
/// @brief Sends a message to the CommuneServer telling it to map a segment
/// from a remote CommuneServer to the local one.
//=============================================================================
int CClient::mapSegment( const string& host, int port,
                         const string& rsegment, const string& lsegment,
                         int usec_timeout )
{
    QString str = QString("map:") + host.c_str() + ':' + 
                  QString::number(port) + ':' +
                  rsegment.c_str() + ':' + QString( lsegment.c_str() );
    POTextMsg msg(str.latin1());
    m_mbox->sendMessage( &msg, m_serveradr );
    return 0;
}

//=============================================================================
// Function Name: unmapSegment
/// @brief Tells the CommuneServer to unmap a remote shared memory segment.
//=============================================================================
int CClient::unmapSegment( const string& host, const string& rseg, 
                           const string& lseg, int usec_timeout )
{
    QString str = QString("unmap:") + host.c_str() + ':' + 
                  rseg.c_str() + ':' + QString( lseg.c_str() );
    POTextMsg msg(str.latin1());
    m_mbox->sendMessage( &msg, m_serveradr );
    return 0;
} 

//=============================================================================
/// @brief Handles error messages.
/// @param msg - the error messgae to be handled.
//=============================================================================
void CClient::handleErrorMsg( POErrorMsg* msg )
{
    Tracer::error( "CClient::handleErrorMsg( %s ) not implimented",
                  msg->getText().c_str() );
}

//=============================================================================
/// @brief Handles text messages.
/// @param msg - the text message to be handled.
//=============================================================================
void CClient::handleTextMsg( POTextMsg* msg )
{
    Tracer::warn( "CClient::handleTextMsg( %s ) not implimented",
                  msg->getText().c_str() );
}
