#include"CommuneServer.h"
#include<pexpr/PostOffice.h>
#include<pexpr/POBox.h>
#include<pexpr/POMessage.h>
#include<pexpr/POAddress.h>
#include<pexpr/POErrorMsg.h>
#include<pexpr/POTextMsg.h>
#include"CSCreateSegmentMsg.h"
#include"CSSegmentListMsg.h"
#include"MemorySegment.h"
#include"CClient.h"
#include"ShMem.h"
#include<dutil/DTime.h>
#include<dutil/ColonList.h>
#include<tracer/Tracer.h>
#ifdef WIN32
#define getpid GetCurrentProcessId
#else
#include<unistd.h>
#endif /* WIN32 */
#include<iostream>

//=============================================================================
/// @brief Constructor.
/// @param name - the name of the CommuneServer.
/// @param port - the port which should be used for shared memory.
//=============================================================================
CommuneServer::CommuneServer( const string& name, int port )
{
    m_po = new PostOffice( name );
    m_po->servicePort( port );
    m_box = m_po->registerPOBox( "CommuneServer" );

    connect( m_box, SIGNAL( messageArrived() ),
             this, SLOT( handleMessage() ) );
}

//=============================================================================
/// @brief Handles all messages for the memory segment.
//=============================================================================
void CommuneServer::handleMessage()
{
    const POMessage* msg = NULL;
    while( m_box->hasMessages() )
    {
        msg = m_box->popMessage();
        if( msg->messageType() == "CSCreateSegmentMsg" )
            handleCreateSegment( (const CSCreateSegmentMsg*)msg );
        else if( msg->messageType() == "CSSegmentListMsg" )
            handleListSegments( (const CSSegmentListMsg*)msg );
        else if( msg->messageType() == "POTextMsg" )
            handleMapMessage( (const POTextMsg*)msg );
        delete msg;
    }
}

//=============================================================================
// Function Name: handleMapMessage
/// @brief Parses a text message for a command in the form 
/// map:host:port:remotesegment:localsegment 
/// and then executes the command.
/// @param msg - the message containing the command.
//=============================================================================
void CommuneServer::handleMapMessage( const POTextMsg* msg )
{
    ColonList cmd(msg->getText());
    // Check for errors
    if( cmd[0] == "map" )
    {
        if( cmd.size() < 5 )
        {
            Tracer::warn( "CommuneServer::handleMapMessage()\n"
                          "\t  expected text of the form \"map:host:"
                                      "port:remotesegment:localsegment\"\n"
                          "\t  but received message %s.", 
                          msg->getText().c_str() );
            return;
        }
        mapSegment( cmd[1], atoi(cmd[2].c_str()), cmd[3], cmd[4] );
    }
    else if( cmd[0] == "unmap" )
    {
        if( cmd.size() < 4 )
        {
            Tracer::warn( "CommuneServer::handleMapMessage()\n"
                          "\t  expected text of the form \"unmap:host:"
                                      "remotesegment:localsegment\"\n"
                          "\t  but received message %s.", 
                          msg->getText().c_str() );
            return;
        }
        unmapSegment( cmd[1], cmd[2], cmd[3] );
    }
}

//=============================================================================
// Function Name: mapSegment
/// @brief Maps shared memory from a remote system to the local one.
/// Used so changes are reflected on both systems.
/// @param host - hostname or IP of remote system.
/// @param port - port number of remote system.
/// @param rseg - the remote segment name.
/// @param lseg - the local segement name.
//=============================================================================
int CommuneServer::mapSegment( const string& host, int port,
                                const string& rseg, const string& lseg )
{
    Tracer::stat( "Mapping memory segment %s on host %s:%d to \n"
                  "\t\t local memory segment %s.", rseg.c_str(),
                  host.c_str(), port, lseg.c_str() );

    map<string, CClient*>::iterator itr = m_remoteshm.find( host );
    if( itr == m_remoteshm.end() )
    {
        // Create a client object to the host/port
        CClient* cl = new CClient( string("CommuneServerClient_")
                                   + QString::number(getpid()).latin1(),
                                   host, port );

        m_remoteshm[host] = cl;
        cl->createSegment(rseg, 0 );
      
        // 10 second time out if we can't get the segment
        DTime timeout;
        timeout.start();
        ShMem* mem = cl->getSegment( rseg, false, 1000000 );
        if( timeout.elapsed() >= DTime( 10, 0 ) )
        {
            Tracer::warn( "Couldn't map shared memory, connection"
                          "timed out after 10 seconds." );
            return 1; 
        }   
        if( mem == NULL )
        {
            Tracer::error( "CommuneServer::mapSegment() \n"
                          "\t\tCouldn't get remote segment %s", rseg.c_str() );
            exit(2);
            return 2;
        }
        createSegment( lseg, 0 );
        m_local2remote[lseg] = mem;
        m_remote2local[rseg] = m_memory[lseg];
        // When the remote shared memory changes, forward to the
        // local shared mem
        connect( mem, SIGNAL( changed( const string&, char*, int, int ) ),
                 this, SLOT( forwardRemoteChange( const string&, char*, int, int ) ) );

        // When the local shared mem changes, forward it to the remote
        connect( m_memory[lseg], SIGNAL( changed
                                    ( const string&, char*, int, int ) ),
                 this,           SLOT ( forwardLocalChange
                                    ( const string&, char*, int, int ) ) );
    }
    return 0;
}

//=============================================================================
// Function Name: handleCreateSegment
/// @brief Responds to create segment message.
//=============================================================================
void CommuneServer::handleCreateSegment( const CSCreateSegmentMsg* msg )
{
    if( createSegment( msg->getSegmentName(), msg->getSize() ) == 1 )
    {
        POErrorMsg errmsg("Couldn't create segment - another segnemt"
                          " with that name" );
        m_box->sendMessage( &errmsg, msg->getSource() );
        return;
    }
    CSSegmentListMsg listmsg;
    listmsg.addSegment( msg->getSegmentName() );
    m_box->sendMessage( &listmsg, msg->getSource() );
}

//=============================================================================
// Function Name: createSegment
/// @brief Creates a new segment.
/// @param segname - the name of the segment.
/// @param size - the size of the segment.
/// @return 0 on success, 1 if segment exists.
//=============================================================================
int CommuneServer::createSegment( const string& segname, int size )
{
    map<string, MemorySegment*>::iterator itr = m_memory.find(segname);
    if( itr != m_memory.end() )
        return 1;
    MemorySegment* seg = new MemorySegment( m_po, segname, 
                                                  size );
    m_memory[ segname ] = seg;
    return 0;
}

//=============================================================================
/// @brief Handles all list segment messages.
/// @param msg - the list message to be handled.
//=============================================================================
void CommuneServer::handleListSegments( const CSSegmentListMsg* msg )
{
    CSSegmentListMsg listmsg;
    map<string, MemorySegment*>::iterator itr = m_memory.begin();
    while( itr != m_memory.end() )
    {
        listmsg.addSegment( itr->first );
        itr++;
    }
    m_box->sendMessage( &listmsg, msg->getSource() );
}

//=============================================================================
/// @brief Not implemented.
//=============================================================================
int CommuneServer::unmapSegment( const string& host, const string& rsegment, 
                                      const string& lsegment )
{
    Tracer::error("CommuneServer::unmapSegment() - not implimented");
    return 0;
}

//=============================================================================
/// @brief Forward a change to a local segment.
/// @param lsegname - the name of the local segment to change.
/// @param buffer - the buffer containing the segment data.
/// @param start - the index of the start of the segment.
/// @param size - the size of the segment.
//=============================================================================
void CommuneServer::forwardLocalChange( const string& lsegname,
                                        char* buffer, int start, int size )
{
    map<string, ShMem*>::iterator itr = m_local2remote.find(lsegname);
    if( itr == m_local2remote.end() )
    {
        Tracer::warn( "CommuneServer::forwardLocalChange( %s, ...)\n"
                      "\t\tUnknown segment.", lsegname.c_str() );
        return;
    }
    itr->second->write( buffer, start, size );
}

//=============================================================================
/// @brief Forward a change to a remote segment.
/// @param rsegname - the name of the remote segment to change.
/// @param buffer - the buffer containing the segment data.
/// @param start - the index of the start of the segment.
/// @param size - the size of the segment.
//=============================================================================
void CommuneServer::forwardRemoteChange( const string& rsegname,
                                         char* buffer, int start, int size )
{
    map<string, MemorySegment*>::iterator itr = m_remote2local.find(rsegname);
    if( itr == m_remote2local.end() )
    {
        Tracer::warn( "CommuneServer::forwardRemoteChange( %s, ...)\n"
                      "\t\tUnknown segment.", rsegname.c_str() );
        return;
    }
    itr->second->write( buffer, start, size );
    
}
