#include"MemorySegment.h"
#include<pexpr/POTextMsg.h>
#include<pexpr/PostOffice.h>
#include<pexpr/POBox.h>
#include"MSWriteMsg.h"
#include<string.h>
#include<pexpr/POTextMsg.h>
#include<tracer/Tracer.h>

#include<iostream>
using namespace std;

//=============================================================================
/// @brief Constructor.
/// @param po - the PostOffice to which the memory segment belongs.
/// @param name - the name of the memory segment.
/// @param size - the size of the memory segment.
//=============================================================================
MemorySegment::MemorySegment( PostOffice* po, const string& name, int size )
    :m_name(name)
{
    m_box = po->registerPOBox( name );
    m_size = size;
    m_memory = new char[size];
    memset( m_memory, 0, size );
    connect( m_box, SIGNAL( messageArrived() ),
             this, SLOT( handleMessage() ) );
}

//=============================================================================
/// @brief Destructor.
//=============================================================================
MemorySegment::~MemorySegment()
{
    delete[] m_memory;
}

//=============================================================================
/// @brief Handles all messages for the memory segment.
//=============================================================================
void MemorySegment::handleMessage()
{
    const POMessage* msg = NULL;
    while( m_box->hasMessages(true) ) // limit latency and check sockets now
    {
        msg = m_box->popMessage();
        if( msg->messageType() == "POTextMsg" )
        {
            const POTextMsg* tmsg = (const POTextMsg*)msg;
            if( tmsg->getText() == "CONNECT" )
                handleConnect( tmsg );
            else if( tmsg->getText() == "DISCONNECT" )
                handleDisconnect( tmsg );
        }
        else if( msg->messageType() == "MSWriteMsg" )
            handleWrite( (const MSWriteMsg*)msg );
        delete msg;
    }
}

//=============================================================================
/// @brief Handles the creation of a connection to another POAddress.
/// @param msg - the message telling the segment to connect.
//=============================================================================
void MemorySegment::handleConnect( const POTextMsg* msg )
{
    // TBD Prevent Duplicate Connections
    m_connections.push_back( msg->getSource() );
    MSWriteMsg wmsg( m_memory, 0, m_size );
    m_box->sendMessage( &wmsg, m_connections.back() );
}

//=============================================================================
/// @brief Handles a disconnection between this POAddress and another.
/// @param msg - the message telling the segment to disconnect.
//=============================================================================
void MemorySegment::handleDisconnect( const POTextMsg* msg )
{
    vector<POAddress>::iterator itr = m_connections.begin();
    while( itr != m_connections.end() )
    {
        if( *itr == msg->getSource() )
        {
            m_connections.erase( itr );
            return;
        }
        itr++;
    }
}

//=============================================================================
/// @brief Resize the shared memory segment.
/// @param size - the new size of the segment.
//=============================================================================
void MemorySegment::resize( int size )
{
    char* tmp = new char[size];
    if( size > m_size )
        memcpy( tmp, m_memory, m_size );
    else
        memcpy( tmp, m_memory, size );
    delete[] m_memory;
    m_memory = tmp;
    m_size = size;
}

//=============================================================================
/// @brief Stores messages in itself and emits signals to allow the messages to
/// be processed.
/// @param msg - the message to be handled.
//=============================================================================
void MemorySegment::handleWrite( const MSWriteMsg* msg )
{
    if( msg->getStartIndex() + msg->getSize() > size() )
        resize( msg->getStartIndex() + msg->getSize() );
    memcpy( &m_memory[msg->getStartIndex()], msg->getData(), msg->getSize() );
//    write( msg->getData(), msg->getStartIndex(), msg->getSize() );
    MSWriteMsg* wmsg = (MSWriteMsg*)msg->clone();
    vector<POAddress> errors;
	unsigned int i;
    for( i = 0; i < m_connections.size(); i++ )
    {
        if( m_connections[i] == msg->getSource() ) continue;
        if( !m_box->sendMessage( wmsg, m_connections[i], true ) ) // limit latency
        {
            errors.push_back(m_connections[i]);
            Tracer::stat("Disconnecting from memory segment");
        }
    }
    delete wmsg;

    // we need to clean up and disconnect any address which we 
    // couldn't send to
    for( i = 0; i < errors.size(); i++ )
    {
        vector<POAddress>::iterator itr = m_connections.begin();
        while( itr != m_connections.end() )
        {
            if( *itr == errors[i] )
            {
                m_connections.erase( itr );
                continue;
            }
            itr++;
        }
    }
    emit changed();
    emit changed( name(), msg->getData(), 
                  msg->getStartIndex(), msg->getSize() );
}


//=============================================================================
// Function Name: write
/// @brief Writes the buffer to shared memory starting at the specified index.
/// @param buffer - a pointer to the buffer
/// @param start - where to start the copy
/// @param size - the size of the buffer
//=============================================================================
void MemorySegment::write( char* buffer, int start, int size )
{
    if( m_memory == NULL )
    {
        Tracer::warn( "MemorySegment::write(...) can't write until initialized" );
        return;
    }
    if( start+size > m_size )
    {
        Tracer::error( "MemorySegment::write( ..., %d, %d ) exceeds size %d"
                       " of memory segment %s", 
                       start, size, m_size, m_name.c_str() );
        exit(1);
    }
    MSWriteMsg wmsg(buffer, start, size);
    //==================================================
    // Brodcast changes
    //=================================================
    vector<POAddress> errors;
	unsigned int i;
    for( i = 0; i < m_connections.size(); i++ )
    {
        if( !m_box->sendMessage( &wmsg, m_connections[i], true ) ) // limit latency
        {
            errors.push_back(m_connections[i]);
            Tracer::stat("Disconnecting from memory segment");
        }
    }

    // we need to clean up and disconnect any address which we 
    // couldn't send to
    for( i = 0; i < errors.size(); i++ )
    {
        vector<POAddress>::iterator itr = m_connections.begin();
        while( itr != m_connections.end() )
        {
            if( *itr == errors[i] )
            {
                m_connections.erase( itr );
                continue;
            }
            itr++;
        }
    }
    memcpy( &m_memory[start], buffer, size );
//    emit changed();
//    emit changed( name(), buffer, 
//                  start, size );
}

//=============================================================================
// Function Name: read
/// @brief Copies shared memory to a user alocated block of the specified size 
/// from the given starting location.
/// @param buffer - the buffer to copy into.
/// @param start - where to start in the shared mem.
/// @param size - the number of bytes to read.
//=============================================================================
void MemorySegment::read( char* buffer, int start, int size )const
{
    if( start+size > m_size )
    {
        Tracer::error( "MemorySegment::read( ..., %d, %d ) exceeds size %d"
                       " of memory segment %s", 
                       start, size, m_size, name().c_str() );
        exit(1);
    }
    memcpy( buffer, &m_memory[start], size );
}

//=============================================================================
/// @brief Get the size of the memory segment.
/// @return the size of the memory segment.
//=============================================================================
int MemorySegment::size()const
{
    return m_size;
}

//=============================================================================
/// @brief Get the name of the memory segment.
/// @return the name of the memory segment.
//=============================================================================
string MemorySegment::name()const
{
    return m_name;
}
