#include"ShMem.h"
#include<pexpr/POTextMsg.h>
#include<pexpr/PostOffice.h>
#include<pexpr/POBox.h>
#include"MSWriteMsg.h"
#include<string.h>
#include<pexpr/POTextMsg.h>

#include<tracer/Tracer.h>
#include<iostream>

//=============================================================================
/// @brief Constructor.
/// @param po - the PostOffice to which the shared memory segment belongs.
/// @param segaddr - the address that the memory segment has been given.
//=============================================================================
ShMem::ShMem( PostOffice* po, const POAddress& segaddr )
{
    m_box = po->registerPOBox( segaddr.getBox() );
    m_size = 0;
    m_memory = NULL;
    m_segaddr = segaddr;
    POTextMsg con("CONNECT");
    m_addr = segaddr;
    m_box->sendMessage( &con, m_addr );

    connect( m_box, SIGNAL( messageArrived() ),
             this,  SLOT  ( handleMessage() ) );
}

//=============================================================================
/// @brief Destructor.
//=============================================================================
ShMem::~ShMem()
{
    m_box->getPostOffice()->closePOBox( m_segaddr.getBox() );
    delete[] m_memory;
}

//=============================================================================
/// @brief Handles all messages of type "MSWriteMsg"
//=============================================================================
void ShMem::handleMessage()
{
    const POMessage* msg = NULL;
    while( m_box->hasMessages(true) )
    {
        msg = m_box->popMessage();
        if( msg->messageType() == "MSWriteMsg" )
            handleWrite( (const MSWriteMsg*)msg );
        delete msg;
    }
}

//=============================================================================
/// @brief Stores messages in itself and emits signals to allow the messages to
/// be processed.
/// @param msg - the message to be handled.
//=============================================================================
void ShMem::handleWrite( const MSWriteMsg* msg )
{
    if( m_memory == NULL )
    {
        m_size = msg->getSize() + msg->getStartIndex();
        m_memory = new char[m_size];
    }
    if( m_size < msg->getStartIndex()+msg->getSize() )
    {
        Tracer::error( "ShMem::handleWrite(...)\n"
                       "\t\t Writing %d bytes from %d exceeds segment size %d" );
        exit(1);
    }
    memcpy( &m_memory[msg->getStartIndex()], msg->getData(), msg->getSize() );
    emit changed();
    emit changed( name(), msg->getData(), 
                  msg->getStartIndex(), msg->getSize() );
}

//=============================================================================
// Function Name: write
/// @brief Write the the buffer to shared memory.
/// @param buffer - a pre-allocated buffer of size >= the shared memory size.
//=============================================================================
void ShMem::write( void* buffer )
{
    write( (char*)buffer, 0, size() );
}

//=============================================================================
// Function Name: write
/// @brief Writes the buffer to shared memory starting at the specified index.
/// @param buffer - a pre-allocated buffer of size >= the shared memory size.
/// @param start - where to start the copy.
/// @param size - the size of the buffer.
//=============================================================================
void ShMem::write( char* buffer, int start, int size )
{
    if( m_memory == NULL )
    {
        Tracer::warn( "ShMem::write(...) can't write until initialized" );
        return;
    }
    if( start+size > m_size )
    {
        resize( start + size );
        Tracer::warn( "ShMem::write( ..., %d, %d ) exceeds size %d"
                       " of memory segment %s, resizing memory segment", 
                       start, size, m_size, m_addr.getPO().c_str() );
    }
    MSWriteMsg msg(buffer, start, size);
    m_box->sendMessage( &msg, m_addr, true );
    memcpy( &m_memory[start], buffer, size );
    emit changed();
    emit changed( name(), buffer, 
                  start, size );
}

//=============================================================================
// Function Name: read
/// @brief Read shared memory into a buffer.
/// DANGEROUS, use read( void*, int start, int size ) for safer operation.
/// @param buffer - a pre-allocated buffer of size >= the shared memory size.
/// @return the size of shared memory.
//=============================================================================
int ShMem::read( void* buffer )
{
    read( (char*)buffer, 0, size() );
    return size();
}
//=============================================================================
// Function Name: read
/// @brief Copies shared memory to a user alocated block of the specified size 
/// from the given starting location.
/// @param buffer - the buffer to copy into.
/// @param start - where to start in the shared mem.
/// @param size - the number of bytes to read.
//=============================================================================
void ShMem::read( char* buffer, int start, int size )const
{
    if( start+size > m_size )
    {
        Tracer::error( "ShMem::read( ..., %d, %d ) exceeds size %d"
                       " of memory segment %s", 
                       start, size, m_size, m_addr.getPO().c_str() );
        exit(1);
    }
    memcpy( buffer, &m_memory[start], size );
}

//=============================================================================
/// @brief Resize the shared memory segment.
/// @param size - the new size of the segment.
//=============================================================================
void ShMem::resize( int size )
{
    char* tmp = new char[size];
    if( size > m_size )
        memcpy( tmp, m_memory, m_size );
    else
        memcpy( tmp, m_memory, size );
    delete[] m_memory;
    m_memory = tmp;
    m_size = size;
}

//=============================================================================
/// @brief Get the size of the memory segment.
/// @return the size of the memory segment.
//=============================================================================
int ShMem::size()const
{
    return m_size;
}

//=============================================================================
/// @brief Get the name of the memory segment.
/// @return the name of the memory segment.
//=============================================================================
string ShMem::name()const
{
    return m_addr.getBox();
}
