#include"DSoundNode.h"
#include<tracer/Tracer.h>

//=============================================================================
/// @brief Constructor.
/// @param filename - the name of the sound file to open and use at this node.
/// @param prio - the priority of the sound node.
//=============================================================================
DSoundNode::DSoundNode( const string& filename, int prio )
{
    m_fname = filename;
    m_prio = prio;

    m_sndptr = FSOUND_Sample_Load( FSOUND_UNMANAGED, filename.c_str(),
                                   FSOUND_HW3D | FSOUND_LOOP_NORMAL, 0 );

    if( !m_sndptr )
    {
        Tracer::error( "DSoundNode::DSoundNode(...) - Failed to load sound file: %s",
                        filename.c_str() );
        return;
    }

    m_mindist = 4;
    m_maxdist = 10000;

    m_volume = 128; // 50%
    m_channel = -1;

    FSOUND_Sample_SetMinMaxDistance( m_sndptr, m_mindist, m_maxdist );
    FSOUND_Sample_SetMode( m_sndptr, FSOUND_LOOP_NORMAL );
}

//=============================================================================
/// @brief Destructor.
//=============================================================================
DSoundNode::~DSoundNode()
{
    FSOUND_Sample_Free( m_sndptr );
}

//=============================================================================
/// @brief Update the sound node to determine its properties based on position
/// and velocity.
//=============================================================================
void DSoundNode::update()
{
    float dt = m_lastupdate.reset().sec();
    m_vel = (m_curpos - m_lastpos)/dt; 
    m_lastpos = m_curpos;
    m_lastdir = m_curdir;
    m_lastup  = m_curup;

    FSOUND_3D_SetAttributes( m_channel, m_curpos.ptr(), m_vel.ptr() );
}

//=============================================================================
/// @brief Set the current position and orientation from the specified matrix.
/// @param mat - the matrix to use to calculate the position and orientation.
/// @param update - if true, sets the last position and orientation vectors to the new
/// position and orientation vectors, else only the current vectors are changed.
//=============================================================================
void DSoundNode::setPositionMatrix( const dMatrix& mat, bool update )
{
    m_curpos = mat * dVec3(0,0,0);
    m_curdir = mat * dVec3(0,1,0);
    m_curup  = mat * dVec3(0,0,1);
    if( update )
    {
        m_lastpos = m_curpos;
        m_lastdir = m_curdir;
        m_lastup  = m_curup;
    }
}

//=============================================================================
/// @brief Set the position and orientation from the specified vectors.
/// @param pos - the position vector.
/// @param dir - the direction vector.
/// @param up - the up vector.
/// @param update - if true, sets the last position and orientation vectors to the new
/// position and orientation vectors, else only the current vectors are changed.
//=============================================================================
void DSoundNode::setPosition( const dVec3& pos, const dVec3& dir, const dVec3& up, bool update )
{
    m_curpos = pos;
    m_curdir = dir;
    m_curup = up;
    if( update )
    {
        m_lastpos = m_curpos;
        m_lastdir = m_curdir;
        m_lastup  = m_curup;
    }
}

//=============================================================================
/// @brief Play the sound
//=============================================================================
void DSoundNode::play()
{
    m_channel = FSOUND_PlaySoundEx( FSOUND_FREE, m_sndptr, NULL, true );
    FSOUND_3D_SetAttributes( m_channel, m_curpos.ptr(), m_vel.ptr() );
    FSOUND_SetVolume( m_channel, m_volume );
    if( !FSOUND_SetPaused( m_channel, false ) )
        Tracer::error( "DSoundNode::play(%s) - %s",
                        m_fname.c_str(), FMOD_ErrorString( FSOUND_GetError() ) );
}

//=============================================================================
/// @brief Pause the sound.
//=============================================================================
void DSoundNode::pause()
{
    FSOUND_SetPaused( m_channel, true );
}

//=============================================================================
/// @brief Determines if the sound is playing.
/// @return true if the sound is playing, else returns false.
//=============================================================================
bool DSoundNode::isPlaying()
{
    return FSOUND_GetPaused( m_channel );
}

//=============================================================================
/// @brief Set the minimum and maximum distances at which the sound will be heard.
///
/// The minimum distance which is the closest distance to the node at which the
/// sound will stop increasing in volume. Between the minimum and maximum distance
/// the volume will decrease. The maximum distance is the distance at which the
/// sound will no longer be played and will not be heard.
/// @param min - the minimum distance.
/// @param max - the maximum distance.
//=============================================================================
void DSoundNode::setMinMaxDistance( float min, float max )
{
    FSOUND_Sample_SetMinMaxDistance( m_sndptr, min, max );
}

//=============================================================================
/// Set the volume for the sound node.
/// @param vol - the volume of the sound.
///
/// Valid ranges are from 0 (silent) to 255 (full volume).
//=============================================================================
void DSoundNode::setVolume( int vol )
{
    m_volume = vol;
    FSOUND_SetVolume( m_channel, vol );
}

//=============================================================================
/// @brief Set the priority of the sound node.
/// @param p - the priority value to be set.
///
/// Valid range of values are from 0 (lowest) to 255 (highest).
//=============================================================================
void DSoundNode::setPriority( int p )
{
    m_prio = p;
    FSOUND_SetPriority( m_channel, p );
}
