//=============================================================================
// Tracer.h
//  - This class provides an elegant way to manage printing the stack as 
//  different functions are called and indents their start and stop tags to
//  make seeing recursion easy
//  
//  Created 10/5/02 By Daniel Larimer
//=============================================================================
#ifndef _TRACER_H_
#define _TRACER_H_
#include<string>
#include<vector>

#ifdef WIN32
#ifdef TRACER_EXPORTS
#define TRACER_API __declspec(dllexport)
#else
#define TRACER_API __declspec(dllimport)
#endif /* TRACER_EXPORTS */
#else
#define TRACER_API
#endif /* WIN32 */

using namespace std;

//=============================================================================
/// @class Tracer
/// @brief The tracer is used to spew formatted output to the console window.
///
/// Tracer spew output has several types ( TRACE, WARN, DEBUG, STAT, and ERROR )
/// and the level of spew can be set for each except for TRACE which is either on
/// or off.
///
/// The tracer spew output is set through the environment variable 
/// TRACER_SPEW or can also be set using the QDDisplay Settings dialog in the Tracer
/// Panel. If you are using a saved configuration file ( diverse.xml ), the tracer
/// settings can be set within the configuration file. The arguments listed in the
/// TRACER_SPEW environment variable should be separated by a colon and can include
/// any combination of TRACE, DEBUG, WARN, STAT, and NOERROR.
///
/// Example:
///
///     > export TRACER_SPEW=TRACE:DEBUG:WARN:STAT
///
/// Tracer levels are used to control the amount of output that is generated.
/// If a tracer statement is sent with a level equal to or less than the set
/// level then it will be output. If the tracer level of the message is less
/// than the level set then it will not be displayed. Using the environment
/// variable will result in default levels for each type of spew.
///
/// Defaults:
///
///     TRACE   - true ( on or off - no levels )
///
///     DEBUG   - 5
///
///     WARN    - 5
///
///     STAT    - 5
///
///     NOERROR - 0 ( No errors will be shown if this parameter is passed through
///                   TRACER_SPEW )
///
/// The levels of spew can be changed through the QDDisplay Settings dialog in Tracer
/// Panel.
///
/// The character string passed to the tracer functions uses standard formatting
/// as is used by standard C printf functions.
//=============================================================================
class TRACER_API Tracer 
{
    public:
        static void init();
        Tracer( const string& enter, bool print = true );
        ~Tracer();
        static void enter( const string& function, bool print = true );
        static void debug  ( const char* format, ... );
        static void error  ( const char* error, ... );
        static void warn   ( const char* warn, ...  );
        static void stat   ( const char* stat, ...  );
        static void debug  ( int level, const char* format, ... );
        static void error  ( int level, const char* error, ... );
        static void warn   ( int level, const char* warn, ...  );
        static void stat   ( int level, const char* stat, ...  );

        template<class ttype>
        static ttype& leave( ttype& rtn_value );
        static void leave();
	//===================================================================
	/// @brief Similar to leave.
	/// @param x - an integer value.
	/// @return an integer value that was passed as the x parameter.
	//===================================================================
	static int leave( int x ) { leave(); return x; }
        static void printTrace ( bool state ) { m_printtrace = state; }
        static void printDebug ( int state )  { m_printdebug = state; }
        static void printError ( int state ) { m_printerror = state; }
        static void printWarn  ( int state ) { m_printwarn  = state; }
        static void printStatus( int state ) { m_printstat  = state; }
        static void setIndentSpaces( int num ) { m_indent = num;     }
    private:
        static void printIndent();

        static bool m_initialized;
        static vector< pair<string, bool> > m_stack;
        static bool m_printtrace;
        static int m_printdebug;
        static int m_printerror;
        static int m_printwarn;
        static int m_printstat;
        static int  m_level;
        static int  m_indent;
};

//=============================================================================
//  Function Name: Tracer::leave
/// @brief Leaves a level, printing the function name on the level it was called from
/// and then decreasing the indent.
/// @param rtn_value - the value to be returned.
//=============================================================================
template<class ttype>
ttype& Tracer::leave(ttype& rtn_value)
{
    if( !m_printtrace || m_stack.size() == 0 ) return rtn_value;
    if( !m_stack.back().second ) 
    {
        m_stack.pop_back();
        return rtn_value;
    }
    printIndent();
    m_level--;
//    cout<<"Exit: "<<m_stack[m_stack.size()-1].first<<endl;
    m_stack.pop_back();
    return rtn_value;
}
#endif
