//=============================================================================
// wandJoystickNav.cpp
// - Navigation plugin class for wand joystick controller
//
// Created 6/4/03 By Patrick Shinpaugh
//=============================================================================

#include "wandJoystickNav.h"
#include <tracer/Tracer.h>
#include <dmath/dMatrix.h>
#include <dreg/Registry.h>
#include <eventmgr/EventManager.h>

using namespace std;

wandJoystickNav::wandJoystickNav( WJNAVTYPE navtype )
{
    m_navtype = navtype;
    m_active = false;
    NavManager::addNavigation( this );
    switch (navtype)
    {
        case WJNAV:
		NavManager::setNavigation("wandJoystickNav");
        case WJPITCHNAV:
		NavManager::setNavigation("wandJoystickPitchNav");
        case WJROLLNAV:
		NavManager::setNavigation("wandJoystickRollNav");
        case WJHPRNAV:
		NavManager::setNavigation("wandJoystickHPRNav");
    }
    NavManager::setNavigation("wandJoystickNav");
    m_deltapos = 0.0f;
    m_deltadir = 0.0f;

    EventManager::addWandListener( this );

    //===============================
    // Speed and Twist are per second
    //===============================
    m_speedfactor = atof( Registry::getValue( "/wandJoystickNav/Speed", "2.0").c_str() );
    m_twist = atof( Registry::getValue( "/wandJoystickNav/Twist", "45.0").c_str() );

    //===============================
    // Trans and Rot dead zones are
    // tolerances within which no
    // movement will be registered
    //===============================
    m_transdeadzone = atof( Registry::getValue( "/wandJoystickNav/TransDeadZone", "0.05").c_str() );
    m_rotdeadzone = atof( Registry::getValue( "/wandJoystickNav/RotDeadZone", "0.1").c_str() );

    Tracer::debug( "wandJoystickNav constructor finished" );

}

wandJoystickNav::~wandJoystickNav()
{
    stop();
}

dMatrix wandJoystickNav::getNavMatrix()const
{
    return m_navmat;
}

void wandJoystickNav::setNavMatrix( const dMatrix& navmat )
{
    m_navmat = navmat;
}

float wandJoystickNav::getSpeedFactor() const
{
    return m_speedfactor;
}

void wandJoystickNav::setRotateFactor( float rfactor )
{
    m_rfactor = rfactor;
}

void wandJoystickNav::setSpeedFactor( float speedfactor )
{
    m_speedfactor = speedfactor;
}

bool wandJoystickNav::onWandEvent( DWandEvent* evt )
{
    if ( !m_active )
        return false;

    if ( evt->action() == DWandEvent::JOYSTICKMOVE )
    {
        float x = evt->getJoystickX();
        float y = evt->getJoystickY();

        if ( fabs(x) <= m_rotdeadzone )
            m_deltadir = 0;
        else
            m_deltadir = 
                (x > 0 ? x - m_rotdeadzone : x + m_rotdeadzone) * m_twist * m_rfactor;

        if ( fabs(y) <= m_transdeadzone )
            m_deltapos = 0.0f;
        else
            m_deltapos = (y > 0 ? y - m_transdeadzone : y + m_transdeadzone) * m_speedfactor;

        return false;
    }
    else if ( evt->action() == DWandEvent::TRACK )
    {
        m_wandpos = evt->getPosition();
        m_wanddir = evt->getDirection();
        return false;
    }

    return false;
}

void wandJoystickNav::process()
{
    if ( !m_active )
        return;

    float delta_t = m_time.reset().sec();
    float delta_dir = DEG2RAD( m_deltadir * delta_t);
    float delta_pos = m_deltapos * delta_t;
    dMatrix rotmat;
                
    dVec3 wanddir( 0, 1, 0 );
    dMatrix euler;
    euler.makeRotate( DEG2RAD(m_wanddir.h()), dVec3( 0, 0, -1 ),
                      DEG2RAD(m_wanddir.p()), dVec3( -1, 0, 0 ),
                      DEG2RAD(m_wanddir.r()), dVec3( 0, 1, 0 ) );
    wanddir = euler * wanddir;
    
	m_velocity = wanddir * m_deltapos;

    switch (m_navtype)
    {
        //==============================
        // WJNAV - joystick left/right
        // rotates about world z axis
        //==============================
        case WJNAV:
        {
//            rotmat.rotate( delta_dir, 0.0f, 0.0f );
//            rotmat.makeRotate( delta_dir, 0, 0, 1 ); // rotate around Z  for heading
            rotmat.makeRotate( delta_dir, 0, 0, 1 );
            break;
        }
        //==============================
        // WJPITCHNAV - joystick left/right
        // rotates about world x axis
        //==============================
        case WJPITCHNAV:
        {
//            rotmat.rotate( 0.0f, delta_dir, 0.0f );
            rotmat.makeRotate( delta_dir, -1, 0, 0 );
            break;
        }
        //==============================
        // WJROLLNAV - joystick left/right
        // rotates about world y axis
        //==============================
        case WJROLLNAV:
        {
            rotmat.makeRotate( delta_dir, 0, -1, 0 );
            //rotmat.rotate( 0.0f, 0.0f, delta_dir );
            break;
        }
        //==============================
        // WJHPRNAV - joystick left/right
        // rotates about wand axis
        //==============================
        case WJHPRNAV:
        {
            if (delta_dir)
            {
                rotmat.makeRotate( delta_dir, wanddir.x(), wanddir.y(), wanddir.z() );
            }
            break;
        }
    }

    // translate to wand, rotate, translate back
    rotmat =  dMatrix::translate( -m_wandpos ) 
                * rotmat * dMatrix::translate( m_wandpos );

    // we have now rotated so that 0, 1, 0 is facing the right dir, now move in that dir

    dMatrix trans(rotmat * dMatrix::translate( wanddir * -delta_pos ));

    // now multiply this delta trans by the previous nav matrix
    //  PRVNAV * DRX * DTX 
    m_navmat = m_navmat * trans;
}

void wandJoystickNav::start()
{
    EventManager::addWandListener( this );
    m_speedfactor = NavManager::getSpeedFactor();
    m_rfactor = NavManager::getRotateFactor();
    m_active = true;
    m_time.reset();
    //COPIED FROM ON WAND EVENT TO UPDATE THE STATE
    DWandEvent* evt = EventManager::getWandState();
    //If we have no state, do nothing
    if (evt == NULL)
	    return;
    //Else update our state
        float x = evt->getJoystickX();
        float y = evt->getJoystickY();

        if ( fabs(x) <= m_rotdeadzone )
            m_deltadir = 0;
        else
            m_deltadir = 
                (x > 0 ? x - m_rotdeadzone : x + m_rotdeadzone) * m_twist * m_rfactor;

        if ( fabs(y) <= m_transdeadzone )
            m_deltapos = 0.0f;
        else
            m_deltapos = (y > 0 ? y - m_transdeadzone : y + m_transdeadzone) * m_speedfactor;

        m_wandpos = evt->getPosition();
        m_wanddir = evt->getDirection();
	m_velocity = m_wanddir * m_deltapos;
}

void wandJoystickNav::stop()
{
    EventManager::removeWandListener( this );
    m_active = false;
}

bool wandJoystickNav::isRunning() const
{
    return m_active;
}

string wandJoystickNav::getNavigationName()const
{
    switch (m_navtype)
    {
        case WJNAV:
            return "wandJoystickNav";
        case WJPITCHNAV:
            return "wandJoystickPitchNav";
        case WJROLLNAV:
            return "wandJoystickRollNav";
        case WJHPRNAV:
            return "wandJoystickHPRNav";
    }
    return "wandJoystickNav - Unknown";
}

dVec3 wandJoystickNav::getPivot()
{
    switch (m_navtype)
    {
        case WJNAV:
            return dVec3( 0, 0, 1 );
        case WJPITCHNAV:
            return dVec3( 1, 0, 0 );
        case WJROLLNAV:
            return dVec3( 0, 1, 0 );
        case WJHPRNAV:
		{
			dVec3 wandvec;
			wandvec.x() = -sin( DEG2RAD( m_wanddir.h() ) );
			wandvec.y() = cos( DEG2RAD( m_wanddir.h() ) );
			wandvec.z() = sin( DEG2RAD( m_wanddir.p() ) );
			wandvec.normalize();
            return wandvec;
		}
    }
    return dVec3();
}
