//=============================================================================
// DebugHeadsUp.cpp
// - Navigation plugin class for wand joystick controller
//
// Created 6/19/03 By Patrick Shinpaugh
//=============================================================================

#include "DebugHeadsUp.h"
#include <tracer/Tracer.h>
#include <dmath/dMatrix.h>
#include <navmgr/NavManager.h>
#include <navmgr/NavInterface.h>
#include <commander/CommandManager.h>
#include <commander/CommandOut.h>
#include <fontmgr/FontManager.h>
#include <fontmgr/dglfont.h>
#include <dgl/dglcore/DGL.h>
#include <dmath/Rect.h>
#include <GL/gl.h>
#include <stdarg.h>
#include <stdio.h>

REGISTER_PLUGIN(DebugHeadsUp);

using namespace std;

static char* float9( char *s, float f );
static char* format( char *s, const char *fmt, ... );

char* float9( char *s, float f )
{
    float a;
    a = f < 0 ? -f : f;

    if (a==0.0f)
        sprintf(s,"%-+8.6f",f) ;
    else if (a>=100000000.f || a<=.0001f)
        sprintf(s,"%-+8.2e",f) ;
    else if (a<10.f)
        sprintf(s,"%-+8.6f",f) ;
    else if (a<100.f)
        sprintf(s,"%-+8.5f",f) ;
    else if (a<1000.f)
        sprintf(s,"%-+8.4f",f) ;
    else if (a<10000.f)
        sprintf(s,"%-+8.3f",f) ;
    else if (a<100000.f)
        sprintf(s,"%-+8.2f",f) ;
    else if (a<1000000.f)
        sprintf(s,"%-+8.1f",f) ;
    else if (a<10000000.f)
        sprintf(s,"%-+8.0f",f) ;
    else
        sprintf(s,"%-+8.0f",f) ;
    return s;
}

static char* format( char *s, const char *fmt, ... )
{
    va_list args;
    va_start(args, fmt);
    vsprintf( s, fmt, args );
    va_end(args);
    return s;
}

DebugHeadsUp::DebugHeadsUp( const string& name )
:Plugin( name ), m_curfont(NULL)
{
    m_visible = false;
    EventManager::addKeyListener( this );
    setDescription( "DGL debug heads up display - toggle display of current info" );

    FontManager::init();
    CommandManager.addCommandHandler( "debugHeadsUp", this );

    Tracer::debug( "DebugHeadsUp constructor finished" );
}

CommandOut* DebugHeadsUp::execute( const Command& cmd )
{
    CommandOut* out = new CommandOut( cmd );
    if( cmd.getArgs().size() < 2 || (cmd.getArgs()[1] != "show" && cmd.getArgs()[1] != "hide") ) 
    {
        out->print("Usage: \n\tdebugHeadsUp [show|hide]");
        return out;
    }
    if( cmd.getArgs()[1] == "show" )
    {
        out->print("Showing Debug Heads Up Display" );
        m_visible = true;
    }
    if( cmd.getArgs()[1] == "hide" )
    {
        out->print("Hiding Debug Heads Up Display" );
        m_visible = false;
    } 
    return out;
}

bool DebugHeadsUp::onKeyEvent( DKeyEvent* evt)
{
    switch ( evt->action() )
    {
        case DKeyEvent::DOWN:
            if ( evt->ascii() == '?' || evt->ascii() == '/' )
                m_visible = ! m_visible;
            break;
        case DKeyEvent::UP:
            break;
        case DKeyEvent::REPEAT:
            break;
    }
    return false;
}

bool DebugHeadsUp::initFont()
{
    if ( !m_curfont.isDefined() || m_curfont.get() == NULL )
    {
#ifdef WIN32
        dglfont* font = FontManager::requestFont( ANSI_FIXED_FONT, 0, 255 );
#else
        dglfont* font = FontManager::requestFont( "fixed", 0, 255 );
#endif /* WIN32 */
        if( !font )
        {
            Tracer::error( "Couldn't get font from font manager" );
            return false;
        }
        m_curfont.set( font );
        return true;
    }
    return false;
}
void DebugHeadsUp::load()
{
    Plugin::load();
    DisplayManager::addOverlayCallback( this );   
}

void DebugHeadsUp::unload()
{
    DisplayManager::removeOverlayCallback( this );   
    Plugin::unload();
}

void DebugHeadsUp::overlay()
{
    if ( !m_visible )
        return;

    initFont();
    
    int winsize[4];
    glGetIntegerv( GL_VIEWPORT, winsize );
    float width = winsize[2], height = winsize[3];

    glMatrixMode( GL_PROJECTION );
    glPushMatrix();
    glLoadIdentity();
    glOrtho( 0.0f, width, 0.0f, height, -1.0f, 1.0f );
    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();
    glLoadIdentity();

    glPushAttrib( GL_ALL_ATTRIB_BITS );
    glDisable( GL_DEPTH_TEST );
    glDisable( GL_LIGHTING );
    glDisable( GL_TEXTURE_1D );
    glDisable( GL_TEXTURE_2D );
#ifdef GL_TEXTURE_3D
    glDisable( GL_TEXTURE_3D );
#endif
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

    string statstr;
    char tempstr[128];
    int row = 15;
    DEvent *curheadstat = EventManager::getTrackerState( "headtracker" ),
            *curwandstat = EventManager::getWandState(),
            *curmousestat = EventManager::getMouseState();

	if ( !curwandstat && !curmousestat )
		row--;
	if ( !curwandstat && !curheadstat )
		row--;
	if ( !curwandstat )
		row -= 2;
	if ( !curheadstat )
		row -= 2;

    glColor4f( 0.5f, 0.5f, 0.5f, 0.5f );
    glBegin(GL_QUADS);
    glVertex3f( 0.0f, 0.0f, 0.0f );
    glVertex3f( width , 0.0f, 0.0f );
    glVertex3f( width , (float)( row + 2 ) * m_curfont.get()->height(), 0.0f );
    glVertex3f( 0.0f, (float)( row + 2 ) * m_curfont.get()->height(), 0.0f );
    glEnd();

    glDisable( GL_BLEND );

    glColor3f( 1.0f, 1.0f, 1.0f );

    columntext( "X         Y         Z         H    P    R", 10, row-- );
    columntext( "WORLD", 0, row-- );

    dVec3 curpos, curdir;
    dMatrix navmat;
    Coord worldCoord;

    columntext( "ORIGIN", 2, row );
    worldCoord = worldToLocal( curpos, curdir );
    curpos = worldCoord.xyz;
    curdir = worldCoord.hpr;
    columntext( float9( tempstr, curpos.x()), 9, row );
    columntext( float9( tempstr, curpos.y()), 19, row );
    columntext( float9( tempstr, curpos.z()), 29, row ); 
    columntext( format( tempstr, "%-+3.f  ", curdir.h() ), 39, row );
    columntext( format( tempstr, "%-+3.f  ", curdir.p() ), 44, row );
    columntext( format( tempstr, "%-+3.f  ", curdir.r() ), 49, row-- );

    if ( curheadstat )
    {
        curpos = ((DTrackerEvent*)curheadstat)->getPosition();
        curdir = ((DTrackerEvent*)curheadstat)->getDirection();

        columntext( "HEAD", 4, row );
        columntext( float9( tempstr, curpos.x() ), 9, row );
        columntext( float9( tempstr, curpos.y() ), 19, row );
        columntext( float9( tempstr, curpos.z() ), 29, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.h() ), 39, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.p() ), 44, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.r() ), 49, row-- );
    }
    else
        Tracer::debug( 10, "DebugHeadsUp - Head Tracker state was NULL" );

    if ( curwandstat )
    {
        curpos = ((DWandEvent*)curwandstat)->getPosition();
        curdir = ((DWandEvent*)curwandstat)->getDirection();

        columntext( "WAND", 4, row );
        columntext( float9( tempstr, curpos.x() ), 9, row );
        columntext( float9( tempstr, curpos.y() ), 19, row );
        columntext( float9( tempstr, curpos.z() ), 29, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.h() ), 39, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.p() ), 44, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.r() ), 49, row-- );
    }
    else
        Tracer::debug( 10, "DebugHeadsUp - Wand Tracker state was NULL" );

    if ( curheadstat || curwandstat )
    {
        columntext( "DIVERSE", 0, row-- );
    }

    if ( curheadstat )
    {
        curpos = ((DTrackerEvent*)curheadstat)->getPosition();
        curdir = ((DTrackerEvent*)curheadstat)->getDirection();
        Coord local = worldToLocal( curpos, curdir );
        curpos = local.xyz;
        curdir = local.hpr;

        columntext( "HEAD", 4, row );
        columntext( float9( tempstr, curpos.x() ), 9, row );
        columntext( float9( tempstr, curpos.y() ), 19, row );
        columntext( float9( tempstr, curpos.z() ), 29, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.h() ), 39, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.p() ), 44, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.r() ), 49, row-- );
    }

    if ( curwandstat )
    {
        curpos = ((DWandEvent*)curwandstat)->getPosition();
        curdir = ((DWandEvent*)curwandstat)->getDirection();
        Coord local = worldToLocal( curpos, curdir );
        curpos = local.xyz;
        curdir = local.hpr;

        columntext( "WAND", 4, row );
        columntext( float9( tempstr, curpos.x() ), 9, row );
        columntext( float9( tempstr, curpos.y() ), 19, row );
        columntext( float9( tempstr, curpos.z() ), 29, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.h() ), 39, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.p() ), 44, row );
        columntext( format( tempstr, "%-+3.f  ", curdir.r() ), 49, row-- );
    }

    if ( curwandstat )
    {
        columntext( "JOYSTICK", 0, row );
        columntext( format( tempstr, "%-+5.2f", ((DWandEvent*)curwandstat)->getJoystickX() ), 9, row );
        columntext( format( tempstr, "%-+5.2f", ((DWandEvent*)curwandstat)->getJoystickY() ), 15, row );
    }

    if ( curwandstat || curmousestat )
    {
        columntext( "BUTTONS", 30, row );
        statstr = "";
        if ( curwandstat )
        {
            for ( int i=0;i<((DWandEvent*)curwandstat)->getNumButtons();i++)
            {
                if ( !((DWandEvent*)curwandstat)->getButton( i ) )
                    statstr += "-";
                else
                {
                    switch ( i )
                    {
                    case 0:
                        statstr += "L";
                        break;
                    case 1:
                        statstr += "M";
                        break;
                    case 2:
                        statstr += "R";
                        break;
                    case 3:
                        statstr += "4";
                       break;
                    }
                }
            }
        }
        else if ( curmousestat )
        {
            for ( int i=0;i<3;i++)
            {
                if (!((DMouseEvent*)curmousestat)->getButton( i ))
                    statstr += "-";
                else
                {
                    switch ( i )
                    {
                    case DMouseEvent::LEFT:
                        statstr += "L";
                        break;
                    case DMouseEvent::MIDDLE:
                        statstr += "M";
                        break;
                    case DMouseEvent::RIGHT:
                        statstr += "R";
                        break;
                    }
                }
            }
        }
        columntext( statstr, 39, row-- );
    }
    row--;

	navmat = NavManager::getMatrix();
	navmat = navmat.inverse( navmat );
    curpos = navmat.getTrans();
    curdir = navmat.getHPR();
    statstr = "Navigation - ";
    if ( NavManager::getCurrentNav() )
        statstr += NavManager::getCurrentNav()->getNavigationName();
    else
        statstr += "None loaded";
    columntext( statstr, 0, row-- );

    columntext( "DIVERSE", 0, row-- );
    columntext( "LOCATION", 0, row );
    columntext( float9( tempstr, curpos.x() ), 9, row );
    columntext( float9( tempstr, curpos.y() ), 19, row );
    columntext( float9( tempstr, curpos.z() ), 29, row );
    columntext( format( tempstr, "%-+3.f  ", curdir.h() ), 39, row );
    columntext( format( tempstr, "%-+3.f  ", curdir.p() ), 44, row );
    columntext( format( tempstr, "%-+3.f  ", curdir.r() ), 49, row-- );

    // The following items have nopt been implemented
    curpos = NavManager::getCurrentNav()->getPivot();
	if ( curpos != dVec3() )
		curdir = vecToDir( curpos );
	else
		curdir = dVec3();
    columntext( "PIVOT", 0, row );
    columntext( float9( tempstr, curpos.x() ), 9, row );
    columntext( float9( tempstr, curpos.y() ), 19, row );
    columntext( float9( tempstr, curpos.z() ), 29, row );
    columntext( format( tempstr, "%-+3.f  ", curdir.h() ), 39, row );
    columntext( format( tempstr, "%-+3.f  ", curdir.p() ), 44, row );
    columntext( format( tempstr, "%-+3.f  ", curdir.r() ), 49, row-- );

	curpos = NavManager::getVelocity();
	if ( curpos != dVec3() )
		curdir = vecToDir( curpos );
	else
		curdir = dVec3();
    columntext( "VELOCITY", 0, row );
    columntext( float9( tempstr, curpos.x() ), 9, row );
    columntext( float9( tempstr, curpos.y() ), 19, row );
    columntext( float9( tempstr, curpos.z() ), 29, row );
    columntext( format( tempstr, "%-+3.f  ", curdir.h() ), 39, row );
    columntext( format( tempstr, "%-+3.f  ", curdir.p() ), 44, row );
    columntext( format( tempstr, "%-+3.f  ", curdir.r() ), 49, row-- );

    columntext( "SCALE:", 0, row );
    columntext( float9( tempstr, NavManager::getMatrix().getScale().x() ), 6, row );
    columntext( float9( tempstr, NavManager::getMatrix().getScale().y() ), 16, row );
    columntext( float9( tempstr, NavManager::getMatrix().getScale().z() ), 26, row );

    columntext( "METERS:", 37, row );
    columntext( float9( tempstr, DisplayManager::getMetersPerUnit() ), 44, row );

    glPopAttrib();
    glPopMatrix();
    glMatrixMode( GL_PROJECTION );
    glPopMatrix();
    glMatrixMode( GL_MODELVIEW );
}

dVec3 DebugHeadsUp::vecToDir( dVec3 vector )
{
	dVec3 hpr;
	hpr.h() = RAD2DEG( atan2( -vector.x(), vector.y() ) );
	hpr.p() = RAD2DEG( atan2( vector.z(), sqrt( vector.x() * vector.x() + vector.y() * vector.y() ) ) );
	hpr.r() = 0;
	return hpr;
}

void DebugHeadsUp::printDirection( dMatrix mat, int column, int row )const
{
    if ( !m_curfont.isDefined() )
        Tracer::warn( "DebugHeadsUp::printDirection - attempt to use NULL dglfont" );

    int i;
    dVec3 axis;
    char temp[12];
    for ( i=0;i<3;i++ )
    {
        switch (i)
        {
            case 0:
                axis = dVec3( 1.0f, 0.0f, 0.0f );
                break;
            case 1:
                axis = dVec3( 0.0f, 1.0f, 0.0f );
                break;
            case 2:
                axis = dVec3( 0.0f, 0.0f, 1.0f );
                break;
        }
        columntext( format( temp, "%5.3f", axis.x() ), column, row );
        column += strlen( temp );
        columntext( format( temp, "%5.3f", axis.y() ), column, row );
        column += strlen( temp );
        columntext( format( temp, "%5.3f", axis.z() ), column, row );
        column += strlen( temp );
    }
}

Coord DebugHeadsUp::worldToLocal( dVec3 pos, dVec3 dir )const
{
    Coord worldCoord;
    dMatrix navmat = NavManager::getMatrix();
    dMatrix mat;
    mat.makeRotate( dir.h(), dVec3( 0.0f, 0.0f, 1.0f ), dir.p(),
            dVec3( 1.0f, 0.0f, 0.0f ), dir.r(), dVec3( 0.0f, 1.0f, 0.0f ) );
    mat.makeTranslate( pos );
    mat = mat * navmat;
//	mat = mat.inverse( mat );
    worldCoord.xyz = mat.getTrans();
    worldCoord.hpr = mat.getHPR();

    return worldCoord;
}

//===================================================
// columntext will call the dglfont class to display
// bitmap text in a row, column from gl origin of
// window based on the maximum width and height of
// the current font referenced by m_curfont
//===================================================
void DebugHeadsUp::columntext( const char* str, int column, int row )const
{
    if ( !m_curfont.isDefined() )
        Tracer::warn( "DebugHeadsUp::columntext - attempt to use NULL dglfont" );
    float x = column * m_curfont.get()->width() + m_curfont.get()->width() / 2;
	float y = row * m_curfont.get()->height() + m_curfont.get()->height() / 2;
    m_curfont.get()->dglPrintPos( str, x, y, 0.0f );
}

void DebugHeadsUp::columntext( const string& str, int column, int row )const
{
        columntext( str.c_str(), column, row );
}

