#include "IS900.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <termios.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <eventmgr/EventManager.h>
#include <tracer/Tracer.h>
#include <dreg/Registry.h>

IS900::IS900( const string& name )
: GService( name )
{
    /****************** initialize all object data *****************/
    m_ptr = m_end = m_buffer;
    m_buttons_event = NULL;
    m_joystick_event = NULL;
    for (int i=0;i<2;i++)
    {
        m_tracker_event[i] = NULL;
        m_joystick[i] = 0.0f;
    }
    for (int i=0;i<4;i++)
        m_buttons[i] = false;

    m_tolerance = atof( Registry::getValue( "/IS900/tolerance", "0.01" ).c_str() );

#ifdef CALIBRATE
    m_cal_event = NULL;
#endif

    /******************* setup device file for **********************/

    m_fd = open( DEVICE_FILE, O_RDWR|O_NDELAY );
    if (m_fd < 0)
    {
        Tracer::error( "IS900::IS900()"
            " failed: open(\"%s\", "
            "O_RDWR|O_NDELAY) failed.\n",
            DEVICE_FILE );
        return;
    }

    /* deep voodoo */
    struct termios terminfo;

    terminfo.c_iflag = 0;
    terminfo.c_oflag = 0;
    terminfo.c_lflag = 0;
#ifdef __sgi
    terminfo.c_cflag = CS8|CREAD|CLOCAL|CNEW_RTSCTS;
    terminfo.c_ispeed = terminfo.c_ospeed = B38400;
#else /* Linux and what */
    terminfo.c_cflag = CS8|CREAD|CLOCAL;
    cfsetospeed(&terminfo, B115200);
    cfsetispeed(&terminfo, B115200);
#endif

    int i;

    for (i=0; i<NCCS; i++)
        terminfo.c_cc[i] = 0;

    if(tcsetattr(m_fd, TCSANOW, &terminfo ) == -1)
    {
        Tracer::error( "IS900::IS900() failed: "
            "tcsetattr(%d,TCSANOW) failed "
            "for device file %s.",
            m_fd, DEVICE_FILE );
        clean_up();
        return;
    }

    if(tcflush(m_fd, TCIOFLUSH) == -1)
    {
        Tracer::error( "IS900::IS900() failed: "
            "tcsetattr(%d,TCIOFLUSH) failed "
            "for device file %s.",
            m_fd, DEVICE_FILE );
        clean_up();
        return;
    }


    i = write(m_fd, setupstring, strlen(setupstring));
    if ((size_t) i != strlen(setupstring))
    {
        Tracer::error( "IS900::IS900() failed: "
            "error sending configuration "
            "string to tracker:\n"
            "write(fd=%d,,) to device file %s failed.",
            m_fd, DEVICE_FILE );
        clean_up();
        return;
    }

    // Translate the point that is being tracked can be displaced a
    // little from the tracker for calibration.
    m_transHead.translate(0.0f,-0.06f,-0.035f);
	m_delta_time.start();

    Tracer::debug( "IS900 service created" );
    return;
}

IS900::~IS900(void)
{
  clean_up();
}

void IS900::clean_up(void)
{
    if(m_fd != -1)
    {
        close(m_fd);
        m_fd = -1;
    }
    m_end = m_ptr = m_buffer;

    for(int i=0;i<NUM_STATIONS;i++)
        if(m_tracker_event[i])
        {
            delete m_tracker_event[i];
            m_tracker_event[i] = NULL;
        }
    if(m_buttons_event)
    {
        delete m_buttons_event;
        m_buttons_event = NULL;
    }
    if(m_joystick_event)
    {
        delete m_joystick_event;
        m_joystick_event = NULL;
    }

#ifdef CALIBRATE
    if(m_cal_event)
    {
        delete m_cal_event;
        m_cal_event = NULL;
    }
#endif
}

void IS900::preProcess()
{
/*    if ( m_delta_time.elapsed().sec() > 1.0f/60.0f )
        m_delta_time.reset();
    else
        return;*/

    // end  points just past the end of the data read
    // ptr  points to the data not processed yet

    // check for buffer memory wrapping
    if( BUFFER_SIZE - (size_t) (m_end - m_buffer) < CHUCK)
    {
        size_t cpy_size =
            (CHUCK < (size_t) (m_end - m_ptr))?CHUCK:(size_t) (m_end - m_ptr);
        memcpy(m_buffer, m_ptr, cpy_size);
        m_end = m_buffer + cpy_size;
        m_ptr = m_buffer;
    }

    // read to end of data in the buffer
    int bytes_read = read(m_fd, m_end, BUFFER_SIZE - (size_t)(m_end - m_buffer));

    if (bytes_read < 0)
    {
        Tracer::error( "tracker::serve(): "
            "read(fd=%d,,) from device file %s failed.",
            m_fd, DEVICE_FILE );
        return;
    }
    else if(bytes_read == 0)
    {
        Tracer::warn( "tracker::serve(): "
            "read() returned 0.\n");
        return;
    }
    else
    m_end += bytes_read;

#if(0)
    {
        unsigned char *p = m_ptr;
        while( (size_t)(m_end - p) > (size_t) 0)
        {
            printf("%3.3o ",*p);
            p++;
        }
        fflush(stdout);
    }
#endif

    /* look for data records */

    for(;m_ptr <= m_end - HEAD_RECORDLEN; m_ptr++)
    {
        int j = 0; // station number starting at 0

        if( *m_ptr == '0' && *(m_ptr+2) == ' '
        && (j = ((*(m_ptr+1)-'0')-1)) > -1 && j < NUM_STATIONS )
        {
            // HEAD
            if( *(m_ptr+27) == '\r' && *(m_ptr+28) == '\n' && j == 0 )
            {
                /** EDIT * SCALING for head tracker **/
                /******** change shift and offsets *******/

                m_statepos[j].x() = (getFloat(&m_ptr[Y]) - 15.522769f)/152.4f;
                m_statepos[j].y() = (getFloat(&m_ptr[X]) - 342.853088f)/152.4f;
                m_statepos[j].z() = -(getFloat(&m_ptr[Z]) + 164.330032f)/152.4f;
                m_statedir[j].h() = -getFloat(&m_ptr[H]);
                m_statedir[j].p() =  getFloat(&m_ptr[P]);
                m_statedir[j].r() =  getFloat(&m_ptr[R]);
                /** END_EDIT **/
#ifdef CALIBRATE
                ////////////////////////////////////////////////
                // used to calibrate the rotation point offset

                m_cal_event->read(m_offset);

                // Adjust translation offsets

                m_transHead.identity();
                m_transHead.translate(m_offset[0],m_offset[1],m_offset[2]);


                ////////////////////////////////////////////////
#endif

                dMatrix tmat = m_transHead;
                cout<<"check pre/post mult here: IS900.cpp:238";
                tmat = m_transHead * dMatrix::rotate( m_statedir[j].h(), dVec3( 0,0,1),
                                                      m_statedir[j].p(), dVec3( 1,0,0),
                                                      m_statedir[j].p(), dVec3( 0,1,0));

                tmat.translate( m_statepos[j] );
                m_statepos[j] = tmat.getTrans();

                bool changed = false;
                for (int i=0;i<2;i++)
                {
                    if ( ( m_statepos[i] - m_headpos[i] ).length() > m_tolerance )
                    {
                        changed = true;
                        break;
                    }
                    if ( ( m_statedir[i] - m_headdir[i] ).length() > m_tolerance )
                    {
                        changed = true;
                        break;
                    }
                }
                if ( changed )
                {
                    m_tracker_event[j] = new DTrackerEvent( "head", m_statepos[j],
                            m_statedir[j] );
                    if ( m_tracker_event[j] )
                        EventManager::pushEvent( m_tracker_event[j] );
                }

#if(0) /* debugging print */
                printf("\n");
                printf("                                       "
                    "%+2.2f %+2.2f %+2.2f  \r",
                    state[j][3], state[j][4], state[j][5]);
                printf("               %+2.2f %+2.2f %+2.2f\r",
                    state[j][0], state[j][1], state[j][2]);
                printf("got station(%d)\n",j);
                fflush(stdout);
#endif
                m_ptr += HEAD_RECORDLEN -1; // ptr++ above
            }

            // WAND
            else if( m_ptr <= m_end - WAND_RECORDLEN && *(m_ptr+30) == '\r'
                && *(m_ptr+31) == '\n' && j == 1 )
            {
                /** EDIT * SCALING for wand tracker **/
                /******** change shift and offsets *******/

                m_statepos[j].x() = ( getFloat(&m_ptr[Y]) - 17.608776f)/152.4f;
                m_statepos[j].y() = ( getFloat(&m_ptr[X]) - 344.141479f)/152.4f;
                m_statepos[j].z() = (-getFloat(&m_ptr[Z]) - 159.391495f)/152.4f;
                m_statedir[j].h() = -getFloat(&m_ptr[H]);
                m_statedir[j].p() =  getFloat(&m_ptr[P]);
                m_statedir[j].r() =  getFloat(&m_ptr[R]);

                /** END_EDIT **/

                unsigned char buttons = m_ptr[BUTTONS];
                buttons =
                ((buttons & 02) >> 1) |
                ((buttons & 01) << 1) |
                ((buttons & 04) << 0) |
                ((buttons & 010) << 0);

                m_buttons_event = new DWandEvent( DWandEvent::BUTTONPRESS,
                        "buttons", m_statepos[j], m_statedir[j] );
                for(int i=0;i<4;i++)
                    ((DWandEvent*)m_buttons_event)->setButton( i,
                            (&buttons)[i/8] & (01 << (i %8)) );
                bool changed = false;
                for (int i=0;i<4;i++)
                {
                    if ( ((DWandEvent*)m_buttons_event)->getButton( i ) != m_buttons[i] )
                    {
                        changed = true;
                        break;
                    }
                }
                if ( changed )
                    if ( m_buttons_event )
                        EventManager::pushEvent( m_buttons_event );

                float joystick[2] =
                { m_ptr[JOYSTICK] - 127.0f, m_ptr[JOYSTICK+1] - 127.0f };
                if(joystick[0] > 0.0f)
                    joystick[0] /= 128.0f;
                else
                    joystick[0] /= 127.0f;
                if(joystick[1] > 0.0f)
                    joystick[1] /= 128.0f;
                else
                    joystick[1] /= 127.0f;

                m_joystick_event = new DWandEvent(DWandEvent::JOYSTICKMOVE,
                        "joystick", m_statepos[j], m_statedir[j] );
                ((DWandEvent*)m_joystick_event)->setJoystickPos( joystick[0], joystick[1] );
                changed = false;
                for ( int i=0;i<2;i++ )
                {
                    if ( fabs( joystick[i] - m_joystick[i] ) > m_tolerance )
                    {
                        changed = true;
                        break;
                    }
                }
                if ( changed )
                    if ( m_joystick_event )
                        EventManager::pushEvent( m_joystick_event );

                changed = false;
                for (int i=0;i<2;i++)
                {
                    if ( ( m_statepos[i] - m_headpos[i] ).length() > m_tolerance )
                    {
                        changed = true;
                        break;
                    }
                    if ( ( m_statedir[i] - m_headdir[i] ).length() > m_tolerance )
                    {
                        changed = true;
                        break;
                    }
                }
                if ( changed )
                {
                    m_tracker_event[j] = new DWandEvent( DWandEvent:: TRACK, "wand", m_statepos[j], m_statedir[j] );
                    if ( m_tracker_event[j] )
                        EventManager::pushEvent( m_tracker_event[j] );
                }

#if(0) /* debugging print */
                printf("\n");
                printf("                                       "
                    "                                       "
                    "0%3.3o\r",buttons);
                printf("                                       "
                    "%+2.2f %+2.2f %+2.2f  \r",
                    state[j][3], state[j][4], state[j][5]);
                printf("               %+2.2f %+2.2f %+2.2f\r",
                    state[j][0], state[j][1], state[j][2]);
                printf("got station(%d)\n",j);
                fflush(stdout);
#endif
                m_ptr += WAND_RECORDLEN -1; // m_ptr++ above
            }
        }
    }

    return;
}
